#include "stdafx.h"
#include "XarOutputDev.h"
#include "common/ColourConv.h"
#include "poppler/Page.h"
#include "poppler/Link.h"
#include "poppler/OptionalContent.h"

std::wstring TextStringToUnicode(const GooString* textStr);

std::wstring ASCIIStringToUnicode(const GooString* str)
{
	std::wstring res;
	if (!str)
		return res;
	int len = str->getLength();	
	if (!len)
		return res;
	const char *s = str->c_str();

	res.reserve(len);
	while (len--)
		res.push_back((unsigned char)(*s++));

	return res;
}

void AppendWebAddress(VXfTreeNode* pNode, CXfWebAddress* pWeb)
{
	size_t childrenCount = pNode->ChildrenCount();
	for (size_t i = 0; i < childrenCount; ++i)
	{
		VXfTreeNode *pChild = pNode->GetChild(i);
		if (pChild && TAG_URL_COLOUR_OVERRIDE == pChild->GetTag())
		{
			pNode->InsertBefore(i, pWeb);
			return;
		}
	}

	pNode->AppendChild(pWeb);
}

/******************************************************************************
   Function:	CLinkProps::CalculateValuableTextLength
   Author:		Paul Gruzdev
   Purpose:     Given a text string, calculate the actual length needed to gather
				statistics to pick the determinant URL colour. For example, if the
				string solely consists of many '.' chars, we won't use them all.
   Returns:		int
   Parameters:	std::wstring str  
   Notes:       
******************************************************************************/
int CLinkProps::CalculateValuableTextLength(std::wstring str)
{
	int iSize = 0;
	std::set<wchar_t> sNonValuable;

	for (size_t i = 0; i < str.length(); i++)
	{
		wchar_t ch = str[i];

		if (ch == '.' || ch == ',' || ch == '-' || ch == '_' || ch == '\'')
		{
			if (sNonValuable.find(ch) == sNonValuable.end())
			{
				sNonValuable.insert(ch);
				iSize++;
			}
		}
		else if (ch == 32 || ch <= 9)
			continue;
		else
			iSize++;
	}
	return iSize;
}

int AnnotationContext::GetPageNum(const LinkDest *pDest) const
{
	if (!pDest) return 0;

	if (pDest->isPageRef())
	{
		PageNumMap::const_iterator it = m_PageNumMap.find(pDest->getPageRef());
		return (m_PageNumMap.end() == it) ? 0 : it->second;
	}
	else
		return pDest->getPageNum();
}

/******************************************************************************
   Function:    AnnotationContext::ProcessAnnotations
   Author:      Paul Gruzdev
   Purpose:		Run through the annotation array of the page and try to import
				all hyperlinks/bookmarks/etc. Put all annotations as invisible 
				rectangles on a separate layer called 'Annotations'. This is the
				first and most simple method of importing annotations. Used for
				debug purposes only. Now the quad tree is used to match annotation
				and every object on the page to create link from object, not from
				a special layer
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool AnnotationContext::ProcessAnnotations(PDFDoc* pdfDoc, XarOutputDev *pOutDev)
{
	START_STOPWATCH_FN
	if (!m_pAnnots || !m_pAnnots->getNumAnnots()) return false;

	VXfTreeNode* pLayer = nullptr;

	if (g_settings.bNoLayers)
		pLayer = pOutDev->GetDefLayer();

	if (!pLayer)
	{
		pLayer = new CXfLayer();
		pOutDev->GetCurrentPage()->AppendChild(pLayer);
		pLayer->AppendChild(new CXfLayerDetails(TAG_LAYER_FLAG_VISIBLE | TAG_LAYER_FLAG_PRINTABLE, L"Annotations"));
	}

	for(int i = 0; i < m_pAnnots->getNumAnnots(); i++)
	{
		Annot* pAnnot = m_pAnnots->getAnnot(i);
		LinkAction* pAction = NULL;
		switch (pAnnot->getType())
		{
		case Annot::typeLink:
			pAction = ((AnnotLink*)pAnnot)->getAction();
			break;
		case Annot::typeScreen:
			pAction = ((AnnotScreen*)pAnnot)->getAction();
			break;
		case Annot::typeWidget:
			{
				AnnotWidget* pWidget = (AnnotWidget*)pAnnot;
				pAction = pWidget->getAction();
				LinkAction* pEnter = pWidget->getAdditionalAction(Annot::actionCursorEntering);
				LinkAction* pLeave = pWidget->getAdditionalAction(Annot::actionCursorLeaving);
				pAction = pEnter;
			}
			break;
		}

		if(!pAction) continue;

		double x1, y1, x2, y2;
		pAnnot->getRect(&x1, &y1, &x2, &y2);
		state->transform(x1, y1, &x1, &y1);
		state->transform(x2, y2, &x2, &y2);
		//Rect rAnnot(min(x1, x2), min(y1, y2), fabs(x2-x1), fabs(y2-y1));

		double dXCenter = (x1+x2)/2, dYCenter = (y1+y2)/2, dw = fabs(x2-x1), dh = fabs(y2-y1);
		//state->transform((x1+x2)/2, (y1+y2)/2, &dXCenter, &dYCenter);
		//dw = state->transformWidth(x2-x1);
		//dh = state->transformWidth(y2-y1);

		switch(pAction->getKind())
		{
		case actionGoToR:
			{
				LinkGoToR* pGoto = (LinkGoToR*)pAction;
				const LinkDest* pDest = pGoto->getDest();
				if (!pDest && pGoto->getNamedDest())
					pDest = pdfDoc->getCatalog()->findDest(pGoto->getNamedDest());

				CXfRect* pObj = new CXfRect((LONG)dXCenter, (LONG)dYCenter, (LONG)(dSQRT2*dw), (LONG)(dSQRT2*dh));

				//pObj->AppendChild(new CXfLineColor(CXfLineColor::None));
				pObj->AppendChild(new CXfSimpleColor(0, 128, 192));
				pObj->AppendChild(new CXfLineColor(pObj->GetLastChild()));
				pObj->AppendChild(new CXfLineWidth(501));
				pObj->AppendChild(new CXfFlatFill(CXfFlatFill::None));

				const GooString* pFile = pGoto->getFileName();
				if (pFile)
				{
					CXfWebAddress* pWeb = new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0),
						CXfWebAddress::File, TextStringToUnicode(pFile).c_str(), L"");
					AppendWebAddress(pObj, pWeb);
				}

				pLayer->AppendChild(pObj);
			}
			break;
		case actionLaunch:
			{
				LinkLaunch* pLaunch = (LinkLaunch*)pAction;
				CXfRect* pObj = new CXfRect((LONG)dXCenter, (LONG)dYCenter, (LONG)(dSQRT2*dw), (LONG)(dSQRT2*dh));

				//pObj->AppendChild(new CXfLineColor(CXfLineColor::None));
				pObj->AppendChild(new CXfSimpleColor(225, 45, 225));
				pObj->AppendChild(new CXfLineColor(pObj->GetLastChild()));
				pObj->AppendChild(new CXfLineWidth(501));
				pObj->AppendChild(new CXfFlatFill(CXfFlatFill::None));

				const GooString* pFile = pLaunch->getFileName();
				if (pFile)
				{
					CXfWebAddress* pWeb = new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0),
						CXfWebAddress::File, TextStringToUnicode(pFile).c_str(), L"");
					AppendWebAddress(pObj, pWeb);
				}

				pLayer->AppendChild(pObj);
			}
			break;
		case actionGoTo:
			{
				LinkGoTo* pGoto = (LinkGoTo*)pAction;
				const LinkDest* pDest = pGoto->getDest();
				if (!pDest && pGoto->getNamedDest())
					pDest = pdfDoc->getCatalog()->findDest(pGoto->getNamedDest());

				if (int iPageNum = GetPageNum(pDest))
				{
					CXfRect* pObj = new CXfRect((LONG)dXCenter, (LONG)dYCenter, (LONG)(dSQRT2*dw), (LONG)(dSQRT2*dh));

					//pObj->AppendChild(new CXfLineColor(CXfLineColor::None));
					pObj->AppendChild(new CXfSimpleColor(45, 225, 45));
					pObj->AppendChild(new CXfLineColor(pObj->GetLastChild()));
					pObj->AppendChild(new CXfLineWidth(501));
					pObj->AppendChild(new CXfFlatFill(CXfFlatFill::None));

					CXfWebAddress* pWeb = new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0), 
						CXfWebAddress::Page, FormatStringW(L"%d", iPageNum).c_str(), L"");
					AppendWebAddress(pObj, pWeb);
					pLayer->AppendChild(pObj);
				}
			}
			break;
		case actionURI:
			{
				LinkURI* pURI = (LinkURI*)pAction;
				CXfRect* pObj = new CXfRect((LONG)dXCenter, (LONG)dYCenter, (LONG)(dSQRT2*dw), (LONG)(dSQRT2*dh));

				//pObj->AppendChild(new CXfLineColor(CXfLineColor::None));
				pObj->AppendChild(new CXfSimpleColor(255, 75, 0));
				pObj->AppendChild(new CXfLineColor(pObj->GetLastChild()));
				pObj->AppendChild(new CXfLineWidth(501));
				pObj->AppendChild(new CXfFlatFill(CXfFlatFill::None));

				AppendWebAddress(pObj, new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0), 
					CXfWebAddress::External, ASCIIStringToUnicode(pURI->getURI()).c_str(), L""));

				pLayer->AppendChild(pObj);
			}
			break;
		case actionOCGState:
			{
				LinkOCGState* pOCG = (LinkOCGState*)pAction;
				CXfRect* pObj = new CXfRect((LONG)dXCenter, (LONG)dYCenter, (LONG)(dSQRT2*dw), (LONG)(dSQRT2*dh));
				//pObj->AppendChild(new CXfLineColor(CXfLineColor::None));
				pObj->AppendChild(new CXfSimpleColor(45, 45, 225));
				pObj->AppendChild(new CXfLineColor(pObj->GetLastChild()));
				pObj->AppendChild(new CXfLineWidth(501));
				pObj->AppendChild(new CXfFlatFill(CXfFlatFill::None));

				for (int i = 0; i < pOCG->getStateList()->size(); i++)
				{
					LinkOCGState::StateList* pState = (*pOCG->getStateList())[i];
					if (pState->st == LinkOCGState::On)
					{
						int iLen = pState->list->size();
						Ref* ref = (*pState->list)[0];
						// Get the last ref in this list as it is the actual Layer popup name
						Ref* refLast = (*pState->list)[iLen - 1];
						Object ocgObj = pdfDoc->getXRef()->fetch(refLast->num, refLast->gen);
						Object oName = ocgObj.getDict()->lookup("Name");
						if (!oName.isNull() && oName.isString())
						{
							AppendWebAddress(pObj, new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0), 
								CXfWebAddress::Layer, TextStringToUnicode(oName.getString()).c_str(), L""));
							pLayer->AppendChild(pObj);
						}
					}
				}
			}
			break;
		default:
			{
				CXfRect* pObj = new CXfRect((LONG)dXCenter, (LONG)dYCenter, (LONG)(dSQRT2*dw), (LONG)(dSQRT2*dh));
				
				pObj->AppendChild(new CXfSimpleColor(165, 165, 165));
				pObj->AppendChild(new CXfLineColor(pObj->GetLastChild()));
				pObj->AppendChild(new CXfLineWidth(501));
				pObj->AppendChild(new CXfFlatFill(CXfFlatFill::None));

				std::wstring strAction;
				switch(pAction->getKind())
				{
				case actionNamed:			// named action
					strAction = L"actionNamed";
					break;
				case actionMovie:			// movie action
					strAction = L"actionMovie";
					break;
				case actionRendition:		// rendition action
					strAction = L"actionRendition";
					break;
				case actionSound:			// sound action
					strAction = L"actionSound";
					break;
				case actionJavaScript:		// JavaScript action
					strAction = L"actionJavaScript";
					break;
				case actionUnknown:			// anything else
					strAction = L"actionUnknown";
					break;
				}

				AppendWebAddress(pObj, new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0), 
					CXfWebAddress::External, strAction.c_str(), L""));
				pLayer->AppendChild(pObj);
			}
		}
	}

STOP_STOPWATCH(__FUNCTION__)
	return true;
}

/*
static inline double Log2(double n)
{  
	return log( n ) / log( 2 );  
}
*/
/******************************************************************************
   Function:	AnnotationContext::Create
   Author:		Paul Gruzdev
   Purpose:     Create annotation context for a given page
   Returns:		bool
   Parameters:	Page * pPage  
   Parameters:	int iWidth  
   Parameters:	int iHeight  
   Notes:       
******************************************************************************/
bool AnnotationContext::Create(Page* pPage, int iWidth, int iHeight)
{
	Reset();
	m_pAnnots = pPage->getAnnots();
	if (true)//pAnnots && pAnnots->getNumAnnots())
	{
		int rotate = pPage->getRotate();
		if (rotate >= 360) {
			rotate -= 360;
		} else if (rotate < 0) {
			rotate += 360;
		}
		state = new GfxState(72000, 72000, pPage->getCropBox(), rotate, false);
		m_tree.push_back(CQuadtree());
		m_tree.back().Create(iWidth, iHeight, 8);
		m_pageToTreeIdx[pPage->getNum()] = (int)m_tree.size() - 1;

		return true;
	}
	else m_pAnnots = nullptr;

	return false;
}

/******************************************************************************
   Function:	AnnotationContext::AddElement
   Author:		Paul Gruzdev
   Purpose:     Add a renderable object to the quad tree according to it's bounds
   Returns:		bool
   Parameters:	VXfTreeNode * pNode  
   Notes:       
******************************************************************************/
bool AnnotationContext::AddElement(VXfBounded* pNode, int page)
{
	//return false; // temporary disable for stable fix of DDES-1003 prior the release
	if (((m_pdfDoc->getOutline() && m_pdfDoc->getOutline()->getItems()) || (m_pAnnots && m_pAnnots->getNumAnnots())) && HasTree(page))
		return GetTree(page).AddElement(pNode) == 0;
	return false;
}

/******************************************************************************
   Function:	AnnotationContext::CalcAnnotationCoverage
   Author:		Paul Gruzdev
   Purpose:     Calculate the fraction at which the annotation covers the
				object.
   Returns:		double
   Parameters:	Rect & rAnnot  
   Parameters:	Rect & rcObject  
   Notes:       
******************************************************************************/
double AnnotationContext::CalcAnnotationCoverage(Rect& rAnnot, Rect& rcObject)
{
	// First simple rule - if an object is completely covered by annotation - put a link on it.
	if(rAnnot.Contains(rcObject))
		return 1.0;

	Rect rcIntersect;
	Rect::Intersect(rcIntersect, rcObject, rAnnot);
	double dObjectSquare = (double)rcObject.width*rcObject.height;
	double dIntersectSquare = (double)rcIntersect.width*rcIntersect.height;
	double dSquareDiffRatio;

	// Object is a line orthogonal to X or Y axis - let's apply a line rule: annotation contains >= threshold object line width
	if(dObjectSquare == 0)
		dSquareDiffRatio = (double)(rcIntersect.width ? rcIntersect.width : rcIntersect.height) / 
		(double)(rcIntersect.width ? rcObject.width : rcObject.height);
	else
		// Square rule - annotation covers >= 0.8 of object's area
		dSquareDiffRatio = dIntersectSquare/dObjectSquare;

	return dSquareDiffRatio;
}


bool AnnotListSort(std::pair<double, VElementRect*> p1, std::pair<double, VElementRect*> p2)
{
	return p1.first > p2.first;
}

/******************************************************************************
   Function:	AnnotationContext::CreateAnnotationCoverageList
   Author:		Paul Gruzdev
   Purpose:     Create a sorted list of pairs: [coverage by annotation, element].
				The first pair in the list has the most coverage by annot.
   Returns:		void
   Parameters:	std::vector<VElementRect * > & elem  
   Parameters:	Rect & rAnnot  
   Parameters:	CoveragePairVector & sortedList  
   Notes:       
******************************************************************************/
void AnnotationContext::CreateAnnotationCoverageList(std::vector<VElementRect*>& elem, Rect& rAnnot, CoveragePairVector& sortedList)
{
	for(std::vector<VElementRect*>::iterator it = elem.begin(); it != elem.end(); ++it)
	{
		VElementRect* pElem = *it;
		int x1, y1, x2, y2;
		pElem->GetRect(x1, y1, x2, y2);
		VXfBounded* pObject = (VXfBounded*)pElem;
		Rect rcObject = pObject->Bounds();
		sortedList.push_back(std::make_pair(CalcAnnotationCoverage(rAnnot, rcObject), pElem));
	}

	std::sort(sortedList.begin(), sortedList.end(), AnnotListSort);
}

/******************************************************************************
   Function:	AnnotationContext::ShouldApplyAnnotation
   Author:		Paul Gruzdev
   Purpose:     Match annotation and object rect to decide whether the annotaion
				(link) can be 'put on' this object.
   Returns:		bool
   Parameters:	Rect & rAnnot  
   Parameters:	Rect & rcObject  
   Notes:       
******************************************************************************/
bool AnnotationContext::ShouldApplyAnnotation(Rect& rAnnot, Rect& rcObject)
{
	// First simple rule - if an object is completely covered by annotation - put a link on it.
	if(rAnnot.Contains(rcObject))
		return true;

	bool bAnnotInside = rcObject.Contains(rAnnot);
	Rect rcIntersect;
	Rect::Intersect(rcIntersect, rcObject, rAnnot);
	double dObjectSquare = (double)rcObject.width*rcObject.height;
	double dIntersectSquare = (double)rcIntersect.width*rcIntersect.height;
	double dSquareDiffRatio;

	// Object is a line orthogonal to X or Y axis - let's apply a line rule: annotation contains >= 0.8 object line width
	if(dObjectSquare == 0)
		dSquareDiffRatio = (double)(rcIntersect.width ? rcIntersect.width : rcIntersect.height) / 
						   (double)(rcIntersect.width ? rcObject.width : rcObject.height);
	else
		// Square rule - annotation covers >= 0.8 of object's area
		dSquareDiffRatio = dIntersectSquare/dObjectSquare;

	return dSquareDiffRatio >= (bAnnotInside ? 0.45 : 0.75);
}

/******************************************************************************
   Function:	AnnotationContext::RemoveURLProperties
   Author:		Paul Gruzdev
   Purpose:     Since we evaluated default URL colour and underline, we need
				to remove these properties from every link that has the defaults
   Returns:		void
   Parameters:	VXfTextRenderable * pString  
   Notes:       
******************************************************************************/
void AnnotationContext::RemoveURLProperties(VXfTextRenderable* pString)
{
	if (m_bUnderlineURL)
	{
		VXfTreeNode* pUnderline = pString->FindFirstNode(TAG_TEXT_UNDERLINE_ON, false);
		if (pUnderline) // Remove the underline if it is a default URL behaviour set in the document
			pUnderline->SetRemoved(true);
	}
	if (m_pURLColour)
	{
		VXFFillStyleNode* pStoryFillStyle = ((VXfTextRenderable*)pString->GetParent()->GetParent())->Context().m_pFill;
		CXfFlatFill* pStoryFill = pStoryFillStyle->IsFlatFill() ? (CXfFlatFill*)pStoryFillStyle : NULL;
		CXfFlatFill* pFill = (CXfFlatFill*)pString->Context().m_pFill;
		if (pFill)
		{
			// If the link colour is the same as the default - remove it
			if (m_pURLColour->Compare((CXfComplexColor*)pFill->pReference))
			{
				if (!pFill->IsRemoved())
				{
					bool bDoRemove = false;
					VXfTreeNode* pParent = pFill->GetParent();
					if (pParent->GetTag() == TAG_TEXT_LINE)
					{
						CXfTextLine* pLine = (CXfTextLine*)pParent;
						bDoRemove = pLine->GetStringCount() == pLine->GetURLCount();
					}
					else if (pParent->IsTextStory())
					{
						CXfTextStorySimple* pStory = (CXfTextStorySimple*)pParent;
						bDoRemove = pStory->GetStringCount() == pStory->GetURLCount();
					}

					// Now, if not removed, test all child text nodes vs colour.
					if (!bDoRemove && (pParent->IsTextStory() || pParent->GetTag() == TAG_TEXT_LINE))
					{
						bool bHasColorApplied = true;
						VXfTreeNodeVector v;
						pParent->FindNodes(CTagSet(TAG_TEXT_CHAR, TAG_TEXT_STRING), v, true);
						for (size_t i = 0; i < v.size(); i++)
						{
							VXfRenderable* pNode = (VXfRenderable*)v[i];
							if (!pNode->FindFirstWebaddressNode(false, false))
							{
								if (pNode->Context().m_pFill == pFill) // This text char/string has no URL but the same fill applied: can't delete it
								{
									bHasColorApplied = false;
									break;
								}
							}
						}
						bDoRemove = bHasColorApplied;
					}
					// Now, we need to remove it, but only if it's a string child or all strings in line or story
					// are URLs
					if (pParent->IsTextString() || bDoRemove)
					{
						/* DDES-14408
						   This is no longer correct since colours can be shared between text elements
						pFill->pReference->SetRemoved(true);*/
						pFill->SetRemoved(true);

						if (pParent->IsTextStory()) // if we are removing the default text story colour, replace it with black
						{
							pParent->InsertAfterChild(pFill, new CXfFlatFill(CXfFlatFill::Black));
						}
					}
				}
			}
			else
				// Otherwise tell Camelot it's a custom link colour that it should keep
				pString->AppendChild(new CXfZeroSizedNode(TAG_URL_COLOUR_OVERRIDE));
		}
	}
}
/******************************************************************************
   Function:	AnnotationContext::AddLink
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		void
   Parameters:	std::vector<VXfTextRenderable * > & textLinks  
   Parameters:	VXfBounded * pObject  
   Notes:       
******************************************************************************/
void AnnotationContext::AddLink(std::vector<VXfTextRenderable*>& textLinks, VXfBounded* pObject)
{
	if (pObject->IsTextString() || pObject->GetTag() == TAG_TEXT_KERN)
	{
		if (!m_bDefURLPropertiesEvaluated)
		{
			VXfTextRenderable* pstr = (VXfTextRenderable*)pObject;
			//if (pstr->Context().m_pFill && pstr->Context().m_pFill->IsFlatFill())
			textLinks.push_back(pstr);
		}
		else
			RemoveURLProperties((VXfTextRenderable*)pObject);
	}

}

/******************************************************************************
   Function:	FindObject
   Author:		Paul Gruzdev
   Purpose:		Find a Xara bounded object around the destination point
   Returns:		VXfBounded*
   Parameters:	LinkDest * pDest
   Notes:       
******************************************************************************/
VXfBounded* AnnotationContext::FindObject(const LinkDest* pDest)
{
	int iPageNum = GetPageNum(pDest);
	// if non of the coords are defined, simply return link to a page
	if (/*!(pDest->getChangeLeft() || pDest->getChangeTop()) ||*/ !HasTree(iPageNum))
		return NULL;

	std::vector<PDFRectangle> rcs;

	double x1, y1, x2, y2;
	switch (pDest->getKind())
	{
	case destXYZ:
	case destFitR:
		{
			const double d = 10, d2 = d*2; // 10pt around the annotation centre

			x1 = pDest->getLeft() - d;
			y1 = pDest->getTop() - d;
			x2 = x1 + d2;
			y2 = y1 + d2;

			rcs.push_back(PDFRectangle(x1, y1, x2, y2)); // first one in search is around our destination point
			rcs.push_back(PDFRectangle(x2, y1, x2 + d2, y2)); // then search to the right (our PDFEXport for example shifts (-8,8) when specifying dest point
			rcs.push_back(PDFRectangle(x2, y1 - d2, x2 + d2, y2)); // then to the right bottom
			rcs.push_back(PDFRectangle(x1, y1 - d2, x2 + d2, y1)); // then bottom
			rcs.push_back(PDFRectangle(x1 - d2, y1 - d2, x1, y1)); // then bottom left
			rcs.push_back(PDFRectangle(x1 - d2, y1, x1, y2)); // then left
			rcs.push_back(PDFRectangle(x1 - d2, y2, x1, y2 + d2)); // then left top
			rcs.push_back(PDFRectangle(x1, y2, x2, y2 + d2)); // then top
			rcs.push_back(PDFRectangle(x2, y2, x2 + d2, y2 + d2)); // then top right

		}
		break;
	/*case destFitR:
		x1 = pDest->getLeft();
		y1 = pDest->getBottom();
		x2 = pDest->getRight();
		y2 = pDest->getTop();

		rcs.push_back(PDFRectangle(x1, y1, x2, y2));
		break;*/
	default:
		return NULL;
	}		

	CQuadtree& tree = GetTree(iPageNum);

	for (size_t i = 0; i < rcs.size(); i++)
	{
		PDFRectangle r = rcs[i];
		state->transform(r.x1, r.y1, &x1, &y1);
		state->transform(r.x2, r.y2, &x2, &y2);
		Rect rAnnot((int)min(x1, x2), (int)min(y1, y2), (int)fabs(x2 - x1), (int)fabs(y2 - y1));

		std::vector<VElementRect*> elem;
		if (tree.FindIntersectElements(elem, rAnnot))
			return (VXfBounded*)elem.back();
	}

	return NULL;
}

void AnnotationContext::AddLinkPair(const LinkDest* pDest, VXfBounded* pSrcObject)
{
	int page = GetPageNum(pDest);
	if (m_ForwardLinkMap.find(page) == m_ForwardLinkMap.end())
		m_ForwardLinkMap[page] = std::vector<LinkPair>();
	m_ForwardLinkMap[page].push_back(std::make_pair(pDest, pSrcObject));
}

CQuadtree& AnnotationContext::GetTree(int page)
{
	if (HasTree(page) == false)
	{
		_ASSERT(FALSE);
		static CQuadtree temp;
		return temp;
	}
	return page < 0 ? m_tree.back() : m_tree[m_pageToTreeIdx[page]];
}

/******************************************************************************
   Function:	AssignLink
   Author:		Paul Gruzdev
   Purpose:		Find a destination object on the page and assign link from source
				to it, or link to this page if not found
   Returns:		void
   Parameters:	int iPageNum
   Parameters:	LinkDest * pDest
   Parameters:	VXfBounded * pSrcObject
   Notes:       
******************************************************************************/
std::wstring AnnotationContext::AssignLink(const LinkDest* pDest, VXfBounded* pSrcObject)
{
	std::wstring str;
	const std::wstring strObjName(L"ObjectName");
	static std::set<VXfBounded*> srcObjCache;
	if (pSrcObject && srcObjCache.find(pSrcObject) != srcObjCache.end())
		return str; // One object may have only one link to destination object

	VXfBounded* pDstObject = FindObject(pDest);
	if (pDstObject)
	{
		CXfWizop* pNamedLink = (CXfWizop*)pDstObject->FindFirstNode(TAG_WIZOP);

		if (pNamedLink && pNamedLink->Name.getstr() == strObjName)
			str = pNamedLink->Parameter.getstr();
		else
			str = GetNextNamedLink(); // Generate a unique name for the link

		if (pSrcObject)
		{
			CXfWebAddress* pWeb = new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0),
				CXfWebAddress::Object, str.c_str(), L"");
			AppendWebAddress(pSrcObject, pWeb); // add link to the source
		}

		if (pNamedLink == NULL)
		{
			pNamedLink = new CXfWizop(strObjName.c_str(), L"", str.c_str(), L"");
			pDstObject->AppendChild(pNamedLink); // add name to the destination
		}
	}
	else if (pSrcObject)
	{
		CXfWebAddress* pWeb = new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0),
			CXfWebAddress::Page, FormatStringW(L"%d", GetPageNum(pDest)).c_str(), L"");
		AppendWebAddress(pSrcObject, pWeb);
	}

	if (pSrcObject)
		srcObjCache.insert(pSrcObject);

	return str;
}

/******************************************************************************
   Function:	AnnotationContext::ProcessAnnotationsQuadTree
   Author:		Paul Gruzdev
   Purpose:     This should be called once per page to iterate through
				all annotations and find corresponding objects and put
				links on them
   Returns:		bool
   Parameters:	PDFDoc* pdfDoc
   Notes:       
******************************************************************************/
bool AnnotationContext::ProcessAnnotationsQuadTree(int iCurrentPage, PDFDoc* pdfDoc, VXfTreeNode* pDoc)
{
	if (m_PageNumMap.empty())
	{
		for (int page = 1; page <= pdfDoc->getNumPages(); page++)
		{
			if (Ref *ref = pdfDoc->getCatalog()->getPageRef(page))
				m_PageNumMap[*ref] = page;
		}
	}

	// First of all: scan the forward links to this page to set them to previous src objects
	std::map<int, std::vector<LinkPair>>::iterator it = m_ForwardLinkMap.find(iCurrentPage);
	if (it != m_ForwardLinkMap.end())
	{
		for (size_t i = 0; i < it->second.size(); i++)
		{
			const LinkDest* pDest = it->second[i].first;
			VXfBounded* pSrcObject = it->second[i].second;
			AssignLink(pDest, pSrcObject);
		}
	}

	std::vector<VXfTextRenderable*> textLinks;

	for (int i = 0; i < m_pAnnots->getNumAnnots(); i++)
	{
		Annot* pAnnot = m_pAnnots->getAnnot(i);
		LinkAction* pAction = NULL;
		switch (pAnnot->getType())
		{
		case Annot::typeLink:
			pAction = ((AnnotLink*)pAnnot)->getAction();
			break;
		case Annot::typeScreen:
			pAction = ((AnnotScreen*)pAnnot)->getAction();
			break;
		case Annot::typeWidget:
			{
				AnnotWidget* pWidget = (AnnotWidget*)pAnnot;
				pAction = pWidget->getAction();
				LinkAction* pEnter = pWidget->getAdditionalAction(Annot::actionCursorEntering);
				LinkAction* pLeave = pWidget->getAdditionalAction(Annot::actionCursorLeaving);
				pAction = pEnter;
			}
			break;
		}

		if (!pAction) continue;

		double x1, y1, x2, y2;
		pAnnot->getRect(&x1, &y1, &x2, &y2);
		state->transform(x1, y1, &x1, &y1);
		state->transform(x2, y2, &x2, &y2);
		Rect rAnnot((int)min(x1, x2), (int)min(y1, y2), (int)fabs(x2-x1), (int)fabs(y2-y1));

		std::vector<VElementRect*> elem;
		if (GetTree().FindIntersectElements(elem, rAnnot))
		{
			CoveragePairVector sortedList;
			CreateAnnotationCoverageList(elem, rAnnot, sortedList);

			double dHighestMeasure = sortedList.begin()->first;
			for (CoveragePairVector::iterator it = sortedList.begin(); it != sortedList.end(); ++it)
			{
				double dMeasure = it->first;
				VElementRect* pElem = it->second;
				int x1, y1, x2, y2;
				pElem->GetRect(x1, y1, x2, y2);
				VXfBounded* pSrcObject = (VXfBounded*)pElem;
				Rect rcObject = pSrcObject->Bounds();
				bool bAnnotInside = rcObject.Contains(rAnnot);
				bool bApplyStrict = dMeasure >= (bAnnotInside ? 0.45 : 0.75);
				bool bApplyMild = fabs(dMeasure-dHighestMeasure) <= 0.02;

				if (bApplyStrict || bApplyMild)
				{
					switch(pAction->getKind())
					{
					case actionGoToR:
						{
							LinkGoToR* pGoto = (LinkGoToR*)pAction;
							const LinkDest* pDest = pGoto->getDest();
							if (!pDest && pGoto->getNamedDest())
								pDest = pdfDoc->getCatalog()->findDest(pGoto->getNamedDest());

							const GooString* pFile = pGoto->getFileName();
							if (pFile)
							{
								CXfWebAddress* pWeb = new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0),
									CXfWebAddress::File, TextStringToUnicode(pFile).c_str(), L"");
								AppendWebAddress(pSrcObject, pWeb);

								AddLink(textLinks, pSrcObject);
							}
						}
						break;
					case actionLaunch:
						{
							LinkLaunch* pLaunch = (LinkLaunch*)pAction;
							const GooString* pFile = pLaunch->getFileName();
							if (pFile)
							{
								CXfWebAddress* pWeb = new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0),
									CXfWebAddress::File, TextStringToUnicode(pFile).c_str(), L"");
								AppendWebAddress(pSrcObject, pWeb);

								AddLink(textLinks, pSrcObject);
							}
						}
						break;				
					case actionGoTo:
						{
							LinkGoTo* pGoto = (LinkGoTo*)pAction;
							const LinkDest* pDest = pGoto->getDest();
							if (!pDest && pGoto->getNamedDest())
								pDest = pdfDoc->getCatalog()->findDest(pGoto->getNamedDest());

							//double dMediaHeight = pPage->getMediaHeight();

							if (int iDstPageNum = GetPageNum(pDest))
								//pDest->getLeft() == 0.0 && pDest->getTop() == dMediaHeight)
							{
								if (iDstPageNum <= iCurrentPage)
								{
									// Link to an object on a previous page - just add a link to it
									AssignLink(pDest, pSrcObject);
								}
								else
								{
									// add link for further processing when the page iPageNum happens
									AddLinkPair(pDest, pSrcObject);
								}

								AddLink(textLinks, pSrcObject);
							}
						}
						break;
					case actionURI:
						{
							LinkURI* pURI = (LinkURI*)pAction;
							AppendWebAddress(pSrcObject, new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0), 
								CXfWebAddress::External, ASCIIStringToUnicode(pURI->getURI()).c_str(), L""));
							AddLink(textLinks, pSrcObject);
						}
						break;
					case actionOCGState:
						{
							LinkOCGState* pOCG = (LinkOCGState*)pAction;
							if (pOCG->getStateList()->size())
							{
								for (int i = 0; i < pOCG->getStateList()->size(); i++)
								{
									LinkOCGState::StateList* pState = (*pOCG->getStateList())[i];
									int iLen = pState->list->size();
									if (pState->st == LinkOCGState::On && iLen)
									{
										Ref* ref = (*pState->list)[0];
										// Get the last ref in this list as it is the actual Layer popup name
										Ref* refLast = (*pState->list)[iLen - 1];
										Object ocgObj = pdfDoc->getXRef()->fetch(refLast->num, refLast->gen);
										Object oName = ocgObj.getDict()->lookup("Name");
										if (!oName.isNull() && oName.isString())
										{
											AppendWebAddress(pSrcObject, new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0), 
												CXfWebAddress::Layer, TextStringToUnicode(oName.getString()).c_str(), L""));
											AddLink(textLinks, pSrcObject);
										}
									}
								}
							}
						}
						break;
					} // end switch
				}// Should apply annot
				else
					break; // since we have a sorted measure list, the latter elems will be even farther
			}
		}
	}

	for (auto itLink = m_links.begin(); itLink != m_links.end(); ++itLink)
	{
		AppendWebAddress(*itLink, new CXfWebAddressBoundingBox(DocCoord(0, 0), DocCoord(0, 0), 
			CXfWebAddress::External, (*itLink)->m_str.c_str(), L""));
		AddLink(textLinks, *itLink);
	}

	if (textLinks.size())
	{
		// Define a default URL colour and underline for the whole document here
		UINT uiChapterIdx = 0;
		pDoc->FindFirstNode(TAG_CHAPTER, false, &uiChapterIdx);
		CLinkProps props;
		bool bRes = EvaluateDefaultURLProperties(textLinks, props);
		m_bUnderlineURL = props.m_bUnderline;
		if (bRes && m_bUnderlineURL)
			pDoc->InsertBefore(uiChapterIdx, new CXfZeroSizedNode(TAG_UNDERLINE_URLS));
		if (bRes && props.m_pColor)
		{
			m_bDefURLPropertiesEvaluated = true;
			m_pURLColour = (CXfComplexColor*)pDoc->InsertBefore(uiChapterIdx + 1, new CXfComplexColor(*props.m_pColor));
			pDoc->InsertBefore(uiChapterIdx + 2, new CXfDocumentURLColour(m_pURLColour));
			// Now also add the default mouse over and visited URL colours
			VXfTreeNode* pMouseOver = pDoc->InsertBefore(uiChapterIdx + 3, new CXfComplexColor(0, 0.666667, 1, 0.933333));
			pDoc->InsertBefore(uiChapterIdx + 4, new CXfDocumentMouseOverURLColour(pMouseOver));
			VXfTreeNode* pVisited = pDoc->InsertBefore(uiChapterIdx + 5, new CXfComplexColor(0, 0.755162, 0.81295, 0.545098));
			pDoc->InsertBefore(uiChapterIdx + 6, new CXfDocumentVisitedURLColour(pVisited));
		}

		if (m_bDefURLPropertiesEvaluated)
		{
			for (int i = 0; i < textLinks.size(); i++)
			{
				VXfTextRenderable* pString = textLinks[i];
				RemoveURLProperties(pString);
			}
		}
	}

	return true;
}

/******************************************************************************
   Function:	XarOutputDev::BeginAnnotations
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		void
   Notes:       
******************************************************************************/
void XarOutputDev::BeginAnnotations()
{
	VXfTreeNode* pLayer = nullptr;

	if (g_settings.bNoLayers)
		pLayer = GetDefLayer();

	if (!pLayer)
	{
		pLayer = new CXfLayer();
		m_pPage->AppendChild(pLayer);
		pLayer->AppendChild(new CXfLayerDetails(TAG_LAYER_FLAG_VISIBLE | TAG_LAYER_FLAG_PRINTABLE, L"Annotations"));
	}

	SetParentNode(pLayer);
}

void XarOutputDev::AddToAnnotationContext(VXfBounded* pNode)
{
	if (m_pAnnots)
	{
		m_pAnnots->AddElement(pNode);

		if (pNode->IsTextString())
		{
			CXfTextString *pTextString = (CXfTextString *)pNode;

			size_t spacePos = pTextString->m_str.find_first_of(L" ");

			if (spacePos >= pTextString->m_str.length() - 1 &&
			    (pTextString->m_str.length() >= 8 && 0 == _wcsnicmp(pTextString->m_str.c_str(), L"https://", 8) ||
			     pTextString->m_str.length() >= 7 && 0 == _wcsnicmp(pTextString->m_str.c_str(), L"http://", 7) ||
			     pTextString->m_str.length() >= 6 && 0 == _wcsnicmp(pTextString->m_str.c_str(), L"ftp://", 6) ||
			     pTextString->m_str.length() >= 4 && 0 == _wcsnicmp(pTextString->m_str.c_str(), L"www.", 4)))
				 m_pAnnots->m_links.insert(pTextString);
		}
	}
}

/******************************************************************************
   Function:	ResortLinkProps
   Author:		Paul Gruzdev
   Purpose:     Do a bubble insertion - move the current up if it has
				more count value than the previous link prop object
   Returns:		void
   Parameters:	std::vector<CLinkProps> & lsa  
   Parameters:	int idx  
   Notes:       
******************************************************************************/
void ResortLinkProps(std::vector<CLinkProps>& lsa, int idx)
{
	while (idx && lsa[idx].m_uiCount > lsa[idx - 1].m_uiCount)
	{
		std::swap(lsa[idx], lsa[idx - 1]);
		idx--;
	}
}
/******************************************************************************
   Function:	AddSortLink
   Author:		Paul Gruzdev
   Purpose:     Add a link property to sorted array and update the array
				to remain sorted
   Returns:		void
   Parameters:	std::vector<CLinkProps> & lsa  
   Parameters:	CLinkProps lp  
   Notes:       
******************************************************************************/
void AddSortLink(std::vector<CLinkProps>& lsa, CLinkProps lp)
{
	int iInsertIdx = 0;
	bool bInserted = false;
	for (int i = 0; i < lsa.size(); i++)
	{
		if (lp == lsa[i])
		{
			lsa[i].m_uiCount += lp.m_uiCount;
			ResortLinkProps(lsa, i);
			bInserted = true;
			break;
		}
		else if (lsa[i].m_uiCount >= lp.m_uiCount)
			iInsertIdx = i + 1;
	}
	if (!bInserted)
	{
		if (iInsertIdx >= lsa.size())
			lsa.push_back(lp);
		else
			lsa.insert(lsa.begin() + iInsertIdx, lp);
	}
}

/******************************************************************************
   Function:	AnnotationContext::EvaluateDefaultURLProperties
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		void
   Parameters:	std::vector<CXfTextString * > & textLinks  
   Notes:       
******************************************************************************/
bool AnnotationContext::EvaluateDefaultURLProperties(std::vector<VXfTextRenderable*>& textLinks, CLinkProps& out)
{
	std::vector<CLinkProps> linkSortArray;
	for (int i = 0; i < textLinks.size(); i++)
	{
		if (!textLinks[i]->IsTextString()) // skip kerns as they don't contain any words
			continue;
		CXfTextString* pString = (CXfTextString*)textLinks[i];
		if (!pString->Context().m_pFill || !pString->Context().m_pFill->IsFlatFill()) continue;

		CXfFlatFill* pFill = (CXfFlatFill*)pString->Context().m_pFill;

		bool bUnderline = pString->TextContext().m_bUnderlineOn;
		CLinkProps lnk((CXfComplexColor*)pFill->pReference, bUnderline, CLinkProps::CalculateValuableTextLength(pString->m_str));

		AddSortLink(linkSortArray, lnk);
	}

	if (linkSortArray.size())
		out = linkSortArray[0];

	return linkSortArray.size() > 0;
}

/******************************************************************************
   Function:	AnnotationContext::ConvertURLColourHSV
   Author:		Paul Gruzdev
   Purpose:     After all processing of all pages - convert the URL colour to 
				HSV as required
   Returns:		void
   Notes:       
******************************************************************************/
void AnnotationContext::ConvertURLColourHSV()
{
	if (m_pURLColour)
	{
		double H, S, V;
		COLORREF cr =m_pURLColour->Color;
		RGB2HSV(GetRValue(cr) / 255.0, GetGValue(cr) / 255.0, GetBValue(cr) / 255.0, H, S, V);
		m_pURLColour->ColourModel = CXfComplexColor::cmHSV;
		m_pURLColour->Desc1 = H;
		m_pURLColour->Desc2 = S;
		m_pURLColour->Desc3 = V;
		m_pURLColour->Desc4 = 0;
	}
}

std::wstring UTF32ToUTF16(const Unicode *utf32, size_t len)
{
	std::wstring res;
	res.reserve(len);

	for (size_t i = 0; i < len; ++i)
	{
		Unicode c = utf32[i];

		if (c > 0xFFFF)
		{
			c -= 0x10000;
			res.append(1, 0xD800 + 0x03FF & (c >> 10));
			res.append(1, 0xDC00 + 0x03FF & c);
		}
		else
			res.append(1, c);
	}

	return res;
}

void AnnotationContext::ImportOutlineItem(PDFDoc* pdfDoc, OutlineItem *pItem, rapidjson::GenericValue<rapidjson::UTF16<>> &jsonObj, rapidjson::GenericDocument<rapidjson::UTF16<>> &d)
{
	using namespace rapidjson;

	std::wstring title = UTF32ToUTF16(pItem->getTitle(), pItem->getTitleLength());

	jsonObj.AddMember(L"Title", GenericValue<UTF16<>>(title.c_str(), (SizeType)title.length(), d.GetAllocator()), d.GetAllocator());
	jsonObj.AddMember(L"IsOpen", pItem->isOpen(), d.GetAllocator());

	if (const LinkAction *pAction = pItem->getAction())
	{
		switch (pAction->getKind())
		{
		case actionURI:
			{
				LinkURI *pURIAction = (LinkURI *)pAction;
				GenericValue<UTF16<>> action(kObjectType);

				const wchar_t *type = L"URI";
				action.AddMember(L"Type", GenericValue<UTF16<>>(type, (SizeType)wcslen(type), d.GetAllocator()), d.GetAllocator());

				std::wstring uri = TextStringToUnicode(pURIAction->getURI());
				action.AddMember(L"URI", GenericValue<UTF16<>>(uri.c_str(), (SizeType)uri.length(), d.GetAllocator()), d.GetAllocator());

				jsonObj.AddMember(L"Action", action, d.GetAllocator());
			}
			break;
		case actionGoTo:
			{
				LinkGoTo *pGoToAction = (LinkGoTo *)pAction;
				GenericValue<UTF16<>> action(kObjectType);

				const wchar_t *type = L"GoTo";
				action.AddMember(L"Type", GenericValue<UTF16<>>(type, (SizeType)wcslen(type), d.GetAllocator()), d.GetAllocator());

				const LinkDest* pDest = pGoToAction->getDest();
				if (!pDest && pGoToAction->getNamedDest())
					pDest = pdfDoc->getCatalog()->findDest(pGoToAction->getNamedDest());

				if (pDest)
				{
					GenericValue<UTF16<>> dest(kObjectType);

					int pageNum = GetPageNum(pDest);
					Page *page = nullptr;

					if (pageNum)
						page = m_pdfDoc->getPage(pageNum);

					double x1 = 0, y1 = 0;

					if (page)
					{
						const PDFRectangle *cropBox = page->getCropBox();
						x1 = cropBox->x1;
						y1 = cropBox->y1;
					}

					/*std::wstring destName = AssignLink(pDest);
					if (destName.empty())
					{
						if (pageNum)
						{
							auto it = m_pageToTreeIdx.find(pageNum);
							if (m_pageToTreeIdx.end() != it)
							{
								pageNum = it->second + 1;
								dest.AddMember(L"PageNum", pageNum, d.GetAllocator());
							}
						}
					}
					else
						dest.AddMember(L"DestName", GenericValue<UTF16<>>(destName.c_str(), (SizeType)destName.length(), d.GetAllocator()), d.GetAllocator());*/

					std::wstring pageName = GetPageName(pageNum);
					dest.AddMember(L"PageName", GenericValue<UTF16<>>(pageName.c_str(), (SizeType)pageName.length(), d.GetAllocator()), d.GetAllocator());

					const wchar_t *destType = nullptr;
					switch (pDest->getKind())
					{
					case destXYZ:
						destType = L"XYZ";
						dest.AddMember(L"Left", pDest->getLeft() - x1, d.GetAllocator());
						dest.AddMember(L"Top", pDest->getTop() - y1, d.GetAllocator());
						dest.AddMember(L"Zoom", pDest->getZoom(), d.GetAllocator());
						break;
					case destFit:
						destType = L"Fit";
						break;
					case destFitH:
						destType = L"FitH";
						dest.AddMember(L"Top", pDest->getTop() - y1, d.GetAllocator());
						break;
					case destFitV:
						destType = L"FitV";
						dest.AddMember(L"Left", pDest->getLeft() - x1, d.GetAllocator());
						break;
					case destFitR:
						destType = L"FitR";
						dest.AddMember(L"Left", pDest->getLeft() - x1, d.GetAllocator());
						dest.AddMember(L"Top", pDest->getTop() - y1, d.GetAllocator());
						dest.AddMember(L"Right", pDest->getRight() - x1, d.GetAllocator());
						dest.AddMember(L"Bottom", pDest->getBottom() - y1, d.GetAllocator());
						break;
					case destFitB:
						destType = L"FitB";
						break;
					case destFitBH:
						destType = L"FitBH";
						dest.AddMember(L"Top", pDest->getTop() - y1, d.GetAllocator());
						break;
					case destFitBV:
						destType = L"FitBV";
						dest.AddMember(L"Left", pDest->getLeft() - x1, d.GetAllocator());
						break;
					}

					if (destType)
						dest.AddMember(L"Type", GenericValue<UTF16<>>(destType, (SizeType)wcslen(destType), d.GetAllocator()), d.GetAllocator());

					action.AddMember(L"Dest", dest, d.GetAllocator());
				}

				jsonObj.AddMember(L"Action", action, d.GetAllocator());
			}
			break;
		}
	}

	if (pItem->hasKids())
	{
		pItem->open();

		if (auto pKids = pItem->getKids())
		{
			GenericValue<UTF16<>> kids(kArrayType);

			for (size_t i = 0; i < pKids->size(); ++i)
			{
				GenericValue<UTF16<>> item(kObjectType);
				ImportOutlineItem(pdfDoc, (*pKids)[i], item, d);
				kids.GetArray().PushBack(item, d.GetAllocator());
			}

			jsonObj.AddMember(L"Kids", kids, d.GetAllocator());
		}

		pItem->close();
	}
}

bool AnnotationContext::ProcessTOC(PDFDoc* pdfDoc, VXfTreeNode* pNode)
{
	using namespace rapidjson;

	if (Outline *pOutline = pdfDoc->getOutline())
	{
		if (auto pItems = pOutline->getItems())
		{
			if (size_t itemCount = pItems->size())
			{
				GenericDocument<UTF16<>> d;
				d.SetObject();
				GenericValue<UTF16<>> v(kArrayType);

				for (size_t i = 0; i < itemCount; ++i)
				{
					GenericValue<UTF16<>> item(kObjectType);
					ImportOutlineItem(pdfDoc, (*pItems)[i], item, d);
					v.GetArray().PushBack(item, d.GetAllocator());				
				}

				d.AddMember(L"Outline", v, d.GetAllocator());

				GenericStringBuffer<UTF16<>> buffer;
				Writer<GenericStringBuffer<UTF16<>>, UTF16<>, UTF16<>> writer(buffer);
				d.Accept(writer);

				// Should be located before any Spread objects (DDES-14847)
				pNode->InsertFirstChild(new CXfUserValue(L"PDF\\JSON", (const wchar_t *)buffer.GetString()));

				//`std::cout << (const wchar_t *)buffer.GetString() << std::endl;
				return true;
			}
		}
	}

	return false;
}

std::wstring AnnotationContext::GetPageName(int index) const
{
	std::wstring name;

	wchar_t buf[100];
	swprintf(buf, L"page%d", index);
	name = buf;

	return name;
}
