/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
//
// XarOutputDev.cc
//
#include "stdafx.h"
#include "XarOutputDev.h"
#include "PathHelpers.h"

/******************************************************************************
   Function:	XarOutputDev::DoClip
   Author:		Paul Gruzdev
   Purpose:     Perform even-odd/non-zero clip operation to Xara's tree.
				Take current gfx path and make a clipview controller out of it.
				If clip to stroke is set, make an outline out of the stroked path
				and clip to it.
   Returns:		void
   Parameters:	GfxState * state  
   Parameters:	bool bClipToStroke - clip to stroke path
   Parameters:	BYTE byWindingRule - even-odd or non-zero
   Notes:       
******************************************************************************/
void XarOutputDev::DoClip(GfxState* state, bool bClipToStroke, BYTE byWindingRule)
{
START_STOPWATCH_FN
	MergeSmallBitmaps();

	GfxPath* path = NULL;

	if (bClipToStroke)
	{
		// We need to create an outline of the current path, as Camelot does not allow clipping to stroke.
		if (path = CreatePathOutline(state->getPath(), state))
		{
			m_clip.Add(path, state, CXF_WINDING_RULE_NONZERO);
			delete path;
		}
	}
	else
		m_clip.Add(state->getPath(), state, byWindingRule);

STOP_STOPWATCH(__FUNCTION__)
}

XarOutputDev::ClipState::ClipState(): m_bNeedRebuild(false), m_pClipController(nullptr)
{
	Init();
}

XarOutputDev::ClipState::~ClipState()
{
	FreeClipPaths();
}

void XarOutputDev::ClipState::Init()
{
	m_bInfinity = true;
	m_windingRule = CXF_WINDING_RULE_NONZERO;
	m_polygons.clear();
}

void XarOutputDev::ClipState::FreeClipPaths()
{
	for (PathStack::const_iterator it = m_paths.begin(); it != m_paths.end(); ++it)
		if (it->first) delete it->first;
	m_paths.clear();
}

void XarOutputDev::ClipState::Reset(INT x, INT y, INT width, INT height)
{
	FreeClipPaths();
	Init();
	m_bNeedRebuild = false;

	GfxPath path;
	AddRect(&path, x, y, width, height);
	Add(&path, NULL, CXF_WINDING_RULE_NONZERO);
}

void XarOutputDev::ClipState::Save()
{
	m_paths.push_back(std::make_pair((ClipperLib::Paths *)NULL, CXF_WINDING_RULE_NONZERO));
}

const ClipperLib::Paths *XarOutputDev::ClipState::GetTopPath() const
{
	return m_paths.size() > 1 ? m_paths.back().first : NULL;
}

bool XarOutputDev::ClipState::Restore()
{
	bool bChanged = false;
	while (m_paths.size() > 1 && m_paths.back().first)
	{
		delete m_paths.back().first;
		m_paths.pop_back();
		bChanged = true;
		m_pClipController = nullptr;
	}
	if (m_paths.size() > 1) m_paths.pop_back();

	if (bChanged && !m_bNeedRebuild)
	{
		Init();
		m_bNeedRebuild = true;
	}

	return bChanged;
}

bool XarOutputDev::ClipState::Empty()
{
	return GetPolygons().empty();
}

void XarOutputDev::ClipState::Add(GfxPath *path, GfxState* state, BYTE windingRule)
{
	ClipperLib::Paths *polygons = new ClipperLib::Paths();
	AddPath(*polygons, path, state);
	if (polygons->empty())
	{
		delete polygons;
		return;
	}

	m_pClipController = nullptr;

	Rebuild();
	m_paths.push_back(std::make_pair(polygons, windingRule));
	Update(polygons, windingRule);
}

void XarOutputDev::ClipState::Rebuild()
{
	if (m_bNeedRebuild)
	{
		m_bNeedRebuild = false;
		for (PathStack::const_iterator it = m_paths.begin(); it != m_paths.end(); ++it)
		{
			if (it->first) Update(it->first, it->second);
		}
	}
}

void XarOutputDev::ClipState::Update(ClipperLib::Paths *polygons, BYTE windingRule)
{
	if (m_bInfinity)
	{
		m_bInfinity = false;
		m_windingRule = windingRule;
		m_polygons = *polygons;
	}
	else if (!m_polygons.empty())
	{
		try
		{
			ClipperLib::Clipper clipper;
			clipper.AddPaths(*polygons, ClipperLib::ptSubject, true);
			clipper.AddPaths(m_polygons, ClipperLib::ptClip, true);
			if (clipper.Execute(ClipperLib::ctIntersection, m_polygons, XaraFillRuleToPoly(windingRule), XaraFillRuleToPoly(m_windingRule)))
			{
				m_windingRule = CXF_WINDING_RULE_NONZERO;
			}
		}
		catch (ClipperLib::clipperException& e) { e; }
	}
}

const ClipperLib::Paths &XarOutputDev::ClipState::GetPolygons()
{
	Rebuild();
	return m_polygons;
}

BYTE XarOutputDev::ClipState::GetWindingRule()
{
	Rebuild();
	return m_windingRule;
}

CXfPath* XarOutputDev::ClipState::CreateXaraPath(UINT tag)
{
	CXfPath* pPath = new CXfPath(tag);
	PolygonsToPath(pPath, GetPolygons());
	return pPath;
}

CXfClipviewController *XarOutputDev::ClipState::GetClipviewController(bool bCreate/* = false*/)
{
	if (!m_pClipController || bCreate)
	{
		CXfPath* pClipPath = CreateXaraPath(TAG_PATH_FILLED_STROKED);
		m_pClipController = CreateClipviewController(pClipPath, GetWindingRule());
	}
	return m_pClipController;
}
