/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#include "stdafx.h"
#include "Panose.h"
#include "Fonts.h"
#include "PathHelpers.h"
#include "../../poppler/fofi/FoFiTrueType.h"
#include "../../poppler/splash/SplashPath.h"
#include "../../poppler/poppler/Gfx.h"
#include "Type3GlyphVectorizer.h"

#include "../../common/XFPathHelpers.h"

#include <ft2build.h>
#include FT_FREETYPE_H

#define BI_CODE(bBold, bItalic) (bBold ? 1 : 0) + (bItalic ? 2 : 0)
#define IS_BOLD(val) ((val & 1) ? true : false)
#define IS_ITALIC(val) ((val & 2) ? true : false)

#define CHAR_PAIR2KEY(c1,c2) ((((DWORD32)c1) << 16) | (DWORD32)c2)
#define KEY2CHAR1(key) ((wchar_t)(key >> 16))
#define KEY2CHAR2(key) ((wchar_t)(key & 0x0000ffff))

#ifdef _DEBUG
#define LOG_FONT(a) std::cout << a
#define LOGW_FONT(a) std::wcout << a
#else
#define LOG_FONT(a) 
#define LOGW_FONT(a) 
#endif

const wchar_t* g_pszEmbPostfix = L"-emb";

extern UINT MapUnicode2Wingdins(UINT uUnicode);
extern UINT MapUnicode2Symbol  (UINT uUnicode);
extern UINT MapWingdingsName2Uni(const char* name);

#ifdef FILTER_DEBUG
#include "common/HiResolutionTimer.h"
double g_dCompareGlyphTime = 0;
#endif

std::wstring Utf8ToUtf16(const std::string &s)
{
	std::wstring ret;
	int len = MultiByteToWideChar(CP_UTF8, 0, s.c_str(), (int)s.length(), NULL, 0);
	if (len > 0)
	{
		ret.resize(len);
		MultiByteToWideChar(CP_UTF8, 0, s.c_str(), (int)s.length(), const_cast<wchar_t*>(ret.c_str()), len);
	}
	return ret;
}

/******************************************************************************
   Function:	FontManager::CompareOutlines
   Author:		Paul Gruzdev
   Purpose:     Compare two glyph outlines the following way: convert each [bezier] outline into a set of line vertexes,
				with certain accuracy, then use Clipper lib to generate a difference polygon. Then calc it's 
				square and return it. Normal measure of a glyph to differ is 1.5% of square difference between glyphs.
   Returns:		bool
   Parameters:	PBYTE Verbs1  
   Parameters:	DocCoord * Coords1  
   Parameters:	DWORD dwCount1  
   Parameters:	PBYTE Verbs2  
   Parameters:	DocCoord * Coords2  
   Parameters:	DWORD dwCount2  
   Parameters:	double dThreshold  
   Notes:       
******************************************************************************/
double FontManager::CompareOutlines(PBYTE Verbs1, DocCoord* Coords1, DWORD dwCount1, PBYTE Verbs2, DocCoord* Coords2, DWORD dwCount2, CGlyphDiff& diff)
{
	ClipperLib::Paths out1, out2;
	RECT rcb1, rcb2;
	diff.iNumSubpaths1 = FlattenConvertPathEx(Verbs1, Coords1, dwCount1, out1, g_dFlatnessThreshold, rcb1);
	diff.iNumSubpaths2 = FlattenConvertPathEx(Verbs2, Coords2, dwCount2, out2, g_dFlatnessThreshold, rcb2);

	ClipperLib::Paths result;
	ClipperLib::Clipper c;
	c.AddPaths(out1, ClipperLib::ptSubject, true);
	c.AddPaths(out2, ClipperLib::ptClip, true);
	bool bRes = c.Execute(ClipperLib::ctXor, result, ClipperLib::pftNonZero, ClipperLib::pftNonZero);

	double dArea = 0;
	for(size_t i = 0; i < result.size(); i++)
	{
		dArea += ClipperLib::Area(result[i]);
	}

	double dGlyphSquare = max((double)(rcb1.right-rcb1.left)*(double)(rcb1.bottom-rcb1.top), (double)(rcb2.right-rcb2.left)*(double)(rcb2.bottom-rcb2.top));
	diff.xor = dArea;
	diff.square = dGlyphSquare;

	//dArea /= dGlyphSquare > 0 ? dGlyphSquare : 72000.0*72000.0;
	dArea /= 72000.0*72000.0;

	return dArea;
}

/******************************************************************************
   Function:	FontManager::CompareGlyphs
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		bool
   Parameters:	CXfFontGlyph * pGlyph1  
   Parameters:	CXfFontGlyph * pGlyph2  
   Notes:       
******************************************************************************/
bool FontManager::CompareGlyphs(CXfFontGlyph* pGlyph1, CXfFontGlyph* pGlyph2)
{
	if(pGlyph1->isBold() != pGlyph2->isBold() || pGlyph1->isItalic() != pGlyph2->isItalic())
		return false;

	LONG lAbsWidthDiff = abs(pGlyph1->Width.get() - pGlyph2->Width.get());
	if(pGlyph1->Verbs.size() == 0)
	{
		return pGlyph2->Verbs.size() == 0 ? lAbsWidthDiff <= GLYPH_WIDTH_DIFF_THRESGHOLD : false;
	}
	if(pGlyph2->Verbs.size() == 0)
	{
		return false; // first glyphs has >0 verbs
	}

	if(lAbsWidthDiff > GLYPH_WIDTH_DIFF_THRESGHOLD)
		return false;

	CGlyphDiff diff(pGlyph1->CharCode);
	double dArea = CompareOutlines(&pGlyph1->Verbs[0], &pGlyph1->Points[0], (DWORD)pGlyph1->Verbs.size(), 
		&pGlyph2->Verbs[0], &pGlyph2->Points[0], (DWORD)pGlyph2->Verbs.size(), diff);

	return diff.GlyphsSimilar();
}

/******************************************************************************
   Function:	EnumerateDocumentGlyphs
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		CXfEmbeddedGlyphSizeHint*
   Parameters:	VXfTreeNode * pChapter  
   Notes:       
******************************************************************************/
CXfEmbeddedGlyphSizeHint* EnumerateDocumentGlyphs(VXfTreeNode* pChapter)
{
	START_STOPWATCH_FN;
	VXfTreeNodeVector fonts;
	pChapter->FindNodes(CTagSet(TAG_FONT_DEF_TRUETYPE, TAG_FONT_DEF_ATM), fonts, false);

	if(!fonts.size()) return NULL;
	CXfEmbeddedGlyphSizeHint* pHint = new CXfEmbeddedGlyphSizeHint();
	for(VXfTreeNodeVector::iterator it = fonts.begin(); it != fonts.end(); ++it)
	{
		CXfFontDef* pDef = (CXfFontDef*)*it;
		pHint->NumFontVariants = pHint->NumFontVariants + pDef->NumVariants();
		pHint->NumGlyphs = pHint->NumGlyphs + (ULONG)pDef->NumGlyphs();
		pHint->TotalNumPoints = pHint->TotalNumPoints + (ULONG)pDef->CalcGlyphVerbCount();
	}
	STOP_STOPWATCH(__FUNCTION__)
	return pHint;
}

/******************************************************************************
   Function:    GetFTCharacterUnicode
   Author:      Pavel Ostapenko
   Purpose:		Get unicode value of character code within character map 
                specified by FreeType enumeration value:
				http://www.freetype.org/freetype2/docs/reference/ft2-base_interface.html#FT_Encoding
				Standard character maps use Win API for conversion, official 
				code page list for conversion is available here:
				http://msdn.microsoft.com/en-us/library/windows/desktop/dd317756(v=vs.85).aspx
				Non standard maps use custom conversion tables recommended in
				FreeType documentation which can be found here:
				http://www.kostis.net/charsets/
   Returns:     unicode value
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
FT_ULong GetFTCharacterUnicode(FT_UInt chr, FT_Encoding encoding)
{
	FT_ULong out = 0;
	UINT codePage = 0;
	switch (encoding)
	{
	case FT_ENCODING_SJIS:        codePage = 932;   break;
	case FT_ENCODING_GB2312:      codePage = 936;   break;
	case FT_ENCODING_BIG5:        codePage = 950;   break;
	case FT_ENCODING_WANSUNG:     codePage = 20949; break;
	case FT_ENCODING_JOHAB:       codePage = 1361;  break;
	case FT_ENCODING_OLD_LATIN_2: codePage = 852;   break;
	case FT_ENCODING_APPLE_ROMAN: codePage = 10000; break;
	case FT_ENCODING_MS_SYMBOL:
		{
			static const FT_ULong table[0x100] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,8704,35,8707,37,38,8715,40,41,8727,43,44,8722,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,8773,913,914,935,916,917,934,915,919,921,977,922,923,924,925,927,928,920,929,931,932,933,962,937,926,936,918,91,8756,93,8869,95,8254,945,946,967,948,949,966,947,951,953,981,954,955,956,957,959,960,952,961,963,964,965,982,969,958,968,950,123,124,125,8764,978,8242,8804,8725,8734,402,9827,9830,9829,9824,8596,8592,8593,8594,8595,176,177,8243,8805,215,8733,8706,8729,247,8800,8801,8776,8943,8629,8465,8476,8472,8855,8853,8709,8745,8746,8835,8839,8836,8834,8838,8712,8713,8736,8711,174,169,8482,8719,8730,8901,172,8743,8744,8660,8656,8657,8658,8659,9674,9001,8721,9002,8747,8992,8993,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};
			return chr > 0xFF ? static_cast<FT_ULong>(chr) : table[chr];
		}
	case FT_ENCODING_ADOBE_STANDARD:
		{
			static const FT_ULong table[0x100] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,161,162,163,8260,165,402,167,164,8220,171,8249,8250,64257,64258,8211,8224,8225,183,182,8729,8218,8222,8221,187,8943,8240,191,768,180,770,771,175,728,729,168,176,184,733,731,711,8212,198,170,321,216,338,186,230,305,322,248,339,223,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};
			return chr > 0xFF ? static_cast<FT_ULong>(chr) : table[chr];
		}
	case FT_ENCODING_ADOBE_LATIN_1:
		{
			static const FT_ULong table[0x100] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,126,305,768,769,770,771,772,774,775,776,778,807,779,808,780,160,161,162,163,164,165,166,167,168,169,170,171,172,173,174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255,237,238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,254,255};
			return chr > 0xFF ? static_cast<FT_ULong>(chr) : table[chr];
		}
	}

	if (!codePage)
		return static_cast<FT_ULong>(chr);
	MultiByteToWideChar(codePage, 0, reinterpret_cast<LPCSTR>(&chr), (int)strnlen(reinterpret_cast<const char *>(&chr),sizeof(chr)), reinterpret_cast<wchar_t*>(&out), sizeof(out));
//	std::cout << std::hex << chr << " --> " << out << std::endl;
	return out;
}

class SplashFTFontFile_ : public SplashFontFile
{
public:
	virtual ~SplashFTFontFile_() {}
	virtual SplashFont *makeFont(SplashCoord *mat, SplashCoord *textMat) {}

/******************************************************************************
   Function:    FindGlyphsByCodeToGid
   Author:      Paul Gruzdev
   Purpose:		Scan existing charcode to glyph index map, find non-zero
				entries and move them to glyph index to charcode and to unicode
				maps
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
	bool FindGlyphsByCodeToGid(std::map<int, int>& mGlyphsToCharcode, std::map<int, int>& mGlyphsToUnicode, 
						   FT_Encoding encoding)
	{START_STOPWATCH_FN
		bool bRet = false;
		if (codeToGID && codeToGIDLen)
		{
			for (int i = 0; i < codeToGIDLen; i++)
			{
				if (codeToGID[i] > 0) // it also could be -1 which is invalid obv
				{
					mGlyphsToCharcode.insert(std::make_pair(codeToGID[i], i));
					mGlyphsToUnicode.insert(std::make_pair(codeToGID[i], GetFTCharacterUnicode(i, encoding)));
					bRet = true;
				}
			}
STOP_STOPWATCH(__FUNCTION__)
			return bRet;
		}
		return bRet;
	}

public:
	SplashFTFontEngine *engine;
	FT_Face face;
	int* codeToGID;
	int codeToGIDLen;
	bool trueType;
	bool type1;
};

int parseCharName(GfxFont *font, int charCode, Unicode *uBuf, int uLen)
{
	if (!font->isCIDFont() && fontType3 != font->getType() && font->hasToUnicodeCMap() && !font->isSymbolic())
	{
		if (const char *charName = ((Gfx8BitFont *)font)->getCharName(charCode))
		{
			if (strcmp(charName, ".notdef"))
			{
				return GfxFont::parseCharName((char *)charName, uBuf, uLen, true, true, false, false, true);
			}
		}
	}

	return 0;
}

void AdjustUnicode(Unicode *u, int uLen)
{
	if (1 == uLen)
	{
		switch (*u)
		{
		case 0x09: // TAB
			*u = ' ';
			break;

		case 0xAD: // SOFT HYPHEN
			*u = '-';
			break;
		}
	}
}

/******************************************************************************
   Function:	MapLigature
   Author:		Paul Gruzdev
   Purpose:     Given a ligature notation in u, map it to it's native Unicode
				value. This is done because the embedded font may not have the
				individual glyphs that combine this ligature, i.e. the font has
				a glyph 'fi', but may not have 'f' and 'i'
   Returns:	Unicode
   Parameters:	Unicode * u  
   Parameters:	int uLen  
   Notes:       
******************************************************************************/
Unicode MapLigature(const Unicode* u, int uLen)
{
	switch (uLen)
	{
	case 2:
		switch (u[0])
		{
		case 'f':
			switch (u[1])
			{
			case 'f':
				return 0xFB00;
			case 'i':
				return 0xFB01;
			case 'l':
				return 0xFB02;
			case 't':
				return 0xFB05;
			}
			break;
		case 's':
			if ('t' == u[1])
				return 0xFB06;
			break;
		}
		break;
	case 3:
		if ('f' == u[0] && 'f' == u[1]) switch (u[2])
		{
		case 'i':
			return 0xFB03;
		case 'l':
			return 0xFB04;
		}
		break;
	}

	return 0;
}

int MapLigature(Unicode u, Unicode *buf)
{
	switch (u)
	{
	case 0xFB00:
		buf[0] = 'f';
		buf[1] = 'f';
		return 2;
	case 0xFB01:
		buf[0] = 'f';
		buf[1] = 'i';
		return 2;
	case 0xFB02:
		buf[0] = 'f';
		buf[1] = 'l';
		return 2;
	case 0xFB03:
		buf[0] = 'f';
		buf[1] = 'f';
		buf[2] = 'i';
		return 3;
	case 0xFB04:
		buf[0] = 'f';
		buf[1] = 'f';
		buf[2] = 'l';
		return 3;
	case 0xFB05:
		buf[0] = 'f';
		buf[1] = 't';
		return 2;
	case 0xFB06:
		buf[0] = 's';
		buf[1] = 't';
		return 2;
	}
	return 0;
}

/******************************************************************************
   Function:	MappedFontDesc::GetFontDef
   Author:		Paul Gruzdev
   Purpose:     Get the font def for this charcode. It may be altered if the character
				is placed in an extra font def: it either a glyph that does not exist in
				the system font, glyph collision second variant or a non-standard ligature.
   Returns:		CXfFontDef*
   Parameters:	int charCode  
   Parameters:	bool & bGlyphSubstituted  
   Notes:       
******************************************************************************/
CXfFontDef* MappedFontDesc::GetFontDef(int charCode, bool& bGlyphSubstituted)
{
	std::map<int, CXfFontDef*>::iterator it = extraNode.find(charCode);
	bGlyphSubstituted = it != extraNode.end();
	return bGlyphSubstituted ? it->second : node;
}


//---- Utility functions

// Windows-specific font em detector, copied with little modification from SVGExport/import/SVGTextImport.cpp
MILLIPOINT GetWindowsFontEm(const std::wstring &strFontName, DWORD weight, bool bItalic, wchar_t chr/*= 'M'*/)
{
	INT width = -1;

	CFontDC hdc(strFontName.c_str(), weight, bItalic);

	static const size_t LENGTH_FONT_FACE = 512;
	wchar_t fontFace[LENGTH_FONT_FACE];
	GetTextFaceW(hdc, LENGTH_FONT_FACE, fontFace);
	if (!wcsncmp(fontFace, strFontName.c_str(), LENGTH_FONT_FACE))
	{
		ABC abc;
		if (GetCharABCWidths(hdc, chr, chr, &abc))
			width = abc.abcA + abc.abcB + abc.abcC;
		else if (!GetCharWidth32(hdc, chr, chr, &width))
			width = -1;
	}
	else
	{
		_ASSERT(0); // We are trying to use the font typeface name that is NOT present in the system.
	}

	// That's exactly the width that Camelot calculates, for comparison with us
	//int iw = MulDiv(iWidth, 72000, 2048); 
	return -1 != width ? MulDiv(width, 72000, 2048) : width;
}

static std::map<SystemFontInfo, SystemFontMetrics> g_SystemFontCache;

// OS-specific character width
MILLIPOINT MappedFontDesc::GetCharWidthOS(wchar_t u)
{
	const SystemFontInfo *pFontKey = NULL;

	if (substitute.fontName.empty())
		pFontKey = this;
	else
		pFontKey = &substitute;

	auto itFont = g_SystemFontCache.find(*pFontKey);
	if (itFont == g_SystemFontCache.end())
		itFont = g_SystemFontCache.insert(std::make_pair(*pFontKey, SystemFontMetrics())).first;

	auto itWidth = itFont->second.widths.find(u);
	if (itWidth == itFont->second.widths.end())
		itWidth = itFont->second.widths.insert(std::make_pair(u, GetWindowsFontEm(pFontKey->fontName, pFontKey->weight, pFontKey->isItalic, u) )).first;

	return itWidth->second;
}

MILLIPOINT MappedFontDesc::GetKernSizeOS(wchar_t c1, wchar_t c2)
{
	const SystemFontInfo *pFontKey = NULL;

	if (substitute.fontName.empty())
		pFontKey = this;
	else
		pFontKey = &substitute;

	auto itFont = g_SystemFontCache.find(*pFontKey);
	if (itFont == g_SystemFontCache.end())
		itFont = g_SystemFontCache.insert(std::make_pair(*pFontKey, SystemFontMetrics())).first;

	if (itFont->second.kerns.empty())
	{
		CFontDC dc(pFontKey->fontName.c_str(), pFontKey->weight, pFontKey->isItalic);

		DWORD dwNumPairs = GetKerningPairs(dc, 0, NULL);
		KERNINGPAIR* pPair = new KERNINGPAIR[dwNumPairs];
		if (GetKerningPairs(dc, dwNumPairs, pPair) == dwNumPairs)
		{
			for (DWORD i = 0; i < dwNumPairs; ++i)
			{
				itFont->second.kerns[CHAR_PAIR2KEY(pPair[i].wFirst, pPair[i].wSecond)] = MulDiv(pPair[i].iKernAmount, 72000, 2048);
			}
		}

		delete [] pPair;
	}

	auto itKern = itFont->second.kerns.find(CHAR_PAIR2KEY(c1, c2));

	return itFont->second.kerns.end() == itKern ? 0 : itKern->second;
}

/******************************************************************************
   Function:	MappedFontDesc::GetCharBoundingBoxOS
   Author:		Paul Gruzdev
   Purpose:     Get the bbox of a system char of a given font and it's variant. 
				Store for faster use it in the bbox map
   Returns:		bool
   Parameters:	wchar_t u  
   Parameters:	Rect & rcBounds  
   Notes:       
******************************************************************************/
bool MappedFontDesc::GetCharBoundingBoxOS(wchar_t u, Rect& rcBounds)
{
	std::map<wchar_t, Rect>::iterator it = bboxMap.find(u);
	if(it != bboxMap.end())
	{
		rcBounds = it->second;
		return true;
	}

	CFontDC hdc(fontName.c_str(), weight, isItalic);

	if (!hdc.GlyphExists(u)) return false;

	CharDescription desc(u, fontName, weight, isItalic);
	DocCoord* pCoord = NULL;
	PathVerb* pVerb = NULL;
	ULONG ulNumCoords = 0;
	if (!TextManager::GetTTCharPath(desc, &pCoord, &pVerb, &ulNumCoords, NULL)) return false;

	rcBounds = FindXFPathBounds(pVerb, pCoord, ulNumCoords);
	bboxMap[u] = rcBounds;

	return true;
}

#pragma pack(push, 1)

struct TTF_HEAD_TABLE
{
	INT   version;
	INT   fontRevision;
	UINT  checkSumAdjustment;
	UINT  magicNumber;
	WORD  flags;
	WORD  unitsPerEm;
};

struct TTF_OS2_TABLE
{
	WORD  version;
	SHORT xAvgCharWidth;
	WORD  usWeightClass;
	WORD  usWidthClass;
	SHORT fsType;
	SHORT ySubscriptXSize;
	SHORT ySubscriptYSize;
	SHORT ySubscriptXOffset;
	SHORT ySubscriptYOffset;
	SHORT ySuperscriptXSize;
	SHORT ySuperscriptYSize;
	SHORT ySuperscriptXOffset;
	SHORT ySuperscriptYOffset;
	SHORT yStrikeoutSize;
	SHORT yStrikeoutPosition;
	SHORT sFamilyClass;
	BYTE  panose[10];
	DWORD ulUnicodeRange1;
	DWORD ulUnicodeRange2;
	DWORD ulUnicodeRange3;
	DWORD ulUnicodeRange4;
	CHAR  achVendID[4];
	WORD  fsSelection;
	WORD  usFirstCharIndex;
	WORD  usLastCharIndex;
	SHORT sTypoAscender;
	SHORT sTypoDescender;
	SHORT sTypoLineGap;
};

#pragma pack(pop)

#define ntohs(x) MAKEWORD(HIBYTE(x), LOBYTE(x))
#define ntohl(x) MAKELONG(ntohs(HIWORD(x)), ntohs(LOWORD(x)))

FontMetrics GetFontMetricsOS(const wchar_t *fontName, int weight, bool isItalic)
{
	FontMetrics res;
	CFontDC hdc(fontName, weight, isItalic);

	static const size_t LENGTH_FONT_FACE = 512;
	wchar_t fontFace[LENGTH_FONT_FACE];
	GetTextFaceW(hdc, LENGTH_FONT_FACE, fontFace);
	if (!wcsncmp(fontFace, fontName, LENGTH_FONT_FACE))
	{
		TTF_HEAD_TABLE head;
		TTF_OS2_TABLE  os2;
		if (sizeof(os2) == GetFontData(hdc, ntohl('OS/2'), 0, &os2, sizeof(os2)) &&
		    ntohs(os2.version) >= 4 && 0 != (ntohs(os2.fsSelection) & 128) &&
		    sizeof(head) == GetFontData(hdc, ntohl('head'), 0, &head, sizeof(head)))
		{
			res.ascent = MulDiv((SHORT)ntohs(os2.sTypoAscender), 72000, (WORD)ntohs(head.unitsPerEm));
#ifdef _ENABLE_RELATIVE_LINE_SPACING
			res.descent = MulDiv((SHORT)ntohs(os2.sTypoDescender - os2.sTypoLineGap), 72000, (WORD)ntohs(head.unitsPerEm));
#else
			res.descent = MulDiv((SHORT)ntohs(os2.sTypoDescender), 72000, (WORD)ntohs(head.unitsPerEm));
#endif
			return res;
		}
		
		TEXTMETRIC metrics;
		if (GetTextMetrics(hdc, &metrics))
		{
			res.ascent  = MulDiv(metrics.tmAscent, 72000, 2048);
#ifdef _ENABLE_RELATIVE_LINE_SPACING
			res.descent = -MulDiv(metrics.tmDescent + metrics.tmExternalLeading, 72000, 2048);
#else
			res.descent = -MulDiv(metrics.tmDescent, 72000, 2048);
#endif
		}
	}
	return res;
}

const SystemFontInfo &MappedFontDesc::GetSubstitute(CXfFontDef* pFontDef, bool bBold, bool bItalic)
{
	if (substitute.fontName.empty())
	{
		if (bPresentInTheSystem)
		{
			substitute.fontName = fontName;
			substitute.weight   = weight;
			substitute.isItalic = isItalic;
		}
		else
			substitute.fontName = pFontManager->FindFontSubstitute(pFontDef, bBold, bItalic, substitute.weight, substitute.isItalic);
	}
	return substitute;
}

FontMetrics MappedFontDesc::GetFontMetricsOS()
{
	typedef std::map<std::wstring, FontMetrics> MetricsCache;
	static MetricsCache cache;

	MetricsCache::iterator metricsCacheIt = cache.find(fontName);
	if (metricsCacheIt == cache.end())
	{
		FontMetrics metricsCached = ::GetFontMetricsOS(fontName.c_str(), weight, isItalic);
		metricsCacheIt = cache.insert(std::make_pair(fontName, metricsCached)).first;
	}

	if (metricsCacheIt == cache.end())
		return FontMetrics();

	return metricsCacheIt->second;
}

/******************************************************************************
   Function:    MappedFontDesc::AdjustPreciseCharRect
   Author:      Paul Gruzdev
   Purpose:		Find corresponding glyph in the font def and adjust char bbox taking into accoutn
			starting and trailing spaces and glyph bbox.
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
/*bool MappedFontDesc::AdjustPreciseCharRect(wchar_t c, Rect& rc)
{
	if (CXfFontGlyph* pGlyph = node->GetGlyph(c, isBold, isItalic))
	{
		rc = pGlyph->Bounds();
	}
	else
	{
		if (!GetCharBoundingBoxOS(c, rc))
			return false;
	}

	return true;
}*/

FontManager::FontManager(const wchar_t* filename, VXfTreeNode* pChapter, PDFDoc* pDoc, std::string strFilterDataDir, GlobalParams* pGlobalParams):
fontMapper(true), m_filename(filename), m_pChapter(pChapter), m_pDoc(pDoc), m_pGlobalParams(pGlobalParams), m_type3FontNameCount(0), m_dumpGlyphs(filename)
{
	fontEngine = new SplashFontEngine(
		globalParams->getEnableFreeType(),
		false, //enableFreeTypeHinting,
		false, //enableSlightHinting,
		false /* allowAntialias &&
		//globalParams->getAntialias() &&
		colorMode != splashModeMono1*/);

	if (!g_settings.bTextAsOutline)
	{
		InitialiseBuiltInFonts(strFilterDataDir);
		fontMapper.LoadNgramData(); //.PopulateFromResource(_Module.GetResourceInstance(), IDR_PS_FONT_MAP1, L"PS_FONT_MAP");
		/*
		CFileDisk f1(L"current.bin");
		f1.Open(vcFile::orOpenForRead);
		fontMapper.LoadNgramData(&f1);
		f1.Close();
		*/

#ifndef _ENABLE_PREPASS
		LOG_FONT("----------------------\n");

		FontInfoScanner fontScanner(pDoc);
		GooList *fonts = fontScanner.scan(pDoc->getNumPages());
		for (int i=0; i<fonts->getLength(); ++i)
		{
			FontInfo *pFontInfo = reinterpret_cast<FontInfo *>(fonts->get(i));

			Ref &ref = pFontInfo->getRef();
			GetFontDesc(ref);
		}
		delete fonts;

		CreateFontDefs(g_settings.bLoadEmbeddedGlyphs, false);

// 		if(m_settings.bLoadEmbeddedGlyphs)
// 			EmbedDocumentFontGlyphs(false, m_pChapter);

		LOG_FONT("----------------------\n");
#endif
	}
}

FontManager::~FontManager()
{
	if (fontEngine)
		delete fontEngine;
}

/******************************************************************************
   Function:	FontManager::MapCharacter
   Author:		Pavel Ostapenko
   Purpose:     Map Unicode character to be ready to pass to Xara.
   Returns:		UINT
   Parameters:	UINT uiUnicode  
   Notes:       
******************************************************************************/
UINT FontManager::MapCharacter(MappedFontDesc& mfd, UINT uiUnicode, UINT uiCharcode, bool bEmptyGlyph)
{
	std::map<int, int>::iterator it = mfd.charToUni.find(uiCharcode);
	if(it != mfd.charToUni.end())
		return it->second;

	// These two always map to a correct Unicode range (>=32)
	if (mfd.fontName == L"Wingdings")
		uiUnicode = MapUnicode2Wingdins(uiUnicode);
	// Always map symbol, as we don't know whether font will be embedded or not, up till the end
	else if (mfd.fontName == L"Symbol" || mfd.fontName == L"SymbolPS")
		uiUnicode = MapUnicode2Symbol(uiUnicode);
	else
	{
		// this is non-empty SPACE character, need to relocate it
		if ((32 == uiUnicode || 0xA0 == uiUnicode) && !bEmptyGlyph)
			uiUnicode = 0;
		// this is an empty TAB - map it to SPACE, the width will be calculated accordingly. If it's not an empty TAB,
		// relocate it similarly to SPACE
		if (uiUnicode == 9 && bEmptyGlyph)
			uiUnicode = 32;

		// The last resort: we still have a 'bad' character value, map it through our internal map to Unicode private use area
		// 0xFFFD -> DDES-10754, where several glyph indexes were mapped to this 'replacement character'. Probably Unicode map corruptiom.
		if(uiUnicode < 32 || uiUnicode >= 0xFFFD)
		{
			uiUnicode = 0; // reset it so later we could find whether we assigned it or not
			auto it = mfd.badmap.find(uiCharcode);
			if (mfd.badmap.end() != it)
				uiUnicode = it->second;
			else if (mfd.m_pFont && mfd.m_bEmbedded && mfd.m_pFont->getType() <= fontTrueTypeOT)
			{
				Gfx8BitFont* pFont = (Gfx8BitFont*)mfd.m_pFont;
				const char* name = pFont->getCharName(uiCharcode);
				if (name) // encguide.pdf - getCharName returns NULL pointer
				{
					Unicode uni = globalParams->mapNameToUnicodeAll(name);
					if (!uni)
						uni = MapWingdingsName2Uni(name);

					if (uni)
					{
						mfd.badmap[uiCharcode] = uni;
						uiUnicode = uni;
					}
				}
			}

			if (!uiUnicode)
			{
				uiUnicode = mfdMap.GetNextPrivateUnicode(mfd.fontName, mfd.weight, mfd.isItalic)++;
				mfd.badmap[uiCharcode] = uiUnicode;
			}
		}
	}

	mfd.charToUni[uiCharcode] = uiUnicode;
	return uiUnicode;
}

/******************************************************************************
   Function:	FontManager::MapNonStandardLigature
   Author:		Paul Gruzdev
   Purpose:     Non-standard ligature - we map it into Unicode private use area
   Returns:		int
   Parameters:	Unicode * u  
   Parameters:	int uLen  
   Notes:       
******************************************************************************/
Unicode FontManager::MapNonStandardLigature(MappedFontDesc& mfd, Unicode* u, int uLen)
{
	std::wstring str;

	for (int i = 0; i < uLen; i++)
		str.push_back((wchar_t)u[i]);

	auto it = mfd.nsLigatureMap.find(str);
	if (it != mfd.nsLigatureMap.end())
		return it->second;

	Unicode res = mfdMap.GetNextPrivateUnicode(mfd.fontName, mfd.weight, mfd.isItalic)++;
	mfd.nsLigatureMap[str] = res;

	return res;
}

/******************************************************************************
   Function:	FontManager::FindOrCreateExtraFontDef
   Author:		Paul Gruzdev
   Purpose:     Each embedded font that is also present in the system should have at least
				one 'extra font definition (CXfFontDef)' where all the embedded glyphs
				are added. They are also compared against the system glyphs and if a difference
				is found, the 'normal' font def is switched to use the extra font def for 
				all the glyphs of this typeface name. Here 
   Returns:		CXfFontDef*
   Parameters:	const MappedFontDesc &mfd
   Parameters:	CXfFontGlyph * pGlyph
   Notes:       
******************************************************************************/
CXfFontDef* FontManager::FindOrCreateExtraFontDef(const MappedFontDesc &mfd, CXfFontGlyph* pGlyph, bool *pGlyphExists)
{
	if (pGlyphExists) *pGlyphExists = true;

	ExtraFontDefMap::iterator it = extraDefs.find(mfd.pFontDefOrg);
	if(it != extraDefs.end())
	{
		if (pGlyph)
		{
			wchar_t wUnicode = pGlyph->CharCode;
			bool bBold = pGlyph->isBold();
			bool bItalic = pGlyph->isItalic();

			FontDefList& list = it->second;
			for(FontDefList::iterator i = list.begin(); i != list.end(); ++i)
			{
				CXfFontGlyph* pGlyph1 = (*i)->GetGlyph(wUnicode, bBold, bItalic);
				if(pGlyph1 == NULL)
				{
					(*i)->AddGlyph(wUnicode, pGlyph, true);
					(*i)->UpdateEmbeddingPolicy(pGlyph->Flags, mfd.m_fsType);
					if (pGlyphExists) *pGlyphExists = false;
					return (*i);
				}

				if(CompareGlyphs(pGlyph, pGlyph1))
					return (*i);
			}
		}
		else
			return it->second.front();
	}
	else
		it = extraDefs.insert(std::make_pair(mfd.pFontDefOrg, FontDefList())).first;

	std::wstring strBase = mfd.pFontDefOrg->FullFontName;
	/*size_t pos = strBase.find(g_pszEmbPostfix);
	if(pos != std::wstring::npos)
		strBase = strBase.substr(0, pos);*/

	std::wstring strName = strBase + (it->second.size() ? FormatStringW(L"%s%d", g_pszEmbPostfix, it->second.size()) : g_pszEmbPostfix);
	CXfFontDef* pExtraDef = new CXfFontDef(*mfd.pFontDefOrg);
	pExtraDef->FullFontName = strName;
	pExtraDef->TypeFaceName = strName;

	it->second.push_back(pExtraDef);
	m_pChapter->InsertFirstChild(pExtraDef);

	if (pGlyph)
	{
		pExtraDef->AddGlyph(pGlyph->CharCode, pGlyph, true);
		pExtraDef->UpdateEmbeddingPolicy(pGlyph->Flags, mfd.m_fsType);
		if (pGlyphExists) *pGlyphExists = false;
	}

	return pExtraDef;
}

/******************************************************************************
   Function:	FontManager::FindFontSubstitute
   Author:		Denis Onishchenko
   Purpose:     Given that glyphs is NOT embedded, and this font typeface name is 
			NOT present in the system, pick the font substitute name exact the
			same name as Camelot does it
   Returns:	std::wstring, substituteWeight, substituteIsItalic
   Parameters:	CXfFontDef* pFontDef
   Parameters:	bool isBold
   Parameters:	bool isItalic
   Notes:       
******************************************************************************/

std::wstring FontManager::FindFontSubstitute(CXfFontDef* pFontDef, bool isBold, bool isItalic, int &substituteWeight, bool &substituteIsItalic)
{
	std::wstring substituteName = L"Times New Roman";
	substituteWeight   = isBold ? 700 : 400;
	substituteIsItalic = isItalic;

	CCPanose panose;
	panose.SetAll(pFontDef->PANOSE[0], pFontDef->PANOSE[1], pFontDef->PANOSE[2], pFontDef->PANOSE[3], pFontDef->PANOSE[4],
		pFontDef->PANOSE[5], pFontDef->PANOSE[6], pFontDef->PANOSE[7], pFontDef->PANOSE[8], pFontDef->PANOSE[9]);

	// check, is this a 'simple' panose number with just Weight and Letterform components?
	// These 'simple' panose numbers get saved out for old files and ATM fonts. Obviously
	// we can't do any sensible font matching, so just use the default font.
	if (
		PAN_ANY != panose.GetFamilyType() ||
		PAN_ANY != panose.GetSerifStyle() ||
		PAN_ANY != panose.GetProportion() ||
		PAN_ANY != panose.GetContrast() ||
		PAN_ANY != panose.GetStrokeVariation() ||
		PAN_ANY != panose.GetArmStyle() ||
		PAN_ANY != panose.GetMidline() ||
		PAN_ANY != panose.GetXHeight())
	{
		FontMapDataSet::const_iterator it = fontMapper.FindClosestFont(panose);
		if (fontMapper.m_DataSet.end() != it)
		{
			substituteName     = it->m_strTypeface;
			substituteWeight   = it->m_iWeight;
			substituteIsItalic = it->m_bItalic;
		}
	}

	return substituteName;
}

/******************************************************************************
   Function:	FontManager::GetGlyphWidth
   Author:		Paul Gruzdev
   Purpose:     Given the char code (the code that appears in PDF /Contents stream),
				and possible Unicode (system query) - find and store the glyph width 
				of a particular font variant (mfd)
   Returns:		MILLIPOINT
   Parameters:	int charCode  
   Parameters:	wchar_t wUnicode  
   Parameters:	MappedFontDesc * pFontDesc  
   Notes:       You should only call this function after ALL fonts are enumerated and
				all font defs are created and all space characters reconstructed.
******************************************************************************/
/*MILLIPOINT FontManager::GetGlyphWidth(int charCode, wchar_t wUnicode, MappedFontDesc* pFontDesc)
{
	std::map<int, MILLIPOINT>::iterator it = pFontDesc->mGlyphWidthCache.find(charCode);
	if(it != pFontDesc->mGlyphWidthCache.end())
		return it->second;

	bool bGlyphSubstituted = false;
	CXfFontDef* pCurFontDef = pFontDesc->GetFontDef(charCode, bGlyphSubstituted);
	MILLIPOINT glyphWidth = 0;

	// If the glyph is substituted, then it uses another font def record (-emb) that always not present in the system.
	if (pFontDesc->bPresentInTheSystem && !bGlyphSubstituted)
		glyphWidth = pFontDesc->GetCharWidthOS(wUnicode);
	else
	{
		CXfFontDef* pDef = !bGlyphSubstituted ?
			// glyph is NOT ns ligatures, nor absent from the system. 
			(pFontDesc->bPresentInTheSystem ? FindOrCreateExtraFontDef(*pFontDesc) : pFontDesc->node)
			: pCurFontDef;

		//if (pFontDesc->m_bEmbedded)
		//{
		if (CXfFontGlyph* pGlyph = pDef->GetGlyph(wUnicode, pFontDesc->isBold, pFontDesc->isItalic))
		{
			glyphWidth = pGlyph->Width;
		}
		//}
		else
		{
			// The renderer needs a glyph width, that isn't embedded, NOR present in the system. Camelot picks a font
			// substitute in this case and renders with it, so we need to do the same and get the width from the substitute.
			pFontDesc->GetSubstitute(pDef, pFontDesc->isBold, pFontDesc->isItalic);
			// this utilises font substitute name (that IS present in the system).
			glyphWidth = pFontDesc->GetCharWidthOS(wUnicode);
		}
	}

	pFontDesc->mGlyphWidthCache[charCode] = glyphWidth;
	return glyphWidth;
}*/

void FontManager::CreateType3MappedFontDesc(MappedFontDesc &mfd)
{
	mfd.pFontManager = this;

	const double *fontMatrix = mfd.m_pFont->getFontMatrix();
	const double *fontBBox = mfd.m_pFont->getFontBBox();

	double bboxHeight = fontBBox[3] - fontBBox[1];

	mfd.scale = bboxHeight * fontMatrix[3];

	if (bboxHeight)
	{
		mfd.ascent  = _round(fontBBox[3] * 72000. / bboxHeight);
		mfd.descent = _round(fontBBox[1] * 72000. / bboxHeight);
	}
	else
	{
		mfd.ascent  = _round(mfd.m_pFont->getAscent() * 72000.);
		mfd.descent = _round(mfd.m_pFont->getDescent() * 72000.);
	}

	if (mfd.m_fontObj.isDict())
	{
		Dict *fontDict = mfd.m_fontObj.getDict();

		// we should always use generated names for Type3 fonts to avoid using the same name for different fonts (TEXT.pdf, pages 29-30)
		/*Object obj = fontDict->lookup("Name");
		if (obj.isName())
			mfd.fontName = Utf8ToUtf16(obj.getName());*/
	}

	if (mfd.fontName.empty())
		mfd.fontName = FormatStringW(L"T3Font_%d", m_type3FontNameCount++);

	mfd.fontNamePS = mfd.fontNameRaw = mfd.fontName;

	mfd.weight = 400;
	if (GfxFont::WeightNotDefined != mfd.m_pFont->getWeight())
		mfd.weight = (int)mfd.m_pFont->getWeight() * 100;
	mfd.isBold = mfd.m_pFont->isBold();
	mfd.isItalic = mfd.m_pFont->isItalic();

	mfd.bPresentInTheSystem = false;

	CXfFontDef* pFontDef = new CXfFontDef(mfd.fontName, mfd.fontName);
	mfd.pFontDefOrg = mfd.node = pFontDef;
	mfd.emWidth = 72000;

	pFontDef->weight = mfd.weight;

	// Add our font prior any other fonts. Only reliable place, or we need to scan through children and insert before
	// any spread node
	m_pChapter->InsertFirstChild(mfd.node, false);

	if (mfd.m_fontObj.isDict())
	{
		Dict *fontDict = mfd.m_fontObj.getDict();
		BYTE byPanose[10];
		ConstructPANOSE(byPanose, mfd.m_pFont, fontDict);
		MergePANOSE(pFontDef->PANOSE, byPanose);
		//memcpy(pFontDef->PANOSE, (PBYTE)"\x2\xB\x6\x4\x3\x5\x4\xB\x2\x4", 10);
		//MergePANOSE(pFontDef->PANOSE, (PBYTE)"\x2\x0\x5\x3\x3\x0\x0\x2\x0\x3");
		BYTE byAltPANOSE[10]; memset(byAltPANOSE, 0, 10);
		BYTE byClassID = 0, bySubclassID = 0;
		if (ExtractPANOSEandIBMInfo(fontDict, byClassID, bySubclassID, byAltPANOSE))
			MergePANOSE(pFontDef->PANOSE, byAltPANOSE);
	}

	CharCodeToUnicode *ctu = (CharCodeToUnicode *)mfd.m_pFont->getToUnicode();

	for (CharCode c = 0; c < 256; ++c)
	{
		Object charProc = ((Gfx8BitFont *)mfd.m_pFont)->getCharProc(c);

		if (charProc.isStream())
		{
			ULONG ulFlags = mfd.isBold ? FG_BOLD : 0;
			ulFlags |= mfd.isItalic ? FG_ITALIC : 0;

			CXfFontGlyph* pGlyph = new CXfFontGlyph(ulFlags, 0, 72000, mfd.ascent, mfd.descent, mfd.emWidth);

			Type3GlyphVectorizer vectorizer(pGlyph->Points, pGlyph->Verbs);
			{
				Gfx gfx(m_pDoc, &vectorizer, ((Gfx8BitFont *)mfd.m_pFont)->getResources(), &PDFRectangle(0, 0, 0, 0), NULL, Type3GlyphVectorizer::abortCheckCbk, &vectorizer);
				if (fontMatrix)
					gfx.getState()->concatCTM(fontMatrix[0], fontMatrix[1], fontMatrix[2], fontMatrix[3], fontMatrix[4], fontMatrix[5]);
				gfx.getState()->concatCTM(72000 / fabs(mfd.scale), 0, 0, 72000 / mfd.scale, 0, 0);
				gfx.display(&charProc, false);
			}

			if (vectorizer.m_bStatus)
			{
				Rect bounds;
				if (!pGlyph->Points.empty())
				{
					std::vector<DocCoord>::const_iterator it = pGlyph->Points.begin();
					bounds.x = it->x;
					bounds.y = it->y;
					for( ++it; it != pGlyph->Points.end(); ++it)
						Rect::InflateRectToContainPoint(bounds, it->x, it->y);
				}
				pGlyph->SetBounds(bounds);

				pGlyph->Width = _round(vectorizer.m_advanceWidth * fontMatrix[0] * 72000.0 / fabs(mfd.scale));

				Unicode *u = NULL;
				int uLen = ctu ? ctu->mapToUnicode(c, &u) : 0;

				AdjustUnicode(u, uLen);

				Unicode uBuf[8];
				if (int len = parseCharName(mfd.m_pFont, c, uBuf, sizeof(uBuf) / sizeof(uBuf[0])))
				{
					u = uBuf;
					uLen = len;
				}

				Unicode uLigature;
				if (uLen > 1)
				{
					uLigature = MapLigature(u, uLen);
					if (0 == uLigature)
						uLigature = MapNonStandardLigature(mfd, u, uLen);

					uLen = 1;
					u = &uLigature;
				}
				else if(0 == uLen)
				{
					uLen = 1;
					uLigature = 0;
					u = &uLigature;
				}

				pGlyph->CharCode = MapCharacter(mfd, *u, c, pGlyph->Points.empty());							

				bool bDeleteGlyph = true;
				if (CXfFontGlyph* pMapGlyph = pFontDef->GetGlyph(pGlyph->CharCode, mfd.isBold, mfd.isItalic))
				{
					// Ok, the last resort: probably a wrong Unicode map. Or just a regular collision from a different subset
					// We need to create another font def object and store the collided glyphs there: JonPeddieResearch.pdf
					// has the actual different glyphs for the same font name UEUPZH+SabonLTStd-Roman, fontRef=473, 294.
					// Page 22, 'Ben Obi-Wan Kenobi' all caps got lost as a result.
					if(CompareGlyphs(pGlyph, pMapGlyph) == false)
						mfd.extraNode[c] = FindOrCreateExtraFontDef(mfd, pGlyph, &bDeleteGlyph);
				}
				else // AddGlyph
				{
					pFontDef->AddGlyph(pGlyph->CharCode, pGlyph, true); // add glyph and add it to xar tree also
					pFontDef->UpdateEmbeddingPolicy(pGlyph->Flags, mfd.m_fsType);
					bDeleteGlyph = false;
					pGlyph->SetObjectName(mfd.fontNameRaw + FormatStringW(L"+%d", mfd.m_pFont->getID()->num));
				}

				if (bDeleteGlyph)
					delete pGlyph;
			}
			else
				delete pGlyph;
		}
	}

	/*
	if (ctu)
		ctu->decRefCnt();
	*/

	// We are embedded, and font is not present in the system, set the whole font def as embedded
	// Paul: now we embed everything and test each glyph against the system one, detach the font if it has completely same
	// glyphs as the system font
	mfd.node->SetEmbedded(mfd.node->IsEmbedded() || mfd.m_bEmbedded ? true : false);

	// define a space, find a font substitute if necessary
	AssignSpaceCharacters(&mfd);
}

/******************************************************************************
   Function:    FontManager::ProcessFontGlyphs
   Author:      Paul Gruzdev
   Purpose:		Given a font definition record and font info record and a 
				font description, scan the particular font for glyphs and 
				embed them (if necessary) into Xar stream. Note: currently,
				if font of a given name and weight/italic exists in the system,
				we don't pass the glyphs to Xara as they will be ignored by
				Camelot anyway.
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool FontManager::ProcessFontGlyphs(CXfFontDef* pFontDef, MappedFontDesc& mfd, bool embed, bool bSubset, bool bAlterFontDef)
{
	//USES_CONVERSION;
	pFontDef->SetExternal(!pFontDef->IsEmbedded());

	// 20tipkid.pdf: we help Camelot to pick a correct variant of the font - not the one defined in the same doc but with different
	// weight (bold) applied
	if (bAlterFontDef)
		pFontDef->weight = mfd.weight;

	if (!mfd.m_bEmbedded)
	{
		// Font does not exist in the system and we have no glyphs - map the font name from Mac to Windows
		if (!mfd.bPresentInTheSystem)
		{
			pFontDef->TypeFaceName = CFontMapper::MapMacFontName(mfd.fontName);
			pFontDef->FullFontName = pFontDef->TypeFaceName;
			wprintf(L"Mac font change: %s ---> %s\n", mfd.fontName.c_str(), pFontDef->TypeFaceName.get());
		}

		return false;
	}

	// Now we have a PostScript font name ----> Windows font name map of all the fonts found in the system. If the current
	// font exists, we won't pass the glyphs as they will be ignored by Camelot anyway.
	// An update: if Regular variant of some particular font is present in the system, but Italic is not, we can't embed just
	// italic glyphs as Camelot starts to render regular glyphs using Times... so again - we need to embed all glyphs
	// cause we don't know yet about the upcoming variant of the same font
	if (!embed || !pFontDef->IsEmbedded())
		return true;

	if (!mfd.bPresentInTheSystem)
	{
		BYTE byPanose[10];
		ConstructPANOSE(byPanose, mfd.m_pFont, mfd.m_fontObj.getDict());
		MergePANOSE(pFontDef->PANOSE, byPanose);
		//memcpy(pFontDef->PANOSE, (PBYTE)"\x2\xB\x6\x4\x3\x5\x4\xB\x2\x4", 10);
		//MergePANOSE(pFontDef->PANOSE, (PBYTE)"\x2\x0\x5\x3\x3\x0\x0\x2\x0\x3");
		BYTE byAltPANOSE[10]; memset(byAltPANOSE, 0, 10);
		BYTE byClassID = 0, bySubclassID = 0;
		if (ExtractPANOSEandIBMInfo(mfd.m_fontObj.getDict(), byClassID, bySubclassID, byAltPANOSE))
			MergePANOSE(pFontDef->PANOSE, byAltPANOSE);
	}
	else
	{
		if(pFontDef->PANOSE[0] == 0 || !mfd.isBold && !mfd.isItalic) // fill-in only if the PANOSE is not initialized yet
			GetSystemPANOSE(mfd.fontName.c_str(), mfd.weight, mfd.isItalic, pFontDef->PANOSE);
	}

	bool bItalic = mfd.m_pFont->isItalic() ? true : false;
	//bool bBold = pFont->isBold() ? true : false;
	bool bBold = mfd.isBold;

	// TODO: we have to subst our weight taken from font name with a real one from /FontWeight 
	// through GfxFont, if it is defined!!
	// Here it goes:
	// Paul: an update - we *have* to operate the same way as we do when rendering text, so we need to disable the weight 
	// calculation - or add this weight taken from gfxfont there
	/*	if(pFont->getWeight() != GfxFont::WeightNotDefined)
	{
	pFontDef->weight = ((int)pFont->getWeight() - (int)GfxFont::W100 + 1)*100;
	mfd.weight = pFontDef->weight;
	}
	else*/
	// We need to set the font weight and a 'multiweight' flag
	if (bAlterFontDef)
		pFontDef->weight = mfd.weight;

	//	bBold |= pFontDef->weight >= 700; // The same method of defining bold as in rendering glyphs - take the flag from the font AND a weight component
	bItalic |= mfd.isItalic;
	mfd.isItalic = bItalic;

	GfxFontType type = mfd.m_pFont->getType();

	SplashFTFontFile_* pFTFile = (SplashFTFontFile_*)mfd.m_pSplashFont->getFontFile();
/*
{
	FILE* out = fopen("d:\\temp\\charmap.txt", "at");
	fwprintf(out, L"font: %s ------------------------------------------------\n\n", strName.c_str());
	CharCodeToUnicode* pc = pFont->getToUnicode();
	int ccc=5;
	CharCode icnt = pc->getLength();
	for(int i = 0; i < icnt; i++)
	{
		Unicode* u;
		if(pc->mapToUnicode(i, &u))
			fprintf(out, "%d[%X] unicode: %X, gid: %d\n", i, i, *u, pFTFile->codeToGID[i]);
		else
			fprintf(out, "%d[%X] unicode: no code. gid: %d\n", i, i, pFTFile->codeToGID[i]);
	}
	pc->decRefCnt();
	fclose(out);
}
*/

	FT_Face face = pFTFile->face;
	FT_UShort uEMWidth = face->units_per_EM;

	_ASSERT(mfd.ascent >= 0 && mfd.descent <= 0);
	// Now as we take ascent/descent from the /FontDescriptor dictionary, not from the font file, there is no need to extract
	// the em width and scale by it
	if (mfd.emWidth < 0) // This now is not used - em extraction can be removed
		mfd.emWidth = MulDiv(face->units_per_EM, 72000, face->height);

	FT_CharMap* charmaps = face->charmaps;
	int iNumMaps = face->num_charmaps;

	bool bFontDiffersFromSystem = false;

	if (true)//mfd.m_charSet.empty())
	{
		int ng = face->num_glyphs;
		if (ng <= 0 || ng > 65535)
			return false; // bail

		std::map<int, int> mGlyphsToCharcode;
		std::map<int, int> mGlyphsToUnicode;

		// First try to enum the font glyphs using it's codeToGID table
		bool bResFind = pFTFile->FindGlyphsByCodeToGid(mGlyphsToCharcode, mGlyphsToUnicode, iNumMaps ? charmaps[0]->encoding : FT_ENCODING_NONE);

		// If codeToGID enum failed, try to use FreeType's glyph enumeration
		// An update: FreeType's glyph enumeration is unreliable - it misses many glyphs out there. So just scan through all of them
		// instead.
		if (0)//!bResFind && charmaps)
		{
			std::wcout << mfd.fontName << L": Enumerating glyphs via FT_Get_First_Char" << std::endl;
			for (int i = 0; i < iNumMaps; i++)
			{
				int rCode = FT_Set_Charmap(face, charmaps[i]);

				// the count will be less 
				int cc = 0;
				FT_UInt leftGlyphIndex;
				FT_ULong leftCharCode = FT_Get_First_Char(face, &leftGlyphIndex);

				while (leftGlyphIndex)
				{
					mGlyphsToCharcode.insert(std::make_pair(leftGlyphIndex, leftCharCode));
					mGlyphsToUnicode.insert(std::make_pair(leftGlyphIndex, GetFTCharacterUnicode(leftCharCode,charmaps[i]->encoding)));
					++cc;
					leftCharCode = FT_Get_Next_Char(face, leftCharCode, &leftGlyphIndex);
					if (cc >= ng)
						break; // exceeded claimed glyph count!
				}
			}
		}
		else if (!bResFind)// && !charmaps)
		{
			std::wcout << mfd.fontName << L": Enumerating glyphs by scanning all IDs" << std::endl;
			for (int charcode = 1; charcode <= ng; charcode++)
			{
				mGlyphsToCharcode[charcode] = charcode;
				mGlyphsToUnicode[charcode] = charcode;
			}
		}
/*
		else
		{
			int i = 0;
			for(std::set<int>::iterator it = mfd.unicodeSubset.begin(); it != mfd.unicodeSubset.end(); ++it)
			{
				mGlyphsToCharcode[i] = *it;
				mGlyphsToUnicode[i++] = *it;
			}
		}
*/

		std::map<int, int>::iterator itU = mGlyphsToUnicode.begin();
		std::map<int, int>::iterator it  = mGlyphsToCharcode.begin();
		for (; itU != mGlyphsToUnicode.end(); ++it, ++itU)
		{
			// Paul: I'm not sure about the font-type-based pick, let's try to pick based whether we have code --> GID map
			//int iRealCode = type >= fontCIDType2 ? iGlyphIdx : iCharcode;
			CharCode charCode = pFTFile->codeToGID ? it->second : it->first;

			EmbedGlyph(mfd, charCode, itU->second, bSubset, false, true, bFontDiffersFromSystem);
		}
	}
	else
	{
		for (auto it = mfd.m_charSet.begin(); it != mfd.m_charSet.end(); ++it)
			EmbedGlyph(mfd, *it, GetFTCharacterUnicode(*it, iNumMaps ? charmaps[0]->encoding : FT_ENCODING_NONE), bSubset, false, true, bFontDiffersFromSystem);
	}

	if (bFontDiffersFromSystem)
	{
		// The whole mfd is marked as '-emb'.
		mfd.bPresentInTheSystem = false;
		// Now do the bold flag back thing: if the [this bold variant] font IS present in the system, then we forcibly need to set-up it's bold
		// flag because Camelot takes it in this way. But: now the font is not present anymore, we alter it's name, so - we get rid
		// of the bold flag. Note: this function ProcessFontGlyphs can't be used after text processing (subset flag), because
		// then we would need to process all text strings that use this particular font and clear a bold flag to them as well.
		if (bAlterFontDef && mfd.isBold)
		{
			mfd.isBold = false;
			std::vector<CXfFontGlyph*> glyphs;
			pFontDef->GetGlyphs(glyphs, bBold, bItalic);
			for(size_t i = 0; i < glyphs.size(); i++)
				glyphs[i]->setBold(false);
		}
	}

	return true;
}

/******************************************************************************
   Function:	FontManager::FindGlyph
   Author:		Paul Gruzdev
   Purpose:     Find the glyph in the main font def provided, and in all it's
				-emb variants. This function is used to reliably find a SPACE
				character of the current font typeface
   Returns:		CXfFontGlyph*
   Parameters:	wchar_t wUnicode  
   Parameters:	CXfFontDef * pFontDef  
   Parameters:	bool bBold  
   Parameters:	bool bItalic  
   Notes:       
******************************************************************************/
CXfFontGlyph* FontManager::FindGlyph(wchar_t wUnicode, CXfFontDef* pFontDef, bool bBold, bool bItalic)
{
	// First search the glyph in the main font def (we put glyphs there only if it's not present in the system).
	if (CXfFontGlyph *pGlyph = pFontDef->GetGlyph(wUnicode, bBold, bItalic))
		return pGlyph;

	// Now search the extra defs
	// Since we now store our glyphs in extra definitions, not in the maim pFontDef (and then switch),
	// find our space character in the extra defs list
	FontDefList& list = extraDefs[pFontDef];
	for (FontDefList::const_iterator it = list.begin(); it != list.end(); ++it)
	{
		if (CXfFontGlyph *pGlyph = (*it)->GetGlyph(wUnicode, bBold, bItalic))
			return pGlyph;
	}

	return nullptr;
}

/******************************************************************************
   Function:	FontManager::EmbedGlyph
   Author:		Paul Gruzdev
   Purpose:     Given current loaded into XarOutputDev::font, embed Xara glyph.
   Returns:		bool
   Parameters:	CXfFontDef * pFontDef  
   Parameters:	MappedFontDesc & mfd  
   Parameters:	int iRealCode  
   Parameters:	int iUnicode  
   Parameters:	bool bSubset  
   Parameters:  bool bNeedGlyph - indicates whether we are enumerating or we are actually drawing text. Set to false if
				you are enumerating all *possible* glyphs in the font, but set it to true if you 100% sure the glyph exists
				in the font
   Parameters:	bool & bFontDiffersFromSystem  
   Notes:       
******************************************************************************/
bool FontManager::EmbedGlyph(MappedFontDesc& mfd, CharCode charCode, int iUnicode, bool bSubset, bool bNeedGlyph, bool bCanChangeNode, bool& bFontDiffersFromSystem)
{
	CXfFontDef* pFontDef = mfd.node; // the font def we are adding our glyph at.
	const bool bBold = mfd.isBold;
	const bool bItalic = mfd.isItalic;

	bool bNullUnicodeMapping = false;
	//bool bNonStandardLigature = false;

	std::wstring ligChars;

	if (CharCodeToUnicode *pCTU = (CharCodeToUnicode *)mfd.m_pFont->getToUnicode())
	{
		Unicode* u;
		int uLen = pCTU->mapToUnicode(charCode, &u);

		AdjustUnicode(u, uLen);

		Unicode uBuf[8];
		if (int len = parseCharName(mfd.m_pFont, charCode, uBuf, sizeof(uBuf) / sizeof(uBuf[0])))
		{
			u = uBuf;
			uLen = len;
		}

		if (uLen && u)
		{
			if (uLen > 1)
			{

				//if (uLen <= LIGATURE_MAX_CHARS)
				//{
					ligChars.resize(uLen);
					for (int i = 0; i < uLen; ++i)
						ligChars[i] = u[i];
				//}					

				Unicode uLig = MapLigature(u, uLen);
				if (uLig) iUnicode = uLig;
				// Ok, we've found a non-standard ligature, we can't just drop it, we need to embed it also
				else
				{
					iUnicode = MapNonStandardLigature(mfd, u, uLen);
					// Since we map the ligature to a Unicode private use area, now we need to embed the whole font;
					//bNonStandardLigature = true;
				}
			}
			else
				iUnicode = *u;
		}
		else
		{
			bNullUnicodeMapping = true;
#ifdef _DEBUG
			std::wcout << L"Null Unicode mapping of char 0x" << std::hex << iUnicode << L", typeface: " << mfd.fontName << std::endl;
			#endif
			//wprintf(L"Null Unicode mapping of char %c, typeface: %s\n", iUnicode, strName.c_str());
		}
		//pCTU->decRefCnt();
	}

	// Don't embed a non-existing glyph if we pass all glyphs to Camelot, not a subset
	if(!bSubset && bNullUnicodeMapping && !bNeedGlyph)
		return false;

	// we strip to two bytes here, AND when adding actual text
	// Leave this unicode char out if it is not used in the document
	// DDES-1530: we also need to include all Latin1 plain in order to
	// make documents more editable in Camelot
// Now we never get here
/*
	if(wUnicode < 32)
	{
		_ASSERT(bNullUnicodeMapping == true); // We have missed the glyph as it's Unicode value is still < 32, probably need to make glyph relocation
		return false;
	}

	if(bSubset && !mfd.CharacterExists(wUnicode) && (wUnicode < 32 || wUnicode > 127 || bNullUnicodeMapping))
		return false;
*/
	if (!mfd.m_pSplashFont)
		return false;
	SplashPath* pPath = mfd.m_pSplashFont->getGlyphPath(charCode);
	double dAdvance = mfd.m_pSplashFont->getGlyphAdvance(charCode);

	// DDES-888: Very strange output of pCTU->mapToUnicode could be - characters with more than 2bytes like 0x00022DE5,
	// lscape.pdf, need to explicitly convert to wchar_t, dropping unwanted bytes
	wchar_t wUnicode = iUnicode;
	wUnicode = MapCharacter(mfd, wUnicode, charCode, IsPathEmpty(pPath));

	// credit card numbers.pdf - skip empty glyphs that are not used in the document
	if (pPath && IsPathEmpty(pPath) && ((bSubset && !mfd.CharacterExists(wUnicode)) || bNullUnicodeMapping) && !bNeedGlyph)
	{
		delete pPath;
		return false;
	}

	// DDES-882 fix - libertine.pdf - an artificial glyph has a zero advance!!! so we can't ignore that
	if (pPath /*&& dAdvance > 0*/)
	{
		// Paul: DDES-284 fix. It appeared that various glyphs could have empty paths and we still need to pass them
		// for the correct metrics
		//if(pPath->getLength() || iUnicode == 32)
		{
			ULONG ulFlags = bBold ? FG_BOLD : 0;
			ulFlags |= bItalic ? FG_ITALIC : 0;
			CXfFontGlyph* pGlyph = new CXfFontGlyph(ulFlags, wUnicode, (ULONG)(dAdvance * 72000 + 0.5), mfd.ascent, mfd.descent, mfd.emWidth);
			SplashPathToGlyph(pPath, pGlyph, 1000.0);
			delete pPath;

			bool bGlyphAdded = false, bDeleteGlyph = true;
			/*if (bNonStandardLigature)
			{




				//wprintf(L"extra: nslig: %X(%X) %s%s%s ref:%d\n", charCode, wUnicode, mfd.fontName.c_str(), bBold ? L"-Bold" : L"", bItalic ? L"-Italic" : L"", mfd.m_pFont->getID()->num);
				mfd.extraNode[charCode] = FindOrCreateExtraFontDef(mfd, pGlyph, &bDeleteGlyph);
				bGlyphAdded = true;
			}
			else */if (CXfFontGlyph* pMapGlyph = pFontDef->GetGlyph(wUnicode, bBold, bItalic))
			{
				// Ok, the last resort: probably a wrong Unicode map. Or just a regular collision from a different subset
				// We need to create another font def object and store the collided glyphs there: JonPeddieResearch.pdf
				// has the actual different glyphs for the same font name UEUPZH+SabonLTStd-Roman, fontRef=473, 294.
				// Page 22, 'Ben Obi-Wan Kenobi' all caps got lost as a result.
				if(CompareGlyphs(pGlyph, pMapGlyph) == false)
				{
					wprintf(L"[%s]-->[%s] w: %d, it: %s, glyph collision: %X, exists in [%s] ", mfd.fontNameRaw.c_str(), mfd.fontName.c_str(), mfd.weight, mfd.isItalic ? L"true" : L"false", wUnicode, pMapGlyph->GetObjectName().c_str());

					//wprintf(L"extra: glyph collision: %X(%X) %s%s%s ref:%d\n", charCode, wUnicode, mfd.fontName.c_str(), bBold ? L"-Bold" : L"", bItalic ? L"-Italic" : L"", mfd.m_pFont->getID()->num);
					mfd.extraNode[charCode] = FindOrCreateExtraFontDef(mfd, pGlyph, &bDeleteGlyph);
					bGlyphAdded = true;
				}
			}
			else // AddGlyph
			{
				if(mfd.bPresentInTheSystem && !bFontDiffersFromSystem)
				{
#ifdef FILTER_DEBUG
					CHiResolutionTimer tm;
#endif
					bool bNewBold = mfd.isBold, bNewItalic = mfd.isItalic;
					CGlyphDiff diff(pGlyph->CharCode);
					bool bEqual = true;

					if (mfd.m_charSet.empty() || mfd.m_charSet.end() != mfd.m_charSet.find(charCode) )
						bEqual = CompareGlyphWithSystem(mfd, pGlyph, diff, ligChars);
#ifdef FILTER_DEBUG
					g_dCompareGlyphTime += tm.Current();
#endif
					// If the font is present in the system, add to an extra def '-emb' anyway.
					CXfFontDef* pExtraDef = FindOrCreateExtraFontDef(mfd, pGlyph, &bDeleteGlyph);
					bGlyphAdded = true;

					if(!bEqual)
					{
						//wprintf(L"extra: glyph diff: %X(%X) %s%s%s ref:%d\n", charCode, wUnicode, mfd.fontName.c_str(), bBold ? L"-Bold" : L"", bItalic ? L"-Italic" : L"", mfd.m_pFont->getID()->num);
						if(diff.GlyphNotPresent() || !bCanChangeNode) // relocate just this glyph to '-emb', rare case. 
						{
							if(!bCanChangeNode)
								wprintf(L"%s: re-embedded %X moved to '-emb' alone.\n", mfd.fontNameRaw.c_str(), pGlyph->CharCode);
							// Or we missed this glyph when scanning fonts, and found it when importing text. 
							// Also move only this glyph to '-emb' variant as we can't change the entire system font to embedded - 
							// we already rendered previous text using this font's metrics (system glyph widths).
							mfd.extraNode[charCode] = pExtraDef;
						}
						else // outline differ or width mismatch: need to move whole mfd to an '-emb' font def
						{
							FontDefList& vec = extraDefs[pFontDef];
							if(mfd.node != vec.front())
							{
								bFontDiffersFromSystem = true;
								mfd.node = vec.front();
								pFontDef = mfd.node;
							}

							if(pExtraDef != mfd.node)
								mfd.extraNode[charCode] = pExtraDef;
						}
					}
					else
					{
						// Add our bbox to the bbox map - as it's equal to the system glyph's bbox.
						if(mfd.bboxMap.find(wUnicode) == mfd.bboxMap.end())
							mfd.bboxMap[wUnicode] = pGlyph->Bounds();
						// New rule: since we move all glyphs that differ from system to extra font defs, there is no need
						// to add a glyph that is equal to system to the regular font def
						bGlyphAdded = true;
						//delete pGlyph; pGlyph = NULL;
					}
				}

				if (!bGlyphAdded)
				{
					pFontDef->AddGlyph(wUnicode, pGlyph, true); // add glyph and add it to xar tree also
					pFontDef->UpdateEmbeddingPolicy(pGlyph->Flags, mfd.m_fsType);
					bDeleteGlyph = false;
				}
				if (pGlyph)
					pGlyph->SetObjectName(mfd.fontNameRaw+FormatStringW(L"+%d", mfd.m_pFont->getID()->num));
			}

			if (bDeleteGlyph)
				delete pGlyph;
		}
	}
	return true;
}

/******************************************************************************
   Function:	FontManager::InitMappedFontDesc
   Author:		Paul Gruzdev
   Purpose:     Split from CreateNewFont(): this creates an mfd out of the FontInfo,
				especially forming the font name that will be represented by Camelot.
				Given that PDF font names look like this: [AADFA]-ArialBoldItalicMT,
				we need to present a more sane font name, also matching it vs system
				fonts, deciding about a bold flag (if matched vs the system), mapping Mac --> Win
				name, if not embedded and not present in the system, and doint Adobe base14
				font mapping. So as a result, we get a font name and bold/italic flags ready
				to be passed to Camelot.
   Returns:		void
   Parameters:	[in] - MappedFontDesc & fontDesc  
   Notes:       
******************************************************************************/
void FontManager::InitMappedFontDesc(MappedFontDesc &fontDesc)
{
	fontDesc.pFontManager = this;

	std::wstring name, nameWithPrefix;

	bool bBuiltInFont = false;

	fontDesc.ascent = _round(fontDesc.m_pFont->getAscent() * 72000.);
	fontDesc.descent = _round(fontDesc.m_pFont->getDescent() * 72000.);

	std::string psName;
	if (fontDesc.m_pFont->getName())
		psName = fontDesc.m_pFont->getName()->c_str();

	// DDES-11074: now we have a list of fonts distributed with the pdf2xar. If the font is not embedded, but is in that
	// list, treat it like embedded one and take glyphs from it
	if (mBuiltInFonts.find(psName) != mBuiltInFonts.end())
		bBuiltInFont = true;

	// DDES-15046 (Bold text in pdf not being imported as Bold)
	if (fontDesc.m_pSplashFont)
	{
		if (SplashFTFontFile_* pFTFile = (SplashFTFontFile_*)fontDesc.m_pSplashFont->getFontFile())
		{
			FT_Face face = pFTFile->face;
			const char *ps_name = FT_Get_Postscript_Name(face);
			if (ps_name && ps_name[0])
				psName = ps_name;

			fontDesc.m_fsType = FT_Get_FSType_Flags(face);

			fontDesc.ascent  = MulDiv(face->ascender,  72000, face->units_per_EM);
			fontDesc.descent = MulDiv(face->descender, 72000, face->units_per_EM);

			if (fontDesc.ascent - fontDesc.descent > 96000)
			{
				fontDesc.ascent  = MulDiv(face->ascender,  72000, face->height);
				fontDesc.descent = MulDiv(face->descender, 72000, face->height);
			}
		}
	}

	if (psName.empty())
	{
		nameWithPrefix = name = L"Times New Roman";
//#ifdef _DEBUG
//		nameWithPrefix = name = L"UNKNOWN";
//#endif

		fontDesc.fontName = fontDesc.fontNamePS = fontDesc.fontNameRaw = nameWithPrefix;
	}
	else
	{
		nameWithPrefix = name = Utf8ToUtf16(psName);
		if (nameWithPrefix.length() > 7 && nameWithPrefix.at(6) == '+')
			nameWithPrefix.erase(0, 7);

		fontDesc.fontNameRaw = name;
		fontDesc.fontNamePS = nameWithPrefix;

		//fontMapper.ApproximateFontFace(nameWithPrefix, fontDesc.fontName, fontDesc.weight, fontDesc.isItalic);
		bool bRes = fontMapper.ApproximateFontFaceStrict(nameWithPrefix, fontDesc.fontName, fontDesc.weight, fontDesc.isBold, fontDesc.isItalic);

		if (g_settings.bRawFontNames)
		{
			fontDesc.fontName = fontDesc.fontNameRaw;
			fontDesc.fontNamePS = fontDesc.fontNameRaw;
			fontDesc.isBold = false;
		}

		fontDesc.bPresentInTheSystem = FontExists(fontDesc);

		// DDES-11074: if we have a built in font (comes along with pdf2xar) that is not present in the system, change it name
		// if required, so that for example, 'Zapf Dingbats' becomes 'Dingbats' due to copyright issues.
		if (!fontDesc.bPresentInTheSystem && bBuiltInFont)
		{
			fontDesc.fontName = mBuiltInFonts[psName];
			// And check the system again
			fontDesc.bPresentInTheSystem = FontExists(fontDesc);
		}

		if (fontDesc.bPresentInTheSystem)
		{
			// DDES-1654 fix: we need not only fix-up the font name if it's present in the system, but
			// also it's bold flag given weight, as Camelot does not understand multi-weight font records
			// for installed fonts, it needs a bold flag set for actual glyphs
			fontDesc.fontName = fontMapper.FixUpFontProperties(fontDesc.fontName, fontDesc.weight, fontDesc.isItalic, fontDesc.isBold);
		}

		// No mapping - try to subst Mac ---> Windows and try again
		if (!fontDesc.bPresentInTheSystem && !fontDesc.m_bEmbedded)
		{
			std::wstring strNew = CFontMapper::MapMacFontName(fontDesc.fontName);
			if (strNew != fontDesc.fontName)
			{
				fontDesc.fontName = strNew;
				fontMapper.FindFontBoldFlag(strNew, fontDesc.weight, fontDesc.isBold, fontDesc.isItalic);
				// Re-check whether the font exists in the system as we changed it's name
				fontDesc.bPresentInTheSystem = FontExists(fontDesc);
			}
		}

		// Now - if the font is embedded, let's see whether it's one of the common Mac Base14 fonts and we have all the
		// similar glyphs in the system, in this case - change the font name to Windows equivalent and don't embed any glyphs
		if (!g_settings.bNoBase14Substitution && fontDesc.m_bEmbedded)
		{
			std::wstring strWinName = CFontMapper::MapBase14FontName(fontDesc.fontName);

			if (strWinName != fontDesc.fontName)
			{
				MappedFontDesc mfd1(this);
				mfd1.fontName = strWinName;
				mfd1.weight = fontDesc.weight; mfd1.isItalic = fontDesc.isItalic;
				if (FontExists(mfd1))
				{
					// Obviously we can't check if all glyphs used in the document are present in the system font here: 
					// as we haven't gathered a font subset yet. We have to create a font here on the other hand, because of 
					// the 'bold' flag to be decided here - Camelot discrepancy issue
					//bool bOk = CheckGlyphs(strWinName, mfd.weight, mfd.isItalic, mfd.unicodeSubset);
					/*
					#ifdef FILTER_DEBUG
					FILE* out = fopen("GDIFontLog.txt", "at");
					fwprintf(out, L"%s: ****Mapped %s ---> %s\n", filename.c_str(), fontDesc.fontName.c_str(), strWinName.c_str());
					fclose(out);
					#endif*/
					fontDesc.fontName = strWinName;
					// DDES-10382. Now we still don't know yet, whether we keep the font as embedded or not. So we can't
					// drop that flag here obviously, only after outline glyphs check
					//fontDesc.m_bEmbedded = false;
					fontDesc.bPresentInTheSystem = true;
					fontMapper.FindFontBoldFlag(strWinName, fontDesc.weight, fontDesc.isBold, fontDesc.isItalic);
				}
			}
		}
		LOGW_FONT(L" nameSrc=" << nameWithPrefix << L" name=" << fontDesc.fontName << L" w=" << fontDesc.weight << (fontDesc.isItalic ? L'I' : L' '));

		/*
		static int is=0;
		wchar_t psz[128];
		swprintf(psz, L"pdf%d_", ++is);
		fontDesc.fontName = std::wstring(psz) + fontDesc.fontName;*/
	}
}

/******************************************************************************
Function:    FontManager::CreateNewFontEx
Author:      Paul Gruzdev
Purpose:	Given a PDF font info struct, search a font variant in our Xara font
			map, possibly creating a new CXfFontDef or adding to an existing
			one. We add to an existing one if we have the same font name and
			the same weight already created before. 
Returns:
Exceptions:
Parameters:
Notes:		That's the new alternative way, this fn is 
			called ONLY for fonts that were not enumerated at the start
			when creating the map. For example for the fonts, defined in a strange
			place like annotation XObject that are invisible for the Poppler's font
			scanner.
******************************************************************************/
/*CXfFontDef* FontManager::CreateNewFontEx(MappedFontDesc& mappedFontDesc, bool bEmbedGlyphs)
{
	CXfFontDef* newFont = NULL;

	if (mfdMap.CheckWeight(mappedFontDesc.fontName, mappedFontDesc.weight))
	{
		newFont = mfdMap.GetFontDef(mappedFontDesc.fontName, mappedFontDesc.weight);
		mappedFontDesc.emWidth = -1;
		mappedFontDesc.node = newFont;
		mappedFontDesc.pFontDefOrg = newFont;
	}
	else
	{
		newFont = new CXfFontDef(mappedFontDesc.fontName, mappedFontDesc.fontName);

		mappedFontDesc.emWidth = -1;
		mappedFontDesc.node = newFont;
		mappedFontDesc.pFontDefOrg = newFont;

		// Here we probably can enumerate and embed all glyphs
		if (bEmbedGlyphs)
		{
			std::map<CXfFontDef*, bool> fontDefPresent;
			newFont->SetRemoved(false); // In case we removed it earlier
			ProcessFontGlyphs(newFont, mappedFontDesc, g_settings.bLoadEmbeddedGlyphs, false, false);

			if (mappedFontDesc.bPresentInTheSystem && fontDefPresent.find(newFont) == fontDefPresent.end())
				fontDefPresent[newFont] = true;
			else if (!mappedFontDesc.bPresentInTheSystem && fontDefPresent.find(newFont) != fontDefPresent.end())
				fontDefPresent[newFont] = false;

			for (std::map<CXfFontDef*, bool>::iterator it = fontDefPresent.begin(); it != fontDefPresent.end(); ++it)
			{
				if (it->second)
				{
					CXfFontDef* pFontDef = it->first;
					pFontDef->MarkGlyphsRemoved(false, false);
					pFontDef->MarkGlyphsRemoved(false, true);
					pFontDef->MarkGlyphsRemoved(true, false);
					pFontDef->MarkGlyphsRemoved(true, true);
				}
			}
		}

		// Add our font prior any other fonts. Only reliable place, or we need to scan through children and insert before
		// any spread node
		m_pChapter->InsertFirstChild(newFont, false);
	}

	// We are embedded, and font is not present in the system, set the whole font def as embedded
	// Paul: now we embed everything and test each glyph against the system one, detach the font if it has completely same
	// glyphs as the system font
	newFont->SetEmbedded(newFont->IsEmbedded() || mappedFontDesc.m_bEmbedded ? true : false);
	// define a space, find a font substitute if necessary
	AssignSpaceCharacters(&mappedFontDesc);

	return newFont;
}*/

/******************************************************************************
   Function:	FontManager::AssignSpaceCharacter
   Author:		Paul Gruzdev
   Purpose:     Given a (newly created) MappedFontDesc, find and assign a corresponding
				SPACE charater to it. It's better to first create
				ALL mfd's for all fonts in the document, and then run through them and
				assign spaces, because the SPACE glyphs will be shared between them.
   Returns:		void
   Parameters:	MappedFontDesc * pmfd  
   Notes:       
******************************************************************************/
void FontManager::AssignSpaceCharacter(MappedFontDesc* pmfd, wchar_t ch)
{
	static const double GLYPH_UNIT_PER_EM_DEFAULT = 72000;
	static const double GLYPH_ADVANCE_WHITESPACE_DEFAULT = .25;

	GlyphMap& map = pmfd->node->GetGlyphMap(pmfd->isBold, pmfd->isItalic);

	// The font is not embedded (or has no glyphs), nor present in the system. Camelot will use a substitute font for it: let's
	// pick the substitute just like Camelot does it
	if (map.empty() && !pmfd->bPresentInTheSystem)
		pmfd->GetSubstitute(pmfd->node, pmfd->isBold, pmfd->isItalic);

	if (!pmfd->bPresentInTheSystem)
	{	
		// Need to search for extra defs also?
		CXfFontGlyph* pGlyph = FindGlyph(ch, pmfd->pFontDefOrg, pmfd->isBold, pmfd->isItalic);

		if(pGlyph == NULL && !map.empty()) // no space in font def: need to create and add one
		{
			LONG dwAdvance = LONG(GLYPH_UNIT_PER_EM_DEFAULT * GLYPH_ADVANCE_WHITESPACE_DEFAULT);

			// This variant of our font has glyphs but does not have a space char. First of all - find it
			// in the neighbourhood 
			pGlyph = pmfd->node->GetGlyph(ch, pmfd->isBold, !pmfd->isItalic);
			if (pGlyph)
				dwAdvance = pGlyph->Width;
			else
			{
				pGlyph = pmfd->node->GetGlyph(ch, !pmfd->isBold, pmfd->isItalic);
				if(pGlyph)
					dwAdvance = pGlyph->Width;
				else
				{
					pGlyph = pmfd->node->GetGlyph(ch, !pmfd->isBold, !pmfd->isItalic);
					if(pGlyph)
						dwAdvance = pGlyph->Width;
				}
			}

			ULONG ulFlags = pmfd->isBold ? FG_BOLD : 0;
			ulFlags |= pmfd->isItalic ? FG_ITALIC : 0;
			pGlyph = new CXfFontGlyph(
				ulFlags, 
				ch, 
				dwAdvance,
				map.begin()->second->FontAscent.get(), 
				map.begin()->second->FontDescent.get(),
				map.begin()->second->FontEMWidth.get()
				);
			InitEmptyGlyph(pGlyph);
			pmfd->node->AddGlyph(ch, pGlyph, true); // add glyph and add it to xar tree also
			pmfd->node->UpdateEmbeddingPolicy(pGlyph->Flags, pmfd->m_fsType);
		}
	}
}

void FontManager::AssignSpaceCharacters(MappedFontDesc* pmfd)
{
	AssignSpaceCharacter(pmfd, L' '); // SPACE
	AssignSpaceCharacter(pmfd, 0xA0); // NO-BREAK SPACE
}

void AdjustFontMetrics(CXfFontDef* pFontDef, bool bBold, bool bItalic)
{
	GlyphMap &glyphs = pFontDef->GetGlyphMap(bBold, bItalic);
	if (glyphs.empty())
		return;

	GlyphMap::iterator it = glyphs.begin();	

	MILLIPOINT ascent  = it->second->FontAscent;
	MILLIPOINT descent = it->second->FontDescent;

	for (++it; it != glyphs.end(); ++it)
	{
		if (it->second->FontAscent > ascent)
			ascent = it->second->FontAscent;

		if (it->second->FontDescent < descent)
			descent = it->second->FontDescent;
	}

	/*if (ascent + descent > 96000)
	{
		for (it = glyphs.begin(); it != glyphs.end(); ++it)
		{
			size_t count = min(it->second->Verbs.size(), it->second->Points.size());
			if (!count)
				continue;
			Rect bounds = FindXFPathBounds(&it->second->Verbs[0], &it->second->Points[0], count);

			ascent = bounds.y + bounds.height;
			descent = bounds.y;
		}
	}*/

	for (it = glyphs.begin(); it != glyphs.end(); ++it)
	{
		it->second->FontAscent = ascent;
		it->second->FontDescent = descent;
	}
}

void AdjustFontMetrics(CXfFontDef* pFontDef)
{
	AdjustFontMetrics(pFontDef, false, false);
	AdjustFontMetrics(pFontDef, false, true);
	AdjustFontMetrics(pFontDef, true, false);
	AdjustFontMetrics(pFontDef, true, true);
}

/******************************************************************************
   Function:	FontManager::CreateFontDefs
   Author:		Paul Gruzdev
   Purpose:     A modern way to create a set of CXfFontDef out of much 
				bigger set of mfd's. We try to do this the way that
				Camelot does. We enumerate system fonts and divide them 
				into non-multiweight and multiweight. Then we scan embedded fonts
				and also divide. Any non-multiweight variant can contain up to 2 weights
				defined by weight and altweight properties of a font def, and up to 4
				different font variants: regular, bold, italic, bold-italic. Note
				that flag BOLD is used both for bold glyphs and text to select the second
				weight. Each multiweight font record contains only ONE weight - so
				it may contain only two variants - regular and italic.
   Returns:		void
   Notes:       
******************************************************************************/
void FontManager::CreateFontDefs(bool bEmbedGlyphs, bool bSubset)
{
	std::map<CXfFontDef*, bool> fontDefPresent;
	//MfdMap map; // put all our mfd's into a map typeface -> set of weights, set of mfd's for each weight
	for (FontMap::iterator it = m_fontMap.begin(); it != m_fontMap.end(); ++it)
	{
		if (fontType3 != it->second.m_pFont->getType())
		{
			InitMappedFontDesc(it->second);
			mfdMap[it->second.fontName][it->second.weight].push_back(&(it->second));
		}
	}
	// Now run through each typeface name. One CXfFontDef for one non-multiweight set of mfd's,
	// One CXfFontDef for each weight of a multi-weight font typeface name.
	for (MfdMap::iterator it = mfdMap.begin(); it != mfdMap.end(); ++it)
	{
		MfdWeightCollection& mwc = it->second; // get our map: weight ---> vector of mfd's
		int iBaseWeight = mwc.begin()->first;
		bool bIt = mwc.begin()->second.front()->isItalic;

		bool bMultiweight = false;
		FontMapData fmd(it->first, L"", iBaseWeight, bIt);
		FontMapDataSet::iterator itFnt = fontMapper.m_DataSet.find(fmd);
		if (itFnt != fontMapper.m_DataSet.end())
			bMultiweight = itFnt->m_bMultiweight;
		else
			bMultiweight = mwc.size() > 2; // number of weights in the map > 2

		if (!bMultiweight)
		{
			int iAltWeight = 0;
			if (mwc.size() > 1)
				iAltWeight = (++mwc.begin())->first;
			else if (iBaseWeight > 500) // DDES-15609 (document properties incorrectly shows the font as installable, Not installed.)
			{
				iAltWeight = iBaseWeight;
				iBaseWeight = 400;
			}

			CXfFontDef* pFontDef = new CXfFontDef(it->first, it->first, iBaseWeight, 0, 0, iAltWeight ? CXfFontDef::AlternativeWeight : 0);
			if (iAltWeight) pFontDef->altweight = iAltWeight;

			m_pChapter->AppendChild(pFontDef);

			for (MfdWeightCollection::iterator it1 = mwc.begin(); it1 != mwc.end(); ++it1)
			{
				bool bBold = mwc.size() > 1 ? it1 != mwc.begin() :
					0 != iAltWeight; // DDES-15609 (document properties incorrectly shows the font as installable, Not installed.)
				MfdWeightData& v = it1->second;
				for (MfdWeightData::iterator i = v.begin(); i != v.end(); ++i)
				{
					MappedFontDesc& mfd = *(*i);
					if (!mfd.bPresentInTheSystem) mfd.isBold = bBold;
					mfd.node = pFontDef;
					mfd.pFontDefOrg = pFontDef;

					mfd.emWidth = -1;
					pFontDef->SetEmbedded(pFontDef->IsEmbedded() || mfd.m_bEmbedded ? true : false);
					
					// Here we probably can enumerate and embed all glyphs
					if (bEmbedGlyphs)
					{
						pFontDef->SetRemoved(false); // In case we removed it earlier
						ProcessFontGlyphs(pFontDef, mfd, g_settings.bLoadEmbeddedGlyphs, bSubset, false);

						auto itPresent = fontDefPresent.find(pFontDef);

						if (mfd.bPresentInTheSystem && itPresent == fontDefPresent.end())
							fontDefPresent[pFontDef] = true;
						else if (!mfd.bPresentInTheSystem && itPresent != fontDefPresent.end())
							itPresent->second = false;
					}
				}
			}

			AdjustFontMetrics(pFontDef);
		}
		else
		{
			for (MfdWeightCollection::iterator it1 = mwc.begin(); it1 != mwc.end(); ++it1)
			{
				CXfFontDef* pFontDef = new CXfFontDef(it->first, it->first, it1->first, 0, 0, CXfFontDef::MultiWeight);
				m_pChapter->AppendChild(pFontDef);

				MfdWeightData& v = it1->second;
				for (MfdWeightData::iterator i = v.begin(); i != v.end(); ++i)
				{
					MappedFontDesc& mfd = *(*i);
					mfd.isBold = false;
					mfd.node = pFontDef;
					mfd.pFontDefOrg = pFontDef;

					mfd.emWidth = -1;
					pFontDef->SetEmbedded(pFontDef->IsEmbedded() || mfd.m_bEmbedded ? true : false);

					// Here we probably can enumerate and embed all glyphs
					if (bEmbedGlyphs)
					{
						pFontDef->SetRemoved(false); // In case we removed it earlier
						ProcessFontGlyphs(pFontDef, mfd, g_settings.bLoadEmbeddedGlyphs, bSubset, false);

						auto itPresent = fontDefPresent.find(pFontDef);

						if (mfd.bPresentInTheSystem && itPresent == fontDefPresent.end())
							fontDefPresent[pFontDef] = true;
						else if (!mfd.bPresentInTheSystem && itPresent != fontDefPresent.end())
							itPresent->second = false;
					}
				}

				AdjustFontMetrics(pFontDef);
			}
		}
	}

	// Now check all our mfd's that they have a space char width defined. This is absolutely needed for proper formatting
	// of text later when rendering: space char width should be defined
	for (MfdMap::iterator it = mfdMap.begin(); it != mfdMap.end(); ++it)
	{
		MfdWeightCollection& mwc = it->second; // get our map: weight ---> vector of mfd's
		for (MfdWeightCollection::iterator itw = mwc.begin(); itw != mwc.end(); ++itw)
		{
			MfdWeightData& mwd = itw->second;
			for (MfdWeightData::iterator itd = mwd.begin(); itd != mwd.end(); ++itd)
			{
				MappedFontDesc* pmfd = *itd;
				AssignSpaceCharacters(pmfd);
			}
		}
	}

	for(std::map<CXfFontDef*, bool>::iterator it = fontDefPresent.begin(); it != fontDefPresent.end(); ++it)
	{
		if(it->second == true)
		{
			CXfFontDef* pFontDef = it->first;
			pFontDef->MarkGlyphsRemoved(false, false);
			pFontDef->MarkGlyphsRemoved(false, true);
			pFontDef->MarkGlyphsRemoved(true, false);
			pFontDef->MarkGlyphsRemoved(true, true);
			pFontDef->SetEmbedded(false);
		}
	}

	for (FontMap::iterator it = m_fontMap.begin(); it != m_fontMap.end(); ++it)
	{
		if (fontType3 == it->second.m_pFont->getType())
			CreateType3MappedFontDesc(it->second);
	}

#ifdef _DEBUG
	for (FontMap::iterator it = m_fontMap.begin(); it != m_fontMap.end(); ++it)
	{
		for(std::map<int, CXfFontDef*>::iterator itt = it->second.extraNode.begin(); itt != it->second.extraNode.end(); ++itt)
		{
			printf("ref:%d, code:%d\n", it->second.m_pFont->getID()->num, itt->first);
		}
	}
#endif
}

bool IsStandardFont(const std::wstring &fontName)
{
	if (fontName == L"Arial")
		return true;

	if (fontName == L"Times New Roman")
		return true;

	if (fontName == L"Courier New")
		return true;

	return false;
}

/******************************************************************************
   Function:	FontManager::CompareGlyphWithSystem
   Author:		Paul Gruzdev
   Purpose:     Compare the embedded glyph [pGlyph] with the system one of
				the same font.
   Returns:		bool
   Parameters:	MappedFontDesc & mfd  
   Parameters:	CXfFontGlyph * pGlyph  
   Notes:       
******************************************************************************/
bool FontManager::CompareGlyphWithSystem(MappedFontDesc& mfd, CXfFontGlyph* pGlyph, CGlyphDiff& diff, const std::wstring& ligChars)
{
	if (g_settings.bAlwaysUseEmbeddedFonts)
		return false;

	CFontDC hdc(mfd.fontName.c_str(), mfd.weight, mfd.isItalic);

	DWORD glyphID = 0;
	if (!ligChars.empty())
	{
		CXfFontOTLigatures *pFontOTLigatures = GetLigatures(mfd.pFontDefOrg, mfd.isBold, mfd.isItalic, false);
		if (!pFontOTLigatures)
		{
			pFontOTLigatures = GetLigatures(mfd.pFontDefOrg, mfd.isBold, mfd.isItalic, true);
			OTFDataLoader otfDataLoader;
			COTFDataLoaderCallback callback(pFontOTLigatures);
			otfDataLoader.ParseFontData(hdc, &callback);
			otfDataLoader.SetSubstitutionData(&callback);
		}

		std::vector<XarLigatureEntry>::iterator itLigature = GetLigatureEntry(pFontOTLigatures, ligChars);
		if (pFontOTLigatures->ligs.end() != itLigature)
			glyphID = itLigature->GlyphID;
	}

	wchar_t wUnicode = pGlyph->CharCode;

	// First step - check if the glyph exists in the system font
	if (!glyphID && !hdc.GlyphExists(wUnicode))
	{
		diff.bGlyphNotPresent = true;
		std::wstring strMessage = FormatStringW(L"<%s> glyph %X: not present in the system\n", mfd.fontNameRaw.c_str(), wUnicode);
		wprintf(strMessage.c_str());
		m_dumpGlyphs.AddDiffGlyph(mfd, pGlyph->GetVerbs(), pGlyph->GetCoords(), pGlyph->NumCoords(), NULL, NULL, 0, strMessage);
		return false;
	}

	CharDescription desc(glyphID ? (WCHAR)glyphID : wUnicode, mfd.fontName, mfd.weight, mfd.isItalic, glyphID ? TRUE : FALSE);
	DocCoord* pCoord = NULL;
	PathVerb* pVerb = NULL;
	ULONG ulNumCoords = 0;
	BOOL bRes = TextManager::GetTTCharPath(desc, &pCoord, &pVerb, &ulNumCoords, NULL);
	bool bMissingGlyph = pGlyph->Points.size() == 0 && ulNumCoords != 0;
	//LONG lCharWidth = glyphID ? hdc.GetGlyphWidth(glyphID) : mfd.GetCharWidthOS(wUnicode);

	if (bRes)
	{
		/*if (pGlyph->Points.size() == 0 && ulNumCoords == 0 && abs(pGlyph->Width.get() - lCharWidth) > GLYPH_WIDTH_DIFF_THRESGHOLD)
		{
			diff.bWidthMismatch = true;
			wprintf(L"empty<%s> glyph %X: %d char width vs sys %d char width\n", mfd.fontNameRaw.c_str(), wUnicode, pGlyph->Width.get(), lCharWidth);
			return false;
		}

		if (!bMissingGlyph && abs(pGlyph->Width.get() - lCharWidth) > GLYPH_WIDTH_DIFF_THRESGHOLD)
		{
			diff.bWidthMismatch = true;
			wprintf(L"<%s> glyph %X: %d char width vs sys %d char width\n", mfd.fontNameRaw.c_str(), wUnicode, pGlyph->Width.get(), lCharWidth);
			return false;
		}*/

		if (!bMissingGlyph && pGlyph->Points.size() && ulNumCoords && !IsStandardFont(mfd.fontName))
		{
			double dArea = CompareOutlines(&pGlyph->Verbs[0], &pGlyph->Points[0], (DWORD)pGlyph->Verbs.size(), pVerb, pCoord, (DWORD)ulNumCoords, diff);
			bool bSimilar = diff.GlyphsSimilar();

			if (!bSimilar)
			{
				std::wstring strMessage = FormatStringW(L"<%s> glyph '%c'[%X]: %.4f area diff\n", mfd.fontNameRaw.c_str(), wUnicode, wUnicode, (float)dArea);
				wprintf(strMessage.c_str());
				m_dumpGlyphs.AddDiffGlyph(mfd, pGlyph->GetVerbs(), pGlyph->GetCoords(), pGlyph->NumCoords(), pVerb, pCoord, ulNumCoords, strMessage);
			}

			return bSimilar;
			//wprintf(L"<%s> glyph %X: %d points vs sys %d points\n", mfd.fontNameRaw.c_str(), pGlyph->CharCode, pGlyph->Points.size(), ulNumCoords);
			//return false;
		}

		// This could be missing glyph in the font file. But if we start to draw with it using system font,
		// we draw regular glyphs instead of whitespaces: OnePage.pdf, DDES-1523, DDES-1185. So in a case of missing
		// glyph in the font file but existing non-whitespace in the system, return false
		if (bMissingGlyph)
		{
			return false;
		}
	}

	return true;
}

/******************************************************************************
   Function:    FontManager::FontExists
   Author:      Paul Gruzdev
   Purpose:		Determine whether a particular font exist in the system using
				global font map
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool FontManager::FontExists(const MappedFontDesc& mfd)
{
	std::wstring fontKey = mfd.fontNamePS;
	std::transform(fontKey.begin(), fontKey.end(), fontKey.begin(), std::tolower);

	if (fontMapper.m_FontMap.find(fontKey) != fontMapper.m_FontMap.end())
		return true;

	bool bTypefacePresent = CFontMapper::TypefacePresentInTheSystem(mfd.fontName);
/*
	FontMapData fmd(mfd.fontName, L"", mfd.weight, mfd.isItalic);
	//return fontMapper.m_DataSet.find(fmd) != fontMapper.m_DataSet.end();
	bool bVariantPresent = fontMapper.m_DataSet.find(fmd) != fontMapper.m_DataSet.end();
	if(bVariantPresent != bTypefacePresent)
	{
		FILE* out = _wfopen(L"d:/xara/powershell/font_typeface_present_diff.txt", L"at");
		fwprintf(out, L"%s: %s: %s, %s-%d-%s: %s\n", g_pszInputName, mfd.fontName.c_str(), bTypefacePresent ? L"present" : L"absent",
			mfd.fontName.c_str(), mfd.weight, mfd.isItalic ? L"it" : L"reg", bVariantPresent ? L"present" : L"absent");
		fclose(out);
	}*/
	return bTypefacePresent;
}

/******************************************************************************
   Function:	FontManager::EmbedDocumentFontGlyphs
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		bool
   Parameters:	bool bSubset  
   Parameters:	VXfTreeNode * pChapter  
   Notes:       
******************************************************************************/
/*bool FontManager::EmbedDocumentFontGlyphs(bool bSubset, VXfTreeNode* pChapter)
{
	START_STOPWATCH_FN;

#ifdef FILTER_DEBUG
	FILE* out = fopen("d:/Xara/Powershell/kpairs_dump.txt", "at");
	if (out)
	{
		fwprintf(out, L"------------------ %s\n", m_filename);
		fclose(out);
	}
#endif

	std::map<CXfFontDef*, bool> fontDefPresent;
	for (TextFontMap::iterator it = textFontMap.begin(); it != textFontMap.end(); ++it)
	{
		MappedFontDesc& mfd = it->second;
		CXfFontDef* pFontDef = mfd.node;
		if(!bSubset || !mfd.unicodeSubset.empty())
		{
			pFontDef->SetRemoved(false); // In case we removed it earlier
			ProcessFontGlyphs(pFontDef, mfd, g_settings.bLoadEmbeddedGlyphs, bSubset, false);

			if(mfd.bPresentInTheSystem && fontDefPresent.find(pFontDef) == fontDefPresent.end())
				fontDefPresent[pFontDef] = true;
			else if(!mfd.bPresentInTheSystem && fontDefPresent.find(pFontDef) != fontDefPresent.end())
				fontDefPresent[pFontDef] = false;
		}
		else
		{
			// Font subset is empty - don't add this font def at all! This happens if we import only particular pages, 
			// and they do not use this particular font
			// temp comment - should be uncommented
			if(!pFontDef->HasGlyphs() && !pFontDef->IsExternal())
				pFontDef->SetRemoved(true);
		}
	}

	for(std::map<CXfFontDef*, bool>::iterator it = fontDefPresent.begin(); it != fontDefPresent.end(); ++it)
	{
		if(it->second == true)
		{
			CXfFontDef* pFontDef = it->first;
			pFontDef->MarkGlyphsRemoved(false, false);
			pFontDef->MarkGlyphsRemoved(false, true);
			pFontDef->MarkGlyphsRemoved(true, false);
			pFontDef->MarkGlyphsRemoved(true, true);
		}
	}
	//for (TextFontMap::iterator it = textFontMap.begin(); it != textFontMap.end(); ++it)
	//{
	//	delete it->second.pFont;
	//}

	STOP_STOPWATCH(__FUNCTION__)
	return true;
}*/

class SplashOutFontFileID: public SplashFontFileID {
public:

	SplashOutFontFileID(const Ref *rA) { r = *rA; }

	~SplashOutFontFileID() {}

	bool matches(SplashFontFileID *id) {
		return ((SplashOutFontFileID *)id)->r.num == r.num &&
			((SplashOutFontFileID *)id)->r.gen == r.gen;
	}

private:

	Ref r;
};

// -------------------------------------------------------------------------
SplashFontSrc *CreateSplashFontSrc(GfxFont *gfxFont, XRef *xref, GfxFontType &fontType)
{
	SplashFontSrc *fontSrc = NULL;
	if (gfxFont)
	{
		fontType = gfxFont->getType();

		if (GfxFontLoc *fontLoc = gfxFont->locateFont(xref, NULL))
		{
			switch (fontLoc->locType)
			{
			case gfxFontLocEmbedded:
				{
					int tmpBufLen = 0;
					if (char *tmpBuf = gfxFont->readEmbFontFile(xref, &tmpBufLen))
					{
						fontSrc = new SplashFontSrc();
						fontSrc->setBuf(tmpBuf, tmpBufLen, true);
					}
				}
				break;
			case gfxFontLocExternal:
				if (fontLoc->path)
				{
					fontSrc = new SplashFontSrc();
					fontSrc->setFile(fontLoc->path, false);
					fontType = fontLoc->fontType;
				}
				break;
			//case gfxFontLocResident:
			//	break;
			}
			delete fontLoc;
		}
	}
	else
		fontType = fontUnknownType;
	return fontSrc;
}

SplashFontFile *CreateSplashFontFile(SplashFontEngine *fontEngine, GfxFont *gfxFont, SplashFontFileID *id, SplashFontSrc *fontsrc, GfxFontType fontType)
{
	FoFiTrueType *ff;
	int *codeToGID;
	int codeToGIDLen;

	if (gfxFont->isCIDFont())
	{
		switch (fontType)
		{
		//----- GfxCIDFont
		case fontCIDType0:
		case fontCIDType0C:
			return fontEngine->loadCIDFont(id, fontsrc);
		case fontCIDType0COT:
			if (int *cidToGID = ((GfxCIDFont *)gfxFont)->getCIDToGID())
			{
				codeToGIDLen = ((GfxCIDFont *)gfxFont)->getCIDToGIDLen();
				codeToGID = (int *)gmallocn(codeToGIDLen, sizeof(int));
				memcpy(codeToGID, cidToGID, codeToGIDLen * sizeof(int));
			}
			else
			{
				codeToGID = NULL;
				codeToGIDLen = 0;
			}
			return fontEngine->loadOpenTypeCFFFont(id, fontsrc, codeToGID, codeToGIDLen);
		case fontCIDType2:
		case fontCIDType2OT:
			codeToGID = NULL;
			codeToGIDLen = 0;
			if (int *cidToGID = ((GfxCIDFont *)gfxFont)->getCIDToGID())
			{
				if (codeToGIDLen = ((GfxCIDFont *)gfxFont)->getCIDToGIDLen())
				{
					codeToGID = (int *)gmallocn(codeToGIDLen, sizeof(int));
					memcpy(codeToGID, cidToGID, codeToGIDLen * sizeof(int));
				}
			}
			else
			{
				if (fontsrc->isFile)
					ff = FoFiTrueType::load(fontsrc->fileName->c_str());
				else
					ff = FoFiTrueType::make(fontsrc->buf, fontsrc->bufLen);
				if (ff)
				{
					codeToGID = ((GfxCIDFont *)gfxFont)->getCodeToGIDMap(ff, &codeToGIDLen);
					delete ff;
				}
				else
					return NULL;
			}
			return fontEngine->loadTrueTypeFont(id, fontsrc, codeToGID, codeToGIDLen, 0);
		}
	}
	else
	{
		switch (fontType)
		{
		case fontType1:
			return fontEngine->loadType1Font(id, fontsrc, (const char **)((Gfx8BitFont *)gfxFont)->getEncoding());
		case fontType1C:
			return fontEngine->loadType1CFont(id, fontsrc, (const char **)((Gfx8BitFont *)gfxFont)->getEncoding());
		case fontType1COT:
			return fontEngine->loadOpenTypeT1CFont(id, fontsrc, (const char **)((Gfx8BitFont *)gfxFont)->getEncoding());
		case fontTrueType:
		case fontTrueTypeOT:
			if (fontsrc->isFile)
				ff = FoFiTrueType::load(fontsrc->fileName->c_str());
			else
				ff = FoFiTrueType::make(fontsrc->buf, fontsrc->bufLen);
			if (ff)
			{
				codeToGID = ((Gfx8BitFont *)gfxFont)->getCodeToGIDMap(ff);
				codeToGIDLen = 256;
				delete ff;
				// if we're substituting for a non-TrueType font, we need to mark
				// all notdef codes as "do not draw" (rather than drawing TrueType
				// notdef glyphs)
				if (gfxFont->getType() != fontTrueType && gfxFont->getType() != fontTrueTypeOT)
				{
					for (int i = 0; i < 256; ++i)
						if (codeToGID[i] == 0)
							codeToGID[i] = -1;
				}
			}
			else
			{
				codeToGID = NULL;
				codeToGIDLen = 0;
			}
			return fontEngine->loadTrueTypeFont(id, fontsrc, codeToGID, codeToGIDLen);
		};
	}

	return NULL;
}

SplashFont* FontManager::CreateSplashFont(GfxFont *gfxFont, SplashCoord mat[4], bool bUseFontCache)
{
	if (!gfxFont)
		return NULL;

	SplashFontFile *fontFile;

	// check the font file cache
	SplashOutFontFileID *id = new SplashOutFontFileID(gfxFont->getID());
	if ((fontFile = fontEngine->getFontFile(id)))
		delete id;
	else
	{
		GfxFontType fontType = fontUnknownType;
		if (SplashFontSrc *fontsrc = CreateSplashFontSrc(gfxFont, m_pDoc->getXRef(), fontType))
		{
			// load the font file
			fontFile = CreateSplashFontFile(fontEngine, gfxFont, id, fontsrc, fontType);
			fontsrc->unref();
		}

		if (!fontFile)
		{
			delete id;
			return NULL;
		}
	}

	// create the scaled font
	if (bUseFontCache)
	{
		SplashCoord ctm[4] = {1, 0, 0, 1};
		return fontEngine->getFont(fontFile, mat, ctm); 
	}
	else
	{
		SplashCoord mat1[4] = {mat[0], -mat[1], mat[2], -mat[3]};
		return fontFile->makeFont(mat1, mat);
	}
}

void FontManager::CreateSplashFont(MappedFontDesc &fontDesc)
{
	if (fontType3 == fontDesc.m_pFont->getType())
	{
		fontDesc.m_bEmbedded = true;
	}
	else
	{
		Ref embRef;
		fontDesc.m_bEmbedded = fontDesc.m_pFont->getEmbeddedFontID(&embRef);

		if (!fontDesc.m_bEmbedded && fontDesc.m_pFont->getName() && mBuiltInFonts.end() != mBuiltInFonts.find(fontDesc.m_pFont->getName()->c_str()))
			fontDesc.m_bEmbedded = true;

		if (fontDesc.m_bEmbedded)
		{
			SplashCoord mat[4] = {72, 0, 0, 72};
			fontDesc.m_pSplashFont = CreateSplashFont(fontDesc.m_pFont, mat, false);			

			// InsuranceDoc(edited).pdf: fontRef: 65, Verdana fails to load via Poppler (freetype). We need to mark it as not embedded thus.
			if (!fontDesc.m_pSplashFont)
				fontDesc.m_bEmbedded = false;
		}
	}
}

MappedFontDesc &FontManager::GetFontDesc(const Ref &ref)
{
	auto it = m_fontMap.find(ref);
	if (m_fontMap.end() != it)
		return it->second;
	MappedFontDesc &fontDesc = m_fontMap[ref];

	fontDesc.m_fontObj = m_pDoc->getXRef()->fetch(ref.num, ref.gen);

	fontDesc.m_pFont = GfxFont::makeFont(m_pDoc->getXRef(), "temp font", ref, fontDesc.m_fontObj.getDict());
	if (!fontDesc.m_pFont)
		return fontDesc;

	fontDesc.m_pFont->incRefCnt();

	CreateSplashFont(fontDesc);

	return fontDesc;
}

MappedFontDesc &FontManager::GetFontDesc(GfxFont* pFont)
{
	const Ref &ref = *(pFont->getID());
	auto it = m_fontMap.find(ref);
	if (m_fontMap.end() != it)
		return it->second;
	MappedFontDesc &fontDesc = m_fontMap[ref];

	fontDesc.m_fontObj = m_pDoc->getXRef()->fetch(ref.num, ref.gen);

	fontDesc.m_pFont = pFont;
	fontDesc.m_pFont->incRefCnt();

	CreateSplashFont(fontDesc);

	return fontDesc;
}

MappedFontDesc *FontManager::GetFont(GfxState *state)
{
	GfxFont* pFont = state->getFont();

	FontMap::iterator itFontDesc = m_fontMap.find(*(pFont->getID()));

	// DDES-516 fix. Sometimes we get here and state refers to unknown font for us! Without drawing anything with it.
	// In this case just bail.
	// An update! Font scanner does not scan deeply inside - for example XObject (group) with text inside being a part of a Soft Mask
	// (DDES-516, page 237) - is not being scanned by font scanner so the whole font gets bailed by the importer. So we need to
	// add this new font here, or create a better font scanner at the start
	if (itFontDesc == m_fontMap.end())
		return nullptr;

	MappedFontDesc& mfd = itFontDesc->second;

	mfd.ascent  = _round(pFont->getAscent() * 72000. + 0.5);
	mfd.descent = _round(pFont->getDescent() * 72000. + 0.5);

	if (mfd.bPresentInTheSystem)
	{
		FontMetrics metrics = mfd.GetFontMetricsOS();
		mfd.ascent  = metrics.ascent;
		mfd.descent = metrics.descent;
	}

	return &itFontDesc->second;
}

static bool FileExists(const char *path)
{
	FILE * f = fopen(path, "rb");
	if (f) {
		fclose(f);
		return true;
	}
	return false;
}

/******************************************************************************
   Function:	FontManager::InitialiseBuiltInFonts
   Author:		Paul Gruzdev
   Purpose:     Initialise global params with paths to fonts that are distributed
				along with pdf2xar. ZapfDingbats currently
   Returns:		void
   Notes:       
******************************************************************************/
void FontManager::InitialiseBuiltInFonts(std::string strFilterDataDir)
{
	GooString* str1 = new GooString(std::string(strFilterDataDir + "\\d050000l.pfb").c_str());
	if (FileExists(str1->c_str()))
	{
		GooString* pDing = new GooString("Dingbats");
		m_pGlobalParams->addFontFile(pDing, str1);
		mBuiltInFonts[pDing->c_str()] = L"Dingbats";

		GooString* pZapf = new GooString("ZapfDingbats");
		m_pGlobalParams->addFontFile(pZapf, new GooString(str1));
		mBuiltInFonts[pZapf->c_str()] = L"Dingbats";
	}
}

GlyphDiffDocument::GlyphDiffDocument(std::wstring filename)
{
	m_strSource = filename;
	m_bDumpEmbeddedFonts = g_settings.bDumpEmbeddedFonts;

	if (g_settings.bDumpGlyphDiffs)
	{
		m_bActive = true;
		m_strDumpFilename = filename.substr(0, filename.length() - 4) + L"-glyph-diffs.xar";

		m_tree.GetRoot()->AppendChild(new CXfFileHeader("pdf2xar", "001", "glyph diff"));
		VXfTreeNode* pDoc = new CXfZeroSizedNode(TAG_DOCUMENT);
		m_tree.GetRoot()->AppendChild(pDoc);

		CXfViewport* pViewport = new CXfViewport(DocCoord(0, 0), DocCoord(72000, 72000));
		pDoc->AppendChild(pViewport);
		// Create and add a new document info tag that is now *required* for all multi-page imported documents
		CXfDocumentInformation* pDocInfo = new CXfDocumentInformation(1); // we have only one chapter
		pDocInfo->AddChapterInfo(1); // we fill this in later at the end
		pDoc->AppendChild(pDocInfo);

		pDoc->AppendChild(new CXfDocumentFlags(DOCFLAG_EMPTYLINESSIGNIFICANT | DOCFLAG_WORD_LINE_SPACING/* | DOCFLAG_PROMOTEURLS*/));
		//DOCFLAG_LEGACYAPTS | DOCFLAG_ALLPAGESTHESAME | DOCFLAG_SAMELAYERSONALLPAGES | DOCFLAG_EMPTYLINESSIGNIFICANT));
		//DOCFLAG_EMPTYLINESSIGNIFICANT | DOCFLAG_SIMULATEWEBLENGTHS | DOCFLAG_PROMOTEURLS));
		/*| DOCFLAG_MULTILAYER | DOCFLAG_ALLLAYERSVISIBLE*/

		m_pChapter = new CXfZeroSizedNode(TAG_CHAPTER);
		pDoc->AppendChild(m_pChapter);
	}
}

extern BOOL CXfTextWriteTextString(VXfTreeNode* pInsert, std::wstring str, std::wstring strTypeface, ULONG ulFontSize, bool bBold,
	bool bItalic, bool bUnderline, COLORREF crTextColor);

void GlyphDiffDocument::AddDiffGlyph(MappedFontDesc& mfd, PathVerb* pVerb, DocCoord* pCoord, ULONG ulNumCoords, PathVerb* pSysVerb, DocCoord* pSysCoord, ULONG ulNumSysCoords, std::wstring strMessage)
{
	const COLORREF crGlyph = RGB(46, 48, 224);
	const COLORREF crSystem = RGB(224, 175, 34);

	if (m_bActive)
	{
		// For the first page we create TAG_SPREAD, for each consecutive page it should be PHASE2
		VXfTreeNode *pPage = new CXfZeroSizedNode(m_bFirstPage ? TAG_SPREAD : TAG_SPREAD_PHASE2);
		m_bFirstPage = false;
		m_pChapter->AddChild(pPage);

		DWORD dwWidth = 72000;
		DWORD dwHeight = 72000;

		// flags=2: Show drop shadow = true, double page spread = false
		pPage->AddChild(new CXfSpreadInformation(dwWidth, dwHeight, 72000 * 12, 0, 2));

		CXfLayer* pLayer = new CXfLayer();
		pPage->AppendChild(pLayer);
		pLayer->AppendChild(new CXfLayerDetails(TAG_LAYER_FLAG_VISIBLE | TAG_LAYER_FLAG_PRINTABLE | TAG_LAYER_FLAG_ACTIVE,
			L"Layer 1"));

		if (pVerb && pCoord && ulNumCoords)
		{
			CXfPath* pPath = new CXfPath(TAG_PATH_FILLED_STROKED);
			pLayer->AppendChild(pPath);
			pPath->CopyPoints(pVerb, pCoord, ulNumCoords);

			pPath->AppendChild(new CXfSimpleColor(GetRValue(crGlyph), GetGValue(crGlyph), GetBValue(crGlyph)));
			pPath->AppendChild(new CXfFlatFill(pPath->GetLastChild()));
// 			pPath->AppendChild(new CXfFlatTransparentFill(byTrans, CXF_TRANSPARANCY_TYPE_MIX));
// 			pPath->AppendChild(new CXfLineTransparency(byLineTrans, CXF_TRANSPARANCY_TYPE_MIX));
			pPath->AppendChild(new CXfWindingRule(CXF_WINDING_RULE_NONZERO));
		}

		if (pSysVerb && pSysCoord && ulNumSysCoords)
		{
			CXfPath* pPath = new CXfPath(TAG_PATH_FILLED_STROKED);
			pLayer->AppendChild(pPath);
			pPath->CopyPoints(pSysVerb, pSysCoord, ulNumSysCoords);

			pPath->AppendChild(new CXfSimpleColor(GetRValue(crSystem), GetGValue(crSystem), GetBValue(crSystem)));
			pPath->AppendChild(new CXfFlatFill(pPath->GetLastChild()));
			pPath->AppendChild(new CXfFlatTransparentFill(92, CXF_TRANSPARANCY_TYPE_MIX));
			pPath->AppendChild(new CXfLineTransparency(92, CXF_TRANSPARANCY_TYPE_MIX));
			pPath->AppendChild(new CXfWindingRule(CXF_WINDING_RULE_NONZERO));
		}

		if (strMessage.length())
		{
			CXfTextStorySimple* pStory = new CXfTextStorySimple(1000, 1000, 1);
			pLayer->AppendChild(pStory);
			pStory->AppendChild(new CXfTextLine());

			CXfTextWriteTextString(pStory->GetLastChild(), strMessage, L"Open Sans", 9000, false, false, false, RGB(0, 0, 0));
			pStory->GetLastChild()->AppendChild(new CXfZeroSizedNode(TAG_TEXT_EOL));
		}
	}
}

void GlyphDiffDocument::ScanAddFonts(VXfTreeNode* pRoot)
{
	VXfTreeNodeVector fonts;
	pRoot->FindNodes(CTagSet(TAG_FONT_DEF_TRUETYPE, TAG_FONT_DEF_ATM), fonts);

	for (size_t i = 0; i < fonts.size(); i++)
	{
		CXfFontDef* pFont = (CXfFontDef*)fonts[i];
		if (pFont->IsEmbedded())
			AddFont(pFont);
	}
}

GlyphDiffDocument::~GlyphDiffDocument()
{
	if (m_bDumpEmbeddedFonts && m_fonts.size())
	{
		const COLORREF crGlyph = RGB(46, 48, 224);
		const wchar_t* pszVariant[4] = { L"Regular", L"Bold", L"Italic", L"Bold-Italic"};
		std::wstring strFilename = m_strSource.substr(0, m_strSource.length() - 4) + L"-EmbeddedFonts.xar";

		CXfXaraTree tree;
		tree.GetRoot()->AppendChild(new CXfFileHeader("pdf2xar", "001", "EmbFonts"));
		VXfTreeNode* pDoc = new CXfZeroSizedNode(TAG_DOCUMENT);
		tree.GetRoot()->AppendChild(pDoc);

		CXfViewport* pViewport = new CXfViewport(DocCoord(0, 0), DocCoord(96000 * 5, 96000 * 5));
		pDoc->AppendChild(pViewport);
		// Create and add a new document info tag that is now *required* for all multi-page imported documents
		CXfDocumentInformation* pDocInfo = new CXfDocumentInformation(1); // we have only one chapter
		pDocInfo->AddChapterInfo(1); // we fill this in later at the end
		pDoc->AppendChild(pDocInfo);

		pDoc->AppendChild(new CXfDocumentFlags(DOCFLAG_EMPTYLINESSIGNIFICANT | DOCFLAG_WORD_LINE_SPACING));

		VXfTreeNode* pChapter = new CXfZeroSizedNode(TAG_CHAPTER);
		pDoc->AppendChild(pChapter);

		for (size_t fnt = 0; fnt < m_fonts.size(); fnt++)
		{
			CXfFontDef* pFont = m_fonts[fnt];
			size_t szMaxCnt = max(max(pFont->GetGlyphMap(false, false).size(), pFont->GetGlyphMap(false, true).size()),
				max(pFont->GetGlyphMap(true, false).size(), pFont->GetGlyphMap(true, true).size()));
			if (szMaxCnt == 0)
				continue;

			// For the first page we create TAG_SPREAD, for each consecutive page it should be PHASE2
			VXfTreeNode *pPage = new CXfZeroSizedNode(fnt == 0 ? TAG_SPREAD : TAG_SPREAD_PHASE2);
			pChapter->AddChild(pPage);

			DWORD dwWidth = 96000*(DWORD)szMaxCnt;
			DWORD dwHeight = 96000*5;

			// flags=2: Show drop shadow = true, double page spread = false
			pPage->AddChild(new CXfSpreadInformation(dwWidth, dwHeight, 72000 * 12, 0, 2));

			CXfLayer* pLayer = new CXfLayer();
			pPage->AppendChild(pLayer);
			pLayer->AppendChild(new CXfLayerDetails(TAG_LAYER_FLAG_VISIBLE | TAG_LAYER_FLAG_PRINTABLE | TAG_LAYER_FLAG_ACTIVE,
				L"Layer 1"));

			// Draw the font name first
			{
				CXfTextStorySimple* pStory = new CXfTextStorySimple(12000, 12000, 1);
				pLayer->AppendChild(pStory);
				pStory->AppendChild(new CXfTextLine());
				std::wstring str = FormatStringW(L"FontName: %s", pFont->FullFontName.get());
				CXfTextWriteTextString(pStory->GetLastChild(), str, L"Open Sans", 36000, false, false, false, RGB(0, 0, 0));
				pStory->GetLastChild()->AppendChild(new CXfZeroSizedNode(TAG_TEXT_EOL));
			}

			for (size_t j = 0; j < 4; j++)
			{
				bool bBold = (j & 1) ? true : false;
				bool bItalic = (j & 2) ? true : false;

				std::vector<CXfFontGlyph*> glyphs;
				pFont->GetGlyphs(glyphs, bBold, bItalic);
				ULONG x = 12000;
				ULONG y = (4 - (ULONG)j) * 96000 + 12000;

				// Draw the font variant
				{
					CXfTextStorySimple* pStory = new CXfTextStorySimple(x, y-12000, 1);
					pLayer->AppendChild(pStory);
					pStory->AppendChild(new CXfTextLine());
					CXfTextWriteTextString(pStory->GetLastChild(), pszVariant[j], L"Open Sans", 12000, false, false, false, RGB(0, 0, 0));
					pStory->GetLastChild()->AppendChild(new CXfZeroSizedNode(TAG_TEXT_EOL));
				}

				for (size_t i = 0; i < glyphs.size(); i++)
				{
					CXfFontGlyph* pGlyph = glyphs[i];
					ULONG ulNumCoords = pGlyph->NumCoords();

					{
						CXfPath* p = new CXfPath(TAG_PATH_FILLED_STROKED);
						p->Add(PT_MOVETO, DocCoord(x, y));
						p->Add(PT_LINETO, DocCoord(x + 72000, y));
						p->Add(PT_LINETO, DocCoord(x + 72000, y + 72000));
						p->Add(PT_LINETO + PT_CLOSEFIGURE, DocCoord(x, y + 72000));
						pLayer->AppendChild(p);
						p->AppendChild(new CXfZeroSizedNode(TAG_FLATFILL_NONE));
						p->AppendChild(new CXfZeroSizedNode(TAG_LINECOLOUR_BLACK));
						p->AppendChild(new CXfLineWidth(500));
						p->AppendChild(new CXfLineTransparency(127, CXF_TRANSPARANCY_TYPE_MIX));
					}
					if (ulNumCoords)
					{
						PBYTE pVerb = pGlyph->GetVerbs();
						DocCoord* pCoord = pGlyph->GetCoords();

						CXfPath* pPath = new CXfPath(TAG_PATH_FILLED_STROKED);
						pLayer->AppendChild(pPath);
						pPath->CopyPoints(pVerb, pCoord, ulNumCoords);

						XaraMatrix mat(DocCoord(x, y));
						mat.transform(&pPath->Coords[0], (UINT)pPath->Coords.size());

						pPath->AppendChild(new CXfZeroSizedNode(TAG_FLATFILL_BLACK));
// 						pPath->AppendChild(new CXfSimpleColor(GetRValue(crGlyph), GetGValue(crGlyph), GetBValue(crGlyph)));
// 						pPath->AppendChild(new CXfFlatFill(pPath->GetLastChild()));
						// 			pPath->AppendChild(new CXfFlatTransparentFill(byTrans, CXF_TRANSPARANCY_TYPE_MIX));
						// 			pPath->AppendChild(new CXfLineTransparency(byLineTrans, CXF_TRANSPARANCY_TYPE_MIX));
						pPath->AppendChild(new CXfWindingRule(CXF_WINDING_RULE_NONZERO));
					}

					x += 96000;
				}
			}
		}

		tree.GetRoot()->AppendChild(new CXfZeroSizedNode(TAG_ENDOFFILE));

		CXarExport* pExporter = XarLibrary::CreateExporter();
		IStream* strOut = NULL;
		BOOL ok = FALSE;

		{
			HRESULT hRes = SHCreateStreamOnFileEx(strFilename.c_str(), STGM_CREATE | STGM_WRITE, FILE_ATTRIBUTE_NORMAL, TRUE, NULL, &strOut);
			ok = hRes == S_OK && strOut != NULL;
			ok = ok && pExporter->StartExport(strOut);
			tree.Write(tree.GetRoot(), pExporter);
			strOut->Release();
		}

		delete pExporter;

	}

	if (!m_bFirstPage && m_strDumpFilename.length())
	{
		m_tree.GetRoot()->AppendChild(new CXfZeroSizedNode(TAG_ENDOFFILE));

		CXarExport* pExporter = XarLibrary::CreateExporter();
		IStream* strOut = NULL;
		BOOL ok = FALSE;

		{
			HRESULT hRes = SHCreateStreamOnFileEx(m_strDumpFilename.c_str(), STGM_CREATE | STGM_WRITE, FILE_ATTRIBUTE_NORMAL, TRUE, NULL, &strOut);
			ok = hRes == S_OK && strOut != NULL;
			ok = ok && pExporter->StartExport(strOut);
			m_tree.Write(m_tree.GetRoot(), pExporter);
			strOut->Release();
		}

		delete pExporter;
	}
}
