#include "stdafx.h"
#include "XarOutputDev.h"
#include "poppler/Page.h"
#include "poppler/Link.h"
#include "poppler/OptionalContent.h"
#include "poppler/PDFDocEncoding.h"

/******************************************************************************
   Function:	TextStringToUnicode
   Author:	Paul Gruzdev/Denis Onishchenko (correctly use PDFDocEncoding instead of system code page)
   Purpose:     Convert from PDF text string object to Unicode string, possibly
			decoding Unicode string that starts with 0xFEFF in PDF with
			byte order swapped
   Returns:	std::wstring
   Parameters:	GooString* textStr
   Notes:       
******************************************************************************/
std::wstring TextStringToUnicode(const GooString* textStr)
{
	std::wstring res;
	if (!textStr)
		return res;
	int len = textStr->getLength();
	if (len == 0)
		return res;
	const char *s = textStr->c_str();
	if (textStr->hasUnicodeMarker())
	{
		len = len / 2 - 1;
		if (len > 0)
		{
			res.reserve(len);
			s += 2;
			while (len--)
			{
				wchar_t c = ((*s++) & 0xff) << 8;
				c |= ((*s++) & 0xff);
				res.push_back(c);
			}
		}
	}
	else
	{
		res.reserve(len);
		while (len--)
			res.push_back(pdfDocEncoding[(*s++) & 0xff]);
	}
	return res;
}

/******************************************************************************
   Function:	GetLayerName
   Author:	Denis Onishchenko
   Purpose:     Computes layer name from Optional Content Group Dictionary or 
				from Optional Content Membership Dictionary
				(see "8.11 Optional Content" in PDF Reference)
   Returns:	std::wstring
   Parameters:	Dict *dict
   Notes:       
******************************************************************************/
std::wstring XarOutputDev::GetLayerName(Dict *dict)
{
	if (dict)
	{
		if (dict->is("OCG")) // Optional Content Group Dictionary
		{
			Object name = dict->lookup("Name");
			if (name.isString())
			return TextStringToUnicode(name.getString());
		}
		else if (dict->is("OCMD") && !dict->hasKey("P") && !dict->hasKey("VE")) // Optional Content Membership Dictionary
		{
			Object ocgs = dict->lookup("OCGs");
			if (ocgs.isDict())
				return GetLayerName(ocgs.getDict());
			else if (ocgs.isArray())
			{
				Array *arr = ocgs.getArray();
				int len = arr->getLength();
				if (1 == len)
				{
					Object ocg = arr->get(0);
					if (ocg.isDict())
						return GetLayerName(ocg.getDict());
				}
			}
		}
	}
	return FormatStringW(L"Layer %d", ++m_layerCount);
}

/******************************************************************************
   Function:	XarOutputDev::BeginLayer
   Author:		Paul Gruzdev/Denis Onishchenko
   Purpose:     Called when a new layer is appeared as an optional contents in
				a /Contents stream. Start new Xara layer here
   Returns:		void
   Parameters:	char * name  
   Parameters:	bool bVisible  
   Notes:       
******************************************************************************/

void XarOutputDev::BeginLayer(const wchar_t *name, bool bVisible)
{
	// We always create the initial layer at the page start. So here we need to decide, either we already have
	// the first layer, or we need to create an additional layer object
	if (m_pCurLayer->ChildrenCount() < 2)
	{
		CXfLayerDetails* pDetails = (CXfLayerDetails*)m_pCurLayer->FindFirstNode(TAG_LAYERDETAILS, false);
		pDetails->Name = name;
		pDetails->SetVisible(bVisible != 0 ? true : false);
	}
	else
	{
		CXfLayer* pLayer = new CXfLayer();

		m_pCurLayer->GetParent()->AppendChild(pLayer);
		BYTE byFlags = TAG_LAYER_FLAG_PRINTABLE | TAG_LAYER_FLAG_ACTIVE;
		if(bVisible) byFlags |= TAG_LAYER_FLAG_VISIBLE;
		CXfLayerDetails* pDetails = new CXfLayerDetails(byFlags, name);
		pLayer->AppendChild(pDetails);
		SetParentNode(pLayer);
		m_pCurLayer = pLayer;
	}
}

/******************************************************************************
   Function:	XarOutputDev::BeginOptionalContent
   Author:		Denis Onishchenko
   Purpose:     Called when an optional content sequence is begun in a /Contents stream.
   Returns:		void
   Parameters:	Object *dictRef
   Parameters:	bool bVisible
   Notes:       
******************************************************************************/
void XarOutputDev::BeginOptionalContent(Object *dictRef, bool bVisible)
{
	if (g_settings.bNoLayers)
		return;

	Object dictObj = dictRef->fetch(m_pDoc->getXRef());
	std::wstring name = GetLayerName(1 == m_OCStack.size() && dictObj.isDict() ? dictObj.getDict() : NULL);
	m_OCStack.push_back(std::make_pair(name, bVisible));
	BeginLayer(name.c_str(), bVisible);
}

/******************************************************************************
   Function:	XarOutputDev::EndOptionalContent
   Author:		Denis Onishchenko
   Purpose:     Called when an optional content sequence is ended in a /Contents stream.
   Returns:		void
   Notes:       
******************************************************************************/
void XarOutputDev::EndOptionalContent()
{
	if (g_settings.bNoLayers)
		return;

	if (m_OCStack.size() < 2 ) return;	
	m_OCStack.pop_back();
	BeginLayer(m_OCStack.back().first.c_str(), m_OCStack.back().second);
}
