/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#include "stdafx.h"
#include "PNGWriterEx.h"

#include <zlib.h>
#include <stdlib.h>
#include <string.h>

#include "poppler/Error.h"
#include "poppler/splash/SplashBitmap.h"
#include "goo/gmem.h"

#include <png.h>
#include "../../png/pngstruct.h"

extern void PNGAPI PngWriteDataToMemFile(png_structp png_ptr, png_bytep data, png_size_t length);
extern void PNGAPI PngReadDataFromMemFile(png_structp png_ptr, png_bytep data, png_size_t length);

struct PNGWriterPrivateEx 
{
	PNGWriterEx::Format format;
	png_structp png_ptr;
	png_infop info_ptr;
	unsigned char *icc_data;
	int icc_data_size;
	char *icc_name;
	bool sRGB_profile;
};

PNGWriterEx::PNGWriterEx(Format formatA)
{
	priv = new PNGWriterPrivateEx;
	priv->format = formatA;
	priv->icc_data = NULL;
	priv->icc_data_size = 0;
	priv->icc_name = NULL;
	priv->sRGB_profile = false;
}

PNGWriterEx::~PNGWriterEx()
{
	/* cleanup heap allocation */
	png_destroy_write_struct(&priv->png_ptr, &priv->info_ptr);
	if (priv->icc_data) {
		gfree(priv->icc_data);
		free(priv->icc_name);
	}

	delete priv;
}

void PNGWriterEx::setICCProfile(const char *name, unsigned char *data, int size)
{
	priv->icc_data = (unsigned char *)gmalloc(size);
	memcpy(priv->icc_data, data, size);
	priv->icc_data_size = size;
	priv->icc_name = strdup(name);
}

void PNGWriterEx::setSRGBProfile()
{
	priv->sRGB_profile = true;
}

bool PNGWriterEx::init(FILE *f, int width, int height, int hDPI, int vDPI)
{
	/* libpng changed the png_set_iCCP() prototype in 1.5.0 */
#if PNG_LIBPNG_VER < 10500
	png_charp icc_data_ptr = (png_charp)priv->icc_data;
#else
	png_const_bytep icc_data_ptr = (png_const_bytep)priv->icc_data;
#endif

	/* initialize stuff */
	priv->png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!priv->png_ptr) {
		error(errInternal, -1, "png_create_write_struct failed");
		return false;
	}

	priv->info_ptr = png_create_info_struct(priv->png_ptr);
	if (!priv->info_ptr) {
		error(errInternal, -1, "png_create_info_struct failed");
		return false;
	}

	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "png_jmpbuf failed");
		return false;
	}

	priv->png_ptr->io_ptr = (png_voidp)f;
	priv->png_ptr->write_data_fn = PngWriteDataToMemFile;

	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "Error during writing header");
		return false;
	}

	// Set up the type of PNG image and the compression level
	png_set_compression_level(priv->png_ptr, Z_BEST_SPEED);

	// Silence silly gcc
	png_byte bit_depth = -1;
	png_byte color_type = -1;
	switch (priv->format) {
	case RGB:
		bit_depth = 8;
		color_type = PNG_COLOR_TYPE_RGB;
		break;
	case RGBA:
		bit_depth = 8;
		color_type = PNG_COLOR_TYPE_RGB_ALPHA;
		break;
	case GRAY:
		bit_depth = 8;
		color_type = PNG_COLOR_TYPE_GRAY;
		break;
	case MONOCHROME:
		bit_depth = 1;
		color_type = PNG_COLOR_TYPE_GRAY;
		break;
	}
	png_byte interlace_type = PNG_INTERLACE_NONE;

	png_set_IHDR(priv->png_ptr, priv->info_ptr, width, height, bit_depth, color_type, interlace_type, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	png_set_pHYs(priv->png_ptr, priv->info_ptr, (png_uint_32)(hDPI/0.0254), (png_uint_32)(vDPI/0.0254), PNG_RESOLUTION_METER);

	if (priv->icc_data)
		png_set_iCCP(priv->png_ptr, priv->info_ptr, priv->icc_name, PNG_COMPRESSION_TYPE_BASE, icc_data_ptr, priv->icc_data_size);
	else if (priv->sRGB_profile)
		png_set_sRGB(priv->png_ptr, priv->info_ptr, PNG_sRGB_INTENT_RELATIVE);

	png_write_info(priv->png_ptr, priv->info_ptr);
	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "error during writing png info bytes");
		return false;
	}

	// pack 1 pixel/byte rows into 8 pixels/byte
	if (priv->format == MONOCHROME)
		png_set_packing(priv->png_ptr);

	return true;
}

bool PNGWriterEx::writePointers(unsigned char **rowPointers, int rowCount)
{
	png_write_image(priv->png_ptr, rowPointers);
	/* write bytes */
	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "Error during writing bytes");
		return false;
	}

	return true;
}

bool PNGWriterEx::writeRow(unsigned char **row)
{
	// Write the row to the file
	png_write_rows(priv->png_ptr, row, 1);
	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "error during png row write");
		return false;
	}

	return true;
}

bool PNGWriterEx::close()
{
	/* end write */
	png_write_end(priv->png_ptr, priv->info_ptr);
	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "Error during end of write");
		return false;
	}

	return true;
}

bool PNGWriterEx::writeImgFile(SplashBitmap* bitmap, FILE *f, int hDPI, int vDPI) 
{
	int width = bitmap->getWidth();
	int height = bitmap->getHeight();
	SplashColorMode mode = bitmap->getMode();

	if (mode != splashModeXBGR8)
	{
		return false;
	}

	if (!init(f, width, height, hDPI, vDPI)) 
	{
		return false;
	}

	switch (mode) 
	{
	case splashModeXBGR8:
		{
			unsigned char* row = new unsigned char[4 * width];
			SplashColorPtr pBits = bitmap->getDataPtr();

			for (int y = 0; y < height; y++)
			{
				SplashColorPtr pSrc = pBits + y * bitmap->getRowSize();
				unsigned char* ptr = row;
				// Convert into a PNG row
				for (int x = 0; x < width; x++)
				{
					*ptr++ = pSrc[x * 4 + 2];
					*ptr++ = pSrc[x * 4 + 1];
					*ptr++ = pSrc[x * 4];
					*ptr++ = pSrc[x * 4 + 3];
				}

				if (!writeRow(&row)) 
				{
					delete[] row;
					return false;
				}
			}
			delete[] row;
		}
		break;
	}

	if (!close())
		return false;

	return true;
}
/*
bool PNGWriterEx::initPDFWrite(FILE *f, int width, int height, PBYTE pPalette, int iPaletteCount, int iBitsPerComp, int iNComps)
{
#if PNG_LIBPNG_VER < 10500
	png_charp icc_data_ptr = (png_charp)priv->icc_data;
#else
	png_const_bytep icc_data_ptr = (png_const_bytep)priv->icc_data;
#endif

	priv->png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!priv->png_ptr) {
		error(errInternal, -1, "png_create_write_struct failed");
		return false;
	}

	priv->info_ptr = png_create_info_struct(priv->png_ptr);
	if (!priv->info_ptr) {
		error(errInternal, -1, "png_create_info_struct failed");
		return false;
	}

	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "png_jmpbuf failed");
		return false;
	}

	priv->png_ptr->io_ptr = (png_voidp)f;
	priv->png_ptr->write_data_fn = PngWriteDataToMemFile;

	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "Error during writing header");
		return false;
	}

	// Set up the type of PNG image and the compression level
	png_set_compression_level(priv->png_ptr, Z_NO_COMPRESSION);

	// Silence silly gcc
	png_byte bit_depth = iBitsPerComp;
	png_byte color_type = -1;

	if(iNComps == 1 && !pPalette)
		color_type = PNG_COLOR_TYPE_GRAY;
	else if(iNComps == 1 && pPalette && iPaletteCount)
		color_type = PNG_COLOR_TYPE_PALETTE;
	else if(iNComps == 3)
		color_type = PNG_COLOR_TYPE_RGB;
	else if(iNComps == 4)
		color_type = PNG_COLOR_TYPE_RGB_ALPHA; // actually CMYK but we should pass it as this

	png_byte interlace_type = PNG_INTERLACE_NONE;

	png_set_IHDR(priv->png_ptr, priv->info_ptr, width, height, bit_depth, color_type, interlace_type, PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

	png_set_pHYs(priv->png_ptr, priv->info_ptr, (png_uint_32)(72/0.0254), (png_uint_32)(72/0.0254), PNG_RESOLUTION_METER);

	if (priv->icc_data)
		png_set_iCCP(priv->png_ptr, priv->info_ptr, priv->icc_name, PNG_COMPRESSION_TYPE_BASE, icc_data_ptr, priv->icc_data_size);
	else if (priv->sRGB_profile)
		png_set_sRGB(priv->png_ptr, priv->info_ptr, PNG_sRGB_INTENT_RELATIVE);

	png_write_info(priv->png_ptr, priv->info_ptr);
	if (setjmp(png_jmpbuf(priv->png_ptr))) {
		error(errInternal, -1, "error during writing png info bytes");
		return false;
	}

	// pack 1 pixel/byte rows into 8 pixels/byte
	if (iBitsPerComp == 1)
		png_set_packing(priv->png_ptr);

	return true;
}

bool PNGWriterEx::writePDFToPNGFile(Stream* strIn, FILE *fOut, int width, int height, GfxImageColorMap *colorMap)
{
	GfxColorSpace* pSpace = colorMap->getColorSpace();
	int len, iBPC = colorMap->getBits();
	unsigned char* buf = strIn->toUnsignedChars(&len);
	int iNComps = colorMap->getNumPixelComps();
	PBYTE pbyPal = NULL;
	int iPalCount = 0;

	if(pSpace->getMode() == csIndexed)
	{
		GfxIndexedColorSpace* pIndexed = (GfxIndexedColorSpace*)pSpace;
		int h = pIndexed->getIndexHigh();
		iPalCount = h + 1; // The high or max index plus one is the number of colours in the palette
		//iBPC = h == 1 ? 1 : (h <= 3 ? 2 : (h <= 15 ? 4 : (h <= 255 ? 8 : 0)));
		pbyPal = pIndexed->getLookup();
		iNComps = pIndexed->getBase()->getNComps();

		if(iNComps == 4)
			for(int i = 0; i < iPalCount; i++)
			{
				pbyPal[i*4+3] = 255-pbyPal[i*4+3];
			}
	}
	else if(pSpace->getMode() != csCalRGB

	if (!initPDFWrite(f, width, height, pbyPal, iPalCount, iBPC, iNComps)) 
	{
		return false;
	}

	switch (mode) 
	{
	case splashModeXBGR8:
		{
			unsigned char* row = new unsigned char[4 * width];
			SplashColorPtr pBits = bitmap->getDataPtr();

			for (int y = 0; y < height; y++)
			{
				SplashColorPtr pSrc = pBits + y * bitmap->getRowSize();
				unsigned char* ptr = row;
				// Convert into a PNG row
				for (int x = 0; x < width; x++)
				{
					*ptr++ = pSrc[x * 4 + 2];
					*ptr++ = pSrc[x * 4 + 1];
					*ptr++ = pSrc[x * 4];
					*ptr++ = pSrc[x * 4 + 3];
				}

				if (!writeRow(&row)) 
				{
					delete[] row;
					return false;
				}
			}
			delete[] row;
		}
		break;
	}

	if (!close())
		return false;

	return true;
}
*/