/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#ifndef XarOutputDev_H
#define XarOutputDev_H

#ifdef USE_GCC_PRAGMAS
#pragma interface
#endif

#include "stdafx.h"
#include <string>
#include <iostream>
#include <map>
#include <list>
#include <tuple>
#include "OutputDev.h"
#include "GfxState.h"
#include "GlobalParams.h"

#include "../../common/XFTree/XFTree.h"
#include "../../common/AffineTransform.h"
#include "../../poppler/poppler/PDFDoc.h"
#include "../../poppler/splash/SplashPath.h"
#include "../../poppler/poppler/Outline.h"

#include "clipper/clipper.hpp"

#include <xarlib.h>
#include "common/quadtree.h"
#include "LinkProps.h"
#include "PathHelpers.h"
#include "Fonts.h"

#include "rapidjson/document.h"
#include "rapidjson/writer.h"
#include "rapidjson/stringbuffer.h"

inline GfxColorSpace *GetBaseColorSpace(GfxColorSpace *pSpace)
{
	while (true)
	{
		switch (pSpace->getMode())
		{
		case csICCBased:
			pSpace = ((GfxICCBasedColorSpace *)pSpace)->getAlt();
			break;
		case csIndexed:
			pSpace = ((GfxIndexedColorSpace *)pSpace)->getBase();
			break;
		case csSeparation:
			pSpace = ((GfxSeparationColorSpace *)pSpace)->getAlt();
			break;
		case csDeviceN:
			pSpace = ((GfxDeviceNColorSpace *)pSpace)->getAlt();
			break;
		default:
			return pSpace;
		}
	}
}

typedef std::pair<double, VElementRect*> CoveragePair;
typedef std::vector<CoveragePair> CoveragePairVector;

// small utility class for determining most commonly used attribute values
template<class T> class SeriesCollector
{
public:
	SeriesCollector() :needUpdateCommon(false) {}
	SeriesCollector(const T &default) :needUpdateCommon(false), common(default) {}
	void Register(const T &);
	const T &operator()(const T &default) const;
	const T &operator()() const;
	bool IsDefined() const {return !usage.empty();}
	void Clear() { usage.clear(); needUpdateCommon = false; }

private:
	std::map<T,int> usage;
	mutable T       common;
	mutable bool    needUpdateCommon;
};

class AnnotationContext
{
public:
	AnnotationContext(PDFDoc* pdfDoc) : m_pdfDoc(pdfDoc), m_pAnnots(nullptr), state(nullptr), m_bDefURLPropertiesEvaluated(false), m_bUnderlineURL(false), m_pURLColour(0) {}
	~AnnotationContext() { Reset(); }

	bool Create(Page* pPage, int iWidth, int iHeight);
	bool AddElement(VXfBounded* pNode, int page = -1);
	CQuadtree& GetTree(int page = -1);

	/////////////////////////////////////////////
	// Annotations/hyperlinks
	bool ProcessAnnotations(PDFDoc* pdfDoc, class XarOutputDev *pOutDev);
	bool ProcessAnnotationsQuadTree(int iCurrentPage, PDFDoc* pdfDoc, VXfTreeNode* pDoc);
	VXfBounded* FindObject(const LinkDest* pDest);
	void AddLinkPair(const LinkDest* pDest, VXfBounded* pSrcObject);
	std::wstring GetNextNamedLink() { return FormatStringW(L"PDF.NamedLink.%d", m_szNextNameID++); }
	std::wstring AssignLink(const LinkDest* pDest, VXfBounded* pSrcObject = nullptr);

	// table of contents
	void ImportOutlineItem(PDFDoc* pdfDoc, OutlineItem *pItem, rapidjson::GenericValue<rapidjson::UTF16<>> &jsonObj, rapidjson::GenericDocument<rapidjson::UTF16<>> &d);
	bool ProcessTOC(PDFDoc* pdfDoc, VXfTreeNode* pDoc);

	std::wstring GetPageName(int index) const;

	void ConvertURLColourHSV();

	std::set<CXfTextString *> m_links;

private:
	PDFDoc* m_pdfDoc;
	Annots* m_pAnnots;
	GfxState* state;
	std::map<int, int> m_pageToTreeIdx; // maps page number to m_tree index, as we can import only part of pages
	std::vector<CQuadtree> m_tree;
	size_t m_szNextNameID = 0;
	typedef std::pair<const LinkDest *, VXfBounded*> LinkPair; // pair: dest, source object
	std::map<int, std::vector<LinkPair>> m_ForwardLinkMap;

	void Reset()
	{
		if (state)
		{
			delete state;
			state = nullptr;
		}
		m_pAnnots = nullptr;
		m_links.clear();
	}
	bool HasTree(int page = -1) { return page < 0 ? m_tree.size() : m_pageToTreeIdx.find(page) != m_pageToTreeIdx.end(); }

	static bool ShouldApplyAnnotation(Rect& rAnnot, Rect& rcObject);
	static double CalcAnnotationCoverage(Rect& rAnnot, Rect& rcObject);
	static void CreateAnnotationCoverageList(std::vector<VElementRect*>& elem, Rect& rAnnot, CoveragePairVector& sortedList);

	typedef std::map<Ref, int> PageNumMap;
	PageNumMap m_PageNumMap;

	int GetPageNum(const LinkDest *pDest) const;

	//////////////////////////////////////////////////////////////
	// URL links functions
	bool EvaluateDefaultURLProperties(std::vector<VXfTextRenderable*>& textLinks, CLinkProps& out);
	void AddLink(std::vector<VXfTextRenderable*>& textLinks, VXfBounded* pObject);
	void RemoveURLProperties(VXfTextRenderable* pString);

	// Default URL colour/underline attributes
	bool m_bDefURLPropertiesEvaluated;
	bool m_bUnderlineURL;

	CXfComplexColor* m_pURLColour;
};

//------------------------------------------------------------------------
// XarOutputDev
//------------------------------------------------------------------------

class XarOutputDev: public OutputDev 
{
public:
	enum // Text rendering modes
	{
		eTextFill = 0,			// Fill text.
		eTextStroke = 1,		// Stroke text.
		eTextFillStroke = 2,	// Fill, then stroke text.
		eTextNone = 3,			// Neither fill nor stroke text (invisible).
		eTextFillClip = 4,		// Fill text and add to path for clipping (see above).
		eTextStrokeClip = 5,	// Stroke text and add to path for clipping.
		eTextFillStrokeClip = 6,// Fill, then stroke text and add to path for clipping.
		eTextClip = 7			//Add text to path for clipping.
	};
	// Here are shading types - for some reason not constants but just numbers in the poppler library. Define them:
	enum { 
		Shading_None=0,
		Shading_FunctionBased=1, 
		Shading_Axial=2, 
		Shading_Radial=3, 
		Shading_FreeformGouraudMeshes=4, 
		Shading_LatticeformGouraudMeshes=5, 
		Shading_CoonsPatchMeshes=6,
		Shading_TensorProductPatchMeshes=7
	};
	// And here are all function types, also just numbers in the poppler lib
	enum {
		Function_identity = -1,
		Function_sampled=0,
		Function_exponential=2,
		Function_stitching=3,
		Function_PostScript=4
	};

	// Constructor.
	XarOutputDev(const wchar_t* _filename, VXfTreeNode* pChapter, PDFDoc* pDoc, FontManager *pFontManager, AnnotationContext *pAnnots);

	// Destructor.
	virtual ~XarOutputDev();

	void SetGfx(Gfx *pGfx) {m_pGfx = pGfx;}
	void DrawType3Char(GfxState * state, double x, double y, CharCode code);

	//----- get info about output device

	// Does this device use tilingPatternFill()?  If this returns false,
	// tiling pattern fills will be reduced to a series of other drawing
	// operations.
	virtual bool useTilingPatternFill() { return false; }

	// Does this device use functionShadedFill(), axialShadedFill(), and
	// radialShadedFill()?  If this returns false, these shaded fills
	// will be reduced to a series of other drawing operations.
	virtual bool useShadedFills(int type)
	{
		switch(type)
		{
		case Shading_None:
		case Shading_FunctionBased:
		case Shading_Axial:
		case Shading_Radial:
		case Shading_FreeformGouraudMeshes:
		case Shading_LatticeformGouraudMeshes:
		case Shading_CoonsPatchMeshes:
		case Shading_TensorProductPatchMeshes: // We need this so we can render the mesh into the Splash bitmap

			return true;
		}
		return false;
	}

	virtual bool useFillColorStop() { return false; }
	// Does this device use upside-down coordinates?
	// (Upside-down means (0,0) is the top left corner of the page.)
	virtual bool upsideDown() { return false; }

	// Does this device use drawChar() or drawString()?
	virtual bool useDrawChar() { return g_settings.bTextAsOutline; }

 	// Does this device use drawForm()?  If this returns false,
	// form-type XObjects will be interpreted (i.e., unrolled).
	virtual bool useDrawForm() { return /*true*/false; }

	// Does this device use beginType3Char/endType3Char?  Otherwise,
	// text in Type 3 fonts will be drawn with drawChar/drawString.
	virtual bool interpretType3Chars() { return false/*true*/; }

	// Does this device require incCharCount to be called for text on
	// non-shown layers?
	virtual bool needCharCount() { return false; }

	//----- initialization and control

	// Start a page.
	virtual void startPage(int pageNum, GfxState *state, XRef *xref);

	// End a page.
	virtual void endPage();

	//----- save/restore graphics state
	virtual void saveState(GfxState *state);
	virtual void restoreState(GfxState *state);

	//----- update graphics state
	virtual void updateAll(GfxState *state);
	virtual void updateCTM(GfxState *state, double m11, double m12, double m21, double m22, double m31, double m32);
	virtual void updateLineDash(GfxState *state);
	virtual void updateFlatness(GfxState *state);
	virtual void updateLineJoin(GfxState *state);
	virtual void updateLineCap(GfxState *state);
	virtual void updateMiterLimit(GfxState *state);
	virtual void updateLineWidth(GfxState *state);
	virtual void updateStrokeAdjust(GfxState *state);
	virtual void updateFillColorSpace(GfxState *state);
	virtual void updateStrokeColorSpace(GfxState *state);
	virtual void updateFillColor(GfxState *state);
	virtual void updateStrokeColor(GfxState *state);
	virtual void updateBlendMode(GfxState *state);
	virtual void updateFillOpacity(GfxState *state);
	virtual void updateStrokeOpacity(GfxState *state);
	virtual void updateFillOverprint(GfxState *state);
	virtual void updateStrokeOverprint(GfxState *state);
	virtual void updateOverprintMode(GfxState *state);
	virtual void updateTransfer(GfxState *state);

	//----- update text state
	virtual void updateFont(GfxState *state);
	virtual void updateTextMat(GfxState * /*state*/);
	virtual void updateCharSpace(GfxState * /*state*/);
	virtual void updateRender(GfxState * /*state*/);
	virtual void updateRise(GfxState * /*state*/);
	virtual void updateWordSpace(GfxState * /*state*/);
	virtual void updateHorizScaling(GfxState * /*state*/);
	virtual void updateTextPos(GfxState * /*state*/);
	virtual void updateTextShift(GfxState * /*state*/, double /*shift*/);
	virtual void saveTextPos(GfxState * /*state*/);
	virtual void restoreTextPos(GfxState * /*state*/);

	//----- path painting
	virtual void stroke(GfxState *state);
	virtual void fill(GfxState *state);
	virtual void eoFill(GfxState *state);
	virtual bool tilingPatternFill(GfxState *state, Gfx *gfx, Catalog *catalog, Object *str,
		const double *pmat, int paintType, int tilingType, Dict *resDict,
		const double *mat, const double *bbox,
		int x0, int y0, int x1, int y1,
		double xStep, double yStep);
	virtual bool functionShadedFill(GfxState* state, GfxFunctionShading* shading);
	virtual bool axialShadedFill(GfxState *state, GfxAxialShading *shading, double tMin, double tMax);
	virtual bool radialShadedFill(GfxState *state, GfxRadialShading *shading, double tMin, double tMax);
	virtual bool gouraudTriangleShadedFill(GfxState *state, GfxGouraudTriangleShading *shading);
	virtual bool patchMeshShadedFill(GfxState *state, GfxPatchMeshShading *shading);

	//----- path clipping
	virtual void clip(GfxState *state);
	virtual void eoClip(GfxState *state);
	virtual void clipToStrokePath(GfxState *state);

	//----- text drawing
	virtual void beginStringOp(GfxState * /*state*/);
	virtual void endStringOp(GfxState * /*state*/);
	virtual void beginString(GfxState * /*state*/, const GooString * /*s*/);
	virtual void endString(GfxState * /*state*/);
	virtual void drawChar(GfxState * /*state*/, double /*x*/, double /*y*/,
			double /*dx*/, double /*dy*/,
			double /*originX*/, double /*originY*/,
			CharCode /*code*/, int /*nBytes*/, Unicode * /*u*/, int /*uLen*/);
	virtual void drawString(GfxState * /*state*/, const GooString * /*s*/);
	virtual bool beginType3Char(GfxState * /*state*/, double /*x*/, double /*y*/,
			       double /*dx*/, double /*dy*/,
			       CharCode /*code*/, Unicode * /*u*/, int /*uLen*/);
	virtual void endType3Char(GfxState * /*state*/);
	virtual void beginTextObject(GfxState * /*state*/);
	virtual void endTextObject(GfxState * /*state*/);
	virtual void incCharCount(int /*nChars*/);
	virtual void beginActualText(GfxState * /*state*/, const GooString * /*text*/ );
	virtual void endActualText(GfxState * /*state*/);

	//----- image drawing
	virtual void drawImageMask(GfxState *state, Object *ref, Stream *str,
		int width, int height, bool invert,
		bool interpolate, bool inlineImg);
	virtual void setSoftMaskFromImageMask(GfxState *state,
		Object *ref, Stream *str,
		int width, int height, bool invert,
		bool inlineImg, double *baseMatrix);
	virtual void unsetSoftMaskFromImageMask(GfxState *state, double *baseMatrix);
	virtual void drawImage(GfxState *state, Object *ref, Stream *str,
		int width, int height, GfxImageColorMap *colorMap,
		bool interpolate, int *maskColors, bool inlineImg);
	virtual void drawMaskedImage(GfxState *state, Object *ref, Stream *str,
		int width, int height,
		GfxImageColorMap *colorMap,
		bool interpolate,
		Stream *maskStr, int maskWidth, int maskHeight,
		bool maskInvert, bool maskInterpolate);
	virtual void drawSoftMaskedImage(GfxState *state, Object *ref, Stream *str,
		int width, int height,
		GfxImageColorMap *colorMap,
		bool interpolate,
		Stream *maskStr,
		int maskWidth, int maskHeight,
		GfxImageColorMap *maskColorMap,
		bool maskInterpolate);

	//----- Type 3 font operators
	virtual void type3D0(GfxState *state, double wx, double wy);
	virtual void type3D1(GfxState *state, double wx, double wy,
		double llx, double lly, double urx, double ury);

	//----- form XObjects
	virtual void drawForm(Ref id);
	virtual void beginForm(Stream *str, GfxState * state);
	virtual void endForm();

	//----- transparency groups and soft masks
	virtual bool checkTransparencyGroup(GfxState *state, bool knockout);
	virtual void beginTransparencyGroup(GfxState *state, const double *bbox,
		GfxColorSpace *blendingColorSpace,
		bool isolated, bool knockout,
		bool forSoftMask);
	virtual void endTransparencyGroup(GfxState *state);
	virtual void paintTransparencyGroup(GfxState *state, const double *bbox);
	virtual void setSoftMask(GfxState *state, const double *bbox, bool alpha,
		Function *transferFunc, GfxColor *backdropColor);
	virtual void clearSoftMask(GfxState *state);
	///////////////////////////////////////////////////////////////////////////////////
	// Marked contents i.t. Acrobat layers
	virtual bool DisplayHiddenContents() { return !g_settings.bNoLayers; }
	std::wstring GetLayerName(Dict *dict);
	void BeginLayer(const wchar_t *name, bool bVisible);
	virtual void BeginOptionalContent(Object *dictRef, bool bVisible);
	virtual void EndOptionalContent();

	void BeginAnnotations();
	void AddToAnnotationContext(VXfBounded* pNode);

	// Xar node creation methods

	VXfTreeNode* GetCurrentPage() const { return m_pPage; }
	VXfTreeNode* GetDefLayer() const { return m_pDefLayer; }

	bool DoFill(GfxState* state, BYTE byWindingRule);
	void DoClip(GfxState* state, bool bClipToStroke, BYTE byWindingRule);
	CXfPath* DoDrawImage(GfxState *state, Object *ref, Stream *str,
		int width, int height, GfxImageColorMap *colorMap,
		bool interpolate, int *maskColors, bool inlineImg, bool allowMerging, PDFRectangle *bbox = NULL);
	void AppendAttributes(VXfTreeNode* pNode, GfxState* state, bool bStroke, bool bFill, bool bFillOpacity = false, bool bStrokeColor = false);
	CXfComplexColor* CreateComplexColour(GfxColorSpace* pSpace, const GfxColor* pColor);
	//CXfPath* CreateXaraPath(GfxPath* path, GfxState* state, UINT uiTag);
	static void ConvertOpacity(double dOpacity, GfxBlendMode mode, BYTE& byTrans, BYTE& byType);
	VXfTreeNode* AppendAxialFill(VXfTreeNode* pNode, GfxState* state, GfxAxialShading* shading);
	bool CanConvertRadialFill(GfxRadialShading* shading);
	VXfTreeNode* AppendRadialFill(VXfTreeNode* pNode, GfxState* state, GfxRadialShading* shading);
	CXfDefineBitmap* CreateDefineBitMask(unsigned char* pbySrc, int width, int height, GfxImageColorMap* colorMap, int* maskColors);
	VXfBounded* EncloseInSoftMaskGroup(VXfBounded* pBounded);
	/////////////////////////////////////////////
	// Splash renderer methods
	bool CreateSplashRenderer(GfxState* state);
	void GrowBitmapObject(SplashBitmap* bitmap);
	bool InterpolateStitchingFunction(GfxUnivariateShading* shading, std::vector<double>& bounds);
	bool InterpolateExponential(GfxUnivariateShading* shading, ExponentialFunction* pExp, std::vector<double>& bounds, double dx1, double dx2);

	static bool IsTextStroked(GfxState* state);
	static bool IsTextFilled(GfxState* state);
	/////////////////////////////////////////////////////////////
	// Clean-up functions
	bool FindRemoveBottomWhiteRects(VXfTreeNode* pPage);

private:
	void renderDump(const char *func, int di = 0, bool newline = true)
	{
		static const std::string delim("::");
		static const size_t delimL = delim.length();
		static int indent = 0;

		std::string f(func);
		size_t delimI = f.find(delim);
		if (delimI != std::string::npos)
			f.erase(0, delimI+delimL);

		std::cout << "\t" << std::string(indent*4, ' ') << f << std::flush;
		//std::cout << " insert: " << m_pInsertTo;
		indent += di;

		if (newline)
			std::cout << std::endl;
	}

	template<typename T> void renderDumpArg(const char *func, int di = 0, bool newline = true, const T &pdVal = NULL)
	{
		renderDump(func,di,false);
		std::cout << " " << pdVal;
		if (newline)
			std::cout << std::endl;
	}

	template<> void renderDumpArg<std::wstring>(const char *func, int di, bool newline, const std::wstring &pdVal)
	{
		renderDump(func,di,false);
		std::wcout << " " << pdVal;
		if (newline)
			std::cout << std::endl;
	}

	const wchar_t* filename;
	PDFDoc*          m_pDoc;
	VXfTreeNode*	 m_pChapter;
	VXfTreeNode*	 m_pPage;
	VXfTreeNode*	 m_pInsertTo;
	Gfx*             m_pGfx;
	// Pages, layers
	int m_layerCount;
	std::list<std::pair<std::wstring, bool> > m_OCStack;
	VXfTreeNode *m_pDefLayer, *m_pCurLayer, *m_pBBoxLayer, *m_pTextCommentLayer;
	LONG m_XTextNew, m_YTextNew;
	double m_xBaseOld, m_yBaseOld;

	class ClipState
	{
		typedef std::list<std::pair<ClipperLib::Paths *, BYTE> > PathStack;
		PathStack m_paths; // These paths are already transformed via CTM so they are in Xara coordiate space
		bool m_bInfinity, m_bNeedRebuild;
		BYTE m_windingRule;
		ClipperLib::Paths m_polygons;

		void Init();
		void FreeClipPaths();
		void Rebuild();
		void Update(ClipperLib::Paths *polygons, BYTE windingRule);
	public:

		CXfClipviewController *m_pClipController;

		ClipState();
		~ClipState();

		void Reset(INT x, INT y, INT width, INT height);
		void Add(GfxPath *path, GfxState* state, BYTE windingRule);
		void Save();
		bool Restore();

		const ClipperLib::Paths *GetTopPath() const;
		bool Empty();

		const ClipperLib::Paths &GetPolygons();
		BYTE GetWindingRule();

		CXfPath* CreateXaraPath(UINT tag = TAG_PATH_FILLED_STROKED);
		CXfClipviewController *GetClipviewController(bool bCreate = false);
	} m_clip;
	CXfClipviewController *m_pTextClipController;

	std::list<CXfClipviewController*> m_clipControllers;
	std::list<CXfGroup*> m_formGroups;

	std::list<std::pair<VXfTreeNode *, GfxColorSpace *> > m_transparencyGroupStack;
	typedef std::list<CXfSoftMaskGroupTransparentFill *> SoftMaskStack;
	SoftMaskStack m_SoftMaskStack;
	////////////////////////////////////////
	// Splash renderer to bitmap - for mesh fills and other fills that need rasterising
	SplashBitmap* bitmap;
	Splash* splash;

	////////////////////////////////////////
	// Text handling members

	bool m_bType3Char;
	GfxColorSpace *m_type3FillColorSpace;
	GfxColorSpace *m_type3StrokeColorSpace;
	GfxColor m_type3FillColor;
	GfxColor m_type3StrokeColor;

	GfxColorSpace *GetFillColorSpace(GfxState *state);
	GfxColorSpace *GetStrokeColorSpace(GfxState *state);
	const GfxColor *GetFillColor(GfxState *state);
	const GfxColor *GetStrokeColor(GfxState *state);

	struct Margins
	{
		double left, bottom, right, top;

		Margins(double l = 0, double b = 0, double r = 0, double t = 0): left(l), bottom(b), right(r), top(t) {}
	};

	struct TextChar;

	struct Color
	{
		GfxColorSpaceMode colorSpaceMode;
		COLORREF          crRGB;
		union
		{
			GfxColorComp  gray;
			GfxCMYK       cmyk;
			GfxRGB        rgb;
		};

		Color(): colorSpaceMode(csDeviceGray), crRGB(0), gray(0)
		{
		}

		static Color MakeFromState(GfxColorSpace *pSpace, const GfxColor *pColor);
		CXfComplexColor *CreateComplexColour() const;
		bool operator<(const Color &) const;
		bool operator!=(const Color &) const;
	};

	std::map<Color, CXfComplexColor *> m_colorMap;

	CXfComplexColor *CreateComplexColour(const Color &color);

	struct Opacity
	{
		BYTE mode;
		BYTE value;
		bool operator<(const Opacity &) const;
		bool operator!=(const Opacity &) const;
		static Opacity MakeFromState(double dOpacity, GfxBlendMode mode);
		Opacity();
	};

	struct TextAttrs
	{
		FIXED16         aspectRatio;
		MappedFontDesc *pFontDesc;
		MILLIPOINT      miterLimit, fontSize, lineWidth;
		typedef std::vector<MILLIPOINT> Dash;
		Dash            dash;
		bool
		                isBold    : 1,
		                isItalic  : 1,
		                isStroked : 1,
		                isFilled  : 1;
		BYTE            lineJoin, lineCap;

		Color colorStroke, colorFill;
		Opacity opacityStroke, opacityFill;

		bool operator != (const TextAttrs &attrs) const
		{
			return isFilled != attrs.isFilled || isStroked != attrs.isStroked ||
			       (isFilled && (colorFill != attrs.colorFill || opacityFill != attrs.opacityFill)) ||
			       (isStroked && (colorStroke != attrs.colorStroke || opacityStroke != attrs.opacityStroke)) ||
			       lineJoin != attrs.lineJoin || lineCap != attrs.lineCap || miterLimit != attrs.miterLimit ||
			       lineWidth != attrs.lineWidth || dash != attrs.dash ||
			       aspectRatio != attrs.aspectRatio || fontSize != attrs.fontSize ||
			       isBold != attrs.isBold || isItalic != attrs.isItalic;
		}

		bool operator == (const TextAttrs &attrs) const
		{
			return !(*this != attrs);
		}

		TextAttrs()
			:lineJoin(CXF_JOIN_STYLE_BEVELLED), lineCap(CXF_CAP_STYLE_BUTT), lineWidth(0), miterLimit(4000)
			,isBold(false), isItalic(false), isFilled(false), isStroked(false), pFontDesc(NULL), fontSize(10000), aspectRatio(1)
		{}
	};

	struct TextAttrsCollector
	{
		SeriesCollector<BYTE>       lineJoin;
		SeriesCollector<BYTE>       lineCap;
		SeriesCollector<MILLIPOINT> lineWidth;
		SeriesCollector<MILLIPOINT> miterLimit;
		SeriesCollector<bool>       isFilled;
		SeriesCollector<bool>       isStroked;
		SeriesCollector<MILLIPOINT> fontSize;
		//SeriesCollector<MILLIPOINT> baseLineShift;
		SeriesCollector<LONG>       tracking;
		SeriesCollector<bool>       isBold;
		SeriesCollector<bool>       isItalic;
		SeriesCollector<FIXED16>    aspectRatio;
		SeriesCollector<Color>   colorStroke, colorFill;
		SeriesCollector<Opacity> opacityStroke, opacityFill;
		SeriesCollector<TextAttrs::Dash>    dash;
		SeriesCollector<CXfFontDef*>                   pFontDef;

		TextAttrsCollector()
			:lineJoin(CXF_JOIN_STYLE_BEVELLED), lineCap(CXF_CAP_STYLE_BUTT), lineWidth(0), miterLimit(4000)
			,isFilled(false), isStroked(false), isBold(false), isItalic(false)
			,pFontDef(NULL), fontSize(10000), /*baseLineShift(0),*/ tracking(0), aspectRatio(1) {}
	};

	struct GraphicsLine
	{
		Color   color;
		Opacity opacity;
		DocCoord           coords[4];
		CXfPath *          nodePath;

		bool TransformToRect(const AffineTransform &matrix, Rect &rect) const;
	};

	bool CreateGraphicsLine(CXfPath* pPath, GfxState* state, bool bStroke, GraphicsLine &out);

	struct TextChar
	{
		double x, y;
		MILLIPOINT  width, ascent, descent, verticalOffset;
		MILLIPOINT  baseLineShift;
		FIXED16     scriptScale, scriptOffset;
		LONG        tracking, kern;
		CXfFontDef *pFontDef;
		Rect        bbox; // this is a precise bbox - based on each glyph Y and height and excluding starting and trailing spaces
		bool
		     isUnderline             : 1,
		     isStrikethrough         : 1,
		     isTabStop               : 1,
		     isAutoListChar          : 1,
		     isAnchorPos             : 1,
		     isCacheValid            : 1;
		UINT32     c;

		std::wstring	   ligChars;
		UINT16             ligType;

		TextAttrs attrs;

		int GetUnicodeMask() const;

		bool IsSpace() const;
		bool IsHyphen() const;
		bool IsBreak() const;
		void ApplyScriptAttributes(VXfTreeNode *pTreeNode) const;
		MILLIPOINT GetFontSize() const;
		MILLIPOINT GetWidth(MILLIPOINT cachedWidth) const;
		MILLIPOINT GetWidth() const;
		MILLIPOINT GetWidth(wchar_t ch) const;
		MILLIPOINT GetFontAscent() const;
		MILLIPOINT GetFontDescent() const;
		MILLIPOINT GetTrackingWidth(LONG tracking) const;
		MILLIPOINT GetTrackingWidth() const;
		MILLIPOINT GetKernWidth(LONG kern) const;
		MILLIPOINT GetKernWidth() const;
		MILLIPOINT GetAdvance(const TextChar *pNextChar = nullptr) const;
		MILLIPOINT GetAutoKernSize(const TextChar *pNextChar) const;

		MILLIPOINT GetStrikethroughWidth() const;
		MILLIPOINT GetStrikethroughPos() const;


		Rect GetBBox(bool bPrecise = false) const;

		TextChar(wchar_t _c, double _x, double _y): c(_c), x(_x), y(_y), ligType(0), width(0), ascent(0), descent(0), verticalOffset(0), baseLineShift(0), tracking(0), kern(0), scriptOffset(0), scriptScale(1), pFontDef(NULL), isUnderline(false), isStrikethrough(false), isTabStop(false), isAutoListChar(false), isAnchorPos(false), isCacheValid(false) {}
	//private:
		double GetScaleY() const;
		double GetScaleX(double scaleY) const;
		double GetScaleX() const;

	private:
		void GetMetrics();
	};

	void SetAttributes(TextAttrs &attrs, GfxState *state);

	struct TextChars: public std::list<TextChar>
	{
		Rect bbox;
	};

	struct TextLine: public TextChars
	{
		MILLIPOINT     baseLine, ascent, descent, leftMargin, rightMargin, endMargin, maxFontSize, firstKern;

		UINT32         number, numberAlt, startNumber, level;
		std::wstring   numberingFormat, numberingFormatAlt;

		TextAttrsCollector attrs;

		void UpdateAscentDescent();

		MILLIPOINT GetUnderlinePos(const_iterator itBegin, const_iterator itEnd, MILLIPOINT &width) const;
		bool CheckUnderlineStrikethrough(const Rect &rect, const Color &color, const Opacity &opacity);

		static void GetText(const TextChars &chars, std::wstring &text);

		size_t GetListHeadCharCount() const;
		void ClearAutoListItemFlags();

		static void GetBoundingPolygons(const TextChars &chars, ClipperLib::Paths &polygons);
		void GetBoundingPolygons(ClipperLib::Paths &polygons) const;

		//MILLIPOINT GetWidth() const;
		MILLIPOINT GetMaxTrackingWidth(LONG tracking) const;
		void PositionChars();
		bool CheckWrapping(MILLIPOINT fitWidth, const TextLine *pNextLine, int nbsp = 0) const;
		MILLIPOINT GetMaxWidth(const TextLine *pNextLine, int nbsp = 0) const;
		bool RemoveKernings(MILLIPOINT fitWidth, MILLIPOINT justifyWidth, LONG tracking, LONG maxRelativeOffset, const TextLine *pNextLine, bool bCheckOnly);
		bool RemoveKernings(MILLIPOINT fitWidth, MILLIPOINT justifyWidth, LONG maxRelativeOffset, const TextLine *pNextLine, bool bCheckOnly = false);

		double GetAscentDescentRatio() const;

		LONG CalcKernReplacementTracking(const_iterator itBegin, const_iterator itEnd, bool isSpaces) const;
		LONG CalcKernReplacementTracking() const;
		bool ReplaceKernings(iterator itBegin, iterator itEnd, bool isSpaces);
		void ReplaceKernings();

		void RegisterAttributes();
		void RecognizeListItemHead();

		void AppendSpaces(const_iterator it, const TextChar &ref, size_t count = 1, bool bUpdateAscentDescent = true);
		void ReconstructWhiteSpace(const_iterator itPrev, const_iterator itNext);
		void ReconstructWhiteSpaces();

		bool Append(TextChars &chars);

		TextLine(): baseLine(0), ascent(0), descent(0), leftMargin(0), rightMargin(0), endMargin(0), maxFontSize(0), firstKern(0), number(0), numberAlt(0), startNumber(0), level(0) {}
	};

	typedef std::list<TextLine> TextLines;

	enum ALIGN
	{
		ALIGN_LEFT   = 0x1,
		ALIGN_CENTER = 0x2,
		ALIGN_RIGHT  = 0x4,
		ALIGN_FULL   = 0x8
	};

	struct Paragraph: public TextLines
	{
		bool isStoryBreak  : 1,
		     isColumnBreak : 1;
		UINT possibleAlign : 4;
		MILLIPOINT firstMargin, leftMargin, rightMargin, spaceBefore, spaceAfter;

#ifdef _ENABLE_RELATIVE_LINE_SPACING
		FIXED16 linespaceRatio;
#else
		MILLIPOINT linespacing;
#endif

		Paragraph *pPrevInFlow, *pNextInFlow;

		Paragraph *GetFirstInFlow();
		UINT GetAlign() const;
		MILLIPOINT GetMaxWidth() const;

		void ConvertSpaces(int nbsp = 0);

		Paragraph():
#ifdef _ENABLE_RELATIVE_LINE_SPACING
			linespaceRatio(1),
#else
			linespacing(0),
#endif
			firstMargin(0), leftMargin(0), rightMargin(0), spaceBefore(0), spaceAfter(0), possibleAlign(ALIGN_LEFT | ALIGN_CENTER | ALIGN_RIGHT | ALIGN_FULL), isStoryBreak(false), isColumnBreak(false), pPrevInFlow(NULL), pNextInFlow(NULL) {}
	};

	typedef std::list<Paragraph> Paragraphs;

	struct TextStory: public Paragraphs
	{
		AffineTransform matrix, matrixInv;
		CXfTextStoryComplex *nodeStory;
		bool bNeedClipview;
		ClipperLib::Paths clipPolygons;
		BYTE clipWindingRule;

		MILLIPOINT leftMargin, rightMargin;

		TextLines lines;

		TextAttrsCollector attrs;

		Margins margins;

		double leftMarginSum;
		double leftMarginSqrSum;

		Paragraphs::iterator Finalize(Paragraphs::iterator itBegin, Paragraphs::iterator itEnd);
		void RenderClipview(Paragraphs::iterator itBegin, Paragraphs::iterator itEnd);

		void DetectTabStops();
		void DetectParagraphs();

		double GetLeftMarginMean() const;
		double GetLeftMarginDeviation() const;

		TextStory(): nodeStory(NULL), bNeedClipview(false), clipWindingRule(CXF_WINDING_RULE_NONZERO), leftMargin(0), rightMargin(0), leftMarginSum(0), leftMarginSqrSum(0) {}

		static double GetLineSpaceRatio(const TextLine &l1, const TextLine &l2);
		static double GetLineSpaceRatio(const TextLines &lines);		

		static double GetAbsoluteLineSpacing(const TextLine &l1, const TextLine &l2, size_t m = 0);
		static double GetParagraphSpacing(const Paragraph &p1, const Paragraph &p2);

		static bool IsEndOfStatement(const TextLine &cur, const TextLine &next);

	private:
		bool RevertOneItemList();
	};
	typedef std::list<TextStory> TextStories;
	TextStories textStories; // deque because vector would reallocate TextStories while expanding

	void AddOptimizedAttributeStringTags(const TextChar &ch, const TextAttrsCollector &lineAttrs, VXfTreeNode *node, const TextChar *pListTextFirstChar);
	void AddOptimizedAttributeLineTags(const TextAttrsCollector &attrs, const TextAttrsCollector &lineAttrs, VXfTreeNode *node);
	void AddDefaultAttributeTagsForStory(const TextAttrsCollector &attrs, VXfTreeNode *node);

	// Shared data
	typedef std::map<Goffset, CXfDefineBitmap *> BitmapMap;
	BitmapMap m_bitmaps;
	UINT32 m_anchodedGroupTag;
	UINT32 m_textAreaID;

	TextChars textChars;
	AffineTransform textMatrix, textMatrixInv;
	double textXScaleNorm, textYScale, xText, yText, textShearFactor;

	typedef std::list<GraphicsLine> QueueUnderlineCheck;
	QueueUnderlineCheck queueUnderline;

	void   Append(TextStory& textStory, TextChars &chars);
	void   DetectUnderlineStrikethrough(TextStory& textStory, TextLine &line);
	void   DetectEmbeddedGraphics(TextStory& textStory, TextLine &line);

	void   FinalizeLine(TextStory& textStory, TextLine &line);

	void   RemoveTextMarkupPath(GraphicsLine &line);

	void   AddToUnderlineStrikethroughCheck(GraphicsLine &line);
	void   ClearUnderlineStrikethrough();

	void   TextReconstruct();
	void   TextReconstruct(TextChars &strElements);
	void   TextParseString(GfxState *state, const GooString *s, double scaleX, double scaleY, double xText, double yText, double shearFactor);
	void   OnEndPage();
	void   OnEndTextObject();
	void   SetParentNode(VXfTreeNode *pParent);
	void   TextBreakStory();	
	void   RenderToOptimizedTags(TextStory& textStory);

	static CXfClipviewController *CreateClipviewController(CXfPath* pPath, BYTE windingRule);
	bool   AddGraphics(CXfPath* pPath, CXfPath *pOutline = NULL, BYTE windingRule = CXF_WINDING_RULE_NONZERO, bool bClip = true, GfxState* state = NULL, bool bStroke = false);

	bool m_bFirstPage;

	VXfTreeNode *CreatePage(int pageNum, GfxState *state);


	FontManager *m_pFontManager;
	AnnotationContext *m_pAnnots;
public:

	struct BitmapData
	{
		BitmapData(): buf(NULL), cmykAlphaBuf(NULL), isCMYK(false), width(0), height(0), pbyPal(NULL), iPalCount(0), iBPC(0), iNComps(0)
		{
		}

		unsigned char* buf;
		int width;
		int height;
		PBYTE pbyPal;
		int iPalCount;
		int iBPC;
		int iNComps;

		bool isCMYK;
		unsigned char* cmykAlphaBuf;
	};

	struct BitmapMergeData: public BitmapData
	{
		BitmapMergeData(): left(0), right(0), top(0), bottom(0)
		{
		}

		double left, right, top, bottom;
	};

	double bitmapMergePixWidth, bitmapMergePixHeight;

	typedef std::list<BitmapMergeData> BitmapMergeList;
	BitmapMergeList bitmapMergeList;

	BitmapMergeData MergeBitmapData(BitmapMergeList &bitmaps);
	void MergeSmallBitmaps();

	CXfDefineICC_CMYK *m_cmykProfile;

	CXfDefineBitmap* CreateDefineBitmap(Stream* str, int width, int height, GfxImageColorMap* colorMap, int* maskColors = NULL, CXfDefineBitmap** pMask = NULL, BitmapData* bitmapData = NULL, Rect *pBounds = NULL);
	CXfDefineICC_CMYK *GetCMYKProfile();

private:
	void sAddTextCommentEx(int iColorIdx, LONG x, LONG y, const wchar_t* fmt, wchar_t* psz);
	void vaAddTextCommentEx(int iColorIdx, LONG x, LONG y, const wchar_t* fmt, va_list argList);
	void AddTextCommentEx(int iColorIdx, LONG x, LONG y, const wchar_t* fmt, ...);
	void AddTextComment(LONG x, LONG y, const wchar_t* fmt, ...);
	void AddTextCommentLine(LONG x, LONG y, const wchar_t* fmt, ...);

	void _Debug_FlowLine(VXfTreeNode *layer, double xStart, double yStart, double xEnd, double yEnd, bool isNewStory);
};

#endif
