/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#include "stdafx.h"
#include "FloatQuad.h"

#include <algorithm>

using std::max;
using std::min;

static inline float min4(float a, float b, float c, float d)
{
    return min(min(a, b), min(c, d));
}

static inline float max4(float a, float b, float c, float d)
{
    return max(max(a, b), max(c, d));
}

inline float dot(const FloatSize& a, const FloatSize& b)
{
    return a.width() * b.width() + a.height() * b.height();
}

inline bool isPointInTriangle(const FloatPoint& p, const FloatPoint& t1, const FloatPoint& t2, const FloatPoint& t3)
{
    // Compute vectors        
    FloatSize v0 = t3 - t1;
    FloatSize v1 = t2 - t1;
    FloatSize v2 = p - t1;
    
    // Compute dot products
    float dot00 = dot(v0, v0);
    float dot01 = dot(v0, v1);
    float dot02 = dot(v0, v2);
    float dot11 = dot(v1, v1);
    float dot12 = dot(v1, v2);

    // Compute barycentric coordinates
    float invDenom = 1.0f / (dot00 * dot11 - dot01 * dot01);
    float u = (dot11 * dot02 - dot01 * dot12) * invDenom;
    float v = (dot00 * dot12 - dot01 * dot02) * invDenom;

    // Check if point is in triangle
    return (u >= 0) && (v >= 0) && (u + v <= 1);
}

FloatRect FloatQuad::boundingBox() const
{
    float left   = min4(m_p1.x(), m_p2.x(), m_p3.x(), m_p4.x());
    float top    = min4(m_p1.y(), m_p2.y(), m_p3.y(), m_p4.y());

    float right  = max4(m_p1.x(), m_p2.x(), m_p3.x(), m_p4.x());
    float bottom = max4(m_p1.y(), m_p2.y(), m_p3.y(), m_p4.y());
    
    return FloatRect(left, top, right - left, bottom - top);
}

bool FloatQuad::isRectilinear() const
{
    return (m_p1.x() == m_p2.x() && m_p2.y() == m_p3.y() && m_p3.x() == m_p4.x() && m_p4.y() == m_p1.y())
        || (m_p1.y() == m_p2.y() && m_p2.x() == m_p3.x() && m_p3.y() == m_p4.y() && m_p4.x() == m_p1.x());
}

bool FloatQuad::containsPoint(const FloatPoint& p) const
{
    return isPointInTriangle(p, m_p1, m_p2, m_p3) || isPointInTriangle(p, m_p1, m_p3, m_p4);
} 

// Note that we only handle convex quads here.
bool FloatQuad::containsQuad(const FloatQuad& other) const
{
    return containsPoint(other.p1()) && containsPoint(other.p2()) && containsPoint(other.p3()) && containsPoint(other.p4());
}

bool IntersectionOfLineSegments(const FloatPoint& p1, const FloatPoint& p2, const FloatPoint& p3, const FloatPoint& p4, FloatPoint* out=NULL)
{
	float fUaNum = (p4.x()-p3.x())*(p1.y()-p3.y()) - (p4.y()-p3.y())*(p1.x()-p3.x());
	float fUaDenum = (p4.y()-p3.y())*(p2.x()-p1.x()) - (p4.x()-p3.x())*(p2.y()-p1.y());
	float fUbNum = (p2.x()-p1.x())*(p1.y()-p3.y()) - (p2.y()-p1.y())*(p1.x()-p3.x());

	if(fUaDenum == 0.0f) // Parallel case
		return false;

	float fUa = fUaNum/fUaDenum;
	float fUb = fUbNum/fUaDenum;

	if(out)
	{
		out->setX(p1.x() + fUa*(p2.x() - p1.x()));
		out->setY(p1.y() + fUa*(p2.y() - p1.y()));
	}

	return fUa >= 0.0f && fUa <= 1.0f && fUb >= 0.0f && fUb <= 1.0f;
}

bool FloatQuad::MeIntersectRect(FloatRect& r) const
{
	if(r.contains(p1()) || r.contains(p2()) || r.contains(p3()) || r.contains(p4()))
		return true;

	FloatQuad qr(r);

	bool bRes = IntersectionOfLineSegments(m_p1, m_p2, qr.p1(), qr.p2()) ||
				IntersectionOfLineSegments(m_p1, m_p2, qr.p2(), qr.p3()) ||
				IntersectionOfLineSegments(m_p1, m_p2, qr.p3(), qr.p4()) ||
				IntersectionOfLineSegments(m_p1, m_p2, qr.p4(), qr.p1());
	if(bRes)
		return true;
	bRes = IntersectionOfLineSegments(m_p2, m_p3, qr.p1(), qr.p2()) ||
		IntersectionOfLineSegments(m_p2, m_p3, qr.p2(), qr.p3()) ||
		IntersectionOfLineSegments(m_p2, m_p3, qr.p3(), qr.p4()) ||
		IntersectionOfLineSegments(m_p2, m_p3, qr.p4(), qr.p1());
	if(bRes)
		return true;
	bRes = IntersectionOfLineSegments(m_p3, m_p4, qr.p1(), qr.p2()) ||
		IntersectionOfLineSegments(m_p3, m_p4, qr.p2(), qr.p3()) ||
		IntersectionOfLineSegments(m_p3, m_p4, qr.p3(), qr.p4()) ||
		IntersectionOfLineSegments(m_p3, m_p4, qr.p4(), qr.p1());
	if(bRes)
		return true;
	bRes = IntersectionOfLineSegments(m_p4, m_p1, qr.p1(), qr.p2()) ||
		IntersectionOfLineSegments(m_p4, m_p1, qr.p2(), qr.p3()) ||
		IntersectionOfLineSegments(m_p4, m_p1, qr.p3(), qr.p4()) ||
		IntersectionOfLineSegments(m_p4, m_p1, qr.p4(), qr.p1());
	return bRes;
}
