/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#pragma once

#include "../../XarLib_vc7/doccoord.h"

//--------------------------------------------------------------------------
// Represents a dimension in a 2D coordinate system (integer coordinates)
//--------------------------------------------------------------------------

class Size
{
public:
    Size()
    {
        cx = cy = 0;
    }

    Size(IN const Size& size)
    {
        cx = size.cx;
        cy = size.cy;
    }

    Size(IN INT width1,
         IN INT height1)
    {
        cx = width1;
        cy = height1;
    }

    Size operator+(IN const Size& sz) const
    {
        return Size(cx + sz.cx,
                    cy + sz.cy);
    }

    Size operator-(IN const Size& sz) const
    {
        return Size(cx - sz.cx,
                    cy - sz.cy);
    }

    bool operator==(IN const Size& sz) const
    {
        return (cx == sz.cx) && (cy == sz.cy);
    }

	bool operator<(const Size& rhs) const
	{
		return cx+cy < rhs.cx + rhs.cy;
	}

    BOOL empty() const
    {
        return (cx == 0 && cy == 0);
    }

public:

    INT cx;
    INT cy;
};

//--------------------------------------------------------------------------
// Represents a location in a 2D coordinate system (integer coordinates)
//--------------------------------------------------------------------------

class Point
{
public:
	Point()
	{
		x = y = 0;
	}

	Point(IN const Point &point)
	{
		x = point.x;
		y = point.y;
	}

	Point(IN const Size& size)
	{
		x = size.cx;
		y = size.cy;
	}

	Point(IN INT x1,
		IN INT y1)
	{
		x = x1;
		y = y1;
	}

	Point operator+(IN const Point& point) const
	{
		return Point(x + point.x,
			y + point.y);
	}

	Point operator-(IN const Point& point) const
	{
		return Point(x - point.x,
			y - point.y);
	}

	bool operator==(IN const Point& point)
	{
		return (x == point.x) && (y == point.y);
	}

public:

	INT x;
	INT y;
};

//--------------------------------------------------------------------------
// Represents a rectangle in a 2D coordinate system (integer coordinates)
//--------------------------------------------------------------------------

class Rect
{
public:

    Rect()
    {
        x = y = width = height = 0;
    }

    Rect(IN INT x1,
         IN INT y1,
         IN INT width1,
         IN INT height1)
    {
        x = x1;
        y = y1;
        width = width1;
        height = height1;
    }

    Rect(IN const Point& location,
         IN const Size& size)
    {
        x = location.x;
        y = location.y;
        width = size.cx;
        height = size.cy;
    }

    void operator=(const Rect& other)
    {
		x = other.x;
		y = other.y;
		width = other.width;
		height = other.height;
    }

    Point location() const
    {
        return Point(x, y);
    }

	Size size() const
    {
		return Size(width, height);
    }

    INT left() const
    {
        return x;
    }

    INT top() const
    {
        return y;
    }

	INT _top() const // this is the upright version of a rectangle top
	{
		return y+height;
	}

    INT right() const
    {
        return x+width;
    }

    INT bottom() const
    {
        return y+height;
    }

	INT _bottom() const // this is the upright version of a rectangle bottom, i.e. in Xara space
	{
		return y;
	}

	Point center()
	{
		return Point(x+width/2, y+height/2);
	}

    bool IsEmptyArea() const
    {
        return (width <= 0) || (height <= 0);
    }

	bool empty() const { return width == 0 && height == 0; }
	bool isNull() const { return x == 0 && y == 0 && width == 0 && height == 0; }
	void clear() { x = 0; y = 0; width = 0; height = 0; }

    bool operator==(IN const Rect & rect) const
    {
        return x == rect.x &&
               y == rect.y &&
               width == rect.width &&
               height == rect.height;
    }

	bool FuzzyEqual(IN const Rect & rect, int iDelta) const
	{
		return	(abs(x - rect.x) < iDelta) &&
				(abs(y - rect.y) < iDelta) &&
				(abs(right() - rect.right()) < iDelta) &&
				(abs(bottom() - rect.bottom()) < iDelta);
	}

    bool Contains(IN INT x1,
                  IN INT y1) const
    {
        return x1 >= x && x1 < x+width &&
               y1 >= y && y1 < y+height;
    }

    bool Contains(IN const Point& pt) const
    {
        return Contains(pt.x, pt.y);
    }

    bool Contains(IN Rect& rect) const
    {
        return (x <= rect.x) && (rect.right() <= right()) &&
               (y <= rect.y) && (rect.bottom() <= bottom());
    }

    void Inflate(IN INT dx,
                 IN INT dy)
    {
        x -= dx;
        y -= dy;
        width += 2*dx;
        height += 2*dy;
    }

    void Inflate(IN const Point& point)
    {
        Inflate(point.x, point.y);
    }

    bool Intersect(IN const Rect& rect)
    {
        return Intersect(*this, *this, rect);
    }

    static bool Intersect(OUT Rect& c,
                          IN const Rect& a,
                          IN const Rect& b)
    {
        INT right = min(a.right(), b.right());
        INT bottom = min(a.bottom(), b.bottom());
        INT left = max(a.left(), b.left());
        INT top = max(a.top(), b.top());

        c.x = left;
        c.y = top;
        c.width = right - left;
        c.height = bottom - top;
        return !c.IsEmptyArea();
    }

    bool IntersectsWith(IN const Rect& rect) const
    {
        return (left() < rect.right() &&
                top() < rect.bottom() &&
                right() > rect.left() &&
                bottom() > rect.top());
    }

    static BOOL Union(OUT Rect& c,
                      IN const Rect& a,
                      IN const Rect& b)
    {
        INT r = max(a.right(), b.right());
        INT bt = max(a.bottom(), b.bottom());
        INT l = min(a.left(), b.left());
        INT t = min(a.top(), b.top());

        c.x = l;
        c.y = t;
        c.width = r - l;
        c.height = bt - t;
        return !c.IsEmptyArea();
    }

    VOID Offset(IN const Point& point)
    {
        Offset(point.x, point.y);
    }

    VOID Offset(IN INT dx,
                IN INT dy)
    {
        x += dx;
        y += dy;
    }

	static void InflateRectToContainPoint(Rect& r, int x, int y)
	{
		if (r.empty() && r.x == 0 && r.y == 0)
		{
			r.x = x;
			r.y = y;
			//r.setSize(FloatSize(1, 1));
			return;
		}
		if (x < r.x) {
			r.width = r.right() - x;
			r.x = x;
		} else {
			int w = x - r.x;// + 1;
			if (w > r.width)
				r.width = w;
		}
		if (y < r.y) {
			r.height = r.bottom() - y;
			r.y = y;
		} else {
			int h =  y - r.y;// + 1;
			if (h > r.height)
				r.height = h;
		}
	}

/******************************************************************************
   Function:    FindBezierBounds
   Author:      various artists
   Purpose:		Find bounds of a bezier segment. Note that this method is not
				optimal timewise - it runs through the bezier with a step of
				0.01 to find the bounds. If time of executions matters, we'd better
				to calc extremums of the bezier curve instead.
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
	static void FindBezierBounds(Rect& rBounds, const DocCoord P0, const DocCoord P1, const DocCoord P2, const DocCoord P3)
	{
		float A = (float)P3.x - 3 * (float)P2.x + 3 * (float)P1.x - (float)P0.x;
		float B = 3 * (float)P2.x - 6 * (float)P1.x + 3 * (float)P0.x;
		float C = 3 * (float)P1.x - 3 * (float)P0.x;
		float D = (float)P0.x;

		float E = (float)P3.y - 3 * (float)P2.y + 3 * (float)P1.y - (float)P0.y;
		float F = 3 * (float)P2.y - 6 * (float)P1.y + 3 * (float)P0.y;
		float G = 3 * (float)P1.y - 3 * (float)P0.y;
		float H = (float)P0.y;

		float x, y;
		float xMin = (float)INT_MAX;
		float yMin = (float)INT_MAX;
		float xMax = 0;
		float yMax = 0;

		for (float t = 0.0f; t <= 1.0f; t += 0.01f)
		{
			x = A * t * t * t + B * t * t + C * t + D;
			if (x < xMin)
				xMin = x;
			if (x > xMax)
				xMax = x;
			y = E * t * t * t + F * t * t + G * t + H;
			if (y < yMin)
				yMin = y;
			if (y > yMax)
				yMax = y; 
		}

		rBounds.x = (LONG)xMin;
		rBounds.y = (LONG)yMin;
		rBounds.width = (LONG)(xMax-xMin);
		rBounds.height = (LONG)(yMax-yMin);
	}

	static void inflateRectToContainBezier(Rect& r, const DocCoord lastPoint, const DocCoord p0, const DocCoord p1, const DocCoord p2)
	{
		Rect rMe;
		FindBezierBounds(rMe, lastPoint, p0, p1, p2);
		Rect::Union(r, r, rMe);
	}

public:
	INT x;
	INT y;
	INT width;
	INT height;
};

// This is a simple virtual class to obtain a rect from any bounded element
class VElementRect
{
public:
	virtual void GetRect(int& x1, int& y1, int& x2, int& y2) = 0;
};
