/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#include "stdafx.h"

#include "XFTree.h"
#include <atlconv.h>
#include <assert.h>

/******************************************************************************
   Function:    CXfXaraTree::CXfXaraTree
   Author:      Paul Gruzdev
   Purpose:		CXfXaraTree constructor
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
CXfXaraTree::CXfXaraTree() : m_pLast(0), m_pActualFontDef(0)
{
	m_ParentStack.push_back(&m_pRoot);
	m_AttrContext.push_back(CXfAttributeContext());
	m_TextContext.push_back(CXfTextContext());
}

void CXfXaraTree::Delete()
{
	for(size_t i = 0; i < m_pRoot.ChildrenCount(); i++)
		delete m_pRoot.GetChild(i);
	m_pRoot.ClearChildren();
}
/******************************************************************************
   Function:    CXfXaraTree::HandleRecord
   Author:      Paul Gruzdev
   Purpose:		Handle record on export to build a Xara document tree
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
BOOL CXfXaraTree::HandleRecord(CXaraFileRecord* pRecord)
{
	switch(pRecord->GetTag())
	{
	case TAG_DOWN:
		{
			if(m_pLast) // save the parent node in the stack
				m_ParentStack.push_back(m_pLast);
			m_AttrContext.push_back(m_AttrContext.back()); // Clone the attribute context
			m_TextContext.push_back(m_TextContext.back()); // Clone the text context

			// Special treatment of a style def: we ignore all attributes that are global to the style. We only
			// incorporate and store attributes found INSIDE the style def, so we need to reset it here
			if(m_pLast && m_pLast->GetTag() == TAG_STYLE_DEF)
			{
				m_TextContext.back().Reset();
			}
		}
		break;
	case TAG_UP:
		// Update the attr context of a renderable node
		if(m_ParentStack.back()->IsRenderable())
		{
			VXfRenderable* pRenderable = (VXfRenderable*)m_ParentStack.back();
			pRenderable->Context() = m_AttrContext.back();
		}
		// Update the text attr context of a text node
		if(m_ParentStack.back()->IsTextRenderable())
		{
			VXfTextRenderable* pRenderable = (VXfTextRenderable*)m_ParentStack.back();
			pRenderable->TextContext() = m_TextContext.back(); // Update the attr context of a renderable node
		}

		// Get rid of contexts
		m_AttrContext.pop_back();
		m_TextContext.pop_back();
		m_ParentStack.pop_back();
		break;
	default:
		{
			VXfTreeNode* pNewNode = CXfTreeNodeFactory::CreateTreeNode(*pRecord);
			if(pNewNode)
			{
				if(m_ParentStack.size())
				{
					VXfTreeNode* pParent = m_ParentStack.back();
					pNewNode->SetParent(pParent);
					pParent->AddChild(pNewNode);
				}

				m_pLast = pNewNode;

				if(pNewNode->IsRefNode())
					m_Refs[pRecord->GetRecordNumber()] = pNewNode;

				m_AttrContext.back().Update(pNewNode);
				m_TextContext.back().Update(pNewNode);

				switch(pNewNode->GetTag())
				{
				case TAG_FONT_DEF_TRUETYPE:
				case TAG_FONT_DEF_ATM:
					m_pActualFontDef = pNewNode;
					break;
				case TAG_FONT_GLYPH:
				case TAG_FONT_GLYPH_EXTRA_WEIGHT:
					if(m_pActualFontDef)
					{
						CXfFontDef* pFontDef = (CXfFontDef*)m_pActualFontDef;
						CXfFontGlyph* pGlyph = (CXfFontGlyph*)pNewNode;
						pFontDef->AddGlyph(pGlyph->CharCode, pGlyph, false);
					}
					break;
				case TAG_NAVBAR_START:
					m_AttrContext.back().m_bNavBar = true;
					break;
				case TAG_NAVBAR_END:
					m_AttrContext.back().m_bNavBar = false;
					break;
				case TAG_OBJECTBOUNDS:
					if(pNewNode->GetParent() && pNewNode->GetParent()->IsBounded())
					{
						CXfObjectBounds* p = (CXfObjectBounds*)pNewNode;
						((VXfRenderable*)pNewNode->GetParent())->SetBounds(Rect(p->BottomLeft->x, p->BottomLeft->y, 
							p->TopRight->x-p->BottomLeft->x, p->TopRight->y-p->BottomLeft->y));
					}
					break;
				case TAG_WIZOP:
					{
						CXfWizop* pWiz = (CXfWizop*)pNewNode;
						if(pWiz->Name.getstr() == L"ObjectName")
						{
							VXfTreeNode* pParent = pNewNode->GetParent();
							m_ObjectName[pWiz->Parameter.getstr()].push_back(pParent);
							if(pParent)
							{
								if(pWiz->IsHTMLText())
									pParent->SetHTMLText();
								else
									pParent->SetObjectName(pWiz->Parameter.getstr());
							}
						}
					}
					break;
				case TAG_TEXT_STORY_LINK_INFO:
					{
						CXfTextStorySimple* pStory = (CXfTextStorySimple*)pNewNode->GetParent();
						CXfTextStoryLinkInfo* pLinkInfo = (CXfTextStoryLinkInfo*)pNewNode;
						LONG lCurrent = pLinkInfo->ThisID;
						LONG lNext = pLinkInfo->NextInFlowID;

						if(m_PrevLink.find(lCurrent) != m_PrevLink.end())
						{
							CXfTextStorySimple* pPrev = (CXfTextStorySimple*)m_PrevLink[lCurrent];
							pPrev->m_pNextLink = pStory;
							pStory->m_pPrevLink = pPrev;
						}

						if(lNext != -1)
						{
							m_PrevLink[lNext] = pStory;
						}

					}
					break;
				default:
					;
				}
				// Update the attr context of a renderable node - we initialize it with the params collected before
				if(pNewNode->IsRenderable())
				{
					VXfRenderable* pRenderable = (VXfRenderable*)pNewNode;
					pRenderable->Context() = m_AttrContext.back();
				}
				// Update the text attr context of a text node
				if(pNewNode->IsTextRenderable())
				{
					VXfTextRenderable* pRenderable = (VXfTextRenderable*)pNewNode;
					pRenderable->TextContext() = m_TextContext.back(); // Update the attr context of a renderable node
				}
			}
		}
	}
	
	return TRUE;
}

/******************************************************************************
   Function:    VXfTreeNode::GetRoot
   Author:      Paul Gruzdev
   Purpose:		Find a root node
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
VXfTreeNode* VXfTreeNode::GetRoot()
{
	VXfTreeNode* pRoot = this;
	while(pRoot)
	{
		if(pRoot->GetParent())
			pRoot = pRoot->GetParent();
		else
			break;
	}

	return pRoot;
}

/******************************************************************************
   Function:	BoundingBoxUnite
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		bool
   Parameters:	Rect & rcBounds  
   Parameters:	Rect & rcAdd  
   Notes:       
******************************************************************************/
bool BoundingBoxUnite(Rect& rcBounds, Rect& rcAdd)
{
	bool bRes = false;
	if(rcBounds.empty())
	{
		bRes = true;
		rcBounds = rcAdd;
	}
	else if(!rcAdd.empty())
	{
		bRes = true;
		Rect::Union(rcBounds, rcBounds, rcAdd);
	}

	return bRes;
}

/******************************************************************************
   Function:	VXfTreeNode::FindParentBounded
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		VXfBounded*
   Notes:       
******************************************************************************/
VXfBounded* VXfTreeNode::FindParentBounded()
{
	VXfTreeNode* pRes = GetParent();
	while(pRes && !pRes->IsBounded())
		pRes = pRes->GetParent();
	return (VXfBounded*)pRes;
}

/******************************************************************************
   Function:	VXfTreeNode::FindParentNode
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		VXfTreeNode*
   Parameters:	UINT uiTag  
   Notes:       
******************************************************************************/
VXfTreeNode* VXfTreeNode::FindParentNode(UINT uiTag)
{
	VXfTreeNode* pRes = GetParent();
	while(pRes && pRes->GetTag() != uiTag)
		pRes = pRes->GetParent();
	return pRes;
}

/******************************************************************************
   Function:    VXfBounded::UpdateBBox
   Author:      Paul Gruzdev
   Purpose:		Update the bbox of us and all our parents
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfBounded::UpdateBBox(Rect& rcAdd)
{
	bool bRes = BoundingBoxUnite(m_rcBounds, rcAdd);

	VXfBounded* pParentBounded = FindParentBounded();
	if(pParentBounded)
		bRes = pParentBounded->UpdateBBox(m_rcBounds);

	return bRes;
}

void VXfBounded::GetRect(int& x1, int& y1, int& x2, int& y2)
{
	x1 = m_rcBounds.x;
	y1 = m_rcBounds.y;
	x2 = m_rcBounds.right();
	y2 = m_rcBounds.bottom();
}

/******************************************************************************
   Function:    CXfClipviewController::UpdateBBox
   Author:      Paul Gruzdev
   Purpose:		Update the bbox of us and all our parents
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool CXfClipviewController::UpdateBBox(Rect& rcAdd)
{
	Rect rcClip;
	// First of all - find a bbox of all our 'clip' objects - i.e. that come prior the TAG_CLIPVIEW child
	for(size_t i = 0; i < ChildrenCount(); i++)
	{
		if(GetChild(i)->GetTag() == TAG_CLIPVIEW)
			break;

		if(GetChild(i)->IsBounded())
			BoundingBoxUnite(rcClip, ((VXfBounded*)GetChild(i))->Bounds());
	}

	bool bRes = BoundingBoxUnite(m_rcObjects, rcAdd);
	if(!rcClip.empty())
	{
		if(!m_rcObjects.empty())
			Rect::Intersect(m_rcBounds, rcClip, m_rcObjects);
	}
	else
		m_rcBounds = m_rcObjects;

	if(GetParent() && GetParent()->IsBounded())
		bRes = ((VXfBounded*)GetParent())->UpdateBBox(m_rcBounds);

	return bRes;
}

/******************************************************************************
   Function:    CXfClipviewController::CalcObjectsRect
   Author:      Paul Gruzdev
   Purpose:		Controller - calc the underneath objedts rectangle
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool CXfClipviewController::CalcObjectsRect(Rect& rcObjects)
{
	UINT iIdx = 0;
	if(FindFirstNode(TAG_CLIPVIEW, false, &iIdx))
	{
		rcObjects.clear();
		for(size_t i = iIdx + 1; i < ChildrenCount(); i++)
		{
			if(GetChild(i)->IsBounded())
				BoundingBoxUnite(rcObjects, ((VXfBounded*)GetChild(i))->Bounds());
		}
		return true;
	}
	return false;
}

/******************************************************************************
   Function:    VXfTreeNode::UpdateContexts
   Author:      Paul Gruzdev
   Purpose:		When adding a child node manually, not via CXaraFileRecord, 
				but for import filters via AddChild we need to update our
				parent's attribute and text contexts - do it here
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
void VXfTreeNode::UpdateContexts(VXfTreeNode* pChild)
{
	if (IsRenderable())
	{
		VXfRenderable* pRenderable = (VXfRenderable*)this;
		pRenderable->Context().Update(pChild);

		if (pChild->IsRenderable())
		{
			VXfRenderable* pChildRenderable = (VXfRenderable*)pChild;
			pChildRenderable->Context() = pRenderable->Context();
			// We just copied the context. Now we need to run through children and update it
			if (pChild->HasChildren())
			{
				for (size_t i = 0; i < pChild->ChildrenCount(); i++)
					pChildRenderable->Context().Update(pChild->GetChild(i));
			}
		}
	}
	if (IsTextRenderable())
	{
		VXfTextRenderable* pRenderable = (VXfTextRenderable*)this;
		pRenderable->TextContext().Update(pChild);

		if (pChild->IsTextRenderable())
		{
			VXfTextRenderable* pChildRenderable = (VXfTextRenderable*)pChild;
			pChildRenderable->TextContext() = pRenderable->TextContext();
			// We just copied the context. Now we need to run through children and update it
			if (pChild->HasChildren())
			{
				for (size_t i = 0; i < pChild->ChildrenCount(); i++)
					pChildRenderable->TextContext().Update(pChild->GetChild(i));
			}
		}
	}
}

/******************************************************************************
   Function:    VXfTreeNode::RebuildContexts
   Author:      Paul Gruzdev
   Purpose:		When inserting a child node manually, not via CXaraFileRecord, 
				but for import filters via InsertChild we need to rebuild our
				parent's attribute and text contexts - do it here
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
void VXfTreeNode::RebuildContexts()
{
	if (IsRenderable())
	{
		VXfRenderable* pRenderable = (VXfRenderable*)this;
		for (size_t i = 0; i < ChildrenCount(); i++)
			pRenderable->Context().Update(GetChild(i));
	}
	if (IsTextRenderable())
	{
		VXfTextRenderable* pRenderable = (VXfTextRenderable*)this;
		for (size_t i = 0; i < ChildrenCount(); i++)
			pRenderable->TextContext().Update(GetChild(i));
	}
}

/******************************************************************************
   Function:    VXfTreeNode::AddChild
   Author:      Paul Gruzdev
   Purpose:		Add a child node
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
void VXfTreeNode::AddChild(VXfTreeNode* pChild, bool bUpdateBBox)
{
	if (!pChild) return;
	if (m_Childern.size())
	{
		pChild->SetPrev(m_Childern.back());
		m_Childern.back()->SetNext(pChild);
	}

	pChild->SetParent(this);

	if (bUpdateBBox && pChild->IsBounded())
	{
		VXfBounded* pChildBounded = (VXfBounded*)pChild;
		// Do update our bounds/move up recursively in UpdateBBox
		if (IsBounded())
		{
			VXfBounded* pThis = (VXfBounded*)this;
			pThis->UpdateBBox(pChildBounded->Bounds());
		}
		else
		{
			VXfBounded* pBounded = FindParentBounded();
			if (pBounded)
				pBounded->UpdateBBox(pChildBounded->Bounds());
		}
	}
	// Now find a nearest parent clipview controller and update it's objects polygon. 
	if (!IsClipperController())
	{
		CXfClipviewController* pController = (CXfClipviewController*)FindParentClipperController();
		if (pController)
			pController->UpdateObjectsPolygon(pChild);
	}

	UpdateContexts(pChild);
	Reserve();
	m_Childern.push_back(pChild); 
}

void VXfTreeNode::InsertFirstChild(VXfTreeNode* pChild, bool bUpdateBBox)
{
	if (m_Childern.size())
	{
		(*m_Childern.begin())->SetPrev(pChild);
		pChild->SetNext((*m_Childern.begin()));
	}

	if (bUpdateBBox && pChild->IsBounded())
	{
		VXfBounded* pChildBounded = (VXfBounded*)pChild;
		// Do update our bounds/move up recursively in UpdateBBox
		if (IsBounded())
		{
			VXfBounded* pThis = (VXfBounded*)this;
			pThis->UpdateBBox(pChildBounded->Bounds());
		}
		else
		{
			VXfBounded* pBounded = FindParentBounded();
			if (pBounded)
				pBounded->UpdateBBox(pChildBounded->Bounds());
		}
	}

	pChild->SetParent(this);

	Reserve();
	m_Childern.insert(m_Childern.begin(), pChild);// .push_back(pChild); 
	RebuildContexts();
}

VXfTreeNode* VXfTreeNode::InsertBefore(size_t iIdx, VXfTreeNode* pChild, bool bUpdateBBox)
{
	if (iIdx > m_Childern.size())
		return false;
	else if (iIdx == m_Childern.size())
		return AppendChild(pChild, bUpdateBBox);

	if (iIdx)
	{
		m_Childern[iIdx - 1]->SetNext(pChild);
		pChild->SetPrev(m_Childern[iIdx - 1]);
	}

	if (iIdx < m_Childern.size() )
	{
		m_Childern[iIdx]->SetPrev(pChild);
		pChild->SetNext(m_Childern[iIdx]);
	}

	pChild->SetParent(this);
	Reserve();
	m_Childern.insert(m_Childern.begin() + iIdx, pChild);
	RebuildContexts();

	if (bUpdateBBox && pChild->IsBounded())
	{
		VXfBounded* pChildBounded = (VXfBounded*)pChild;
		// Do update our bounds/move up recursively in UpdateBBox
		if (IsBounded())
		{
			VXfBounded* pThis = (VXfBounded*)this;
			pThis->UpdateBBox(pChildBounded->Bounds());
		}
		else
		{
			VXfBounded* pBounded = FindParentBounded();
			if (pBounded)
				pBounded->UpdateBBox(pChildBounded->Bounds());
		}
	}

	return pChild;
}

bool VXfTreeNode::IsSmartTableNode(VXfTreeNode* pNode)
{
	if (pNode->GetTag() != TAG_SMARTSHAPEGROUP && pNode->GetParent() && pNode->GetParent()->GetTag() == TAG_GROUP){
		if (VXfTreeNode* pGroupNode = pNode->FindParentNode(TAG_GROUP)){
			CXfSmartShapeGroup* pSmartShapeGroup = (CXfSmartShapeGroup*)pGroupNode->FindFirstNode(TAG_SMARTSHAPEGROUP, false);
			if (pSmartShapeGroup && pSmartShapeGroup->strType.getstr() == L"SmartTable"){
				return true;
			}
		}
	}
	return false;
}

VXfTreeNode * VXfTreeNode::InsertAfterChild(VXfTreeNode *pAfter, VXfTreeNode *pChild)
{
	if (m_Childern.empty())
	{
		AppendChild(pChild);
		return pChild;
	}

	if (!pAfter)
	{
		InsertFirstChild(pChild);
		return pChild;
	}

	Reserve();

	for (VXfTreeNodeVector::iterator it=m_Childern.begin(); it != m_Childern.end(); ++it)
		if (pAfter == *it)
		{
			if ((*it)->NextSibling())
			{
				pChild->SetNext((*it)->NextSibling());
				(*it)->NextSibling()->SetPrev(pChild);
			}
			(*it)->SetNext(pChild);
			pChild->SetParent(this);
			pChild->SetPrev(*it);
			m_Childern.insert(++it, pChild);
			return pChild;
		}

	if(pChild->IsBounded())
	{
		VXfBounded* pChildBounded = (VXfBounded*)pChild;
		// Do update our bounds/move up recursively in UpdateBBox
		if(IsBounded())
		{
			VXfBounded* pThis = (VXfBounded*)this;
			pThis->UpdateBBox(pChildBounded->Bounds());
		}
		else
		{
			VXfBounded* pBounded = FindParentBounded();
			if(pBounded)
				pBounded->UpdateBBox(pChildBounded->Bounds());
		}
	}

	m_Childern.back()->SetNext(pChild);
	pChild->SetPrev(m_Childern.back());
	m_Childern.push_back(pChild);
	return pChild;
}

/******************************************************************************
   Function:    VXfTreeNode::RemoveChild
   Author:      Paul Gruzdev
   Purpose:		Recursively remove a child node. TODO: we also need to rebuild
				all our contexts in order to keep them up to date.
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::RemoveChild(UINT iIdx, bool bDelete)
{
	if(iIdx >= 0 && iIdx < m_Childern.size())
	{
		VXfTreeNode* pRemove = m_Childern[iIdx];
		VXfTreeNode* pPrev = pRemove->PrevSibling();
		VXfTreeNode* pNext = pRemove->NextSibling();
		if(pPrev) pPrev->SetNext(pNext);
		if(pNext) pNext->SetPrev(pPrev);
		m_Childern.erase(m_Childern.begin() + iIdx);
		if(bDelete)
			delete pRemove;
		return true;
	}

	return false;
}

/******************************************************************************
   Function:    VXfTreeNode::RemoveChild
   Author:      Paul Gruzdev
   Purpose:		Recursively remove a child node
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::RemoveChild(VXfTreeNode* pChild, bool bDelete)
{
	UINT uiIdx = 0;
	for(; uiIdx < m_Childern.size(); uiIdx++)
	{
		if(m_Childern[uiIdx] == pChild)
			break;
	}

	return RemoveChild(uiIdx, bDelete);
}

/******************************************************************************
   Function:    VXfTreeNode::MoveChildrenTo
   Author:      Paul Gruzdev
   Purpose:		Move children from this node to the target node, 
				removing them from us. This is useful if we want to change
				the parent node tag, and so on, say from TAG_GROUP to 
				TAG_SOFTMASK_GROUP
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::MoveChildrenTo(VXfTreeNode* pTarget)
{
	bool bRes = m_Childern.size() > 0;

	if(bRes)
	{
		for(size_t i = 0; i < m_Childern.size(); i++)
			pTarget->AppendChild(m_Childern[i]);

		m_Childern.clear();
	}
	return bRes;
}

/******************************************************************************
   Function:    VXfTreeNode::FindFirstNode
   Author:      Paul Gruzdev
   Purpose:		Find the first node that matches the tag specified
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
VXfTreeNode* VXfTreeNode::FindFirstNode(UINT uiTag, bool bRecursive, UINT* puiIdx)
{
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(pChild->GetTag() == uiTag)
		{
			if(puiIdx) *puiIdx = (UINT)i;
			return pChild;
		}

		if(bRecursive)
		{
			VXfTreeNode* pResult = pChild->FindFirstNode(uiTag, true, puiIdx);
			if(pResult)
				return pResult;
		}
	}

	return NULL;
}

/******************************************************************************
   Function:    VXfTreeNode::FindFirstNode
   Author:      Paul Gruzdev
   Purpose:		Find the first node that matches the set of tags specified
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
VXfTreeNode* VXfTreeNode::FindFirstNode(CTagSet Tags, bool bRecursive, UINT* puiIdx)
{
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(Tags.find(pChild->GetTag()) != Tags.end())
		{
			if(puiIdx) *puiIdx = (UINT)i;
			return pChild;
		}

		if(bRecursive)
		{
			VXfTreeNode* pResult = pChild->FindFirstNode(Tags, true, puiIdx);
			if(pResult)
				return pResult;
		}
	}

	return NULL;
}

/******************************************************************************
   Function:    VXfTreeNode::FindLastNode
   Author:      Paul Gruzdev
   Purpose:		Find the last node that matches a set of tags specified
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
VXfTreeNode* VXfTreeNode::FindLastNode(CTagSet Tags, bool bRecursive, UINT* puiIdx)
{
	for(int i = (int)m_Childern.size()-1; i >= 0; i--)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(Tags.find(pChild->GetTag()) != Tags.end())
		{
			if(puiIdx) *puiIdx = (UINT)i;
			return pChild;
		}

		if(bRecursive)
		{
			VXfTreeNode* pResult = pChild->FindLastNode(Tags, true, puiIdx);
			if(pResult)
				return pResult;
		}
	}

	return NULL;
}

/******************************************************************************
   Function:    VXfTreeNode::FindNodes
   Author:      Paul Gruzdev
   Purpose:		Find all nodes that match the specified tag
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::FindNodes(UINT uiTag, VXfTreeNodeVector& result, bool bRecursive)
{
	bool bRes = false;
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(pChild->GetTag() == uiTag)
			result.push_back(pChild);

		if(bRecursive)
			bRes = pChild->FindNodes(uiTag, result, bRecursive);
	}

	return result.size() > 0;
}

/******************************************************************************
   Function:	VXfTreeNode::NextNode
   Author:		Paul Gruzdev
   Purpose:     For continuous enumeration of all nodes, like in Xar stream
   Returns:		VXfTreeNode*
   Parameters:	bool bRecursive  
   Notes:       
******************************************************************************/
VXfTreeNode* VXfTreeNode::NextNode(bool bRecursive)
{
	// First of all, if recursive - go down one level and enum children
	if(bRecursive && m_Childern.size())
		return GetChild(0);
	// If non-recursive or we have next - return it
	if(!bRecursive || m_pNext) return m_pNext;
	// Here we are: both in recursive mode and no next node - we have to go up as many levels as we find the next sibling
	VXfTreeNode* pParent = GetParent();
	while(pParent && !pParent->m_pNext)
		pParent = pParent->GetParent();

	return pParent ? pParent->m_pNext : NULL;
}

/******************************************************************************
Function:		VXfTreeNode::FindFirstRenderableNode
Author:			Paul Gruzdev
Purpose:		Find first renderable node. Those are the actual objects to be
				rendered
Returns:
Exceptions:
Parameters:
Notes:
******************************************************************************/
VXfTreeNode* VXfTreeNode::FindFirstRenderableNode(bool bRecursive, bool bExcludeClipPath, bool bExcludeLayer, bool bIgnoreRemoved)
{
	for (size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if (pChild->IsRenderable())
		{
			if (!((VXfRenderable*)pChild)->IsClippingPath() || !bExcludeClipPath)
			{
				if (pChild->GetTag() != TAG_LAYER || !bExcludeLayer)
				{
					if (!pChild->IsRemoved() || !bIgnoreRemoved)
						return pChild;
				}
			}
		}

		if (bRecursive)
		{
			VXfTreeNode* pChildRes = pChild->FindFirstRenderableNode(bRecursive, bExcludeClipPath, bExcludeLayer, bIgnoreRemoved);
			if (pChildRes) return pChildRes;
		}
	}

	return NULL;
}

/******************************************************************************
   Function:	VXfTreeNode::FindFirstWebaddressNode
   Author:		Paul Gruzdev
   Purpose:     
   Returns:		VXfTreeNode*
   Parameters:	bool bRecursive  
   Parameters:	bool bIgnoreRemoved  
   Notes:       
******************************************************************************/
VXfTreeNode* VXfTreeNode::FindFirstWebaddressNode(bool bRecursive, bool bIgnoreRemoved)
{
	for (size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if (pChild->IsWebAddress())
		{
			if (!pChild->IsRemoved() || !bIgnoreRemoved)
				return pChild;
		}

		if (bRecursive)
		{
			VXfTreeNode* pChildRes = pChild->FindFirstWebaddressNode(bRecursive, bIgnoreRemoved);
			if (pChildRes) return pChildRes;
		}
	}

	return NULL;
}

/******************************************************************************
   Function:    VXfTreeNode::FindRenderableNodes
   Author:      Paul Gruzdev
   Purpose:		Find all renderable nodes. Those are the actual objects to be
				rendered
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::FindRenderableNodes(VXfTreeNodeVector& result)
{
	bool bRes = false;
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(pChild->IsRenderable())
			result.push_back(pChild);

		bRes = pChild->FindRenderableNodes(result);
	}

	return result.size() > 0;
}

/******************************************************************************
   Function:    VXfTreeNode::FindTextRenderableNodes
   Author:      Paul Gruzdev
   Purpose:		Find all text renderable nodes. TAG_TEXT_CHAR, TAG_TEXT_STRING*
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::FindTextRenderableNodes(VXfTreeNodeVector& result)
{
	bool bRes = false;
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(pChild->IsTextRenderable())
			result.push_back(pChild);

		bRes = pChild->FindTextRenderableNodes(result);
	}

	return result.size() > 0;
}

/******************************************************************************
   Function:    VXfTreeNode::FindPageNodes
   Author:      Paul Gruzdev
   Purpose:		Find all page nodes TAG_SPREAD*
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::FindPageNodes(VXfTreeNodeVector& result)
{
	bool bRes = false;
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(pChild->GetTag() == TAG_SPREAD || pChild->GetTag() == TAG_SPREAD_PHASE2)
			result.push_back(pChild);
		else // Don't go further in the hierarchy if we found our spread node on this level
			bRes = pChild->FindPageNodes(result);
	}

	return result.size() > 0;
}

/******************************************************************************
   Function:    VXfTreeNode::FindAnchoredGroup
   Author:      Paul Gruzdev
   Purpose:		Find an anchored group given it's ID
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
CXfGroup* VXfTreeNode::FindAnchoredGroup(UINT id)
{
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];
		if(pChild->GetTag() == TAG_ANCHORED_GROUP_INFO && ((CXfAnchoredGroupInfo*)pChild)->ImportTag.get() == id)
			return (CXfGroup*)pChild->GetParent();

		CXfGroup* pGroup = pChild->FindAnchoredGroup(id);
		if(pGroup) return pGroup;
	}

	return NULL;
}

/******************************************************************************
   Function:    VXfTreeNode::FindNodes
   Author:      Paul Gruzdev
   Purpose:		Find all nodes that match the specified tag
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::FindNodes(CTagSet Tags, VXfTreeNodeVector& result, bool bRecursive)
{
	bool bRes = false;
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];

		if(Tags.find(pChild->GetTag()) != Tags.end())
			result.push_back(pChild);

		if(bRecursive)
			bRes = pChild->FindNodes(Tags, result, bRecursive);
	}

	return result.size() > 0;
}

/******************************************************************************
   Function:    VXfTreeNode::FindNodesThatContainNodes
   Author:      Paul Gruzdev
   Purpose:		Find all nodes that match the specified tag
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::FindNodesThatContainNodes(CTagSet Tags, VXfTreeNodeVector& result, bool bRecursive, CTagSet tContain, bool bContainRecursive)
{
	bool bRes = false;
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];

		if(Tags.find(pChild->GetTag()) != Tags.end())
		{
			VXfTreeNodeVector v;
			pChild->FindNodes(tContain, v, bContainRecursive);
			if(v.size())
				result.push_back(pChild);
		}

		if(bRecursive)
			bRes = pChild->FindNodesThatContainNodes(Tags, result, bRecursive, tContain, bContainRecursive);
	}

	return result.size() > 0;
}

/******************************************************************************
   Function:    VXfTreeNode::FindNodesThatNotContainNodes
   Author:      Ekaterina Dolgunicheva
   Purpose:		Find all nodes that match the specified tag
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::FindNodesThatNotContainNodes(CTagSet Tags, VXfTreeNodeVector& result, bool bRecursive, CTagSet tContain, bool bContainRecursive)
{
	bool bRes = false;
	for(size_t i = 0; i < m_Childern.size(); i++)
	{
		VXfTreeNode* pChild = m_Childern[i];

		if(Tags.find(pChild->GetTag()) != Tags.end())
		{
			VXfTreeNodeVector v;
			pChild->FindNodes(tContain, v, bContainRecursive);
			if(!v.size())
				result.push_back(pChild);
		}

		if(bRecursive)
			bRes = pChild->FindNodesThatNotContainNodes(Tags, result, bRecursive, tContain, bContainRecursive);
	}

	return result.size() > 0;
}

/******************************************************************************
   Function:    VXfTextRenderable::ShouldApplyURLColor
   Author:      Paul Gruzdev
   Purpose:		Test this renderable node whether an URL colour should be applied
				over it or not. The rules are quite complex, see e-mail from
				Martin 'Re: Hyperlinks highlighting' received on 15 Feb 2013
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTextRenderable::ShouldApplyURLColor()
{
	CTagSet WebTag(TAG_WEBADDRESS, TAG_WEBADDRESS_BOUNDINGBOX, TAG_WEBADDRESS_PAGE, TAG_WEBADDRESS_PAGE_BOUNDINGBOX, 
		TAG_WEBADDRESS_OBJECT, TAG_WEBADDRESS_OBJECT_BOUNDINGBOX);
	// We assume that this is a renderable char node
	UINT uTag = GetTag();
	if(uTag != TAG_TEXT_STRING && uTag != TAG_TEXT_CHAR && uTag != TAG_TEXT_KERN)
		return false;
	// If we do not have a webaddress applied - don't even bother!
	if(!Context().m_pWebAddress)
		return false;

	VXfTreeNode* pStory = GetParent()->GetParent();
	bool bLocal = pStory->FindFirstNode(WebTag, true) != NULL;
	if(bLocal) return true;

	// Now test the controller
	VXfTreeNode* pController = pStory->GetParent();

	if(pController->FindFirstNode(WebTag, false) != NULL)
	{
		UINT uControllerTag = pController->GetTag();

		if(uControllerTag == TAG_SHADOWCONTROLLER || uControllerTag == TAG_BEVEL || uControllerTag == TAG_CONTOURCONTROLLER)
			return true;

		if(uControllerTag == TAG_GROUP && pController->FindFirstNode(TAG_SYMBOLGROUP, false))
			return true;
		// Controller is a group but not a symbol group - should not highlight!
		return false;
	}

	// Controller does not have web address attribute and it is not local, so it is higher, return false
	return false;
}

/******************************************************************************
   Function:    VXfTextRenderable::ShouldApplyURLUnderline
   Author:      Paul Gruzdev
   Purpose:		Test this renderable node whether an URL colour should be applied
				over it or not. The rules are quite complex, see e-mail from
				Martin 'Re: Hyperlinks highlighting' received on 15 Feb 2013
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTextRenderable::ShouldApplyURLUnderline()
{
	CTagSet WebTag(TAG_WEBADDRESS, TAG_WEBADDRESS_BOUNDINGBOX, TAG_WEBADDRESS_PAGE, TAG_WEBADDRESS_PAGE_BOUNDINGBOX, 
		TAG_WEBADDRESS_OBJECT, TAG_WEBADDRESS_OBJECT_BOUNDINGBOX);
	// We assume that this is a renderable char node
	UINT uTag = GetTag();
	if(uTag != TAG_TEXT_STRING && uTag != TAG_TEXT_CHAR && uTag != TAG_TEXT_KERN)
		return false;
	// If we do not have a webaddress applied - don't even bother!
	if(!Context().m_pWebAddress)
		return false;

	VXfTreeNode* pDocument = GetRoot()->FindFirstNode(TAG_DOCUMENT, false);
	// Check the underline url document setting
	if(!pDocument || !pDocument->FindFirstNode(TAG_UNDERLINE_URLS, false))
		return false;

	VXfTreeNode* pStory = GetParent()->GetParent();

	if(pStory->GetTag() == TAG_TEXT_STORY_COMPLEX)
	{
		CXfTextStoryComplex* pComplex = (CXfTextStoryComplex*)pStory;
		XaraMatrix& m = (XaraMatrix)(pComplex->Transform);

		ANGLE aRotation, aShear;
		m.Decompose(NULL, NULL, &aRotation, &aShear, NULL, NULL);
		// If the matrix is share or/and rotation, no underline
		if(aRotation.MakeDouble() != 0 || aShear.MakeDouble() != 0)
			return false;
	}
	else if(pStory->GetTag() != TAG_TEXT_STORY_SIMPLE)
	{
		// We should be in *some* text story, complex or simple
		assert(FALSE);
		return false;
	}

	// Now test if the link is 'local'
	VXfTreeNode* pController = GetParent()->GetParent()->GetParent();
	bool bLocal = pController->FindFirstNode(WebTag, true) != NULL;

	if(!bLocal)
		return false;

	CXfWizop* pWizop = (CXfWizop*)pStory->FindFirstNode(TAG_WIZOP, false);
	bool bHTMLText = pWizop && pWizop->Parameter.getstr() == L"HTMLText";

	if(bHTMLText)
		return true;

	for(VXfTreeNode* pParent = pStory->GetParent(); pParent != NULL;)
	{
		if(pParent->GetTag() == TAG_GROUP && !pParent->FindFirstNode(TAG_SYMBOLGROUP, false))
		{
			return false;
		}
		else if(pParent->GetTag() == TAG_SHADOWCONTROLLER || pParent->GetTag() == TAG_BEVEL || 
			pParent->GetTag() == TAG_CONTOURCONTROLLER || pParent->GetTag() == TAG_LIVE_EFFECT ||
			pParent->GetTag() == TAG_CLIPVIEWCONTROLLER)
			return false;

		pParent = pParent->GetParent();
	}

	return true;
}

/******************************************************************************
   Function:    VXfTreeNode::IsThisAnchoredBitmap
   Author:      Paul Gruzdev
   Purpose:		Find whether this is a regular bitmap that should be rendered,
				or this is a bitmap inside an anchored group that were already
				embedded inside the text.
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
bool VXfTreeNode::IsThisAnchoredBitmap()
{
	if(!IsNodeBitmap()) return false;
	VXfTreeNodeVector v;
	if(GetParent()->GetTag() == TAG_GROUP && GetParent()->FindNodes(TAG_ANCHORED_GROUP_INFO, v))
		return true;

	if(!GetObjectName().empty())
	{
		UINT ut, eType;
		if(CXfWizop::ParseObjectName(GetObjectName(), ut, eType) && (eType == CXfWizop::PrimaryGroup))
			return true;
	}
	return false;
}


/******************************************************************************
Function:    VXfTreeNode::IsThisAnchoredBitmap
Author:      Ekaterina Dolgunicheva
Purpose:		Find whether this is a regular bitmap that should be rendered,
or this is a bitmap inside an anchored group that were already
embedded inside the text. (USING a vector of TAG_WIZOP's)
Returns:
Exceptions:
Parameters:
Notes:
******************************************************************************/
bool VXfTreeNode::IsThisAnchoredBitmap(VXfTreeNodeVector& vWizOps){
	if (!IsNodeBitmap()) return false;
	VXfTreeNodeVector v;
	if (GetParent()->GetTag() == TAG_GROUP && GetParent()->FindNodes(TAG_ANCHORED_GROUP_INFO, v))
		return true;

	for (VXfTreeNodeVector::iterator it = vWizOps.begin(); it != vWizOps.end(); ++it){
		std::wstring strObjName = ((CXfWizop*)*it)->Parameter;

		if (!strObjName.empty()){
			UINT ut, eType;
			if (CXfWizop::ParseObjectName(strObjName, ut, eType) && (eType == CXfWizop::PrimaryGroup))
				return true;
		}
	}
	return false;
}

CXfClipviewController* VXfTreeNode::FindParentClipperController()
{
	VXfTreeNode* pParent = GetParent();
	while(pParent)
	{
		if(pParent->IsClipperController())
		{
			return (CXfClipviewController*)pParent;
		}
		pParent = pParent->GetParent();
	}
	return NULL;
}

bool CXfTextStoryComplex::IsTextStoryActuallySimple()
{
	Matrix mat = Transform;
	FIXED16 m[4];
	LONG ef[2];
	mat.GetComponents(m, ef);

	return fabs(m[0].MakeDouble()-1.0) < 0.01 && fabs(m[1].MakeDouble()) < 0.01 && fabs(m[2].MakeDouble()) < 0.01 &&
		fabs(m[3].MakeDouble()-1.0) < 0.01;
}

/////////////////////////////////////////////////////////////////
// Writing functions
/******************************************************************************
   Function:    CXfXaraTree::Write
   Author:      Paul Gruzdev
   Purpose:		A recursive tree writing routine
   Returns:     
   Exceptions:  
   Parameters:  
   Notes:       
******************************************************************************/
BOOL CXfXaraTree::Write(VXfTreeNode* pNode, CXarExport* pExporter, ULONG* pRecordNum)
{
	if(pNode->IsRemoved()) // If the node was removed, we won't save it and all of it's childern would we?
		return TRUE;

	ULONG ulRef;
	bool bRootNode = pNode->GetTag() == -1;
	// This assert fails if the node being written does not have a correct m_uiTag member set
	assert(bRootNode && (pNode->GetParent() == NULL) || !bRootNode);

	BOOL ok = !bRootNode ? pNode->Write(pExporter, &ulRef) : TRUE;
	if(!bRootNode) pNode->SetRef(ulRef); // set the correct reference after we've written the node

	if(pNode->HasChildren())
	{
		if(ok && !bRootNode) ok = pExporter->WriteZeroSizedRecord(TAG_DOWN);
		for(size_t i = 0; i < pNode->ChildrenCount(); i++)
		{
			ULONG ulChildRef;
			if(ok) ok = Write(pNode->GetChild(i), pExporter, &ulChildRef);
		}
		if(ok && !bRootNode) ok = pExporter->WriteZeroSizedRecord(TAG_UP);
	}

	return ok;
}
