#ifndef FONTMAPPER_NGRAM_H
#define FONTMAPPER_NGRAM_H

#include <string>
#include <vector>
#include <set>
#include <sstream>
#include <cctype>
#include <algorithm>
#include <math.h>

enum {NGRAM_INVALID_INDEX = std::string::npos};

std::vector<std::wstring> TokenizeEmbeddedFontName(const std::wstring &encoded/*, int *firstModTokenI*/)
{
	enum EMBEDDED_FONT_PARSE_STATE
	{
		STATE_DELIMITER = 1, 
		STATE_ELM_NUMBERS = 101, 
		STATE_ELM_ALPHA = 100, 
	};
	EMBEDDED_FONT_PARSE_STATE state = STATE_DELIMITER;
	std::vector<std::wstring> tokens;
	bool isCapital = false;
	//*firstModTokenI = -1;

	for (size_t i=0; i < encoded.length(); ++i)
	{
		const wchar_t c = encoded.at(i);
		if (c == L'~')
			break;
		
		switch (state)
		{
		case STATE_DELIMITER:
			//if (c == L'-' && *firstModTokenI < 0)
			//	*firstModTokenI = tokens.size();
			//else
			if (c >= L'0' && c <= L'9')
			{
				state = STATE_ELM_NUMBERS;
				tokens.push_back(std::wstring(1,c));
			}
			else if ((c >= L'A' && c <= L'Z') || (c >= L'a' && c <= L'z'))
			{
				state = STATE_ELM_ALPHA;
				isCapital = (c >= L'A' && c <= L'Z');
				tokens.push_back(std::wstring(1,c));
			}
			break;
		case STATE_ELM_NUMBERS:
			if (c >= L'0' && c <= L'9')
			{
				tokens[tokens.size()-1] += c;
			}
			else
			{
				state = STATE_DELIMITER;
				--i;
				continue;
			}
			break;
		case STATE_ELM_ALPHA:
			if ((c >= L'A' && c <= L'Z') || (c >= L'a' && c <= L'z'))
			{
				if (!isCapital && (c >= L'A' && c <= L'Z'))
				{
					state = STATE_DELIMITER;
					--i;
					continue;
				}

				if (isCapital && (c >= L'a' && c <= L'z'))
				{
					std::wstring &token = tokens[tokens.size()-1];
					if (token.size() > 1)
					{
						wchar_t oc = token[token.size()-1];
						token.resize(token.size()-1);
						tokens.push_back(std::wstring(1,oc));
						tokens[tokens.size()-1] += c;
						isCapital = false;
						continue;
					}
				}

				tokens[tokens.size()-1] += c;
				if (c >= L'a' && c <= L'z')
					isCapital = false;
			}
			else
			{
				state = STATE_DELIMITER;
				--i;
				continue;
			}
			break;
		}
	}
	return tokens;
}

struct Modifier
{
	std::vector<std::wstring> tokens;
	std::set<size_t>          incompatible;
	float                     penalty;
	float                     weight; // weight < 0 => weight unchanged
	bool                      strictCompare;

	Modifier() :penalty(0.), weight(-1.), strictCompare(false) {}

	typedef std::vector<Modifier> Modifiers;
	typedef std::map<std::wstring,size_t> ModifiersCache;
	static Modifiers modifiers;
	static ModifiersCache modifiersCache;
};

struct FontNameDesc
{
	std::vector<std::wstring> name;
	std::set<size_t> modifiers;
	float weight;
	std::wstring nameStrictCompare;

	FontNameDesc() :weight(.4f) {}

	int GetWeight() { return static_cast<int>(weight * 1000.f + .5f); }
	bool isItalic() { static const size_t i = Modifier::modifiersCache.find(L"italic")->second; return modifiers.find(i) != modifiers.end(); }

	bool InsertModifierCompatible(size_t modI)
	{
		for each (size_t oldModI in modifiers)
		{
			if (Modifier::modifiers[oldModI].incompatible.find(modI) != Modifier::modifiers[oldModI].incompatible.end())
				return false;
		}

		modifiers.insert(modI);
		
		if (Modifier::modifiers[modI].weight > 0.)
		{
			if (Modifier::modifiers[modI].weight > .4 && Modifier::modifiers[modI].weight > weight)
				weight = Modifier::modifiers[modI].weight;
			else
			if (Modifier::modifiers[modI].weight < .4 && Modifier::modifiers[modI].weight < weight)
				weight = Modifier::modifiers[modI].weight;
		}

		return true;
	}

	static float GetModifierDistance(const std::set<size_t> &op1, float weight1, const std::set<size_t> &op2, float weight2)
	{
		float out = 0.;
		bool foundIncompatible = false;
		for each (size_t modI in op1)
		{
			if (op2.find(modI) == op2.end())
				out += Modifier::modifiers[modI].penalty;

			for each (size_t modICompared in op2)
			{
				if (Modifier::modifiers[modI].incompatible.find(modICompared) != Modifier::modifiers[modI].incompatible.end())
				{
					foundIncompatible = true;
					break;
				}
			}
			if (foundIncompatible)
				break;
		}
		if (foundIncompatible)
			return -1.;

		return out + abs(weight1 - weight2);
	}

	static float ModifiersDistance(const std::set<size_t> &op1, float weight1, const std::set<size_t> &op2, float weight2)
	{
		float d1 = GetModifierDistance(op1,weight1,op2,weight2);
		if (d1 < 0)
			return -1.;
		float d2 = GetModifierDistance(op2,weight2,op1,weight1);
		if (d2 < 0)
			return -1.;
		return d1+d2;
	}

	std::wstring GetNameW() const
	{
		std::wstring out;
		bool firstToken = true;
		for each (const std::wstring &token in name)
		{
			if (firstToken)
				firstToken = false;
			else
				out += L' ';
			out += token;
		}
		return out;
	}

	std::wstring GetUnspacedNameW() const
	{
		std::wstring out;
		bool firstToken = true;
		for each (const std::wstring &token in name)
			out += token;

		return out;
	}

	std::wstring ToWstring() const
	{
		std::wostringstream out;
		out << L"'";
		bool firstToken = true;
		for each (const std::wstring &token in name)
		{
			if (firstToken)
				firstToken = false;
			else
				out << L' ';
			out << token;
		}
		out << L"' [w=" << (1000. * weight);
		if (!modifiers.empty())
			out << L", "; 
		bool firstModifier = true;
		for each (size_t modI in modifiers)
		{
			if (firstModifier)
				firstModifier = false;
			else
				out << L' ';
			out << Modifier::modifiers[modI].tokens[0];
		}
		out << L']';
		return out.str();
	}

	std::wstring GetMods() const
	{
		std::wostringstream out;
		out << L" ["; 
		bool firstModifier = true;
		for each (size_t modI in modifiers)
		{
			if (firstModifier)
				firstModifier = false;
			else
				out << L' ';
			out << Modifier::modifiers[modI].tokens[0];
		}
		out << L']';
		return out.str();
	}
};

void ModifiersInit();

FontNameDesc NormalizeEmbeddedFontName(const std::wstring &encoded)
{
	static bool modifiersInit = false;
	typedef std::set<std::wstring> Prefixes;
	static Prefixes prefixes;
	if (!modifiersInit)
	{
		modifiersInit = true;

		prefixes.insert(L"semi");
		prefixes.insert(L"demi");
		prefixes.insert(L"ultra");
		prefixes.insert(L"extra");
	}

	FontNameDesc out;

	std::vector<std::wstring> tokens = TokenizeEmbeddedFontName(encoded/*, &firstModTokenI*/);
	if (tokens.size() >= 3 && tokens[tokens.size()-2] == L"Identity" && tokens[tokens.size()-1] == L"H")
	{
		tokens.pop_back();
		tokens.pop_back();
	}

	int tokenI = 0;
	bool startedModifiers = false;
	std::wstring curToken;
	for each (std::wstring token in tokens)
	{
		std::transform(token.begin(), token.end(), token.begin(), std::tolower);
		if (!tokenI++)
		{
			out.name.push_back(token);
			out.name[out.name.size()-1][0] = std::toupper(out.name[out.name.size()-1][0]);
			out.nameStrictCompare += out.name.back();
			continue;
		}
		
		std::wstring oldToken = curToken;
		Prefixes::const_iterator prefixIt = prefixes.find(token);
		if (prefixIt != prefixes.end())
		{
			curToken = token;
			continue;
		}

		curToken += token;
		Modifier::ModifiersCache::const_iterator modIt = Modifier::modifiersCache.find(curToken);
		if (modIt == Modifier::modifiersCache.end())
		{
			// prefix + token -> failed, trying to add in parts
			if (!oldToken.empty())
			{
				modIt = Modifier::modifiersCache.find(oldToken);
				if (modIt != Modifier::modifiersCache.end())
				{
					// prefix
					startedModifiers = true;
					out.InsertModifierCompatible((*modIt).second);
				}
			}

			modIt = Modifier::modifiersCache.find(token);
			if (modIt != Modifier::modifiersCache.end())
			{
				// token is modifier, but prefix is not
				startedModifiers = true;
				out.InsertModifierCompatible((*modIt).second);
				
				if (Modifier::modifiers[(*modIt).second].strictCompare)
				{
					out.nameStrictCompare += L' ';
					out.nameStrictCompare += curToken;
					out.nameStrictCompare[out.nameStrictCompare.length()-curToken.length()] = std::toupper(out.nameStrictCompare[out.nameStrictCompare.length()-curToken.length()]);
				}
				curToken.clear();
				continue;
			}
			else
			{
				//if (!startedModifiers)
				{
					if (!oldToken.empty())
					{
						out.name.push_back(oldToken);
						out.name[out.name.size()-1][0] = std::toupper(out.name[out.name.size()-1][0]);
					}
					out.name.push_back(token);
					out.name[out.name.size()-1][0] = std::toupper(out.name[out.name.size()-1][0]);

					out.nameStrictCompare += L' ';
					out.nameStrictCompare += out.name.back();
				}
				curToken.clear();
				continue;
			}
		}
	
		if (Modifier::modifiers[(*modIt).second].strictCompare)
		{
			out.nameStrictCompare += L' ';
			out.nameStrictCompare += token;
			out.nameStrictCompare[out.nameStrictCompare.length()-token.length()] = std::toupper(out.nameStrictCompare[out.nameStrictCompare.length()-token.length()]);
		}

		startedModifiers = true;
		out.InsertModifierCompatible((*modIt).second);
		curToken.clear();
	}
	if (!curToken.empty())
	{
		Modifier::ModifiersCache::const_iterator modIt = Modifier::modifiersCache.find(curToken);
		if (modIt != Modifier::modifiersCache.end())
			out.InsertModifierCompatible((*modIt).second);
	}
	return out;
}

bool StrictCompare(const FontNameDesc &op1, const FontNameDesc &op2)
{
	if (fabs(op1.weight - op2.weight) > 0.01)
		return false;

	static const size_t modifiersCacheItalicI = Modifier::modifiersCache.find(L"italic")->second;
	bool isItalic   = op1.modifiers.find(modifiersCacheItalicI) != op1.modifiers.end();
	bool isItalicOp = op2.modifiers.find(modifiersCacheItalicI) != op2.modifiers.end();
	if (isItalic != isItalicOp)
		return false;

	if (op1.nameStrictCompare != op2.nameStrictCompare)
		return false;

	return true;
}


void ModifiersInit()
{
	Modifier mod;

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = 9.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"ultrablack");
	mod.tokens.push_back(L"ultra");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = .9f;
	mod.tokens.clear();
	mod.tokens.push_back(L"black");
	mod.tokens.push_back(L"dark");
	mod.tokens.push_back(L"ultraheavy");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = .8f;
	mod.tokens.clear();
	mod.tokens.push_back(L"heavy");
	mod.tokens.push_back(L"extra");
	mod.tokens.push_back(L"extrabold");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = false;
	mod.penalty = 0.f;
	mod.weight  = .7f;
	mod.tokens.clear();
	mod.tokens.push_back(L"bold");
	mod.tokens.push_back(L"bol");
	mod.tokens.push_back(L"bo");
	mod.tokens.push_back(L"bd");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = .6f;
	mod.tokens.clear();
	mod.tokens.push_back(L"semibold");
	mod.tokens.push_back(L"semibo");
	mod.tokens.push_back(L"semibd");
	mod.tokens.push_back(L"demibold");
	mod.tokens.push_back(L"demibo");
	mod.tokens.push_back(L"demi");
	mod.tokens.push_back(L"semi");
	mod.tokens.push_back(L"demibd");
	mod.tokens.push_back(L"smbd");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = false;
	mod.penalty = 0.f;
	mod.weight  = .5f;
	mod.tokens.clear();
	mod.tokens.push_back(L"medium");
	mod.tokens.push_back(L"medi");
	mod.tokens.push_back(L"med");
	mod.tokens.push_back(L"md");
	mod.tokens.push_back(L"mm");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = false;
	mod.penalty = 0.001f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"regular");
	mod.tokens.push_back(L"regu");
	mod.tokens.push_back(L"reg");
	mod.tokens.push_back(L"rg");
	mod.tokens.push_back(L"normal");
	mod.tokens.push_back(L"norm");
  //mod.tokens.push_back(L"roman");
  //mod.tokens.push_back(L"roma");
  //mod.tokens.push_back(L"rom");
	mod.tokens.push_back(L"plain");
	mod.tokens.push_back(L"pln");
	mod.tokens.push_back(L"standard");
	mod.tokens.push_back(L"std");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.001f;
	mod.weight  = .4f;
	mod.tokens.clear();
	mod.tokens.push_back(L"book");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = false;
	mod.penalty = 0.f;
	mod.weight  = .3f;
	mod.tokens.clear();
	mod.tokens.push_back(L"light");
	mod.tokens.push_back(L"ligh");
	mod.tokens.push_back(L"lgt");
	mod.tokens.push_back(L"lt");
	mod.tokens.push_back(L"ltstd");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = .1f;
	mod.tokens.clear();
	mod.tokens.push_back(L"thin");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = .2f;
	mod.tokens.clear();
	mod.tokens.push_back(L"ultralight");
	mod.tokens.push_back(L"extralight");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = .1f;
	mod.tokens.clear();
	mod.tokens.push_back(L"hairline");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = .5f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"capitals");
	mod.tokens.push_back(L"caps");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = .5f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"smallcaps");
	mod.tokens.push_back(L"small");
	mod.tokens.push_back(L"sm");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = .6f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"narrow");
	mod.tokens.push_back(L"narr");
	mod.tokens.push_back(L"nar");
	mod.tokens.push_back(L"condensed");
	mod.tokens.push_back(L"cond");
	mod.tokens.push_back(L"co");
	mod.tokens.push_back(L"cn");
	mod.tokens.push_back(L"compressed");
	mod.tokens.push_back(L"compress");
	mod.tokens.push_back(L"compr");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = .4f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"seminarrow");
	mod.tokens.push_back(L"seminarr");
	mod.tokens.push_back(L"seminar");
	mod.tokens.push_back(L"semicondensed");
	mod.tokens.push_back(L"semicond");
	mod.tokens.push_back(L"semico");
	mod.tokens.push_back(L"semicn");
	mod.tokens.push_back(L"semicompressed");
	mod.tokens.push_back(L"semicompress");
	mod.tokens.push_back(L"semicompr");
	mod.tokens.push_back(L"deminarrow");
	mod.tokens.push_back(L"deminarr");
	mod.tokens.push_back(L"deminar");
	mod.tokens.push_back(L"demicondensed");
	mod.tokens.push_back(L"demicond");
	mod.tokens.push_back(L"demico");
	mod.tokens.push_back(L"demicn");
	mod.tokens.push_back(L"demicompressed");
	mod.tokens.push_back(L"demicompress");
	mod.tokens.push_back(L"demicompr");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = .7f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"wide");
	mod.tokens.push_back(L"expanded");
	mod.tokens.push_back(L"expand");
	mod.tokens.push_back(L"expnd");
	mod.tokens.push_back(L"expn");
  //mod.tokens.push_back(L"extended");
  //mod.tokens.push_back(L"extend");
	mod.tokens.push_back(L"extnd");
	mod.tokens.push_back(L"extn");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = .5f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"semiwide");
	mod.tokens.push_back(L"semiexpanded");
	mod.tokens.push_back(L"semiexpand");
	mod.tokens.push_back(L"semiexpnd");
	mod.tokens.push_back(L"semiexpn");
	mod.tokens.push_back(L"semiextended");
	mod.tokens.push_back(L"semiextend");
	mod.tokens.push_back(L"semiextnd");
	mod.tokens.push_back(L"semiextn");
	mod.tokens.push_back(L"demiwide");
	mod.tokens.push_back(L"demiexpanded");
	mod.tokens.push_back(L"demiexpand");
	mod.tokens.push_back(L"demiexpnd");
	mod.tokens.push_back(L"demiexpn");
	mod.tokens.push_back(L"demiextended");
	mod.tokens.push_back(L"demiextend");
	mod.tokens.push_back(L"demiextnd");
	mod.tokens.push_back(L"demiextn");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = false;
	mod.penalty = 1.f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"italic");
	mod.tokens.push_back(L"ital");
	mod.tokens.push_back(L"itc");
	mod.tokens.push_back(L"ita");
	mod.tokens.push_back(L"it");
	mod.tokens.push_back(L"oblique");
	mod.tokens.push_back(L"obli");
	mod.tokens.push_back(L"obl");
	mod.tokens.push_back(L"ob");
	mod.tokens.push_back(L"slanted");
	mod.tokens.push_back(L"slant");
	mod.tokens.push_back(L"leftalic");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = .6f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"monospace");
	mod.tokens.push_back(L"monospac");
	mod.tokens.push_back(L"monotype");
	mod.tokens.push_back(L"monotyp");
	mod.tokens.push_back(L"monotp");
	mod.tokens.push_back(L"mono");
	mod.tokens.push_back(L"mon");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"proportional");
	mod.tokens.push_back(L"proportion");
	Modifier::modifiers.push_back(mod);
/*
	mod.strictCompare = true;
	mod.penalty = 0.;
	mod.weight  = -1.;
	mod.tokens.clear();
	mod.tokens.push_back(L"sans");
	mod.incompatible.clear();
	mod.incompatible.push_back(L"serif");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.;
	mod.weight  = -1.;
	mod.tokens.clear();
	mod.tokens.push_back(L"serif");
	mod.incompatible.clear();
	mod.incompatible.push_back(L"sans");
	Modifier::modifiers.push_back(mod);
*/
	mod.strictCompare = false;
	mod.penalty = .04f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"mt");
	mod.tokens.push_back(L"mtps");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = false;
	mod.penalty = 0.03f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"ps");
	mod.tokens.push_back(L"psmt");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.001f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"bt");
	Modifier::modifiers.push_back(mod);

	mod.strictCompare = true;
	mod.penalty = 0.f;
	mod.weight  = -1.f;
	mod.tokens.clear();
	mod.tokens.push_back(L"customsubst"); // modifier to distinguish substituted font faces (i.e. for mac)
	Modifier::modifiers.push_back(mod);

	// modifiers token cache
	size_t modI = 0;
	for (Modifier::Modifiers::const_iterator modIt = Modifier::modifiers.begin(); modIt != Modifier::modifiers.end(); ++modIt, ++modI)
	{
		for each (const std::wstring &token in (*modIt).tokens)
		{
			Modifier::modifiersCache.insert(std::make_pair(token,modI));
		}
	}

	/*
	#define DEFINE_INCOMPATIBLE_MODIFIERS(a,b) Modifier::modifiers[Modifier::modifiersCache.find(a)->second].incompatible.insert(Modifier::modifiersCache.find(b)->second)
	#define DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(a,b) DEFINE_INCOMPATIBLE_MODIFIERS(a,b); DEFINE_INCOMPATIBLE_MODIFIERS(b,a)
	DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(L"regular",L"italic");
	DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(L"capitals",L"smallcaps");
	DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(L"narrow",L"wide");
	DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(L"narrow",L"semiwide");
	DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(L"seminarrow",L"wide");
	DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(L"seminarrow",L"semiwide");
	DEFINE_INCOMPATIBLE_MODIFIERS_PAIR(L"monospace",L"proportional");
	*/
}

void GetNgrams(const std::vector<std::wstring> &v, const size_t N, std::map<std::wstring,float> &out, const float WEIGHT = 1.0)
{
	for each (const std::wstring &op in v)
	{
		if (op.length() <= N)
		{
			std::map<std::wstring,float>::iterator it = out.insert( std::make_pair(op,0.f) ).first;
			(*it).second += WEIGHT*op.length()/N;
			continue;
		}
		for (size_t i=0; i<=op.length()-N; ++i)
		{
			std::map<std::wstring,float>::iterator it = out.insert( std::make_pair(op.substr(i,N),0.f) ).first;
			(*it).second += WEIGHT;
		}
	}
}

template <class SortedCache, class SortedCacheElement>
void GetNgramsInd(const std::vector<std::wstring> &v, const size_t N, SortedCache &ngramCache, std::map<size_t,float> &out, const float WEIGHT = 1.0)
{
	for (std::vector<std::wstring>::const_iterator opIt = v.begin(); opIt != v.end(); ++opIt)
	{
		const std::wstring &op = *opIt;
		for (size_t i=0; i<=std::max<size_t>(0U,op.length()-N); ++i)
		{
			std::wstring opSub;
			float w = WEIGHT;
			bool isSmallNameElement = false;

			if (op.length() > N)
				opSub = op.substr(i,N);
			else
			{
				opSub = op;
				isSmallNameElement = true;
				w *= ((float) op.length()) / N;
			}

			SortedCacheElement elm;
			elm.name = opSub;
			SortedCache::const_iterator cacheIt = std::lower_bound(ngramCache.begin(), ngramCache.end(), elm);
			/*
			{
				size_t index;
				if (cacheIt == ngramCache.end())
					index = ngramCache.size();
				else
				{
					index = cacheIt->index;
					//advance 
					SortedCache::iterator tempCacheIt = ngramCache.begin();
					std::advance(tempCacheIt, std::distance(tempCacheIt, cacheIt));
					for (; tempCacheIt != ngramCache.end(); ++tempCacheIt)
						++tempCacheIt->index;
				}
				FontNgram ngramNew;
				ngramNew.name  = opSub;
				ngramNew.index = index;
				cacheIt = ngramCache.insert(ngramNew);
			}*/
			size_t cacheI = (cacheIt == ngramCache.end() || cacheIt->name != opSub) ? NGRAM_INVALID_INDEX : cacheIt->index;
			std::map<std::size_t,float>::iterator it = out.insert( std::make_pair(cacheI,0.f) ).first;
			(*it).second += w;

			if (isSmallNameElement)
				break;
		}
	}
}

float substring_distance( const std::map<std::wstring,float> &sampleSubs, const std::map<std::wstring,float> &baseSubs)
{
	if (sampleSubs.empty())
		return 1.;

	float sumSubL = 0;
	float sumSubFoundL = 0;

	for each (std::pair<std::wstring,float> sampleSub in sampleSubs)
	{
		float w = sampleSub.first.length()*sampleSub.second;
		sumSubL += w;

		std::map<std::wstring,float>::const_iterator it = baseSubs.find(sampleSub.first);
		if (it == baseSubs.end())
			continue;

		sumSubFoundL += w;
	}

	return ((float)(sumSubL-sumSubFoundL))/(sumSubL);
}

template <class SortedCache>
float substring_distanceInd( const std::map<size_t,float> &sampleSubs, const std::map<size_t,float> &baseSubs, const SortedCache &ngramCache, const size_t N)
{
	if (sampleSubs.empty())
		return 1.;

	float sumSubL = 0;
	float sumSubFoundL = 0;

	for (std::map<size_t,float>::const_iterator sampleSubIt = sampleSubs.begin(); sampleSubIt != sampleSubs.end(); ++sampleSubIt)
	{
		float w = sampleSubIt->first == NGRAM_INVALID_INDEX ? N*sampleSubIt->second : ngramCache[sampleSubIt->first].name.length()*sampleSubIt->second;
		sumSubL += w;
		if (sampleSubIt->first == NGRAM_INVALID_INDEX)
			continue;
		std::map<size_t,float>::const_iterator it = baseSubs.find(sampleSubIt->first);
		if (it == baseSubs.end())
			continue;

		sumSubFoundL += w;
	}

	//dbg += _dbg.str();
	return ((float)(sumSubL-sumSubFoundL))/(sumSubL);
}

#endif