#include "StdAfx.h"
#include "quadtree.h"

//
// class of elements in the NearestEdge priority queue
//
class PQElem 
{
	friend class CQuadtree;
private:
	bool IsEdge;    // true if elem represents an edge, false if a CQuadTreeNode
	double dist;    // distance to query point, the key of the PQ
	CQuadTreeNode* Block;// pointer to CQuadTreeNode if IsEdge==false
	int edge_idx;   // index in E of edge if IsEdge==true
public:
	PQElem(bool isedge = true, double d = -1.0, CQuadTreeNode* b = 0, int eid = -1)
		: IsEdge(isedge), dist(d), Block(b), edge_idx(eid) {}
	friend bool operator<(const PQElem& x, const PQElem& y)
	{
		return (x.dist > y.dist);
	}
	void operator=(const PQElem& RHS)
	{
		IsEdge = RHS.IsEdge;
		dist = RHS.dist;
		Block = RHS.Block;
		edge_idx = RHS.edge_idx;
	}
};        


//
// CQuadTreeNode constructor
//
CQuadTreeNode::CQuadTreeNode(int set_depth)
{
	_depth = set_depth;
	_parent = 0;
	_NE = 0;
	_NW = 0;
	_SW = 0;
	_SE = 0;
	_x = 0;
	_y = 0;
}

//
// CQuadtree constructor from width and height
//
inline LONG ilog2(float x)
{
	DWORD ix = (DWORD&)x;
	DWORD exp = (ix >> 23) & 0xFF;
	LONG log2 = LONG(exp) - 127;

	return log2;
}

void CQuadtree::Create(LONG width, LONG height, int set_k)
{
	_depth = ilog2((float)max(width, height));
	_k = set_k;
}

int CQuadtree::AddElement(VElementRect* pElem)
{   
	int x1, y1, x2, y2;
	CQuadTreeNode* Block;
	std::queue<CQuadTreeNode*> Q;

	if(_root == NULL)
	{
		// setup the root block of the quadtree
		try 
		{
			_root = new CQuadTreeNode(_depth);
		}
		catch (std::bad_alloc) 
		{
			return 1;
		}
	}

	//
	// load segment into the tree
	//    
	{
		// get the coordinates of the current segment
		pElem->GetRect(x1, y1, x2, y2);
// 		x1 = _g->E.x[0][i];
// 		y1 = _g->E.y[0][i];
// 		x2 = _g->E.x[1][i];
// 		y2 = _g->E.y[1][i];

		// enqueue the root of the tree
		Q.push(_root);

		// do while blocks are still queued
		while (Q.size() > 0) 
		{
			// pop the next block from the queue
			Block = Q.front();

			// do if this is a leaf node
			if (Block->_NE == 0)
			{
				// test if rect intersects with or is contained within block
				if (BoxContainsPoint(Block, x1, y1, x2, y2)) 
				{
					// add edge to the list for this block
					Block->_e.push_back(pElem);

					// split the block if we exceed the threshold
					if (Block->_e.size() > _k) 
					{
						if (Split(Block) != 0) 
						{
							return 1;
						}
					}
				}                
			}
			// do if non-leaf node
			else 
			{
				// test each of the four children against the bounding
				// box of the current edge, enqueue any matching blocks

				// NE
				if (BoxContainsPoint(Block->_NE, x1, y1, x2, y2)) {
					Q.push(Block->_NE);
				}
				// NW
				if (BoxContainsPoint(Block->_NW, x1, y1, x2, y2)) {
					Q.push(Block->_NW);
				}
				// SW
				if (BoxContainsPoint(Block->_SW, x1, y1, x2, y2)) {
					Q.push(Block->_SW);
				}
				// SE
				if (BoxContainsPoint(Block->_SE, x1, y1, x2, y2)) {
					Q.push(Block->_SE);
				}
			}
			// end switch on leaf or non-leaf node

			// dequeue the current block
			Q.pop();
		}
		// end while blocks still exist in Q        
	}
	// end of loop for each segment

	return 0;
} 

//
// Build - build Quadtree from the Graph g
//
/*
int CQuadtree::Build()
{   
	int i, x1, y1, x2, y2;
	CQuadTreeNode* Block;
	std::queue<CQuadTreeNode*> Q;

	// assert that we have a valid pointer to graph
	if (_g == 0) return 1;

	// setup the root block of the quadtree
	try {
		_root = new CQuadTreeNode(_depth);
	}
	catch (std::bad_alloc) {
		return 1;
	}

	//
	// load each segment into the tree
	//    
	for (i = 0; i < _g->Ecount; i++) {

		// assert that the queue is initially empty
		if (Q.size() != 0) {
			return 1;
		}

		// get the coordinates of the current segment
		x1 = _g->E.x[0][i];
		y1 = _g->E.y[0][i];
		x2 = _g->E.x[1][i];
		y2 = _g->E.y[1][i];

		// enqueue the root of the tree
		Q.push(_root);

		// do while blocks are still queued
		while (Q.size() > 0) {

			// pop the next block from the queue
			Block = Q.front();

			// do if this is a leaf node
			if (Block->_NE == 0) {                

				// test if edge intersects with or is contained within block
				if (EdgeIntersects(Block, x1, y1, x2, y2)) {

					// add edge to the list for this block
					Block->_e.push_back(i);

					// split the block if we exceed the threshold
					if (Block->_e.size() > _k) {
						if (Split(Block) != 0) {
							return 1;
						}
					}
				}                
			}

			// do if non-leaf node
			else {

				// test each of the four children against the bounding
				// box of the current edge, enqueue any matching blocks

				// NE
				if (BoxIntersects(Block->_NE, x1, y1, x2, y2)) {
					Q.push(Block->_NE);
				}
				// NW
				if (BoxIntersects(Block->_NW, x1, y1, x2, y2)) {
					Q.push(Block->_NW);
				}
				// SW
				if (BoxIntersects(Block->_SW, x1, y1, x2, y2)) {
					Q.push(Block->_SW);
				}
				// SE
				if (BoxIntersects(Block->_SE, x1, y1, x2, y2)) {
					Q.push(Block->_SE);
				}

			}
			// end switch on leaf or non-leaf node

			// dequeue the current block
			Q.pop();

		}
		// end while blocks still exist in Q        
	}
	// end of loop for each segment

	return 0;
} 
*/
//
// Split - split a block into 4 child blocks
//    return 0 on success, non-zero on something bad
int CQuadtree::Split(CQuadTreeNode* Block)
{   
	int         i, j;
	int         x1, y1, x2, y2;
	CQuadTreeNode*   NewBlock;

	// fail if block already has children
	if (Block->_NE != 0)
		return 1;

	// return silently if depth has already been reached
	if (Block->_depth == 0)
		return 0;

	// allocate space for new child blocks
	try 
	{
		Block->_NE = new CQuadTreeNode(Block->_depth - 1);
		Block->_NW = new CQuadTreeNode(Block->_depth - 1);
		Block->_SW = new CQuadTreeNode(Block->_depth - 1);
		Block->_SE = new CQuadTreeNode(Block->_depth - 1);
	}
	catch (std::bad_alloc) 
	{
		return 1;
	}

	// set each new child block's parent
	Block->_NE->_parent = Block;
	Block->_NW->_parent = Block;
	Block->_SW->_parent = Block;
	Block->_SE->_parent = Block;

	int iNewBlockSize = (2 << Block->_NE->_depth);
	// set each child block's origin coordinates 
	Block->_NE->_x = Block->_x + iNewBlockSize;
	Block->_NE->_y = Block->_y + iNewBlockSize;

	Block->_NW->_x = Block->_x;
	Block->_NW->_y = Block->_y + iNewBlockSize;

	Block->_SW->_x = Block->_x;
	Block->_SW->_y = Block->_y;

	Block->_SE->_x = Block->_x + iNewBlockSize;
	Block->_SE->_y = Block->_y;

	// increment the number of tree leaves by 3,
	// remember, previously the parent was considered a leaf 
	_numleaves += 3;

	//
	// Build the child edge lists from the edges in the parent list
	//

	for (i = 0; i < Block->_e.size(); i++) 
	{
		// get the index in E of this edge in e
		VElementRect* pElem = Block->_e[i];
		pElem->GetRect(x1, y1, x2, y2);
		bool bPromoted = false;

		// start with the NE child block
		for (j = 0, NewBlock = Block->_NE; j < 4; j++) 
		{
			// test if rect intersects or is entirely contained within block
			if (BoxContainsPoint(NewBlock, x1, y1, x2, y2)) 
			{
				bPromoted = true;
				// add edge to the list for new block
				NewBlock->_e.push_back(pElem);

				// NOTE: do not split the child block here even if it
				// contains all the elements of its parent, only split
				// the next time a segment is added to it
			}

			// then go to NW, SW, and SE blocks
			if (j == 0)
				NewBlock = Block->_NW;
			else if (j == 1)
				NewBlock = Block->_SW;
			else if (j == 2)
				NewBlock = Block->_SE;
		}   // continue to next child block

		if(!bPromoted)
		{
			assert(false); // The block got lost!!
		}

	}   // continue to next edge in parent's original list

	// now remove all items from the parent block's list
	Block->_e.clear();

	return 0;
}

//
// EdgeIntersects - determine whether a given segment intersects with 
//   a block, or alternatively if the segment is entirely contained 
//   within the block
//
/*
int CQuadtree::EdgeIntersects(CQuadTreeNode* Block, int x1, int y1, int x2, int y2)
{   
	// This algorithm adapted from Kyle Louden's "Mastering Algorithms in C" 

	int     x3, x4, y3, y4;
	int     z1, z2, s1, s2;
	int     i, w, match;

	// define the block as a bounding box 
	w = (2 << Block->_depth);
	x3 = Block->_x;
	y3 = Block->_y;
	x4 = Block->_x + w;
	y4 = Block->_y + w;

	// first test whether the segment bounding box is entirely contained
	// within the block.  If true, the segment is obviously inside the block

	if ((x3 <= MIN(x1, x2)) &&
		(MAX(x1, x2) <= x4) &&
		(y3 <= MIN(y1, y2)) &&
		(MAX(y1, y2) <= y4)) return 1;

	// if segment bounding box is not entirely contained within the block, then
	// the segment itself must intersect with at least one of the four sides
	// of the block.  Otherwise, the segment and block do not intersect 

	// reset y4 such that we are testing the south edge of the block 
	y4 = Block->_y;

	for (i = 0, match = 0; i < 4; i++) {

		// determine orientation of p3 relative to p2, wrt p1 
		if ((z1 = ((x3-x1)*(y2-y1)) - ((y3-y1)*(x2-x1))) < 0)
			s1 = -1;
		else if (z1 > 0)
			s1 = 1;
		else
			s1 = 0;

		// determine orientation of p4 relative to p2, wrt p1 
		if ((z2 = ((x4-x1)*(y2-y1)) - ((y4-y1)*(x2-x1))) < 0)
			s2 = -1;
		else if (z2 > 0)
			s2 = 1;
		else
			s2 = 0;

		// if the signs of z1 and z2 are different, or either is 0, 
		// then signal a match 
		if ((s1 == 0 || s2 == 0) || (s1 != s2)) {
			match = 1;
			break;
		}

		// prepare to test another edge

		// 2nd test = west edge 
		if (i == 0) {
			x4 = Block->_x;
			y4 = Block->_y + w;
		}

		// 3rd test = north edge
		else if (i == 1) {
			x3 = Block->_x + w;
			y3 = Block->_y + w;
		}

		// 4th test = east edge
		else if (i == 2) {
			x4 = Block->_x + w;
			y4 = Block->_y;
		}

	}

	if (match)
		return 1;
	else    
		return 0;

}
*/

//
// BoxIntersects - determine whether a segment bounding box 
//   intersects with block
//
int CQuadtree::BoxIntersects(CQuadTreeNode* Block, int x1, int y1, int x2, int y2)
{
	int x3, x4, y3, y4;
	int w;

	// define the block as a bounding box 
	w = (2 << Block->_depth);
	x3 = Block->_x;
	y3 = Block->_y;
	x4 = Block->_x + w;
	y4 = Block->_y + w;

	return IntersectRect(x1, y1, x2, y2, x3, y3, x4, y4);
}

int CQuadtree::BoxContainsPoint(CQuadTreeNode* Block, int x3, int y3, int x4, int y4)
{
	int x1, x2, y1, y2;
	int w;

	// define the block as a bounding box 
	w = (2 << Block->_depth);
	x1 = Block->_x;
	y1 = Block->_y;
	x2 = Block->_x + w;
	y2 = Block->_y + w;

	return ContainsPoint(x1, y1, x2, y2, x3, y3, x4, y4);
}

//
// NearestEdge - find the segment nearest to a given query point
//   return the index in G->E of the nearest segment if found,
//   else return -1 
/*
int CQuadtree::NearestEdge(int qx, int qy, double radius)
{
	int retval = -1;
	bool found = false;
	double dist;
	int i;
	int edge_count;
	int eid;
	int x1, x2, y1, y2;
	CQuadTreeNode* Block;
	PQElem pqelem;
	std::priority_queue<PQElem> PQ;

	// quick return if query point is out of range
	if (qx < 0 || qx >= (2 << _depth) || qy < 0 || qy >= (2 << _depth))
		return retval;

	// if radius parameter is 0, set to infinite
	if (radius == 0.0) radius = DBL_MAX;

	// set the root block of the quadtree as the first
	// element of the priority queue
	PQ.push(PQElem(false, 0.0, _root, -1));

	//
	// main loop
	//
	while (!found && !PQ.empty()) {

		// pop the top of the priority queue, this element is either
		// an edge or a Block, in either case it is the nearest element
		// to the query point
		pqelem = PQ.top();
		PQ.pop();

		// if this is a segment, return its index in E
		// unless it is beyond the search radius
		if (pqelem.IsEdge) {

			found = true;
			if (pqelem.dist > radius) {
				retval = -1;
			}
			else {
				retval = pqelem.edge_idx;
			}
		}

		// if leaf block, then add its segments to the pq
		else if (pqelem.Block->_NE == 0) {

			edge_count = pqelem.Block->_e.size();

			for (i = 0; i < edge_count; i++) {

				eid = pqelem.Block->_e[i];
				x1 = _g->E.x[0][eid];
				y1 = _g->E.y[0][eid];
				x2 = _g->E.x[1][eid];
				y2 = _g->E.y[1][eid];

				// calculate dsq from query point to this segment
				dist = DistPointSegment(qx, qy, x1, y1, x2, y2);

				// add to PQ if not less than distance to current block
				// and less than radius
				if (dist >= pqelem.dist && dist <= radius) {                    
					PQ.push(PQElem(true, dist, 0, eid));
				}
			}
		}

		// if non-leaf block, add its 4 children to the pq
		else {

			// start with the NE child
			Block = pqelem.Block->_NE;

			// loop 4 times
			for (i = 0; i < 4; i++) {

				// calculate the distance to this block
				dist = DistPointBlock(qx, qy, Block);

				// add the block to the pq if dist is less than radius
				if (dist <= radius) {
					PQ.push(PQElem(false, dist, Block, -1));
				}

				// set to the next child block
				if (i == 0)
					Block = pqelem.Block->_NW;
				else if (i == 1)
					Block = pqelem.Block->_SW;
				else if (i == 2)
					Block = pqelem.Block->_SE;

			}

		}
		// END OF IF STATEMENT BASED ON TYPE OF ELEM IN THE PQ

	}
	// END OF MAIN LOOP WHILE NOT FOUND

	// now free any remaining elems in the pq
	// to be nice to the memory manager
	while (!PQ.empty()) {
		PQ.pop();
	}

	return retval;
}


//
// DistPointSegment - calculate the squared distance between a point and a line segment
//
double CQuadtree::DistPointSegment(int qx, int qy, int segx1, int segy1, int segx2, int segy2)
{    
	// for details, see the comp.graphics.algorithms FAQ

	double  x, y, Ax, Ay, Bx, By;
	double  d, r, s;

	x = (double) qx;
	y = (double) qy;

	Ax = (double) segx1;
	Bx = (double) segx2;
	Ay = (double) segy1;
	By = (double) segy2;

	// calculate squared length of the segment 
	d = (Bx-Ax)*(Bx-Ax) + (By-Ay)*(By-Ay);

	// if distance is zero, then the segment is just a point
	if (d == 0.0) {
		return ((x-Ax)*(x-Ax) + (y-Ay)*(y-Ay));
	}

	// calculate the parameter, r
	r = ((x-Ax)*(Bx-Ax) + (y-Ay)*(By-Ay)) / d;

	// r helps find the perpendicular from segment to query point
	// r < 0 : P falls to the left of segment, return distance to A
	// r = 0 : P = A, return distance to A
	// 0<r<1 : P is on the segment, return length of perp. segment
	// r = 1 : P = B, return distance to B
	// r > 1 : P falls to the right of segment, return distance to B 

	if (r <= 0) {
		return ((x-Ax)*(x-Ax) + (y-Ay)*(y-Ay));
	}
	else if (r >= 1) {
		return ((x-Bx)*(x-Bx) + (y-By)*(y-By));
	}
	else {
		s = ((Ay-y)*(Bx-Ax)-(Ax-x)*(By-Ay));
		return (s * s / d);
	}    
}


//
// DistPointBlock - calculate the squared distance between a point and a CQuadTreeNode
// note: if block contains point, return 0
double CQuadtree::DistPointBlock(int x, int y, CQuadTreeNode* Block)
{
	int     w;
	double  dist = 0.0;
	double  dnw, dne, dsw, dse, min, min2;
	int     v1[3], v2[3], v3[3], v4[3];

	// store the width of the block
	w = (2 << Block->_depth);

	// if block contains point, return 0
	if (Block->_x <= x &&
		x < Block->_x + w &&
		Block->_y <= y &&
		y < Block->_y + w) return dist;

	// calculate the distance from point to each of the corners of the block 
	v1[1] = Block->_x;
	v1[2] = Block->_y;
	v2[1] = Block->_x;
	v2[2] = Block->_y;
	v3[1] = Block->_x;
	v3[2] = Block->_y;
	v4[1] = Block->_x;
	v4[2] = Block->_y;


	dsw = (x-Block->_x)*(x-Block->_x) + (y-Block->_y)*(y-Block->_y);
	dnw = (x-Block->_x)*(x-Block->_x) + (y-(Block->_y+w))*(y-(Block->_y+w));
	dse = (x-(Block->_x+w))*(x-(Block->_x+w)) + (y-Block->_y)*(y-Block->_y);
	dne = (x-(Block->_x+w))*(x-(Block->_x+w)) + (y-(Block->_y+w))*(y-(Block->_y+w));

	// find the nearest point 
	v1[1] = Block->_x;
	v1[2] = Block->_y;
	min = dsw;
	if (dnw < min) {
		v1[1] = Block->_x;
		v1[2] = Block->_y + w;
		min = dnw;
	}
	if (dse < min) {
		v1[1] = Block->_x + w;
		v1[2] = Block->_y;
		min = dse;
	}
	if (dne < min) {
		v1[1] = Block->_x + w;
		v1[2] = Block->_y + w;
		min = dne;
	}

	// find the second nearest point
	v2[1] = Block->_x;
	v2[2] = Block->_y;
	min2 = dsw;
	if (min2 == min) {
		v2[1] = Block->_x;
		v2[2] = Block->_y + w;
		min2 = dnw;
	}
	if (dse < min2 && dse != min) {
		v2[1] = Block->_x + w;
		v2[2] = Block->_y;
		min2 = dse;
	}
	if (dne < min2 && dne != min) {
		v2[1] = Block->_x + w;
		v2[2] = Block->_y + w;
		min2 = dne;
	}

	// get the distance to the segment formed by the two nearest points
	dist = DistPointSegment(x, y, v1[1], v1[2], v2[1], v2[2]);

	return dist;
}
*/

bool CQuadtree::FindIntersectElements(std::vector<VElementRect*>& elems, Rect r)
{
	std::set<VElementRect*> elemSet;
	CQuadTreeNode* Block;
	std::queue<CQuadTreeNode*> Q;
	int x1 = r.x, y1 = r.y, x2 = r.right(), y2 = r._top();

	if (!_root) return false;
	// enqueue the root of the tree
	Q.push(_root);

	// do while blocks are still queued
	while (Q.size() > 0) 
	{
		// pop the next block from the queue
		Block = Q.front();

		// do if this is a leaf node
		if (Block->_NE == 0)
		{
			// test if rect intersects with or is contained within block
			if (1)//BoxContainsPoint(Block, x1, y1, x2, y2)) 
			{
				for(size_t i = 0; i < Block->_e.size(); i++)
				{
					int x3, y3, x4, y4;
					Block->_e[i]->GetRect(x3, y3, x4, y4);
					if(IntersectRect(x1, y1, x2, y2, x3, y3, x4, y4))
					{
						if (elemSet.find(Block->_e[i]) == elemSet.end())
						{
							elemSet.insert(Block->_e[i]);
							elems.push_back(Block->_e[i]);
						}
					}
				}
			}                
		}
		// do if non-leaf node
		else 
		{
			// test each of the four children against the bounding
			// box of the current edge, enqueue any matching blocks

			// NE
			if (BoxIntersects(Block->_NE, x1, y1, x2, y2)) {
				Q.push(Block->_NE);
			}
			// NW
			if (BoxIntersects(Block->_NW, x1, y1, x2, y2)) {
				Q.push(Block->_NW);
			}
			// SW
			if (BoxIntersects(Block->_SW, x1, y1, x2, y2)) {
				Q.push(Block->_SW);
			}
			// SE
			if (BoxIntersects(Block->_SE, x1, y1, x2, y2)) {
				Q.push(Block->_SE);
			}
		}
		// end switch on leaf or non-leaf node

		// dequeue the current block
		Q.pop();
	}
	// end while blocks still exist in Q

	return elems.size() > 0;
}
