#ifndef QUADTREE_H__
#define QUADTREE_H__

#include "common/SimpleTypes.h"
#include <vector>
#include <queue>
#include <float.h>

// MAX and MIN macros used in EdgeInteresects and BoxIntersects functions
#define MIN(x, y) ((x) < (y) ? (x) : (y))
#define MAX(x, y) ((x) > (y) ? (x) : (y))

class CQuadtree;

//
// Quadtree node class
//
class CQuadTreeNode
{
	friend class CQuadtree;

public:
	// constructor
	CQuadTreeNode(int set_depth);

private:
	// A CQuadTreeNode represents a node in the tree

	// Depth: a power of two expressing the level of depth of this 
	// block. The depth of the root block defines the total amount of 
	// space to be subdivided, such that 2^depth == width of one side 
	// of the square block. A child block's depth is one less than its
	// parent block's depth. The root block's depth is thus the maximum
	// number of layers between the root and any given leaf since at a
	// depth of 0, 2^0 == 1, and we have a block that is one unit 
	// square.  Beyond that no further splitting is possible and 
	// additional items added to such a leaf will be added beyond
	// the splitting threshold ad infinitum.
	int _depth;

	// pointer to this block's parent block
	CQuadTreeNode* _parent;

	// Every parent block will have 4 children, arranged in quadrants.
	// A simple test for whether a given block is a leaf or not is to
	// check whether any one of these is NULL.  If so, it is a leaf. 
	CQuadTreeNode* _NE;
	CQuadTreeNode* _NW;
	CQuadTreeNode* _SW;
	CQuadTreeNode* _SE;

	// Vector of elements added to the block
	std::vector<VElementRect*> _e;

	// Origin of block (lower-left) as integer x, y coordinate pair.
	// The root block has an origin of (0,0), and the 4 children
	// of the root block have origins at:
	// NE: (2^(depth - 1), 2^(depth - 1))
	// NW: (0, 2^(depth - 1))
	// SW: (0, 0)
	// SE: (2^(depth - 1), 0)
	// Storing this with every block makes spatial comparisons
	// easier when determining which block a given point or
	// line segment falls into.
	int _x, _y;

};


//
// Quadtree class
//
class CQuadtree
{
public:
	// constructor
	CQuadtree() : _depth(16), _k(8), _root(NULL), _numleaves(0) {}
	void Create(LONG width, LONG height, int set_k = 8);

	// build it
	int AddElement(VElementRect* pElem);

	bool FindIntersectElements(std::vector<VElementRect*>& elems, Rect r);
	// test whether an edge intersects with a Block
	//int EdgeIntersects(CQuadTreeNode* Block, int x1, int y1, int x2, int y2);

	static inline int PtInRect(int x, int y, int x1, int y1, int x2, int y2);
	static inline int IntersectRect(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4);
	static inline int ContainsPoint(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4);
	// test whether the bounding box of an edge intersects with a Block
	int BoxIntersects(CQuadTreeNode* Block, int x1, int y1, int x2, int y2);
	int BoxContainsPoint(CQuadTreeNode* Block, int x3, int y3, int x4, int y4);

	// split a node into 4 children
	int Split(CQuadTreeNode* Block);

	// find the segment nearest to a given query point
	int NearestEdge(int qx, int qy, double radius);

	// debugging
	int GetDepth() const { return _depth; };
	int GetK() const { return _k; };
	int GetNumLeaves() const { return _numleaves; };

	bool empty () { return _root == 0; } 

private:

	// The depth of the root block where 2^depth == length of one
	// side of the square encompassing all the space to be subdivided.
	int _depth;

	// The splitting threshold for each block in the tree.  When items
	// are added to the list of items contained in a given block, if 
	// that insertion brings the list size to k, the block will then be
	// split into 4 children, then each item in the original list will
	// be tested to see which of the 4 children's lists it should be
	// inserted into, before being emptied from the (now-parent) list.
	// The literature suggests that 8 is a good splitting threshold 
	// to use in most applications.
	int _k;

	// pointer to the root block of the tree 
	CQuadTreeNode* _root;

	// The number of leaf nodes in the tree. Note that this is not the
	// same as the current number of used lists used above to handle
	// storage.  When a block is split, the number of leaves increases
	// by 3, not 4, because previously, the parent counted as a leaf.
	// However, we do not free the space allocated for the list of the
	// now-parent, we just empty it, but add pointers in the array to
	// brand new lists for the children. 
	int _numleaves;

	// private member functions

	// calculate the squared distance between a point and a line segment
	double DistPointSegment(int qx, int qy, int segx1, int segy1, int segx2, int segy2);

	// calculate the squared distance between a point and a CQuadTreeNode
	// note: if block contains point, return 0
	double DistPointBlock(int x, int y, CQuadTreeNode* Block);

};

inline int CQuadtree::PtInRect(int x, int y, int x1, int y1, int x2, int y2)
{
	return x >= x1 && x < x2 && y >= y1 && y < y2;
}

inline int CQuadtree::IntersectRect(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4)
{
	// perform the test, return 1 if the boxes intersect 
	if ((MAX(x1, x2) >= MIN(x3, x4)) &&
		(MAX(x3, x4) >= MIN(x1, x2)) &&
		(MAX(y1, y2) >= MIN(y3, y4)) &&
		(MAX(y3, y4) >= MIN(y1, y2))) return 1;

	return 0;        
}

inline int CQuadtree::ContainsPoint(int x1, int y1, int x2, int y2, int x3, int y3, int x4, int y4)
{
	return PtInRect(x3, y3, x1, y1, x2, y2) || PtInRect(x4, y3, x1, y1, x2, y2) || PtInRect(x4, y4, x1, y1, x2, y2) ||
		PtInRect(x3, y4, x1, y1, x2, y2);
}

#endif  // QUADTREE_H__                                              