/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#ifndef	__FILE_H__
#define	__FILE_H__
/******************************************************************************
 *
 *
 *							File base class.
 *
 *
 ******************************************************************************/
#include "FileLimits.h"				// Limits
#include <stdio.h>
#include "errorlib\berrorcodes.h"

#define DEFAULT_READSIZE	32767			// Default size for reading and writing
#define DEFAULT_GRANULARITY	(1024*1024)			// Default granularity of FileMemory
#define	FILESIZE_ERROR		0xFFFFFFFF

/*------------------------------------------------------------------------
Class:					vcFile
Base Class:				(none)
Description:			This is the virtual base class for a file object. It
						provides the interface to which other libraries write to files.
Usable:					No
Public Methods:		
Public Data:			
Usage Notes:			
Inheritable:			Yes
Overridable Methods:	
Override Notes:
		- These methods should throw a BException class when errors occur.
		- Read() should be implemented so as to raise an exception when
		  bReadExact is true and less then uCount bytes are read.
-------------------------------------------------------------------------*/
class vcFile 
{
public:
	/* The OpenReason type indicates the reason for opening the file.
	 */
	typedef enum OpenReason {
		orOpenNone = 0,							// Uninitialized
		orOpenForRead,							// Use to read a file
		orOpenForReadWrite,						// Use to read or write a file
		orOpenForOverwrite,						// Use to overwrite an existing file
	};

	/* This specifies the memory mode of a memfile
	 */
	typedef enum FileMemoryMode {
		moNormal = 0,							// Normal internal memory allocations
		moSharedPointerExternal = 1
	};

	/* This specifies from where a file seek should be performed.	 */
	typedef enum OriginForSeek {
		ofsBegin = 0,
		ofsCurrent,
		ofsEnd
	};

	/* This specifies how to read  */
	typedef enum eByteOrder {
		boIntelByteOrder = 0,
		boMotorolaByteOrder,
	};


	/* Construction/Destruction.	 */
	vcFile() {;};
	virtual ~vcFile() {;};

	/* Methods required by JFF. Some may be no-op functions, but
	 * the implementor must decide.	 */

	virtual void Open(const OpenReason ReasonForOpening) = 0;
	virtual void Write(const PVOID pBuffer, const DWORD dwCount) = 0;
	virtual void Flush() = 0;
	virtual DWORD  Read(PVOID pBuffer, DWORD dwCount, bool bReadExact = false) = 0;
	virtual void Close() = 0;
	virtual void Delete() = 0;
	virtual bool Eof() = 0;

	virtual DWORD GetFileSize()  = 0;
	virtual DWORD Seek(LONG lDistance, OriginForSeek eFrom) = 0;

	virtual LPCTSTR GetFilePathname() const = 0;

	/* These are helper functions that are implemented
	 * in the base class (though they call virtual functions
	 * that are implemented in the child classes).	 */
	inline DWORD Seek(LONG lDistance);
	inline DWORD SeekFromBeginning(LONG lDistance);
	inline DWORD SeekFromEnd(LONG lDistance);
	inline DWORD GetCurrentPosition();

	int GetString(char* psz, size_t iMaxSize);
	DWORD ReadDWord(eByteOrder ByteOrder = boIntelByteOrder);
	WORD ReadWord(eByteOrder ByteOrder = boIntelByteOrder);

	void WriteDWord(const DWORD dwData, eByteOrder ByteOrder = boIntelByteOrder);
	void WriteWord(const WORD dwData, eByteOrder ByteOrder = boIntelByteOrder);
	void WriteByte(const BYTE byData);
	void WriteString(const wchar_t* psz, eByteOrder ByteOrder = boIntelByteOrder);

	void fprintf(LPCTSTR lpszFormat, ...);
	static void __fprintf(vcFile* out, LPCTSTR lpszFormat, ...);

	void Copy(vcFile* pSource);
	void Append(vcFile* pSource);
};

/* These base class implementations are "shortcuts" to child class
 * functionality.
 *
 * Note that it is left to inheriting classes to 
 * implement the Seek(LONG, OriginForSeek) method.
 */
inline DWORD vcFile::Seek(LONG lDistance) {
	return Seek(lDistance, vcFile::ofsCurrent);
}

inline DWORD vcFile::SeekFromBeginning(LONG lDistance) {
	return Seek(lDistance, vcFile::ofsBegin);
}

inline DWORD vcFile::SeekFromEnd(LONG lDistance) {
	return Seek(lDistance, vcFile::ofsEnd);
}

inline DWORD vcFile::GetCurrentPosition() {
	return Seek(0L);
}


/******************************************************************************
 *
 *
 *			File child class that reads/writes to disk (CFileDisk).
 *
 *
 ******************************************************************************/

/*------------------------------------------------------------------------
Class:					CFileDisk
Base Class:				vcFile
Description:			This class is the vcFile implementation that
						reads from and writes to a physical file on disk.
Usable:					Yes
Public Methods:		
Public Data:			
Usage Notes:			
Inheritable:			No
Overridable Methods:	
Override Notes:		
-------------------------------------------------------------------------*/
class CFileDisk : public vcFile 
{
public:
	/* Construction/Destruction.	 */
	CFileDisk();
#ifndef _UNICODE
	CFileDisk(LPCTSTR szFilePathname);
#endif
	CFileDisk(const wchar_t* szFilePathname);
	virtual ~CFileDisk();

	/* Override the following methods from the base class.	 */
	void Open(const OpenReason ReasonForOpening);
	void Write(const PVOID pBuffer, const DWORD dwCount);
	void Flush();
	DWORD Read(PVOID pBuffer, DWORD dwCount, bool bReadExact = false);
	void Close();
	void Delete();
	bool Eof();

	DWORD GetFileSize();
	static LONG GetFileSize(FILE* f);

	DWORD Seek(LONG lDistance, OriginForSeek eFrom);

	// The following methods are not defined in the base class
	void SetFilePathname(LPCTSTR fileName);
	LPCTSTR GetFilePathname() const;
	int GetError(){return m_iErrorCode; };

protected:
	typedef enum Activity {
		aNone,
		aReading,
		aReadingWriting,
		aWriting,
	};

	//char				m_abyFilePathname[MAX_FILENAME_LEN];
	std::string		m_strFilename;
	std::wstring	m_strFilenameW;

	FILE*				m_pDiskFile;
	Activity			m_CurrentActivity;
	int					m_iErrorCode;
};

/******************************************************************************
 *
 *
 *		vcFile child class that reads/writes to memory buffer (CFileMemory).
 *
 *
 ******************************************************************************/

/*------------------------------------------------------------------------
Class:					CFileMemory
Base Class:				vcFile
Author:					Paul Gruzdev
Description:			This class is the vcFile implementation that
						reads from and writes to a memory buffer.
Usable:					Yes
Public Methods:		
Public Data:			
Usage Notes:			
Inheritable:			No
Overridable Methods:	
Override Notes:		
-------------------------------------------------------------------------*/
class CFileMemory : public vcFile 
{

  public:

	  /* Construction/Destruction. */

	  CFileMemory();
	  CFileMemory(const OpenReason ReasonForOpening);
	  virtual ~CFileMemory();

	  /* Overriding these methods. */

	  void Open(const OpenReason ReasonForOpening);
	  void Write(const PVOID pBuffer, const DWORD dwCount);
	  void Flush();
	  DWORD Read(PVOID pBuffer, DWORD dwCount, bool bReadExact = false);
	  void Close();
	  void Delete();
	  bool Eof();

	  DWORD GetFileSize();
	  DWORD Seek(LONG lDistance, OriginForSeek eFrom);
	  BYTE* GetTempMemPtr();
	  void SetSize(DWORD dwCount);
	  void SetGranularity(DWORD dwCount);
	  BOOL SetPointerMemoryMode(BYTE* pbyMemory, DWORD dwMemorySize);
	  PVOID Detach();
	  BOOL SetMemoryOrigin(DWORD dwOrigin);
  	  PVOID GetMemoryBuffer() { return m_pMemoryBuffer; }
      LPCTSTR GetFilePathname() const { return(_T("")); };

  protected:

	BOOL ReAllocate(DWORD dwNewSize);


	FileMemoryMode	m_Mode;						// What memory mode are we in?
	PVOID			m_pMemoryBuffer;			// Pointer to memory buffer
	DWORD			m_dwMemorySize;				// Bytes currently used in memory block
    DWORD			m_dwMemoryPosition;			// current position in memory block
	DWORD			m_dwMemoryAllocated;		// Bytes allocated for memory block (may be larger than m_dwmemorysize)
	DWORD			m_dwGranularity;			// Minimum allocation size unit
	DWORD			m_dwMemoryOrigin;			// Origin (offset) of start of file
	OpenReason		m_Reason;					// What do they want us to do?
};

#endif	// __FILE_H__
