/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#include "stdafx.h"

#include "BException.h"
#include <assert.h>
#include <tchar.h>
#include <atlconv.h>
#pragma warning(disable:4711)

// This static exception object is declared here so that it does not have to
// be allocated when the system is low on memory.  It will be allocated at
// program startup.
#pragma warning(push)
#pragma warning(disable:4710)
static CBException s_MemoryError(EID_MEMORY_ERROR);
#pragma warning(pop)

// This global boolean flag can be used in the debugger to turn on and off
// automatic logging of exceptions.  By default it is off.

bool g_bLogCBException = false;

/******************************************************************************
   Function:    CBException::CBException
   Author:      Michael Tatarnikov
   Purpose:     Copy constructor.
   Returns:     void
   Exceptions:  
   Parameters:  const CBException &rkoException - existing exception object
						to copy.
   Notes:       If this is a debug build and the global flag g_bLogCBException
				is true, a dump of the object information will be sent to the
				debugging window.
******************************************************************************/
CBException::CBException(const CBException &rkoException)
{
	m_ErrorCode = rkoException.m_ErrorCode;
	m_ulRefCount = 0;
	m_bAutoDelete = false;
	m_bAutoThrow = false;
	m_pszMsg = NULL;
	if (rkoException.m_pszMsg != NULL)
		SetWhat(rkoException.m_pszMsg);
	m_pszDebugMsg = rkoException.m_pszDebugMsg;
	m_pszFile = rkoException.m_pszFile;
	m_iFileLine = rkoException.m_iFileLine;

	// thrown exceptions are always copied, so this makes a perfect place to
	// output the debug information
#ifdef _DEBUG
	OutputDebug();
#endif
}

/******************************************************************************
   Function:    operator =
   Author:      Michael Tatarnikov
   Purpose:     Copy operator.
   Returns:     const CBException & - reference to this object.
   Exceptions:  
   Parameters:  const CBException & rkoException - existing exception object
						to copy.
   Notes:       
******************************************************************************/
const CBException & CBException::operator =(const CBException & rkoException)
{
	m_ErrorCode = rkoException.m_ErrorCode;
	m_bAutoDelete = false;
	m_bAutoThrow = false;
	if ((m_pszMsg != NULL) && (m_pszMsg != m_szMsgBuf))
		delete [] m_pszMsg;
	m_pszMsg = NULL;
	if (rkoException.m_pszMsg != NULL)
		SetWhat(rkoException.m_pszMsg);
	m_pszDebugMsg = rkoException.m_pszDebugMsg;
	m_pszFile = rkoException.m_pszFile;
	m_iFileLine = rkoException.m_iFileLine;

	return *this;
}

/******************************************************************************
   Function:    CBException::~CBException
   Author:      Michael Tatarnikov
   Purpose:     Destructor.
   Returns:     void
   Exceptions:  
   Parameters:  None
   Notes:       If the global variable g_bLogCBException is set, an error message
				will be sent to the debugger.  This could have been done in the
				constructor, or where the error code is set, but putting it here
				ensures the message is only produced once.  The theory is that
				an exception object may change the error it holds at any time,
				but is unlikely to change after the error has been thrown.
******************************************************************************/
CBException::~CBException()
{
	// delete the buffer holding the what() string
	if ((m_pszMsg != NULL) && (m_pszMsg != m_szMsgBuf))
		delete [] m_pszMsg;
}

/******************************************************************************
   Function:    CBException::AddRef
   Author:      Michael Tatarnikov
   Purpose:     Add one to the reference count.
   Returns:     ULONG - the resulting reference count.  May not be accurate in
						a multi-threaded environment or under Windows 95 (see
						documentation for InterlockedIncrement).
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
ULONG CBException::AddRef()
{
	return (ULONG) InterlockedIncrement((LPLONG) &m_ulRefCount);
}

/******************************************************************************
   Function:    CBException::Release
   Author:      Michael Tatarnikov
   Purpose:     Release a reference, and delete the object when zero.
   Returns:     ULONG - the resulting reference count.  May not be accurate in
						a multi-threaded environment.
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
ULONG CBException::Release()
{
	if (InterlockedDecrement((LPLONG) &m_ulRefCount) > 0)
		return m_ulRefCount;

	// ignore auto-throw if we are already inside exception processing
	if (m_bAutoThrow && !std::uncaught_exception() && (m_ErrorCode != EID_NO_ERROR))
	{
		// If we have auto-throw AND auto-delete, we need to delete the object
		// AFTER the throw.  The auto_ptr will delete the object in its destructor.
		std::auto_ptr<CBException> spTemp(m_bAutoDelete ? this : NULL);
		Throw();
	}

	if (m_bAutoDelete)
		delete this;

	return 0;
}

/******************************************************************************
   Function:    CBException::Throw
   Author:      Michael Tatarnikov
   Purpose:     Throw an exception and delete the object.
   Returns:     void
   Exceptions:  
   Parameters:  None
   Notes:       This function MUST be overridden in any derived class.  If it
				is not, the exception thrown will be a CBException, not the
				derived class!
******************************************************************************/
void CBException::Throw()
{
	throw *this;
}

#define STRSIZE 512

void GetErrorMessageString(HINSTANCE hResourceHandle, BErrorCode jecRetVal, LPTSTR cStr, short nLenStr)
{
    TCHAR cStr2[STRSIZE];

    if (!LoadString(hResourceHandle, jecRetVal, cStr, nLenStr)) 
	{
	    if (!LoadString(hResourceHandle, BE_UNDEFINED_ERROR, cStr2, STRSIZE-1)) 
			_tcscpy(cStr2, _T("Undefined error - %u"));

		_stprintf(cStr, cStr2, jecRetVal);
	}
}

/******************************************************************************
   Function:    CBException::what
   Author:      Michael Tatarnikov
   Purpose:     Return an error message.
   Returns:     LPCTSTR - the error message, suitable for presenting to
						the user.
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
LPCSTR CBException::what() const
{
	// if a message was set earlier, return it now
	if (m_pszMsg != NULL)
#ifndef _UNICODE
		return m_pszMsg;
#else
		return m_szMsgBufA;
#endif

	// get the message from JLem
	m_pszMsg = m_szMsgBuf;
	try
	{
		GetErrorMessageString(GetModuleHandle(NULL), m_ErrorCode, m_pszMsg, sizeof(m_szMsgBuf)/sizeof(m_szMsgBuf[0]));
	}
	catch (...)
	{
		m_szMsgBuf[0] = 0;
	}

#ifdef _UNICODE
	USES_CONVERSION;
	char* psz = W2A(m_pszMsg);
	strcpy(m_szMsgBufA, psz);
	return m_szMsgBufA;
#else
	return m_pszMsg;
#endif
}

/******************************************************************************
   Function:    CBException::what
   Author:      Michael Tatarnikov
   Purpose:     Return an error message.
   Returns:     LPCTSTR - the error message, suitable for presenting to
						the user.
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
LPCSTR CBException::what(HINSTANCE hResource) const
{
	// if a message was set earlier, return it now
	if (m_pszMsg != NULL)
#ifndef _UNICODE
		return m_pszMsg;
#else
		return m_szMsgBufA;
#endif

	// get the message from JLem
	m_pszMsg = m_szMsgBuf;
	try
	{
		GetErrorMessageString(hResource, m_ErrorCode, m_pszMsg, sizeof(m_szMsgBuf)/sizeof(m_szMsgBuf[0]));
	}
	catch (...)
	{
		m_szMsgBuf[0] = 0;
	}
#ifdef _UNICODE
	USES_CONVERSION;
	char* psz = W2A(m_pszMsg);
	strcpy(m_szMsgBufA, psz);
	return m_szMsgBufA;
#else
	return m_pszMsg;
#endif
}

/******************************************************************************
   Function:    CBException::SetWhat
   Author:      Michael Tatarnikov
   Purpose:     Set an error message to be returned.
   Returns:     void
   Exceptions:  
   Parameters:  LPCTSTR pszMessage - the error message.
   Notes:       The error messages returned by what() are intended for display
				to the user.  This means that they must be in the user's language,
				and should not contain excessive technical jargon.
******************************************************************************/
void CBException::SetWhat(LPCTSTR pszMessage)
{
	// delete any old message
	if ((m_pszMsg != NULL) && (m_pszMsg != m_szMsgBuf))
		delete [] m_pszMsg;

	// allocate a buffer for a copy of the message
	// TODO: check the following for proper operation on multi-byte strings
	size_t iLen = _tcslen(pszMessage) + 1;
	if (iLen <= sizeof(m_szMsgBuf)/sizeof(m_szMsgBuf[0]))
		m_pszMsg = m_szMsgBuf;
	else
		m_pszMsg = new TCHAR[iLen];

	// copy the message
	if (m_pszMsg != NULL)
		_tcscpy(m_pszMsg, pszMessage);
}

/******************************************************************************
   Function:    CBException::OutputDebug
   Author:      Michael Tatarnikov
   Purpose:     Output debugging information.
   Returns:     void
   Exceptions:  
   Parameters:  None
   Notes:       This function does nothing in a release build.
******************************************************************************/
void CBException::OutputDebug() const
{
#ifdef _DEBUG
	if (g_bLogCBException && (m_ErrorCode != EID_NO_ERROR))
	{
		// Instead of declaring a huge buffer and using sprintf to format a
		// single line, we use a small static buffer and use a series of
		// OutputDebugStrings.  This is to allow correct operation in a low
		// memory situation.
		static TCHAR s_szFormat1[] = _T("(%d) : Exception %d : ");
		static TCHAR s_szFormat2[] = _T("Exception %d : ");
		static TCHAR s_szLine[sizeof(s_szFormat1)/sizeof(s_szFormat1[0]) + 2*9];

		// output file name, line number, and error code
		if (m_pszFile != NULL)
		{
			OutputDebugStringA(m_pszFile);
			_stprintf(s_szLine, s_szFormat1, m_iFileLine, m_ErrorCode);
		}
		else
			_stprintf(s_szLine, s_szFormat2, m_ErrorCode);
		OutputDebugString(s_szLine);

		// output user-level description
		LPCSTR pszWhat = what();
		if (pszWhat[0] == 0)
		{
			if (m_ErrorCode == EID_USER_ABORTED)
				pszWhat = "USER_ABORTED";
			else if (m_ErrorCode == EID_INVALID_PARAMETER_PASSED)
				pszWhat = "INVALID_PARAMETER_PASSED";
		}
		OutputDebugStringA(pszWhat);

		// finally, output the debugging string provided by the caller
		if (m_pszDebugMsg != NULL)
		{
			OutputDebugString(_T(" : "));
			OutputDebugString(m_pszDebugMsg);
		}

		// top it all off with a newline
		OutputDebugString(_T("\n"));
	}
#endif	// _DEBUG
}

/***********************************************************************************
***********************************************************************************/

/******************************************************************************
   Function:    CErrorAutoThrow::CErrorAutoThrow
   Author:      Michael Tatarnikov
   Purpose:     Constructor from error code.
   Returns:     void
   Exceptions:  
   Parameters:  BErrorCode jec - error code; see BErrorCodes.h.
                LPCTSTR pszDebugMsg - message that can be seen in the debugger.
						This message will be automatically output to the debug
						window if this is a debug build, and the flag
						g_bLogCBException is set to true.  This message will
						only be seen by programmers, and does not have to be
						localized.  This string is NOT copied, so the caller
						must ensure its validity for as long as it might be
						used.  A string constant works best.
                LPCSTR pszFile - the name of the source file where the error
						occurred.  This string is NOT copied, but this is not
						usually a problem if the source is the __FILE__ macro,
						which is a static constant.
                int iLine - the line number of the source file where the
						error occurred.
   Notes:       
******************************************************************************/
CErrorAutoThrow::CErrorAutoThrow(BErrorCode jec, LPCTSTR pszDebugMsg, LPCSTR pszFile, int iLine)
{
	m_pEx = NULL;
	if (jec != EID_NO_ERROR)
		SetError(jec, pszDebugMsg, pszFile, iLine);
}

/******************************************************************************
   Function:    CErrorAutoThrow::SetError
   Author:      Michael Tatarnikov
   Purpose:     Set the error code.
   Returns:     void
   Exceptions:  
   Parameters:  BErrorCode jec - error code; see BErrorCodes.h.
                LPCTSTR pszDebugMsg - message that can be seen in the debugger.
						This message will be automatically output to the debug
						window if this is a debug build, and the flag
						g_bLogCBException is set to true.  This message will
						only be seen by programmers, and does not have to be
						localized.  This string is NOT copied, so the caller
						must ensure its validity for as long as it might be
						used.  A string constant works best.
                LPCSTR pszFile - the name of the source file where the error
						occurred.  This string is NOT copied, but this is not
						usually a problem if the source is the __FILE__ macro,
						which is a static constant.
                int iLine - the line number of the source file where the
						error occurred.
   Notes:       
******************************************************************************/
void CErrorAutoThrow::SetError(BErrorCode jec, LPCTSTR pszDebugMsg, LPCSTR pszFile, int iLine)
{
	// throw away any old value
	if (m_pEx != NULL)
	{
		m_pEx->AutoThrow(false);
		m_pEx->Release();
	}

	// no error?  No need for a CBException object
	if (jec == EID_NO_ERROR)
		m_pEx = NULL;

	// the general case
	else
	{
		// create a CBException to hold the error
		try
		{
			m_pEx = new CBException(jec, pszDebugMsg, true, true, pszFile, iLine);
		}
		catch(std::bad_alloc &)
		{
			m_pEx = NULL;
		}

		// in case of memory error, use the predefined memory error object
		if (m_pEx == NULL)
		{
			m_pEx = &s_MemoryError;
			m_pEx->AutoThrow();
		}

		// tell the object we have a reference
		m_pEx->AddRef();
	}
}

/******************************************************************************
   Function:    CErrorAutoThrow::CErrorAutoThrow
   Author:      Michael Tatarnikov
   Purpose:     Constructor from copy of CBException object.
   Returns:     void
   Exceptions:  
   Parameters:  const CBException & rkException - reference to exception object.
   Notes:       Use this form when returning an error after catching an
				exception.  Since the caught exception is a temporary object,
				a copy needs to be made.
******************************************************************************/
CErrorAutoThrow::CErrorAutoThrow(const CBException & rkException)
{
	// no error?  No need for a CBException object
	if (rkException.GetErrorCode() == EID_NO_ERROR)
		m_pEx = NULL;

	// the general case
	else
	{
		// create a CBException to hold the error
		try
		{
			m_pEx = new CBException(rkException);
		}
		catch(std::bad_alloc &)
		{
			m_pEx = NULL;
		}

		// in case of memory error, use the predefined memory error object
		if (m_pEx == NULL)
			m_pEx = &s_MemoryError;

		// otherwise, set to delete itself
		else
			m_pEx->AutoDelete();

		// arm the new exception
		m_pEx->AutoThrow();

		// tell the object we have a reference
		m_pEx->AddRef();
	}
}



/******************************************************************************
   Function:    operator =
   Author:      Michael Tatarnikov
   Purpose:     Assign to error code.
   Returns:     CErrorAutoThrow & - reference to *this.
   Exceptions:  
   Parameters:  BErrorCode jec - error code; see BErrorCodes.h.
   Notes:       
******************************************************************************/
CErrorAutoThrow & CErrorAutoThrow::operator =(BErrorCode jec)
{
	// throw away any old value
	if (m_pEx != NULL)
	{
		m_pEx->AutoThrow(false);
		m_pEx->Release();
	}

	// no error?  No need for a CBException object
	if (jec == EID_NO_ERROR)
		m_pEx = NULL;

	// memory error is a special case; we don't want to allocate new memory
	else if (jec == EID_MEMORY_ERROR)
	{
		m_pEx = &s_MemoryError;
		m_pEx->AddRef();
		m_pEx->AutoThrow();
	}
	// the general case
	else
	{
		m_pEx = new CBException(jec, NULL, true, true);	// auto-throw and auto-delete
		m_pEx->AddRef();
	}

	return *this;
}

/******************************************************************************
   Function:    CErrorAutoThrow::operator =
   Author:      Michael Tatarnikov
   Purpose:     Copy operator.
   Returns:     CErrorAutoThrow & - reference to *this.
   Exceptions:  
   Parameters:  const CErrorAutoThrow & rkoAT - object to copy.
   Notes:       If the object previously contained an error that has not been
				disarmed, that error will be thrown.
				The exception will be rearmed so that when ALL CErrorAutoThrow
				copies have been released, a CBException will be thrown.
******************************************************************************/
CErrorAutoThrow & CErrorAutoThrow::operator =(const CErrorAutoThrow & rkoAT)
{
	// check for same value; don't want to accidentally release
	if (m_pEx != rkoAT.m_pEx)
	{
		// get rid of any old data
		if (m_pEx != NULL)
			m_pEx->Release();

		// set the new value
		m_pEx = rkoAT.m_pEx;

		// do reference counting on the new pointer
		if (m_pEx != NULL)
			m_pEx->AddRef();
	}

	// rearm the exception
	if ((m_pEx != NULL) && (m_pEx->GetErrorCode() != EID_NO_ERROR))
		m_pEx->AutoThrow();

	return *this;
}

/******************************************************************************
   Function:    operator ==
   Author:      Michael Tatarnikov
   Purpose:     Compare to error code.
   Returns:     bool - true if the error is equal to the error code.
   Exceptions:  
   Parameters:  BErrorCode jec - error code; see BErrorCodes.h.
   Notes:       This function disarms the exception so that it will not
				auto-throw in the destructor.
******************************************************************************/
bool CErrorAutoThrow::operator ==(BErrorCode jec)
{
	if (m_pEx == NULL)
		return (jec == EID_NO_ERROR);

	// disarm the exception, it has been checked
	m_pEx->AutoThrow(false);

	// compare the error codes
	return (jec == m_pEx->GetErrorCode());
}

/******************************************************************************
   Function:    operator !
   Author:      Michael Tatarnikov
   Purpose:     Return true if there is no error.
   Returns:     bool - true if there is no error.
   Exceptions:  
   Parameters:  None
   Notes:       This function disarms the exception so that it will not
				auto-throw in the destructor.
******************************************************************************/
bool CErrorAutoThrow::operator !()
{
	if (m_pEx == NULL)
		return true;

	// disarm the exception, it has been checked
	m_pEx->AutoThrow(false);

	// return the error state
	return (m_pEx->GetErrorCode() == EID_NO_ERROR);
}

/******************************************************************************
   Function:    operator BErrorCode (cast)
   Author:      Michael Tatarnikov
   Purpose:     Get the underlying error code.
   Returns:     BErrorCode - the error code contained within the object.
   Exceptions:  
   Parameters:  None
   Notes:       This function disarms the exception so that it will not
				auto-throw in the destructor.
******************************************************************************/
CErrorAutoThrow::operator BErrorCode()
{
	if (m_pEx == NULL)
		return EID_NO_ERROR;

	// disarm the exception, it has been checked
	m_pEx->AutoThrow(false);

	// return the error state
	return m_pEx->GetErrorCode();
}

/******************************************************************************
   Function:    CErrorAutoThrow::GetException
   Author:      Michael Tatarnikov
   Purpose:     Get the underlying exception object.
   Returns:     CBException * - a pointer to the exception object contained
						within this object, or NULL if there is no error.
   Exceptions:  
   Parameters:  None
   Notes:       This function disarms the exception so that it will not
				auto-throw in the destructor.
******************************************************************************/
CBException * CErrorAutoThrow::GetException()
{
	if ((m_pEx == NULL) || (m_pEx->GetErrorCode() == EID_NO_ERROR))
		return NULL;

	// disarm the exception, it has been checked
	m_pEx->AutoThrow(false);

	// return the exception object
	return m_pEx;
}
