/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
#ifndef	BException_H
#define	BException_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include <exception>						// std::exception
#include "BErrorCodes.h"	// Error codes


//	The following global is used to turn logging of exceptions on or off:
#pragma warning(push)
#pragma warning(disable:4273)
extern bool g_bLogCBException;
#pragma warning(pop)

// macros to assert in debug, and throw an exception following the assertion
#ifdef _DEBUG
	#define VALIDATEPARAMRETURNJID( szExpr ) _ASSERTE(szExpr); if ( !(szExpr) ) return EID_INVALID_PARAMETER_PASSED;
	#define VALIDATEPARAMTHROWEXC( szExpr )	 _ASSERTE(szExpr); if ( !(szExpr) ) throw CBException( EID_INVALID_PARAMETER_PASSED, "Parameter validation failed", false, false, __FILE__, __LINE__ );
#else
	#define VALIDATEPARAMRETURNJID( szExpr ) if ( !(szExpr) ) return EID_INVALID_PARAMETER_PASSED;
	#define VALIDATEPARAMTHROWEXC( szExpr )  if ( !(szExpr) ) throw CBException( EID_INVALID_PARAMETER_PASSED );
#endif


#ifdef _DEBUG
	#define VALIDATEPTRRETURNJID( szExpr ) _ASSERTE(szExpr); if ( !(szExpr) ) return EID_MEMORY_ERROR;
	#define VALIDATEPTRTHROWEXC( szExpr )  _ASSERTE(szExpr); if ( !(szExpr) ) throw CBException( EID_MEMORY_ERROR, "Value must not be NULL", false, false, __FILE__, __LINE__ );
#else
	#define VALIDATEPTRRETURNJID( szExpr ) if ( !(szExpr) ) return EID_MEMORY_ERROR;
	#define VALIDATEPTRTHROWEXC( szExpr )  if ( !(szExpr) ) throw CBException( EID_MEMORY_ERROR );
#endif

// macros to assert in debug, and throw an exception following the assertion
#ifdef _DEBUG
	#define VALIDATEPARAM( szExpr )		_ASSERTE(szExpr); if ( !(szExpr) ) \
	throw CBException( EID_INVALID_PARAMETER_PASSED, "Parameter validation failed", false, false, __FILE__, __LINE__ );
#else
	#define VALIDATEPARAM( szExpr ) if ( !(szExpr) ) throw CBException( EID_INVALID_PARAMETER_PASSED );
#endif


// Macro to throw an exception, with file and line number where generated:
#ifdef _DEBUG
	#define BE_THROW(jecError, szDebugMsg) throw CBException(jecError, szDebugMsg, false, false, __FILE__, __LINE__)
#else
	#define BE_THROW(jecError, szDebugMsg) throw CBException(jecError)
#endif

// Macro to create a temp CErrorAutoThrow object with file and line number:
#ifdef _DEBUG
	#define AUTO_THROW(jecError, szDebugMsg) CErrorAutoThrow(jecError, szDebugMsg, __FILE__, __LINE__)
#else
	#define AUTO_THROW(jecError, szDebugMsg) CErrorAutoThrow(jecError)
#endif

#pragma warning(push)
#pragma warning(disable:4275)
/*************************************************************************
Class:				CBException
Base Class:			std::exception
Author:				Mike Tatarnikov / Paul Gruzdev / Oleg Monakhov
Description:		the exception class.	
Pure Virtual:		Yes
Known Issues:		None
Usage Notes:		Use this class to throw an exception based on a
					BErrorCode.
Overridable Methods:	Throw, what.
Override Notes:		This class can be subclassed to create specialized error
					classes for an application.  EVERY derived class should
					override the Throw() function for proper operation; the
					base class code can be copied verbatim.  You may also
					want to override the what() function to provide a string
					describing the error, especially if you can provide
					additional information about the error that would not be
					available to the catcher of the exception.  This string
					should be localized to the language of the app.
**************************************************************************/
class CBException : public std::exception
{
public:
	// Constructors/destructors.
	CBException();
	CBException(BErrorCode jecError, LPCTSTR pszDebugMsg = NULL, bool bAutoDelete = false,
		bool bAutoThrow = false, LPCSTR pszFile = NULL, int iLine = 0);
	CBException(const CBException &rkoException);
	virtual ~CBException();

	// Member access methods.
	inline void SetErrorCode(BErrorCode jecError) {m_ErrorCode = jecError;}
	inline BErrorCode GetErrorCode() const {return m_ErrorCode;}

	// Automatic flags
	void AutoDelete(bool bDelete = true);
	void AutoThrow(bool bThrow = true);

	// Reference counting
	ULONG AddRef();
	ULONG Release();

	// Helper for throwing a derived exception
	virtual void Throw();

	// Error messages
	virtual LPCSTR what() const;
	virtual LPCSTR what(HINSTANCE hResource) const;

	void SetWhat(LPCTSTR pszMessage);
	void OutputDebug() const;

	// Copy
	const CBException & operator =(const CBException & rkoException);

private:
	// Holds error code.
	BErrorCode m_ErrorCode;

	// Reference count
	ULONG	m_ulRefCount;

	// flags for automatic actions on last release
	bool	m_bAutoDelete;
	bool	m_bAutoThrow;

	// error message
	mutable LPTSTR	m_pszMsg;
	mutable TCHAR	m_szMsgBuf[512];	// do not access directly - use m_pszMsg
#ifdef _UNICODE
	mutable char	m_szMsgBufA[512];	// do not access directly - use m_pszMsg
#endif
	// debug info
	int		m_iFileLine;
	LPCSTR	m_pszFile;
	LPCTSTR	m_pszDebugMsg;
};

#pragma warning(pop)

/******************************************************************************
   Function:    CBException::CBException
   Author:      Michael Tatarnikov
   Purpose:     Default constructor.
   Returns:     void
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
inline CBException::CBException()
{
	m_ErrorCode = EID_NO_ERROR;
	m_ulRefCount = 0;
	m_bAutoDelete = false;
	m_bAutoThrow = false;
	m_pszMsg = NULL;
	m_pszDebugMsg = NULL;
	m_pszFile = NULL;
	m_iFileLine = 0;
}

/******************************************************************************
   Function:    CBException::CBException
   Author:      Michael Tatarnikov
   Purpose:     Constructor from error code.
   Returns:     void
   Exceptions:  
   Parameters:  BErrorCode jecError - error code; see BErrorCodes.h.
                LPCTSTR pszDebugMsg - message that can be seen in the debugger.
						This message will only be seen by programmers, and does
						not have to be localized.  This string is NOT copied,
						so the caller must ensure its validity for the life of
						the exception object.  Default is NULL.
                bool bAutoDelete - true if the exception object should delete
						itself when its reference count goes to 0 in the
						Release function.  Default is false.
                bool bAutoThrow - true if the exception object should be
						thrown when its reference count goes to 0 in the
						Release function.  Default is false.
                LPCSTR pszFile - the name of the source file where the error
						occurred.  This string is NOT copied, but this is not
						usually a problem since the __FILE__ macro is a static
						constant.  Default is NULL, meaning that the file was
						not specified.
                int iLine - the line number of the source file where the
						error occurred.  Default is 0, meaning the line number
						was not specified.
   Notes:       
******************************************************************************/
inline CBException::CBException(BErrorCode jecError, LPCTSTR pszDebugMsg,
								bool bAutoDelete, bool bAutoThrow, LPCSTR pszFile,
								int iLine)
{
	m_ErrorCode = jecError;
	m_ulRefCount = 0;
	m_bAutoDelete = bAutoDelete;
	m_bAutoThrow = bAutoThrow;
	m_pszMsg = NULL;
	m_pszDebugMsg = pszDebugMsg;
	m_pszFile = pszFile;
	m_iFileLine = iLine;
}

/******************************************************************************
   Function:    CBException::AutoDelete
   Author:      Michael Tatarnikov
   Purpose:     Set to auto-delete on last release.
   Returns:     Void
   Exceptions:  
   Parameters:  bool bDelete - true to delete, false to not delete.  Defaults
						to true.
   Notes:       Use this function only if the object was created on the heap
				via "new".  Setting this to true will delete the object when
				the reference count goes to zero, allowing its use in a smart
				pointer.
******************************************************************************/
inline void CBException::AutoDelete(bool bDelete)
{
	m_bAutoDelete = bDelete;
}

/******************************************************************************
   Function:    CBException::AutoThrow
   Author:      Michael Tatarnikov
   Purpose:     Set to throw an exception automatically on last release.
   Returns:     Void
   Exceptions:  
   Parameters:  bool bThrow - true to throw, false to not throw.  Defaults to
						true.
   Notes:       Setting auto-throw to true will cause the exception to throw
				itself when the reference count goes to zero.
******************************************************************************/
inline void CBException::AutoThrow(bool bThrow)
{
	m_bAutoThrow = bThrow;
}


/***********************************************************************************
***********************************************************************************/

/*************************************************************************
Class:				CErrorAutoThrow
Base Class:			(none)
Author:				Michael Tatarnikov
Description:		This class is intended as a return from other functions.
					It allows an either-or use of error returns - if the
					error code is retained or checked, it acts just like a
					BErrorCode.  If the return value is ignored, and it
					represents an error, an exception will be thrown.
Pure Virtual:		Yes
Known Issues:		None
Usage Notes:		This class will freely convert itself to or from a
					BErrorCode, but will throw an exception if you are
					not careful.  It should only be used as the return
					value from a function.
Overridable Methods:	none.
Override Notes:		NO derived classes should be created from this class.
					It was created as a very light-weight class, with only
					one data member and no virtual functions.  Because the
					destructor is not virtual, it cannot be expected to
					work properly if overridden.
**************************************************************************/
class CErrorAutoThrow
{
public:
	// Constructors / destructors
	CErrorAutoThrow();
	CErrorAutoThrow(const CErrorAutoThrow& rkoAT);

	void SetError(BErrorCode jec, LPCTSTR pszDebugMsg = NULL, LPCSTR pszFile = NULL, int iLine = 0);
	CErrorAutoThrow(BErrorCode jec, LPCTSTR pszDebugMsg = NULL, LPCSTR pszFile = NULL, int iLine = 0);

	CErrorAutoThrow(const CBException & rkException);
	~CErrorAutoThrow();


	// Operators
	CErrorAutoThrow &	operator =(BErrorCode jec);			// assignment
	CErrorAutoThrow &	operator =(const CErrorAutoThrow & rkoAT);
	bool				operator ==(BErrorCode jec);			// comparison
	inline bool			operator !=(BErrorCode jec) { return !(*this == jec); }
	bool				operator !();
						operator BErrorCode();


	// Accessors
	CBException *		GetException();
	bool				IsError();
	BErrorCode		Peek();

private:
	CBException *	m_pEx;
};

	
/******************************************************************************
   Function:    CErrorAutoThrow::CErrorAutoThrow
   Author:      Michael Tatarnikov
   Purpose:     Default constructor.
   Returns:     void
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
inline CErrorAutoThrow::CErrorAutoThrow()
{
	m_pEx = NULL;
}

/******************************************************************************
   Function:    CErrorAutoThrow::CErrorAutoThrow
   Author:      Michael Tatarnikov
   Purpose:     Copy constructor.
   Returns:     void
   Exceptions:  
   Parameters:  const CErrorAutoThrow & rkoAT - object to copy.
   Notes:       The exception will be rearmed so that when ALL CErrorAutoThrow
				copies have been released, a CBException will be thrown.
******************************************************************************/
inline CErrorAutoThrow::CErrorAutoThrow(const CErrorAutoThrow& rkoAT) 
{
	m_pEx = NULL;
	if (rkoAT.m_pEx != NULL)
		*this = rkoAT;
}

/******************************************************************************
   Function:    CErrorAutoThrow::~CErrorAutoThrow
   Author:      Michael Tatarnikov
   Purpose:     Destructor.
   Returns:     void
   Exceptions:  
   Parameters:  None
   Notes:       This destructor will cause a CBException to be thrown if the
				following conditions are met:
				1.	The object was set to an error code other than EID_NO_ERROR.
				2.	None of the disarming functions were called.  See the Notes
					under each function to see which will disarm the auto-throw.
				3.  This is the last object which points to the CBException
					object as a smart pointer.
				4.  The destructor was not called as part of a throw operation.
				The code to implement these conditions are implemented throughout
				this class and CBException::Release.
******************************************************************************/
inline CErrorAutoThrow::~CErrorAutoThrow()
{
	// The call to Release needs to come at the end of the function, since it may
	// throw an exception; ordinarily it is bad form to throw an exception from a
	// destructor, because you may leave an object un-destructed.
	if (m_pEx != NULL)
		m_pEx->Release();
}

/******************************************************************************
   Function:    CErrorAutoThrow::IsError
   Author:      Michael Tatarnikov
   Purpose:     Return error status without disarming auto-throw.
   Returns:     bool - true if there is an error.
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
inline bool CErrorAutoThrow::IsError()
{
	return ((m_pEx != NULL) && (m_pEx->GetErrorCode() != EID_NO_ERROR));
}

/******************************************************************************
   Function:    CErrorAutoThrow::Peek
   Author:      Michael Tatarnikov
   Purpose:     Get error code without disarming auto-throw.
   Returns:     BErrorCode - error code, or EID_NO_ERROR if none.
   Exceptions:  
   Parameters:  None
   Notes:       
******************************************************************************/
inline BErrorCode CErrorAutoThrow::Peek()
{
	return (m_pEx == NULL) ? EID_NO_ERROR : m_pEx->GetErrorCode();
}


#endif	// BException_H
