
#ifndef INC_CCFILE
#define INC_CCFILE

#include <fstream>
#include "pathname.h"
#include "memory.h"

using namespace std;
typedef INT_PTR			filedesc;
typedef INT32			FilePos;

typedef ios MyIOS;

// Default File Buffer Size

#define CCFILE_DEFAULTSIZE (1024)

// Default String Sizes

#define DEF_STRING_SIZE	32
#define STRING_64_SIZE	64
#define STRING_128_SIZE	128
#define STRING_256_SIZE	256

// CCMemTextFile::Read terminators 

#define END_OF_LINE '\n'
#define END_OF_FILE '\0'


// Enumeration of all the File Access rights available
enum FileAccess {CCMemRead, CCMemWrite};


class GZipFile; 
class GZipStream;
class costream;
struct IStream;

template< class chType = char > class CBaseSimpleStringBuffer
{
public:
	virtual operator const chType*() const PURE;
	virtual operator chType*() PURE;

	virtual size_t GetSize() const PURE;
};


template< int cch, class chType = char > class CSimpleStringBuffer : public CBaseSimpleStringBuffer< chType >
{
private:
	const size_t		m_cch;
	chType				m_Buffer[cch];

public:
	CSimpleStringBuffer() : m_cch( cch )
	{
		m_Buffer[0] = 0;
	}

	operator const chType*() const
	{
		return m_Buffer;
	}

	operator chType*()
	{
		return m_Buffer;
	}

	size_t GetSize() const
	{
		return m_cch;
	}
};


/********************************************************************************************

>	class CCFile : public CCObject

	Author:		Mario Shamtani
	Created:	31/8/1993
	Base Class:	CCObject
	Purpose:	Represents an abstract file class from which all other file classes will be 
				derived. It therefore acts as a generic interface to file handling of all
				sorts. A routine can read or write to a CCFile object without caring whether
				it is a real file on disk, or to the clipboard, or to a block of RAM
				somewhere.
				This class contains many pure virtual functions, which should be implemented
				by the derived classes.  The basic functions of this class are supplemented
				by the derived class CCLexFile, which provides functions to perform lexical
				analysis on a file object. (Note that CCLexFile is also	an abstract base
				class.) The main use of this class will be for the export and import filters.
				The class and its derivatives in this file used to just call InformError()
				about any errors that were encountered by the routines. The caller had no 
				control over this so the likelihood was that users would see a number of 
				errors instead of just one. This is still the default but now all error
				handling goes through a GotError function which can be made to call
				InformError() and/or throw an exception. This is controlled by two protected
				BOOL values	and can be either set in the constructor or via the public
				SetReportErrors	and	SetThrowExceptions calls. GotError also tries to force
				the bad(), fail() functions to fail on the next call so that no more file
				accesses should happen. 
	SeeAlso:	CCLexFile; CCDiskFile; CCMemFile; CCMemTextFile

********************************************************************************************/

class CCAPI CCFile
{
public:
	// Constructor with default parameters 
	//CCFile() {}
	CCFile(BOOL ErrorReporting = TRUE, BOOL ExceptionThrowing = FALSE);
	virtual ~CCFile() { /* empty */ }
	
	// Switch between text and binary modes. (filebuf::text is non-standard and the same as 0)
	virtual BOOL  setMode(int fileMode = 0)				PURE;

	// Check if the file is open or not.
	virtual BOOL  	 isOpen() const						PURE;

	// File pointer access/control.
	virtual CCFile&	 seekIn(FilePos) 					PURE;
	virtual CCFile&	 seekIn(LONG Offset,
							ios::seekdir Dir) 			PURE;
	virtual FilePos  tellIn() 							PURE;
	virtual CCFile&	 seek(FilePos pos) 					PURE;
	virtual FilePos	 tell() 							PURE;

	// Read functions.
	virtual CCFile&  read(void *buf, UINT length = 1)	PURE;
	virtual CCFile&  read( CBaseSimpleStringBuffer<char>* pBuf ) PURE;
	virtual CCFile&  read(char& buf)					PURE;

	// Write functions.
	virtual CCFile&  write(const void *buf, 
						   UINT length = 1)				PURE;
	virtual CCFile&  write(const std::string& buf, 
						   UINT length = 0) 			PURE;
	virtual CCFile&  write(char buf)					PURE;

	// Get the size of an open file in bytes.
	virtual LONG     Size() 			  				PURE;

	// Status functions.
	virtual BOOL     eof() const  		  				PURE;

	// Close the file.
	virtual void     close()			  				PURE;

	// Non-pure status functions.
	virtual BOOL     good() const;
	virtual BOOL     bad() const;
	virtual BOOL     fail() const;

	// Pure status setting functions.
	virtual void	SetBadState()						PURE;
	virtual void	SetGoodState()						PURE;

	// Allows access to the underlying fstream, is so required.
	// Mainly required for PNG/ZLib bits
	// All non-disk file classes just return NULL
	virtual iostream* GetIOFile()						PURE;

	// Flush the file object's buffer (if any).
	virtual CCFile&	 flush();

	// Character access.
	virtual CCFile&  get(char&);
	virtual CCFile&  put(char);
	
	// Stream operators for Strings.
	virtual CCFile&  operator>>(StringBase *buf);
	virtual CCFile&  operator<<(const StringBase& buf);

	// Error checking operators
	operator void*() const	{ return fail() ? 0 : (void*) this; }
	int operator!() const	{ return fail(); }

	// Public way of setting the two protected variables and reading their state
	virtual BOOL SetReportErrors( BOOL );
	virtual BOOL SetThrowExceptions( BOOL );
	virtual BOOL IsReportingSet();
	virtual BOOL IsThrowingSet();
	virtual BOOL IsMemFile() { return FALSE; }

	// Uses the above flags to determine whether to report errors and/or throw exceptions
	virtual void GotError( UINT errorID, const TCHAR* errorString);
	virtual void GotError( UINT errorID);

	// Find out whether we have a useful name to return which we can append to say
	// error messages to give feedback on what went wrong.
	virtual BOOL GetName(StringBase*) const;
	virtual PathName GetPathName() const;

	// Used to make short but informative error messages.
	virtual BOOL SetFileErrorMessage(UINT ErrorID);
	virtual BOOL SetZipFileErrorMessage(UINT ErrorID);

	// Get access to the file descriptor or handle for the file
	virtual filedesc GetFileHandle() const				PURE;

	// Allow setting of compression of output data
	virtual BOOL IsCompressionSet( );
	virtual BOOL SetCompression( BOOL );

	virtual BOOL InitCompression(BOOL Header = FALSE)	PURE;
	virtual BOOL StartCompression()						PURE;
	virtual BOOL StopCompression()						PURE;

protected:
	// Determines whether InformError is to be called and/or exception throwing is required  
	BOOL ReportErrors;						// Flag to say is we report errors
	BOOL ThrowExceptions;					// Flag to say whether we throw exceptions

	BOOL CompressData;						// Flag to say compress the output
	UINT GZipErrorID;						// current zip file error id

	// The size of the token buffer used for lexical analysis in the CLexFile class
	// Paul: I think now we can increase the buf size given that we got troubles on deleting it, for some 
	// EPS files, namely bug #5302, ferrari image.
	//const int TokenBufSize = 256;
	const static int lexTokenBufSize = 4096;
};



/********************************************************************************************

>	class CCLexFile : public CCFile

	Author:		Tim
	Created:	05/04/94
	Purpose:	Provide a derivation of the CCFile class that provides support for lexical
				analysis of a file. This allows a stream of tokens to be extracted from
				an ASCII text file.  This is an abstract class - you should not try to
				create an object of this class; use one of the derivations such as 
				CCDiskFile or CCMemFile.
				Note that the lexical analysis facilities are present in all concrete
				file classes, but the only overhead is a few tens of bytes on the size
				of the object - no buffers are allocated for lexical analysis unless you
				call the CCLexFile::InitLexer() function (and hence lexical analysis will
				not work *unless* you call this function!).
	SeeAlso:	CCFile; CCDiskFile; CCMemFile; CCMemTextFile

********************************************************************************************/

typedef enum
{
	TOKEN_EOF,
	TOKEN_EOL,
	TOKEN_NORMAL,
	TOKEN_COMMENT,
	TOKEN_STRING,
	TOKEN_LINE
} LexTokenType;

class CCLexFile : public CCFile
{
public:
	CCLexFile(BOOL ErrorReporting = TRUE, BOOL ExceptionThrowing = FALSE);
	~CCLexFile();

	BOOL InitLexer(BOOL IsSeekingRequired = FALSE,BOOL DoIgnoreStringEscapeCodes = FALSE);
	void DeinitLexer();

	// Change behaviour of token parser
	void SetWhitespace(const char*);
	void SetDelimiters(const char*);
	// Paul Gruzdev: add special delimiters that could be of more than one character wide.
	// This is for AI dictionaries that are enclosed in "<< >>"
	void AddSpecialDelimiter(const char* str);
	void SetIgnoreSpecialPrivateComment(bool b) {m_bIgnoreSpecialPrivateComment = b;}
	bool GetIgnoreSpecialPrivateComment(){return m_bIgnoreSpecialPrivateComment;}
	void SetCommentMarker(char);
	void SetStringDelimiters(const char*);

	// Read in the next token.
	BOOL GetToken();

	// Read in hexadecimal data
	BOOL GetHexToken();
	// Paul Gruzdev: Read in ASCII85 - encoded data skipping end of line
	int GetASCII85Token(BYTE* pbyBuffer, DWORD dwBufSize, bool bIgnoreComment=false);

	// Read until next token is TOKEN_NORMAL or TOKEN_STRING.
	// Delimiter tokens are skipped. 
	// Return FALSE if error occurs or if TOKEN_EOF encountered
	BOOL GetSimpleToken();

	// Read in a token in a line-based manner.  If the current input position is in the
	// middle of a line, the data up until the end of the line is read. If at the start
	// of a line, the whole line is returned.
	// The token is still examined to work out what it is, but obviously it may not
	// match any proper construct, in which case the type of the token is set to TOKEN_LINE.
	// (NB this is the only time that TOKEN_LINE is used)
	BOOL GetLineToken();

	// Put the current token back onto the input stream - i.e. the next time GetToken()
	// is called it will return the token it returned last time.
	// This is not nestable - you can only put back one token at once.
	void UngetToken();

	// Get status information...
	ULONG 		 GetLineNumber()	{ return Line;      }
	UINT  		 GetCharOffset()	{ return Ofs;       }
	virtual	LONG GetCharsRead()		{ return CharsRead; }
	LexTokenType GetTokenType() 	{ return TokenType; }
	const char  *GetTokenBuf() 		{ return TokenBuf;  }
	BOOL 		 IsLexerInitialised() { return LexerInitialised; }

	// Token writing routines.
	// These very simple functions allow your token output code to look
	// far neater. Besides, if we provide a GetToken() routine, it is only
	// common decency to provide a PutToken() routine too...
	BOOL	PutString(const StringBase& str,UINT length = 0,TCHAR* Sep=_T(" "));
	BOOL	PutToken(StringBase& str,UINT length = 0,TCHAR* Sep=_T(" "));
	BOOL	PutToken(const char* buf, char* Sep = " ");
	BOOL	PutToken(LONG n,TCHAR* Sep=_T(" "));
	BOOL	PutNewLine();
	void	IncIndent();
	void	DecIndent();
	void	SetIndentDelta(UINT d) { IndentDelta = d; }

	// The following 3 functions shouldn't really be public...pretend they're not here.
	// It is possible to call these directly, but it's not recommended unless you absolutely
	// have to, because you might stuff up the one character look-ahead.
	// (see epsfiltr.cpp for example of how to do it properly).
	// In fact, if you need to call these, tell me (Tim) and I'll work out what sort of
	// function we need to have here...I'll probably change it soon, anyway.
	void GetLine();
	void GetCh();

	// Helper function in case you ever need to call GetLine() directly and want to see
	// what you've read in.
	const char *GetLineBuf() { return LineBuf; }

	// this might look like a bodge but really it's a cunning way to avoid failure in 
	// LoadInitialSegment() which sadly always requests 1K in the asynch csae 
	// ( cause we dont know the filesize till the downloads over
	virtual void SetDontFail(BOOL state) { DontFail = state; };

	//Graham's HTML parsing code
public:
	//Entry points
	BOOL GetHTMLToken(BOOL fIgnoreEOL=TRUE, BOOL fCorrectCase=TRUE);

	char* GetHTMLTokenBuffer()
	{
		return m_pcHTMLBuffer;
	}

	BOOL IsHTMLTag()
	{
		return m_fIsTag;
	}

	BOOL IsEndOfHTMLFile()
	{
		return m_fEndOfHTMLFile;
	}

	PCSTR GetHTMLTagName( char strToReturn[256] );
	PCSTR GetHTMLParameterValue(char strToReturn[256], const String_256& strParameterName, BOOL fCorrectCase=TRUE);
	PCSTR GetHTMLParameterValue(char strToReturn[256], PCSTR szParameterName, BOOL fCorrectCase=TRUE);


	
protected:
	//HTML code member variables
	char* m_pcHTMLBuffer;		//The HTML buffer
	int m_iCharsInHTMLBuffer;	//Number of chars in the buffer
	int m_iLengthOfHTMLBuffer;	//Length of the buffer

	BOOL m_fIsTag;				//Whether the token in the HTML buffer is a tag

	BOOL m_fIsCharacterWaiting;	//Whether a character is waiting to be read
	char m_cWaitingCharacter;	//The character waiting to be read
		
	BOOL m_fEndOfHTMLFile;		//Whether we have reached the end of this file
	

protected:
	//HTML code internal functions
	void InitHTMLLexer();				//Initialises the HTML lexer

	void DeleteHTMLBuffer();			//Deletes the HTML buffer

	void AddToHTMLBuffer(char cToAdd);//Adds a character to the HTML buffer
	
	char PeekNextHTMLChar();	//Looks at the next character to be read, without reading it
	char ReadNextHTMLChar();	//Reads the next character from the file

	BOOL IsWhitespace(char cToTest);

	PSTR FindStringWithoutCase( PSTR strToSearch, PCSTR strToFind);


protected:
	// Variable to cache eof status; optimises lexical analysis.
	BOOL EOFFound;

	BOOL DontFail;

	// Input manipulation/parsing routines
	BOOL IsDelim();
	BOOL IsWhitespace();
	BOOL GetStringToken();
	int IsSpecialDelim();

	// Flag set to TRUE if the token read by GetToken was a delimiter
	BOOL DelimiterFound;

	// Flag to indicate whether or not to return the current token from GetToken(),
	// or get a new one from the file (the latter is the usual behaviour).
	BOOL TokenIsCached;

	// Buffer to hold the current line being read
	CSimpleStringBuffer<256, char>	LineBuf;

	// Buffer to hold the current token
	char		TokenBuf[lexTokenBufSize];

	// Indicates what kind of token was read.
	LexTokenType TokenType;
	
	// The current line number
	ULONG Line;
	
	// Index into the buffer (the current character)
	UINT Ofs;

	// The current character
	char Ch;
	
	// How many characters have been read.
	LONG CharsRead;

	// File position of start of last line read - so we can return the correct
	// tell info...
	std::streamoff LastLinePos;

	// Should we do the extra seeking code (which is a bit slower)
	BOOL SeekingRequired;

	// Should escape codes in strings be ignored?
	// e.g. Should "hello\r" turn into "hello" + CR, or should we ignore the \r code and hence leave
	// the string unaltered?
	// I.e. if IgnoreStringEscapeCodes is TRUE, then the '\' character in strings is left alone.
	BOOL IgnoreStringEscapeCodes;

	// Delimiter/white space definitions.
	// (These point to char arrays that hold a null-terminated set of characters that are
	// considered to be delimiters and whitespace respectively).
	char WhitespaceChars[100];
	char DelimiterChars[100];

	// Paul Gruzdev: add special delimiters that could be of more than one character wide.
	// This is for AI dictionaries that are enclosed in "<< >>"
	std::vector<std::string> m_SpecialDelimiter;
	// Paul: this flag is set when we want to decode tokens stored inside the special comments of a AI13 private
	// data. This is '%_' starting at the line start
	bool m_bIgnoreSpecialPrivateComment;

	// Comment definition - similar to whitespace and delimiters, except that it is
	// holds one character that starts a comment.
	// e.g. for PostScript, it would be "%".
	// Comments are deemed to last until the end of the line (i.e. the first CR or LF found).
	char CommentMarker;

	// These two characters are what bracket string tokens.
	// e.g. for C it would be "\"\"" and for PostScript it would be "()".
	char StringDelimiters[100];

	// The number of spaces to output at the start of every line.
	// This can be changed using IncIndent() and DecIndent() and is used
	// by PutNewLine().  Together they provide a mechanism for outputting 
	// structures of tokens.
	// IndentDelta is the amount that IndentSpaces is increased/decreased by
	// IncIndent() and DecIndent().  It can be changed using SetIndentDelta()
	LONG IndentSpaces;
	UINT IndentDelta;

private:
	// A flag to tell when the lexer has been initialised.
	BOOL LexerInitialised;
};



/********************************************************************************************
>	class CCStreamFile : public CCLexFile

	Author:		JustinF (based on a Mario-ism)
	Created:	13/8/96
	Purpose:	Represents a file class that uses an iostream type object for direct I/O.
				CCDiskFile is derived from this, using a standard fstream object for I/O.
				CCMemFile should be derived from this, but isn't, using a standard strstream
				object.  CCOleStream is also derived from this, using a custom costream
				object for I/O.
********************************************************************************************/

class CCAPI CCStreamFile : public CCLexFile
{
protected:
	char*		buffer;						// Fixed Size Buffer
	BOOL 		StreamFileInitialised;
											// Inherited from CCFile
protected:
	
	iostream* 	IOFile;						// Underlying implementation is an iostream
	GZipFile*	GZFile;						// the gz file class we might use 
	GZipStream*	GZStream;					// the gz stream class we might use 
	int			ModeOfFile;					// file mode we are currently using, use ios flags	

public:
											// Constructor
	CCStreamFile(iostream* pstream,
				 UINT bufferSize = CCFILE_DEFAULTSIZE,
				 BOOL ErrorReporting = TRUE,
				 BOOL ExceptionThrowing = FALSE);

											// Constructor that opens stream
	CCStreamFile(const TCHAR* pcszPath, int fileMode,
				 int fileAccess  = MyIOS::_Openprot,
				 UINT bufferSize = CCFILE_DEFAULTSIZE,
				 BOOL ErrorReporting = TRUE,
				 BOOL ExceptionThrowing = FALSE);

	virtual ~CCStreamFile();

	// Paul: introduce one more open flag for CCDiskFile: real_binary, used in eps filter only
	static const int real_binary = 0x100;
	virtual CCFile& txt_read(CBaseSimpleStringBuffer<char>* pString);
	int& GetModeOfFile(){return ModeOfFile;}

	// As at present the constructors have errors in them due to memory claims rather
	// than doing these in a init function, we will introduce a flag to say whether
	// things have gone well so that if errors are suppressed then a user can check. 
	BOOL IsInited();

	// Call this to change the iostream this classes uses.
	virtual iostream* SetStream(iostream* pStream);

	// Stream files are always open and cannot be closed.
	virtual BOOL isOpen() const OVERRIDE;

	// This may be very slow for generic iostreams, or even impossible.
	virtual LONG Size() OVERRIDE;

	// This closes a CCStreamFile.  It is a no-op.
	virtual void close() OVERRIDE;

	// File pointer access/control.
	virtual CCFile&	 seekIn(FilePos) OVERRIDE;
	virtual CCFile&	 seekIn(LONG Offset, ios::seekdir Dir) OVERRIDE;
	virtual FilePos  tellIn() OVERRIDE;
	virtual CCFile&  seek(FilePos pos) OVERRIDE;
	virtual FilePos	 tell() OVERRIDE;

	// Read functions.
	virtual CCFile&  read(void* buf, UINT length = 1) OVERRIDE;
	virtual CCFile&  read( CBaseSimpleStringBuffer<char>* pString ) OVERRIDE;
	virtual CCFile&  read(char& buf) OVERRIDE;

	// Write functions.
	virtual CCFile&  write(const void* buf, UINT length = 1) OVERRIDE;
	virtual CCFile&  write( const std::string& buf, UINT length = 0) OVERRIDE;
	virtual CCFile&  write(char buf) OVERRIDE;

	// Character access
	virtual CCFile&  get(char& buf) OVERRIDE;
	virtual CCFile&  put(char buf) OVERRIDE;

	// Flush the file object's buffer (if any).
	virtual CCFile&	 flush() OVERRIDE;

	// Status functions.
	virtual BOOL     good() const OVERRIDE;
	virtual BOOL     bad() const OVERRIDE;
	virtual BOOL     fail() const OVERRIDE;
	virtual BOOL     eof() const OVERRIDE;

	// Status setting function so that bad() will fail and somebody can force a good state.
	virtual void	SetBadState() OVERRIDE;
	virtual void	SetGoodState() OVERRIDE;

	// Allows access to the underlying iostream, if so required.
	// Mainly required for PNG/ZLib bits.  All non-disk file classes just return NULL.
	virtual iostream* GetIOFile() OVERRIDE { return IOFile; }

	virtual BOOL InitCompression(BOOL Header = FALSE) OVERRIDE;
	virtual BOOL StartCompression() OVERRIDE;
	virtual BOOL StopCompression() OVERRIDE;

	virtual	LONG GetCharsRead() OVERRIDE;

	// Get access to the file descriptor or handle for the file.  This is faked for
	// plain iostreams.
	virtual filedesc GetFileHandle() const OVERRIDE;
};



/********************************************************************************************
>	class CCOleStream : public CCStreamFile

	Author:		JustinF
	Created:	17/8/96
	Purpose:	Represents an IStream within an OLE Structured Storage.  Uses the custom
				iostream class costream for its implementation.
	SeeAlso:	IStream; IStorage; costream; costreambuf; CCStreamFile; CCDiskFile; CCLexFile
********************************************************************************************/

class CCAPI CCOleStream : public CCStreamFile
{
public:

	// Creation & destruction.
	CCOleStream(IStream* pIStream = 0,
				UINT bufferSize = CCFILE_DEFAULTSIZE,
				BOOL ErrorReporting = TRUE,
				BOOL ExceptionThrowing = FALSE);

	CCOleStream(IStream* pIStream,
				char* pBuf,
				UINT bufferSize,
				BOOL ErrorReporting = TRUE,
				BOOL ExceptionThrowing = FALSE);

	CCOleStream(PathName path, int fileMode,
				int fileAccess = 0,
				UINT bufferSize = CCFILE_DEFAULTSIZE,
				BOOL ErrorReporting = TRUE,
				BOOL ExceptionThrowing = FALSE);

	// Attributes.
	virtual BOOL isOpen() const OVERRIDE;
	virtual LONG Size() OVERRIDE;
	virtual void SetSize(LONG NewSize);
	virtual BOOL setMode(int fileMode = 0) OVERRIDE;

	// Operations.
	virtual BOOL open_file(PathName fPath, int fileMode, int fileAccess = 0);
	virtual BOOL attach(IStream* pIstream);
	virtual void close() OVERRIDE;

protected:

	// These are shorthand.  Use them to refer to CCStreamFile::IOFile as a costream instead
	// of a plain ol' iostream.

	// following two cause memory exceptions !!!!!!  seems to be related to use of inline
	// always return TRASH !!!!!  Only solution was to make NOT inline ....

	// see more notes in ccfile.cpp (search on "causes it to throw a wobobley") ....
	
	//inline costream& GetCoStream() { return (costream&) *IOFile; }
	//inline const costream& GetCoStream() const { return (const costream&) *IOFile; }

	costream& GetCoStream() const;
};


// WEBSTER-Martin-29/12/96
// no accusoft stuff
#ifndef WEBSTER

/********************************************************************************************
>	class CCOleAccusoftStream : public CCOleStream

	Author:		JustinF
	Created:	3/10/96
	Purpose:	CCOleStream that manages the Accusoft stream switching as well.
	SeeAlso:	CCOleStream
********************************************************************************************/

class CCOleAccusoftStream : public CCOleStream
{
public:

	// Switch Accusoft to read from OLE streams.
	CCOleAccusoftStream(IStream* pIStream = 0,
						UINT bufferSize = CCFILE_DEFAULTSIZE,
						BOOL ErrorReporting = TRUE,
						BOOL ExceptionThrowing = FALSE);

	// Switch Accusoft to read from OLE streams.
	CCOleAccusoftStream(PathName path, int fileMode,
						int fileAccess = 0,
						UINT bufferSize = CCFILE_DEFAULTSIZE,
						BOOL ErrorReporting = TRUE,
						BOOL ExceptionThrowing = FALSE);

	// Switch Accusoft back again.
	virtual ~CCOleAccusoftStream();
};

#endif //WEBSTER


/********************************************************************************************
>	class CCDiskFile : public CCStreamFile

	Author:		Mario Shamtani
	Created:	31/8/1993          
	Purpose:	Represents a file class that can offer buffered disk file I/O in both
				text and binary modes.
				Derived from the CCLexFile class so you can use lexical analysis
				functions on it.
********************************************************************************************/

class CCAPI CCDiskFile : public CCStreamFile
{
protected:

	PathName    filePath;					// PathName of File
	mutable int	m_fdClient;					// The filedesc (if any) created for a caller

	// These are shorthand.  Use them to refer to CCStreamFile::IOFile as an fstream instead
	// of a plain ol' iostream.
	inline fstream& GetFileStream() { return (fstream&) *IOFile; }
	inline const fstream& GetFileStream() const { return (const fstream&) *IOFile; }

public:

	// Creation & destruction.
	CCDiskFile(UINT bufferSize = CCFILE_DEFAULTSIZE,
			   BOOL ErrorReporting = TRUE,
			   BOOL ExceptionThrowing = FALSE);

	  // Creation & destruction.
	CCDiskFile(fstream * pfstream,
			   UINT bufferSize = CCFILE_DEFAULTSIZE,
			   BOOL ErrorReporting = TRUE,
			   BOOL ExceptionThrowing = FALSE);

	CCDiskFile(PathName fPath, int fileMode, 
			   int fileAccess  = MyIOS::_Openprot,
			   UINT bufferSize = CCFILE_DEFAULTSIZE,
			   BOOL ErrorReporting = TRUE,
			   BOOL ExceptionThrowing = FALSE);

	CCDiskFile(fstream * pfstream,
			   PathName fPath, int fileMode, 
			   int fileAccess  = MyIOS::_Openprot,
			   UINT bufferSize = CCFILE_DEFAULTSIZE,
			   BOOL ErrorReporting = TRUE,
			   BOOL ExceptionThrowing = FALSE);

	~CCDiskFile();

	// Open a file on a disk.
	virtual BOOL open(PathName fPath, int fileMode, int fileAccess = MyIOS::_Openprot);

	// Switch between text and binary modes.
	virtual BOOL setMode(int fileMode = 0) OVERRIDE;

	// Check if the file is open or not.
	virtual BOOL isOpen() const OVERRIDE;

	// This closes a CCDiskFile.
	virtual void close() OVERRIDE;

	// Get the size of an open file in bytes.
	virtual LONG Size() OVERRIDE;

	// Find out whether we have a useful name to return which we can append to say
	// error messages to give feedback on what went wrong.
	virtual BOOL GetName( StringBase *name) const OVERRIDE;
	virtual PathName GetPathName() const OVERRIDE;

	// Get access to the file descriptor or handle for the file
	virtual filedesc GetFileHandle() const OVERRIDE;

	inline CCFile& WriteASCII( const StringBase& buf )
	{
#if defined(_UNICODE)
		USES_CONVERSION;
        return write( (void*)W2A( PCTSTR(buf) ), buf.Length() );
#else
		return write( buf );
#endif
	}
};



/********************************************************************************************

>	class CCMemFile : public CCLexFile

	Author:		Mario Shamtani
	Created:	6/9/1993          
	Purpose:	Represents a memory IO file - IO is in byte units.
	Errors:		None.

********************************************************************************************/

class CCAPI CCMemFile : public CCLexFile
{
protected:
	
	MHANDLE MemHandle;
	BYTE*	MemFile;
	ULONG	AllocSize;		// this is the pre-allocated size. The file can be written up to this limit, then reallocated again
	ULONG 	FileSize;		// this is the actual file size. When we seek to the end - we seek to this pos
	ULONG	CurrentPos;		// it's the position we are about to write to
	BOOL	IsOpen;
	UINT	FileProt;
	BOOL	WasError;									// TRUE if error occurred
	BOOL	OwnMemHandle; // When opened to write, we sometimes want to then read from the mem file. This flag tells we need to free mem block on close after reading.
	
	// A flag to tell when the disk file has been initialised ok.
	BOOL 		MemFileInitialised;

public:
	CCMemFile(BOOL ErrorReporting = TRUE, BOOL ExceptionThrowing = FALSE);
	
	CCMemFile(void* pFile, UINT size, FileAccess fProt = CCMemRead,
	          BOOL ErrorReporting = TRUE, BOOL ExceptionThrowing = FALSE);
	
	~CCMemFile();
	
	// As at present the constructors have errors in them due to memory claims rather
	// than doing these in a init function, we will introduce a flag to say whether
	// things have gone well so that if errors are supressed then a user can check. 
	BOOL IsMemFileInited();

	BOOL GetBuffer(BYTE** ppBuffer, ULONG* pSize);
	BOOL GrowMemFile(UINT increase = CCFILE_DEFAULTSIZE);
	BOOL ChangeFileAccess(FileAccess fProt = CCMemRead);

	BOOL open(void* pFile, UINT size, FileAccess fProt = CCMemRead);
	MHANDLE Detach();
	PBYTE GetMemoryBuffer();

	// Switch between text and binary modes.
	virtual BOOL  setMode(int fileMode = 0) OVERRIDE;

	// Check if the file is open or not.
	virtual BOOL  	 isOpen() const OVERRIDE;
	virtual BOOL IsMemFile() { return TRUE; }

	// File pointer access/control.
	virtual CCFile&	 seekIn(FilePos) OVERRIDE;
	virtual CCFile&	 seekIn(LONG Offset, 
							ios::seekdir Dir) OVERRIDE;
	virtual FilePos  tellIn() OVERRIDE;
	virtual CCFile&  seek(FilePos pos) OVERRIDE;
	virtual FilePos	 tell() OVERRIDE;

	// Read functions.
	virtual CCFile&  read(void *buf, UINT length = 1) OVERRIDE;
	virtual CCFile&  read( CBaseSimpleStringBuffer<char>* pBuf) OVERRIDE;
	virtual CCFile&  read(char& buf) OVERRIDE;

	// Write functions.
	virtual CCFile&  write(const void *buf, UINT length = 1) OVERRIDE;
	virtual CCFile&  write(const std::string& buf, UINT length = 0) OVERRIDE;
	virtual CCFile&  write(char buf) OVERRIDE;

	// Get the size of an open file in bytes.
	virtual LONG     Size() OVERRIDE;

	// Status functions.
	virtual BOOL     good() const OVERRIDE;
	virtual BOOL     bad() const OVERRIDE;
	virtual BOOL     fail() const OVERRIDE;
	virtual BOOL     eof() const OVERRIDE;

	// Status setting function so that bad() will fail and somebody can force a good state.
	virtual void	SetBadState() OVERRIDE;
	virtual void	SetGoodState() OVERRIDE;

	// Allows access to the underlying fstream, is so required.
	// Mainly required for PNG/ZLib bits
	// All non-disk file classes just return NULL
	virtual iostream* GetIOFile() OVERRIDE { return NULL; }

	// Find out whether we have a useful name to return which we can append to say
	// error messages to give feedback on what went wrong.
	virtual BOOL GetName( StringBase *name ) const OVERRIDE;

	// Get access to the file descriptor or handle for the file
	virtual filedesc GetFileHandle() const OVERRIDE;

	// Close the file.
	virtual void     close() OVERRIDE;

	virtual BOOL InitCompression(BOOL Header = FALSE) OVERRIDE;
	virtual BOOL StartCompression() OVERRIDE;
	virtual BOOL StopCompression() OVERRIDE;
};



/*********************************************************************************************

>	class CCMemTextFile : public CCMemFile

	Author:		Mario Shamtani
	Created:	8/9/1993          
	Purpose:	Represents a character memory IO file which can cope with unicode.
	Errors:		None.

*********************************************************************************************/

class CCAPI CCMemTextFile : public CCMemFile
{
public:
	CCMemTextFile(BOOL ErrorReporting = TRUE, BOOL ExceptionThrowing = FALSE);
	
	CCMemTextFile(char* pFile, UINT size, FileAccess fProt = CCMemRead,
				  BOOL ErrorReporting = TRUE,
				  BOOL ExceptionThrowing = FALSE);
	
	~CCMemTextFile();
	
	BOOL open(TCHAR* pFile, UINT size, FileAccess fProt = CCMemRead);

	// Read functions.
	//virtual CCFile&  read(void *buf, UINT length = 1);
	virtual CCFile&  read( CBaseSimpleStringBuffer<char>* pBuf) OVERRIDE;
	virtual	CCFile&  read(char& buf) OVERRIDE;

	// Write functions.
	//virtual CCFile&  write(const void *buf, UINT length = 1);
	virtual CCFile&  write( const std::string& buf, UINT length = 0 ) OVERRIDE;
	virtual	CCFile&  write(char buf) OVERRIDE;

	// Status functions.
	virtual BOOL     eof() const OVERRIDE;
};




/*
	Alternate C-style I/O interfaces.  These global functions substitute for the standard
	C library's read, write, and seek functions, in particular those calls used by the
	Accusoft library.  The "file-descriptor" passed as the first parameter should be
	obtained by calling the fake_filedesc function.

	CAUTION: this won't work if sizeof(int) != sizeof(iostream*) !!

	Some wobbly code to show the general idea:-
		
		// Some C++ code that sets up and calls ...
		fstream fs("c:\\in.txt", ios::in);
		costream cs("c:\\out.txt", ios::out | costream::convert | costream::transacted);
		int fd_fs = fake_filedesc(fs);
		int fd_os = fake_filedesc(cs);
		
		// ... some C code that does I/O with C++ streams.
		char buf[256];
		int n = stream_read(fd_fs, buf, sizeof(buf));
		if (n > 0 && stream_seek(fd_os, 0, SEEK_END)) stream_write(fd_os, buf, n);
*/

INT_PTR fake_filedesc(iostream& io);
iostream* unfake_filedesc(INT_PTR nFileDesc);

#define ACCUAPI_IO		FAR __cdecl

extern "C"
{
	int  ACCUAPI_IO stream_read(int nFileDesc,  char* pchBuffer, unsigned nBufSize);
	int  ACCUAPI_IO stream_write(int nFileDesc, char* pchBuffer, unsigned nBufSize);
	long ACCUAPI_IO stream_seek(int nFileDesc,  long nOffset,	 unsigned fSeekType);
}


/*********************************************************************************************

>	class DisableFileError

	Author:		LukeH
	Created:	31/03/2015          
	Purpose:	A scoped setr to stop CCLexFile reporting an error
	Errors:		None.

*********************************************************************************************/

class DisableFileError
{
public:
	DisableFileError(CCLexFile* pFile) : m_pFile(pFile)
	{
		m_bOldReportErr = pFile->SetReportErrors(FALSE);
	}

	~DisableFileError()
	{
		m_pFile->SetReportErrors(m_bOldReportErr);
	}

private:
	CCLexFile*		m_pFile;
	BOOL			m_bOldReportErr;
};

#endif
