/* ================================XARAHEADERSTART===========================
 
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/

=================================XARAHEADEREND============================*/
/*
	$Header: /Camelot/winoil/costream.h 1     21/08/96 2:25p Justinf $

	costream.h

	This class is a stream derivative that can read and write COM structured storage.
	It is similar to the standard fstream disk-file class.
*/

#ifndef COSTREAM_H
#define COSTREAM_H

#include <iostream>
#include <vector>
#include <ole2.h>


/********************************************************************************************
>	class costream : public iostream

	Author:		JustinF
	Created:	5/8/96
	Purpose:	Provides a programming interface to COM structured storage with very similar
				semantics to the standard fstream class.

				Some points to remember:-
					(i) The costream and fstream classes are both derived from iostream.
						Anything that works with istream, ostream, or iostream, which is
						most things, especially the Standard Template Library, will work
						with costream and fstream in a similar manner.

				   (ii) There is no icostream or ocostream classes, as yet.  A plain
						costream can serve as either if opened in the appropriate access
						mode (ios::in for input, ios::out for output).

				  (iii) An costream, like an fstream, does not maintain independent read
						and write seek pointers, or independent read and write buffers.
						There is one buffer and one pointer for both read and write, ie.
						tellg() and tellp() always return the same value.

				   (iv) There is no "text" or "binary" mode.  All streams are "binary",
						and the ios::text open-mode is ignored.

				   (v)  Not all access modes, as defined by the OLE docs, are currently
						supported by this release of structured storage.  This is a
						limitation of COM, not the costream class.  In particular you
						can only open storages and streams with costream::sh_none, and
						streams do not support costream::convert or costream::transacted.
						The costream class will check for unimplemented flags and ignore
						them as appropriate (note that it will complain if it is passed
						flags that are invalid, rather than unimplemented).

				   (vi) The classes are not yet thread-safe, which would require bracketing
						functions that set the state of the objects with calls to ios::lock
						and ios::unlock, to guarantee synchronisation.

				  (vii) The costream object assumes responsibility for releasing an IStream
						interface it uses.  If you use an IStream interface beyond the
						lifetime of a costream that uses it, you must call AddRef on the
						IStream before passing it to the costream, for example:-

							// Get an IStream interface.
							IStream* pIStream;
							SomeFunc(&pIStream);

							// Possibly do something with it and a costream.
							if (fSomeTest)
							{
								// Attach a costream to the IStream.
								costream cs(pIStream);

								// Make sure costream won't release the IStream when it is
								// destroyed at the end of this block.
								cs->AddRef();
							}

							// Do something else with the IStream, it still exists.
							pIStream->Revert();

						The extra IStream::AddRef "cancels out" the call to IStream::Release
						made by the costream destructor.

				 (viii) The Structured Storage specification allows for streams up to 2^64
						bytes in size, while the C++ stream library, and the Windows
						implementation of the specification, only allows streams to contain
						at most 2^32 bytes.					

	SeeAlso:	iostream; fstream; costreambuf; IStream; IStorage; IRootStorage;
				StgOpenStorage
********************************************************************************************/

class costreambuf;

class costream : public iostream
{
public:

	// Access modes for creating and opening storages and streams.
	enum open_mode
	{
		simple	   = ios::binary << 1,	// binary assumed to be the highest ios::open_mode
		transacted = simple << 1,
		priority   = transacted << 1,
		temp	   = priority << 1,
		convert	   = temp << 1,
		noscratch  = convert << 1
	};

	// Access modes for sharing storages and streams.
	enum share_mode
	{
		sh_none  = 1,					// all assumed to be equal to filebuf::share_mode
		sh_read	 = 2,
		sh_write = 4
	};

	// Creation & destruction.
	costream(IStream* pIStream = 0);
	costream(const TCHAR* pcszFilePath, int nMode, int nProt = sh_read | sh_write);
	costream(IStorage* pIStorage, const TCHAR* pcszStreamName,
			 int nMode, int nProt = sh_read | sh_write);
#if defined(NEED_TO_SPEC_BUFFER_IN_COSTRAMBUF)
	costream(IStream* pIStream, char* pchReserve, int nLength);
#endif
	costream(const costream& other);
	virtual ~costream();

	// Attributes.
	virtual int is_open() const;
	virtual operator IStream*() const;
	virtual costreambuf* rdbuf() const;
	virtual int get_access_mode(int* pnMode, int* pnProt) const;

	// Operations.
	virtual streambuf* setbuf(char* pchReserve, int nLength);
	virtual HRESULT set_size(size_t cbNewSize);
	virtual HRESULT attach(IStream* pIStream);
	virtual HRESULT close();
	virtual HRESULT open_file(const TCHAR* pcszFilePath,
							  int nMode, int nProt = sh_read | sh_write);
	virtual HRESULT open_stream(IStorage* pIStorage, const TCHAR* pcszStreamName,
								int nMode, int nProt = sh_read | sh_write);

protected:

	// Override this to be informed immediately after a stream error is detected.
	virtual void on_fail(HRESULT hr);
};



/********************************************************************************************
>	class costreambuf : public streambuf

	Author:		JustinF
	Created:	5/8/96
	Purpose:	Provides a concrete streambuf derivative that can read and write IStreams in
				COM structured storage files.  This class manages the direct connection with
				the COM libraries and is the key to integrating them with the C++ stream
				library.
	SeeAlso:	streambuf; ios; costream
********************************************************************************************/

// This is the default for the size of the reserve area of an costreambuf.  It
// is the same as the COM implementation's stream "granularity".
const static size_t COM_BUFSIZ = 512;

class costreambuf : public streambuf
{
public:
	// Creation & destruction.
	costreambuf(IStream* pIStream = 0, size_t cbBuffer = COM_BUFSIZ);
	costreambuf(const TCHAR* pcszFilePath,
				int nMode, int nProt = costream::sh_read | costream::sh_write, 
				size_t cbBuffer = COM_BUFSIZ);
	costreambuf(IStorage* pIStorage, const TCHAR* pcszStreamName,
				int nMode, int nProt = costream::sh_read | costream::sh_write, 
				size_t cbBuffer = COM_BUFSIZ);
#if defined(NEED_TO_SPEC_BUFFER_IN_COSTRAMBUF)
	costreambuf(IStream* pIStream, char* pchReserve, int nLength, 
				size_t cbBuffer = COM_BUFSIZ);
#endif
	costreambuf(const costreambuf& other);
	virtual ~costreambuf();

	// Attributes.
	virtual int is_open() const;
	virtual operator IStream*() const;
	virtual HRESULT get_access_mode(int* pnMode, int* pnProt) const;

	// Operations.
	virtual costreambuf& operator=(const costreambuf& other);
	virtual streambuf* setbuf(char* pchReserve, int nLength);
	virtual HRESULT set_size(size_t cbNewSize);
	virtual HRESULT attach(IStream* pIStream);
	virtual HRESULT close();
	virtual HRESULT open_file(const TCHAR* pcszFilePath, int nMode,
							  int nProt = costream::sh_read | costream::sh_write);
	virtual HRESULT open_stream(IStorage* pIStorage, const TCHAR* pcszStreamName,
								int nMode,
								int nProt = costream::sh_read | costream::sh_write);
	virtual pos_type seekoff(off_type pos, ios_base::seekdir eDir, ios_base::openmode nMode = ios::in | ios::out);
	virtual pos_type seekpos(pos_type pos, ios_base::openmode nMode = ios::in | ios::out);
protected:

	// Stream implementation.  These are required by the streambuf abstract base class.
	virtual int overflow(int nChar = EOF);
	virtual int underflow();
	virtual int sync();

	// These perform the hard work of the above overridables, and return better error
	// codes as well.
	virtual HRESULT hr_overflow(int nChar = EOF);
	virtual HRESULT hr_underflow(int* pnChar);
	virtual HRESULT hr_sync();
	virtual HRESULT hr_seekoff(pos_type pos, ios_base::seekdir eDir,
						pos_type* pnNewPos = 0, ios_base::openmode nMode = ios::in | ios::out);
	
	// Make this costreambuf into an independent copy with an identical state of the
	// other costreambuf. 
	virtual HRESULT do_clone(const costreambuf& other);

	// This translates an ios type of access mode, eg. ios::noreplace, into a structured
	// storage access mode, eg. STGM_FAILIFTHERE.  Derived classes can override to use
	// new structured storage flags, as they are defined.
	virtual int ios_to_stgm(int* pnMode, int* pnProt, DWORD* pgrfResult) const;

	// This performs the reverse mapping, from a STGM_* access mode to the ios/costream
	// equivalent.
	virtual int stgm_to_ios(DWORD* pgrfResult, int* pnMode, int* pnProt) const;

	// This converts an ANSI string into Unicode.  The pointer returned is temporary and
	// not thread-safe.
	static const WCHAR* convert_string(const CHAR* pcsz);

private:

	// Data members.
	IStream*  m_pIStream;		// the IStream this costreambuf is attached to
	IStorage* m_pIStorage;		// the IStream's parent IStorage, if opened by STGM_CONVERT
	short	  m_fAppend;		// if non-zero then append writes to the end of the stream

	std::vector<char>	m_vecBufferIn;
	std::vector<char>	m_vecBufferOut;
	enum{ eEither, eRead, eWrite }	m_eDirection;

	void InitBuffers();

	// This class makes sure that the structured storage libraries are initialised before
	// any costreambuf objects are created.
	class init
	{
	public:

		// Creation & destruction.
		init(LPVOID lpvReserved = 0);
		~init();

	private:

		// Data members.
		short m_fInitialised;
		static init the_instance;
	};
};

#endif	// !COSTREAM_H
