/* ================================XARAHEADERSTART===========================
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/
=================================XARAHEADEREND============================*/
// $Header: /Camelot/winoil/fixmem.cpp 12    8/10/96 15:44 Richard $

// fixed memory allocator (used in preference to malloc/free etc)


#include "stdafx.h"
#include "fixmem.h"
#include <new.h>
#include <malloc.h>
//#include "resource.h"

#if WIN32
#define	checksize(x)	FALSE
#else
#define	checksize(x)	( (x) >= _HEAP_MAXREQ )
#endif


#ifdef _DEBUG

typedef struct CCMallocList
{
	LONG GuardWord;
	ULONG Size;
	CCMallocList *Next;
	CCMallocList *Prev;
	char   Filename[22];
	int	   LineNum;
} CCMallocList;

static CCMallocList *TraceListHead = NULL;


#define	MAGIC_LONGWORD	0xDEADBEEF
#define	MAGIC_EXTRA		sizeof(CCMallocList)


#ifdef _OUTPUT_DEBUG
static LONG TotalMalloced = 0;
#endif

/********************************************************************************************

	static void *CheckMagic( void* Block, BOOL Destroy )

	Author: 	Andy
	Created:	15/6/94
	Inputs:		Block - pointer to memory block. If Destroy is TRUE will remove magic
				marker so detect double-freeing.
	Outputs:	None
	Returns:	Pointer before magic in block. Returns NULL if Block was null on input, or
				if the block is not recognised as one of ours.
	Purpose:	For debug builds to ensure the magic longword marker is in the block.
				If Destroy is TRUE then removes block from memory tracing list.
				Stubbed away in retail builds.
	Scope:		Static to fixmem.cpp

********************************************************************************************/

static void *CheckMagic( void* Block, BOOL Destroy )
{
	if (Block != NULL)
	{
		CCMallocList *p = (CCMallocList*)Block;

		// decrement pointer by sizeof(CCMallocList)
		p--;

		if ( p->GuardWord == MAGIC_LONGWORD )
		{
			if (Destroy)
			{
				p->GuardWord = 0xF000BAAA;						// if it was our marker then destroy it

				if (p == TraceListHead)
					TraceListHead = p->Next;

				if (p->Next != NULL)
					p->Next->Prev = p->Prev;

				if (p->Prev != NULL)
					p->Prev->Next = p->Next;
			}
		}
		else
		{
			// if marker not there then either we didn't alloc it, or it has been freed for
			// the second time
			ENSURE (FALSE, "CCFree/realloc called on memory that was not CCMalloced");
			return NULL;								// so no-one tries to use it
		}
		Block = p;										// use real address to free
	}
	return Block;
}

/********************************************************************************************

	static void *StoreMagic( void *Block , ULONG Size , LPCSTR Filename , int Line )

	Author: 	Andy (and Gerry)
	Created:	15/6/94
	Inputs:		Block - pointer to memory block. Stores magic longword in there.
	Outputs:	None
	Returns:	The address just past that to the caller for use by the user.
	Purpose:	For debug builds to insert magic longword marker and other memory tracing
				info in the block. Stubbed away in retail builds.
	Scope:		Static to fixmem.cpp

********************************************************************************************/

static void *StoreMagic( void *Block , ULONG Size , LPCSTR Filename , int Line )
{
	if (Block != NULL)
	{
		CCMallocList *p = (CCMallocList*)Block;

		p->GuardWord = MAGIC_LONGWORD;		 							// store magic

		p->Size = Size;

		// Fill in the details about the file position
		if (Filename != NULL)
		{	
			// copy the string passed in and note the line number
			int Count = strlen(Filename);
			if (Count > 21)
			{
				// The string was too long, so we have to copy just the last 22 chars in it
				Filename += (Count - 21); // *sizeof(TCHAR);
			}

			// copy the end of the string into the space I have for it
			strcpy_s(p->Filename, Filename);
			p->LineNum = Line;
		}
		else
		{
			// set the fields to null values
			strcpy_s(p->Filename, "Unknown File");
			p->LineNum = 0;
		}

		Block = ((char*)Block) + MAGIC_EXTRA;  							// and return user bigger pointer
	}
	return Block;
}


/********************************************************************************************

	static void LinkMagic( void *Block )

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		Block - pointer to memory block.
	Outputs:	None
	Purpose:	For debug builds to link the block into the tracking list. Stubbed away in
				retail builds.
	Scope:		Static to fixmem.cpp

********************************************************************************************/

static void LinkMagic( void *Block )
{
	if (Block != NULL)
	{
		CCMallocList *p = (CCMallocList*)Block;

		p->GuardWord = MAGIC_LONGWORD;		 	// store magic word

		p->Next = TraceListHead;
		p->Prev = NULL;

		if (TraceListHead != NULL)
			TraceListHead->Prev = p;

		TraceListHead = p;
	}
}


/********************************************************************************************

>	void DumpCCMallocTrace(void)

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		None
	Outputs:	None
	Returns:	None
	Purpose:	Lists any unreleased memory blocks
	Errors:		None
	SeeAlso:	LinkMagic;CheckMagic;StoreMagic
	
********************************************************************************************/

void DumpCCMallocTrace(void)
{
	if (TraceListHead != NULL)
	{
		TRACE("\n\n...CCMalloc leaks detected...\n\n");

		CCMallocList *Ptr = TraceListHead;

		while (Ptr != NULL)
		{
			TRACE("%ld bytes allocated in %s at line %ld\n",Ptr->Size, Ptr->Filename, Ptr->LineNum);
			Ptr = Ptr->Next;
		}

		TRACE("\n...End of CCMalloc leaks...\n\n");
	}
}


/********************************************************************************************

>	LPVOID DebugCCMalloc(ULONG size, LPCSTR Filename, int Line)

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		size of block required
				source filename called from (supplied by macro)
				line number of call (supplied by macro)
	Outputs:	None
	Returns:	None
	Purpose:	A version of malloc which does not throw exceptions and provides source file
				and line number tracing.
	Errors:		Out of memory
	SeeAlso:	RetailCCMalloc;malloc

********************************************************************************************/

LPVOID DebugCCMalloc(ULONG Size, LPCSTR Filename, int Line)
{
	if (Size==0L)															// size of 0 is illegal
	{
		ENSURE(FALSE, "Someone asked for 0 bytes");
		return NULL;
	}

#ifdef _OUTPUT_DEBUG
	TotalMalloced += Size;
	TRACE("Malloced %ld (+%ld in %s at line %ld)\n",TotalMalloced,Size, Filename, Line);
#endif

	ULONG NewSize = Size + MAGIC_EXTRA;

	if (checksize(NewSize))
		return NULL;														// if too big

	// save VC++s internal fail-memory pointer and reset it so nothing special occurs
	// if memory runs out
	_PNH old_handler = _set_new_handler( NULL );

	// call malloc to get the memory or a NULL pointer
	void *p = malloc( (size_t)NewSize );

	// The next line is for debug purposes (comment out above line)
	//void *p = NULL;
	
	// repair the function pointer to its old value and return our pointer
	_set_new_handler( old_handler );

	// set the error and return if it didn't work
	if (p == NULL)
	{
		Error::SetError(IDS_OUT_OF_MEMORY);
		return NULL;
	}

	// Link block in to list
	LinkMagic(p);

	// add the magic longword
	return StoreMagic(p, Size, Filename, Line);
}


/********************************************************************************************

	void DebugCCRealloc(LPVOID Buf, ULONG Size, LPCSTR Filename, int Line)

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		pointer to memory allocated with CCMalloc (or NULL)
				new size for block
				source filename called from (supplied by macro)
				line number of call (supplied by macro)
	Outputs:	None
	Returns:	None
	Purpose:	A version of realloc which does not throw exceptions and provides source file
				and line number tracing.  DO NOT CALL.  Use CCRealloc instead.
	Errors:		Out of memory
	SeeAlso:	RetailCCRealloc;realloc

********************************************************************************************/

LPVOID DebugCCRealloc(LPVOID Buf, ULONG Size, LPCSTR Filename, int Line)
{
	if (Size==0L)															// size of 0 is illegal
	{
		ENSURE(FALSE, "Someone asked for 0 bytes");
		return NULL;
	}

	// check magic longword (and remove it)
	const LPVOID NewBuf = CheckMagic(Buf, TRUE);	// Unlink block and set deleted
	if (
		(NewBuf==NULL) &&
		(Buf!=NULL)
	   )
	{
		// CheckMagic decided that it wasn't one of our blocks, so don't try to grow it
		return NULL;
	}

	ULONG NewSize = Size + MAGIC_EXTRA;

	if (checksize(NewSize))
		return NULL;															// if too big

	// save VC++s internal fail-memory pointer and reset it so nothing special occurs
	// if memory runs out
	_PNH old_handler = _set_new_handler( NULL );

	// call realloc 
	void *p = realloc( NewBuf, (size_t)NewSize );

	// The next line is for test purposes (comment out above line)
	//void *p = NULL;
	
	// finally repair the function pointer to its old value and return our pointer
	_set_new_handler( old_handler );

	// set the error and return if it didn't work
	if (p == NULL)
	{
		Error::SetError(IDS_OUT_OF_MEMORY);

		// Re-link old block into list
		LinkMagic(NewBuf);

		// return NULL as realloc
		return NULL;
	}
	else
	{
		// Link new block into list
		LinkMagic(p);

		// return result, storing magic word
		return StoreMagic(p, Size, Filename, Line);
	}
}



/********************************************************************************************

>	LONG DebugCCGetBlockSize(LPVOID Block)

	Author:		Rik
	Created:	1/5/95
	Inputs:		Block - the alloc'ed block we want to know the size of
	Returns:	the size of the block that was asked for
	Purpose:	Finds out how big a block of ram someone asked for when they called CCMalloc

********************************************************************************************/

LONG DebugCCGetBlockSize(LPVOID Block)
{
	// default to 0
	LONG Size = 0;

	// make sure it is OK
	if (Block != NULL)
	{
		// Find out the size
		CCMallocList *p = (CCMallocList*) (((char*)Block)-MAGIC_EXTRA);
		Size = p->Size;
	}

	// return it
	return Size;
}


#else
// retail builds stub out those fns just in case
// the only one that is used is CheckMagic (in CCFree)
#define	CheckMagic(p,k)	(p)
#define	StoreMagic(p,s,f,l)	(p)
#define LinkMagic(p)
#define	MAGIC_EXTRA		0

/********************************************************************************************

	LPVOID RetailCCMalloc( ULONG size )

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		size of memory required (must be non-zero)
	Outputs:	None
	Returns:	Pointer to new block, or NULL if failed.
	Purpose:	Like malloc but does not throw exceptions.
	Errors:		None.
	SeeAlso:	CCFree;RetailCCRealloc;malloc

********************************************************************************************/

LPVOID RetailCCMalloc( ULONG Size )
{
	if (Size==0L)															// size of 0 is illegal
		return NULL;

	if (checksize(Size))
		return NULL;														// if too big

	// save VC++s internal fail-memory pointer and reset it so nothing special occurs
	// if memory runs out
	_PNH old_handler = _set_new_handler( NULL );

	// call malloc to get the memory or a NULL pointer
	void *p = malloc( (size_t)Size );
	
	// repair the function pointer to its old value and return our pointer
	_set_new_handler( old_handler );

	// set the error if it didn't work
	if (p == NULL)
		Error::SetError(IDS_OUT_OF_MEMORY);

	return p;
}




/********************************************************************************************

	LPVOID RetailCCRealloc( LPVOID buf, ULONG size )

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		Pointer to block allocated with CCMalloc (can be NULL) and new size (must be non-zero)
	Returns:	Pointer to new block (might have changed) or NULL if cannot reallocate (old block
				will remain intact)
	Purpose:	Change size of memory block, similar to realloc but does not throw exceptions.
	Errors:		None.
	SeeAlso:	RetailCCMalloc;CCFree;realloc

********************************************************************************************/

LPVOID RetailCCRealloc( LPVOID Buf, ULONG Size )
{
	if (Size==0L)															// size of 0 is illegal
		return NULL;

	if (checksize(Size))
		return NULL;															// if too big

	// save VC++s internal fail-memory pointer and reset it so nothing special occurs
	// if memory runs out
	_PNH old_handler = _set_new_handler( NULL );

	// call realloc
	void *p = realloc( Buf, (size_t)Size );
	
	// finally repair the function pointer to its old value and return our pointer
	_set_new_handler( old_handler );

	// set the error if it didn't work
	if (p == NULL)
		Error::SetError(IDS_OUT_OF_MEMORY);

	// return result, storing magic word
	return p;
}




/********************************************************************************************

>	LONG RetailCCGetBlockSize(LPVOID Block)

	Author:		Rik
	Created:	1/5/95
	Inputs:		Block - the alloc'ed block we want to know the size of
	Returns:	the size of the block that was asked for
	Purpose:	Finds out how big a block of ram someone asked for when they called CCMalloc

********************************************************************************************/

LONG RetailCCGetBlockSize(LPVOID Block)
{
	if (Block==NULL)
		return 0;

	return _msize(Block);
}


#endif



/********************************************************************************************

>	LPVOID CCMalloc( ULONG size )

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		size of memory required (must be non-zero)
	Outputs:	None
	Returns:	Pointer to new block, or NULL if failed.
	Purpose:	Like malloc but does not throw exceptions.
	Errors:		None.
	SeeAlso:	malloc;CCFree;CCRealloc

********************************************************************************************/
/*
Technical notes:
This has been implemented as a macro which calls either RetailCCMalloc or DebugCCMalloc
depending on whether it is a retail or debug build.

********************************************************************************************/


/********************************************************************************************

>	LPVOID CCRealloc( LPVOID buf, ULONG size )

	Author: 	Gerry
	Created:	1/9/94
	Inputs:		Pointer to block allocated with CCMalloc (can be NULL) and new size (must be non-zero)
	Returns:	Pointer to new block (might have changed) or NULL if cannot reallocate (old block
				will remain intact)
	Purpose:	Change size of memory block, similar to realloc but does not throw exceptions.
	Errors:		None.
	SeeAlso:	RetailCCMalloc;CCFree;realloc

********************************************************************************************/
/*
Technical notes:
This has been implemented as a macro which calls either RetailCCRealloc or DebugCCRealloc
depending on whether it is a retail or debug build.

********************************************************************************************/


/********************************************************************************************

>	void CCFree( LPBYTE buf )

	Author: 	Andy
	Created:	25/10/93
	Inputs:		pointer to memory allocated with CCMalloc (or NULL)
	Outputs:	None
	Returns:	None
	Purpose:	Like free.
	Errors:		None
	SeeAlso:	CCMalloc;free

********************************************************************************************/

void CCFree( LPVOID Buf )
{
#ifdef _OUTPUT_DEBUG
	if (Buf)
	{
		CCMallocList* p = (CCMallocList*) Buf;
		p--;
		TotalMalloced -= p->Size;
		TRACE("UnMalloced %ld (-%ld in %s at %ld)\n",TotalMalloced, p->Size, p->Filename, p->LineNum);
	}

#endif
	// check magic longword (and remove it to find double-frees)
	Buf = CheckMagic( Buf, TRUE );

	// then free it
	free( Buf );
}
