/* ================================XARAHEADERSTART===========================
 
               PDF2XAR, a PDF file to XAR file converter
                    Copyright (C) 1993-2014 Xara Group Ltd.
       Copyright on certain contributions may be held in joint with their
              respective authors. 


LICENSE TO USE AND MODIFY SOFTWARE
----------------------------------

This file is part of PDF2XAR.

PDF2XAR is free software; you can redistribute it and/or modify it under the 
terms of the GNU General Public License version 2 or version 3 as published
by the Free Software Foundation.

PDF2XAR and its component source files are distributed in the hope
that it will be useful, but WITHOUT ANY WARRANTY; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with PDF2XAR (see the file GPL in the root directory of the
distribution); if not, write to the Free Software Foundation, Inc., 51
Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

SCOPE OF LICENSE
----------------

This license applies to this program (PDF2XAR) and its constituent source
files only, and does not apply to other Xara products which may
in part share the same code base, and are subject to their own licensing
terms.

This license does not apply to the binary libraries (if any) within
the "libs" directory, which are subject to a separate license contained
within that directory in the file "LIBS-LICENSE".


ARRANGEMENTS FOR CONTRIBUTION OF MODIFICATIONS
----------------------------------------------

If you wish to contribute modifications please contact Xara via 
www.xara.com/contact

Please note that Xara will not accept modifications which modify any of
the text between the start and end of this header (marked
XARAHEADERSTART and XARAHEADEREND).

MARKS
-----

Xara, PDF2XAR, Xara Web Designer, Xara Photo & Graphic Designer, Xara Page & 
Layout Designer, Xara Designer Pro and Xara designs are trademarks, design-marks, 
and/or service marks of Xara Group Ltd. All rights in these marks are reserved.


      Xara Group Ltd, Gaddesden Place, Hemel Hempstead, HP2 6EX, UK.
                        http://www.xara.com/

=================================XARAHEADEREND============================*/
// $Header: /Camelot/kernel/matrix.h 8     30/01/04 15:56 Phil $

/*
 * */


#ifndef INC_MATRIX
#define INC_MATRIX

#include "ccmaths.h"
#include "doccoord.h"

class DocRect;

/********************************************************************************************

<	TransformType; enum

	Author:		Rik
	Created:	19/10/93
	Comment:	This enum is used to determine what type of matrix we have. It can be one of
				the following values :- MonoOn
				TRANS_IDENTITY		// An Identity XaraMatrix
				TRANS_TRANSLATION	// A Translation XaraMatrix
				TRANS_ROTATION		// A Rotation XaraMatrix
				TRANS_SCALE			// A Scale XaraMatrix
				TRANS_SHEAR			// A Shear XaraMatrix
				TRANS_COMPLEX		// A XaraMatrix that is a combination of the above
				MonoOff

********************************************************************************************/

typedef enum
{
	TRANS_IDENTITY,
	TRANS_TRANSLATION,
	TRANS_ROTATION,
	TRANS_SCALE,
	TRANS_SHEAR,
	TRANS_COMPLEX
} TransformType;


/***********************************************************************************************

>	class XaraMatrix

	Author:		Mario Shamtani (extended by Rik)
	Created:	13/5/1993 (extended 19/10/93)
	Purpose:	Represents a mathematical transformation matrix. Transformation matrices can 
				transform the coordinate pairs of one coordinate space into another coordinate
				space.
				Matrices are used in Camelot for the rendering process in various ways. These 
				include coordinate conversion, scaling, rotation and translation. 
				
				Depending on the constructor applied it is possible to specify any of the above
				types of matrix. MonoOn
                 eg.	XaraMatrix();              			//creates the identity matrix
                 		XaraMatrix(ANGLE(90));              //creates a rotation matrix by 90 deg
                 		XaraMatrix(ScaleX, ScaleY);         //creates a scaling matrix
                 		XaraMatrix(xOffset, yOffset);       //creates a translation matrix
                MonoOff
                The XaraMatrix also contains a couple of Public vars that may or may not be
                of any use to anyone. These are as follows :- MonoOn
                TransformType	Type;	// The type of transformation this matrix represents
                ANGLE			Angle;	// The angle of rotation for a rotation matrix
                MonoOff
	SeeAlso:	TransformType
                                    
***********************************************************************************************/

class XaraMatrix
{
friend class OSRenderRegion;
friend class XMatrix;

private:
	FIXED16 a;
	FIXED16 b;
	FIXED16 c;
	FIXED16 d;
	LONG e;
	LONG f;

public:
	TransformType	Type;	// The type of transform this matrix represents
	ANGLE			Angle;	// If it is a rotation, this holds the angle of rotation

public:
	// Constructors et all
	XaraMatrix();
	XaraMatrix(const FIXED16&, const FIXED16&);
	XaraMatrix(const ANGLE& theta);
	XaraMatrix(const DocCoord&);
	XaraMatrix(const DocCoord&, const ANGLE&);
	XaraMatrix(const LONG, const LONG);
	XaraMatrix(const FIXED16&, const FIXED16&, const FIXED16&, const FIXED16&, const LONG, const LONG);
	XaraMatrix(const DocRect&, const DocRect&);
	
	// Transformations using the matrix
	void transform(DocCoord*) const;
	void transform(DocCoord[], UINT count) const;
	void transform(DocCoord[], const DocCoord[], UINT) const;

	void translate(const LONG TransX, const LONG TransY);

	BOOL TransformBounds(DocRect* pRect) const;
	BOOL Decompose(FIXED16* pScale=NULL, FIXED16* pAspect=NULL,
				   ANGLE* pRotation=NULL, ANGLE* pShear=NULL,
				   DocCoord* pTranslate=NULL, FIXED16* pScaleY=NULL);
	BOOL Compose(FIXED16 Scale=1.0, FIXED16 Aspect=1.0,
				 ANGLE Rotation=0, ANGLE Shear=0, DocCoord Translation=DocCoord(0,0));

	BOOL IsRotation(const double epsilon = 0.0) const;
	BOOL IsReflection(const double epsilon = 0.0) const;
	BOOL IsTranslation(const double epsilon = 0.0) const;
	BOOL IsIdentity(/*DOUBLEPARAM epsilon*/) const;

	// XaraMatrix copy and multiplication stuff
	XaraMatrix& operator=(const XaraMatrix&);
	BOOL operator==(const XaraMatrix&) const;
	friend XaraMatrix operator*(const XaraMatrix&, const XaraMatrix&);
	XaraMatrix& operator*=(const XaraMatrix&);

	// XaraMatrix Inversion
	XaraMatrix Inverse() const;

	// obtain matrix compnents
	void GetComponents( FIXED16 *, LONG * ) const;
	void GetTranslation(DocCoord&) const;
	void GetTranslation(LONG&,LONG&) const;
	void SetTranslation(const DocCoord&);
	void SetTranslation(const LONG&, const LONG&);

	// Karim 23/07/2000
	// ok, here are some factory methods so you don't have to trawl
	// through the codebase every time you want a certain type of matrix.
	static XaraMatrix CreateTransMatrix(const DocCoord& dcTrans);
	static XaraMatrix CreateScaleMatrix(const double& xScale, const double& yScale);
	static XaraMatrix CreateScaleMatrix(const double& ScaleFactor);
	static XaraMatrix CreateShearMatrix(const double& ShearAngle);
	static XaraMatrix CreateRotateMatrix(const double& RotateAngle);

	// Karim 04/09/2000
	// two methods to obtain the trace and determinant of a matrix.
	double Trace() const;
	double Det() const;

	// Karim 09/2001
	// A set of matrix fns useful for geometric tweening.
	void RatioMatrix (const double ratio);
	XaraMatrix Interpolate(const XaraMatrix& op, const double ratio) const;
	XaraMatrix PostRotate(DocCoord dcCentre, const ANGLE& Angle) const;
};

#endif 
