#pragma once

#ifndef CAMELOT_PRODUCT
#include <unordered_map>
#include <memory>
#include <array>
#include "../XPFilter/common/GSUBEntry.h"

using StylisticSetDataSet = std::array<std::unordered_map<WCHAR, WCHAR>, 20>;
using StylisticSetNameSet = std::array<std::unordered_map<LANGID, std::string>, 20>;
using StylisticSetSizeSet = std::array<ULONG, 20>;

struct TypoMetrics
{
	BOOL bUseTypoMetrics{ FALSE };
	MILLIPOINT TypoAscent{ 0 };
	MILLIPOINT TypoDescent{ 0 };
	MILLIPOINT TypoLineGap{ 0 };
};
#else
struct TypoMetrics;
#endif

struct OTLookup;
struct LookupSubTable; // fwd

class VAddOpenTypeFeature
{
public:
	VAddOpenTypeFeature() {}
	virtual ~VAddOpenTypeFeature() {}

	virtual LookupSubTable* AddLigature(std::wstring key, WCHAR wLig, DWORD dwLigType) = 0;
	virtual LookupSubTable* AddSubstitute(WCHAR c1, WCHAR c2, DWORD dwTableSubtype) = 0;
	virtual bool AddSubstRule(DWORD dwTableSubtype, std::vector<std::wstring>& pre, std::vector<std::wstring>& post, std::vector<std::wstring>& keys, OTLookup* lookup, UINT16 priority, std::vector<UINT32>& langs) = 0;
};
/********************************************************************************************
>	class VGlyphSource:

	Author:		Paul Gruzdev
	Created:	10/11/2016
	Purpose:	The abstract interface that works as a source of glyph outline data, either
				from Camelot or the external filters.
				The source should provide: 
					- font properties like 
						- font name
						- ascent
						- descent
						- character width
						- font weight
						- italic flag
					- begin(), end() functions of type std::set<WCHAR>::iterator so that
						the consumer can easily iterate through all of the Unicode values of 
						glyphs being provided
					- bounding box calculation routine for a given bezier (verbs, coords)
					- glyph outline data itself for a given Unicode value
					- kerning pairs enumeration mechanism
********************************************************************************************/

class VGlyphSource
{
public:
	enum GlyphState { keError = 0, keRegular, keSubstituted, keEmbedded, kePackedCFF };

	VGlyphSource() {}
	virtual ~VGlyphSource() {}
	typedef std::set<WCHAR>::iterator iterator;

	virtual bool IsGood() = 0;
	// This is written to string index=1 of the 'name' table
	virtual std::wstring GetFontFamilyName() = 0;
	// This is written to string index=4 of the 'name' table
	virtual std::wstring GetFullFontName() = 0;
	// This is written to string index=6 of the 'name' table. Optional: if not specified, fontlib will generate it
	virtual std::wstring GetPostScriptFontName() { return L""; }
	virtual LONG GetWeight() = 0;
	virtual BOOL IsItalic() = 0;

	virtual bool CalcBBox(DocCoord* pCoord, BYTE* pVerb, ULONG ulCount, RECT& rc) = 0;
	virtual GlyphState GetGlyphPath(WCHAR wUni, PBYTE* verb, DocCoord** coord, ULONG* pulNumCoords) = 0;
	virtual LONG GetCharWidth(WCHAR wUni) = 0;
	virtual LONG GetFontAscent() = 0;
	virtual LONG GetFontDescent() = 0;

	virtual iterator begin() = 0;
	virtual iterator end() = 0;
	// Kerning pair support
	virtual bool GetFirstKerningPair(WORD& wFirst, WORD& wSecond, MILLIPOINT& iAmount) = 0;
	virtual bool GetNextKerningPair(WORD& wFirst, WORD& wSecond, MILLIPOINT& iAmount) = 0;
	// Open Type ligatures support
	virtual bool GetFirstLigature(std::wstring& key, wchar_t& wLigature, DWORD& dwLigType) = 0;
	virtual bool GetNextLigature(std::wstring& key, wchar_t& wLigature, DWORD& dwLigType) = 0;
	virtual bool AddOTFeatures(VAddOpenTypeFeature* pAddFeature) { return false; }
};

class VGlyphSource2
{
public:
	virtual RECT GetFontBBox() = 0;
	virtual LONG GetFontCapHeight() = 0;
	virtual LONG GetFontXHeight() = 0;
	virtual LONG GetFontNominalWidth() = 0;
	virtual LONG GetFontDefaultWidth() = 0;
	virtual bool HasLocalSubs() = 0;
	virtual bool GetLocalSubs(PBYTE& pSubs, size_t& szSubSize) = 0;
	virtual bool GetStdHW(int& iStdHW) = 0;
	virtual bool GetStdVW(int& iStdVW) = 0;
};

/********************************************************************************************
>	class VFontBase:

Author:		Paul Gruzdev
Created:	16/03/2017
Purpose:	The abstract interface that works together with otfTools.cpp/h to parse GPOS/GSUB
			tables and pass kerning/ligature info to Camelot or to filters.
********************************************************************************************/

class VFontBase
{
public:
	virtual bool UseOpenTypeSubstitutionData() = 0;
	virtual void SetUnicodeMap(const std::unordered_map<WCHAR, DWORD>& Map) = 0;
	virtual void SetLigatureData(std::unique_ptr<LigatureEntry[]> pLigEntries, const UINT& LigCnt) = 0;
	virtual void SetStylisticSetData(std::unique_ptr<StylisticSetDataSet> pSSetDataSet, std::unique_ptr<StylisticSetSizeSet> pSSetSizeSet) = 0;
	virtual void SetStylisticSetNames(std::unique_ptr<StylisticSetNameSet> pStylisticSetNames) = 0;
	virtual void SetMaxLigatureSize(INT MaxSize) = 0;
	virtual void SetTypoMetricsData(const TypoMetrics& i_metrics) = 0;

#ifdef BUILD_TEXT_ADVANCED_OPENTYPE
	virtual	void SetSubstitutionRuleSet(std::unique_ptr<SubstitutionRuleSet> RuleSet) = 0;
	virtual	SubstitutionRuleSet* GetSubRuleSet() = 0;

	virtual void SetTotalMaxSizes(const RuleMaxSizes& i_maxSizes) = 0;
	virtual RuleMaxSizes GetTotalMaxSizesForFont() = 0;
	virtual	RuleMaxSizes GetMaxSizesForRuleSet(const GlyphSetEnum& Set) = 0;

	virtual void SetSubstitutionTable(std::unique_ptr<SubstitutionTableSet> SubTableSet) = 0;
	virtual SubstitutionTable* GetSubTable(const UINT16& SubTableID) = 0;
	virtual SubstitutionTableSet* GetSubTableSet() = 0;

	virtual GlyphSetEnum GetOTFeatureByPriority(const UINT16& Priority) = 0;
	virtual void SetPriorityToFeatureMap(std::unique_ptr<std::unordered_map<UINT16, GlyphSetEnum>> PriorityToFeatureMap) = 0;
#endif
};
