/********************************************************************************************

>	cff.cpp

Author:		Paul Gruzdev
Created:	15/07/2016
Purpose:	CFF: Compact Font Format saving routine. See 5176.CFF.pdf:
			The Compact Font Format Specification.
			The SFNTWriter class actually wraps CFF into SFNT (whole 9 tables), so 
			that usual recipients like Windows, Chrome etc do understand the format. Currently 
			written tables are: 'CFF ', 'GPOS', 'GSUB'(empty), 'OS/2', 'cmap', 'head', 'hhea',
			'hmtx', 'maxp', 'name', 'post'.

********************************************************************************************/

#include "stdafx.h"
#include "sfnt.h"
#include "AdobeGlyphList.h"
#include "otfTools.h"

// Enables improved memory tracking. Must be defined after CC_IMPLEMENT_DYNCREATE macros.
#ifdef CAMELOT_PRODUCT
	#define new CAM_DEBUG_NEW
#endif

// Standard names for cff: see 5176.CFF.pdf, Appendix A. They should be here so we
// can use indexes instead of strings.
const CHAR* g_pszCFFNames[] =
{
	".notdef",
	"space",
	"exclam",
	"quotedbl",
	"numbersign",
	"dollar",
	"percent",
	"ampersand",
	"quoteright",
	"parenleft",
	"parenright",
	"asterisk",
	"plus",
	"comma",
	"hyphen",
	"period",
	"slash",
	"zero",
	"one",
	"two",
	"three",
	"four",
	"five",
	"six",
	"seven",
	"eight",
	"nine",
	"colon",
	"semicolon",
	"less",
	"equal",
	"greater",
	"question",
	"at",
	"A",
	"B",
	"C",
	"D",
	"E",
	"F",
	"G",
	"H",
	"I",
	"J",
	"K",
	"L",
	"M",
	"N",
	"O",
	"P",
	"Q",
	"R",
	"S",
	"T",
	"U",
	"V",
	"W",
	"X",
	"Y",
	"Z",
	"bracketleft",
	"backslash",
	"bracketright",
	"asciicircum",
	"underscore",
	"quoteleft",
	"a",
	"b",
	"c",
	"d",
	"e",
	"f",
	"g",
	"h",
	"i",
	"j",
	"k",
	"l",
	"m",
	"n",
	"o",
	"p",
	"q",
	"r",
	"s",
	"t",
	"u",
	"v",
	"w",
	"x",
	"y",
	"z",
	"braceleft",
	"bar",
	"braceright",
	"asciitilde",
	"exclamdown",
	"cent",
	"sterling",
	"fraction",
	"yen",
	"florin",
	"section",
	"currency",
	"quotesingle",
	"quotedblleft",
	"guillemotleft",
	"guilsinglleft",
	"guilsinglright",
	"fi",
	"fl",
	"endash",
	"dagger",
	"daggerdbl",
	"periodcentered",
	"paragraph",
	"bullet",
	"quotesinglbase",
	"quotedblbase",
	"quotedblright",
	"guillemotright",
	"ellipsis",
	"perthousand",
	"questiondown",
	"grave",
	"acute",
	"circumflex",
	"tilde",
	"macron",
	"breve",
	"dotaccent",
	"dieresis",
	"ring",
	"cedilla",
	"hungarumlaut",
	"ogonek",
	"caron",
	"emdash",
	"AE",
	"ordfeminine",
	"Lslash",
	"Oslash",
	"OE",
	"ordmasculine",
	"ae",
	"dotlessi",
	"lslash",
	"oslash",
	"oe",
	"germandbls",
	"onesuperior",
	"logicalnot",
	"mu",
	"trademark",
	"Eth",
	"onehalf",
	"plusminus",
	"Thorn",
	"onequarter",
	"divide",
	"brokenbar",
	"degree",
	"thorn",
	"threequarters",
	"twosuperior",
	"registered",
	"minus",
	"eth",
	"multiply",
	"threesuperior",
	"copyright",
	"Aacute",
	"Acircumflex",
	"Adieresis",
	"Agrave",
	"Aring",
	"Atilde",
	"Ccedilla",
	"Eacute",
	"Ecircumflex",
	"Edieresis",
	"Egrave",
	"Iacute",
	"Icircumflex",
	"Idieresis",
	"Igrave",
	"Ntilde",
	"Oacute",
	"Ocircumflex",
	"Odieresis",
	"Ograve",
	"Otilde",
	"Scaron",
	"Uacute",
	"Ucircumflex",
	"Udieresis",
	"Ugrave",
	"Yacute",
	"Ydieresis",
	"Zcaron",
	"aacute",
	"acircumflex",
	"adieresis",
	"agrave",
	"aring",
	"atilde",
	"ccedilla",
	"eacute",
	"ecircumflex",
	"edieresis",
	"egrave",
	"iacute",
	"icircumflex",
	"idieresis",
	"igrave",
	"ntilde",
	"oacute",
	"ocircumflex",
	"odieresis",
	"ograve",
	"otilde",
	"scaron",
	"uacute",
	"ucircumflex",
	"udieresis",
	"ugrave",
	"yacute",
	"ydieresis",
	"zcaron",
	"exclamsmall",
	"Hungarumlautsmall",
	"dollaroldstyle",
	"dollarsuperior",
	"ampersandsmall",
	"Acutesmall",
	"parenleftsuperior",
	"parenrightsuperior",
	"twodotenleader",
	"onedotenleader",
	"zerooldstyle",
	"oneoldstyle",
	"twooldstyle",
	"threeoldstyle",
	"fouroldstyle",
	"fiveoldstyle",
	"sixoldstyle",
	"sevenoldstyle",
	"eightoldstyle",
	"nineoldstyle",
	"commasuperior",
	"threequartersemdash",
	"periodsuperior",
	"questionsmall",
	"asuperior",
	"bsuperior",
	"centsuperior",
	"dsuperior",
	"esuperior",
	"isuperior",
	"lsuperior",
	"msuperior",
	"nsuperior",
	"osuperior",
	"rsuperior",
	"ssuperior",
	"tsuperior",
	"ff",
	"ffi",
	"ffl",
	"parenleftinferior",
	"parenrightinferior",
	"Circumflexsmall",
	"hyphensuperior",
	"Gravesmall",
	"Asmall",
	"Bsmall",
	"Csmall",
	"Dsmall",
	"Esmall",
	"Fsmall",
	"Gsmall",
	"Hsmall",
	"Ismall",
	"Jsmall",
	"Ksmall",
	"Lsmall",
	"Msmall",
	"Nsmall",
	"Osmall",
	"Psmall",
	"Qsmall",
	"Rsmall",
	"Ssmall",
	"Tsmall",
	"Usmall",
	"Vsmall",
	"Wsmall",
	"Xsmall",
	"Ysmall",
	"Zsmall",
	"colonmonetary",
	"onefitted",
	"rupiah",
	"Tildesmall",
	"exclamdownsmall",
	"centoldstyle",
	"Lslashsmall",
	"Scaronsmall",
	"Zcaronsmall",
	"Dieresissmall",
	"Brevesmall",
	"Caronsmall",
	"Dotaccentsmall",
	"Macronsmall",
	"figuredash",
	"hypheninferior",
	"Ogoneksmall",
	"Ringsmall",
	"Cedillasmall",
	"questiondownsmall",
	"oneeighth",
	"threeeighths",
	"fiveeighths",
	"seveneighths",
	"onethird",
	"twothirds",
	"zerosuperior",
	"foursuperior",
	"fivesuperior",
	"sixsuperior",
	"sevensuperior",
	"eightsuperior",
	"ninesuperior",
	"zeroinferior",
	"oneinferior",
	"twoinferior",
	"threeinferior",
	"fourinferior",
	"fiveinferior",
	"sixinferior",
	"seveninferior",
	"eightinferior",
	"nineinferior",
	"centinferior",
	"dollarinferior",
	"periodinferior",
	"commainferior",
	"Agravesmall",
	"Aacutesmall",
	"Acircumflexsmall",
	"Atildesmall",
	"Adieresissmall",
	"Aringsmall",
	"AEsmall",
	"Ccedillasmall",
	"Egravesmall",
	"Eacutesmall",
	"Ecircumflexsmall",
	"Edieresissmall",
	"Igravesmall",
	"Iacutesmall",
	"Icircumflexsmall",
	"Idieresissmall",
	"Ethsmall",
	"Ntildesmall",
	"Ogravesmall",
	"Oacutesmall",
	"Ocircumflexsmall",
	"Otildesmall",
	"Odieresissmall",
	"OEsmall",
	"Oslashsmall",
	"Ugravesmall",
	"Uacutesmall",
	"Ucircumflexsmall",
	"Udieresissmall",
	"Yacutesmall",
	"Thornsmall",
	"Ydieresissmall",
	"001.000",
	"001.001",
	"001.002",
	"001.003",
	"Black",
	"Bold",
	"Book",
	"Light",
	"Medium",
	"Regular",
	"Roman",
	"Semibold",
	NULL
};

INT nStdStrings = sizeof(g_pszCFFNames) / sizeof(g_pszCFFNames[0]) - 1;

namespace SFNT
{

static void dumpint(CCFile* pFile, int num)
{
	if (num >= -107 && num <= 107)
		pFile->write((char)(num + 139));
	else if (num >= 108 && num <= 1131)
	{
		num -= 108;
		pFile->write((char)((num >> 8) + 247));
		pFile->write(num & 0xff);
	}
	else if (num >= -1131 && num <= -108)
	{
		num = -num;
		num -= 108;
		pFile->write((char)((num >> 8) + 251));
		pFile->write(num & 0xff);
	}
	else if (num >= -32768 && num < 32768)
	{
		pFile->write(28);
		pFile->write((char)(num >> 8));
		pFile->write(num & 0xff);
	}
	else
	{	// In dict data we have 4 byte ints, in type2 strings we don't
		pFile->write((char)29);
		pFile->write((char)((num >> 24) & 0xff));
		pFile->write((char)((num >> 16) & 0xff));
		pFile->write((char)((num >> 8) & 0xff));
		pFile->write((char)(num & 0xff));
	}
}

static void dumpdbl(CCFile* pFile, double d)
{
	if (d - rint(d) > -.00001 && d - rint(d) < .00001)
		dumpint(pFile, (int)d);
	else
	{
		static _locale_t pLoc = NULL;
		if (pLoc == NULL)
			pLoc = _create_locale(LC_NUMERIC, "C");

		// The type2 strings have a fixed format, but the dict data does not
		char buffer[20], *pt;
		int sofar, n, odd;
		_sprintf_s_l(buffer, 20, "%g", pLoc, d);
		sofar = 0; odd = true;
		pFile->write((char)30);		// Start a double

		for (pt = buffer; *pt; ++pt)
		{
			if (isdigit(*pt))
				n = *pt - '0';
			else if (*pt == '.')
				n = 0xa;
			else if (*pt == '-')
				n = 0xe;
			else if ((*pt == 'E' || *pt == 'e') && pt[1] == '-') {
				n = 0xc;
				++pt;
			}
			else if (*pt == 'E' || *pt == 'e')
				n = 0xb;
			else
				n = 0;		// Should never happen

			if (odd)
			{
				sofar = n << 4;
				odd = false;
			}
			else
			{
				pFile->write((char)(sofar | n));
				sofar = 0;
				odd = true;
			}
		}

		if (sofar == 0)
			pFile->write((char)0xff);
		else
			pFile->write((char)(sofar | 0xf));
	}
}

static void dumpoffset(CCFile* pFile, int offsize, int val) 
{
	if (offsize == 1)
		pFile->write((char)val);
	else if (offsize == 2)
		putshort(pFile, (short)val);
	else if (offsize == 3) 
	{
		pFile->write((char)((val >> 16) & 0xff));
		pFile->write((char)((val >> 8) & 0xff));
		pFile->write((char)(val & 0xff));
	}
	else
		putlong(pFile, val);
}

static void dumpoper(CCFile* pFile, int oper)
{
	if (oper != -1)
	{
		if (oper >= 256)
			pFile->write((char)(oper >> 8));

		pFile->write((char)(oper & 0xff));
	}
}

static void dumpintoper(CCFile* pFile, int v, int oper)
{
	dumpint(pFile, v);
	dumpoper(pFile, oper);
}

static void dumpdbloper(CCFile* pFile, double d, int oper) 
{
	dumpdbl(pFile, d);
	dumpoper(pFile, oper);
}

static void dumpsizedint(CCFile* pFile, bool big, int num, int oper)
{
	if (big) 
	{
		pFile->write(29);
		pFile->write((char)((num >> 24) & 0xff));
		pFile->write((char)((num >> 16) & 0xff));
		pFile->write((char)((num >> 8) & 0xff));
		pFile->write((char)(num & 0xff));
	}
	else
	{
		pFile->write(28);
		pFile->write((char)(num >> 8));
		pFile->write((char)(num & 0xff));
	}
	dumpoper(pFile, oper);
}

static void DumpFloatArray(float* arr, int n, CCFile* cfff, int oper) 
{
	int mi, i;

	for (mi = n - 1; mi >= 0 && arr[mi] == 0; --mi);

	if (mi < 0)
		return;

	dumpdbl(cfff, arr[0]);

	for (i = 1; i <= mi; ++i)
		dumpdbl(cfff, arr[i] - arr[i - 1]);

	dumpoper(cfff, oper);
}

void SFNTWriter::WriteCFFHeader(CCFile* pFile)
{
	pFile->write('\1');		// Major version: 1
	pFile->write('\0');		// Minor version: 0
	pFile->write('\4');		// Header size in bytes
	pFile->write('\4');		// Absolute Offset size.
}

void SFNTWriter::WriteCFFNames(CCFile* pFile, const char* pszFontName)
{
	size_t len = strlen(pszFontName);
	putshort(pFile, 1);		// One font name
	pFile->write((char)1);		// Offset size
	dumpoffset(pFile, 1, 1);	// Offset to first name
	dumpoffset(pFile, 1, 1 + (int)len);
	pFile->write(pszFontName, (UINT)len);
}

/******************************************************************************
   Function:	SFNT::GetGlyphName
   Author:		Paul Gruzdev
   Purpose:     Given the Unicode, find a glyph name in Adobe Glyph list
   Returns:		std::string
   Parameters:	int uCode  
   Notes:       AGL initially is name --> unicode map, so we should form
				the opposite map the first time using std::map, and then 
				just use it.
******************************************************************************/
std::string GetGlyphName(int uCode)
{
	static std::map<int, std::string> s_AdobeGlyphMap;
	if (s_AdobeGlyphMap.empty())
	{
		for (size_t i = 0; i < ADOBE_GLYPH_LIST_COUNT; i++)
			s_AdobeGlyphMap[g_sgnAdobeGlyphName[i].wCode] = g_sgnAdobeGlyphName[i].pszName;
	}

	std::map<int, std::string>::iterator it = s_AdobeGlyphMap.find(uCode);
	if (it != s_AdobeGlyphMap.end())
		return it->second;

	char buffer[64];
	if (uCode >= 0xD800 && uCode < 0xE000 || uCode >= 0x10000)
		sprintf_s(buffer, 64, "u%04X", uCode);
	else
		sprintf_s(buffer, 64, "uni%04X", uCode);

	return std::string(buffer);
}

static float myround2(float pos, bool round)
{
	//int iVal;
	if (round)
		return((float)rint(pos));
	return pos;
	// Paul: it's better to round the coordinates themselves, not here
	//iVal = pos > 0 ? (int)((65536 * pos) + 0.5) : (int)((65536 * pos) - 0.5);
	//iVal /= 65536;
	//return (float)iVal;
}

static void AddNumber2(ByteVector& buf, float pos, bool round)
{
	int val, factor;

	pos = float(pos > 0 ? (int)((65536 * pos) + 0.5) : (int)((65536 * pos) - 0.5));
	pos /= 65536;

	if (round)
		pos = (float)rint(pos);

	if (pos > 32767.99 || pos<-32768)
	{
		// same logic for big ints and reals
		if (pos>0x3fffffff || pos<-0x40000000)
		{
			ERROR3_2("Number out of range: %g in type2 output (must be [-65536,65535])\n", pos);
			if (pos>0) 
				pos = (float)0x3fffffff;
			else 
				pos = -0x40000000;
		}

		for (factor = 2; factor < 32768; factor <<= 2)
			if (pos / factor<32767.99 && pos / factor>-32768)
				break;

		AddNumber2(buf, pos / factor, false);
		AddNumber2(buf, (float)factor, false);

		buf.push_back(0x0c);		// Multiply operator
		buf.push_back(0x18);
	}
	else if (pos != floor(pos))
	{
		val = int(pos * 65536);
		buf.push_back(0xff);
		buf.push_back((val >> 24) & 0xff);
		buf.push_back((val >> 16) & 0xff);
		buf.push_back((val >> 8) & 0xff);
		buf.push_back(val & 0xff);
	}
	else
	{
		val = (int)rint(pos);

		if (pos >= -107 && pos <= 107)
			buf.push_back((BYTE)(val + 139));
		else if (pos >= 108 && pos <= 1131) 
		{
			val -= 108;
			buf.push_back((BYTE)((val >> 8) + 247));
			buf.push_back(val & 0xff);
		}
		else if (pos >= -1131 && pos <= -108) 
		{
			val = -val;
			val -= 108;
			buf.push_back((BYTE)((val >> 8) + 251));
			buf.push_back(val & 0xff);
		}
		else 
		{
			buf.push_back(28);
			buf.push_back((val >> 8) & 0xff);
			buf.push_back(val & 0xff);
		}
	}
}

/******************************************************************************
	Function:	SFNTWriter::AddString
	Author:		Paul Gruzdev
	Purpose:    Add a string returning it's SID. If the string appeared to be
				the one of the standard ones - just return it's index. Otherwise,
				add it to the string's storage to save later to CFF.
	Returns:	size_t
	Parameters:	std::string str
	Notes:
******************************************************************************/
size_t SFNTWriter::AddString(std::string str)
{
	static std::map<std::string, size_t> s_cffNames;
	if (s_cffNames.empty())
	{
		for (size_t i = 0; i < (size_t)nStdStrings; i++)
			s_cffNames[g_pszCFFNames[i]] = i;
	}

	std::map<std::string, size_t>::iterator it = s_cffNames.find(str);
	if (it != s_cffNames.end())
		return it->second;

	m_strings.push_back(str);
	m_szTotalStringsLength += str.length();
	return nStdStrings + m_strings.size() - 1;
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::WriteSID
   Author:		Paul Gruzdev
   Purpose:     Put the string into the String INDEX assigning it's SID,
				and write out SID and operation
   Returns:		void
   Parameters:	CCFile * pFile  
   Parameters:	std::string str  
   Parameters:	int oper  
   Notes:       
******************************************************************************/
void SFNTWriter::WriteSID(CCFile* pFile, std::string str, int oper)
{
	if (str.empty())
		return;
	dumpint(pFile, (int)AddString(str));
	dumpoper(pFile, oper);
}

int SFNTWriter::DumpCFFDict(CCFile* pFile, int iSIDPostScriptName)
{
	// according to the PSRef Man v3, only fontname, fontmatrix and private appear in this dictionary
	dumpintoper(pFile, iSIDPostScriptName, (12 << 8) | 38);
	{
		dumpdbl(pFile, 1.0);// / g_iEM);
		dumpint(pFile, 0);
		dumpint(pFile, 0);
		dumpdbl(pFile, 1.0);// / g_iEM);
		dumpint(pFile, 0);
		dumpintoper(pFile, 0, (12 << 8) | 7);
	}

	int pstart = pFile->tell();
	dumpsizedint(pFile, false, 0, -1);	// private length
	dumpsizedint(pFile, true, 0, 18);	// private offset
	return pstart;
}

void SFNTWriter::CreateFDArray(CCMemFile* pFile, std::vector<BYTE>& vFDArray)
{
	vFDArray.resize(pFile->tell());
	memcpy(&vFDArray[0], pFile->GetMemoryBuffer(), pFile->tell());
}

void SFNTWriter::DumpCFFDictIndex(CCFile* fdarray, std::vector<FontDict2Data>& vfds, int iSIDPostScriptName)
{
	putshort(fdarray, 1);	// We have only one sub-font
	fdarray->write(2);		// DICTs aren't very big, and there are at most 255
	putshort(fdarray, 1);	// Offset to first dict

	putshort(fdarray, 0);	// Dump offset placeholder (note there's one extra to mark the end)

	FontDict2Data fds;
	int pos = fdarray->tell() - 1;
	fds.iFillinDictMark = DumpCFFDict(fdarray, iSIDPostScriptName);
	fds.iEoDictMark = fdarray->tell();
	if (fds.iEoDictMark > 65536)
		ERROR3("The DICT INDEX got too big, result won't work");

	fdarray->seek(2 * sizeof(short) + sizeof(char));
	putshort(fdarray, fds.iEoDictMark - pos);
	fdarray->seekIn(0, ios_base::end);
	vfds.push_back(fds);
}

/******************************************************************************
   Function:	SFNTWriter::WriteTopDict
   Author:		Paul Gruzdev
   Purpose:     Write CFF's Top Dictionary
   Returns:		void
   Parameters:	CCFile * pFile  
   Notes:       
******************************************************************************/
void SFNTWriter::WriteTopDict(CCFile* pFile)
{
	CreatePrivate();		// Create Private data block
	putshort(pFile, 1);		// One top dict
	pFile->write('\2');		// Offset size
	putshort(pFile, 1);		// Offset to topdict
	FilePos iFinalPos = pFile->tell();
	putshort(pFile, 0);		// placeholder for final position (final offset in index points beyond last element)

	WriteSID(pFile, m_sVersion, 0);
	WriteSID(pFile, m_sCopyright, 1);
	WriteSID(pFile, m_sFullName, 2);
	WriteSID(pFile, m_sFamilyName, 3);
	WriteSID(pFile, m_sWeight, 4);

	if (m_fsType)
	{
		char sPostScript[100];
		sprintf_s(sPostScript, 100, "/FSType %u def", m_fsType);
		WriteSID(pFile, sPostScript, (12 << 8 ) | 21);
	}
	
	if (m_bFixedWidth)
		dumpintoper(pFile, 1, (12 << 8) | 1);

	if (m_dItalicAngle != 0)
		dumpdbloper(pFile, m_dItalicAngle, (12 << 8) | 2);

	if (m_dUnderlinePos != -100)
		dumpdbloper(pFile, m_dUnderlinePos, (12 << 8) | 3);

	if (m_dUnderlineWidth != 50) 
		dumpdbloper(pFile, m_dUnderlineWidth, (12 << 8) | 4);

	//double dPointSize = (double)xar2ttf(m_lAscent - m_lDescent);
	//if (dPointSize != 1000 && dPointSize != 0)
	{
		dumpdbl(pFile, 1.0 / g_iEM);
		dumpint(pFile, 0);
		dumpint(pFile, 0);
		dumpdbl(pFile, 1.0 / g_iEM);
		dumpint(pFile, 0);
		dumpintoper(pFile, 0, (12 << 8) | 7);
	}

	dumpdbl(pFile, floor((float)xar2ttf(m_rcFontBounds.left)));
	dumpdbl(pFile, floor((float)xar2ttf(m_rcFontBounds.top)));
	dumpdbl(pFile, ceil((float)xar2ttf(m_rcFontBounds.right)));
	dumpdbloper(pFile, ceil((float)xar2ttf(m_rcFontBounds.bottom)), 5);

	size_t szStrings = GetStringsSize();
	size_t szGlobal = sizeof(short);	// Single entry: 0, no globals
	size_t szEncoding = 0;				// no encoding
	size_t szCharset = GetCharsetSize();
	size_t szCharstrings = GetCharstringsSize();
	size_t szPrivate = m_Private.size();
	size_t szBase = pFile->tell();
	bool bLongOffset = false;

	if (szBase + 6 * 3 + szStrings + szGlobal + szEncoding + szCharset + szCharstrings + szPrivate > 32767)
	{
		bLongOffset = true;
		szBase += 5 * 5 + 4;
	}
	else
		szBase += 5 * 3 + 4;

	int iCharsetOff = (int)(szBase + szStrings + szGlobal);
	int iCharstringsOff = (int)(szBase + szStrings + szGlobal + szCharset + szEncoding);
	int iPrivateOff = (int)(szBase + szStrings + szGlobal + szCharset + szEncoding + szCharstrings);

#ifdef _DEBUG
	m_iCharsetOff = iCharsetOff;
	m_iCharstringsOff = iCharstringsOff;
	m_iPrivateOff = iPrivateOff;
#endif
	// Offset to charset (oper=15)
	dumpsizedint(pFile, bLongOffset, iCharsetOff, 15);
	// Offset to encoding (oper=16), (not for CID)
	dumpsizedint(pFile, bLongOffset, 0, 16);
	// Offset to charstrings (oper=17)
	dumpsizedint(pFile, bLongOffset, iCharstringsOff, 17);
	// Length of, and Offset to private (oper=18) here (not for CID )
	dumpsizedint(pFile, bLongOffset, (int)szPrivate, -1);
	dumpsizedint(pFile, bLongOffset, iPrivateOff, 18);

	size_t szEotop = szBase - iFinalPos - 1;

	pFile->seekIn(iFinalPos, ios_base::beg);
	putshort(pFile, (short int)szEotop);
	pFile->seekIn(0, ios_base::end);
}

/******************************************************************************
Function:	SFNTWriter::WriteCIDTopDict
Author:		Paul Gruzdev
Purpose:    Write CFF's Top Dictionary with CID info structure
Returns:	void
Parameters:	CCFile * pFile
Notes:
******************************************************************************/
void SFNTWriter::WriteCIDTopDict(CCFile* pFile)
{
	CreatePrivate();	// Create Private data block
	CreateFDSelect(m_FDSelect, m_CharProps.size());	// Create FDSelect
	int iSIDPostScriptFontName = (int)AddString(m_sPostScriptName);
	CCMemFile fda(NULL, 0, CCMemWrite);
	DumpCFFDictIndex(&fda, m_fds, iSIDPostScriptFontName);

	putshort(pFile, 1);		// One top dict
	FilePos iFinalPos;
	pFile->write('\2');		// Offset size
	putshort(pFile, 1);		// Offset to topdict
	iFinalPos = pFile->tell();
	putshort(pFile, 0);		// placeholder for final position (final offset in index points beyond last element)

	if (m_fsType)
	{
		char sPostScript[100];
		sprintf_s(sPostScript, 100, "/FSType %u def", m_fsType);
		WriteSID(pFile, sPostScript, (12 << 8 ) | 21);
	}

	WriteSID(pFile, "Adobe", -1);				// Registry
	WriteSID(pFile, "Identity", -1);			// Ordering
	dumpintoper(pFile, 0, (12 << 8) | 30);		// Supplement
	//dumpintoper(pFile, (int)m_CharProps.size(), (12 << 8) | 34);	// CIDCount
	// P9052-4827 fix: apparently CIDCount is not the number of glyphs, but the number of possible CIDs
	//dumpintoper(pFile, (int)m_CharProps.back().wUnicode + 1, (12 << 8) | 34);	// CIDCount
	dumpintoper(pFile, (int)65535, (12 << 8) | 34);	// CIDCount
	// font bbox
	dumpdbl(pFile, floor((float)xar2ttf(m_rcFontBounds.left)));
	dumpdbl(pFile, floor((float)xar2ttf(m_rcFontBounds.top)));
	dumpdbl(pFile, ceil((float)xar2ttf(m_rcFontBounds.right)));
	dumpdbloper(pFile, ceil((float)xar2ttf(m_rcFontBounds.bottom)), 5);

	{
		dumpdbl(pFile, 1.0 / g_iEM);
		dumpint(pFile, 0);
		dumpint(pFile, 0);
		dumpdbl(pFile, 1.0 / g_iEM);
		dumpint(pFile, 0);
		dumpintoper(pFile, 0, (12 << 8) | 7);
	}

	dumpdbloper(pFile, 5.0, (12 << 8) | 31);	// CIDFontVersion

// 	WriteSID(pFile, m_sVersion, 0);
// 	WriteSID(pFile, m_sCopyright, 1);
// 	WriteSID(pFile, m_sFullName, 2);
// 	WriteSID(pFile, m_sFamilyName, 3);
// 	WriteSID(pFile, m_sWeight, 4);

	if (m_bFixedWidth)
		dumpintoper(pFile, 1, (12 << 8) | 1);

	if (m_dItalicAngle != 0)
		dumpdbloper(pFile, m_dItalicAngle, (12 << 8) | 2);

	if (m_dUnderlinePos != -100)
		dumpdbloper(pFile, m_dUnderlinePos, (12 << 8) | 3);

	if (m_dUnderlineWidth != 50)
		dumpdbloper(pFile, m_dUnderlineWidth, (12 << 8) | 4);

	size_t szStrings = GetStringsSize();
	size_t szGlobal = sizeof(short);	// Single entry: 0, no globals
	size_t szEncoding = 0;				// no encoding
	size_t szCharset = GetCharsetSize();
	size_t szCharstrings = GetCharstringsSize();
	size_t szPrivate = m_Private.size();
	size_t szFDSelect = m_FDSelect.size();
	size_t szFDArray = fda.Size();
	size_t szBase = pFile->tell();
	bool bLongOffset = true;

	szBase += 5 * 4 + 4 + 2;		// two of the opers below are two byte opers

	int iCharsetOff =		(int)(szBase + szStrings + szGlobal);
	int iFDSelectOff =		(int)(szBase + szStrings + szGlobal + szCharset);
	int iCharstringsOff =	(int)(szBase + szStrings + szGlobal + szCharset + szFDSelect);
	int iFDArrayOff =		(int)(szBase + szStrings + szGlobal + szCharset + szFDSelect + szCharstrings);
	int iPrivateOff =		(int)(szBase + szStrings + szGlobal + szCharset + szFDSelect + szCharstrings + szFDArray);

	size_t szPrivateAccum = 0;
	for (size_t i = 0; i < m_fds.size(); ++i)
	{
		fda.seek(m_fds[i].iFillinDictMark);
		dumpsizedint(&fda, false, (int)szPrivate, -1);	// Private len
		dumpsizedint(&fda, true, iPrivateOff + (int)szPrivateAccum, 18);	// Private offset
		szPrivateAccum += szPrivate;
	}

	SFNTWriter::CreateFDArray(&fda, m_FDArray);
	fda.close();

#ifdef _DEBUG
	m_iCharsetOff = iCharsetOff;
	m_iCharstringsOff = iCharstringsOff;
	m_iPrivateOff = iPrivateOff;
#endif
	// Offset to charset (oper=15)
	dumpsizedint(pFile, bLongOffset, iCharsetOff, 15);
	// Offset to fdselect
	dumpsizedint(pFile, bLongOffset, iFDSelectOff, (12 << 8) | 37);
	// Offset to charstrings (oper=17)
	dumpsizedint(pFile, bLongOffset, iCharstringsOff, 17);
	// Offset to fdarray
	dumpsizedint(pFile, bLongOffset, iFDArrayOff, (12 << 8) | 36);

	size_t szEotop = szBase - iFinalPos - 1;

	pFile->seekIn(iFinalPos, ios_base::beg);
	putshort(pFile, (short int)szEotop);
	pFile->seekIn(0, ios_base::end);
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::CreatePrivate
   Author:		Paul Gruzdev
   Purpose:     Create CFF private block as a memory blob (with size), ready
				to write to CFF when needed
   Returns:		void
   Notes:       
******************************************************************************/
void SFNTWriter::CreatePrivate()
{
// 	FILE* in = fopen("D:\\temp\\ttf\\cff.private", "rb");
// 	fseek(in, 0, SEEK_END);
// 	long sz = ftell(in);
// 	fseek(in, 0, SEEK_SET);
// 	m_Private.resize(sz);
// 	fread(&m_Private[0], sz, 1, in);
// 	fclose(in);
// 	return;
// 	float bluevalues[14], otherblues[10];
// 	memset(bluevalues, 0, 14 * sizeof(float));
// 	memset(otherblues, 0, 10 * sizeof(float));
// 	bluevalues[0] = -16.f; bluevalues[1] = 0;
// 	bluevalues[2] = 447.f; bluevalues[3] = 460.f;
// 	bluevalues[4] = 662.f; bluevalues[5] = 694.f;
// 	otherblues[0] = otherblues[1] = -214.f;

	// pdftron values:
	// blue values :
// 	bluevalues[0] = -16, 0,
// 		bluevalues[0] = 424, 438
// 		bluevalues[0] = 656, 670
// 		bluevalues[0] = 676, 692
// 		bluevalues[0] = 738, 746
// 
// 		// other blues :
// 		bluevalues[0] = -244, -244
// 		bluevalues[0] = 260, 266
// 		bluevalues[0] = 400, 400
// 
// 		// family blues :
// 		bluevalues[0] = -16, 0
// 		bluevalues[0] = 416, 428
// 		bluevalues[0] = 646, 658
// 		bluevalues[0] = 676, 692
// 		bluevalues[0] = 732, 742

	CCMemFile mem(NULL, 0, CCMemWrite);

// 	DumpFloatArray(bluevalues, sizeof(bluevalues) / sizeof(bluevalues[0]), &mem, 6);
// 	DumpFloatArray(otherblues, sizeof(otherblues) / sizeof(otherblues[0]), &mem, 7);
// 	DumpFloatArray(bluevalues, sizeof(bluevalues) / sizeof(bluevalues[0]), &mem, 8);
// 	DumpFloatArray(otherblues, sizeof(otherblues) / sizeof(otherblues[0]), &mem, 9);

	dumpintoper(&mem, (int)rint(xar2ttf(m_lDefaultWidth)), 20);		// Default Width
	dumpintoper(&mem, (int)rint(xar2ttf(m_lNominalWidth)), 21);		// Nominative Width

	int iStdHW = 36, iStdVW = 136;
	FilePos fpLocalSubsNumberOff = 0;
	if (m_pSource2)
	{
		m_pSource2->GetStdHW(iStdHW);
		m_pSource2->GetStdVW(iStdVW);
		if (m_pSource2->HasLocalSubs())
		{
			fpLocalSubsNumberOff = mem.tell();
			dumpintoper(&mem, 52, 19);
		}
	}

	dumpdbloper(&mem, iStdHW, 10);		// StdHW - these values make Adobe Reader rasteriser do a better job at lower resolutions
	dumpdbloper(&mem, iStdVW, 11);		// StdVW

	size_t szPrivateSize = (size_t)mem.tell();
	if (fpLocalSubsNumberOff)
	{
		mem.seek(fpLocalSubsNumberOff);
		dumpintoper(&mem, (int)szPrivateSize, 19);
	}

	m_Private.resize(szPrivateSize);
	memcpy(&m_Private[0], mem.GetMemoryBuffer(), szPrivateSize);
	mem.close();
}

/******************************************************************************
   Function:	SFNTWriter::WriteStrings
   Author:		Paul Gruzdev
   Purpose:     Write out string index
   Returns:		void
   Parameters:	CCFile * pFile  
   Notes:       
******************************************************************************/
void SFNTWriter::WriteStrings(CCFile* pFile)
{
	putshort(pFile, (WORD)m_strings.size());

	if (m_strings.empty()) // empty string index is written as a 2-byte zeros
		return;

	bool bByteOff = false;
	if (m_szTotalStringsLength + 1 < 256)
		bByteOff = true;

	pFile->write(bByteOff ? '\1' : '\2');
	size_t offset = 1;

	for (std::list<std::string>::iterator it = m_strings.begin(); it != m_strings.end(); ++it)
	{
		if (bByteOff)
			pFile->write((char)offset);
		else
			putshort(pFile, (short int)offset);
		offset += it->length();
	}

	// Write out the last offset to the n + 1 NULL string
	if (bByteOff)
		pFile->write((char)offset);
	else
		putshort(pFile, (short int)offset);

	for (std::list<std::string>::iterator it = m_strings.begin(); it != m_strings.end(); ++it)
	{
		pFile->write((void*)it->c_str(), (UINT)it->length());
	}
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::GetStringsSize
   Author:		Paul Gruzdev
   Purpose:     Calculate our total strings INDEX size
   Returns:		size_t
   Notes:       
******************************************************************************/
size_t SFNTWriter::GetStringsSize()
{
	size_t size = 2;
	if (m_strings.size() == 0)
		return size;

	bool bByteOff = false;
	if (m_szTotalStringsLength + 1 < 256)
		bByteOff = true;

	//size += bByteOff ? 1 : 2;
	size += 1; // byte indicating 1 for byte offsets and 2 for short offsets
	size += bByteOff ? m_strings.size()+1 : (m_strings.size()+1) * 2;
	size += m_szTotalStringsLength;
	return size;
}

void SFNTWriter::WriteCharstrings(CCFile* pFile)
{
	SFNTWriter::WriteCharstrings(pFile, m_Charstrings, m_szTotalCharstringsLength);
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::WriteCharstrings
   Author:		Paul Gruzdev
   Purpose:     Write out char strings INDEX
   Returns:		void
   Parameters:	CCFile * pFile  
   Notes:       
******************************************************************************/
void SFNTWriter::WriteCharstrings(CCFile* pFile, std::vector<std::vector<BYTE>>& vCharstrings, size_t szTotalCharstringsLength)
{
	putshort(pFile, (WORD)vCharstrings.size());

	if (vCharstrings.empty()) // empty index is written as a 2-byte zeros
		return;

	size_t len = szTotalCharstringsLength + 1;
	size_t offsize = len <= 255 ? 1 : len <= 65535 ? 2 : len <= 0xffffff ? 3 : 4;

	pFile->write((char)offsize);
	int offset = 1;

	for (std::vector<std::vector<BYTE>>::iterator it = vCharstrings.begin(); it != vCharstrings.end(); ++it)
	{
		dumpoffset(pFile, (int)offsize, offset);
		offset += (int)it->size();
	}

	// Write out the last offset to the n + 1 NULL string
	dumpoffset(pFile, (int)offsize, offset);

	for (std::vector<std::vector<BYTE>>::iterator it = vCharstrings.begin(); it != vCharstrings.end(); ++it)
	{
		if (it->size())
			pFile->write((void*)&(*it)[0], (UINT)it->size());
	}
}

size_t SFNTWriter::GetCharstringsSize()
{
	return GetCharstringsSize(m_Charstrings, m_szTotalCharstringsLength);
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::GetCharstringsSize
   Author:		Paul Gruzdev
   Purpose:     Calculate our charstrings INDEX size. Re-calculate if szTotalStringSize
				is zero
   Returns:		size_t
   Notes:       
******************************************************************************/
size_t SFNTWriter::GetCharstringsSize(std::vector<std::vector<BYTE>>& vCharStrings, size_t& szTotalStringSize)
{
	size_t size = 2;
	if (vCharStrings.empty()) // empty index is written as a 2-byte zeros
		return size;

	if (szTotalStringSize == 0)
	{
		for (size_t i = 0; i < vCharStrings.size(); i++)
			szTotalStringSize += vCharStrings[i].size();
	}

	size_t len = szTotalStringSize + 1;
	size_t offsize = len <= 255 ? 1 : len <= 65535 ? 2 : len <= 0xffffff ? 3 : 4;

	size += 1;

	size += (vCharStrings.size() + 1)*offsize;
	size += szTotalStringSize;

	return size;
}

/******************************************************************************
   Function:	SFNTWriter::CreateCharset
   Author:		Paul Gruzdev
   Purpose:     Create a charset - an array of strings of each glyph name
   Returns:		void
   Parameters:	CCFile * pFile  
   Notes:       
******************************************************************************/
void SFNTWriter::CreateCharset()
{
	for (size_t i = 1; i < m_CharProps.size(); i++)
	{
		if (m_bWriteCID)
			m_Charset.push_back(m_CharProps[i].wUnicode); // CID == Unicode in our implementation
		else
		{
			std::string strGlyphName = GetGlyphName(m_CharProps[i].wUnicode);
			size_t szSID = AddString(strGlyphName);
			m_Charset.push_back(szSID);
		}
	}
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::WriteCharset
   Author:		Paul Gruzdev
   Purpose:     Write out the charset
   Returns:		void
   Parameters:	CCFile * pFile  
   Notes:       
******************************************************************************/
void SFNTWriter::WriteCharset(CCFile* pFile)
{
	// Here we use a format 0 charset. ie. an array of SIDs in random order
	// First element must be ".notdef" and is omitted
	pFile->write('\0');

	for (std::list<size_t>::iterator it = m_Charset.begin(); it != m_Charset.end(); ++it)
		putshort(pFile, (short int)*it);
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::moveto2
   Author:		Paul Gruzdev
   Purpose:     Append an encoded type2 moveto command into buffer
   Returns:		void
   Parameters:	std::vector<BYTE> & buf  
   Parameters:	LONG cx  
   Parameters:	LONG cy  
   Parameters:	bool round  
   Notes:       
******************************************************************************/
void SFNTWriter::moveto2(ByteVector& buf, LONG cx, LONG cy, bool round)
{
	float x = xar2ttf(cx);
	float y = xar2ttf(cy);

	if (round)
	{
		x = (float)rint(x);
		y = (float)rint(y);
	}

	if (m_fCurX == x)
	{
		AddNumber2(buf, y - m_fCurY, round);
		buf.push_back(4);		// v move to
	}
	else if (m_fCurY == y)
	{
		AddNumber2(buf, x - m_fCurX, round);
		buf.push_back(22);		// h move to
	}
	else
	{
		AddNumber2(buf, x - m_fCurX, round);
		AddNumber2(buf, y - m_fCurY, round);
		buf.push_back(21);		// r move to
	}

	m_fCurX = x; //rint(32768 * x) / 32768;
	m_fCurY = y; //rint(32768 * y) / 32768;
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::lineto2
   Author:		Paul Gruzdev
   Purpose:     Append encoded type2 lineto to the buffer
   Returns:		size_t
   Parameters:	std::vector<BYTE> & buf  
   Parameters:	LONG cx  
   Parameters:	LONG cy  
   Parameters:	bool round  
   Notes:       
******************************************************************************/
UINT SFNTWriter::lineto2(ByteVector& buf, PBYTE verb, DocCoord* pCoord, UINT size, bool round)
{
	UINT cnt;

	for (cnt = 1; (verb[cnt] & ~(BYTE)PT_CLOSEFIGURE) == PT_LINETO && cnt < 15 && cnt < size; cnt++);

	UINT lastgood = cnt - 1, lasthvgood, hvcnt = 1;

	float x = xar2ttf(pCoord[0].x);
	float y = xar2ttf(pCoord[0].y);

	BYTE bFirstHV = 0;

	int hv = -1;

	if (m_fCurX == x)
	{
		hv = 1;		// Vertical
		bFirstHV = 7; // pick the right operator
	}
	else if (m_fCurY == y)
	{
		hv = 0;		// Horizontal
		bFirstHV = 6;
	}

	bool donehm = true;
	DocCoord temp1, temp2;

	if (hv != -1)
	{
		lasthvgood = 0; hvcnt = 1;

		if (cnt != 1)
		{
			for (UINT test = 1; test < cnt; test++)
			{
				float frommX = xar2ttf(pCoord[test - 1].x);
				float frommY = xar2ttf(pCoord[test - 1].y);
				if (round)
				{
					frommX = (float)rint(frommX);
					frommY = (float)rint(frommY);
				}

				float tomX = xar2ttf(pCoord[test].x);
				float tomY = xar2ttf(pCoord[test].y);

				if (round)
				{
					tomX = (float)rint(tomX);
					tomY = (float)rint(tomY);
				}

				if (hv == 1 && tomY == frommY)
					hv = 0;
				else if (hv == 0 && tomX == frommX)
					hv = 1;
				else
					break;

				lasthvgood = test;
				++hvcnt;

				if (test == lastgood)
					break;
			}
		}

		donehm = true;

		if (hvcnt == cnt || hvcnt >= 2)
		{
			UINT test;
			// It's more efficient to do some h/v linetos
			for (test = 0; test <= lasthvgood; test++)
			{
				//if (!donehm)
				//	break;
				donehm = false;
				float frommX = test == 0 ? m_fCurX : xar2ttf(pCoord[test - 1].x);
				float frommY = test == 0 ? m_fCurY : xar2ttf(pCoord[test - 1].y);
				if (round) 
				{
					frommX = (float)rint(frommX);
					frommY = (float)rint(frommY);
				}

				float tomX = xar2ttf(pCoord[test].x);
				float tomY = xar2ttf(pCoord[test].y);

				if (round) 
				{
					tomX = (float)rint(tomX);
					tomY = (float)rint(tomY);
				}

				if (frommX == tomX)
					AddNumber2(buf, tomY - frommY, round);
				else
					AddNumber2(buf, tomX - frommX, round);

				m_fCurX = tomX; //rint(32768 * tomX) / 32768;
				m_fCurY = tomY; //rint(32768 * tomY) / 32768;
			}

			buf.push_back(bFirstHV);

			return test;
		}
	}

	for (size_t i = 0; i < cnt; i++)
	{
		float x = xar2ttf(pCoord[i].x);
		float y = xar2ttf(pCoord[i].y);

		float fromX = round ? rint(m_fCurX) : m_fCurX;
		float fromY = round ? rint(m_fCurY) : m_fCurY;

		if (round)
		{
			x = (float)rint(x);
			y = (float)rint(y);
		}

		AddNumber2(buf, x - fromX, round);
		AddNumber2(buf, y - fromY, round);

		m_fCurX = x; //rint(32768 * x) / 32768;
		m_fCurY = y; //rint(32768 * y) / 32768;
	}

	buf.push_back(5);	// r line to
	return cnt;
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::curveto2
   Author:		Paul Gruzdev
   Purpose:     Append type2-encoded curveto to the end of the buffer
   Returns:		size_t
   Parameters:	std::vector<BYTE> & buf  
   Parameters:	DocCoord dcPrevCP  
   Parameters:	DocCoord dcNextCP  
   Parameters:	DocCoord dcTo  
   Parameters:	bool round  
   Notes:       
******************************************************************************/
UINT SFNTWriter::curveto2(ByteVector& buf, PBYTE verb, DocCoord* pCoord, UINT size, bool round)
{
	UINT cnt = 0, idx = 0;
	int hv = -1;
	bool donehm = true;

	float fPrevCPX = xar2ttf(pCoord[0].x);
	float fPrevCPY = xar2ttf(pCoord[0].y);
	float fNextCPX = xar2ttf(pCoord[1].x);
	float fNextCPY = xar2ttf(pCoord[1].y);
	float fToX = xar2ttf(pCoord[2].x);
	float fToY = xar2ttf(pCoord[2].y);
	BYTE bFirstHV = 0;

	if (m_fCurX == myround2(fPrevCPX, round) && myround2(fNextCPY, round) == myround2(fToY, round))
	{
		bFirstHV = 30; // vhcurveto
		hv = 1;
	}
	else if (m_fCurY == myround2(fPrevCPY, round) && myround2(fNextCPX, round) == myround2(fToX, round))
	{
		bFirstHV = 31; // hvcurveto
		hv = 0;
	}

	if (hv != -1)
	{
		while (
			(hv == 1 && m_fCurX == myround2(fPrevCPX, round) &&
			myround2(fNextCPY, round) == myround2(fToY, round)) ||
			(hv == 0 && m_fCurY == myround2(fPrevCPY, round) &&
			myround2(fNextCPX, round) == myround2(fToX, round)))
		{
			donehm = false;

			if (hv == 1)
			{
				AddNumber2(buf, myround2(fPrevCPY, round) - m_fCurY, round);
				AddNumber2(buf, myround2(fNextCPX, round) - myround2(fPrevCPX, round), round);
				AddNumber2(buf, myround2(fNextCPY, round) - myround2(fPrevCPY, round), round);
				AddNumber2(buf, myround2(fToX, round) - myround2(fNextCPX, round), round);
				hv = 0;
			}
			else
			{
				AddNumber2(buf, myround2(fPrevCPX, round) - m_fCurX, round);
				AddNumber2(buf, myround2(fNextCPX, round) - myround2(fPrevCPX, round), round);
				AddNumber2(buf, myround2(fNextCPY, round) - myround2(fPrevCPY, round), round);
				AddNumber2(buf, myround2(fToY, round) - myround2(fNextCPY, round), round);
				hv = 1;
			}

			m_fCurX = myround2(fToX, round);
			m_fCurY = myround2(fToY, round);

			++cnt;
			idx += 3;

			if (idx >= size || cnt > 9 || (verb[idx] & ~(BYTE)PT_CLOSEFIGURE) != PT_BEZIERTO)
				break;

			fPrevCPX = xar2ttf(pCoord[idx].x);
			fPrevCPY = xar2ttf(pCoord[idx].y);
			fNextCPX = xar2ttf(pCoord[idx + 1].x);
			fNextCPY = xar2ttf(pCoord[idx + 1].y);
			fToX = xar2ttf(pCoord[idx + 2].x);
			fToY = xar2ttf(pCoord[idx + 2].y);
		}

		buf.push_back(bFirstHV); // vhcurveto or hvcurveto

		return idx;
	}

	while (cnt < 6)
	{
		if (m_fCurX == myround2(fPrevCPX, round) &&
			myround2(fNextCPY, round) == myround2(fToY, round) && 
			idx + 5 < size && (verb[idx + 3] & ~(BYTE)PT_CLOSEFIGURE) == PT_BEZIERTO &&
			myround2(fToY, round) == myround2(xar2ttf(pCoord[idx+3].y), round) &&
			myround2(xar2ttf(pCoord[idx + 4].x), round) == myround2(xar2ttf(pCoord[idx + 5].x), round))
			break;
		else if (m_fCurY == myround2(fPrevCPY, round) &&
			myround2(fNextCPX, round) == myround2(fToX, round) &&
			idx + 5 < size && (verb[idx + 3] & ~(BYTE)PT_CLOSEFIGURE) == PT_BEZIERTO &&
			myround2(fToX, round) == myround2(xar2ttf(pCoord[idx + 3].x), round) &&
			myround2(xar2ttf(pCoord[idx + 4].y), round) == myround2(xar2ttf(pCoord[idx + 5].y), round))
			break;

		AddNumber2(buf, myround2(fPrevCPX, round) - m_fCurX, round);
		AddNumber2(buf, myround2(fPrevCPY, round) - m_fCurY, round);
		AddNumber2(buf, myround2(fNextCPX, round) - myround2(fPrevCPX, round), round);
		AddNumber2(buf, myround2(fNextCPY, round) - myround2(fPrevCPY, round), round);
		AddNumber2(buf, myround2(fToX, round) - myround2(fNextCPX, round), round);
		AddNumber2(buf, myround2(fToY, round) - myround2(fNextCPY, round), round);
		m_fCurX = myround2(fToX, round);
		m_fCurY = myround2(fToY, round);
		++cnt;
		idx += 3;

		if (idx >= size || (verb[idx] & ~(BYTE)PT_CLOSEFIGURE) != PT_BEZIERTO)
			break;

		fPrevCPX = xar2ttf(pCoord[idx].x);
		fPrevCPY = xar2ttf(pCoord[idx].y);
		fNextCPX = xar2ttf(pCoord[idx + 1].x);
		fNextCPY = xar2ttf(pCoord[idx + 1].y);
		fToX = xar2ttf(pCoord[idx + 2].x);
		fToY = xar2ttf(pCoord[idx + 2].y);
	}

	buf.push_back(8);		// rrcurveto

	return idx;
}

/*
UINT SFNTWriter::curveto2(ByteVector& buf, DocCoord dcPrevCP, DocCoord dcNextCP, DocCoord dcTo, bool round)
{
	float fPrevCPX = xar2ttf(dcPrevCP.x);
	float fPrevCPY = xar2ttf(dcPrevCP.y);
	float fNextCPX = xar2ttf(dcNextCP.x);
	float fNextCPY = xar2ttf(dcNextCP.y);
	float fToX = xar2ttf(dcTo.x);
	float fToY = xar2ttf(dcTo.y);

	AddNumber2(buf, myround2(fPrevCPX, round) - m_fCurX, round);
	AddNumber2(buf, myround2(fPrevCPY, round) - m_fCurY, round);
	AddNumber2(buf, myround2(fNextCPX, round) - myround2(fPrevCPX, round), round);
	AddNumber2(buf, myround2(fNextCPY, round) - myround2(fPrevCPY, round), round);
	AddNumber2(buf, myround2(fToX, round) - myround2(fNextCPX, round), round);
	AddNumber2(buf, myround2(fToY, round) - myround2(fNextCPY, round), round);
	m_fCurX = myround2(xar2ttf(dcTo.x), round);
	m_fCurY = myround2(xar2ttf(dcTo.y), round);
	buf.push_back(8);		// rrcurveto
	return 1;
}
*/
/******************************************************************************
   Function:	SFNT::SFNTWriter::WriteGlyph
   Author:		Paul Gruzdev
   Purpose:     Given verbs and coords arrays, encode the glyph as type2 CFF
				into the buffer
   Returns:		void
   Parameters:	std::vector<BYTE> & buf  
   Parameters:	LONG width  
   Parameters:	BYTE * verb  
   Parameters:	DocCoord * pCoord  
   Parameters:	UINT count  
   Parameters:	bool round  
   Notes:       
******************************************************************************/
void SFNTWriter::WriteGlyph(ByteVector& buf, LONG width, BYTE* verb, DocCoord* pCoord, UINT count, bool round)
{
	m_fCurX = m_fCurY = 0; // Reset the current point
	int iCffWidth = (int)rint(xar2ttf(width));
	int iDelta = iCffWidth - (int)rint(xar2ttf(m_lNominalWidth));

	if (iCffWidth != rint(xar2ttf(m_lDefaultWidth)))
		// Nominal and default widths are stored as INTs in the private area, so here we need to
		// calc the difference between int values too
		// And they should match to the hmtx value of the width, where they are rounded up, so we need to
		// round up here as well! Otherwise the Acrobat preflight fails
		AddNumber2(buf, (float)iDelta, round);

	for (UINT i = 0; i < count; i++)
	{
		switch (verb[i] & ~(BYTE)PT_CLOSEFIGURE)
		{
		case PT_MOVETO:
			moveto2(buf, pCoord[i].x, pCoord[i].y, round);
			break;
		case PT_LINETO:
			{
				UINT cnt = lineto2(buf, &verb[i], &pCoord[i], count - i, round);
				i += cnt - 1; // i++ will compensate '-1'
			}
			break;
		case PT_BEZIERTO:
			{
				UINT cnt = curveto2(buf, &verb[i], &pCoord[i], count - i, round);
				i += cnt - 1; // i++ will compensate '-1'
			}
			break;
		}
	}

	buf.push_back(14); // end of char 
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::CreateCharstrings
   Author:		Paul Gruzdev
   Purpose:     Encode each Xara glyph as a CFF glyph char string, the format is
				described in 5177.Type2.pdf, and store it in the list of
				charstrings, computing the font's bounding box by the way
   Returns:		bool
   Parameters:	
   Notes:       Font bounding box is also calculated here, as well as xHeight
				and CapHeight
******************************************************************************/
bool SFNTWriter::CreateCharstrings()
{
	RECT bbox;
	//int idx = 0;
	m_szTotalCharstringsLength = 0;
	// First of all - create a .notdef charstring
	{
		ByteVector verb;
		std::vector<DocCoord> coord;
		LONG lWidth = 0;

		CreateNotDefPath(verb, coord, lWidth);

		{
			m_pSource->CalcBBox(&coord[0], &verb[0], (DWORD)verb.size(), bbox);
			m_CharProps[0].lCharWidth = lWidth;
			m_CharProps[0].rcBBox = bbox;
		}

		m_Charstrings.push_back(ByteVector());
		WriteGlyph(m_Charstrings.back(), lWidth, &verb[0], &coord[0], (UINT)verb.size(), false);
		
		{
/*			char psz[128];
			//sprintf_s(psz, 128, "d:/temp/cff/chrstr%.3d.cff", idx++);
			sprintf_s(psz, 128, "d:/temp/cff/chrstr%.3d.cff-my", idx++);
			FILE* fp = fopen(psz, "wb");
			//fseek(fp, 0, SEEK_END);
			//long len = ftell(fp);
			//rewind(fp);
			//m_Charstrings.back().resize(len);
			//fread(&m_Charstrings.back()[0], len, 1, fp);
			fwrite(&m_Charstrings.back()[0], m_Charstrings.back().size(), 1, fp);
			fclose(fp);*/
		}

		m_szTotalCharstringsLength += m_Charstrings.back().size();
	}

	bool bAllGood = true, bGetFontBBox = false;
	m_lXHeight = m_lCapHeight = 0;
	LONG lCapHeightCnt = 0, lXHeightCnt = 0;

	// Now iterate through all 'normal' glyphs
	for (size_t i = 1; i < m_CharProps.size(); i++)
	{
		CharProps* props = &m_CharProps[i];
		WCHAR wUnicode = props->wUnicode;
		PathVerb* pVerbs = NULL;
		DocCoord* pCoords = NULL;
		ULONG ulNumCoords = 0;

		VGlyphSource::GlyphState GState = m_pSource->GetGlyphPath(wUnicode, &pVerbs, &pCoords, &ulNumCoords);

		if (GState == VGlyphSource::keRegular || GState == VGlyphSource::keEmbedded)
		{
			RECT rcBox;
			{
				m_pSource->CalcBBox(pCoords, pVerbs, ulNumCoords, rcBox);
				//ok = GD->CalcBBox((POINT*)pCoords, pVerbs, (DWORD)ulNumCoords, &rcBox, TRUE);
				m_CharProps[i].rcBBox = rcBox;
				::UnionRect(&m_rcFontBounds, &bbox, &rcBox);
				bbox = m_rcFontBounds;
				// accumulate Cap Height to resolve it later
				if (wUnicode >= 'A' && wUnicode <= 'Z')
				{
					m_lCapHeight += rcBox.bottom;
					lCapHeightCnt++;
				}
				else if (wUnicode >= 'a' && wUnicode <= 'z')
				{
					m_lXHeight += rcBox.bottom;
					lXHeightCnt++;
				}
			}

			m_Charstrings.push_back(ByteVector());
			WriteGlyph(m_Charstrings.back(), m_CharProps[i].lCharWidth, pVerbs, pCoords, ulNumCoords, false);
			{
/*				char psz[128];
				//sprintf_s(psz, 128, "d:/temp/cff/chrstr%.3d.cff", idx++);
				sprintf_s(psz, 128, "d:/temp/cff/chrstr%.3d.cff-my", idx++);
				FILE* fp = fopen(psz, "wb");
				//fseek(fp, 0, SEEK_END);
				//long len = ftell(fp);
				//rewind(fp);
				//m_Charstrings.back().resize(len);
				//fread(&m_Charstrings.back()[0], len, 1, fp);
				fwrite(&m_Charstrings.back()[0], m_Charstrings.back().size(), 1, fp);
				fclose(fp);*/
			}

			m_szTotalCharstringsLength += m_Charstrings.back().size();
		}
		else if (GState == VGlyphSource::kePackedCFF) // our strings are already packed in pVerbs
		{
			m_Charstrings.push_back(ByteVector());
			m_Charstrings.back().resize(ulNumCoords);
			memcpy(&m_Charstrings.back()[0], pVerbs, ulNumCoords);
			bGetFontBBox = true;
			m_szTotalCharstringsLength += m_Charstrings.back().size();
		}
		else
			bAllGood = false;
	}

	if (bGetFontBBox && m_pSource2)
	{
		//VGlyphSource2* pSource2 = (VGlyphSource2*)m_pSource;
		m_rcFontBounds = m_pSource2->GetFontBBox();
		m_lCapHeight = m_pSource2->GetFontCapHeight();
		m_lXHeight = m_pSource2->GetFontXHeight();
	}
	else
	{
		if (lCapHeightCnt)
			m_lCapHeight /= lCapHeightCnt;
		if (lXHeightCnt)
			m_lXHeight /= lXHeightCnt;
	}

	return bAllGood;
}

void SFNTWriter::CreateFDSelect(std::vector<BYTE>& vFDSelect, size_t szGlyphCount)
{
	CCMemFile mem(NULL, 0, CCMemWrite);
	mem.write('\3');		// FDSelect version 3: ranges
	putshort(&mem, 1);		// we have only one range
	putshort(&mem, 0);		// the first GID in range should always be 0
	mem.write('\0');		// we have only one sub-font, it's index is 0
	putshort(&mem, (short)szGlyphCount); // The sentinel GID is set equal to the number of glyphs in the font.
	// That is, its value is 1 greater than the last GID in the font.)
	vFDSelect.resize(mem.tell());
	memcpy(&vFDSelect[0], mem.GetMemoryBuffer(), mem.tell());
	mem.close();
}

// void ReadPSChars(std::vector<std::vector<BYTE>>& strs, size_t& szTotalSize)
// {
// 	szTotalSize = 0;
// 	FILE* in = fopen("D:\\temp\\ttf\\chars.pschars", "rb");
// 	DWORD dwSize;
// 	fread(&dwSize, 4, 1, in);
// 	strs.resize(dwSize);
// 	for (DWORD i = 0; i < dwSize; i++)
// 	{
// 		DWORD dwCnt;
// 		fread(&dwCnt, 4, 1, in);
// 		strs[i].resize(dwCnt);
// 		fread(&strs[i][0], dwCnt, 1, in);
// 		szTotalSize += dwCnt;
// 	}
// 	fclose(in);
// }

/******************************************************************************
Function:	SFNT::SFNTWriter::WriteCFF
Author:		Paul Gruzdev
Purpose:     Create and write the whole CFF into a file
Returns:		bool
Parameters:	CCFile * pFile
Notes:
******************************************************************************/
LONG SFNTWriter::WriteCFF(CCFile* pFile)
{
	m_OutputFormat = keOpenType;
	CreateCharset();
	CreateCharstrings();
	WriteCFFHeader(pFile);
	WriteCFFNames(pFile, m_sPostScriptName.c_str());//m_sFullName.c_str());
	WriteTopDict(pFile);
	WriteStrings(pFile);

	// Global Subrs
	putshort(pFile, 0);

	ERROR3IF(m_iCharsetOff != pFile->tell(), "Calculated offset to charset differs from the real one");
	WriteCharset(pFile);
	ERROR3IF(m_iCharstringsOff != pFile->tell(), "Calculated offset to charstrings differs from the real one");
	WriteCharstrings(pFile);
	/*
	#ifdef _DEBUG
	CCDiskFile file;
	file.open(PathName(_T("d:/temp/cff/charstrings.cff")), ios::in | ios::out | ios::trunc | ios::binary);
	WriteCharstrings(&file);
	file.close();
	#endif*/

	ERROR3IF(m_iPrivateOff != pFile->tell(), "Calculated offset to private differs from the real one");
	// Write private
	pFile->write((void*)&m_Private[0], (UINT)m_Private.size());

	if (m_pSource2 && m_pSource2->HasLocalSubs())
	{
		PBYTE pbyLocalSubs = NULL;
		size_t szLocalSize = 0;
		m_pSource2->GetLocalSubs(pbyLocalSubs, szLocalSize);
		pFile->write((void*)pbyLocalSubs, (UINT)szLocalSize);
	}
	return pFile->tell();
}

/******************************************************************************
   Function:	SFNT::SFNTWriter::WriteCFFCID
   Author:		Paul Gruzdev
   Purpose:     Create and write the whole CFF into a file as CID font
   Returns:		bool
   Parameters:	CCFile * pFile  
   Notes:       
******************************************************************************/
LONG SFNTWriter::WriteCFFCID(CCFile* pFile)
{
	m_OutputFormat = keOpenType;
	CreateCharset();
	CreateCharstrings();
	WriteCFFHeader(pFile);
	WriteCFFNames(pFile, m_sPostScriptName.c_str());//m_sFullName.c_str());
	// fdselect and dict index are written in WriteCIDTopDict
	WriteCIDTopDict(pFile);
	WriteStrings(pFile);

	// Global Subrs
	putshort(pFile, 0);

	ERROR3IF(m_iCharsetOff != pFile->tell(), "Calculated offset to charset differs from the real one");
	WriteCharset(pFile);

	// Write fdselect
	pFile->write((void*)&m_FDSelect[0], (UINT)m_FDSelect.size());

	ERROR3IF(m_iCharstringsOff != pFile->tell(), "Calculated offset to charstrings differs from the real one");
	WriteCharstrings(pFile);

	// Write fdarray
	pFile->write((void*)&m_FDArray[0], (UINT)m_FDArray.size());

	ERROR3IF(m_iPrivateOff != pFile->tell(), "Calculated offset to private differs from the real one");
	// Write private. TODO: check whether it's actually needed in CID? no reference to it is written in the top dict
	pFile->write((void*)&m_Private[0], (UINT)m_Private.size());

	if (m_pSource2 && m_pSource2->HasLocalSubs())
	{
		PBYTE pbyLocalSubs = NULL;
		size_t szLocalSize = 0;
		m_pSource2->GetLocalSubs(pbyLocalSubs, szLocalSize);
		pFile->write((void*)pbyLocalSubs, (UINT)szLocalSize);
	}

	return pFile->tell();
}

static int getoffset(CCFile* ttf, int offsize)
{
	if (offsize == 1)
	{
		char off;
		ttf->read(off);
		return (int)off;
	}
	else if (offsize == 2)
		return getushort(ttf);
	else if (offsize == 3)
		return get3byte(ttf);

	return getlong(ttf);
}

//******************************************************************************
//  Function:	SFNT::SFNTReader::RepackOTF
//  Author:		Paul Gruzdev
//  Purpose:	As oppose to SubsetOTF, this re-creates and subsets the OTF font from scratch
//				thus hugely reducing file size. Glyph charstrings and local subrs
//				are taken from the source font.
//  Returns:	bool
//  Parameters:	const CharSet & sSubset
//  Parameters:	bool bCID
//  Notes:		
//******************************************************************************
bool SFNTReader::RepackOTF(const CharSet& sSubset, bool bCID)
{
	if (m_format != keOpenType || sSubset.Count() == 0)
		return false;

	bool bRetVal = false;

	if (!m_OS2.IsRead())
	{
		m_OS2.Read(map()[TABLEID_OS2]);
		m_head.Read(map()[TABLEID_HEAD]);
		m_hhea.Read(map()[TABLEID_HHEA]);
		m_maxp.Read(map()[TABLEID_MAXP]);
		ReadHmtx(map()[TABLEID_HMTX]);
		ReadCmap(map()[TABLEID_CMAP]);
	}

	CCMemFile* pMemFile = map()[TABLEID_CFF];
	OTFDataLoader loader;
	loader.OTFData.ClearData();
	loader.OTFData.uCFFStart = 0;
	loader.OTFData.uCFFLength = (UINT32)pMemFile->Size();
	PBYTE pbyCFF = pMemFile->GetMemoryBuffer();
	loader.ParseCFFGlyphs(pbyCFF, OTFDataLoader::ParseOptions::keKeepGlyphData);

	if (loader.OTFData.vpTopDicts.size() == 0 || loader.OTFData.vpTopDicts[0] == NULL)
		return false;

	m_pLoader = &loader;
	m_uni.clear();
	for (UINT i = 0; i < 65536; i++)
	{
		if (sSubset.Exists(i))
			m_uni.insert((WCHAR)i);
	}

	float fOldXaraEM = g_fXaraEM;
	g_fXaraEM = 8388607.0f; // increase the precision hugely
	SFNTWriter writer(this, SFNT::keOpenType, this);
	writer.SetWriteCID(bCID);

	TableMap map;
	writer.m_dwLastError = writer.GenerateTables(map);
	if (writer.GetLastError() == NO_ERROR)
	{
		this->m_tableMap.m_map.swap(map.m_map);
		m_format = SFNT::keOpenType;
		m_hmtx.clear();

		for (TableMap::iterator it = m_tableMap.begin(); it != m_tableMap.end(); ++it)
		{
			it->second->ChangeFileAccess(CCMemRead);
			it->second->seek(0);
		}
		bRetVal = true;
	}
	g_fXaraEM = fOldXaraEM;
	return bRetVal;
}

bool SFNTReader::SubsetCFF(CCMemFile* pMemFile, CCFile* pFile, CharSet& subset, std::set<ULONG>& subsetGID)
{
	if (m_CodeToGID.size() == 0)
		return false;

	OTFDataLoader loader;
	//memset(&loader.OTFData, 0, sizeof(OpenTypeFontData));
	loader.OTFData.ClearData();
	loader.OTFData.uCFFStart = 0;
	loader.OTFData.uCFFLength = (UINT32)pMemFile->Size();
	PBYTE pbyCFF = pMemFile->GetMemoryBuffer();
	loader.ParseCFFGlyphs(pbyCFF, OTFDataLoader::ParseOptions::keKeepGlyphData);

	if (loader.OTFData.vpTopDicts.size() == 0 || loader.OTFData.vpTopDicts[0] == NULL)
		return false;

	TopDicts* pTopDict = loader.OTFData.vpTopDicts[0];

	// Create an empty glyph with zero 
	// Now subset: free all non-used glyphs
	for (int i = 1; i < (int)pTopDict->Glyphs.vValues.size(); i++)
	{
		bool bKeepGlyph = GIDExists((ULONG)i, subset) || subsetGID.find(i) != subsetGID.end();

		if (!bKeepGlyph)
		{
			pTopDict->Glyphs.vValues[i].clear();
			if (pTopDict->iDefaultWidthX != 0)
				AddNumber2(pTopDict->Glyphs.vValues[i], (short)-pTopDict->iNominalWidthX, false);
			pTopDict->Glyphs.vValues[i].push_back(14); // add end of string char
		}
	}

	pFile->write('\1');		// Major version: 1
	pFile->write('\0');		// Minor version: 0
	pFile->write('\4');		// Header size in bytes
	pFile->write('\4');		// Absolute Offset size.

	size_t len = loader.OTFData.strCFFFontName.length();
	putshort(pFile, 1);		// One font name
	pFile->write('\1');		// Offset size
	pFile->write('\1');		// Offset to first name
	pFile->write('\1' + (char)len);
	pFile->write(loader.OTFData.strCFFFontName.c_str(), (UINT)len);

	putshort(pFile, 1);		// One top dict
	pFile->write('\2');		// Offset size
	putshort(pFile, 1);		// Offset to topdict
	FilePos iFinalPos = pFile->tell();
	putshort(pFile, 0);		// placeholder for final position (final offset in index points beyond last element)

	bool bCID = false;
	ByteVector vFDSelect;
	std::vector<FontDict2Data> fds;
	CCMemFile fda(NULL, 0, CCMemWrite);

	// CID extended fields
	if (pTopDict->iRosRegistry > 0)
	{
		bCID = true;
		dumpintoper(pFile, pTopDict->iRosRegistry, -1);		// Registry
		dumpintoper(pFile, pTopDict->iRosOrdering, -1);		// Ordering
		dumpintoper(pFile, pTopDict->iRosSupplement, (12 << 8) | 30);		// Supplement
		dumpdbloper(pFile, pTopDict->fCidFontVersion, (12 << 8) | 31);	// CIDFontVersion
		dumpintoper(pFile, pTopDict->iCidCount, (12 << 8) | 34);	// CIDCount

		SFNTWriter::CreateFDSelect(vFDSelect, pTopDict->vecTDCharset.size());	// Create FDSelect
		SFNTWriter::DumpCFFDictIndex(&fda, fds, loader.OTFData.vpSubDicts[0]->iSidFontname);
	}

	if (pTopDict->iVersion > 0)
		dumpintoper(pFile, pTopDict->iVersion, 0);
	if (pTopDict->iNotice > 0)
		dumpintoper(pFile, pTopDict->iNotice, 1);
	if (pTopDict->iFullName > 0)
		dumpintoper(pFile, pTopDict->iFullName, 2);
	if (pTopDict->iFamilyName > 0)
		dumpintoper(pFile, pTopDict->iFamilyName, 3);
	if (pTopDict->iWeight > 0)
		dumpintoper(pFile, pTopDict->iWeight, 4);

	if (pTopDict->iCopyright > 0)
		dumpintoper(pFile, pTopDict->iCopyright, (12 << 8) + 0);

	if (pTopDict->iPostScriptCode > 0)
		dumpintoper(pFile, pTopDict->iPostScriptCode, (12 << 8) + 21);
	if (pTopDict->iBaseFontName > 0)
		dumpintoper(pFile, pTopDict->iBaseFontName, (12 << 8) + 22);
	if (pTopDict->iSidFontname > 0)
		dumpintoper(pFile, pTopDict->iSidFontname, (12 << 8) + 38);

	if (pTopDict->iIsFixedPitch)
		dumpintoper(pFile, 1, (12 << 8) | 1);

	if (pTopDict->fItalicAngle != 0)
		dumpdbloper(pFile, pTopDict->fItalicAngle, (12 << 8) | 2);

	if (pTopDict->fUnderlinePos != -100)
		dumpdbloper(pFile, pTopDict->fUnderlinePos, (12 << 8) | 3);

	if (pTopDict->fUnderlineWidth != 50)
		dumpdbloper(pFile, pTopDict->fUnderlineWidth, (12 << 8) | 4);

	if (!bCID && (pTopDict->fFontMatrix[0] != 0.001f || pTopDict->fFontMatrix[1] != 0.0f || pTopDict->fFontMatrix[2] != 0.0f ||
		pTopDict->fFontMatrix[3] != 0.001f || pTopDict->fFontMatrix[4] != 0.0f || pTopDict->fFontMatrix[5] != 0.0f))
	{
		dumpdbl(pFile, pTopDict->fFontMatrix[0]);
		dumpint(pFile, (int)pTopDict->fFontMatrix[1]);
		dumpint(pFile, (int)pTopDict->fFontMatrix[2]);
		dumpdbl(pFile, pTopDict->fFontMatrix[3]);
		dumpint(pFile, (int)pTopDict->fFontMatrix[4]);
		dumpintoper(pFile, (int)pTopDict->fFontMatrix[5], (12 << 8) | 7);
	}

	dumpdbl(pFile, pTopDict->fFontBoundingBox[0]);
	dumpdbl(pFile, pTopDict->fFontBoundingBox[1]);
	dumpdbl(pFile, pTopDict->fFontBoundingBox[2]);
	dumpdbloper(pFile, pTopDict->fFontBoundingBox[3], 5);

	size_t szStrings = pTopDict->iStringsSize;
	//size_t szGlobal = sizeof(short);	// Single entry: 0, no globals
	size_t szGlobal = pTopDict->iGSubsOff > 0 ? pTopDict->iGSubsSize : sizeof(short);
	size_t szEncoding = 0;				// no encoding
	// otfTools fills the first charset value for .notdef but it's not saved to file
	size_t szCharset = (pTopDict->vecTDCharset.size() - 1) * sizeof(short int) + 1;
	size_t szTotalCharstringsLen = 0;
	size_t szCharstrings = SFNTWriter::GetCharstringsSize(pTopDict->Glyphs.vValues, szTotalCharstringsLen);
	size_t szPrivate = pTopDict->iPrivateSize;
	size_t szFDSelect = vFDSelect.size();
	size_t szFDArray = fda.Size();
	size_t szBase = pFile->tell();
	bool bLongOffset = false;

	if (bCID || (szBase + 6 * 3 + szStrings + szGlobal + szEncoding + szCharset + szCharstrings + szPrivate > 32767))
	{
		bLongOffset = true;
		szBase += bCID ? 5 * 4 + 4 + 2 : 5 * 5 + 4;
	}
	else
		szBase += 5 * 3 + 4;

	int iCharsetOff =     (int)(szBase + szStrings + szGlobal);
	int iFDSelectOff =    (int)(szBase + szStrings + szGlobal + szCharset);
	int iCharstringsOff = (int)(szBase + szStrings + szGlobal + szCharset + szFDSelect);
	int iFDArrayOff =     (int)(szBase + szStrings + szGlobal + szCharset + szFDSelect + szCharstrings);
	int iPrivateOff =     (int)(szBase + szStrings + szGlobal + szCharset + szFDSelect + szCharstrings + szFDArray);

	ByteVector vFDArray;
	if (bCID)
	{
		size_t szPrivateAccum = 0;
		for (size_t i = 0; i < fds.size(); ++i)
		{
			fda.seek(fds[i].iFillinDictMark);
			dumpsizedint(&fda, false, (int)loader.OTFData.vpSubDicts[i]->iPrivateSize, -1);	// Private len
			dumpsizedint(&fda, true, iPrivateOff + (int)szPrivateAccum, 18);	// Private offset
			szPrivateAccum += loader.OTFData.vpSubDicts[i]->iPrivateSize;
		}

		SFNTWriter::CreateFDArray(&fda, vFDArray);
		fda.close();
	}

#ifdef _DEBUG
	m_iCharsetOff = iCharsetOff;
	m_iCharstringsOff = iCharstringsOff;
	m_iPrivateOff = iPrivateOff;
#endif

	// Offset to charset (oper=15)
	dumpsizedint(pFile, bLongOffset, iCharsetOff, 15);

	if (bCID)
		// Offset to fdselect
		dumpsizedint(pFile, bLongOffset, iFDSelectOff, (12 << 8) | 37);
	else
		// Offset to encoding (oper=16), (not for CID)
		dumpsizedint(pFile, bLongOffset, 0, 16);
	// Offset to charstrings (oper=17)
	dumpsizedint(pFile, bLongOffset, iCharstringsOff, 17);

	if (bCID)
		// Offset to fdarray
		dumpsizedint(pFile, bLongOffset, iFDArrayOff, (12 << 8) | 36);
	else
	{
		// Length of, and Offset to private (oper=18) here (not for CID )
		dumpsizedint(pFile, bLongOffset, (int)szPrivate, -1);
		dumpsizedint(pFile, bLongOffset, iPrivateOff, 18);
	}

	size_t szEotop = szBase - iFinalPos - 1;

	pFile->seekIn(iFinalPos, ios_base::beg);
	putshort(pFile, (short int)szEotop);
	pFile->seekIn(0, ios_base::end);

	// Now write strings
	pFile->write(&pbyCFF[pTopDict->iStringsOff], pTopDict->iStringsSize);

	// Global Subrs
	if (pTopDict->iGSubsOff > 0)
		pFile->write(&pbyCFF[pTopDict->iGSubsOff], pTopDict->iGSubsSize);
	else
		putshort(pFile, 0);

	ERROR3IF(m_iCharsetOff != pFile->tell(), "Calculated offset to charset differs from the real one");
	// Write the charset
	// Here we use a format 0 charset. ie. an array of SIDs in random order
	// First element must be ".notdef" and is omitted
	pFile->write('\0');
	for (std::vector<UINT16>::iterator it = pTopDict->vecTDCharset.begin() + 1; it != pTopDict->vecTDCharset.end(); ++it)
		putshort(pFile, (short int)*it);

	if (bCID)
		// Write fdselect
		pFile->write((void*)&vFDSelect[0], (UINT)vFDSelect.size());

	ERROR3IF(m_iCharstringsOff != pFile->tell(), "Calculated offset to charstrings differs from the real one");
	SFNTWriter::WriteCharstrings(pFile, pTopDict->Glyphs.vValues, szTotalCharstringsLen);

	if (bCID)
		// Write fdarray
		pFile->write((void*)&vFDArray[0], (UINT)vFDArray.size());

	ERROR3IF(m_iPrivateOff != pFile->tell(), "Calculated offset to private differs from the real one");
	if (bCID)
	{
		for (size_t i = 0; loader.OTFData.vpSubDicts[i] != NULL; i++)
			if (loader.OTFData.vpSubDicts[i]->iPrivateSize)
				pFile->write(&pbyCFF[loader.OTFData.vpSubDicts[i]->iPrivateOffset], loader.OTFData.vpSubDicts[i]->iPrivateSize);
	}
	else
		// Now write the private data
		pFile->write(&pbyCFF[pTopDict->iPrivateOffset], pTopDict->iPrivateSize);

	// Now write the local subrs
	if (pTopDict->iLocalSubSize)
		pFile->write(&pbyCFF[pTopDict->iPrivateOffset + pTopDict->iSubrsOff], pTopDict->iLocalSubSize);

	return pFile->good() ? true : false;
}

} // namespace SFNT
