/********************************************************************************************

>	gpos.cpp

Author:		Paul Gruzdev
Created:	30/08/2016
Purpose:	Write out the GPOS (glyph positioning) table. This can be a part of a SFNT
			file that represents a font in CFF format. GPOS can be used to provide 
			kerning information along with CFF.

********************************************************************************************/

#include "stdafx.h"
#include "sfnt.h"
#include "gpos.h"

#ifdef _DEBUG
#include "../XPFilter/common/StringUtils.h"
#endif
// Enables improved memory tracking. Must be defined after CC_IMPLEMENT_DYNCREATE macros.
#ifdef CAMELOT_PRODUCT
#define new CAM_DEBUG_NEW
#endif
//#define DUMP_EXTRA_INFO		// Write out extra debug files into d:/temp/cff*
// MS has a parallel set of script tags for their new
//  Indic font shaper
//#define USE_SECOND_INDIC_SCRIPTS

// OpenType does not document 'dflt' as a language, but we'll use it anyway.
// (Adobe uses it too) we'll turn it into a default entry when we output it.
#define DEFAULT_LANG		CHR('d','f','l','t')
// The OpenType spec says in one place that the default script is 'dflt' and
//  in another that it is 'DFLT'. 'DFLT' is correct
#define DEFAULT_SCRIPT		CHR('D','F','L','T')
#define REQUIRED_FEATURE	CHR(' ','R','Q','D')

namespace SFNT
{

// This file contains routines to create the otf gpos and gsub tables and their 
//  attendant subtables 

// Undocumented fact: ATM (which does kerning for otf fonts in Word) can't handle features with multiple lookups 

// Undocumented fact: Only one feature with a given tag allowed per script/lang 
//  So if we have multiple lookups with the same tag they must be merged into 
//  one feature with many lookups 

// scripts (for opentype) that I understand 
// see also list in fontforge/lookups.c mapping script tags to friendly names 

static ULONG g_scripts[][15] = 
{
	{ CHR('a', 'r', 'a', 'b'), 0x0600, 0x06ff, 0xfb50, 0xfdff, 0xfe70, 0xfefe },	// Arabic 
	{ CHR('a', 'r', 'a', 'm'), 0x820, 0x83f },										// Aramaic 
	{ CHR('a', 'r', 'm', 'n'), 0x0530, 0x058f, 0xfb13, 0xfb17 },					// Armenian 
	{ CHR('b', 'a', 'l', 'i'), 0x1b00, 0x1b7f },									// Balinese 
	{ CHR('b', 'e', 'n', 'g'), 0x0980, 0x09ff },									// Bengali 
	{ CHR('b', 'l', 'i', 's'), 0x12200, 0x124ff },									// Bliss symb 
	{ CHR('b', 'o', 'p', 'o'), 0x3100, 0x312f, 0x31a0, 0x31bf },					// Bopomofo 
	{ CHR('b', 'r', 'a', 'i'), 0x2800, 0x28ff },									// Braille 
	{ CHR('b', 'u', 'g', 'i'), 0x1a00, 0x1a1f },									// Buginese 
	{ CHR('b', 'u', 'h', 'd'), 0x1740, 0x1753 },									// Buhid 
	{ CHR('b', 'y', 'z', 'm'), 0x1d000, 0x1d0ff },									// Byzantine M
	{ CHR('c', 'a', 'n', 's'), 0x1400, 0x167f },									// Canadian Syl
	{ CHR('c', 'a', 'r', 'i'), 0x0, 0x0 },											// Carian 
	{ CHR('c', 'h', 'a', 'm'), 0x0, 0x0 },											// Cham 
	{ CHR('c', 'h', 'e', 'r'), 0x13a0, 0x13ff },									// Cherokee 
	{ CHR('c', 'i', 'r', 't'), 0x12080, 0x120ff },									// Cirth 
	{ CHR('h', 'a', 'n', 'i'), 0x3300, 0x9fff, 0xf900, 0xfaff, 0x020000, 0x02ffff },// CJKIdeogra 
	{ CHR('c', 'o', 'p', 't'), 0x2c80, 0x2cff },									// Coptic 
	{ CHR('c', 'p', 'm', 'n'), 0x10800, 0x1083f },									// Cypriot 
	{ CHR('c', 'y', 'r', 'l'), 0x0400, 0x052f, 0x1d2b, 0x1d2b, 0x1d78, 0x1d78,
	0x2de0, 0x2dff, 0xa640, 0xa6ff },												// Cyrillic 
	{ CHR('d', 's', 'r', 't'), 0x10400, 0x1044f },									// Deseret 
	{ CHR('d', 'e', 'v', 'a'), 0x0900, 0x097f },									// Devanagari 
	{ CHR('e', 't', 'h', 'i'), 0x1200, 0x139f },									// Ethiopic 
	{ CHR('g', 'e', 'o', 'r'), 0x1080, 0x10ff },									// Georgian 
	{ CHR('g', 'l', 'a', 'g'), 0x1080, 0x10ff },									// Glagolitic 
	{ CHR('g', 'o', 't', 'h'), 0x10330, 0x1034a },									// Gothic 
	{ CHR('g', 'r', 'e', 'k'), 0x0370, 0x03ff, 0x1f00, 0x1fff },					// Greek 
	{ CHR('g', 'u', 'j', 'r'), 0x0a80, 0x0aff },									// Gujarati 
	{ CHR('g', 'u', 'r', 'u'), 0x0a00, 0x0a7f },									// Gurmukhi 
	{ CHR('h', 'a', 'n', 'g'), 0xac00, 0xd7af, 0x3130, 0x319f, 0xffa0, 0xff9f },	// Hangul 
	{ CHR('h', 'a', 'n', 'o'), 0x1720, 0x1734 },									// Hanunoo 
	// I'm not sure what the difference is between the 'hang' tag and the 'jamo' 
	//  tag. 'Jamo' is said to be the precomposed forms, but what's 'hang'? 
	{ CHR('h', 'e', 'b', 'r'), 0x0590, 0x05ff, 0xfb1e, 0xfb4f },					// Hebrew 
#if 0	// Hiragana used to have its own tag, but has since been merged with katakana 
	// Hiragana { CHR('h', 'i', 'r', 'a'), 0x3040, 0x309f },
#endif
	{ CHR('j', 'a', 'm', 'o'), 0x1100, 0x11ff, 0x3130, 0x319f, 0xffa0, 0xffdf },	// Hangul Jamo
	{ CHR('j', 'a', 'v', 'a'), 0 },	// Javanese. MS has a tag, but there is no unicode range 
	{ CHR('k', 'a', 'n', 'a'), 0x3040, 0x30ff, 0xff60, 0xff9f },					// Katakana 
	{ CHR('k', 'a', 'l', 'i'), 0 },													// Kayah Li 
	{ CHR('k', 'n', 'd', 'a'), 0x0c80, 0x0cff },									// Kannada 
	{ CHR('k', 'h', 'a', 'r'), 0x10a00, 0x10a5f },									// Kharosthi 
	{ CHR('k', 'h', 'm', 'r'), 0x1780, 0x17ff },									// Khmer 
	{ CHR('l', 'a', 't', 'n'), 0x0041, 0x005a, 0x0061, 0x007a,						// Latin 
	0x00c0, 0x02af, 0x1d00, 0x1eff, 0xfb00, 0xfb0f, 0xff00, 0xff5f, 0xa770, 0xa7ff },
	{ CHR('l', 'a', 'o', ' '), 0x0e80, 0x0eff },									// Lao 
	{ CHR('l', 'e', 'p', 'c'), 0 },													// Lepcha 
	{ CHR('l', 'i', 'm', 'b'), 0x1900, 0x194f },									// Limbu 
//	{ CHR('l','i','n','a'), 0x10180, 0x102cf }, // What happened to linear A?		// Linear A 	
	{ CHR('l', 'i', 'n', 'b'), 0x10000, 0x100fa },									// Linear B 
	{ CHR('l', 'y', 'c', 'i'), 0 },													// Lycian 
	{ CHR('l', 'y', 'd', 'i'), 0 },													// Lydian 
	{ CHR('m', 'l', 'y', 'm'), 0x0d00, 0x0d7f },									// Malayalam 
	// Mathematical Alphanumeric Symbols 
	{ CHR('m', 'a', 't', 'h'), 0x1d400, 0x1d7ff },
	{ CHR('m', 'o', 'n', 'g'), 0x1800, 0x18af },									// Mongolian 
	{ CHR('m', 'u', 's', 'c'), 0x1d100, 0x1d1ff },									// Musical 
	{ CHR('m', 'y', 'm', 'r'), 0x1000, 0x107f },									// Myanmar 
	{ CHR('t', 'a', 'l', 'u'), 0 },													// New Tai Lue
	{ CHR('n', 'k', 'o', ' '), 0x07c0, 0x07fa },									// N'Ko 
	{ CHR('o', 'g', 'a', 'm'), 0x1680, 0x169f },									// Ogham 
	{ CHR('o', 'l', 'c', 'k'), 0 },													// Ol Chiki 
	{ CHR('i', 't', 'a', 'l'), 0x10300, 0x1031e },									// Old Italic 
	{ CHR('p', 'e', 'r', 'm'), 0x10350, 0x1037f },									// Old Permic 
	// Old Persian cuneiform 
	{ CHR('x', 'p', 'e', 'o'), 0x103a0, 0x103df },
	{ CHR('o', 'r', 'y', 'a'), 0x0b00, 0x0b7f },									// Oriya 
	{ CHR('o', 's', 'm', 'a'), 0x10480, 0x104a9 },									// Osmanya 
	{ CHR('p', 'h', 'a', 'g'), 0xa840, 0xa87f },									// Phags-pa 
	{ CHR('p', 'h', 'n', 'x'), 0x10900, 0x1091f },									// Phoenician 
	{ CHR('p', 'l', 'r', 'd'), 0x104b0, 0x104d9 },									// Pollard 
	{ CHR('r', 'j', 'n', 'g'), 0 },													// Rejang 
	{ CHR('r', 'o', 'r', 'o'), 0 },													// Rongorongo 
	{ CHR('r', 'u', 'n', 'r'), 0x16a0, 0x16ff },									// Runic 
	{ CHR('s', 'a', 'u', 'r'), 0 },													// Saurashtra
	{ CHR('s', 'h', 'a', 'w'), 0x10450, 0x1047f },									// Shavian 
	{ CHR('s', 'i', 'n', 'h'), 0x0d80, 0x0dff },									// Sinhala 
	// Sumero-Akkadian Cuneiform 
	{ CHR('x', 's', 'u', 'x'), 0x12000, 0x1236e },
	{ CHR('s', 'u', 'n', 'd'), 0 },													// Sundanese 
	// Syloti Nagri 
	{ CHR('s', 'y', 'l', 'o'), 0xa800, 0xa82f },
	{ CHR('s', 'y', 'r', 'c'), 0x0700, 0x074f },									// Syriac 
	{ CHR('t', 'a', 'g', 'l'), 0x1700, 0x1714 },									// Tagalog 
	{ CHR('t', 'a', 'g', 'b'), 0x1760, 0x1773 },									// Tagbanwa 
	{ CHR('t', 'a', 'l', 'e'), 0x1950, 0x1974 },									// Tai Le 
	{ CHR('t', 'a', 'l', 'u'), 0x1980, 0x19df },									// Tai Lu 
	{ CHR('t', 'a', 'm', 'l'), 0x0b80, 0x0bff },									// Tamil 
	{ CHR('t', 'e', 'l', 'u'), 0x0c00, 0x0c7f },									// Telugu 
	{ CHR('t', 'e', 'n', 'g'), 0x12000, 0x1207f },									// Tengwar 
	{ CHR('t', 'h', 'a', 'a'), 0x0780, 0x07bf },									// Thaana 
	{ CHR('t', 'h', 'a', 'i'), 0x0e00, 0x0e7f },									// Thai 
	{ CHR('t', 'i', 'b', 't'), 0x0f00, 0x0fff },									// Tibetan 
	{ CHR('t', 'f', 'n', 'g'), 0x2d30, 0x2d7f },									// Tifinagh 
	{ CHR('u', 'g', 'a', 'r'), 0x10380, 0x1039d },									// Ugaritic 
	{ CHR('y', 'i', ' ', ' '), 0xa000, 0xa4c6 },									// Yi 
	{ 0 }
};

/******************************************************************************
   Function:	ScriptFromUnicode
   Author:		Paul Gruzdev, based on fontforge/tottfgpos.c
   Purpose:		Given a unicode value, return the script tag suitable to put into
				GPOS script table.
   Returns:		ULONG
   Parameters:	int u
   Notes:       
******************************************************************************/
ULONG ScriptFromUnicode(WCHAR u)
{
	if (u == ' ' || u >= '0' && u <= '9') return CHR('l', 'a', 't', 'n');
	int s, k;

	if (u != (WCHAR)-1) 
	{
		for (s = 0; g_scripts[s][0] != 0; ++s) 
		{
			for (k = 1; g_scripts[s][k + 1] != 0; k += 2)
				if (u >= (int)g_scripts[s][k] && u <= g_scripts[s][k + 1])
					break;
			if (g_scripts[s][k + 1] != 0)
				break;
		}

		if (g_scripts[s][0] != 0)
		{
			ULONG script = g_scripts[s][0];

#ifdef USE_SECOND_INDIC_SCRIPTS
			{
				// MS has a parallel set of script tags for their new
				//  Indic font shaper
				if (script == CHR('b', 'e', 'n', 'g')) script = CHR('b', 'n', 'g', '2');
				else if (script == CHR('d', 'e', 'v', 'a')) script = CHR('d', 'e', 'v', '2');
				else if (script == CHR('g', 'u', 'j', 'r')) script = CHR('g', 'j', 'r', '2');
				else if (script == CHR('g', 'u', 'r', 'u')) script = CHR('g', 'u', 'r', '2');
				else if (script == CHR('k', 'n', 'd', 'a')) script = CHR('k', 'n', 'd', '2');
				else if (script == CHR('m', 'l', 'y', 'm')) script = CHR('m', 'l', 'm', '2');
				else if (script == CHR('o', 'r', 'y', 'a')) script = CHR('o', 'r', 'y', '2');
				else if (script == CHR('t', 'a', 'm', 'l')) script = CHR('t', 'm', 'l', '2');
				else if (script == CHR('t', 'e', 'l', 'u')) script = CHR('t', 'e', 'l', '2');
			}
#endif
			return(script);
		}
	}

	return DEFAULT_SCRIPT;
}

ULONG ScriptFromUnicode(std::wstring str)
{
	for (size_t i = 0; i < str.length(); i++)
	{
		ULONG script = ScriptFromUnicode(str[i]);
		if (script != DEFAULT_SCRIPT)
			return script;
	}
	return DEFAULT_SCRIPT;
}

/******************************************************************************
   Function:	_FeatureOrderId
   Author:		Paul Gruzdev
   Purpose:		This is the order in which features should be executed (placed
				in GPOS/GSUB tables)

   Returns:		int
   Parameters:	int isgpos
   Parameters:	UINT32 tag
   Notes:       
******************************************************************************/
int _FeatureOrderId(bool isgpos, UINT32 tag)
{
	if (!isgpos) 
	{
		// GSUB ordering 
		switch (tag)
		{
		case CHR('c', 'c', 'm', 'p'):	// Must be first? 
			return(-2);
		case CHR('l', 'o', 'c', 'l'):	// Language dependent letter forms (serbian uses some different glyphs than russian) 
			return(-1);
		case CHR('i', 's', 'o', 'l'):
			return(0);
		case CHR('j', 'a', 'l', 't'):		// must come after 'isol' 
			return(1);
		case CHR('f', 'i', 'n', 'a'):
			return(2);
		case CHR('f', 'i', 'n', '2'):
		case CHR('f', 'a', 'l', 't'):		// must come after 'fina' 
			return(3);
		case CHR('f', 'i', 'n', '3'):
			return(4);
		case CHR('m', 'e', 'd', 'i'):
			return(5);
		case CHR('m', 'e', 'd', '2'):
			return(6);
		case CHR('i', 'n', 'i', 't'):
			return(7);

		case CHR('r', 't', 'l', 'a'):
			return(100);
		case CHR('s', 'm', 'c', 'p'): case CHR('c', '2', 's', 'c'):
			return(200);

		case CHR('r', 'l', 'i', 'g'):
			return(300);
		case CHR('c', 'a', 'l', 't'):
			return(301);
		case CHR('l', 'i', 'g', 'a'):
			return(302);
		case CHR('d', 'l', 'i', 'g'): case CHR('h', 'l', 'i', 'g'):
			return(303);
		case CHR('c', 's', 'w', 'h'):
			return(304);
		case CHR('m', 's', 'e', 't'):
			return(305);

		case CHR('f', 'r', 'a', 'c'):
			return(306);

			// Indic processing 
		case CHR('n', 'u', 'k', 't'):
		case CHR('p', 'r', 'e', 'f'):
			return(301);
		case CHR('a', 'k', 'h', 'n'):
			return(302);
		case CHR('r', 'p', 'h', 'f'):
			return(303);
		case CHR('b', 'l', 'w', 'f'):
			return(304);
		case CHR('h', 'a', 'l', 'f'):
		case CHR('a', 'b', 'v', 'f'):
			return(305);
		case CHR('p', 's', 't', 'f'):
			return(306);
		case CHR('v', 'a', 't', 'u'):
			return(307);

		case CHR('p', 'r', 'e', 's'):
			return(310);
		case CHR('b', 'l', 'w', 's'):
			return(311);
		case CHR('a', 'b', 'v', 's'):
			return(312);
		case CHR('p', 's', 't', 's'):
			return(313);
		case CHR('c', 'l', 'i', 'g'):
			return(314);

		case CHR('h', 'a', 'l', 'n'):
			return(320);
			// end indic ordering 

		case CHR('a', 'f', 'r', 'c'):
		case CHR('l', 'j', 'm', 'o'):
		case CHR('v', 'j', 'm', 'o'):
			return(350);
		case CHR('v', 'r', 't', '2'): case CHR('v', 'e', 'r', 't'):
			return(1010);		// Documented to come last 

			// Unknown things come after everything but vert/vrt2 
		default:
			return(1000);
		}
	}
	else 
	{
		// GPOS ordering 
		switch (tag)
		{
		case CHR('c', 'u', 'r', 's'):
			return(0);
		case CHR('d', 'i', 's', 't'):
			return(100);
		case CHR('b', 'l', 'w', 'm'):
			return(201);
		case CHR('a', 'b', 'v', 'm'):
			return(202);
		case CHR('k', 'e', 'r', 'n'):
			return(300);
		case CHR('m', 'a', 'r', 'k'):
			return(400);
		case CHR('m', 'k', 'm', 'k'):
			return(500);
			// Unknown things come after everything  
		default:
			return(1000);
		}
	}
}

/******************************************************************************
   Function:	FeatureScriptTagInFeatureScriptList
   Author:		Paul Gruzdev
   Purpose:		Find whether a particular script of a particular feature is present in a feature script lang list
   Returns:		bool
   Parameters:	UINT32 feature
   Parameters:	UINT32 script
   Parameters:	std::vector<FeatureScriptLangList * > & fv
   Notes:       
******************************************************************************/
bool FeatureScriptTagInFeatureScriptList(UINT32 feature, UINT32 script, std::vector<FeatureScriptLangList*>& fv)
{
	for (std::vector<FeatureScriptLangList*>::iterator it = fv.begin(); it != fv.end(); ++it)
	{
		FeatureScriptLangList* fl = *it;
		if (fl->uFeatureTag == feature) 
		{
			for (std::vector<ScriptLangList*>::iterator it = fl->vScripts.begin(); it != fl->vScripts.end(); ++it)
			{
				if ((*it)->uScript == script)
					return true;
			}
		}
	}

	return false;
}

bool OTLookupCompare::operator() (OTLookup* lhs, OTLookup* rhs) const
{
	int left  = _FeatureOrderId(lhs->IsGPos(), lhs->vFeatures.front()->uFeatureTag);
	int right = _FeatureOrderId(rhs->IsGPos(), rhs->vFeatures.front()->uFeatureTag);
	return left < right;
}

bool OTLookupFileCompare::operator() (const std::pair<OTLookup*, DWORD> lhs, const std::pair<OTLookup*, DWORD> rhs) const
{
	if (lhs.second == rhs.second)
		return lhs.first < rhs.first;

	return lhs.second < rhs.second;
}

bool OTLangCompare::operator() (UINT32 _ui1, UINT32 _ui2) const
{
	if (_ui1 == _ui2) return false;
	// The default language is magic, and should come first in the list even
	//  if that is not true alphabetical order
	if (_ui1 == DEFAULT_LANG)
		return true;
	if (_ui2 == DEFAULT_LANG)
		return false;

	if (_ui1 > _ui2)
		return false;
	if (_ui1 < _ui2)
		return true;

	return false;
}

void ProcessLookupFindScripts(OTLookup* test, std::set<UINT32>& scripts)
{
	if (test->uUnused)
		return;

	FeatureScriptLangList* fl;
	ScriptLangList* sl;

	for (size_t fc = 0; fc <test->vFeatures.size(); fc++)
	{
		fl = test->vFeatures[fc];
		if (fl->uIsMac)
			continue;

		for (size_t j = 0; j < fl->vScripts.size(); j++)
		{
			sl = fl->vScripts[j];
			scripts.insert(sl->uScript);
		}
	}
}

void ProcessLookupFindLangs(OTLookup* test, UINT32 script, OTLangSet& langs)
{
	if (test->uUnused)
		return;

	for (size_t fc = 0; fc < test->vFeatures.size(); fc++)
	{
		FeatureScriptLangList* fl = test->vFeatures[fc];

		for (size_t j = 0; j < fl->vScripts.size(); j++)
		{
			ScriptLangList* sl = fl->vScripts[j];
			if (sl->uScript == script)
			{
				for (INT32 l = 0; l < sl->iLangCnt; ++l)
				{
					int lang;
					if (l < MAX_LANG)
						lang = sl->uLangs[l];
					else
						lang = sl->vMoreLangs[l - MAX_LANG];

					langs.insert(lang);
				}
			}
		}
	}
}

void ProcessLookupFindFeatures(SFNTWriter::ETableType type, OTLookup* test, UINT32 script, UINT32 lang, std::set<UINT32>& features)
{
	if (test->uUnused)
		return;

	for (size_t fc = 0; fc < test->vFeatures.size(); fc++)
	{
		FeatureScriptLangList* fl = test->vFeatures[fc];
		if (fl->uIsMac && type != -2)
			continue;

		if (script == 0xffffffff) 
		{
			features.insert(fl->uFeatureTag);
		}
		else
		{
			for (size_t sc = 0; sc < fl->vScripts.size(); sc++)
			{
				ScriptLangList* sl = fl->vScripts[sc];
				if (sl->uScript == script)
				{
					int matched = false;
					if (fl->uIsMac && type == -2)
						matched = true;
					else
					{
						for (INT32 l = 0; l < sl->iLangCnt; ++l)
						{
							UINT32 testlang;
							if (l < MAX_LANG)
								testlang = sl->uLangs[l];
							else
								testlang = sl->vMoreLangs[l - MAX_LANG];

							if (testlang == lang)
							{
								matched = true;
								break;
							}
						}

						if (matched)
							features.insert(fl->uFeatureTag);
					}
				}
			}
		}
	}

}

void ProcessLookupFindFeatureLookups(OTLookup* test, UINT32 script, UINT32 lang, UINT32 feature, std::vector<OTLookup*>& lookups)
{
	if (test->uUnused)
		return;

	for (size_t fc = 0; fc < test->vFeatures.size(); fc++)
	{
		FeatureScriptLangList* fl = test->vFeatures[fc];

		if (fl->uFeatureTag == feature) 
		{
			for (size_t sc = 0; sc < fl->vScripts.size(); sc++)
			{
				ScriptLangList* sl = fl->vScripts[sc];
				if (sl->uScript == script) 
				{
					for (INT32 l = 0; l < sl->iLangCnt; ++l) 
					{
						UINT32 testlang;
						if (l < MAX_LANG)
							testlang = sl->uLangs[l];
						else
							testlang = sl->vMoreLangs[l - MAX_LANG];

						if (testlang == lang) 
						{
							lookups.push_back(test);
							goto found;
						}
					}
				}
			}
		}
	}
found:;

}

bool SFNTWriter::SFScriptsInLookups(ETableType type, std::set<UINT32>& scripts)
{
	// Presumes that either SFFindUnusedLookups or SFFindClearUnusedLookupBits 
	//  has been called first 
	// Since MS will sometimes ignore a script if it isn't found in both 
	//  GPOS and GSUB we want to return the same script list no matter 
	//  what the setting of gpos ... so we totally ignore that argument 
	//  and always look at both sets of lookups 
	/*
	Sergey Malkin from MicroSoft tells me:
	Each shaping engine in Uniscribe can decide on its requirements for
	layout tables - some of them require both GSUB and GPOS, in some cases
	any table present is enough, or it can work without any table.

	Sometimes, purpose of the check is to determine if font is supporting
	particular script - if required tables are not there font is just
	rejected by this shaping engine. Sometimes, shaping engine can not just
	reject the font because there are fonts using older shaping technologies
	we still have to support, so it uses some logic when to fallback to
	legacy layout code.

	In your case this is Hebrew, where both tables are required to use
	OpenType processing. Arabic requires both tables too, Latin requires
	GSUB to execute GPOS. But in general, if you have both tables you should
	be safe with any script to get fully featured OpenType shaping.

	In other words, if we have a Hebrew font with just GPOS features they won't work,
	and MS will not use the font at all. We must add a GSUB table. In the unlikely
	event that we had a hebrew font with only GSUB it would not work either.

	So if we want our lookups to have a chance of executing under Uniscribe we
	better make sure that both tables have the same script set.

	(Sergey says we could optimize a little: A Latin GSUB table will run without
	a GPOS, but he says the GPOS won't work without a GSUB.)
	*/

	// So here always give scripts for both (see comment above) no matter what they asked for
	if (type == keGSUB || type == keBoth)
	{
		for (LookupSet::iterator it = m_GSUBLookups.begin(); it != m_GSUBLookups.end(); ++it)
			ProcessLookupFindScripts(*it, scripts);
	}
	if (type == keGPOS || type == keBoth)
	{
		for (LookupSet::iterator it = m_GPOSLookups.begin(); it != m_GPOSLookups.end(); ++it)
			ProcessLookupFindScripts(*it, scripts);
	}

	return scripts.size() ? true : false;
}

/******************************************************************************
   Function:	SFLangsInScript
   Author:		Paul Gruzdev
   Purpose:		Search GPOS or GSUB lookup tables and find all languages that
				correspond to a given script
   Returns:		bool
   Parameters:	ETableType type
   Parameters:	UINT32 script
   Parameters:	OTLangSet & langs
   Notes:       
******************************************************************************/
bool SFNTWriter::SFLangsInScript(ETableType type, UINT32 script, OTLangSet& langs)
{
	// However, the language lists (I think) are distinct
	// But passing a value of keBoth for gpos will give us the set of languages in
	//  both tables (for this script)
	if (type == keGPOS || type == keBoth)
	{
		for (LookupSet::iterator it = m_GPOSLookups.begin(); it != m_GPOSLookups.end(); ++it)
			ProcessLookupFindLangs(*it, script, langs);
	}
	if (type == keGSUB || type == keBoth)
	{
		for (LookupSet::iterator it = m_GSUBLookups.begin(); it != m_GSUBLookups.end(); ++it)
			ProcessLookupFindLangs(*it, script, langs);
	}

	if (langs.size() == 0)
	{
		// We add dummy script entries. Because Uniscribe will refuse to 
		//  process some scripts if they don't have an entry in both GPOS 
		//  an GSUB. So if a script appears in either table, force it to 
		//  appear in both. That means we can get scripts with no lookups 
		//  and hence no languages. It seems that Uniscribe doesn't like 
		//  that either. So give each such script a dummy default language 
		//  entry. This is what VOLT does 
		langs.insert(DEFAULT_LANG);
	}

	return true;
}

/******************************************************************************
   Function:	SFFeaturesInScriptLang
   Author:		Paul Gruzdev
   Purpose:		Collect and return a set of features for a given script and lang
				(and a GPOS and/or GSUB lookups).
   Returns:		bool
   Parameters:	ETableType type
   Parameters:	UINT32 script
   Parameters:	UINT32 lang
   Parameters:	std::set<UINT32> & features
   Notes:       
******************************************************************************/
bool SFNTWriter::SFFeaturesInScriptLang(ETableType type, UINT32 script, UINT32 lang, std::set<UINT32>& features)
{
	// gpos==0 => GSUB, gpos==1 => GPOS, gpos==-1 => both, gpos==-2 => Both & morx & kern

	if (type == keGSUB || type == keBoth)
	{
		for (LookupSet::iterator it = m_GSUBLookups.begin(); it != m_GSUBLookups.end(); ++it)
			ProcessLookupFindFeatures(type, *it, script, lang, features);
	}
	if (type == keGPOS || type == keBoth)
	{
		for (LookupSet::iterator it = m_GPOSLookups.begin(); it != m_GPOSLookups.end(); ++it)
			ProcessLookupFindFeatures(type, *it, script, lang, features);
	}
/*
	if (bAddDesignSize && type == keGPOS)
	{
		// The 'size' feature is like no other. It has no lookups and so 
		//  we will never find it in the normal course of events. If the 
		//  user has specified a design size, then every script/lang combo 
		//  gets a 'size' feature which contains no lookups but feature 
		//  params 
		features.insert(CHR('s', 'i', 'z', 'e'));
	}
*/
	return true;
}

/******************************************************************************
   Function:	SFLookupsInScriptLangFeature
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	ETableType type
   Parameters:	UINT32 script
   Parameters:	UINT32 lang
   Parameters:	UINT32 feature
   Parameters:	std::vector<OTLookup * > & lookups
   Notes:       
******************************************************************************/
bool SFNTWriter::SFLookupsInScriptLangFeature(ETableType type, UINT32 script, UINT32 lang, UINT32 feature, std::vector<OTLookup*>& lookups)
{
	if (type == keGPOS || type == keBoth)
	{
		for (LookupSet::iterator it = m_GPOSLookups.begin(); it != m_GPOSLookups.end(); ++it)
			ProcessLookupFindFeatureLookups(*it, script, lang, feature, lookups);
	}
	else if (type == keGSUB || type == keBoth)
	{
		for (LookupSet::iterator it = m_GSUBLookups.begin(); it != m_GSUBLookups.end(); ++it)
			ProcessLookupFindFeatureLookups(*it, script, lang, feature, lookups);
	}

	if (lookups.size() == 0)
		return false;

	// lookup order is irrelevant here. might as well leave it in invocation order
	return true;
}

/******************************************************************************
   Function:	FindOrMakeNewFeatureLookup
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		size_t
   Parameters:	GInfo * ginfo
   Parameters:	std::vector<OTLookup * > & lookups
   Parameters:	UINT32 tag
   Notes:       
******************************************************************************/
static size_t FindOrMakeNewFeatureLookup(GInfo* ginfo, std::vector<OTLookup*>& lookups, UINT32 tag) 
{
	size_t i = 0;
	for (; i < ginfo->vFeatLookups.size(); ++i) 
	{
		if (ginfo->vFeatLookups[i].tag != tag)
			continue;

		if (lookups.size() == 0 && ginfo->vFeatLookups[i].vLookups.size() == 0)	// 'size' feature
			return i;

		if (lookups.size() == 0 || ginfo->vFeatLookups[i].vLookups.size() == 0)
			continue;

		size_t j = 0;
		for (; j < lookups.size(); ++j)
		{
			if (j >= ginfo->vFeatLookups[i].vLookups.size())
				break;

			if (ginfo->vFeatLookups[i].vLookups[j] != lookups[j])
				break;
		}
		if (ginfo->vFeatLookups[i].vLookups[j] == lookups[j]) 
		{
			lookups.clear();
			return i;
		}
	}

	ginfo->vFeatLookups.push_back(FeatureLookups());
	ginfo->vFeatLookups.back().iFeatureID = (int)i;
	ginfo->vFeatLookups.back().tag = tag;
	ginfo->vFeatLookups.back().vLookups = lookups;

	return i;
}

static int LangSysMatch(ScriptSet* s, size_t ils1, size_t ils2) 
{
	LangSys* ls1 = &s->vLangSys[ils1];
	LangSys* ls2 = &s->vLangSys[ils2];

	if (ls1->vFeatureID.size() != ls2->vFeatureID.size())
		return false;

	for (size_t i = 0; i < ls1->vFeatureID.size(); ++i)
		if (ls1->vFeatureID[i] != ls2->vFeatureID[i])
			return false;

	return true;
}

bool feat_alphabetize(FeatureLookups fl1, FeatureLookups fl2)
{
	return fl1.tag < fl2.tag ? true : false;
}

/******************************************************************************
   Function:	FindFeatures
   Author:		Paul Gruzdev
   Purpose:		Given the table type (GPOS or GSUB), find all the feature lookups
				and all the scripts we need to write
   Returns:		void
   Parameters:	ETableType type
   Parameters:	GInfo * ginfo
   Notes:       
******************************************************************************/
void SFNTWriter::FindFeatures(ETableType type, GInfo* ginfo)
{
	std::set<UINT32> scripts;

	SFScriptsInLookups(keBoth, scripts);

	if (scripts.size() == 0)	// All lookups unused
		return;

	for (std::set<UINT32>::iterator its = scripts.begin(); its != scripts.end(); ++its)
	{
		OTLangSet langs;
		SFLangsInScript(type, *its, langs);

		ginfo->vScripts.push_back(ScriptSet());
		ginfo->vScripts.back().script = *its;

		for (OTLangSet::iterator itl = langs.begin(); itl != langs.end(); ++itl)
		{
			std::set<UINT32> features;
			SFFeaturesInScriptLang(type, *its, *itl, features);
			
			ginfo->vScripts.back().vLangSys.push_back(LangSys());
			ginfo->vScripts.back().vLangSys.back().lang = *itl;
			ginfo->vScripts.back().vLangSys.back().iSameAs = -1;

			for (std::set<UINT32>::iterator itf = features.begin(); itf != features.end(); ++itf)
			{
				std::vector<OTLookup*> lookups;

				SFLookupsInScriptLangFeature(type, *its, *itl, *itf, lookups);
				ginfo->vScripts.back().vLangSys.back().vFeatureID.push_back((int)FindOrMakeNewFeatureLookup(ginfo, lookups, *itf));
				// lookups is freed or used by FindOrMakeNewFeatureLookup
			}
		}
	}

	std::sort(ginfo->vFeatLookups.begin(), ginfo->vFeatLookups.end(), feat_alphabetize);

	// Now we've disordered the features. Find each feature_id and turn it back
	//  into a feature number
	for (size_t sc = 0; sc < ginfo->vScripts.size(); ++sc)
	{
		for (size_t lc = 0; lc < ginfo->vScripts[sc].vLangSys.size(); ++lc)
		{
			size_t fcmax = ginfo->vScripts[sc].vLangSys[lc].vFeatureID.size();
			std::vector<int>& feature_id = ginfo->vScripts[sc].vLangSys[lc].vFeatureID;

			for (size_t fc = 0; fc < fcmax; ++fc)
			{
				int id = feature_id[fc];
				size_t j = 0;

				for (; j < ginfo->vFeatLookups.size(); ++j)
					if (id == ginfo->vFeatLookups[j].iFeatureID)
						break;

				feature_id[fc] = (int)j;
			}
			std::sort(ginfo->vScripts[sc].vLangSys[lc].vFeatureID.begin(), 
				ginfo->vScripts[sc].vLangSys[lc].vFeatureID.end(), std::less<int>());
		}
		// See if there are langsys tables which use exactly the same features
		//  They can use the same entry in the file. This optimization seems  
		//  to be required for Japanese vertical writing to work in Uniscribe.
		for (size_t lc = 0; lc < ginfo->vScripts[sc].vLangSys.size(); ++lc)
		{
			for (size_t j = 0; j < lc; ++j)
			{
				if (LangSysMatch(&ginfo->vScripts[sc], j, lc))
				{
					ginfo->vScripts[sc].vLangSys[lc].iSameAs = (int)j;
					break;
				}
			}
		}
	}
}

/******************************************************************************
   Function:	GlyphsInSubtable
   Author:		Paul Gruzdev
   Purpose:		Given all kerning pairs we have, find all glyph IDs (sorted), with
				kernings defined that represent the sub-table pSub
   Returns:		void
   Parameters:	std::vector<WCHAR> & glyphs
   Parameters:	LookupSubTable * pSub
   Notes:       
******************************************************************************/
void SFNTWriter::GlyphsInSubtable(std::vector<int>& glyphs, LookupSubTable* pSub)
{
	switch (pSub->pOTLookup->eLookupType)
	{
	case OTL_TYPE_GPOS_PAIR:
		for (KernPairMap::iterator it = m_KPairs.begin(); it != m_KPairs.end(); ++it)
		{
			for (std::set<CKernData>::iterator its = it->second.begin(); its != it->second.end(); ++its)
			{
				if (its->pSubTable == pSub)
				{
					glyphs.push_back(it->first);
					break;
				}
			}
		}
		break;
	case OTL_TYPE_GSUB_LIGATURE:
		for (LigatureMap::iterator it = m_Ligatures.begin(); it != m_Ligatures.end(); ++it)
		{
			for (std::set<CLigatureData>::iterator its = it->second.begin(); its != it->second.end(); ++its)
			{
				if (its->pSubTable == pSub)
				{
					glyphs.push_back(it->first);
					break;
				}
			}
		}
		break;
	case OTL_TYPE_GSUB_SINGLE:
			for (SubstituteMap::iterator it = m_Substitutes.begin(); it != m_Substitutes.end(); ++it)
			{
				for (std::set<CSubstituteData>::iterator its = it->second.begin(); its != it->second.end(); ++its)
				{
					if (its->pSubTable == pSub)
					{
						glyphs.push_back(it->first);
						break;
					}
				}
			}
		break;
	default:
		ASSERT(0); // Unknown sub-table - we have no data for that
		break;
	}
}

/******************************************************************************
   Function:	SubTableFindOrMake
   Author:		Paul Gruzdev
   Purpose:		Create or find a subtable of type lookup_type, given the script
   Returns:		LookupSubTable*
   Parameters:	UINT32 tag
   Parameters:	UINT32 script
   Parameters:	int lookup_type
   Notes:       
******************************************************************************/
LookupSubTable* SFNTWriter::SubTableFindOrMake(UINT32 tag, UINT32 script, int lookup_type)
{
	bool bIsGPos = lookup_type >= OTL_TYPE_GPOS_START && lookup_type <= OTL_TYPE_KERN_STATE_MACHINE;
	LookupSet& lookup = bIsGPos ? m_GPOSLookups : m_GSUBLookups;
	OTLookup* found = NULL;
	LookupSubTable* sub;
	static INT32 s_lookupIndex = 0;

	for (LookupSet::iterator it = lookup.begin(); it != lookup.end(); ++it)
	{
		if ((*it)->eLookupType == lookup_type && FeatureScriptTagInFeatureScriptList(tag, script, (*it)->vFeatures))
		{
			for (std::list<LookupSubTable*>::iterator it1 = (*it)->vSubTables.begin(); it1 != (*it)->vSubTables.end(); it1++)
			{
				sub = *it1;
				if (sub->pKernClass == NULL)
					return sub;
			}
			found = *it;
		}
	}

	if (found == NULL)
	{
		found = new OTLookup;
		found->iLookupIndex = s_lookupIndex++;
		found->eLookupType = (OTLookupType)lookup_type;
		found->vFeatures.push_back(new FeatureScriptLangList);
		found->vFeatures.back()->uFeatureTag = tag;
		found->vFeatures.back()->vScripts.push_back(new ScriptLangList);
		found->vFeatures.back()->vScripts.back()->uScript = script;
		found->vFeatures.back()->vScripts.back()->uLangs[0] = DEFAULT_LANG;
		found->vFeatures.back()->vScripts.back()->iLangCnt = 1;

		lookup.insert(found);
	}

	sub = new LookupSubTable;
	sub->pNext = found->vSubTables.empty() ? NULL : found->vSubTables.front();
	found->vSubTables.insert(found->vSubTables.begin(), sub);
	sub->pOTLookup = found;
	sub->bPerGlyphPstOrKern = TRUE;

	return sub;
}

LookupSubTable* SFNTWriter::MakeSubTable(UINT32 tag, UINT32 script, int lookup_type)
{
	bool bIsGPos = lookup_type >= OTL_TYPE_GPOS_START && lookup_type <= OTL_TYPE_KERN_STATE_MACHINE;
	LookupSet& lookup = bIsGPos ? m_GPOSLookups : m_GSUBLookups;
	OTLookup* found = NULL;
	LookupSubTable* sub;
	static INT32 s_lookupIndex = 0;

	for (LookupSet::iterator it = lookup.begin(); it != lookup.end(); ++it)
	{
		if ((*it)->eLookupType == lookup_type && FeatureScriptTagInFeatureScriptList(tag, script, (*it)->vFeatures))
		{
			found = *it;
			break;
		}
	}

	if (found == NULL)
	{
		found = new OTLookup;
		found->iLookupIndex = s_lookupIndex++;
		found->eLookupType = (OTLookupType)lookup_type;
		found->vFeatures.push_back(new FeatureScriptLangList);
		found->vFeatures.back()->uFeatureTag = tag;
		found->vFeatures.back()->vScripts.push_back(new ScriptLangList);
		found->vFeatures.back()->vScripts.back()->uScript = script;
		found->vFeatures.back()->vScripts.back()->uLangs[0] = DEFAULT_LANG;
		found->vFeatures.back()->vScripts.back()->iLangCnt = 1;

		lookup.insert(found);
	}

	sub = new LookupSubTable;
	sub->pNext = found->vSubTables.empty() ? NULL : found->vSubTables.front();
	found->vSubTables.insert(found->vSubTables.begin(), sub);
	sub->pOTLookup = found;
	sub->bPerGlyphPstOrKern = TRUE;

	return sub;
}

/******************************************************************************
   Function:	AddKerningPair
   Author:		Paul Gruzdev
   Purpose:		Pass a kerning pair info and create a corresponding GPOS
				lookup table that matches the script and a lang of the passed
				pair
   Returns:		void
   Parameters:	WORD wFirst  - Unicode of the first glyph
   Parameters:	WORD wSecond - Unicode of the second glyph
   Parameters:	MILLIPOINT iKernAmount - kerning amount
   Notes:       
******************************************************************************/
void SFNTWriter::AddKerningPair(WORD wFirst, WORD wSecond, MILLIPOINT iKernAmount)
{
	int iFirst = m_CodeToGID[wFirst];	// GPOS/GSUB store solely glyph indices
	int iSecond = m_CodeToGID[wSecond];

	UINT32 script = ScriptFromUnicode(wFirst);
	if (script == DEFAULT_SCRIPT)
		script = ScriptFromUnicode(wSecond);

	LookupSubTable* pSubTable = SubTableFindOrMake(CHR('k', 'e', 'r', 'n'), script, OTL_TYPE_GPOS_PAIR);
	if (m_KPairs.find(iFirst) == m_KPairs.end())
		m_KPairs.insert(std::make_pair(iFirst, std::set<CKernData>()));
	// Now insert our kerning pair
	std::set<CKernData>::_Pairib rib = m_KPairs[iFirst].insert(CKernData(iSecond, iKernAmount, pSubTable));
	if(rib.second == true)
		m_ulPairCount++;
}

bool CLigatureData::operator <(const CLigatureData& rhs) const
{
	ASSERT(to.size() && rhs.to.size());

	if(to.size() != rhs.to.size())
		return to.size() > rhs.to.size();
	for(size_t i = 0; i < to.size(); i++)
	{
		if(to[i] != rhs.to[i])
			return to[i] < rhs.to[i];
	}

	if(pSubTable != rhs.pSubTable)
		return pSubTable < rhs.pSubTable;

	return false;
}

/******************************************************************************
   Function:	AddLigature
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	std::wstring key
   Parameters:	WCHAR wLig
   Parameters:	DWORD dwLigType
   Notes:       
******************************************************************************/
LookupSubTable* SFNTWriter::AddLigature(std::wstring key, WCHAR wLig, DWORD dwLigType)
{
	if (key.empty()) return false; // invalid parameter passed: ligature key can't be empty
	DWORD dwTableSubtype = 0;

	switch (dwLigType)
	{
	case 0x02: // LIG_SUB_STANDARD
		dwTableSubtype = CHR('l', 'i', 'g', 'a');
		break;
	case 0x04: // LIG_SUB_HISTORIC
		dwTableSubtype = CHR('h', 'l', 'i', 'g');
		break;
	case 0x08: // LIG_SUB_DISCRETIONARY
		dwTableSubtype = CHR('d', 'l', 'i', 'g');
		break;
	case 0x10: // LIG_SUB_CONTEXTUAL
		dwTableSubtype = CHR('c', 'l', 'i', 'g');
		break;
	default:
		dwTableSubtype = dwLigType;
		break;
	}
	TRACE(L"Addig liga: %X: '%s' -> %X\n", dwLigType, key.c_str(), wLig);
	// ASSERT(dwTableSubtype != 0);

	ERROR2IF(m_CodeToGID.find(key[0]) == m_CodeToGID.end(), NULL, "Can't add a ligature whos first glyph is absent");
	int iFirst = m_CodeToGID[key[0]];	// GPOS/GSUB store solely glyph indices
	ERROR2IF(m_CodeToGID.find(wLig) == m_CodeToGID.end(), NULL, "Ligature glyph is absent");
	int iLigGID = m_CodeToGID[wLig];

	UINT32 script = ScriptFromUnicode(key);
	LookupSubTable* pSubTable = SubTableFindOrMake(dwTableSubtype, script, OTL_TYPE_GSUB_LIGATURE);

	if (m_Ligatures.find(iFirst) == m_Ligatures.end())
		m_Ligatures.insert(std::make_pair(iFirst, std::set<CLigatureData>()));

	//if(std::find(m_Ligatures[iFirst].begin(), m_Ligatures[iFirst].end(), CLigatureData(iLigGID, pSubTable)) == m_Ligatures[iFirst].end())
	{
		CLigatureData lig(iLigGID, pSubTable);
		// Now save all TO glyph indices
		for (size_t i = 1; i < key.size(); i++)
			lig.AddTo(m_CodeToGID[key[i]]);

		// Now push back our ligature TO data
		m_Ligatures[iFirst].insert(lig);
	}

	return pSubTable;
}

LookupSubTable* SFNTWriter::AddSubstitute(WCHAR c1, WCHAR c2, DWORD dwTableSubtype)
{
	ERROR2IF(m_CodeToGID.find(c1) == m_CodeToGID.end(), NULL, (m_sPostScriptName + ": can't add a substitute whos first glyph absent").c_str());
	//ERROR2IF(m_CodeToGID.find(c2) == m_CodeToGID.end(), NULL, (m_sPostScriptName + ": can't add a substitute whos second glyph absent").c_str());
	if (m_CodeToGID.find(c2) == m_CodeToGID.end()) return NULL;
	TRACE(L"Addig single subst %X: '%c' -> %X\n", dwTableSubtype, c1, c2);

	int iFirst = m_CodeToGID[c1];	// GPOS/GSUB store solely glyph indices
	int iSecond = m_CodeToGID[c2];

	UINT32 script = ScriptFromUnicode(c1);
	if (script == DEFAULT_SCRIPT)
		script = ScriptFromUnicode(c2);

	LookupSubTable* pSubTable = SubTableFindOrMake(dwTableSubtype, script, OTL_TYPE_GSUB_SINGLE);

	if (m_Substitutes.find(iFirst) == m_Substitutes.end())
		m_Substitutes.insert(std::make_pair(iFirst, std::set<CSubstituteData>()));

	m_Substitutes[iFirst].insert(CSubstituteData(iFirst, iSecond, pSubTable));

	return pSubTable;
}

#if defined(_DEBUG) && !defined(CAMELOT_PRODUCT)
std::wstring DumpSubstRule(CSubstRule& rule)
{
	USES_CONVERSION;
	std::wstring pre, post, keys;
	std::wstring langs;

	if (rule.pre.empty()) pre = L"<empty>";
	for (size_t i = 0; i < rule.pre.size(); i++)
	{
		std::wstring s(rule.pre[i].begin(), rule.pre[i].end());
		if (!pre.empty()) pre += L",";
		pre += L"'" + s + L"'";
	}

	if (rule.post.empty()) post = L"<empty>";
	for (size_t i = 0; i < rule.post.size(); i++)
	{
		std::wstring s(rule.post[i].begin(), rule.post[i].end());
		if (!post.empty()) post += L",";
		post += L"'" + s + L"'";
	}

	if (rule.keys.empty()) keys = L"<empty>";
	for (size_t i = 0; i < rule.keys.size(); i++)
	{
		std::wstring s(rule.keys[i].begin(), rule.keys[i].end());
		if (!keys.empty()) keys += L",";
		keys += L"'" + s + L"'";
	}

	for (size_t i = 0; i < rule.langs.size(); i++)
		langs += FormatStringW(L"%d ", rule.langs[i]);

	return FormatStringW(L"ref idx: %d pre: %s post: %s keys: %s langs: %s", rule.pRef ? rule.pRef->iLookupIndex : 65535, pre.c_str(), post.c_str(), keys.c_str(), langs.c_str());
}
#endif

bool SFNTWriter::AddSubstRule(DWORD dwTableSubtype, std::vector<std::wstring>& pre, std::vector<std::wstring>& post, std::vector<std::wstring>& keys, OTLookup* lookup, UINT16 priority, std::vector<UINT32>& langs)
{
	std::wstring& ref = pre.size() ? pre[0] : (post.size() ? post[0] : keys[0]);
	LookupSubTable* pSubTable = MakeSubTable(dwTableSubtype, ScriptFromUnicode(ref), OTL_TYPE_GSUB_CONTEXT_CHAIN);
	m_Rules[pSubTable] = CSubstRule(pre, post, keys, lookup, priority, langs);

#if defined(_DEBUG) && !defined(CAMELOT_PRODUCT)
	OutputDebugStringW(L"Addig rule:");
	OutputDebugStringW(DumpSubstRule(m_Rules[pSubTable]).c_str());
	OutputDebugStringW(L"\n");
#endif

	return true;
}

LookupSubTable* SFNTWriter::AddStylisticSetSubstitute(WCHAR c1, WCHAR c2, DWORD dwSetIdx)
{
	ERROR2IF(m_CodeToGID.find(c1) != m_CodeToGID.end(), false, "Can't add a substitute whos first glyph absent");
	ERROR2IF(m_CodeToGID.find(c2) != m_CodeToGID.end(), false, "Can't add a substitute whos second glyph absent");

	std::stringstream ss;
	ss << "ss";
	ss.width(2);
	ss << (dwSetIdx + 1);
	std::string str = ss.str();

	ERROR2IF(str.length() != 4, false, "Wrong table length");

	DWORD dwTableSubtype = CHR(str[0], str[1], str[2], str[3]);

	int iFirst  = m_CodeToGID[c1];	// GPOS/GSUB store solely glyph indices
	int iSecond = m_CodeToGID[c2];

	UINT32 script = ScriptFromUnicode(c1);
	if (script == DEFAULT_SCRIPT)
		script = ScriptFromUnicode(c2);

	LookupSubTable* pSubTable = SubTableFindOrMake(dwTableSubtype, script, OTL_TYPE_GSUB_SINGLE);

	if (m_Substitutes.find(iFirst) == m_Substitutes.end())
		m_Substitutes.insert(std::make_pair(iFirst, std::set<CSubstituteData>()));

	m_Substitutes[iFirst].insert(CSubstituteData(iFirst, iSecond, pSubTable));

	return pSubTable;
}

/******************************************************************************
   Function:	GatherKerningData
   Author:		Paul Gruzdev
   Purpose:		Given our SFNTWriter, take all the kerning pairs of our font variant
				and place them into the m_KPairs internal structure suitable to
				be written as GPOS/GSUB table later.
   Returns:		bool
   Parameters:
   Notes:       
******************************************************************************/
bool SFNTWriter::GatherKerningData()
{
	WORD wFirst, wSecond; 
	MILLIPOINT  iKernAmount; 

	for (bool bRes = m_pSource->GetFirstKerningPair(wFirst, wSecond, iKernAmount); bRes == true;
		bRes = m_pSource->GetNextKerningPair(wFirst, wSecond, iKernAmount))
	{
		AddKerningPair(wFirst, wSecond, iKernAmount);
	}

	return m_KPairs.size() > 0 ? true : false;
}

bool SFNTWriter::GatherLigatureData()
{
	std::wstring strKey;
	wchar_t wLigature;
	DWORD dwLigType;

	for (bool bRes = m_pSource->GetFirstLigature(strKey, wLigature, dwLigType); bRes == true;
		bRes = m_pSource->GetNextLigature(strKey, wLigature, dwLigType))
		LookupSubTable* pSubTable = AddLigature(strKey, wLigature, dwLigType);

	return m_Ligatures.size() > 0 ? true : false;
}

///////////////////////////////////////////////////////////////////////////////////////
// Actual writing routines
/******************************************************************************
   Function:	DumpCoverageTable
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		void
   Parameters:	CCFile * gpos
   Parameters:	std::vector<UINT16> & glyphs
   Notes:       
******************************************************************************/
static void DumpCoverageTable(CCFile* gpos, std::vector<int>& glyphs)
{
	int last = -2, range_cnt = 0, start, r;
	// the glyph list should already be sorted 
	// figure out whether it is better (smaller) to use an array of glyph ids 
	//  or a set of glyph id ranges 
	for (size_t i = 0; i < glyphs.size(); ++i)
	{
		if (glyphs[i] <= last)
			ERROR2RAW("Glyphs must be ordered when creating coverage table");
		if (glyphs[i] != last + 1)
			++range_cnt;
		last = glyphs[i];
	}

	// I think Windows will only accept format 2 coverage tables?
	if ((int)glyphs.size() <= 3 * range_cnt)
	{
		// We use less space with a list of glyphs than with a set of ranges
		putshort(gpos, 1);						// Coverage format=1 => glyph list
		putshort(gpos, (short)glyphs.size());	// count of glyphs
		for (size_t i = 0; i < glyphs.size(); ++i)
			putshort(gpos, (short)glyphs[i]);			// array of glyph IDs
	}
	else
	{
		putshort(gpos, 2);		// Coverage format=2 => range list
		putshort(gpos, (short)range_cnt);	// count of ranges
		last = -2; start = -2;		// start is a index in our glyph array, last is ttf_glyph
		r = 0;
		for (size_t i = 0; i < glyphs.size(); ++i)
		{
			if (glyphs[i] != last + 1)
			{
				if (last != -2)
				{
					putshort(gpos, (short)glyphs[start]);	// start glyph ID
					putshort(gpos, (short)last);			// end glyph ID
					putshort(gpos, (short)start);			// coverage index of start glyph
					++r;
				}
				start = (int)i;
			}
			last = glyphs[i];
		}

		if (last != -2) 
		{
			putshort(gpos, (short)glyphs[start]);	// start glyph ID
			putshort(gpos, (short)last);	// end glyph ID
			putshort(gpos, (short)start);	// coverage index of start glyph
			++r;
		}
		if (r != range_cnt)
			ERROR2RAW("Miscounted ranges in format 2 coverage table output");
	}
}

/******************************************************************************
   Function:	CreateSameAsTable
   Author:		Paul Gruzdev
   Purpose:		Test m_KPairs table and find the exact similar 'others', then
				fill in the sameAs array so that, only backward rule applies:
				i.e. glyph 5 is same as 2, 10 is sames as 6 etc.
   Returns:		void
   Parameters:	std::vector<int> & glyphs - glyphs to test [GIDs]
   Parameters:	std::vector<size_t> & sameAs - same as result
   Notes:       
******************************************************************************/
void SFNTWriter::CreateSameAsTable(std::vector<int>& glyphs, std::vector<int>& sameAs)
{
	sameAs.assign(glyphs.size(), -1);

	for (size_t i = 1; i < glyphs.size(); i++)
	{
		for (int j = (int)i - 1; j >= 0; j--)
		{
			// Compare two sets of CKernData
			if (m_KPairs[glyphs[i]] == m_KPairs[glyphs[j]])
			{
				sameAs[i] = j;
				break;
			}
		}
	}
}

/******************************************************************************
   Function:	DumpGPOSPairPos
   Author:		Paul Gruzdev
   Purpose:		Write out 'Pair Positioning Adjustment: Format 1'
				https://www.microsoft.com/typography/otspec/gpos.htm
   Returns:		void
   Parameters:	CCFile * gpos
   Parameters:	LookupSubTable * pSub
   Notes:       
******************************************************************************/
void SFNTWriter::DumpGPOSPairPos(CCFile* gpos, LookupSubTable* pSub)
{
	std::vector<int> glyphs, offsets, sameAs;
	GlyphsInSubtable(glyphs, pSub);
	CreateSameAsTable(glyphs, sameAs);

	int offset = 5 * 2 + (int)glyphs.size() * 2;
	for (size_t i = 0; i < glyphs.size(); i++)
	{
		if (sameAs[i] != -1)
			offsets.push_back(offsets[sameAs[i]]);
		else
		{
			offsets.push_back(offset);
			offset += (int)m_KPairs[glyphs[i]].size() * 4 + 2; // count(short)+[to(short), amount(short)]
		}
	}

	putshort(gpos, 1);				// 1 means char pairs (ie. not classes)
	putshort(gpos, (INT16)offset);	// offset to coverage table
	putshort(gpos, 4);				// ValueFormat1: XAdvance: includes horizontal adjustment for advance
	putshort(gpos, 0);				// ValueFormat2 is always 0, we only save horizontal adjustment for now
	putshort(gpos, (INT16)glyphs.size());

	for (size_t i = 0; i < glyphs.size(); i++)
		putshort(gpos, (short)offsets[i]);

	for (size_t i = 0; i < glyphs.size(); i++)
	{
		if (sameAs[i] != -1)	// this glyph's pairs are exactly the same as some other glyph's pairs: we already stored the offset
			continue;				// to them so no further action

		std::set<CKernData>& kern = m_KPairs[glyphs[i]];
		putshort(gpos, (INT16)kern.size());

		for (std::set<CKernData>::iterator it = kern.begin(); it != kern.end(); ++it)
		{
			putshort(gpos, (INT16)it->iGID);
			putshort(gpos, (INT16)xar2ttf(it->wAmount));
		}
	}

	DumpCoverageTable(gpos, glyphs);
}

/******************************************************************************
   Function:	OTFDumpALookup
   Author:		Paul Gruzdev
   Purpose:		Write out a given lookup table to a file provided
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	OTLookup * otl
   Notes:       
******************************************************************************/
void SFNTWriter::OTFDumpALookup(CCFile* pFile, OTLookup* otl)
{
	otl->uLookupOffset = pFile->tell();
	for (std::list<LookupSubTable*>::iterator it = otl->vSubTables.begin(); it != otl->vSubTables.end(); ++it)
	{
		LookupSubTable* pSub = *it;
		if (pSub->uUnused)
			pSub->iSubtableOffset = -1;
		else 
		{
			// This is the offset relative to the start of THIS file, which is a one lookup file, later we need
			// to update this to be relative to the start of ALL lookups
			pSub->iSubtableOffset = pFile->tell();
			switch (otl->eLookupType)
			{
				// GPOS lookup types
			case OTL_TYPE_GPOS_SINGLE:
				// dumpGPOSsimplepos(lfile, sf, sub);
				break;

			case OTL_TYPE_GPOS_PAIR:
				// dumpgposkernclass(lfile, sf, sub, at);
				DumpGPOSPairPos(pFile, pSub);
				break;

			case OTL_TYPE_GPOS_CURSIVE:
				// dumpgposCursiveAttach(lfile, sf, sub, &at->gi);
				break;

			case OTL_TYPE_GPOS_MARK_2_BASE:
			case OTL_TYPE_GPOS_MARK_2_LIGATURE:
			case OTL_TYPE_GPOS_MARK_2_MARK:
				// AnchorsAway(lfile, sf, sub, &at->gi);
				break;

			case OTL_TYPE_GPOS_CONTEXT_CHAIN:
			case OTL_TYPE_GPOS_CONTEXT:
				// dumpg___ContextChain(lfile, sf, sub, at);
				break;

				// GSUB lookup types
			case OTL_TYPE_GSUB_SINGLE:
				// dumpGSUBsimplesubs(lfile, sf, sub);
				DumpGSUBSimpleSubs(pFile, pSub);
				break;

			case OTL_TYPE_GSUB_MULTIPLE:
			case OTL_TYPE_GSUB_ALTERNATE:
				// dumpGSUBmultiplesubs(lfile, sf, sub);
				break;

			case OTL_TYPE_GSUB_LIGATURE:
				DumpGSUBLigData(pFile, pSub);
				break;

			case OTL_TYPE_GSUB_CONTEXT_CHAIN:
				DumpGContextChainCoverage(pFile, pSub);
				break;
			case OTL_TYPE_GSUB_CONTEXT:
			case OTL_TYPE_GSUB_REVERSE_C_CHAIN:
				break;
			}

			if (pFile->tell() - pSub->iSubtableOffset == 0)
			{
				ERROR3_PF(("Lookup sub table, %s in %s, contains no data.\n", pSub->sSubtableName.c_str(), pSub->pOTLookup->sLookupName.c_str()));
				pSub->uUnused = TRUE;
				pSub->iSubtableOffset = -1;
			}
			else if (pSub->vExtraSubTables.empty() && pFile->tell() - pSub->iSubtableOffset > 65535)
			{
				ERROR3_PF(("Lookup sub table, %s in %s, is too big. Will not be useable.\n", pSub->sSubtableName.c_str(), pSub->pOTLookup->sLookupName.c_str()));
			}
		}
	}

	otl->uLookupLength = pFile->tell() - otl->uLookupOffset;
}

/******************************************************************************
   Function:	WriteUpdateLookup
   Author:		Paul Gruzdev
   Purpose:		Write a particular lookup in a GPOS/GSUB file, having it's already
				written to it's own mem file, update offsets etc
   Returns:		void
   Parameters:	OTLookup * otl
   Parameters:	CCFile * pFile
   Parameters:	CCMemFile * pLookupFile
   Parameters:	INT32 & offset
   Notes:       
******************************************************************************/
static void WriteUpdateLookup(OTLookup* otl, CCFile* pFile, CCMemFile* pLookupFile, INT32& offset)
{
	otl->uLookupOffset = pFile->tell();
	LONG lSize = pLookupFile->Size();
	pFile->write(pLookupFile->GetMemoryBuffer(), lSize);

	for (std::list<LookupSubTable*>::iterator its = otl->vSubTables.begin(); its != otl->vSubTables.end(); ++its)
	{
		LookupSubTable* pSub = *its;
		if (!pSub->uUnused)
			// Since we stored each subtable offset relative to the start of the lookup, not the start of the lookups,
			// we need to update in here
			pSub->iSubtableOffset += offset;
	}
	offset += lSize;
}

/******************************************************************************
   Function:	WriteLookups
   Author:		Paul Gruzdev
   Purpose:		Write out all lookups to a given file
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	bool bGPos
   Notes:       Order them so smaller come first, if the total size is greater than
				64K
******************************************************************************/
void SFNTWriter::WriteLookups(CCFile*pFile, bool bGPos)
{
	LookupSet& all = bGPos ? m_GPOSLookups : m_GSUBLookups;
	OTLookupFileMap map;
	std::list<std::pair<OTLookup*, CCFile*>> lst;

	int index = 0;
	for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
	{
		OTLookup* otl = *it;
		if (otl->uUnused /*|| OnlyMac(otl, all)*/)
			otl->iLookupIndex = -1;
		else
			otl->iLookupIndex = index++;
	}

	size_t szTotal = 0;
	for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
	{
		OTLookup* otl = *it;
		if (otl->iLookupIndex != -1)
		{
			CCFile* pMemFile = (CCFile*)new CCMemFile(NULL, 0, CCMemWrite);
			OTFDumpALookup(pMemFile, otl);
#if defined(_DEBUG) && defined(DUMP_EXTRA_INFO)
			char psz[128];
			sprintf_s(psz, 128, "d:/temp/cff/%s%d%s_lookup_%s%d.tst", 
				m_sFullName.c_str(), m_lWeight, m_bItalic ? "i" : "",
				bGPos ? "gpos" : "gsub", 
				otl->iLookupIndex);
			FILE* out = 0; fopen_s(&out, psz, "wb");
			if (out)
			{
				fwrite(((CCMemFile*)pMemFile)->GetMemoryBuffer() + otl->uLookupOffset, otl->uLookupLength, 1, out);
				fclose(out);
			}
#endif

			map.insert(std::make_pair(std::make_pair(otl, pMemFile->tell()), pMemFile));
			lst.push_back(std::make_pair(otl, pMemFile));
			szTotal += pMemFile->Size();
		}
	}

	// Order the lookups so that the smallest ones come first 
	//  thus we are less likely to need extension tables 
	// I think it's better to order the entire lookup rather than ordering the
	//  subtables -- since the extension subtable would be required for all 
	//  subtables in the lookup, so might as well keep them all together 
	INT32 offset = 0;
	if (szTotal < 65536)
	{
		// write them in consequent order, as they were created
		for (std::list<std::pair<OTLookup*, CCFile*>>::iterator it = lst.begin(); it != lst.end(); ++it)
		{
			OTLookup* otl = it->first;
			CCMemFile* pLookupFile = (CCMemFile*)it->second;
			WriteUpdateLookup(otl, pFile, pLookupFile, offset);
			pLookupFile->close();
			delete pLookupFile;
		}
	}
	else
	{
		// Write them in ascending order of lookup file size
		for (OTLookupFileMap::iterator it = map.begin(); it != map.end(); ++it)
		{
			OTLookup* otl = it->first.first;
			CCMemFile* pLookupFile = (CCMemFile*)it->second;
			WriteUpdateLookup(otl, pFile, pLookupFile, offset);
			pLookupFile->close();
			delete pLookupFile;
		}
	}
}

/******************************************************************************
   Function:	DumpScriptTable
   Author:		Paul Gruzdev
   Purpose:		Dump GPOS/GSUB script list table: https://www.microsoft.com/typography/otspec/chapter2.htm
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	ScriptSet * ss
   Parameters:	GInfo * ginfo
   Notes:       
******************************************************************************/
static void DumpScriptTable(CCFile* pFile, ScriptSet* ss, GInfo* ginfo) 
{
	INT32 offset;
	int lcnt = 0, iDefaultLang = -1;

	// Count the languages, and find default 
	for (lcnt = 0; lcnt < (int)ss->vLangSys.size(); ++lcnt)
		if (ss->vLangSys[lcnt].lang == DEFAULT_LANG)
			iDefaultLang = lcnt;
	if (iDefaultLang != -1)
		--lcnt;

	FilePos base = pFile->tell();
	putshort(pFile, 0);				// fill in later, Default Lang Sys
	putshort(pFile, (short)lcnt);
	for (size_t i = 0; i < ss->vLangSys.size(); ++i) 
	{
		if (i != iDefaultLang)
		{
			putlong(pFile, ss->vLangSys[i].lang);	// Language tag
			putshort(pFile, 0);						// Fill in later, offset to langsys
		}
	}

	for (size_t lcnt = 0; lcnt < ss->vLangSys.size(); ++lcnt) 
	{
		if (ss->vLangSys[lcnt].iSameAs != -1)
			offset = ss->vLangSys[ss->vLangSys[lcnt].iSameAs].offset;
		else 
		{
			offset = pFile->tell();
			ss->vLangSys[lcnt].offset = offset;
		}

		pFile->seek(FilePos(lcnt == (size_t)iDefaultLang ? base :
			lcnt < (size_t)iDefaultLang || iDefaultLang == -1 ? base + 4 + lcnt * 6 + 4 :
			base + 4 + (lcnt - 1) * 6 + 4));

		putshort(pFile, (short)(offset - base));
		pFile->seekIn(0, ios_base::end);

		if (ss->vLangSys[lcnt].iSameAs == -1) 
		{
			int req_index = -1;
			for (size_t j = 0; j < ss->vLangSys[lcnt].vFeatureID.size(); ++j) 
			{
				if (ginfo->vFeatLookups[ss->vLangSys[lcnt].vFeatureID[j]].tag == REQUIRED_FEATURE) 
				{
					req_index = ss->vLangSys[lcnt].vFeatureID[j];
					break;
				}
			}
			putshort(pFile, 0);			// LookupOrder, always NULL
			putshort(pFile, (short)req_index);	// index of required feature, if any
			putshort(pFile, (short)(ss->vLangSys[lcnt].vFeatureID.size() - (req_index != -1)));

			// count of non-required features
			for (size_t j = 0; j < ss->vLangSys[lcnt].vFeatureID.size(); ++j) 
				if (ss->vLangSys[lcnt].vFeatureID[j] != req_index)
					putshort(pFile, (short)ss->vLangSys[lcnt].vFeatureID[j]);
		}
	}
}

/******************************************************************************
   Function:	FigureExtensionSubTables
   Author:		Paul Gruzdev
   Purpose:		Write out extension tables (if needed)
   Returns:		bool
   Parameters:	CCFile * pFile
   Parameters:	OTLookup * all
   Parameters:	int startoffset
   Notes:       
******************************************************************************/
static bool FigureExtensionSubTables(CCFile* pFile, LookupSet& all, int startoffset)
{
	LookupSubTable* sub = NULL;
	int offset;
	bool any = false;

	if (all.empty())
		return false;

	bool gotmore = true; 
	int cnt = 0, len = 0;

	while (gotmore) 
	{
		gotmore = false;
		offset = startoffset + 8 * cnt;

		for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
		{
			OTLookup* otf = *it;
			if (otf->iLookupIndex != -1)
			{
				if (otf->uNeedsExtension)
					continue;

				std::list<LookupSubTable*>::iterator its = otf->vSubTables.begin();
				for (; its != otf->vSubTables.end(); ++its)
				{
					sub = *its;

					if (sub->iSubtableOffset == -1)
						continue;

					if (!sub->vExtraSubTables.empty())
					{
						size_t i = 0;
						for (; i < sub->vExtraSubTables.size(); ++i) 
						{
							if (sub->vExtraSubTables[i] + offset > 65535)
								break;
						}
						if (i < sub->vExtraSubTables.size() && sub->vExtraSubTables[i] != -1)
							break;
					}
					else if (sub->iSubtableOffset + offset > 65535)
						break;
				}

				if (its != otf->vSubTables.end())
				{
					if (!any)
					{
						ERROR3_PF(("Lookup %s has an offset bigger than 65535 bytes. This means\nwe must use an extension lookup to output it.\nNot all applications support extension lookups.", otf->sLookupName.c_str()));
						any = true;
					}
					otf->uNeedsExtension = true;
					gotmore = true;
					len += 8 * (int)otf->vSubTables.size();
					++cnt;
				}
				offset -= 6 + 2 * (int)otf->vSubTables.size();
			}
		}
	}

	if (cnt == 0)		// No offset overflows
		return false;

	// Now we've worked out which lookups need extension tables and marked them
	// Generate the extension tables, and update the offsets to reflect the size
	// of the extensions
	int len2 = 0;

	for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
	{
		OTLookup* otf = *it;
		if (otf->iLookupIndex != -1)
		{
			std::list<LookupSubTable*>::iterator its = otf->vSubTables.begin();
			for (; its != otf->vSubTables.end(); ++its)
			{
				sub = *its;
				if (sub->iSubtableOffset == -1)
					continue;

				if (!sub->vExtraSubTables.empty()) 
				{
					for (size_t i = 0; i < sub->vExtraSubTables.size(); ++i)
					{
						sub->vExtraSubTables[i] += len;
						if (otf->uNeedsExtension)
						{
							int off = pFile->tell();
							putshort(pFile, 1);	// extended subtable format (there's only one)
							putshort(pFile, (short)(otf->eLookupType & 0xff));
							putlong(pFile, sub->vExtraSubTables[i] - len2);
							sub->vExtraSubTables[i] = off;
							len2 += 8;
						}
					}
				}
				else 
				{
					sub->iSubtableOffset += len;
					if (otf->uNeedsExtension) 
					{
						int off = pFile->tell();
						putshort(pFile, 1);	// extended subtable format (there's only one)
						putshort(pFile, (short)(otf->eLookupType & 0xff));
						putlong(pFile, sub->iSubtableOffset - len2);
						sub->iSubtableOffset = off;
						len2 += 8;
					}
				}
			}
		}
	}

	return true;
}

/******************************************************************************
   Function:	DumpKern
   Author:		Paul Gruzdev
   Purpose:		Write out a 'kern' table
   Returns:		LONG
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
LONG SFNTWriter::DumpKern(CCFile* pFile)
{
	if (m_ulPairCount > 10920)
	{
		ERROR2RAW("The 'kern' table supports at most 10920 kern pairs in a subtable");
		return 0;
	}

	ULONG c = m_ulPairCount;
	putshort(pFile, 0);			// version. MS only supports 0 so we do this version for now
	putshort(pFile, 1);			// number of subtables

	putshort(pFile, 0);			// subtable version
	putshort(pFile, (short)((7 + 3*c)*sizeof(UINT16))); // subtable length
	putshort(pFile, 1);			// coverage, flags=hor/vert&format=0

	putshort(pFile, (short)c);

	ULONG i, j;
	for (i = 1, j=0; i <= c; i <<= 1, ++j );
	i >>= 1;
	--j;

	putshort(pFile, (short)i*6);		// binary search headers
	putshort(pFile, (short)j);
	putshort(pFile, (short)(6*(c - i)));

	size_t szCntAdded = 0;
	for(KernPairMap::iterator it = m_KPairs.begin(); it != m_KPairs.end(); ++it)
	{
		for(std::set<CKernData>::iterator its = it->second.begin(); its != it->second.end(); ++its)
		{
			putshort(pFile, it->first);
			putshort(pFile, its->iGID);
			putshort(pFile, (INT16)xar2ttf(its->wAmount));
			szCntAdded++;
		}
	}

	if(szCntAdded != m_ulPairCount)
	{
		ERROR3_PF(("%d kern pairs added, %pairs were about to add", szCntAdded, m_ulPairCount));
	}

	return pFile->tell();
}

/******************************************************************************
   Function:	DumpTable
   Author:		Paul Gruzdev
   Purpose		Dump out either a gpos or a gsub table. gpos handles kerns, gsub ligs
				we assume that SFFindUnusedLookups has been called
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	ETableType type
   Notes:       
******************************************************************************/
void SFNTWriter::DumpTable(CCFile* pFile, ETableType type)
{
	GInfo ginfo;
	FindFeatures(type, &ginfo);
	if (ginfo.vScripts.empty())
		return; 

	bool bIsGPos = type == keBoth || type == keGPOS || type == keBothAndExtra;
	CCMemFile lookups(NULL, 0, CCMemWrite);
	WriteLookups(&lookups, bIsGPos);

	if (ginfo.vScripts.empty() && lookups.tell() == 0)
	{
	// ftell(lfile)==0 => There are no lookups for this table 
	// ginfo.sc==0 => There are no scripts. 
	// If both are true then we don't need to output the table 
	// It is perfectly possible to have lookups without scripts 
	//  (if some other table referred to them -- we don't currently 
	//  support this, but we might some day). 
	// It is also possible to have scripts without lookups (to get 
	//  around a bug in Uniscribe which only processes some scripts 
	//  if both GPOS and GSUB entries are present. So we share scripts 
	//  between the two tables 
		return;
	}

	putlong(pFile, 0x10000);	// version number
	putshort(pFile, 10);		// offset to script table
	putshort(pFile, 0);			// offset to features. Come back for this
	putshort(pFile, 0);			// offset to lookups.  Come back for this

	// Now the scripts
	FilePos fpScriptsStartOffset = pFile->tell();
	putshort(pFile, (short)ginfo.vScripts.size());

	for (size_t i = 0; i < ginfo.vScripts.size(); ++i) 
	{
		putlong(pFile, ginfo.vScripts[i].script);
		putshort(pFile, 0);	// fix up later
	}

	// Ok, that was the script_list_table which gives each script an offset 
	// Now for each script we provide a Script table which contains an 
	//  offset to a bunch of features for the default language, and a 
	//  a more complex situation for non-default languages. 
	INT32 offset = 2 + 4;			// To the script pointer at the start of table
	for (size_t i = 0; i < ginfo.vScripts.size(); ++i)
	{
		FilePos here = pFile->tell();
		pFile->seek(fpScriptsStartOffset + offset);
		putshort(pFile, (short)(here - fpScriptsStartOffset));
		offset += 6;
		pFile->seek(here);
		DumpScriptTable(pFile, &(ginfo.vScripts[i]), &ginfo);
	}
	// And that should finish all the scripts/languages

	// Now the features
	FilePos fpFeatureListTableStart = pFile->tell();;
	pFile->seek(6);
	putshort(pFile, (short)fpFeatureListTableStart);
	pFile->seekIn(0, ios_base::end);
	putshort(pFile, (short)ginfo.vFeatLookups.size());	// Number of features
	offset = 2 + 6 * (INT32)ginfo.vFeatLookups.size();			// Offset to start of first feature table from beginning of feature_list

	for (size_t i = 0; i < ginfo.vFeatLookups.size(); ++i)
	{
		putlong(pFile, ginfo.vFeatLookups[i].tag);
		putshort(pFile, (short)offset);
		offset += 4 + 2 * (INT32)ginfo.vFeatLookups[i].vLookups.size();
	}

	// for each feature, one feature table
	INT32 iSizeParamsPtr = 0, iSizeParamsLoc;
	for (size_t i = 0; i < ginfo.vFeatLookups.size(); ++i)
	{
		ginfo.vFeatLookups[i].name_param_ptr = 0;
		if (ginfo.vFeatLookups[i].tag == CHR('s', 'i', 'z', 'e'))
			iSizeParamsPtr = pFile->tell();
		else if (ginfo.vFeatLookups[i].tag >= CHR('s', 's', '0', '1') && ginfo.vFeatLookups[i].tag <= CHR('s', 's', '2', '0'))
			ginfo.vFeatLookups[i].name_param_ptr = pFile->tell();

		putshort(pFile, 0);			// No feature params (we'll come back for 'size')
		putshort(pFile, (short)ginfo.vFeatLookups[i].vLookups.size());	// this many lookups
		for (size_t j = 0; j < ginfo.vFeatLookups[i].vLookups.size(); ++j)
			putshort(pFile, (short)ginfo.vFeatLookups[i].vLookups[j]->iLookupIndex);
		// index of each lookup
	}

	if (iSizeParamsPtr != 0)
	{
		iSizeParamsLoc = pFile->tell();
		pFile->seek(iSizeParamsPtr);
		// When Adobe first released fonts containing the 'size' feature 
		//  they did not follow the spec, and the offset to the size parameters 
		//  was relative to the wrong location. They claim (Aug 2006) that 
		//  this has been fixed. FF used to do what Adobe did. Many programs 
		//  expect broken sizes tables now. Therefore allow the user to chose 
		//  which kind to output 
		// Well don't give the user the choice any more, but retain the old 
		//  code, just in case... 
#if 0
		putshort(pFile, size_params_loc - ((at->gi.flags&ttf_flag_brokensize) ? feature_list_table_start : size_params_ptr));
#else
		putshort(pFile, (short)(iSizeParamsLoc - iSizeParamsPtr));
#endif
		pFile->seek(iSizeParamsLoc);
		putshort(pFile, 160); // 10 << 20);		// sf->design_size - whichever this is?
#if 0
		if (0) //sf->fontstyle_id != 0 || sf->fontstyle_name != NULL)
		{
			putshort(pFile, 0);	// sf->fontstyle_id);
			// at->fontstyle_name_strid = at->next_strid++;
			putshort(pFile, 0); // at->fontstyle_name_strid);
		}
		else
#endif
		{
			putshort(pFile, 0);
			putshort(pFile, 0);
		}
		putshort(pFile, 0); // sf->design_range_bottom);
		putshort(pFile, 0); // sf->design_range_top);
	}
/*	// We are not writing feature names atm
	for (size_t i = 0; i < ginfo.vFeatLookups.size(); ++i)
	{
		if (ginfo.vFeatLookups[i].name_param_ptr != 0 &&
			(fn = findotffeatname(ginfo.feat_lookups[i].tag, sf)) != NULL)
		{
			if (fn->nid == 0)
				fn->nid = at->next_strid++;
			name_param_loc = ftell(g___);
			fseek(g___, ginfo.feat_lookups[i].name_param_ptr, SEEK_SET);
			putshort(g___, name_param_loc - ginfo.feat_lookups[i].name_param_ptr);
			fseek(g___, name_param_loc, SEEK_SET);
			putshort(g___, 0);		// Minor version number
			putshort(g___, fn->nid);
		}
	}*/
	// And that should finish all the features

	// Now the lookups
	int cnt = 0;
	LookupSet& all = type != keGSUB ? m_GPOSLookups : m_GSUBLookups;
	for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
	{
		if ((*it)->iLookupIndex != -1)
			++cnt;
	}

	FilePos fpLookupListTableStart = pFile->tell();
	pFile->seek(8);
	putshort(pFile, (short)fpLookupListTableStart);
	pFile->seekIn(0, ios_base::end);
	putshort(pFile, (short)cnt);
	offset = 2 + 2 * cnt;		// Offset to start of first lookup table from beginning of lookup list

	for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
	{
		if ((*it)->iLookupIndex != -1)
		{
			int scnt = 0;
			std::list<LookupSubTable*>& sub = (*it)->vSubTables;
			std::list<LookupSubTable*>::iterator itSub = sub.begin();
			putshort(pFile, (short)offset);
			for (; itSub != sub.end(); ++itSub)
			{
				if ((*itSub)->iSubtableOffset == -1)
					continue;
				else if (!(*itSub)->vExtraSubTables.empty()) 
					scnt += (int)(*itSub)->vExtraSubTables.size();
				else
					++scnt;
			}

			offset += 6 + 2 * scnt;		// 6 bytes header +2 per lookup
			if ((*it)->uLookupFlags & PST_FLAG_USE_MARK_FILTERING_SET)
				offset += 2;		// For mark filtering set, if used
		}
	}

	offset -= 2 + 2 * cnt;
	// now the lookup tables
	// do we need any extension sub-tables?
	CCMemFile eFile(NULL, 0, CCMemWrite);
	bool bExtensionWritten = FigureExtensionSubTables(&eFile, all, offset);

	for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
	{
		OTLookup* otf = *it;
		if (otf->iLookupIndex != -1)
		{
			putshort(pFile, !otf->uNeedsExtension ? (short)(otf->eLookupType & 0xff) : bIsGPos ? 9 : 7);
			putshort(pFile, (short)(otf->uLookupFlags & 0xffff));
			putshort(pFile, (short)otf->vSubTables.size());

			std::list<LookupSubTable*>::iterator its = otf->vSubTables.begin();
			for (; its != otf->vSubTables.end(); ++its)
			{
				LookupSubTable* sub = *its;
				if (sub->iSubtableOffset == -1)
					continue;
				else if (!sub->vExtraSubTables.empty()) 
				{
					for (size_t i = 0; sub->vExtraSubTables.size(); ++i)
						putshort(pFile, (short)(offset + sub->vExtraSubTables[i]));
				}
				else
					putshort(pFile, (short)(offset + sub->iSubtableOffset));

				// Offset to lookup data which is in the temp file
				// we keep adjusting offset so it reflects the distance between
				// here and the place where the temp file will start, and then
				// we need to skip l->offset bytes in the temp file
				// If it's a big GPOS/SUB table we may also need some extension
				//  pointers, but FigureExtension will adjust for that
			}

			offset -= 6 + 2 * (INT32)otf->vSubTables.size();
			if (otf->uLookupFlags & PST_FLAG_USE_MARK_FILTERING_SET)
			{
				putshort(pFile, (short)otf->uLookupFlags >> 16);
				offset -= 2;
			}
		}
	}

	if (bExtensionWritten)
		pFile->write(eFile.GetMemoryBuffer(), eFile.tell());

	if (lookups.tell())
		pFile->write(lookups.GetMemoryBuffer(), lookups.tell());

	// Once we created a memory file - it's open, so need to explicitly close it to avoid the warning message.
	eFile.close();
	lookups.close();

	for (LookupSet::iterator it = all.begin(); it != all.end(); ++it)
	{
		OTLookup* otf = *it;

		for (std::list<LookupSubTable*>::iterator its = otf->vSubTables.begin(); its != otf->vSubTables.end(); ++its)
			(*its)->vExtraSubTables.clear();
		otf->uNeedsExtension = false;
	}

#if defined(_DEBUG) && defined(DUMP_EXTRA_INFO)
	if (pFile->Size())
	{
		FILE* out = NULL;
		char psz[256];
		sprintf_s(psz, 256, bIsGPos ? "d:/temp/cff/%s%d%s_gpos.tst" : "d:/temp/cff/%s%d%s_gsub.tst", m_sFullName.c_str(), m_lWeight, m_bItalic ? "i" : "");
		fopen_s(&out, psz, "wb");
		if (out) {
			fwrite(((CCMemFile*)pFile)->GetMemoryBuffer(), pFile->Size(), 1, out);
			fclose(out);
		}
	}
#endif

}

/******************************************************************************
   Function:	DumpGSUBLigData
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		void
   Parameters:	CCFile * gsub
   Parameters:	LookupSubTable * pSub
   Notes:       
******************************************************************************/
void SFNTWriter::DumpGSUBLigData(CCFile* gsub, LookupSubTable* pSub)
{
	std::vector<int> glyphs;
	std::vector<UINT16> offsets;

	GlyphsInSubtable(glyphs, pSub);
	size_t cnt = glyphs.size();

	putshort(gsub, 1);		// only one format for ligatures
	FilePos coverage_pos = gsub->tell();
	putshort(gsub, 0);		// offset to coverage table, fill-in later
	putshort(gsub, (short)cnt);
	FilePos next_val_pos = gsub->tell();

	for (size_t i = 0; i < cnt; ++i)
		putshort(gsub, 0);

	for (size_t i = 0; i < cnt; ++i)
	{
		std::vector<UINT16> ligoffsets;
		std::set<CLigatureData>& tos = m_Ligatures[glyphs[i]];
		offsets.push_back(gsub->tell() - coverage_pos + 2);

		size_t pcnt = 0;
		for (std::set<CLigatureData>::iterator its = tos.begin(); its != tos.end(); ++its)
			if (its->pSubTable == pSub)
				pcnt++;

		putshort(gsub, (short)pcnt);
		FilePos lig_list_start = gsub->tell();
		for (size_t j = 0; j < pcnt; ++j)
			putshort(gsub, 0);			// Place holders to fill in later

		for (std::set<CLigatureData>::iterator its = tos.begin(); its != tos.end(); ++its)
		{
			if (its->pSubTable == pSub)
			{
				ligoffsets.push_back(gsub->tell() - lig_list_start + 2);
				putshort(gsub, (short)its->iLigGID);
				putshort(gsub, (short)(its->to.size() + 1));
				//if (lcnt + 1 > at->os2.maxContext)
				//	at->os2.maxContext = lcnt + 1;
				for (size_t j = 0; j < its->to.size(); j++)
					putshort(gsub, (short)its->to[j]);
			}
		}
		gsub->seek(lig_list_start);
		for (size_t j = 0; j < ligoffsets.size(); ++j)
			putshort(gsub, ligoffsets[j]);
		gsub->seekIn(0, std::ios_base::end);
	}

	if (glyphs.size())
	{
		FilePos here = gsub->tell();
		gsub->seek(coverage_pos);
		putshort(gsub, here - coverage_pos + 2);
		gsub->seek(next_val_pos);
		for (size_t i = 0; i < cnt; ++i)
			putshort(gsub, offsets[i]);
		gsub->seek(here);
		DumpCoverageTable(gsub, glyphs);
	}
}

/******************************************************************************
   Function:	DumpGSUBSimpleSubs
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		void
   Parameters:	CCFile * gsub
   Parameters:	LookupSubTable * pSub
   Notes:       
******************************************************************************/
void SFNTWriter::DumpGSUBSimpleSubs(CCFile* gsub, LookupSubTable* pSub)
{
	std::vector<int> coverage;
	std::map<int, int> glyphs;

	for (SubstituteMap::iterator it = m_Substitutes.begin(); it != m_Substitutes.end(); ++it)
	{
		for (std::set<CSubstituteData>::iterator its = it->second.begin(); its != it->second.end(); ++its)
		{
			if (its->pSubTable == pSub)
			{
				glyphs[its->from] = its->to;
				coverage.push_back(it->first);
				break;
			}
		}
	}

	size_t cnt = glyphs.size();

	putshort(gsub, 2);		// glyph list format
	FilePos fpCoverage = gsub->tell();
	putshort(gsub, 0);		// offset to coverage table
	putshort(gsub, (short)cnt);

	for (std::map<int, int>::iterator it = glyphs.begin(); it != glyphs.end(); ++it)
		putshort(gsub, it->second);

	FilePos fpEnd = gsub->tell();
	gsub->seek(fpCoverage);
	putshort(gsub, fpEnd - fpCoverage + 2);
	gsub->seek(fpEnd);
	DumpCoverageTable(gsub, coverage);
}

void SFNTWriter::dumpGSUBmultiplesubs(CCFile* gsub, LookupSubTable* pSub) 
{
	std::vector<int> glyphs;

	GlyphsInSubtable(glyphs, pSub);
	size_t cnt = glyphs.size();

	putshort(gsub, 1);		// glyph list format
	FilePos fpCoverage = gsub->tell();
	putshort(gsub, 0);		// offset to coverage table
	putshort(gsub, (short)cnt);

	size_t offset = 6 + 2 * cnt;
// 	for (cnt = 0; glyphs[cnt] != NULL; ++cnt) 
// 	{
// 		putshort(gsub, offset);
// 		for (int gc = 0; maps[cnt][gc] != NULL; ++gc);
// 		offset += 2 + 2 * gc;
// 	}
// 
// 	for (cnt = 0; glyphs[cnt] != NULL; ++cnt) 
// 	{
// 		for (gc = 0; maps[cnt][gc] != NULL; ++gc);
// 		putshort(gsub, gc);
// 
// 		for (gc = 0; maps[cnt][gc] != NULL; ++gc)
// 			putshort(gsub, maps[cnt][gc]->ttf_glyph);
// 	}

	FilePos end = gsub->tell();
	gsub->seek(fpCoverage);
	putshort(gsub, end - fpCoverage + 2);
	gsub->seek(end);
	DumpCoverageTable(gsub, glyphs);
}

void SFNTWriter::OrderedGlyphsFromString(std::wstring str, std::vector<int>& glyphs) {
	std::set<int> set;
	glyphs.clear();

	for (size_t i = 0; i < str.length(); i++) {
		GIDMap::iterator it = m_CodeToGID.find(str[i]);
		if (it != m_CodeToGID.end())
			set.insert(it->second);
	}
	for (std::set<int>::iterator it = set.begin(); it != set.end(); ++it)
		glyphs.push_back(*it);
}

void SFNTWriter::DumpGContextChainCoverage(CCFile* gsub, LookupSubTable* sub)
{
	FilePos base = gsub->tell();
	std::vector<int> glyphs;
	CSubstRule& rule = m_Rules[sub];
	int lc = rule.pRef ? 1 : 0;

	putshort(gsub, 3);		// Sub format 3 => coverage

	putshort(gsub, (short)rule.pre.size()); // Number of glyphs in the backtracking sequence.
	FilePos bbase = gsub->tell();

	// Array of offsets to coverage tables in backtracking sequence.Offsets are from beginning of substitution subtable, in glyph sequence order.
	for (size_t i = 0; i < rule.pre.size(); ++i)
		putshort(gsub, 0);

	putshort(gsub, (short)rule.keys.size()); // Number of glyphs in input sequence
	FilePos ibase = gsub->tell();
	// Array of offsets to coverage tables in input sequence. Offsets are from beginning of substitution subtable, in glyph sequence order.
	for (size_t i = 0; i < rule.keys.size(); ++i)
		putshort(gsub, 0);

	putshort(gsub, (short)rule.post.size()); // Number of glyphs in lookahead sequence
	FilePos lbase = gsub->tell();

	// Array of offsets to coverage tables in lookahead sequence. Offsets are from beginning of substitution subtable, in glyph sequence order.
	for (size_t i = 0; i < rule.post.size(); ++i)
		putshort(gsub, 0);

	putshort(gsub, lc); // substitutionCount: number of SubstLookupRecords

	if(lc > 0) {
		putshort(gsub, 0);		// glyphSequenceIndex: index into current glyph sequence  first glyph = 0.
		putshort(gsub, rule.pRef ? rule.pRef->iLookupIndex : 65535);	// lookupListIndex: lookup to apply to that position  zero-based index. -1 is the blocking rule
	}

	for (size_t i = 0; i < rule.keys.size(); ++i)
	{
		FilePos pos = gsub->tell();
		gsub->seek(ibase + 2 * (FilePos)i);
		putshort(gsub, pos - base);
		gsub->seek(pos);

		OrderedGlyphsFromString(rule.keys[i], glyphs);
		DumpCoverageTable(gsub, glyphs);
	}
	for (size_t i = 0; i < rule.pre.size(); ++i) {
		FilePos pos = gsub->tell();
		gsub->seek(bbase + 2 * (FilePos)i);
		putshort(gsub, pos - base);
		gsub->seek(pos);

		OrderedGlyphsFromString(rule.pre[i], glyphs);
		DumpCoverageTable(gsub, glyphs);
	}
	for (size_t i = 0; i < rule.post.size(); ++i) {
		FilePos pos = gsub->tell();
		gsub->seek(lbase + 2 * (FilePos)i);
		putshort(gsub, pos - base);
		gsub->seek(pos);

		OrderedGlyphsFromString(rule.post[i], glyphs);
		DumpCoverageTable(gsub, glyphs);
	}
// 
// 	curcontext = fpst->rules[0].u.coverage.ncnt + fpst->rules[0].u.coverage.bcnt + fpst->rules[0].u.coverage.fcnt;
// 	if (curcontext > at->os2.maxContext)
// 		at->os2.maxContext = curcontext;
}

}
