// Header for the OTFDataLoader implementation

#pragma once

struct OTLookup;
struct FeatureParamSubTab;
struct AppleStateMachine;
struct KernClass;

#ifndef SAFE_DELETE
#define SAFE_DELETE(ptr) if (ptr) { delete ptr; ptr = NULL; }
#endif

/*
	Author:     Greg Watroba
	Created:    23/11/15
	Note:		Structures defined here were copied and/or paraphrased from ttf project found in:
	http://svn.xara.com/camelot/trunk/src/PSDFilter/XPFilter/ttf
	*/

/*
	If more than this we allocate more_langs in chunks of MAX_LANG
	*/
#define MAX_LANG 4

/*
	Macro mainly used for comparing tag values
	*/
#define CHR(ch1,ch2,ch3,ch4) (((ch1)<<24)|((ch2)<<16)|((ch3)<<8)|(ch4))

/*
	Macros for moving around BYTE table
*/

#define MOVE8BITS(p) (p += sizeof(UCHAR))
#define MOVE16BITS(p) (p += sizeof(USHORT))
#define MOVE24BITS(p) (p += 3)
#define MOVE32BITS(p) (p += sizeof(LONG))

/*
	Several useful macros
*/

/* 
	OpenType does not document 'dflt' as a language, but we'll use it anyway.
	(Adobe uses it too) we'll turn it into a default entry when we output it. 
*/
#define DEFAULT_LANG		CHR('d','f','l','t')

/* 
	The OpenType spec says in one place that the default script is 'dflt' and
	in another that it is 'DFLT'. 'DFLT' is correct 
*/
#define DEFAULT_SCRIPT		CHR('D','F','L','T')
#define REQUIRED_FEATURE	CHR(' ','R','Q','D')

/*
	Apple State Machine Type
*/
typedef enum
{
	ASM_TYPE_INDIC,
	ASM_TYPE_CONTEXT,
	ASM_TYPE_LIG,
	ASM_TYPE_SIMPLE = 4,
	ASM_TYPE_INSERT,
	ASM_TYPE_KERN = 0x11
}AsmType;

/*
	Apple State Machine Flags
*/
typedef enum
{
	ASM_FLAG_VERT = 0x8000,
	ASM_FLAG_DESCENDING = 0x4000,
	ASM_FLAG_ALWAYS = 0x2000
}AsmFlag;

/*
	Open Type Lookup Type
*/
typedef enum
{
	/* Not a lookup type */
	OTL_TYPE_UNDEF = 0,
	/* Not a lookup type */
	OTL_TYPE_GSUB_START = 0x000,
	OTL_TYPE_GSUB_SINGLE = 0x001,
	OTL_TYPE_GSUB_MULTIPLE = 0x002,
	OTL_TYPE_GSUB_ALTERNATE = 0x003,
	OTL_TYPE_GSUB_LIGATURE = 0x004,
	OTL_TYPE_GSUB_CONTEXT = 0x005,
	OTL_TYPE_GSUB_CONTEXT_CHAIN = 0x006,
	/* GSUB extension 7 */
	OTL_TYPE_GSUB_REVERSE_C_CHAIN = 0x008,
	/* mac state machines */
	OTL_TYPE_MORX_INDIC = 0x0fd,
	OTL_TYPE_MORX_CONTEXT = 0x0fe,
	OTL_TYPE_MORX_INDEX = 0x0ff,
	/* Not a lookup type */
	OTL_TYPE_GPOS_START = 0x100,		
	OTL_TYPE_GPOS_SINGLE = 0x101,
	OTL_TYPE_GPOS_PAIR = 0x102,
	OTL_TYPE_GPOS_CURSIVE = 0x103,
	OTL_TYPE_GPOS_MARK_2_BASE = 0x104,
	OTL_TYPE_GPOS_MARK_2_LIGATURE = 0x105,
	OTL_TYPE_GPOS_MARK_2_MARK = 0x106,
	OTL_TYPE_GPOS_CONTEXT = 0x107,
	OTL_TYPE_GPOS_CONTEXT_CHAIN = 0x108,
	/* GPOS extension 9 */
	OTL_TYPE_KERN_STATE_MACHINE = 0x1ff
	/* otlookup&0xff == lookup type for the appropriate table */
	/* otlookup>>8:     0=>GSUB, 1=>GPOS */
}OTLookupType;

/*
	Position SubType
*/
typedef enum
{ 
	PST_NULL = 0,
	PST_POSITION,
	PST_PAIR,
	PST_SUBSTITUTION,
	PST_ALTERNATE,
	PST_MULTIPLE,
	PST_LIGATURE,
	/* must be PST_MAX-1, see charinfo.c*/
	PST_LCARET,
	PST_MAX,
	/* These are not psts but are related so it's handly to have values for them */
	PST_KERNING = PST_MAX, 
	PST_VERTICAL_KERNING, 
	pst_anchors,
	/* And these are fpsts */
	PST_CONTEXT_POS,
	PST_CONTEXT_SUB, 
	PST_CHAIN_POS,
	PST_CHAIN_SUB,
	PST_REVERSE_SUB,
	FPST_MAX,
	/* And these are used to specify a kerning pair where the current */
	/*  char is the final glyph rather than the initial one */
	/* A kludge used when cutting and pasting features */
	PST_KERN_BACK,
	PST_VERTICAL_KERN_BACK
}PosSubType;

/*
	Feature Positioning Sub Format
*/

typedef enum
{
	FPST_GLYPHS = 0,
	FPST_CLASS,
	FPST_COVERAGE,
	FPST_REVERSECOVERAGE,
	FPST_FORMATMAX
}FeaturePosSubFormat;

/*
	Positioning Sub Tables Flags
*/
typedef enum
{
	PST_FLAG_R2L = 1,
	PST_FLAG_IGNORE_BASE_GLYPHS = 2,
	PST_FLAG_IGNORE_LIGATURES = 4,
	PST_FLAG_IGNORE_COMBINING_MARKS = 8,
	PST_FLAG_USE_MARK_FILTERING_SET = 0x10,
	PST_FLAG_MARK_CLASS = 0xff00, 
	PST_FLAG_MARK_SET = 0xffff0000
}PSTFlag;

/*
	True Type Collection Flags
*/
typedef enum 
{ 
	TTCF_TRY_MERGE = 0x1, 
	TTCF_CFF = 0x2
}TTCFlags;

/*
	Opent Type Flags
*/
typedef enum
{ 
	OTF_FS_TYPE_PERMITTED = 1, 
	OTF_ASK_CMP = 2, 
	OTF_ALL_GLYPHS_IN_TTC = 4,
	OTF_FONT_LINT = 8, 
	OTF_HIDE_WINDOW = 0x10 
}OTFlags;

/*
	Mode to search encodings
*/
typedef enum
{ 
	GIUT_NORMAL = 0, 
	GIUT_JUST_IN_USE,
	GIUT_FIND_NAMES
}GsubInUseType;

/*
	Unicode Interpretation
*/
typedef enum 
{ 
	UI_UNSET= -1, 
	UI_NONE, 
	UI_ADOBE, 
	UI_GREEK, 
	UI_JAPANESE,
	UI_TRAD_CHINESE, 
	UI_SIMP_CHINESE, 
	UI_KOREAN, 
	UI_AMS 
}UnicodeIterp;

#ifndef CAMELOT_PRODUCT
// this enum is used for both GPOS and GSUB subtables in font
// can be used for glyph substitution type identification
typedef enum GlyphSetEnum
{
	GSET_NONE = 0,
	GSET_ACCESS_ALL_ALTERNATES,
	GSET_ABOVE_BASE_FORMS,
	GSET_ABOVE_BASE_MARK,
	GSET_ABOVE_BASE_SUBSTITUTIONS,
	GSET_VERTICAL_FRACTIONS,
	GSET_AKHAND,
	GSET_ANCIENT_LIGATURES,
	GSET_BELOW_BASE_FORMS,
	GSET_BELOW_BASE_MARK,
	GSET_BELOW_BASE_SUBSTITUTIONS,
	GSET_CAPITALS_TO_PETITE_CAPITALS,
	GSET_CAPITALS_TO_SMALL_CAPITALS,
	GSET_CONTEXTUAL_ALTERNATES,
	GSET_CASE_SENSITIVE_FORMS,
	GSET_GLYPH_COMPOSITION_DECOMPOSITION,
	GSET_CONJUNCT_FORM_AFTER_RO,
	GSET_CONJUNCT_FORMS,
	GSET_CONTEXTUAL_LIGATURES,
	GSET_CENTERED_CJK_PUNCTUATION,
	GSET_CAPITAL_SPACING,
	GSET_CONTEXTUAL_SWASH,
	GSET_CURSIVE_ATTACHMENT,
	GSET_CHARACTER_VARIANTS_00,
	GSET_CHARACTER_VARIANTS_01,
	GSET_CHARACTER_VARIANTS_02,
	GSET_CHARACTER_VARIANTS_03,
	GSET_CHARACTER_VARIANTS_04,
	GSET_CHARACTER_VARIANTS_05,
	GSET_CHARACTER_VARIANTS_06,
	GSET_CHARACTER_VARIANTS_07,
	GSET_CHARACTER_VARIANTS_08,
	GSET_CHARACTER_VARIANTS_09,
	GSET_CHARACTER_VARIANTS_10,
	GSET_CHARACTER_VARIANTS_99,
	GSET_DROP_CAPS,
	GSET_DISTANCE,
	GSET_DISCRETIONARY_LIGATURES,
	GSET_DENOMINATORS,
	GSET_DIPTHONGS,
	GSET_DOTLESS_FORMS,
	GSET_EXPERT_FORMS,
	GSET_FINAL_GLYPH_ON_LINE,
	GSET_TERMINAL_FORMS_2,
	GSET_TERMINAL_FORMS_3,
	GSET_TERMINAL_FORMS,
	GSET_FLATTENED_ACCENTS_OVER_CAPITALS,
	GSET_DIAGONAL_FRACTIONS,
	GSET_FULL_WIDTHS,
	GSET_HALF_FORMS,
	GSET_HALANT_FORMS,
	GSET_ALTERNATIVE_HALF_WIDTHS,
	GSET_HISTORICAL_FORMS,
	GSET_HORIZONTAL_KANA_ALTERNATIVES,
	GSET_HISTORIC_LIGATURES,
	GSET_HANJA_TO_HANGUL,
	GSET_HOJO_KANJI_FORMS,
	GSET_HALF_WIDTHS,
	GSET_INITIAL_FORMS,
	GSET_ISOLATED_FORMS,
	GSET_ITALICS,
	GSET_JUSTIFICATION_ALTERNATIVES,
	GSET_JAPANESE_FORMS,
	GSET_JIS2004_FORMS,
	GSET_JIS78_FORMS,
	GSET_JIS83_FORMS,
	GSET_JIS90_FORMS,
	GSET_HORIZONTAL_KERNING,
	GSET_LEFT_BOUNDS,
	GSET_STANDARD_LIGATURES,
	GSET_LEADING_JAMO_FORMS,
	GSET_LINING_FIGURES,
	GSET_LOCALIZED_FORMS,
	GSET_MARK_POSITIONING,
	GSET_MEDIAL_FORMS_2,
	GSET_MEDIAL_FORMS,
	GSET_MATHEMATICAL_GREEK,
	GSET_MARK_TO_MARK,
	GSET_MARK_POSITIONING_VIA_SUBSTITUTION,
	GSET_ALTERNATE_ANNOTATION_FORMS,
	GSET_NLC_KANJI_FORMS,
	GSET_NUKTA_FORMS,
	GSET_NUMERATORS,
	GSET_OLDSTYLE_FIGURES,
	GSET_OPTICAL_BOUNDS,
	GSET_ORDINALS,
	GSET_ORNAMENTS,
	GSET_PROPORTIONAL_ALTERNATE_METRICS,
	GSET_LOWERCASE_TO_PETITE_CAPITALS,
	GSET_PROPORTIONAL_KANA,
	GSET_PROPORTIONAL_NUMBERS,
	GSET_PRE_BASE_FORMS,
	GSET_PRE_BASE_SUBSTITUTIONS,
	GSET_POST_BASE_FORMS,
	GSET_POST_BASE_SUBSTITUTIONS,
	GSET_PROPORTIONAL_WIDTH,
	GSET_QUARTER_WIDTHS,
	GSET_RANDOMIZE,
	GSET_RAKAR_FORMS,
	GSET_REQUIRED_LIGATURES,
	GSET_REPH_FORM,
	GSET_RIGHT_BOUNDS,
	GSET_RIGHT_TO_LEFT_ALTERNATES,
	GSET_RIGHT_TO_LEFT_MIRRORED_FORMS,
	GSET_RUBY_NOTATIONAL_FORMS,
	GSET_STYLISTIC_ALTERNATIVES,
	GSET_SCIENTIFIC_INFERIORS,
	GSET_LOWERCASE_TO_SMALL_CAPITALS,
	GSET_SIMPLIFIED_FORMS,
	GSET_STYLE_SET_1,
	GSET_STYLE_SET_2,
	GSET_STYLE_SET_3,
	GSET_STYLE_SET_4,
	GSET_STYLE_SET_5,
	GSET_STYLE_SET_6,
	GSET_STYLE_SET_7,
	GSET_STYLE_SET_8,
	GSET_STYLE_SET_9,
	GSET_STYLE_SET_10,
	GSET_STYLE_SET_11,
	GSET_STYLE_SET_12,
	GSET_STYLE_SET_13,
	GSET_STYLE_SET_14,
	GSET_STYLE_SET_15,
	GSET_STYLE_SET_16,
	GSET_STYLE_SET_17,
	GSET_STYLE_SET_18,
	GSET_STYLE_SET_19,
	GSET_STYLE_SET_20,
	GSET_SCRIPT_STYLE,
	GSET_SUBSCRIPT,
	GSET_SUPERSCRIPT,
	GSET_SWASH,
	GSET_TITLING,
	GSET_TRAILING_JAMO_FORMS,
	GSET_TRADITIONAL_NAME_FORMS,
	GSET_TABULAR_NUMBERS,
	GSET_TRADITIONAL_FORMS,
	GSET_THIRD_WIDTHS,
	GSET_UNICASE,
	GSET_ALTERNATE_VERTICAL_METRICS,
	GSET_VATTU_VARIANTS,
	GSET_VERTICAL_ALTERNATES,
	GSET_ALTERNATE_VERTICAL_HALF_METRICS,
	GSET_VOWEL_JAMO_FORMS,
	GSET_VERTICAL_KANA_ALTERNATES,
	GSET_VERTICAL_KERNING,
	GSET_PROPORTIONAL_ALTERNATE_VERTICAL_METRICS,
	GSET_VERTICAL_ROTATION_ALTERNATES,
	GSET_SLASHED_ZERO,
	GSET_REQUIRED_FEATURE
};
#endif

/*
	Lookup Sub Table
*/
struct PosSubTab;
struct LookupSubTable
{
	UINT16 UniqueID;
	std::vector<PosSubTab*> vecPosSubTab;
	std::string sSubtableName;
	GlyphSetEnum eGlyphSet;
	// for OTL_TYPE_GSUB_SINGLE, used to find a default replacement
	std::string sSuffix;
	// for OTL_TYPE_GPOS_PAIR, used to guess default kerning values
	INT16 iSeparation;
	INT16 iMinKern;
	OTLookup* pOTLookup;
	UINT32 uUnused : 1;
	UINT32 bPerGlyphPstOrKern : 1;
	UINT32 uAnchorClasses : 1;
	UINT32 uVerticalKerning : 1;
	UINT32 uTicked : 1;
	// for OTL_TYPE_GPOS_PAIR, calculate kerning so that glyphs will touch
	UINT32 uKerningByTouch : 1;
	// for kerning classes
	UINT32 uOnlyCloser : 1;
	// for kerning classes
	UINT32 uDontAutoKern : 1;
	KernClass* pKernClass;
	FeatureParamSubTab* pFPST;
	AppleStateMachine* pASM;
	/* Each time an item is added to a lookup we must place it into a */
	/*  subtable. If it's a kerning class, fpst or state machine it has */
	/*  a subtable all to itself. If it's an anchor class it can share */
	/*  a subtable with other anchor classes (merge with). If it's a glyph */
	/*  PosSubTab it may share a subtable with other PSTs */
	/* Note items may only be placed in lookups in which they fit. Can't */
	/*  put kerning data in a OTL_TYPE_GPOS_SINGLE lookup, etc. */
	LookupSubTable* pNext;
	INT32 iSubtableOffset;
	std::vector<INT32> vExtraSubTables;
	/* If a kerning subtable has too much stuff in it, we are prepared to */
	/*  break it up into several smaller subtables, each of which has */
	/*  an offset in this list (extra-subtables[0]==subtable_offset) */
	/*  the list is terminated by an entry of -1 */
	LookupSubTable() : UniqueID(0), eGlyphSet(GSET_NONE), iSeparation(0), iMinKern(0), uUnused(0), bPerGlyphPstOrKern(0), uAnchorClasses(0), 
					   uVerticalKerning(0), uTicked(0), uKerningByTouch(0), uOnlyCloser(0), uDontAutoKern(0), pKernClass(nullptr), 
					   pFPST(nullptr), pASM(nullptr), pNext(nullptr), iSubtableOffset(0) {}
};

/*
	Sequenced Lookups
*/
struct SeqLookup 
{
	INT32 iSeq;
	INT32 iLookupIndex;
	std::vector<OTLookup*> vLookup;

	SeqLookup() : iSeq(0), iLookupIndex(0) {}
};

typedef enum FPSTRuleType
{
	FPSTR_NONE = 0,
	FPSTR_GLYPH,
	FPSTR_CLASS,
	FPSTR_COVERAGE,
	FPSTR_REV_COVERAGE
};

/*
	Feature Parameters Sub Table Rule
*/
class FPSTRule
{
public:

	INT32 iLookupCnt;
	std::vector<SeqLookup*> vLookups;
	FPSTRuleType eType;

	FPSTRule(FPSTRuleType i_type) : iLookupCnt(0), eType(i_type) {}

	virtual ~FPSTRule()
	{
		for (size_t i = 0; i < vLookups.size(); i++)
			SAFE_DELETE(vLookups[i]);

		vLookups.clear();
	}
};

/*
	Feature Parameters Glyph
*/

class FPSTGlyphRule : public FPSTRule
{
public: 

	std::string sNames;
	std::vector<UINT16> vecNameGlyphs;

	std::string sBack;
	std::vector<UINT16> vecBackGlyphs;

	std::string sFore;
	std::vector<UINT16> vecForeGlyphs;

	FPSTGlyphRule() : FPSTRule(FPSTR_GLYPH) {}

	~FPSTGlyphRule() {}
};

/*
	Feature Parameters Class
*/

class FPSTClassRule : public FPSTRule
{
public:

	INT32 iNCnt;
	INT32 iBCnt;
	INT32 iFCnt;
	std::vector<UINT16> pNClasses;
	std::vector<UINT16> pBClasses;
	std::vector<UINT16> pFClasses;
	std::vector<UINT16> pAllClasses;

	FPSTClassRule() : FPSTRule(FPSTR_CLASS) {}

	~FPSTClassRule() {}
};

/*
	Feature Parameters Value Coverage
*/

class FPSTCoverageRule : public FPSTRule
{
public:

	INT32 iNCnt;
	INT32 iBCnt;
	INT32 iFCnt;
		
	std::vector<std::string> psNCovers;
	std::vector<std::vector<UINT16>> vecNCoverGlyphs;

	std::vector<std::string> psBCovers;
	std::vector<std::vector<UINT16>> vecBCoverGlyphs;

	std::vector<std::string> psFCovers;
	std::vector<std::vector<UINT16>> vecFCoverGlyphs;

	FPSTCoverageRule() : FPSTRule(FPSTR_COVERAGE) {}

	~FPSTCoverageRule() {}
};

/*
	Feature Parameters Value RCoverage
*/

class FPSTRevCoverageRule : public FPSTRule
{
public:

	INT32 iAlwaysL;
	INT32 iBCnt;
	INT32 iFCnt;

	std::vector<std::string> psNCovers;
	std::vector<std::vector<UINT16>> vecNCoverGlyphs;

	std::vector<std::string> psBCovers;
	std::vector<std::vector<UINT16>> vecBCoverGlyphs;

	std::vector<std::string> psFCovers;
	std::vector<std::vector<UINT16>> vecFCoverGlyphs;

	std::string sReplacements;
	std::vector<UINT16> vecReplacementGlyphs;

	FPSTRevCoverageRule() : FPSTRule(FPSTR_REV_COVERAGE) {}

	~FPSTRevCoverageRule() {}
};

/*
	Feature Parameters Sub Tables
*/
struct FeatureParamSubTab 
{
	UINT16 uType;
	UINT16 uFormat;
	std::list<LookupSubTable*> vSubTables;
	FeatureParamSubTab* pNext;
	UINT16 uNcCnt;
	UINT16 uBcCnt;
	UINT16 uFcCnt;
	UINT16 uRuleCnt;
	std::vector<std::string> vsNClass;
	std::vector<std::string> vsBClass;
	std::vector<std::string> vsFClass;
	std::vector<FPSTRule*> vRules;
	UINT8 uTicked;
	UINT8 uEffectivelyByGlyphs;
	std::vector<std::string> vsNClassNames;
	std::vector<std::string> vsBClassNames;
	std::vector<std::string> vsFClassNames;

	FeatureParamSubTab() : uType(PST_NULL), uFormat(0), pNext(nullptr), uNcCnt(0), uBcCnt(0), uFcCnt(0),
						   uRuleCnt(0), uTicked(0), uEffectivelyByGlyphs(0) {}

	~FeatureParamSubTab()
	{
		for (size_t i = 0; i < vRules.size(); i++)
			SAFE_DELETE(vRules[i])

		vRules.clear();
	}
};

/*
	Apple State Machine State
*/
struct ASMState
{
	UINT16 iNextState;
	UINT16 iFlags;

	struct
	{
		/* for contextual glyph subs (tag of a nested lookup) */
		OTLookup* pMarkLookup;
		/* for contextual glyph subs */
		OTLookup* pCurLookup;
	} Context;

	struct
	{
		std::string sMarkIns;
		std::string sCurIns;
	} Insert;

	struct
	{
		std::vector<INT16> pKerns;
		INT32 iKCnt;
	} Kern;

	ASMState() : iNextState(0), iFlags(0) {}
};

/* 
	Apple State Machine 
*/
struct AppleStateMachine
{		
	AppleStateMachine* pNext;
	UINT16 uType;
	// Lookup contains feature setting info
	LookupSubTable* pSubTable;
	// 0x8000=>vert, 0x4000=>r2l, 0x2000=>hor&vert
	UINT16 uFlags;
	UINT8 uTicked;

	UINT16 uClassCnt;
	UINT16 uStateCnt;
	std::vector<std::string> vsClasses;
	ASMState* pState;

	AppleStateMachine() : pNext(nullptr), uType(0), pSubTable(nullptr), uFlags(0), 
						  uTicked(0), uClassCnt(0), uStateCnt(0), pState(nullptr) {}
};

/*
	PSContext
*/

struct PSContext 
{
	INT iIsType2;
	INT iPaintType;
	INT iInstanceCount;
	FLOAT fBlendValues[17];
	INT iBlendWarn;

	PSContext() : iIsType2(0), iPaintType(0), iInstanceCount(0), iBlendWarn(0)
	{
		memset(fBlendValues, 0, 17 * sizeof(FLOAT));
	}
};

/*
	PSCharacters
*/
struct PSChars
{
	INT iCnt;
	INT iNext;
	std::vector<std::vector<CHAR>> vKeys;
	std::vector<std::vector<UINT8>> vValues;
	std::vector<INT> vLens;
	//for Type2 strings
	INT iBias;

	PSChars() : iCnt(0), iNext(0), iBias(0) {}
};

/*
	Script Language List
*/
struct ScriptLangList
{
	UINT32 uScript;
	UINT32 uLangs[MAX_LANG];
	std::vector<UINT32> vMoreLangs;
	INT32 iLangCnt;

	ScriptLangList() : uScript(0), iLangCnt(0)
	{
		memset(uLangs, 0, MAX_LANG * sizeof(UINT32));
	}

	ScriptLangList(INT32 iCnt) : uScript(0), iLangCnt(iCnt)
	{
		memset(uLangs, 0, MAX_LANG * sizeof(UINT32));
	}
};

/*
	Feature Script Language List
*/
struct FeatureScriptLangList
{
	UINT32 uFeatureTag;
	GlyphSetEnum eGlyphSet;
	std::vector<ScriptLangList*> vScripts;
	FeatureScriptLangList* pNext;
	// treat the featuretag as a non-mac feature/setting
	UINT uIsMac;

	FeatureScriptLangList() : uFeatureTag(0), eGlyphSet(GSET_NONE), pNext(nullptr), uIsMac(0) {}

	~FeatureScriptLangList()
	{
		for (size_t i = 0; i < vScripts.size(); i++)
			SAFE_DELETE(vScripts[i]);

		vScripts.clear();
	}
};

struct SubRule
{
	UINT32 iOffset;
	INT iGCnt;
	INT iBCnt;
	INT iFCnt;
	INT iCCnt;
	INT iSCnt;
	std::vector<UINT16> vGlyphs;
	std::vector<UINT16> vBGlyphs;
	std::vector<UINT16> vFGlyphs;
	std::vector<UINT16> vClassIndices;
	std::vector<UINT16> vBClassIndices;
	std::vector<UINT16> vFClassIndices;
	std::vector<SeqLookup*> vSeqLookups;

	SubRule() : iOffset(0), iGCnt(0), iBCnt(0), iFCnt(0), iCCnt(0), iSCnt(0) {}
};

struct Rule
{
	UINT32 iOffset;
	INT iSCnt;
	std::vector<SubRule> vSubRules;

	Rule() : iOffset(0), iSCnt(0) {}
};

/*
	Open Type Lookup
*/
struct OTLookup
{
	OTLookup* pNext;
	OTLookupType eLookupType;
	GlyphSetEnum eGlyphSet;

	bool IsGPos() { return eLookupType >= OTL_TYPE_GPOS_START && eLookupType <= OTL_TYPE_KERN_STATE_MACHINE; }
	/* 
		Low order: traditional flags, High order: markset index, 
		only meaningful if pst_usemarkfilteringset set 
	*/
	UINT32 uLookupFlags;
	std::string sLookupName;
	std::vector<FeatureScriptLangList*> vFeatures;
	std::list<LookupSubTable*> vSubTables;
	/* No subtable is used (call SFFindUnusedLookups before examining) */
	UINT32 uUnused : 1;
	/* No subtable is used, and no anchor classes are used */
	UINT32 uEmpty : 1;
	/* Used for ligatures, some get stored */
	UINT32 uStoreInAfm : 1;
	/*  'liga' generally does, but 'frac' doesn't */
	/* Used during opentype generation */
	UINT32 uNeedsExtension : 1;
	/* Used when decomposing kerning classes into kern pairs for older formats */
	//UINT32 uTemporaryKern : 1;
	UINT32 uDefLangChecked : 1;
	UINT32 uDefLangFound : 1;
	UINT32 uTicked : 1;
	//UINT32 uInGpos : 1;
	//UINT32 uInJSTF : 1;
	//UINT32 uOnlyJSTF : 1;
	/* Actual number of subtables we will output */
	INT16 iSubCnt;
	/* Some of our subtables may contain no data */
	/* Some may be too big and need to be broken up.*/
	/* So this field may be different than just counting the subtables */
	/* used during opentype generation */
	INT32 iLookupIndex;
	UINT32 uLookupOffset;
	UINT32 uLookupLength;
	std::string sTempName;

	OTLookup() : pNext(nullptr), eLookupType(OTL_TYPE_UNDEF), uLookupFlags(0), uUnused(0), 
				 uEmpty(0), uStoreInAfm(0), uNeedsExtension(0), uDefLangChecked(0), uDefLangFound(0), uTicked(0),
				 iSubCnt(0), iLookupIndex(0), uLookupOffset(0), uLookupLength(0), eGlyphSet(GSET_NONE) {}

	~OTLookup()
	{
		for (size_t i = 0; i < vFeatures.size(); i++)
			SAFE_DELETE(vFeatures[i]);

		vFeatures.clear();

		for (std::list<LookupSubTable*>::iterator it = vSubTables.begin(); it != vSubTables.end(); ++it)
			SAFE_DELETE(*it);

		vSubTables.clear();
	}
};

/*
	Open Type Font Language
*/
struct OTFLanguage
{
	UINT32 uTag;
	UINT32 uOffset;
	// required feature index. 0xffff for null
	UINT16 uReq;
	INT32 iFCnt;
	std::vector<UINT16> vFeatures;

	OTFLanguage() : uTag(0), uOffset(0), uReq(0), iFCnt(0) {}
};

/*
	Scripts
*/
struct Scripts
{
	UINT32 uOffset;
	UINT32 uTag;
	// the default language is included as a
	INT32 iLangCnt;
	// normal entry with lang tag 'dflt'
	std::vector<OTFLanguage> vLanguages;

	Scripts() : uOffset(0), uTag(0), iLangCnt(0) {}
};

/*
	Feature
*/
struct Feature
{
	UINT32 uOffset;
	UINT32 uTag;
	INT32 iLCnt;
	std::vector<UINT16> vLookups;
	GlyphSetEnum eGlyphSet;

	Feature() : uOffset(0), uTag(0), iLCnt(0), eGlyphSet(GSET_NONE) {}
};

/*
	Lookup
*/
struct Lookup
{
	UINT16 iType;
	/* 
		low order 16 bits: traditional flags, high order: 
		mark filtering set index if pst_usemarkfilteringset 
	*/
	UINT32 iFlags;
	UINT32 uOffset;
	int iSubTabCnt;
	std::vector<INT32> vSubTabOffsets;

	std::vector<OTLookup*> vOTLookups;

	Lookup() : iType(0), iFlags(0), uOffset(0), iSubTabCnt(0) {}

	~Lookup()
	{
		for (size_t i = 0; i < vOTLookups.size(); i++)
			SAFE_DELETE(vOTLookups[i]);

		vOTLookups.clear();
	}
};

/*
	Open Type Font Name
*/
struct OTFName
{
	// windows language code
	UINT16 iLang;
	// utf8
	std::string sName;

	OTFName() : iLang(0) {}
};

/*
	Open Type Font Feature Name
*/
struct OTFFeatName
{
	/* Feature tag */
	UINT32 uTag;
	std::vector<OTFName> vNames;
	OTFFeatName* pNext;
	/* temporary value */
	UINT16 uNid;

	OTFFeatName() : uTag(0), pNext(nullptr), uNid(0) {}
};

struct FeatureLookups 
{
	UINT32 tag;
	std::vector<OTLookup*> vLookups;
	int iFeatureID;		// Initially consecutive, but may be rearranged by sorting
	UINT32 name_param_ptr;
};

struct LangSys
{
	UINT32 lang;
	std::vector<int> vFeatureID;
	int iSameAs;
	INT32 offset;
};

struct ScriptSet 
{
	UINT32 script;
	std::vector<LangSys> vLangSys;
};

// GPOS/GSUB info: the supplementary structure that is filled-in by CFFWriter::DumpTable
// to hold the info for a specific table GPOS or GSUB, it then writes-off
struct GInfo
{
	std::vector<FeatureLookups> vFeatLookups;
	std::vector<ScriptSet> vScripts;
};
