// Header for the OTFDataLoader implementation

#ifndef INC_OTFTOOLS
#define INC_OTFTOOLS

struct SplineChar;

#ifdef CAMELOT_PRODUCT
#include "fntcache.h"
#include "pathtype.h"
#endif

#include "gpos.h"
#include "GlyphSource.h"

#ifndef CAMELOT_PRODUCT

#ifndef GlyphSubstitutionType
typedef enum GlyphSubstitutionType
{
	GSUB_SINGLE = 0,
	GSUB_MULTIPLE,
	GSUB_LIGATURE,
	GSUB_SINGLE_STANDARD,
	GSUB_MULTIPLE_STANDARD,
	GSUB_LIGATURE_STANDARD
};
#endif

/*
Hashing for unordered_map<std::vector<WCHAR>>
*/

template <class T>
inline void CharHashCombine(std::size_t & seed, const T & v)
{
	std::hash<T> hasher;
	seed ^= hasher(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

namespace std
{
	template<> struct hash<std::vector<WCHAR>>
	{
		inline size_t operator()(const std::vector<WCHAR>& v) const
		{
			size_t seed = 13;

			for (auto& pChar : v)
				::CharHashCombine(seed, pChar);

			return seed;
		}
	};
}

// Single / Multiple / Ligature substitution table
// for further use from substitution rules
class SubstitutionTable
{
public:
	UINT16 UID{ 0 };
	GlyphSubstitutionType SubstitutionType;
	// key chars to substitution chars & glyph IDs
	std::unordered_map<std::wstring, std::vector<std::pair<WCHAR, DWORD>>> SubstitutionMap;
};

// basic substitution rule:
// if any string of WCHARs precedes or follows
// then the substitution should be applied
class SubstitutionRule
{
public:
	// prefix vector
	std::vector<std::wstring> Pre;
	// sufix vector
	std::vector<std::wstring> Post;
	// key vector
	std::vector<std::wstring> Keys;
	// unique table ID
	UINT16 SubstitutionTableID{ 0 };
	// priority - if lower then higher priority we have
	UINT16 Priority{ 0 };
	// language set for given rule - if none, it will apply to any language
	std::vector<UINT32> Langs;
};

// max sizes for substitution rules
// will come in hand while constructing pre and post character lists
class RuleMaxSizes
{
public:
	UINT8 Pre{ 0 };
	UINT8 Key{ 0 };
	UINT8 Post{ 0 };
};

using SubstitutionRuleSet = std::unordered_map < GlyphSetEnum, std::pair<std::vector<SubstitutionRule>, RuleMaxSizes> >;
using SubstitutionTableSet = std::unordered_map < UINT16, SubstitutionTable >;

#endif // !CAMELOT_PRODUCT




/*
	Kerning Class
*/
struct KernClass 
{
	/* Count of classes for first and second chars */
	INT32 iFirstCnt = 0;
	INT32 iSecondCnt = 0;
	/* list of a space separated list of char names */
	std::vector<std::string> vecKernFirsts;
	/*  one entry for each class. Entry 0 is null */
	std::vector<std::string> vecKernSeconds;
	/*  and means everything not specified elsewhere */
	LookupSubTable *pSubTable = nullptr;
	/* Temporary value, used for many things briefly */
	UINT16 uKCId = 0;
	/* array of first_cnt*second_cnt entries */
	std::vector<INT16> vecKernOffsets;
	KernClass *pNext = nullptr;
};


/*
	Kerning Pair
*/
struct KernPair
{
	LookupSubTable *pSubTable = nullptr;
	SplineChar *pSC = nullptr;
	INT16 iOff = -1;
	/* temporary value */
	UINT16 uKCId = 0;
};

/*
	Hashing for unordered_map
*/

template <class T>
inline void hash_combine(std::size_t & seed, const T & v)
{
  std::hash<T> hasher;
  seed ^= hasher(v) + 0x9e3779b9 + (seed << 6) + (seed >> 2);
}

namespace std
{
  template<typename S, typename T> struct hash<pair<S, T>>
  {
	inline size_t operator()(const pair<S, T> & v) const
	{
	  size_t seed = 0;
	  ::hash_combine(seed, v.first);
	  ::hash_combine(seed, v.second);
	  return seed;
	}
  };
}

/*
	Value Record
*/
struct ValueRecord
{
	INT16 iXPlacement = 0;
	INT16 iYPlacement = 0;
	INT16 iXAdvance = 0;
	INT16 iYAdvance = 0;
	UINT16 uOffXPlaceDev = 0;
	UINT16 uOffYPlaceDev = 0;
	UINT16 uOffXAdvanceDev = 0;
	UINT16 uOffYAdvanceDev = 0;
};

/*
	Renaming structure
*/

struct Rename
{
	std::string sFrom;
	std::string sTo;
};

/*
	Name List
*/
struct NameList
{
	NameList* pBasedOn;
	std::string sTitle;
	const CHAR*** pUnicode[17];
	NameList* pNext;
	Rename* pRenames;
	INT iUseUnicode;
	std::string sUtf8Name;
};

/*
	Top Dictionaries
*/
struct TopDicts 
{
	INT32 iCFFStart = 0;
	std::string sFontName;
	INT iVersion = 0;
	INT iNotice = 0;
	INT iCopyright = 0;
	INT iFullName = 0;
	INT iFamilyName = 0;
	INT iWeight = 0;
	INT iIsFixedPitch = 0;
	FLOAT fItalicAngle = 0;
	FLOAT fUnderlinePos = 0;
	FLOAT fUnderlineWidth = 0;
	INT iPaintType = 0;
	INT iCharStringType = 0;
	FLOAT fFontMatrix[6];
	INT iFontMatrixSet = 0;
	INT iUniqueID = 0;
	FLOAT fFontBoundingBox[4];
	FLOAT fStrokeWidth = 0;
	INT iXuid[20];
	INT iStringsOff = -1;
	INT iStringsSize = 0;
	//from start of file
	INT iCharsetOff = -1;
	//from start of file
	INT iEncodingOff = -1;
	//from start of file
	INT iCharStringsOff = -1;	
	INT iPrivateSize = 0;
	//from start of file
	INT iPrivateOffset = -1;
	//font index
	INT iSyntheticBase = 0;
	INT iPostScriptCode = 0;
	//synthetic fonts only (whatever they are)
	INT iBaseFontName = 0;
	//Multiple master/synthetic fonts 
	FLOAT fBaseFontBlend[16];
	//Multiple master fonts only
	INT iBlendAxisTypes[17];
	INT iNMasters = 0;
	INT iNAxes = 0;
	FLOAT iWeightVector[17];
	//No description of what this means
	INT iLenBuildCharArray = 0;
	INT iNormalizeDesignVector = 0;	
	INT iConvertDesignVector = 0;
	//CID fonts only
	INT iRosRegistry = 0;
	INT iRosOrdering = 0;
	INT iRosSupplement = 0;
	FLOAT fCidFontVersion = 0;
	INT iCidFontRevision = 0;
	INT iCidFontType = 0;
	INT iCidCount = 0;
	INT iUidBase = 0;
	//from start of file
	INT iFDArrayOff = -1;
	//from start of file
	INT iFDSelectOff = -1;
	INT iSidFontname = 0;
	FLOAT fBlueValues[14];
	FLOAT fOtherBlues[10];
	FLOAT fFamilyBlues[14];
	FLOAT fFamilyOtherBlues[10];
	FLOAT fBlueScale = 0;
	FLOAT fBlueShift = 0;
	FLOAT fBlueFuzz = 0;
	INT iStdHw = 0;
	INT iStdVw = 0;
	FLOAT fStemsNapH[10];
	FLOAT fStemsNapV[10];
	INT iForceBold = 0;
	FLOAT fForceBoldThreshold = 0;
	INT iLanguageGroup = 0;
	FLOAT fExpansionFactor = 0;
	INT iInitialRandomSeed = 0;
	//from start of this private table
	INT iSubrsOff = -1;
	INT iDefaultWidthX = 0;
	INT iNominalWidthX = 0;

	INT iGSubsOff = -1; // global subroutines data
	INT iGSubsSize = 0;

	PSChars Glyphs;
	PSChars LocalSub;
	INT iLocalSubSize = 0;
	std::vector<UINT16> vecTDCharset;

	TopDicts()
	{
		memset(fFontMatrix, 0, 6 * sizeof(FLOAT));
		memset(fFontBoundingBox, 0, 4 * sizeof(FLOAT));
		memset(iXuid, 0, 20 * sizeof(INT));
		memset(fBaseFontBlend, 0, 16 * sizeof(FLOAT));
		memset(iBlendAxisTypes, 0, 17 * sizeof(INT));
		memset(iWeightVector, 0, 17 * sizeof(FLOAT));
		memset(fBlueValues, 0, 14 * sizeof(FLOAT));
		memset(fOtherBlues, 0, 10 * sizeof(FLOAT));
		memset(fFamilyBlues, 0, 14 * sizeof(FLOAT));
		memset(fFamilyOtherBlues, 0, 10 * sizeof(FLOAT));
		memset(fStemsNapH, 0, 10 * sizeof(FLOAT));
		memset(fStemsNapV, 0, 10 * sizeof(FLOAT));
	}
};

/*
	Value Record
*/
struct ValRec
{
	INT16 iXOff = 0;
	INT16 iYOff = 0;
	INT16 iHAdvOff = 0;
	INT16 iVAdvOff = 0;
};

/*
	Positioning Sub Table Pair
*/
struct PSTPair
{
	std::string sPaired;
	ValRec pVR[2];
};

/*
	Positioning Sub Table
*/
struct PosSubTab
{
	UINT uTicked = 1;
	/* Used in afm ligature closure */
	UINT uTemporary = 1;
	PosSubType uType = PosSubType::PST_NULL;
	LookupSubTable* pSubtable = nullptr;
	PosSubTab *pNext = nullptr;
	UINT16 Priority = 0;
	DWORD GlyphID = 0;

	ValRec Pos;
	PSTPair PstPair;

	struct 
	{ 
		std::string sVariant; 
		std::vector<UINT16> vecGlyphs;
	} PSTSubs;

	struct 
	{ 
		std::string sComponents;
		std::vector<UINT16> vecGlyphs;
	} PSTMult, PSTAlt;

	struct 
	{ 
		std::string sComponents; 
		std::vector<UINT16> vecGlyphs;
		SplineChar* pLig; 
	} PSTLig;
		
	/* Ligature caret positions */
	struct 
	{ 
		std::vector<INT16> vecCaretPos;
		INT32 iCnt; 
	} PSTLCaret;
};

struct PFMInfo
{
	UINT uPfmSet : 1;
	UINT uWinAscentAdd : 1;
	UINT uWinDescentAdd : 1;
	UINT uHHeadAscentAdd : 1;
	UINT uHHeadDescentAdd : 1;
	UINT uTypoAscentAdd : 1;
	UINT uTypoDescentAdd : 1;
	UINT uSubSuperSet : 1;
	UINT uPanoseSet : 1;
	UINT uHHeadSet : 1;
	UINT uVHeadSet : 1;
	UINT uHasCodePages : 1;
	BOOL bHasUnicodeRanges;
	UCHAR uPfmFamily = 0;
	INT16 iWeight = 0;
	INT16 iWidth = 0;
	CHAR cPanose[10];
	INT16 iFStype = 0;
	/* from hhea */
	INT16 iLineGap = 0;
	/* from vhea */
	INT16 iVLineGap = 0;	
	INT16 iHHeadAscent = 0;
	INT16 iHHeadDescent = 0;
	INT16 iOS2TypoAscent = 0;
	INT16 iOS2Typodescent = 0;
	INT16 iOS2Typolinegap = 0;
	INT16 iOS2WinAscent = 0;
	INT16 iOS2WinDescent = 0;
	INT16 iOS2SubxSize = 0;
	INT16 iOS2SubySize = 0;
	INT16 iOS2SubxOff = -1;
	INT16 iOS2SubyOff = -1;
	INT16 iOS2SupxSize = 0;
	INT16 iOS2SupySize = 0;
	INT16 iOS2SupxOff = -1;
	INT16 iOS2SupyOff = -1;
	INT16 iOS2StriKeySize = 0;
	INT16 iOS2StriKeyPos = 0;
	CHAR cOS2Vendor[4];
	INT16 iOS2FamilyClass = 0;
	UINT32 uCodePages[2];
	UINT32 uUnicodeRanges[4];

	PFMInfo() : uPfmSet(0), 
		uWinAscentAdd(0),
		uWinDescentAdd(0),
		uHHeadAscentAdd(0),
		uHHeadDescentAdd(0),
		uTypoAscentAdd(0),
		uTypoDescentAdd(0),
		uSubSuperSet(0),
		uPanoseSet(0),
		uHHeadSet(0),
		uVHeadSet(0),
		uHasCodePages(0),
		bHasUnicodeRanges(false)
	{
		memset(cPanose, 0, 10 * sizeof(CHAR));
		memset(cOS2Vendor, 0, 4 * sizeof(CHAR));
		memset(uCodePages, 0, 2 * sizeof(UINT32));
		memset(uUnicodeRanges, 0, 4 * sizeof(UINT32));
	}
};

/*
	Spline Character
*/
struct SplineChar
{
	INT iTTFGlyph = 0;
	/* 
		Only used when reading in a type1 font
		Or an otf font where it is the subr number of a refered character
		or a ttf font without bit 1 of head.flags set
		or (once upon a time, but no longer) a ttf font with vert metrics where it is the ymax value 
		when we had a font-wide vertical offset
		or when generating morx where it is the mask of tables in which the glyph occurs
	*/
	BOOL bLigCaretCntFixed = FALSE;

	INT16 iLSideBearing = 0;
	INT16 iWidth = 0;
	INT16 iVWidth = 0;

	/* Original position in the glyph list */
	INT32 iOrigPos = 0xffff;
	INT32 iLayersCnt = 0;
	INT32 iUnicodeEnc = 0;
	UINT16 uGlyphClass = 3;
	std::string sName;

	std::vector<PosSubTab*> vecPosSub;
	std::vector<KernPair> vecVerticalKernPairs;
	std::vector<KernPair> vecKernPairs;

	SplineChar(){}

	~SplineChar()
	{
		for (auto &pPS : vecPosSub)
			SAFE_DELETE(pPS);

		vecPosSub.clear();
	}
};

/********************************************************************************************
>	struct LigatureDataSet

	Author:     Greg Watroba
	Created:    6/10/16
	Purpose:    Structure for holding LigatureData
********************************************************************************************/

struct LigatureDataSet
{
	std::wstring vecLigComponents;
	LigatureSubsetFlag LigType{ LIG_SUB_NONE };
	WCHAR LigUnicode{ 0 };
	UINT16 Priority{ 0 };
	DWORD GlyphID{ 0 };

	bool operator == (const LigatureDataSet& A);
};

/********************************************************************************************
>	struct OpenTypeFontData

	Author:     Greg Watroba
	Created:    24/11/15
	Purpose:    Structure for holding Open Type Font Data for further use.
				Partially paraphrased True Type Font Data used in FontForge
********************************************************************************************/

struct OpenTypeFontData
{
	BOOL bBadCFFData;
	BOOL bBadOpenTypeData;
	BOOL bBadGlyphData;
	BOOL bLookupsInitialized;
	BOOL bOnlyStrikes;
	BOOL bIsOrder2Type;
	BOOL bIsTrueTypeCollection;
	BOOL bAlreadyFoundBadGlyphData;
	BOOL bBadBoundingBoxData;
	BOOL bBadOS2Data;
	BOOL bUseTypoMetrics;
	BOOL bWeightWidthSlopeOnly;

	WORD iFontTypeFace;

	OTFlags OTFlags;
	UINT32 uBitmapDataStart;
	UINT32 uBitmapDataLength;
	UINT32 uBitmapLocationStart;
	UINT32 uBitmapLocationLength;
	UINT32 uBSLNStart;
	UINT32 uCFFStart;
	UINT32 uCFFLength;
	UINT16 uDesignSize;
	UINT16 uDesignRangeBottom;
	UINT16 uDesignRangeTop;
	UINT16 uFontStyleId;
	UINT32 uFontRevision;
	UINT32 uFontTimeStamp;
	//grid fitting data, scan conversion
	UINT32 uGASPStart;
	UINT32 uGDEFStart;
	UINT32 uGDEFLength;
	//horizontal metrics
	UINT32 uHMetricsStart;
	//justification data
	UINT32 uJSTFStart;
	UINT32 uJSTFLength;
	UINT32 uKernStart;
	UINT32 uKernLength;
	UINT32 uTyp1Start;
	UINT32 uTyp1Length;
	UINT32 uVHead;
	UINT32 uVMetrics;
	UINT32 uPostScriptStart;
	UINT32 uOS2Start;
	UINT32 uFeatureStart;
	UINT32 uLcarStart;
	UINT32 uOPBDStart;
	UINT32 uPropStart;
	UINT32 uCVTStart;
	UINT32 uCVTLen;
	UINT32 uPrepStart;
	UINT32 uPrepLength;
	UINT32 uFPGMStart;
	UINT32 uFPGMLength;
	UINT32 uPFEDStart;
	UINT32 uTexStart;
	UINT32 uBDFStart;
	UINT32 uCvarStart;
	UINT32 uCvarLength;
	UINT32 uAvarStart;
	UINT32 uAvarLength;
	UINT uOptimisedForClearType;
	UINT uApplyLSB;
	UINT uIndexToLocIsLong;
	UINT uOS2Version;
	UINT uAvgWidth;
	INT iEmSize;
	INT iAscent;
	INT iDescent;
	INT iMacStyle;
	INT iWidthCnt;
	INT iAdvanceWidthMax;
	INT32 iBaseStart;
	INT32 iGposStart;
	INT32 iGposLength;
	INT32 iGsubStart;
	INT32 iGsubLength;
	INT32 iLastSizePos;
	INT32 iFeatureCnt;
	INT32 iLookupCnt;
	INT32 iPositioningSizeBounds;
	INT32 iGDefBounds;
	INT32 iNumTables;
	INT32 iGlyphStart;
	INT32 iGlyphLength;
	INT32 iGlyphLocationStart;
	INT32 iGlyphLocationLength;
	INT32 iGlyphCnt;
	INT32 iEncodingStart;
	INT32 iHeadStart;
	INT32 iHHeaStart;
	INT32 iMaxPStart;
	INT32 iMaxPLen;
	INT32 iMathStart;
	INT32 iMathLength;
	INT32 iMortStart;
	INT32 iMorxStart;
	INT32 iGVarStart;
	INT32 iGVarLength;
	INT32 iFVarStart;
	INT32 iFVarLength;
	INT32 iCopyrightStart;

	INT iMarkClassCnt;
	INT iMarkSetCnt;
	std::string strCFFFontName; // CFF font name taken right after the CFF header
	std::vector<std::string> vMarkClasses;
	std::vector<std::string> vMarkClassNames;
	std::vector<std::string> vMarkSets;
	std::vector<std::string> vMarkSetNames;

	std::vector<TopDicts*> vpTopDicts; // CFF top dictionaries
	std::vector<TopDicts*> vpSubDicts; // CID sub dicts
	std::vector<Lookup> vGPOSLookups;
	std::vector<Lookup> vGSUBLookups;

	std::vector<UINT8> vFDSelect;
	std::unordered_map<UINT16, GlyphSetEnum> mapLookupToGSUBFeature;
	std::unordered_map<UINT16, GlyphSetEnum> mapLookupToGPOSFeature;

	UINT16 LookupSubtableUID;

	SubstitutionRuleSet SubRuleSet;
	RuleMaxSizes TotalMaxSizes;

	std::vector<LookupSubTable*> vecLookupSubTables;
	std::vector<UINT16> vecNeededLookups;
	std::set<UINT16> setStandardSubLookups;

	UnicodeIterp eUniInterp;
	/* x,yMin x,yMax*/
	INT iFontBoundingBoxes[4];
	HDC hDC;
	std::unordered_map<GlyphSetEnum, OTFFeatName> mapFeatNames;
	std::unordered_map<GlyphSetEnum, std::unordered_map<LANGID, std::string>> mapUserFriendlyGlyphSetNames;
	std::vector<OTFName> vFontStyleNames;
	std::vector<SplineChar*> vChars;
	std::list<KernClass*> vVerticalKernClasses;
	std::vector<KernClass*> vKernClasses;
	std::vector<CHAR> vInUse;
	std::list<FeatureParamSubTab*> lstFPST;
	std::string sFontName;
	std::string sOrdering;
	PFMInfo PfmInfo;
	//we need to initialize and resize these vectors only once
	//thus the static type
	static std::vector<wchar_t> s_vCharSet;
	static std::vector<UINT16> s_vGlyphIndices;
	static std::vector<UINT32> s_vScriptsToSupport;
	static std::vector<GlyphSetEnum> s_vSupportedSubstitutions;
	static std::vector<GlyphSetEnum> s_vSupportedRules;

	OpenTypeFontData();
	~OpenTypeFontData() { ClearData(); }

	void ClearData();
};

/********************************************************************************************
>	class OTFDataLoader

	Author:     Greg Watroba
	Created:    23/11/15
	Purpose:    Class that handles data connected with Open Type Fonts.
				First and main purpose so far is to parse GPOS data from otf font file.
********************************************************************************************/

class InstalledFontItem;
struct MillipointKerningPair;

class OTFDataLoader
{
public:
	enum ParseOptions { keDefault = 0, keKeepGlyphData = 1 }; // add other options as bit fields so we can combine them
private:

	USHORT ParseUShortFromPos(const BYTE** pPos);
	LONG ParseLongFromPos(const BYTE** pPos);
	UCHAR ParseCharFromPos(const BYTE** pPos);
	INT Parse3ByteFromPos(const BYTE** pPos);
	INT32 ParseOffsetFromPos(const BYTE** pPos, INT& iOffsetSize);
	void MatMultiply(FLOAT fMat1[6], FLOAT fMat2[6], FLOAT fMatTo[6]);
	std::string OTFDataLoader::GetSID(INT iSID, std::vector<std::string>& vsStrings, INT iSCnt);
	void MapCharEncodings();

	void ParseTTFSizeParameters(const BYTE* pPosData, INT32 iBrokenOffsetPos, INT32 iCorrectOffsetPos);
	void ParseTTFFeatureNameParameters(const BYTE* pPosData, INT32 iPos, UINT32 uTag, const GlyphSetEnum& eGSet);
	void ParseTTFOS2Metrics(const BYTE* pPosData);
	void ParseTTFPostNames(const BYTE* pPosData);
	void ParseTTFGDef(const BYTE* pPosData);
	void ParseTTFProp(const BYTE* pPosData);
	void ParseTTFLcar(const BYTE* pPosData);
	void TagLookupsWithFeature(UINT32& uScriptTag, UINT32& uLangTag, BOOL bRequiredFeature, Feature* pFeature, std::vector<Lookup>* pvLookups);
	void AppendScriptLangOnFeatureList(FeatureScriptLangList* pFSL, UINT32& uScriptTag, UINT32& uLangTag);
	void AppendScriptLangOnFeatureListIteratively(FeatureScriptLangList* pFSL, UINT32& uScriptTag, UINT32& uLangTag);
	void FindAllLangEntries(std::vector<OTFName>* pvOTFNames, const BYTE* pPosData, INT32 iID);
	void SwapBytePairs(BYTE* pDestBytes, const BYTE* pSrcBytes, INT32& iBytesToSwap);
	void ParseValueRecord(ValueRecord* pValueRecord, INT32& iValueFormat, const BYTE** pLookupData);
	void ParseCoverageTable(std::vector<UINT16>* pvGlyphs, const BYTE* pPosData, INT32 iCoverageOffset);
	void AddPairPos(UINT16& uGlyph1, UINT16& uGlyph2, Lookup* pLookup, LookupSubTable* pSubTable, ValueRecord* pValRec1, ValueRecord* pValRec2);
	BOOL AddKernPair(UINT16& uGlyph1, UINT16& uGlyph2, INT16& iOffset, Lookup* pLookup, LookupSubTable* pSubTable, INT32& iIsVertical);
	void GetClassDefinitionTable(std::vector<UINT16>* pvClass, const BYTE* pPosData, INT32 iClassDefOffset);
	void ClassToNames(std::vector<std::string>* pvNames, INT32& iClassCnt, std::vector<UINT16>* pvClass, INT32& iGlyphCnt);
	void CoverageMinusClasses(std::string* sName, std::vector<UINT16>* pvGlyphs, std::vector<UINT16>* pvClass);
	void OTFDataLoader::CFFInfoFillUp(TopDicts* pTopDicts, std::vector<std::string>& vsStrings, INT& iStringCnt);
	SplineChar* OTFDataLoader::PSCharStringToSplines(std::vector<UINT8>& vType1, INT iLen, PSContext* pContext, PSChars* pCharSub, PSChars* pGSubs, std::string sName);

	void ParseGPOSSimplePos(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable);
	void ParseGPOSKernSubTables(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable);
	void ParseGPOSCursiveSubTable(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable);
	void ParseGPOSMarkSubTable(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable);
	void ParseGPOSContextSubTable(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, std::vector<Lookup>* pvLookups);
	void ParseGPOSChainingSubTable(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, std::vector<Lookup>* pvLookups);
	void ParseGPOSExtensionSubTable(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, std::vector<Lookup>* pvLookups);

	void ParseGSUBSimpleSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType);
	void ParseGSUBMultipleSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType);
	void ParseGSUBLigatureSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType);
	void ParseGSUBContextSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups);
	void ParseGSUBContextSubTableType1(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS);
	void ParseGSUBContextSubTableType2(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS);
	void ParseGSUBContextSubTableType3(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS);
	void ParseGSUBChainSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups);
	void ParseGSUBChainSubTableType1(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS);
	void ParseGSUBChainSubTableType2(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS);
	void ParseGSUBChainSubTableType3(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS);
	void ParseGSUBExtensionSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType, std::vector<Lookup>* pvLookups);
	void ParseGSUBReverseChainSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, GsubInUseType eType);

	void ParseGposScripts(std::vector<Scripts>* pvScripts, const BYTE* pPosData, const USHORT& iScriptsOffset, BOOL bIsGPOS);
	void ParseGposFeatures(std::vector<Feature>* pvFeatures, const BYTE* pPosData, const USHORT& iFeaturesOffset, BOOL bIsGPOS);
	void ParseGposLookups(std::vector<Lookup>* pvLookups, const BYTE* pPosData, const USHORT& iLookupsOffset, BOOL bIsGPOS);
	void TagLookupsWithScript(std::vector<Scripts>* pvScripts, std::vector<Feature>* pvFeatures, std::vector<Lookup>* pvLookups);
	void GPOSLookupSwitch(const BYTE* pLookupData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, std::vector<Lookup>* pvLookups);
	void GSUBLookupSwitch(const BYTE* pGsubData, INT32& iSubTableOffset, Lookup* pLookup, LookupSubTable* pSubTable, std::vector<Lookup>* pvLookups, GsubInUseType eType);
	void NameOTLookup(std::vector<OTLookup*>* pvOTLookups);
	void ProcessSubLookups(std::vector<Lookup>* pvLookups, SeqLookup* pSeqLookup, BOOL bGPOS);

	void ReadCFFFontNames(std::vector<std::string>* pvsFontNames, const BYTE** pCFFPos, INT* piCnt);
	void ReadCFFTopDicts(std::vector<TopDicts*>* vpTopDicts, const BYTE** pDictPos, std::vector<std::string>* pvsFontNames, INT32 iCFFOffset, TopDicts* pParentDict);
	void ReadCFFSubs(const BYTE** pCFFPos, PSChars* GSubs);
	void ReadCFFPrivate(const BYTE* pFontData, TopDicts* pTopDicts);
	void ReadCFFSet(const BYTE* pFontData, TopDicts* pTopDicts);
	void ReadCFFFigure(TopDicts* pTopDicts, std::vector<std::string>& vsStrings, INT& iStringCnt, PSChars* pGSubs);
	void ReadFDSelect(std::vector<UINT8>* pvFDSelect, const BYTE** pFDPos, INT32& iGlyphCnt);
	void CIDFigure(TopDicts* pTopDicts, std::vector<std::string>& vsStrings, INT& iStringCnt, PSChars* pGSubs, std::vector<TopDicts*>& vpSubDicts, std::vector<UINT8>& pvFDSelect);
	void ReadCFFEncoding(const BYTE* pFontData, TopDicts* pTopDicts, std::vector<std::string>& vsStrings, INT& iStringCnt);
	void SplineCharConvertToOrder2Type(SplineChar* pSC);
	std::string ReadGlyphName(CHAR* pBuffer, INT iUnicode, UnicodeIterp eUniInterp, NameList* pThisFont);

	SplineChar* ParseTTFGlyph(const BYTE* pFontData, UINT32 iStart, UINT32 iEnd, UINT32 iGid);
	void ParseTTFSimpleGlyph(const BYTE* pFontData, SplineChar* pSC, INT& iPathCnt, INT iGlyphBoundingBox[4]);
	void ParseTTFCompositeGlyph(const BYTE* pFontData, SplineChar* pSC, INT iEndOffset);
	void ReadCFFTopSingleDict(TopDicts** pTopDict, const BYTE** pDictPos, std::string* sFontName, INT iLen);
	void SkipCFF2Thing(const BYTE** pDictPos);
	INT ReadCffThing(const BYTE**pDictPos, INT* iStackValue, FLOAT* fStackValue, INT* iOperatorValue);
	CHAR* AddNibble(CHAR* pChar, int iNib);
	INT ClassFindCnt(std::vector<UINT16>& vClass, INT iTotalOffset);
	std::string GlyphsToNames(std::vector<UINT16>* pvGlyphs, BOOL bMakeUnique);

	BOOL DefaultLangTagInOneScriptList(ScriptLangList* pSLL);
	ScriptLangList* DefaultLangTagInScriptList(std::vector<ScriptLangList*>& pvScripts, BOOL bDefault);
	void TagFullName(std::string* pStrFullName, GlyphSetEnum* pTagEnum, UINT32 uTag, UINT32 uIsMac, BOOL bOnlyIfKnown);
	std::string SuffixFromTags(std::vector<FeatureScriptLangList*>* pvFeatureList);
	
	void ParseFontHeader(const BYTE* pFontData);
	void ParseFontPreglyphData(const BYTE* pFontData);
	void ParseGposGsubData(const BYTE* pPosData, const DWORD& iGPOSTableSize, BOOL bIsGPOS, GsubInUseType eType);

	void ProcessSubstitutionRulesData();
	void ProcessGlyphRule(FPSTGlyphRule* pRule, const GlyphSetEnum& eGlyphSet, UINT16 uPriority, std::vector<UINT32>& vLangs);
	void ProcessClassRule(FPSTClassRule* pRule, const GlyphSetEnum& eGlyphSet, UINT16 uPriority, std::vector<UINT32>& vLangs);
	void ProcessCoverageRule(FPSTCoverageRule* pRule, const GlyphSetEnum& eGlyphSet, UINT16 uPriority, std::vector<UINT32>& vLangs);
	void ProcessRevCoverageRule(FPSTRevCoverageRule* pRule, const GlyphSetEnum& eGlyphSet, UINT16 uPriority, std::vector<UINT32>& vLangs);

	void ProcessSubstitutionData();
	
	std::list<LookupSubTable*> GetSubtableList(FPSTRule* pRule);
	BOOL AllGlyphsInContainerExist(std::vector<UINT16>& srcContainer);
	void GetCharContainerFromGlyphs(std::vector<std::wstring>* pOutResult, std::vector<std::vector<UINT16>>& srcContainer);

	void SetupLigatureData(std::list<LigatureDataSet>* pLigatureList, VFontBase* pFontBase);
	void SetupStylisticSetData(std::array<std::list<StylisticSetEntry>, 20>* pStylisticSets, VFontBase* pFontBase);

	void SetupSubstitutionRules(VFontBase* pFontBase);
	void SetupSubstitutionTables(VFontBase* pFontBase);

	BOOL m_bOTFDataParsingFinished{ FALSE };

public:

	void ParseHeadData(const BYTE* pFontData);
	void ParseHHeaData(const BYTE* pFontData);
	void ParseMaxpData(const BYTE* pFontData);
	void ParseTTFGlyphs(const BYTE* pFontData);
	void ParseTTFNames(const BYTE* pFontData);
	void ParseTTFEncodings(const BYTE* pFontData, INT iJustInUse);
	void ParseTTFGsubUsed(const BYTE* pFontData);
	void ParseOTFMathUsed(const BYTE* pFontData);
	void ParseTTFMortUsed(const BYTE* pFontData);
	void ParseCFFGlyphs(const BYTE* pFontData, bool bKeepGlyphData = false);
	void ParseTyp1Glyphs(const BYTE* pFontData);
	void ParseCopyrightsData(const BYTE* pFontData);

	OpenTypeFontData OTFData;
	void ParseFontData(HDC hDC, VFontBase* pBase, ParseOptions options = ParseOptions::keDefault);
	void SetKernData(INT* iKernCount, MillipointKerningPair** pKernArray, INT* iEmSize);
	void SetSubstitutionData(VFontBase* pBase);

	const BOOL DataParsingFinished() { return m_bOTFDataParsingFinished; }
};

#endif
