/********************************************************************************************

>	reader.cpp

Author:		Paul Gruzdev
Created:	11/01/2017
Purpose:	SFNT: SFNT container format loading routines.

********************************************************************************************/

#include "stdafx.h"
#include "sfnt.h"
#include "ttf.h"
#include "matrix.h"
#include "otfTools.h"
// Enables improved memory tracking. Must be defined after CC_IMPLEMENT_DYNCREATE macros.
#ifdef CAMELOT_PRODUCT
#define new CAM_DEBUG_NEW
#define XaraMatrix Matrix
#endif

namespace SFNT
{

extern bool DecompressData(PBYTE pbyData, DWORD len, DWORD uncomplen, std::vector<BYTE>& out);

static LONG ttf2xar_l(double val)
{
	double x = val*g_fXaraEM / g_iEM;
	return (LONG)x; // rint(x);
}

static DocCoord ttf2xar_dp(DPoint& dp)
{
	double x = dp.x*g_fXaraEM / g_iEM;
	double y = dp.y*g_fXaraEM / g_iEM;
	//return DocCoord((LONG)rint(x), (LONG)rint(y));
	return DocCoord((LONG)(x), (LONG)(y));
}

static void FindBezierBounds(RECT& rBounds, const DocCoord P0, const DocCoord P1, const DocCoord P2, const DocCoord P3)
{
	float A = (float)P3.x - 3 * (float)P2.x + 3 * (float)P1.x - (float)P0.x;
	float B = 3 * (float)P2.x - 6 * (float)P1.x + 3 * (float)P0.x;
	float C = 3 * (float)P1.x - 3 * (float)P0.x;
	float D = (float)P0.x;

	float E = (float)P3.y - 3 * (float)P2.y + 3 * (float)P1.y - (float)P0.y;
	float F = 3 * (float)P2.y - 6 * (float)P1.y + 3 * (float)P0.y;
	float G = 3 * (float)P1.y - 3 * (float)P0.y;
	float H = (float)P0.y;

	float x, y;
	float xMin = (float)INT_MAX;
	float yMin = (float)INT_MAX;
	float xMax = 0;
	float yMax = 0;

	for (float t = 0.0f; t <= 1.0f; t += 0.01f)
	{
		x = A * t * t * t + B * t * t + C * t + D;
		if (x < xMin)
			xMin = x;
		if (x > xMax)
			xMax = x;
		y = E * t * t * t + F * t * t + G * t + H;
		if (y < yMin)
			yMin = y;
		if (y > yMax)
			yMax = y;
	}

	rBounds.left = (LONG)xMin;
	rBounds.top = (LONG)yMin;
	rBounds.right = (LONG)xMax;
	rBounds.bottom = (LONG)yMax;
}

static void InflateRectToContainPoint(RECT& r, LONG x, LONG y)
{
	if (x < r.left)
		r.left = x;
	else if (x > r.right)
		r.right = x;

	if (y < r.top)
		r.top = y;
	else if (y > r.bottom)
		r.bottom = y;
}

/******************************************************************************
Function:    FindPathBounds
Author:      Paul Gruzdev
Purpose:		Calc path bounding box given verbs and points pointers
Returns:
Exceptions:
Parameters:
Notes:
******************************************************************************/
RECT FindPathBounds(PBYTE pVerb, DocCoord* pCoord, ULONG ulCount)
{
	RECT bounds;
	memset(&bounds, 0, sizeof(RECT));

	if (!ulCount)
		return bounds;

	bounds.left = bounds.right = pCoord->x;
	bounds.top = bounds.bottom = pCoord->y;

	DocCoord dcLast;
	std::vector<DocCoord> dcBezier;

	for (ULONG i = 0; i < ulCount; i++)
	{
		BYTE byVerb = pVerb[i];
		DocCoord& dc = pCoord[i];

		if (byVerb == PT_BEZIERTO || dcBezier.size() == 2)
		{
			if (dcBezier.size() == 2) // we already stored two control points, adjust our bb
			{
				RECT rBezier;
				FindBezierBounds(rBezier, dcLast, dcBezier[0], dcBezier[1], dc);
				UnionRect(&bounds, &bounds, &rBezier);

				dcBezier.clear();
				dcLast = dc;
			}
			else
				dcBezier.push_back(dc);
		}
		else
		{
			InflateRectToContainPoint(bounds, dc.x, dc.y);
			dcLast = dc;
		}
	}

	return bounds;
}

/******************************************************************************
   Function:	ReadString
   Author:		Paul Gruzdev
   Purpose:		Read string from the 'name' table
   Returns:		std::wstring
   Parameters:	CCFile* pFile
   Parameters:	ULONG offset
   Parameters:	ULONG len
   Parameters:	int platform
   Notes:       
******************************************************************************/
std::wstring ReadString(CCFile* pFile, ULONG offset, ULONG len, int platform)
{
	USES_CONVERSION;
	long pos = pFile->tell();
	std::wstring str;

	pFile->seek(offset);

	if (platform == 1) // Mac
	{
		// Mac is screwy, there are several different variants of MacRoman
		//  depending on the language, they didn't get it right when they
		//  invented their script system
		for(ULONG i = 0; i < len; ++i)
		{
			char ch;
			pFile->read(&ch, 1);
			str += (wchar_t)((BYTE)ch);
		}
	} 
	else if (platform == 3 || platform == 0)
	{
		for (ULONG i = 0; i < len/2; ++i)
		{
			char ch;
			pFile->read(&ch, 1);
			wchar_t wch = ch << 8;
			pFile->read(&ch, 1);
			str += wch | (wchar_t)ch;
		}
	} 

	pFile->seek(pos);
	return str;
}

/******************************************************************************
   Function:	EncodingFromPlatSpec
   Author:		Paul Gruzdev
   Purpose:		Get the encoding name from the platform ID and platform specific ID
   Returns:		const char*
   Parameters:	int platform
   Parameters:	int specific
   Notes:       
******************************************************************************/
const char* EncodingFromPlatSpec(int platform, int specific)
{
	const char* enc = "Custom";

	if (platform == 0)
	{
		enc = "Unicode";
		if (specific == 4)
			enc = "UnicodeFull";
	} 
	else if (platform == 1)
	{
		if (specific == 0)
			enc = "Mac";
		else if ( specific == 1 )
			enc = "Sjis";
		else if ( specific == 2 )
			enc = "Big5hkscs";		// Or should we just guess big5? Both are wrong sometimes
		else if ( specific==3 )
			enc = "EUC-KR";
		else if ( specific==25 )
			enc = "EUC-CN";
	} 
	else if ( platform == 2 ) 
	{		// obsolete
		if ( specific == 0 )
			enc = "ASCII";
		else if ( specific==1 )
			enc = "Unicode";
		else if ( specific==2 )
			enc = "ISO8859-1";
	} 
	else if ( platform==3 ) 
	{
		if ( specific==1 || specific==0 )	// symbol (sp=0) is just unicode (PUA)
			enc = "Unicode";
		else if ( specific==2 )
			enc = "Sjis";
		else if ( specific==3 )
			enc = "EUC-CN";
		else if ( specific==4 )
			enc = "Big5hkscs";
		else if ( specific==5 )
			enc = "EUC-KR";
		else if ( specific==6 )
			enc = "Johab";
		else if ( specific==10 )
			enc = "UnicodeFull";
	} 
	else if (platform == 7)
	{
		// Used internally in freetype, but
		// there's no harm in looking for it
		// even if it never happens.
		if (specific == 0)		
			enc = "AdobeStandard";
// 		else if (specific == 1)
// 			;// adobe_expert
// 		else if ( specific==2 )
// 			; // adobe_custom 
	}

	return enc;
}

bool SFNTReader::CalcBBox(DocCoord* pCoord, BYTE* pVerb, ULONG ulCount, RECT& rc)
{
	rc = FindPathBounds(pVerb, pCoord, ulCount);
	return true;
}

/******************************************************************************
   Function:	TTFGetFontName
   Author:		Paul Gruzdev
   Purpose:		Reliably read the font name from the SFNT file
   Returns:		std::wstring
   Parameters:	CCFile* pFile
   Parameters:	LONG offset
   Parameters:	LONG off2
   Notes:       Search either Windows (name=4) or PostScript (name=6) font name
				in the 'name' table of the ttc or ttf. If not foudn, returns
				name=1 family name.
******************************************************************************/
std::wstring SFNTReader::TTFGetFontName(CCFile* pFile, LONG offset, LONG off2, bool bPostScript)
{
	std::wstring str;
	LONG nameoffset = 0;

	pFile->seek(offset);
	/* version = */ getlong(pFile);
	size_t num = (size_t)getushort(pFile);
	/* srange = */ getushort(pFile);
	/* esel = */ getushort(pFile);
	/* rshift = */ getushort(pFile);

	for (size_t i = 0; i < num; ++i)
	{
		LONG tag = getlong(pFile);
		/* checksum = */ getlong(pFile);
		LONG tableoffset = off2 + getlong(pFile);
		/* length = */ getlong(pFile);
		if (tag == CHR('n', 'a', 'm', 'e'))
		{
			nameoffset = tableoffset;
			break;
		}
	}

	if (nameoffset == 0)
		return str;

	pFile->seek(nameoffset);
	return GetFontName(pFile, bPostScript, nameoffset);
}

std::wstring SFNTReader::GetFontName(CCFile* pFile, bool bPostScript, LONG nameOffset)
{
	return GetNameString(pFile, bPostScript ? 6 : 4, nameOffset);
}

/******************************************************************************
   Function:	GetNameString
   Author:		Paul Gruzdev
   Purpose:		Read a string from the 'name' table given it's index
   Returns:		std::wstring
   Parameters:	CCFile * pFile
   Parameters:	int nameIndex
   Parameters:	LONG nameOffset
   Notes:       
******************************************************************************/
std::wstring SFNTReader::GetNameString(CCFile* pFile, int nameIndex, LONG nameOffset)
{
	pFile->seek(nameOffset); // First of all seek to the start of the 'name' table
	std::wstring str;
	int plat, spec, lang, name, len, off, val;
	int fullval, fullstr, fulllen, famval, famstr, famlen;
	int fullplat, fullspec, fulllang, famplat, famspec, famlang;
	int locale = 0x409;
	int maclang = 0x409;

	/* format = */ getushort(pFile);
	int num = getushort(pFile);
	LONG stringoffset = nameOffset + getushort(pFile);
	fullval = famval = 0;

	for (size_t i = 0; i < (size_t)num; ++i)
	{
		plat = getushort(pFile);
		spec = getushort(pFile);
		lang = getushort(pFile);
		name = getushort(pFile);
		len = getushort(pFile);
		off = getushort(pFile);

		std::string strEnc = EncodingFromPlatSpec(plat, spec);
		if (strEnc.empty())
			continue;
		bool bCustom = strEnc == "Custom";
		val = 0;
		if (plat == 3 && !bCustom && lang == locale)
			val = 15;
		else if (plat == 3 && !bCustom && (lang & 0xff) == (locale & 0xff))
			val = 14;
		else if ((plat == 0 || plat == 1) && !bCustom && lang == maclang)
			val = 13;
		// Ok, that didn't work, how about an English name?
		else if (plat == 3 && !bCustom && lang == 0x409)
			val = 12;
		else if (plat == 3 && !bCustom && (lang & 0xff) == 0x09)
			val = 11;
		else if ((plat == 0 || plat == 1) && !bCustom && lang == 0)
			val = 10;
		// failing that I'll take what I can get
		else if (!bCustom)
			val = 1;

		if (name == nameIndex && val > fullval)
		{
			fullval = val;
			fullstr = off;
			fulllen = len;
			fullplat = plat;
			fullspec = spec;
			fulllang = lang;
			if (val == 12)
				break;
		} 
		else if (name == 1 && val > famval)
		{
			famval = val;
			famstr = off;
			famlen = len;
			famplat = plat;
			famspec = spec;
			famlang = lang;
		}
	}

	if (fullval == 0) 
	{
		if (famval == 0)
			return str;

		fullstr = famstr;
		fulllen = famlen;
		fullplat = famplat;
		fullspec = famspec;
		fulllang = famlang;
	}

	return ReadString(pFile, stringoffset + fullstr, fulllen, fullplat);
}

/******************************************************************************
   Function:	NamesReadTTF
   Author:		Paul Gruzdev
   Purpose:		Read a set of font names from TTC file, or a single name from SFNT
   Returns:		bool
   Parameters:	char * filename
   Parameters:	std::vector<std::wstring> & names
   Notes:       
******************************************************************************/
bool SFNTReader::NamesReadTTF(CCFile* ttf, std::vector<std::wstring>& names, std::vector<ULONG>& offs)
{
	std::vector<ULONG> offsets;

	long version = getlong(ttf);
	if (version == CHR('t','t','c','f')) 
	{
		/* TTCF version = */ getlong(ttf);
		size_t cnt = (size_t)getlong(ttf);
		offsets.reserve(cnt);

		for (size_t i = 0; i < cnt; ++i)
			offsets.push_back((ULONG)getlong(ttf));

		for (size_t i = 0; i < cnt; ++i)
		{
			std::wstring str = TTFGetFontName(ttf,offsets[i], 0);
			if (!str.empty())
			{
				names.push_back(str);
				offs.push_back(offsets[i]);
			}
		}
	}
	else 
	{
		std::wstring str = TTFGetFontName(ttf,0,0);
		if (!str.empty())
		{
			names.push_back(str);
			offs.push_back(0);
		}
	}

	return names.size() > 0;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Now goes the SFNTReader class

/******************************************************************************
   Function:	ReadSFNT
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		BOOL
   Notes:       
******************************************************************************/
BOOL SFNTReader::ReadSFNT(CCFile* pFile)
{
	bool bWOFF = false;
	pFile->seek(0);
	long lFlavour = getlong(pFile);

	if (lFlavour == CHR('w', 'O', 'F', 'F')) {
		lFlavour = getlong(pFile);
		bWOFF = true;
	}

	switch(lFlavour)
	{
	case CHR('O', 'T', 'T', 'O'):
		m_format = keOpenType;
		break;
	case 0x00010000:
		m_format = keTrueType;
		break;
	default:
		m_format = keUndefined;
		return FALSE;
	}

	if (bWOFF) getlong(pFile); // skip total file size
	int numTables = getushort(pFile);
	getushort(pFile);
	getushort(pFile);
	getushort(pFile);

	if (bWOFF) pFile->seek(pFile->tell() + 6*4); // skip 6 long values

	for (int i = 0; i < numTables; i++)
	{
		long compressedLength = 0;
		DWORD tag = getlong(pFile);
		if(!bWOFF) getlong(pFile);
		long offset = getlong(pFile);
		if (bWOFF) compressedLength = getlong(pFile);
		long length = getlong(pFile);
		if (bWOFF) getlong(pFile); // skip checksum in woff
		int pos = pFile->tell();
		pFile->seek(offset);
		CCMemFile* pMem = new CCMemFile(NULL, 0, CCMemWrite);
		PBYTE buf = new BYTE[bWOFF ? compressedLength : length];
		pFile->read(buf, bWOFF ? compressedLength : length);
		pFile->seek(pos);

		if (bWOFF && compressedLength != length) {
			std::vector<BYTE> data;
			DecompressData(buf, compressedLength, length, data);
			pMem->write(&data[0], length);
		}
		else
			pMem->write(buf, length);

		pMem->ChangeFileAccess(CCMemRead);
		delete []buf;
		m_tableMap[tag] = pMem;
	}

	return TRUE;
}

/******************************************************************************
   Function:	ReadSFNT
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		BOOL
   Notes:       
******************************************************************************/
BOOL SFNTReader::ReadSFNT()
{
	CCDiskFile file;
	if (!file.open(PathName(m_pszFileName), ios::in | ios::binary))
		return FALSE;

	BOOL bRes = ReadSFNT(&file);
	file.close();

	return bRes;
}

/******************************************************************************
   Function:	Read
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
bool SFNTReader::Head::Read(CCFile* pFile)
{
	pFile->seek(0);
	version = (UINT32)getlong(pFile);
	fontRevison = getlong(pFile);
	checkSumAdjustment = (UINT32)getlong(pFile);
	magicNumber = (UINT32)getlong(pFile);
	flags = getushort(pFile);
	unitsPerEm = getushort(pFile);
	createdTime[1] = getlong(pFile);
	createdTime[0] = getlong(pFile);
	modifiedTime[1] = getlong(pFile);
	modifiedTime[0] = getlong(pFile);
	xMin = (INT16)getushort(pFile);
	yMin = (INT16)getushort(pFile);
	xMax = (INT16)getushort(pFile);
	yMax = (INT16)getushort(pFile);
	macStyle = (UINT16)getushort(pFile);
	lowestRecPPEM = (UINT16)getushort(pFile);
	fontDirectionHint = (INT16)getushort(pFile);
	indexToLocFormat = (INT16)getushort(pFile);
	glyphDataFormat = (INT16)getushort(pFile);

	m_bIsRead = pFile->good() ? true : false;
	if (m_bIsRead) g_iEM = (int)unitsPerEm;

	return m_bIsRead;
}

/******************************************************************************
   Function:	Read
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
bool SFNTReader::Hhea::Read(CCFile* pFile)
{
	pFile->seek(0);
	majorVersion = (UINT16)getushort(pFile);
	minorVersion = (UINT16)getushort(pFile);
	Ascender = (INT16)getushort(pFile);
	Descender = (INT16)getushort(pFile);
	LineGap = (INT16)getushort(pFile);
	advanceWidthMax = (UINT16)getushort(pFile);
	minLeftSideBearing = (INT16)getushort(pFile);
	minRightSideBearing = (INT16)getushort(pFile);
	xMaxExtent = (INT16)getushort(pFile);
	caretSlopeRise = (INT16)getushort(pFile);
	caretSlopeRun = (INT16)getushort(pFile);
	caretOffset = (INT16)getushort(pFile);
	reserved1 = (INT16)getushort(pFile);
	reserved2 = (INT16)getushort(pFile);
	reserved3 = (INT16)getushort(pFile);
	reserved4 = (INT16)getushort(pFile);
	metricDataFormat = (INT16)getushort(pFile);
	numberOfHMetrics = (UINT16)getushort(pFile);

	m_bIsRead = pFile->good() ? true : false;
	return m_bIsRead;
}

/******************************************************************************
   Function:	Read
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
bool SFNTReader::OS2::Read(CCFile* pFile)
{
	pFile->seek(0);
	version = (UINT16)getushort(pFile);
	xAvgCharWidth = (INT16)getushort(pFile);
	usWeightClass = (UINT16)getushort(pFile);
	usWidthClass = (UINT16)getushort(pFile);
	fsType = (UINT16)getushort(pFile);
	ySubscriptXSize = (INT16)getushort(pFile);
	ySubscriptYSize = (INT16)getushort(pFile);
	ySubscriptXOffset = (INT16)getushort(pFile);
	ySubscriptYOffset = (INT16)getushort(pFile);
	ySuperscriptXSize = (INT16)getushort(pFile);
	ySuperscriptYSize = (INT16)getushort(pFile);
	ySuperscriptXOffset = (INT16)getushort(pFile);
	ySuperscriptYOffset = (INT16)getushort(pFile);
	yStrikeoutSize = (INT16)getushort(pFile);
	yStrikeoutPosition = (INT16)getushort(pFile);
	sFamilyClass = (INT16)getushort(pFile);
	pFile->read(panose, 10);
	ulUnicodeRange1 = (UINT32)getlong(pFile);	//	Bits 031
	ulUnicodeRange2 = (UINT32)getlong(pFile);	//	Bits 3263
	ulUnicodeRange3 = (UINT32)getlong(pFile);	//	Bits 6495
	ulUnicodeRange4 = (UINT32)getlong(pFile);	//	Bits 96127
	pFile->read(achVendID, 4);		// The four character identifier for the vendor of the given type face.
	fsSelection = (UINT16)getushort(pFile);
	usFirstCharIndex = (UINT16)getushort(pFile);
	usLastCharIndex = (UINT16)getushort(pFile);
	sTypoAscender = (INT16)getushort(pFile);
	sTypoDescender = (INT16)getushort(pFile);
	sTypoLineGap = (INT16)getushort(pFile);
	usWinAscent = (UINT16)getushort(pFile);
	usWinDescent = (UINT16)getushort(pFile);
	if (version >= 1)
	{
		ulCodePageRange1 = (UINT32)getlong(pFile);		//	Bits 031
		ulCodePageRange2 = (UINT32)getlong(pFile);	//	Bits 3263

		if (version >= 2)
		{
			sxHeight = (INT16)getushort(pFile);
			sCapHeight = (INT16)getushort(pFile);
			usDefaultChar = (UINT16)getushort(pFile);
			usBreakChar = (UINT16)getushort(pFile);
			usMaxContext = (UINT16)getushort(pFile);
			if (!pFile->eof())
			{
				usLowerOpticalPointSize = (UINT16)getushort(pFile);
				usUpperOpticalPointSize = (UINT16)getushort(pFile);
			}
		}
	}
	m_bIsRead = pFile->good() ? true : false;
	return m_bIsRead;
}
/******************************************************************************
   Function:	Read
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
bool SFNTReader::Maxp::Read(CCFile* pFile)
{
	pFile->seek(0);
	version = getlong(pFile); // Table version number	0x00010000 for version 1.0.
	numGlyphs = getushort(pFile); //	The number of glyphs in the font.

	if(version == 0x00010000)
	{
		maxPoints = getushort(pFile); //	Maximum points in a non-composite glyph.
		maxContours = getushort(pFile); //	Maximum contours in a non-composite glyph.
		maxCompositePoints = getushort(pFile); //	Maximum points in a composite glyph.
		maxCompositeContours = getushort(pFile); //	Maximum contours in a composite glyph.
		maxZones = getushort(pFile); //	1 if instructions do not use the twilight zone (Z0), or 2 if instructions do use Z0; should be set to 2 in most cases.
		maxTwilightPoints = getushort(pFile); //	Maximum points used in Z0.
		maxStorage = getushort(pFile); //	Number of Storage Area locations.
		maxFunctionDefs = getushort(pFile); //	Number of FDEFs, equals to the highest function number + 1.
		maxInstructionDefs = getushort(pFile); //	Number of IDEFs.
		maxStackElements = getushort(pFile); //	Maximum stack depth2.
		maxSizeOfInstructions = getushort(pFile); //	Maximum byte count for glyph instructions.
		maxComponentElements = getushort(pFile); //	Maximum number of components referenced at top level for any composite glyph.
		maxComponentDepth = getushort(pFile); //	Maximum levels of recursion; 1 for simple components.
	}
	m_bIsRead = pFile->good() ? true : false;
	return m_bIsRead;
}

/******************************************************************************
   Function:	ReadLoca
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
bool SFNTReader::ReadLoca(CCFile* pFile)
{
	pFile->seek(0);
	if(!m_head.IsRead())
		return false; // Need to read the 'head' first - we have to use their indexToLocFormat
	bool bLocaIs32 = m_head.indexToLocFormat == 1;
	if(!m_maxp.IsRead())
		return false; // Need to take the glyph count from 'maxp' table first
	size_t nGlyphsCount = (size_t)m_maxp.numGlyphs;
	size_t szActualCount = bLocaIs32 ? pFile->Size() / 4 : pFile->Size() / 2;
	if(szActualCount < nGlyphsCount)
		nGlyphsCount = szActualCount;

	m_loca.reserve(nGlyphsCount+1);

	if(bLocaIs32)
	{
		for(size_t i = 0; i < nGlyphsCount + 1; i++)
			m_loca.push_back((DWORD)getlong(pFile));
	}
	else
	{
		for(size_t i = 0; i < nGlyphsCount + 1; i++)
			m_loca.push_back(((DWORD)getushort(pFile)) << 1);
	}
	return pFile->good() ? true : false;
}

/******************************************************************************
   Function:	ReadHmtx
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CCFile * pFile: file containing 'hmtx' table
   Notes:       
******************************************************************************/
bool SFNTReader::ReadHmtx(CCFile* pFile)
{
	pFile->seek(0);
	if (!m_hhea.IsRead())
		return false; // we need to read 'hhea' first

	bool bApplyLsb = !(m_head.flags & (1 << 1)) ? true : false;

	size_t nGlyphCount = m_loca.size() ? m_loca.size() - 1 : m_maxp.numGlyphs;
	if (!nGlyphCount || !m_hhea.numberOfHMetrics) return false;

	m_hmtx.resize(nGlyphCount);
	size_t i = 0;
	for (; i < (size_t)m_hhea.numberOfHMetrics; i++)
	{
		m_hmtx[i].advance = (UINT16)getushort(pFile);
		int lsb = (INT16)getushort(pFile);

		if (bApplyLsb)
		{
			if (0 != lsb)
			{
				// translate the glyph here
// 				trans[4] = lsb - sc->lsidebearing;
// 				SplinePointListTransform(sc->layers[ly_fore].splines, trans, tpt_AllPoints);
			}
		}
		else
			m_hmtx[i].lsb = lsb;
	}

	for (; i < nGlyphCount; i++)
	{
		m_hmtx[i].advance = m_hmtx[i - 1].advance;

		if (bApplyLsb)
		{
			int lsb = getushort(pFile);
			if (0 != lsb)
			{
// 				trans[4] = lsb - sc->lsidebearing;
// 				SplinePointListTransform(sc->layers[ly_fore].splines, trans, tpt_AllPoints);
			}
		}
		else
			m_hmtx[i].lsb = 0;
	}

	return pFile->good() ? true : false;
}

bool SFNTReader::HasEncoding(CCFile* pFile, int platformID, int encodingID)
{
	pFile->seek(0);
	int iVersion = getushort(pFile);
	if (iVersion != 0)
		return false;

	size_t szTableCount = (size_t)getushort(pFile);

	for (size_t i = 0; i < szTableCount; i++)
	{
		int iPlatformID = getushort(pFile);
		int iEncodingID = getushort(pFile);

		if (platformID == iPlatformID && encodingID == iEncodingID)
			return true;

		long lOffset = getlong(pFile);
	}

	return false;
}

/******************************************************************************
   Function:	ReadCmap
   Author:		Paul Gruzdev
   Purpose:		Read a 'cmap' table and form internal Unicode --> GID map.
   Returns:		bool
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
bool SFNTReader::ReadCmap(CCFile* pFile)
{
	pFile->seek(0);
	int iVersion = getushort(pFile);
	if (iVersion != 0)
		return false;

	size_t szTableCount = (size_t)getushort(pFile);
	std::map<std::pair<int, int>, LONG> maps;

	for (size_t i = 0; i < szTableCount; i++)
	{
		int iPlatformID = getushort(pFile);
		int iEncodingID = getushort(pFile);
		long lOffset = getlong(pFile);
		maps[std::make_pair(iPlatformID, iEncodingID)] = lOffset;
	}

	int iSubtract = 0;
	std::map<std::pair<int, int>, LONG>::iterator it = maps.find(std::make_pair(3, 1));
	if(it == maps.end()) // Now try the 'Symbol' encoding
	{
		it = maps.find(std::make_pair(3, 0));

		if(it != maps.end())
			iSubtract = 0xF000;
	}

	if(it == maps.end()) // Now try general Mac Unicode encoding
		it = maps.find(std::make_pair(0, 3));

	if(it != maps.end())
	{
		pFile->seek(it->second);
		int iVersion = getushort(pFile);

		ASSERT(iVersion == 4);
		if(iVersion != 4)
			return false;

		ULONG ulLength = (ULONG)getushort(pFile);
		getushort(pFile); // language
		size_t segCountX2 = (size_t)getushort(pFile);
		getushort(pFile); // searchRange
		getushort(pFile); // entrySelector
		getushort(pFile); // rangeShift

		std::vector<WCHAR> segStart, segEnd;
		std::vector<UINT16> RangeOffset;
		std::vector<short> segDelta;

		for (size_t i = 0; i < segCountX2 / 2; ++i)
			segEnd.push_back((WCHAR)getushort(pFile));

		getushort(pFile); // reservedPad

		for (size_t i = 0; i < segCountX2 / 2; ++i)
			segStart.push_back((WCHAR)getushort(pFile));

		for (size_t i = 0; i < segCountX2 / 2; ++i)
			segDelta.push_back((short)getushort(pFile));

		for (size_t i = 0; i < segCountX2 / 2; ++i)
			RangeOffset.push_back((UINT16)getushort(pFile));

		size_t szGlyphIDCnt = (pFile->Size() - pFile->tell()) / 2;
		for(size_t i = 0; i < szGlyphIDCnt; i++)
			RangeOffset.push_back((UINT16)getushort(pFile));

		for(size_t i = 0; i < segCountX2 / 2; ++i)
		{
			for(WCHAR ch = segStart[i]; ch <= segEnd[i] && ch != 0xffff; ++ch)
			{
				if(RangeOffset[i] != 0)
				{
					int iGLyphId = *(RangeOffset[i]/2 + (ch - segStart[i]) + &RangeOffset[i]);
					if(iGLyphId)
					{
						int iGID = (iGLyphId + (int)segDelta[i]) % 65536;
						m_CodeToGID[ch - iSubtract] = iGID;
						m_GIDToCode[iGID].insert(ch - iSubtract);
					}
				}
				else
				{
					int iGID = ((int)ch + (int)segDelta[i]) % 65536;
					m_CodeToGID[ch - iSubtract] = iGID;
					m_GIDToCode[iGID].insert(ch - iSubtract);
				}
			}
		}
	}

	return m_CodeToGID.size() > 0;
}

/******************************************************************************
   Function:	CompositeGlyphExtractComponents
   Author:		Paul Gruzdev
   Purpose:		Go through the composite glyph definition and find all glyphs
				this glyph consists of
   Returns:		bool
   Parameters:	CCFile * ttf
   Parameters:	std::vector<int> & comp
   Notes:       
******************************************************************************/
static bool CompositeGlyphExtractComponents(CCFile* ttf, std::vector<int>& comp)
{
	int flags = 0;
	do
	{
		if (ttf->eof())
			break;

		flags = getushort(ttf);
		int iGlyphIdx = getushort(ttf);
		comp.push_back(iGlyphIdx);

		if (flags & _ARGS_ARE_WORDS)
		{
			getushort(ttf); // arg1
			getushort(ttf); // arg2
		}
		else
			getushort(ttf); // read just two bytes for arg1 and arg2

		if (flags & _SCALE)
			getushort(ttf);
		else if (flags & _XY_SCALE)
		{
			getushort(ttf);
			getushort(ttf);
		}
		else if (flags & _MATRIX)
		{
			getushort(ttf);
			getushort(ttf);
			getushort(ttf);
			getushort(ttf);
		}
	} while (flags & _MORE);

	return !ttf->eof() && comp.size();
}

/******************************************************************************
   Function:	Subset
   Author:		Paul Gruzdev
   Purpose:		General subsetting function. 
   Returns:		bool
   Parameters:	CharSet & subset
   Parameters:	std::set<ULONG> & subsetGID
   Notes:       
******************************************************************************/
bool SFNTReader::Subset(CharSet& subset, std::set<ULONG>& subsetGID)
{
	if (m_format == keTrueType)
		return SubsetTTF(subset, subsetGID);
	else if (m_format == keOpenType)
		return SubsetOTF(subset, subsetGID);
	return false;
}

//******************************************************************************
//  Function:	SFNT::SFNTReader::ConvertToOTFAndSubset
//  Author:		Paul Gruzdev
//  Purpose:	Convert previously read TTF to OTF and optionaly subset.
//				If already OTF - just subset it.
//  Returns:	bool
//  Parameters:	const CharSet * pSubset
//  Parameters:	std::set<ULONG> * pSubsetGID
//  Parameters:	bool bCID
//  Notes:		
//******************************************************************************
bool SFNTReader::ConvertToOTFAndSubset(const CharSet* pSubset, std::set<ULONG>* pSubsetGID, bool bCID)
{
	if (m_format == SFNT::keOpenType && !(pSubset && pSubset->Count()) && !(pSubsetGID && pSubsetGID->size()))
		return true; // nothing to re-create - we got no subsets!

	m_OS2.Read(map()[TABLEID_OS2]);
	m_head.Read(map()[TABLEID_HEAD]);
	m_hhea.Read(map()[TABLEID_HHEA]);
	m_maxp.Read(map()[TABLEID_MAXP]);
	if (m_format == SFNT::keTrueType)
		ReadLoca(map()[TABLEID_LOCA]);
	ReadHmtx(map()[TABLEID_HMTX]);
	ReadCmap(map()[TABLEID_CMAP]);

	// Test whether this file needs subsetting or it contains just the glyphs 
	// listed in subset/subsetGID
	size_t szCount = (pSubset ? pSubset->Count() : 0) + (pSubsetGID ? pSubsetGID->size() : 0);
	if (m_format == SFNT::keOpenType && m_CodeToGID.size() == szCount)
	{
		bool equal = true;
		if (pSubset)
		{
			for (GIDMap::iterator it = m_CodeToGID.begin(); it != m_CodeToGID.end(); ++it)
			if (!pSubset->Exists(it->first))
			{
				equal = false;
				break;
			}
		}

		if (pSubsetGID)
		{
			for (std::set<ULONG>::iterator it = pSubsetGID->begin(); it != pSubsetGID->end() && equal; ++it)
			{
				if (*it >= m_CodeToGID.size())
				{
					equal = false;
					break;
				}
			}
		}
		if (equal) return true;
	}

	bool bRetVal = false;

	if (m_format == SFNT::keTrueType)
	{
		float fOldXaraEM = g_fXaraEM;
		g_fXaraEM = 8388607.0f; // increase the precision hugely
		if (ReadGlyf(map()[TABLEID_GLYF], (CharSet*)pSubset, pSubsetGID))
		{
			// SFNTReader here acts as a VGlyphSource
			SFNTWriter writer(this, SFNT::keOpenType);
			writer.SetWriteCID(bCID);
			TableMap map;
			writer.m_dwLastError = writer.GenerateTables(map);
			if (writer.GetLastError() == NO_ERROR)
			{
				this->m_tableMap.m_map.swap(map.m_map);
				m_format = SFNT::keOpenType;
				m_loca.clear();
				m_hmtx.clear();

				for (TableMap::iterator it = m_tableMap.begin(); it != m_tableMap.end(); ++it)
				{
					it->second->ChangeFileAccess(CCMemRead);
					it->second->seek(0);
				}
				bRetVal = true;
			}
		}
		g_fXaraEM = fOldXaraEM;
	}
	else
	{
		if (pSubsetGID && pSubsetGID->size())
			bRetVal = SubsetOTF((CharSet&)*pSubset, *pSubsetGID);
		else
		{
			CharSet charset;
			if (pSubset == NULL)
			{
				for (GIDMap::iterator it = m_CodeToGID.begin(); it != m_CodeToGID.end(); ++it)
					charset.Add(it->first);
				pSubset = &charset;
			}
			bRetVal = RepackOTF(*pSubset, bCID);
		}
	}

	return bRetVal;
}

//******************************************************************************
//  Function:	SFNT::SFNTReader::GIDExists
//  Author:		Paul Gruzdev
//  Purpose:	Given a glyph index, find whether any of it's Unicode values
//				present in the provided subset
//  Returns:	bool
//  Parameters:	[in] - ULONG uGID - glyph index to test
//  Parameters:	[in] - CharSet& subset - Unicode subset to test in
//  Notes:		
//******************************************************************************
bool SFNTReader::GIDExists(ULONG uGID, CharSet& subset)
{
	GIDToCodeMap::iterator it = m_GIDToCode.find(uGID);
	bool bUnicodeFound = false;
	if (it != m_GIDToCode.end())
	{
		for (auto &wUni : it->second)
		{
			if (subset.Exists(wUni))
			{
				bUnicodeFound = true;
				break;
			}
		}
	}
	return bUnicodeFound;
}

/******************************************************************************
   Function:	TTFCreateGIDCompositeSubset
   Author:		Paul Gruzdev
   Purpose:		Given Unicode subset and GID subset, make a joined GID subset,
				taking into account all composite glyphs that should be examined
				and all component glyphs also included
   Returns:		bool
   Parameters:	[in]  CharSet & subset - Unicode subset
   Parameters:	[in]  std::set<ULONG> & subsetGID - GID subset
   Parameters:	[out] std::set<int> & gids - result GID subset
   Notes:       
******************************************************************************/
bool SFNTReader::TTFCreateGIDCompositeSubset(CharSet& subset, std::set<ULONG>& subsetGID, std::set<ULONG>& gids)
{
	TableMap::iterator itGlyf = map().find(TABLEID_GLYF);
	if (itGlyf == map().end())
		return false;

	PBYTE pby = itGlyf->second->GetMemoryBuffer();

	// Now find the composite glyphs and add the originals they reference to into the subset also
	for (size_t i = 3; i < m_loca.size() - 1; i++)
	{
		bool bDoSubset = GIDExists((ULONG)i, subset) || subsetGID.find((ULONG)i) != subsetGID.end();
		DWORD dwSize = m_loca[i + 1] - m_loca[i];

		if (bDoSubset)
			gids.insert((ULONG)i);

		if (dwSize && bDoSubset)
		{
			CCMemFile glyph(&pby[m_loca[i]], dwSize, CCMemRead);
			int iNumContours = getushort(&glyph);

			if (iNumContours == 0xFFFF) // composite glyph
			{
				getushort(&glyph);	// xMin
				getushort(&glyph);	// yMin
				getushort(&glyph);	// xMax
				getushort(&glyph);	// yMax

				std::vector<int> comps;
				CompositeGlyphExtractComponents(&glyph, comps);
				for (auto &GID : comps)
					gids.insert(GID);
			}

			glyph.close();
		}
	}

	return gids.size() ? true : false;
}

/******************************************************************************
   Function:	SubsetTTF
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CharSet & subset - set of Unicode values to subset
				std::set<ULONG>& subsetGID - set of glyph IDs (indexes) to subset
   Notes:       
******************************************************************************/
bool SFNTReader::SubsetTTF(CharSet& subset, std::set<ULONG>& subsetGID)
{
	if(m_format != keTrueType)
	{
		ERROR3("Wrong font type: use SubsetOTF function");
		return false;
	}

	TableMap::iterator it = map().find(TABLEID_HEAD);
	if(it == map().end())
		return false;

	if(m_head.Read(it->second) == false)
		return false;

	it = map().find(TABLEID_MAXP);
	if(it == map().end())
		return false;

	if(m_maxp.Read(it->second) == false)
		return false;

	TableMap::iterator itLoca = map().find(TABLEID_LOCA);
	if(itLoca == map().end())
		return false;

	if(ReadLoca(itLoca->second) == false)
		return false;

	it = map().find(TABLEID_CMAP);
	if(it == map().end())
		return false;

	if(m_CodeToGID.size() == 0 && ReadCmap(it->second) == false)
		return false;

	TableMap::iterator itGlyf = map().find(TABLEID_GLYF);
	if(itGlyf == map().end())
		return false;

	map().EraseTable(TABLEID_DSIG);
	map().EraseTable(TABLEID_PCLT); // Microsoft no longer recommends including this table
	//map().EraseTable(TABLEID_GSUB); // We won't parse GSUB for now, but we can't retain it because we need to parse
	// Update: we need GSUB for liagtures. But we really need to fully subset it
	map().EraseTable(CHR('M', 'E', 'R', 'G')); // Upsets TTEmbedFontFromFile function
	// it to figure out what ligatures glyphs we need to retain.
	PBYTE pby = itGlyf->second->GetMemoryBuffer();
	std::vector<DWORD> newLoca;
	newLoca.push_back(m_loca.front()); // re-create first 3 default glyphs
	newLoca.push_back(m_loca[1]);
	newLoca.push_back(m_loca[2]);

	std::set<ULONG> gids;
	TTFCreateGIDCompositeSubset(subset, subsetGID, gids);

	CCMemFile* pNewGlyf = new CCMemFile(NULL, 0, CCMemWrite);
	pNewGlyf->write(&pby[m_loca[0]], m_loca[3] - m_loca[0]);

	for(size_t i = 3; i < m_loca.size() - 1; i++)
	{
		newLoca.push_back(pNewGlyf->tell());
		if(gids.find((int)i) != gids.end())
			pNewGlyf->write(&pby[m_loca[i]], m_loca[i+1] - m_loca[i]);
	}

	newLoca.push_back(pNewGlyf->tell());

	CCMemFile* pNewLoca = new CCMemFile(NULL, 0, CCMemWrite);

	if(m_head.indexToLocFormat == 1)
	{
		for(size_t i = 0; i < newLoca.size(); i++)
			putlong(pNewLoca, newLoca[i]);
	}
	else
	{
		for(size_t i = 0; i < newLoca.size(); i++)
			putshort(pNewLoca, (short)((WORD)(newLoca[i] >> 1)));
	}

	itLoca->second->close();
	delete itLoca->second;
	itLoca->second = pNewLoca;
	itGlyf->second->close();
	delete itGlyf->second;
	itGlyf->second = pNewGlyf;
	pNewGlyf->ChangeFileAccess(CCMemRead);
	pNewLoca->ChangeFileAccess(CCMemRead);

	return pNewGlyf->good() * pNewLoca->good() ? true : false;
}

/******************************************************************************
   Function:	SubsetOTF
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	CharSet & subset
   Parameters:	std::set<ULONG> & subsetGID
   Notes:       
******************************************************************************/
bool SFNTReader::SubsetOTF(CharSet& subset, std::set<ULONG>& subsetGID)
{
	if (m_format != keOpenType)
	{
		ERROR3("Call SubsetTTF function instead.");
		return false;
	}

	TableMap::iterator it = map().find(TABLEID_CMAP);
	if (it == map().end())
		return false;

	if (m_CodeToGID.size() == 0 && ReadCmap(it->second) == false)
		return false;

	TableMap::iterator itCFF = map().find(TABLEID_CFF);
	if (itCFF == map().end())
		return false;

	map().EraseTable(TABLEID_DSIG);
	map().EraseTable(TABLEID_PCLT); // Microsoft no longer recommends including this table
	//map().EraseTable(TABLEID_GSUB); // We won't parse GSUB for now, but we can't retain it because we need to parse
	// Update: we need GSUB for ligatures. But we really need to fully subset it
	map().EraseTable(CHR('M', 'E', 'R', 'G')); // Upsets TTEmbedFontFromFile function
	// it to figure out what ligatures glyphs we need to retain.

	CCMemFile* pNewCFF = new CCMemFile(NULL, 0, CCMemWrite);
	bool bRes = SubsetCFF(itCFF->second, pNewCFF, subset, subsetGID);

	if (bRes)
	{
		itCFF->second->close();
		delete itCFF->second;
		itCFF->second = pNewCFF;
		pNewCFF->ChangeFileAccess(CCMemRead);
	}

	return bRes && (pNewCFF->good() ? true : false);
}

/******************************************************************************
   Function:	SFNTReader::GetFontType
   Author:		Paul Gruzdev
   Purpose:     Get type of a font (ttf, otf, type1, etc)
   Returns:		int
   Parameters:	char * filename  
   Notes:       
******************************************************************************/
EFileFormat SFNTReader::GetFontType(TCHAR* filename)
{
	EFileFormat format = keUndefined;

	if (filename == NULL)
		return keUndefined;

	FILE* file = NULL;
	errno_t err = _tfopen_s(&file, filename, _T("rb"));
	if(!file) return keUndefined;

	// Try to guess the file type from the first few characters...
	int ch1 = getc(file);
	int ch2 = getc(file);
	int ch3 = getc(file);
	int ch4 = getc(file);
	int ch5 = getc(file);
	int ch6 = getc(file);
	int ch7 = getc(file);
	int ch9, ch10;
	fseek(file, 98, SEEK_SET);
	ch9 = getc(file);
	ch10 = getc(file);
	rewind(file);

	if (( ch1==0 && ch2==1 && ch3==0 && ch4==0 ) || (ch1=='t' && ch2=='r' && ch3=='u' && ch4=='e'))
		format = keTrueType;
	else if(ch1 == 'O' && ch2 == 'T' && ch3 == 'T' && ch4 == 'O')
		format = keOpenType;
	else if(ch1 == 't' && ch2 == 't' && ch3 == 'c' && ch4 == 'f')
		format = keTTC;
	else if ( ch1 == 'w' && ch2 == 'O' && ch3 == 'F' && ch4 == 'F')
		format = keWOFF;
	else if (( ch1=='%' && ch2=='!' ) || (ch1 == 0x80 && ch2 == '\01')) 
		// PFB header
		format = keType1;
	else if (ch1 == 1 && ch2 == 0 && ch3 == 4) 
		format = keCFF;

	fclose(file);
	return format;
}

//******************************************************************************
//  Function:	SFNT::SFNTReader::ReadGlyf
//  Author:		Paul Gruzdev
//  Purpose:	Parse 'glyf' SFNT table
//  Returns:	bool
//  Parameters:	CCFile * pFile
//  Parameters:	CharSet * pSubset
//  Parameters:	std::set<ULONG> * pSubsetGID
//  Notes:		
//******************************************************************************
bool SFNTReader::ReadGlyf(CCFile* pFile, CharSet* pSubset, std::set<ULONG>* pSubsetGID)
{
	if (!m_loca.size())
		return false;

// 	std::set<ULONG> gids;
// 	bool subset = false;
// 	if (pSubset && pSubsetGID)
// 	{
// 		gids.insert(0);
// 		gids.insert(1);
// 		gids.insert(2); // always read 3 default glyphs
// 		subset = TTFCreateGIDCompositeSubset(*pSubset, *pSubsetGID, gids);
// 	}

	size_t iSrcGlyphCnt = m_loca.size() - 1;
	size_t iDstGlyphCnt = pSubset ? pSubset->Count() : iSrcGlyphCnt;
	if (pSubsetGID) iDstGlyphCnt += pSubsetGID->size();
	iDstGlyphCnt += 1; // add .notdef glyph
	m_chars.resize(iDstGlyphCnt);
	GIDMap gmNew;
	GIDToCodeMap cmNew;
	std::vector<HmtxRecord> hmtxNew;
	std::map<size_t, size_t> backGID;

	// read all the glyphs
	for (size_t i = 0, d = 0; i < iSrcGlyphCnt; ++i)
	{
		bool bDoSubset = i == 0 || ((pSubset == NULL /*&& i > 2*/) || GIDExists((ULONG)i, *pSubset)) || (pSubsetGID && pSubsetGID->find((ULONG)i) != pSubsetGID->end());
//		if (pSubset == NULL || gids.find((ULONG)i) != gids.end())
		if (bDoSubset)
		{
			GIDToCodeMap::iterator it = m_GIDToCode.find((int)i);
			if (it != m_GIDToCode.end())
			{
				for (auto &wUni : it->second)
				{
					if (!pSubset || pSubset->Exists(wUni))
					{
						gmNew[wUni] = (int)d;
						cmNew[(int)d].insert(wUni);
						m_uni.insert(wUni);
						m_chars[d].wUnicode = wUni;
					}
				}
			}

			m_chars[d].uPrevGID = (ULONG)i;
			backGID[i] = d;
			ReadTTFGlyph(pFile, d++, backGID, m_loca[i], m_loca[i + 1], i);
			hmtxNew.push_back(m_hmtx[i]);
		}
	}

	m_CodeToGID = gmNew;
	m_GIDToCode = cmNew;
	m_hmtx = hmtxNew;

	return true;
}

//******************************************************************************
//  Function:	SFNT::FigureControls
//  Author:		Paul Gruzdev
//  Purpose:	Given one ttf control point, calculate two bezier control
//				points and store them in Character
//  Returns:	void
//  Parameters:	Character & ch
//  Parameters:	DPoint & ptPrev
//  Parameters:	DPoint & ptTo
//  Parameters:	DPoint & cp
//  Notes:		
//******************************************************************************
static void FigureControls(Character& ch, DPoint& ptPrev, DPoint& ptTo, DPoint& cp)
{
	float x1 = (float)(1.0 / 3.0) * (ptPrev.x + (float)2 * cp.x);
	float y1 = (float)(1.0 / 3.0) * (ptPrev.y + (float)2 * cp.y);
	float x2 = (float)(1.0 / 3.0) * ((float)2 * cp.x + ptTo.x);
	float y2 = (float)(1.0 / 3.0) * ((float)2 * cp.y + ptTo.y);
	ch.verb.push_back(PT_BEZIERTO);
	ch.verb.push_back(PT_BEZIERTO);
	ch.verb.push_back(PT_BEZIERTO);
	ch.coord.push_back(DocCoord(ttf2xar_l(x1), ttf2xar_l(y1)));
	ch.coord.push_back(DocCoord(ttf2xar_l(x2), ttf2xar_l(y2)));
	ch.coord.push_back(ttf2xar_dp(ptTo));
}

//******************************************************************************
//  Function:	TTFBuildContours
//  Author:		Paul Gruzdev
//  Purpose:	Convert TTF flags/points of quadratic spline into cubic bezier
//				in Xara format
//  Returns:	void
//  Parameters:	Character & ch
//  Parameters:	LONG lPathCnt
//  Parameters:	std::vector<UINT16> & endpt
//  Parameters:	std::vector<char> & flags
//  Parameters:	std::vector<DPoint> & pts
//  Notes:		
//******************************************************************************
static void TTFBuildContours(Character& ch, LONG lPathCnt, std::vector<UINT16>& endpt, std::vector<char>& flags, std::vector<DPoint>& pts)
{
	LONG i = 0;
	for (LONG path = 0; path < lPathCnt; ++path)
	{
		if (endpt[path] < i)	// Sigh. Yes there are fonts with bad endpt info
			continue;

		bool bLastOff = false;
		bool bFirstPt = true;
		DPoint ptPrev, ptFirstOnCurve;
		LONG start = i;

		while (i <= endpt[path])
		{
			if (flags[i] & _On_Curve)
			{
				if (bLastOff)
				{
					if (!bFirstPt)
						FigureControls(ch, ptPrev, pts[i], pts[i - 1]);
					else
					{
						ch.verb.push_back(PT_MOVETO);
						ch.coord.push_back(ttf2xar_dp(pts[i]));
						ptFirstOnCurve = pts[i];
						bFirstPt = false;
					}
					bLastOff = false;
				}
				else
				{
					ch.verb.push_back(bFirstPt ? PT_MOVETO : PT_LINETO);
					ch.coord.push_back(ttf2xar_dp(pts[i]));
					if (bFirstPt)
					{
						ptFirstOnCurve = pts[i];
						bFirstPt = false;
					}
				}
				ptPrev = pts[i];
			}
			else if (bLastOff)
			{
				DPoint ptTo;
				ptTo.x = (pts[i].x + pts[i - 1].x) / 2;
				ptTo.y = (pts[i].y + pts[i - 1].y) / 2;

				if (bFirstPt)
				{
					ch.verb.push_back(PT_MOVETO);
					ch.coord.push_back(ttf2xar_dp(ptTo));
					ptFirstOnCurve = ptTo;
					bFirstPt = false;
				}
				else
					FigureControls(ch, ptPrev, ptTo, pts[i - 1]);

				ptPrev = ptTo;
			}
			else
			{
				bLastOff = true;
			}
			++i;
		}

		if (start == i - 1)
		{
			// MS chinese fonts have contours consisting of a single off curve
			//  point. What on earth do they think that means? 
			// Oh. I see. It's used to position marks and such
		}
		else if (!(flags[start] & _On_Curve) && !(flags[i - 1] & _On_Curve))
		{
			float x3 = (pts[start].x + pts[i - 1].x) / 2;
			float y3 = (pts[start].y + pts[i - 1].y) / 2;
			FigureControls(ch, ptPrev, DPoint(x3, y3), pts[i - 1]);
			FigureControls(ch, DPoint(x3, y3), ptFirstOnCurve, pts[start]);
		}
		else if (!(flags[i - 1] & _On_Curve))
		{
			FigureControls(ch, ptPrev, ptFirstOnCurve, pts[i - 1]);
		}
		else if (!(flags[start] & _On_Curve))
		{
			FigureControls(ch, ptPrev, ptFirstOnCurve, pts[start]);
		}

		if (ch.verb.size())
			ch.verb.back() |= PT_CLOSEFIGURE;
	}
}

/******************************************************************************
   Function:	ReadTTFSimpleGlyph
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		void
   Parameters:	CCFile * ttf
   Parameters:	Character & sc
   Parameters:	LONG lPathCnt
   Parameters:	LONG gbb[4]
   Notes:       
******************************************************************************/
static void ReadTTFSimpleGlyph(CCFile* ttf, Character& sc, LONG lPathCnt, LONG gbb[4])
{
	std::vector<UINT16> endpt(lPathCnt + 1);
	std::vector<char> flags;
	std::vector<DPoint> pts;

	ULONG tot;

	for (LONG i = 0; i < lPathCnt; ++i)
	{
		endpt[i] = getushort(ttf);
		if (i != 0 && endpt[i] < endpt[i - 1])
		{
			ERROR3("Bad tt font: contour ends make no sense in glyph");
			return;
		}
	}

	if (lPathCnt == 0)
	{
		tot = 0;
		pts.resize(1);
	}
	else
	{
		tot = endpt[lPathCnt - 1] + 1;
		pts.resize(tot);
	}

	UINT uiLen = (UINT)getushort(ttf);
	if (uiLen)
	{
		sc.instructions.resize(uiLen);
		ttf->read(&sc.instructions[0], uiLen);
	}
	flags.resize(tot);

	for (ULONG i = 0; i < tot; ++i)
	{
		flags[i] = (char)getchar(ttf);
		if (flags[i] & _Repeat)
		{
			ULONG cnt = (ULONG)getchar(ttf);
			if (i + cnt >= tot)
			{
				//IError("Flag count is wrong (or total is): %d %d", i + cnt, tot);
				cnt = tot - i - 1;
			}

			for (ULONG j = 0; j < cnt; ++j)
				flags[i + j + 1] = flags[i];
			i += cnt;
		}

		if (ttf->eof())
			break;
	}

// 	if (i != tot)
// 		//IError("Flag count is wrong (or total is): %d %d in glyph %d", i, tot, sc->orig_pos);
// 		ERROR3("Flag count is wrong (or total is)");

	int last_pos = 0;
	for (ULONG i = 0; i < tot; ++i)
	{
		if (flags[i] & _X_Short)
		{
			int off = (int)getchar(ttf);
			if (!(flags[i] & _X_Same))
				off = -off;
			pts[i].x = (float)(last_pos + off);
		}
		else if (flags[i] & _X_Same)
			pts[i].x = (float)last_pos;
		else
			pts[i].x = (float)last_pos + (short)getushort(ttf);

		last_pos = (int)pts[i].x;

// 		if ((last_pos < gbb[0] || last_pos > gbb[2]) && (flags[i] & _On_Curve))
// 		{
// 			ERROR3("A point in GID is outside the glyph bounding box");
// 		}
	}

	last_pos = 0;
	for (ULONG i = 0; i < tot; ++i)
	{
		if (flags[i] & _Y_Short)
		{
			int off = getchar(ttf);
			if (!(flags[i] & _Y_Same))
				off = -off;
			pts[i].y = (float)last_pos + off;
		}
		else if (flags[i] & _Y_Same)
			pts[i].y = (float)last_pos;
		else
			pts[i].y = (float)last_pos + (short)getushort(ttf);

		last_pos = (int)pts[i].y;

// 		if ((last_pos < gbb[1] || last_pos>gbb[3]) && (flags[i] & _On_Curve))
// 		{
// 			ERROR3("A point in GID is outside the glyph bounding box");
// 		}
	}

	TTFBuildContours(sc, lPathCnt, endpt, flags, pts);

// 	if (ttf->eof())
// 		ERROR3("Reached end of file when reading simple glyph");
}

//******************************************************************************
//  Function:	SFNT::SFNTReader::ReadTTFCompositGlyph
//  Author:		Paul Gruzdev
//  Purpose:	
//  Returns:	void
//  Parameters:	CCFile * ttf
//  Parameters:	size_t idx
//  Parameters:	std::map<size_t
//  Parameters:	size_t> & backGID
//  Parameters:	DWORD end
//  Notes:		
//******************************************************************************
void SFNTReader::ReadTTFCompositGlyph(CCFile* ttf, Character& sc, size_t idx, std::map<size_t, size_t>& backGID, DWORD end)
{
	int flags = 0, arg1, arg2;
	bool bUseMyMetrics = false;
	float transform[6]; memset(transform, 0, 6 * sizeof(float));

	if (ttf->tell() >= (FilePos)end)
	{
		ERROR3("Empty composite glyph");
		return;
	}

	do
	{
		if (ttf->tell() >= (FilePos)end)
		{
			ERROR3("Bad flags value, implied MORE components at end of glyph");
			break;
		}

		flags = getushort(ttf);
		int gid = getushort(ttf);

		if (flags & _ARGS_ARE_WORDS)
		{
			arg1 = (short)getushort(ttf);
			arg2 = (short)getushort(ttf);
		}
		else
		{
			arg1 = (signed char)getchar(ttf);
			arg2 = (signed char)getchar(ttf);
		}

		bool use_my_metrics = (flags & _USE_MY_METRICS) ? true : false;

		if (use_my_metrics)
		{
			if (bUseMyMetrics)
			{
				ERROR3("Use-my-metrics flag set on at least two components in glyph");
			}
			else
				bUseMyMetrics = true;
		}

		bool bRoundTranslationToGrid = (flags & _ROUND) ? true : false;
		if (flags & _ARGS_ARE_XY)
		{
			// There is some very strange stuff (half-)documented on the apple
			// site about how these should be interpreted when there are 
			// scale factors, or rotations 
			// It isn't well enough described to be comprehensible 
			// http://fonts.apple.com/TTRefMan/RM06/Chap6glyf.html 
			// Microsoft says nothing about this 
			// Adobe implies this is a difference between MS and Apple 
			// MS doesn't do this, Apple does (GRRRGH!!!!) 
			// Adobe says that setting bit 12 means that this will not happen 
			// Adobe says that setting bit 11 means that this will happen 
			// So if either bit is set we know when this happens, if neither 
			// we guess... But I still don't know how to interpret the 
			// apple mode under rotation... 
			// I notice that FreeType does nothing about rotation nor does it 
			// interpret bits 11&12 
			// Ah. It turns out that even Apple does not do what Apple's docs 
			// claim it does. I think I've worked it out (see below), but... 
			// Bleah! 
			transform[4] = (float)arg1;
			transform[5] = (float)arg2;
		}
		else
		{
			// Somehow we can get offsets by looking at the points in the 
			//  points so far generated and comparing them to the points in 
			//  the current component 
			// How exactly is not described on any of the Apple, MS, Adobe 
			// freetype looks up arg1 in the set of points we've got so far 
			//  looks up arg2 in the new component (before renumbering) 
			//  offset.x = arg1.x - arg2.x; offset.y = arg1.y - arg2.y; 
			// This fixup needs to be done later though (after all glyphs 
			//  have been loaded) 
// 			cur->match_pt_base = arg1;
// 			cur->match_pt_ref = arg2;
// 			cur->point_match = true;
			ERROR3("Need source points of the reference glyph");
		}

		transform[0] = transform[3] = 1.0;
		if (flags & _SCALE)
			transform[0] = transform[3] = get2dot14(ttf);
		else if (flags & _XY_SCALE)
		{
			transform[0] = get2dot14(ttf);
			transform[3] = get2dot14(ttf);
		}
		else if (flags & _MATRIX)
		{
			transform[0] = get2dot14(ttf);
			transform[1] = get2dot14(ttf);
			transform[2] = get2dot14(ttf);
			transform[3] = get2dot14(ttf);
		}

		if (flags & _ARGS_ARE_XY)
		{	// Only muck with these guys if they are real offsets and not point matching
			// everywhere else assume unscaled offsets unless told scaled explicitly
			if ((flags & _SCALED_OFFSETS) &&
				(flags & _ARGS_ARE_XY) && (flags&(_SCALE | _XY_SCALE | _MATRIX)))
			{
				// This is not what Apple documents on their website. But it is
				//  what appears to match the behavior of their rasteriser
				// Apple has changed their documentation (without updating their
				//  changelog), but I believe they are still incorrect
				transform[4] *= sqrt(transform[0] * transform[0] + transform[1] * transform[1]);
				transform[5] *= sqrt(transform[2] * transform[2] + transform[3] * transform[3]);
			}
		}

		XaraMatrix m(transform[0], transform[1], transform[2], transform[3], ttf2xar_l(transform[4]), ttf2xar_l(transform[5]));
		std::map<size_t, size_t>::iterator it = backGID.find(gid);
		Character fwd;
		Character& toAdd = fwd;

		if (it == backGID.end())
		{
			FilePos fpPrev = ttf->tell();
			ttf->seek(m_loca[gid]);
			LONG lPathCnt = (LONG)getushort(ttf);
			LONG bbox[4];
			bbox[0] = (LONG)getushort(ttf);
			bbox[1] = (LONG)getushort(ttf);
			bbox[2] = (LONG)getushort(ttf);
			bbox[3] = (LONG)getushort(ttf);

			if (lPathCnt != 0xFFFF)
				ReadTTFSimpleGlyph(ttf, fwd, lPathCnt, bbox);
			else
				ReadTTFCompositGlyph(ttf, fwd, gid, backGID, m_loca[gid+1]);

			ttf->seek(fpPrev);
		}
		else
			toAdd = m_chars[it->second];

		size_t szStartIdx = sc.verb.size();
		size_t szCnt = toAdd.verb.size();
		sc.verb.insert(sc.verb.end(), toAdd.verb.begin(), toAdd.verb.end());
		sc.coord.insert(sc.coord.end(), toAdd.coord.begin(), toAdd.coord.end());
		m.transform(&sc.coord[szStartIdx], (UINT)szCnt);

		/*if (ttf->eof())
		{
			ERROR3("Reached end of file when reading composite glyph");
			break;
		}*/
	} while (flags & _MORE);

	if ((flags & _INSTR) && ttf->tell() < (FilePos)end)
	{
		UINT uiLen = (UINT)getushort(ttf);
		sc.instructions.resize(uiLen);
		ttf->read(&sc.instructions[0], uiLen);
	}
}

//******************************************************************************
//  Function:	SFNT::SFNTReader::ReadTTFGlyph
//  Author:		Paul Gruzdev
//  Purpose:	
//  Returns:	bool
//  Parameters:	CCFile * ttf
//  Parameters:	size_t idx
//  Parameters:	std::map<size_t
//  Parameters:	size_t> & backGID
//  Parameters:	DWORD start
//  Parameters:	DWORD end
//  Parameters:	size_t gid
//  Notes:		
//******************************************************************************
bool SFNTReader::ReadTTFGlyph(CCFile* ttf, size_t idx, std::map<size_t, size_t>& backGID, DWORD start, DWORD end, size_t gid)
{
	LONG gbb[4];

	if (end < start)
	{
		ERROR3("Bad glyph, its data length is negative");
		return false;
	}

	if (start == end)
	{
		// empty glyph, just return
		return true;
	}

	ttf->seek(start);
	LONG lPathCnt = (LONG)getushort(ttf);
	gbb[0] = (LONG)getushort(ttf);
	gbb[1] = (LONG)getushort(ttf);
	gbb[2] = (LONG)getushort(ttf);
	gbb[3] = (LONG)getushort(ttf);

	if (lPathCnt != 0xFFFF)
	{
		ReadTTFSimpleGlyph(ttf, m_chars[idx], lPathCnt, gbb);
	}
	else
		ReadTTFCompositGlyph(ttf, m_chars[idx], idx, backGID, end);

	m_chars[idx].lLeftSideBearing = ttf2xar_l(gbb[0]);

	return true;
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// VGlyphSource implementation

VGlyphSource::GlyphState SFNTReader::GetGlyphPath(WCHAR wUni, PBYTE* verb, DocCoord** coord, ULONG* pulNumCoords)
{
	GIDMap::iterator it = m_CodeToGID.find(wUni);
	if (it == m_CodeToGID.end())
		return VGlyphSource::keError;

	if (m_pLoader)
	{
		TopDicts* pTopDict = m_pLoader->OTFData.vpTopDicts[0];

		if (it->second >= (int)pTopDict->Glyphs.vValues.size())
			return VGlyphSource::keError;

		*pulNumCoords = (ULONG)pTopDict->Glyphs.vValues[it->second].size();
		if (*pulNumCoords)
		{
			*verb = &pTopDict->Glyphs.vValues[it->second][0];
			*coord = NULL;
			return VGlyphSource::kePackedCFF;
		}
		return VGlyphSource::keError;
	}

	*pulNumCoords = (ULONG)m_chars[it->second].verb.size();

	if (*pulNumCoords)
	{
		*verb = &(m_chars[it->second].verb[0]);
		*coord = &(m_chars[it->second].coord[0]);
	}

	return VGlyphSource::keEmbedded;
}

LONG SFNTReader::GetCharWidth(WCHAR wUni)
{
	GIDMap::iterator it = m_CodeToGID.find(wUni);
	if (it == m_CodeToGID.end())
		return 0;

	return ttf2xar_l((double)m_hmtx[it->second].advance);
}

LONG SFNTReader::GetFontAscent()
{
	return ttf2xar_l(m_hhea.Ascender);
}

LONG SFNTReader::GetFontDescent()
{
	return ttf2xar_l(m_hhea.Descender);
}

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// VGlyphSource2 implementation

RECT SFNTReader::GetFontBBox() 
{ 
	RECT rc;
	memset(&rc, 0, sizeof(RECT));

	if (m_pLoader && m_pLoader->OTFData.vpTopDicts.size())
	{
		FLOAT* box = m_pLoader->OTFData.vpTopDicts[0]->fFontBoundingBox;
		rc.left = ttf2xar_l(box[0]);
		rc.top = ttf2xar_l(box[1]);
		rc.right = ttf2xar_l(box[2]);
		rc.bottom = ttf2xar_l(box[3]);
	}
	return rc;
}

LONG SFNTReader::GetFontCapHeight()
{
	return ttf2xar_l(m_OS2.sCapHeight);
}

LONG SFNTReader::GetFontXHeight()
{
	return ttf2xar_l(m_OS2.sxHeight);
}

LONG SFNTReader::GetFontNominalWidth()
{
	if (m_pLoader)
	{
		if (m_pLoader->OTFData.vpSubDicts.size())
			return ttf2xar_l(m_pLoader->OTFData.vpSubDicts[0]->iNominalWidthX);
		if (m_pLoader->OTFData.vpTopDicts.size())
			return ttf2xar_l(m_pLoader->OTFData.vpTopDicts[0]->iNominalWidthX);
	}
	return  0;
}

LONG SFNTReader::GetFontDefaultWidth()
{
	if (m_pLoader)
	{
		if (m_pLoader->OTFData.vpSubDicts.size())
			return ttf2xar_l(m_pLoader->OTFData.vpSubDicts[0]->iDefaultWidthX);
		if (m_pLoader->OTFData.vpTopDicts.size())
			return ttf2xar_l(m_pLoader->OTFData.vpTopDicts[0]->iDefaultWidthX);
	}
	return  0;
}

bool SFNTReader::HasLocalSubs()
{
	return m_pLoader && m_pLoader->OTFData.vpTopDicts.size() && m_pLoader->OTFData.vpTopDicts[0]->iSubrsOff > -1 ? true : false;
}

bool SFNTReader::GetLocalSubs(PBYTE& pSubs, size_t& szSubSize)
{
	if (m_pLoader && m_pLoader->OTFData.vpTopDicts.size())
	{
		TopDicts* pTopDict = m_pLoader->OTFData.vpTopDicts[0];
		PBYTE pbyCFF = map()[TABLEID_CFF]->GetMemoryBuffer();
		szSubSize = (size_t)pTopDict->iLocalSubSize;
		pSubs = pbyCFF + pTopDict->iPrivateOffset + pTopDict->iSubrsOff;
		return true;
	}
	return false;
}

bool SFNTReader::GetStdHW(int& iStdHW)
{
	if (m_pLoader)
	{
		if (m_pLoader->OTFData.vpSubDicts.size())
		{
			iStdHW = m_pLoader->OTFData.vpSubDicts[0]->iStdHw;
			return true;
		}
		if (m_pLoader->OTFData.vpTopDicts.size())
		{
			iStdHW = m_pLoader->OTFData.vpTopDicts[0]->iStdHw;
			return true;
		}
	}
	return  false;
}

bool SFNTReader::GetStdVW(int& iStdVW)
{
	if (m_pLoader)
	{
		if (m_pLoader->OTFData.vpSubDicts.size())
		{
			iStdVW = m_pLoader->OTFData.vpSubDicts[0]->iStdVw;
			return true;
		}
		if (m_pLoader->OTFData.vpTopDicts.size())
		{
			iStdVW = m_pLoader->OTFData.vpTopDicts[0]->iStdVw;
			return true;
		}
	}
	return  false;
}

} // namespace SFNT
