/********************************************************************************************

>	ttf.cpp

Author:		Paul Gruzdev
Created:	08/11/2016
Purpose:	TTF: writing old-style tables that everybody understands: 'glyf' and 'loca'. Also TTF
			requires GASP and 'CVT ' along. Those contain glyph outline data represented in TrueType 
			format, as a set of cubic splines.
			// The most interesting thing this code does,
			//  it does is to figure out a quadratic approximation to the cubic splines 
			//  that PostScript (or Xara) uses. We do this by looking at each spline and running 
			//  from the end toward the beginning, checking approximately every em unit.
			//  There is only one quadratic spline possible for any given interval of the 
			//  cubic. The start and end points are the interval end points (obviously) 
			//  the control point is where the two slopes (at start and end) intersect. 
			// If this spline is a close approximation to the cubic spline (doesn't 
			//  deviate from it by more than an em unit or so), then we use this interval 
			//  as one of our quadratic splines. 
			// It may turn out that the "quadratic" spline above is actually linear. Well 
			//  that's ok. It may also turn out that we can't find a good approximation. 
			//  If that's true then just insert a linear segment for an em unit stretch. 
			//  (actually this failure mode may not be possible), but I'm not sure 
			// Then we play the same trick for the rest of the cubic spline (if any) 

********************************************************************************************/

#include "stdafx.h"
#include "sfnt.h"
#include "ttf.h"

// Enables improved memory tracking. Must be defined after CC_IMPLEMENT_DYNCREATE macros.
#ifdef CAMELOT_PRODUCT
	#define new CAM_DEBUG_NEW
#endif

// An IEEE double has 52 bits of precision. So one unit of rounding error will be
//  the number divided by 2^51
#define D_RE_Factor	(1024.0*1024.0*1024.0*1024.0*1024.0*2.0)

namespace SFNT
{

static DocCoord ttf2xar_dp(DPoint& dp)
{
	double x = dp.x*g_fXaraEM / TTF_EM_HEIGHT;
	double y = dp.y*g_fXaraEM / TTF_EM_HEIGHT;
	return DocCoord((LONG)rint(x), (LONG)rint(y));
}

static DocCoord xar2ttf_dp(DPoint& dp)
{
	double x = dp.x*TTF_EM_HEIGHT / g_fXaraEM;
	double y = dp.y*TTF_EM_HEIGHT / g_fXaraEM;
	return DocCoord((LONG)rint(x), (LONG)rint(y));
}

static DocCoord dp2dc(DPoint& dp)
{
	return DocCoord((LONG)rint(dp.x), (LONG)rint(dp.y));
}

static DocCoord xar2ttf_dc(DocCoord& dc)
{
	double x = (double)dc.x*TTF_EM_HEIGHT / g_fXaraEM;
	double y = (double)dc.y*TTF_EM_HEIGHT / g_fXaraEM;
	return DocCoord((LONG)rint(x), (LONG)rint(y));
}

static DPoint xar2ttf_dcdp(DocCoord& dc)
{
	double x = (double)dc.x*TTF_EM_HEIGHT / g_fXaraEM;
	double y = (double)dc.y*TTF_EM_HEIGHT / g_fXaraEM;
	return DPoint(x, y);
}

static DPoint xar2ttf_dpdp(DPoint& dc)
{
	float x = dc.x*TTF_EM_HEIGHT / (float)g_fXaraEM;
	float y = dc.y*TTF_EM_HEIGHT / (float)g_fXaraEM;
	return DPoint(x, y);
}

static LONG xar2ttf_l(LONG l)
{
	double x = ((double)l)*TTF_EM_HEIGHT / g_fXaraEM;
	return (LONG)rint(x);
}

static inline double xar2ttf_ld(LONG l)
{
	return ((double)l)*TTF_EM_HEIGHT / g_fXaraEM;
}
/*
void AppendFile(const char* file, const wchar_t* fmt, ...)
{
	static std::set<const char*> sInitialised;

	if(sInitialised.find(file) == sInitialised.end())
	{
		_unlink(file);
		sInitialised.insert(file);
	}

	FILE* in=0; fopen_s(&in, file, "at");
	if(in)
	{
		va_list argList;
		va_start(argList, fmt);
		int res = vfwprintf(in, fmt, argList);
		va_end(argList);
		fclose(in);
	}
}

static void DumpQData(Spline* spline, QPoint* pt, int cnt)
{
	int i;
	AppendFile("d:/temp/ttf/qp_xara.txt", L"----------------\n");
	AppendFile("d:/temp/ttf/qp_xara.txt", spline->Dump().c_str());
	for(i = 1; i < cnt; i++)
	{
		DPoint cp(pt[i-1].cp);
		DPoint bp(pt[i].bp);
		AppendFile("d:/temp/ttf/qp_xara.txt", L"cp: %.3f %.3f bp: %.3f %.3f\n", cp.x, cp.y, bp.x, bp.y);
	}
}

static void DumpSpline(wchar_t* psz, wchar_t* pszName, Spline* spline)
{
	wchar_t psz1[2048];
	swprintf_s(psz, 4096, L"%s:(%.3f %.3f)->(%.3f %.3f) (%.3f %.3f) (%.3f %.3f)\n", pszName,
		(float)spline->from.x, (float)spline->from.y,
		(float)spline->to.x, (float)spline->to.y,
		(float)spline->cp1.x, (float)spline->cp1.y,
		(float)spline->cp2.x, (float)spline->cp2.y);
	swprintf_s(psz1, 2048, L"a(%.3f %.3f) b(%.3f %.3f) c(%.3f %.3f) d(%.3f %.3f)\n", (float)spline->splines[0].a, (float)spline->splines[1].a,
		(float)spline->splines[0].b, (float)spline->splines[1].b, (float)spline->splines[0].c, (float)spline->splines[1].c,
		(float)spline->splines[0].d, (float)spline->splines[1].d);

	wcscat_s(psz, 4096, psz1);
}

static void DumpCmpData(Spline* spline, Spline* ttf, int bEq)
{
	int i;
	wchar_t psz[4096];

	AppendFile("d:/temp/ttf/qp_xara.txt", L"Compare: ----------------\n");
	DumpSpline(psz, L"Spline1", spline);
	AppendFile("d:/temp/ttf/qp_xara.txt", psz);
	DumpSpline(psz, L"ttf", ttf);
	AppendFile("d:/temp/ttf/qp_xara.txt", psz);
	AppendFile("d:/temp/ttf/qp_xara.txt", L"*** %s ***\n", bEq ? L"EQUAL" : L"Not Eq");
}

static void DumpPoints(DocCoord* bp, BYTE* fs, int ptcnt)
{
	AppendFile("d:/temp/ttf/points_xara.txt", L"-----------------------------\n");
	for (int i = 0; i < ptcnt; i++)
		AppendFile("d:/temp/ttf/points_xara.txt", L"%d %d %d\n", fs[i], bp[i].x, bp[i].y);
}
*/
static int floatWithin(float a, float b, float fudge) 
{
	return b >= a - fudge && b <= a + fudge;
}

bool floatNear(float a, float b) 
{
	if (a == 0)
		return b > -1e-5 && b < 1e-5;

	if (b == 0)
		return a > -1e-5 && a < 1e-5;

	float d = a/(1024*64.f);
	a -= b;

	if (d < 0)
		return a > d && a < -d;
	else
		return a > -d && a < d;
}

inline bool doubleNear(double a, double b) 
{
	return floatNear((float)a, (float)b);
}

int floatNearish(float a, float b, float precision = 0.001)
{
	a -= b;
	return  a <= precision && a >= -precision;
}

int doubleApprox(double a, double b) 
{
	if (a == 0) 
	{
		if (b < .0001 && b > -.0001)
			return true;
	} 
	else if (b == 0)
	{
		if (a < .0001 && a > -.0001)
			return true;
	} 
	else 
	{
		a /= b;
		if (a >= .95 && a <= 1.05)
			return true;
	}

	return false;
}

bool Within16RoundingErrors(double v1, double v2) 
{
	double temp=v1*v2;
	double re;

	if (temp < 0) // Ok, if the two values are on different sides of 0 there
		return false;	// is no way they can be within a rounding error of each other
	else if (temp == 0) 
	{
		if (v1 == 0)
			return v2 < RE_NearZero && v2 > -RE_NearZero;
		else
			return v1 < RE_NearZero && v1 > -RE_NearZero;
	} 
	else if (v1 > 0) 
	{
		if (v1 > v2) // Rounding error from the biggest absolute value
		{
			re = v1 / (RE_Factor/16);
			return v1-v2 < re;
		} 
		else 
		{
			re = v2/ (RE_Factor/16);
			return v2-v1 < re;
		}
	} 
	else 
	{
		if (v1 < v2) 
		{
			re = v1/ (RE_Factor/16);	// This will be a negative number
			return v1 - v2 > re;
		} 
		else 
		{
			re = v2/ (RE_Factor/16);
			return v2 - v1 > re;
		}
	}
}

/******************************************************************************
   Function:	AlreadyQuadratic
   Author:		Paul Gruzdev
   Purpose:		Test whether this cubic bezier is already quadratic B-spline,
				and calculate the control point of the B-spline
   Returns:		bool
   Parameters:	DocCoord from
   Parameters:	DocCoord * pBezier
   Parameters:	DPoint & cubic
   Notes:       
******************************************************************************/
bool AlreadyQuadratic(DocCoord from, DocCoord* pBezier, DPoint& cubic)
{
	// from p. 393 (Operator Details, curveto) PostScript Lang. Ref. Man. (Red book)
	// x(t) = a*x*t^3 + b*x*t^2 + c*x*t + x0
	// y(t) = a*y*t^3 + b*y*t^2 + c*y*t + y0
/*
	float cx = 3.0f*(pBezier[0].x - from.x);
	float cy = 3.f*(pBezier[0].y - from.y);
	float bx = 3.f*(pBezier[1].x - pBezier[0].x) - cx;
	float by = 3.f*(pBezier[1].y - pBezier[0].y) - cy;
	float ax = pBezier[2].x - from.x - cx - bx;
	float ay = pBezier[2].y - from.y - cy - by;

	cubic.x = (cx + 2*from.x)/2;
	cubic.y = (cy + 2*from.y)/2;

	if ((floatNearish(ax, 0, 3) && floatNearish(ay, 0, 3)) ||
		(bx != 0 && floatNearish(ax / bx, 0, 1)) && (by != 0 && floatNearish(ay / by, 0, 1)))
		return true; // We passed the check
*/

	Spline ps(xar2ttf_dcdp(from), xar2ttf_dcdp(pBezier[0]), xar2ttf_dcdp(pBezier[1]), xar2ttf_dcdp(pBezier[2]));

	if ((floatNearish(ps.splines[0].a, 0) && floatNearish(ps.splines[1].a, 0)) ||
		((ps.splines[0].b != 0 && floatNearish(ps.splines[0].a / ps.splines[0].b, 0)) &&
		(ps.splines[1].b != 0 && floatNearish(ps.splines[1].a / ps.splines[1].b, 0))))
	{
		float cx = 3.0f*(pBezier[0].x - from.x);
		float cy = 3.f*(pBezier[0].y - from.y);
		cubic.x = (cx + 2 * from.x) / 2;
		cubic.y = (cy + 2 * from.y) / 2;
		return true;
	}

	return false;
}

/******************************************************************************
   Function:	CanInterpolate
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	DPoint prevCP
   Parameters:	DPoint me
   Parameters:	DPoint nextCP
   Notes:       Now we are already in TrueType units (NOT Xara units)
******************************************************************************/
bool CanInterpolate(DPoint prevCP, DPoint me, DPoint nextCP)
{
	// Using truetype rules, can we interpolate this point?
	return floatWithin(me.x, (nextCP.x + prevCP.x) / 2.0f, 0.1f) && floatWithin(me.y, (nextCP.y + prevCP.y) / 2.0f, 0.1f);
}

inline static void UpdateBox(DocCoord dc, RECT& rc)
{
	if(dc.x < rc.left) rc.left = dc.x;
	if(dc.x > rc.right) rc.right = dc.x;
	if(dc.y < rc.top) rc.top = dc.y;
	if(dc.y > rc.bottom) rc.bottom = dc.y;
}

/******************************************************************************
   Function:	Spline
   Author:		Paul Gruzdev
   Purpose:		Spline class constructor. Calculate a, b, c, d of cubic spline
   Returns:		
   Parameters:	DPoint dpFrom
   Parameters:	DPoint cp1_
   Parameters:	DPoint cp2_
   Parameters:	DPoint dpTo
   Notes:       
******************************************************************************/
Spline::Spline(DPoint dpFrom, DPoint cp1_, DPoint cp2_, DPoint dpTo) : from(dpFrom), to(dpTo), cp1(cp1_), cp2(cp2_)
{
	Spline1D* xsp = &splines[0], *ysp = &splines[1];

	xsp->d = dpFrom.x; 
	ysp->d = dpFrom.y;

	// from p. 393 (Operator Details, curveto) PostScript Lang. Ref. Man. (Red book)
	xsp->c = 3*(cp1.x - from.x);
	ysp->c = 3*(cp1.y - from.y);
	xsp->b = 3*(cp2.x - cp1.x) - xsp->c;
	ysp->b = 3*(cp2.y - cp1.y) - ysp->c;
	xsp->a = to.x - from.x - xsp->c - xsp->b;
	ysp->a = to.y - from.y - ysp->c - ysp->b;

	if (floatNear(xsp->c, 0)) xsp->c=0;
	if (floatNear(ysp->c, 0)) ysp->c=0;
	if (floatNear(xsp->b, 0)) xsp->b=0;
	if (floatNear(ysp->b, 0)) ysp->b=0;
	if (floatNear(xsp->a, 0)) xsp->a=0;
	if (floatNear(ysp->a, 0)) ysp->a=0;

	if (xsp->a != 0 && (Within16RoundingErrors(xsp->a + from.x, from.x) ||
		Within16RoundingErrors(xsp->a + to.x, to.x)))
		xsp->a = 0;

	if (ysp->a != 0 && (Within16RoundingErrors(ysp->a + from.y, from.y) ||
		Within16RoundingErrors(ysp->a + to.y, to.y)))
		ysp->a = 0;
}

static double CheckExtremaForSingleBitErrors(const Spline1D* sp, double t, double othert) 
{
	double u1, um1;
	double slope, slope1, slopem1;
	int err;
	double diff, factor;

	if (t < 0 || t > 1)
		return t;

	factor = t*0x40000/D_RE_Factor;
	if ((diff = t-othert) < 0) 
		diff = -diff;
	if (factor > diff/4 && diff != 0)		// This little check is to insure we don't skip beyond the well of this extremum into the next
		factor = diff/4;

	slope = (3*(double) sp->a*t+2*sp->b)*t+sp->c;
	if (slope < 0)
		slope = -slope;

	for (err = 0x40000; err!=0; err >>= 1)
	{
		u1 = t + factor;
		slope1 = (3*(double) sp->a*u1+2*sp->b)*u1+sp->c;

		if (slope1 < 0)
			slope1 = -slope1;

		um1 = t-factor;
		slopem1 = (3*(double) sp->a*um1+2*sp->b)*um1+sp->c;
		if (slopem1 < 0) 
			slopem1 = -slopem1;

		if (slope1 < slope && slope1 <= slopem1 && u1 <= 1.0)
		{
			t = u1;
		} 
		else if (slopem1 < slope && slopem1 <= slope1 && um1 >= 0.0)
		{
			t = um1;
		}
		factor /= 2.0;
	}
	// that seems as good as it gets

	return t;
}

void Spline::FindExtrema(const Spline1D* sp, double *_t1, double *_t2)
{
	double t1= -1, t2= -1;
	double b2_fourac;

	// Find the extreme points on the curve 
	//  Set to -1 if there are none or if they are outside the range [0,1] 
	//  Order them so that t1<t2 
	//  If only one valid extremum it will be t1 
	//  (Does not check the end points unless they have derivative==0) 
	//  (Does not check to see if d/dt==0 points are inflection points (rather than extrema) 
	if (sp->a != 0) 
	{
		// cubic, possibly 2 extrema (possibly none)
		b2_fourac = 4*(double) sp->b*sp->b - 12*(double) sp->a*sp->c;

		if (b2_fourac >= 0)
		{
			b2_fourac = sqrt(b2_fourac);
			t1 = (-2*sp->b - b2_fourac) / (6*sp->a);
			t2 = (-2*sp->b + b2_fourac) / (6*sp->a);
			t1 = CheckExtremaForSingleBitErrors(sp, t1, t2);
			t2 = CheckExtremaForSingleBitErrors(sp, t2, t1);

			if (t1 > t2)
			{
				double temp = t1;
				t1 = t2;
				t2 = temp;
			}
			else if ( t1==t2 ) 
				t2 = -1;

			if (doubleNear(t1, 0)) 
				t1 = 0;
			else if (doubleNear(t1, 1)) 
				t1 = 1;

			if (doubleNear(t2, 0)) 
				t2 = 0; 
			else if (doubleNear(t2, 1)) 
				t2 = 1;

			if (t2 <= 0 || t2 >= 1) 
				t2 = -1;

			if (t1 <= 0 || t1 >= 1)
			{ 
				t1 = t2;
				t2 = -1;
			}
		}
	} 
	else if (sp->b != 0)
	{
		// Quadratic, at most one extremum
		t1 = -sp->c / (2.0*(double)sp->b);
		if (t1 <= 0 || t1 >= 1)
			t1 = -1;
	} 
	else //if ( sp->c!=0 )
	{
		// linear, no extrema
	}

	*_t1 = t1; 
	*_t2 = t2;
}

bool Spline::MinMaxWithin()
{
	// We know that this "spline" is basically one dimensional. As long as its
	//  extrema are between the start and end points on that line then we can 
	//  treat it as a line. If the extrema are way outside the line segment 
	//  then it's a line that backtracks on itself 

	double dx = to.x - from.x;
	double dy = to.y - from.y;
	if (dx < 0) dx = -dx;
	if (dy < 0) dy = -dy;
	int which = dx < dy ? 1 : 0;

	double t1, t2;
	FindExtrema(&splines[which], &t1, &t2);

	if (t1 == -1)
		return true;

	double w = ((splines[which].a*t1 + splines[which].b)*t1 + splines[which].c)*t1 + splines[which].d;

	if (floatNear((float)w, (&to.x)[which]) || floatNear((float)w, (&from.x)[which]) )
		// Close enough
		;
	else if ((w < (&to.x)[which] && w < (&from.x)[which]) || (w > (&to.x)[which] && w > (&from.x)[which]) )
		return false;		// Outside

	w = ((splines[which].a*t2 + splines[which].b)*t2 + splines[which].c)*t2 + splines[which].d;

	if (floatNear((float)w, (&to.x)[which]) || floatNear((float)w, (&from.x)[which]))
		// Close enough
		;
	else if ((w < (&to.x)[which] && w < (&from.x)[which]) || (w > (&to.x)[which] && w > (&from.x)[which]) )
		return false;		// Outside

	return true;
}

std::wstring Spline::Dump()
{
	wchar_t psz[4096], psz1[2048];
	swprintf_s(psz, 4096, L"Spline:(%.3f %.3f)->(%.3f %.3f) (%.3f %.3f) (%.3f %.3f)\n", (float)from.x, (float)from.y, (float)to.x, (float)to.y, 
		(float)cp1.x, (float)cp1.y, (float)cp2.x, (float)cp2.y);
	swprintf_s(psz1, 2048, L"a(%.3f %.3f) b(%.3f %.3f) c(%.3f %.3f) d(%.3f %.3f)\n", (float)splines[0].a, (float)splines[1].a, 
		(float)splines[0].b, (float)splines[1].b, (float)splines[0].c, (float)splines[1].c, (float)splines[0].d, (float)splines[1].d);
	wcscat_s(psz, 4096, psz1);
	return psz;
}

/******************************************************************************
   Function:	IsLinear
   Author:		Paul Gruzdev
   Purpose:		Test spline's control points and coeffs and decide it's actually a line
   Returns:		bool
   Notes:       
******************************************************************************/
bool Spline::IsLinear()
{
	bool ret = false;

	if (splines[0].a == 0 && splines[0].b == 0 && splines[1].a == 0 && splines[1].b == 0)
		return true;

	// Something is linear if the control points lie on the line between the two base points

	// Vertical lines
	if (floatNear(from.x, to.x))
	{
		ret = floatNear(from.x, cp1.x) && floatNear(from.x, cp2.x);

		if ( !(
			(cp1.y >= from.y && cp1.y <= to.y && cp2.y >= from.y && cp2.y <= to.y ) ||
			(cp1.y <= from.y && cp1.y >= to.y && cp2.y <= from.y && cp2.y >= to.y )) )
			ret = MinMaxWithin();
	}
	// Horizontal lines
	else if (floatNear(from.y, to.y))
	{
		ret = floatNear(from.y,cp1.y) && floatNear(from.y,cp2.y);

		if ( !(
			(cp1.x >= from.x && cp1.x <= to.x && cp2.x >= from.x && cp2.x <= to.x) ||
			(cp1.x <= from.x && cp1.x >= to.x && cp2.x <= from.x && cp2.x >= to.x)) )
			ret = MinMaxWithin();
	}
	else
	{
		double t1 = (cp1.y-from.y)/(to.y-from.y);
		double t2 = (cp1.x-from.x)/(to.x-from.x);
		double t3 = (to.y-cp2.y)/(to.y-from.y);
		double t4 = (to.x-cp2.x)/(to.x-from.x);

		ret = (Within16RoundingErrors(t1, t2) || (doubleApprox(t1, 0) && doubleApprox(t2, 0))) &&
			(Within16RoundingErrors(t3, t4) || (doubleApprox(t3, 0) && doubleApprox(t4, 0)));

		if (ret) 
		{
			if (t1 < 0 || t2 < 0 || t3 < 0 || t4 < 0 || t1 > 1 || t2 > 1 || t3 > 1 || t4 > 1)
				ret = MinMaxWithin();
		}
	}

	if (ret)
	{
		// A few places that if the spline is knownlinear then its splines[?]
		//  are linear. So give the linear version and not that suggested by
		//  the control points
		splines[0].a = splines[0].b = 0;
		splines[0].d = from.x;
		splines[0].c = to.x - from.x;
		splines[1].a = splines[1].b = 0;
		splines[1].d = from.y;
		splines[1].c = to.y - from.y;
	}

	return ret;
}

// This returns all real solutions, even those out of bounds 
// I use -999999 as an error flag, since we're really only interested in 
//  solns near 0 and 1 that should be ok. -1 is perhaps a little too close 
// Sigh. When solutions are near 0, the rounding errors are appalling.
int _CubicSolve(const Spline1D* sp, double sought, double ts[3]) 
{
	double d, xN, yN, delta2, temp, delta, h, t2, t3, theta;
	double sa=sp->a, sb=sp->b, sc=sp->c, sd=sp->d-sought;
	int i=0;

	ts[0] = ts[1] = ts[2] = -999999;
	if (sd == 0 && sa != 0) 
	{
		// one of the roots is 0, the other two are the soln of a quadratic 
		ts[0] = 0;
		if (sc == 0) 
		{
			ts[1] = -sb/(double) sa;	// two zero roots
		} 
		else 
		{
			temp = sb*(double) sb-4*(double) sa*sc;
			if (floatNear((float)temp, 0))
				ts[1] = -sb/(2*(double) sa);
			else if (temp >= 0)
			{
				temp = sqrt(temp);
				ts[1] = (-sb+temp)/(2*(double) sa);
				ts[2] = (-sb-temp)/(2*(double) sa);
			}
		}
	} 
	else if (sa != 0)
	{
		// http://www.m-a.org.uk/eb/mg/mg077ch.pdf
		// this nifty solution to the cubic neatly avoids complex arithmetic
		xN = -sb/(3*(double) sa);
		yN = ((sa*xN + sb)*xN+sc)*xN + sd;

		delta2 = (sb*(double) sb-3*(double) sa*sc)/(9*(double) sa*sa);
		//if ( RealWithin(delta2,0,.00000001) ) delta2 = 0;

		// the discriminant is yN^2-h^2, but delta might be <0 so avoid using h
		d = yN*yN - 4*sa*sa*delta2*delta2*delta2;
		if (((yN > .01 || yN < -.01) && floatNear((float)(d/yN), 0)) || ((yN <= .01 && yN >= -.01) && floatNear((float)d, 0)))
			d = 0;

		if ( d>0 ) 
		{
			temp = sqrt(d);
			t2 = (-yN-temp)/(2*sa);
			t2 = (t2==0) ? 0 : (t2<0) ? -pow(-t2,1./3.) : pow(t2,1./3.);
			t3 = (-yN+temp)/(2*sa);
			t3 = t3==0 ? 0 : (t3<0) ? -pow(-t3,1./3.) : pow(t3,1./3.);
			ts[0] = xN + t2 + t3;
		} 
		else if (d < 0) 
		{
			if (delta2 >= 0) 
			{
				delta = sqrt(delta2);
				h = 2*sa*delta2*delta;
				temp = -yN/h;
				if (temp >= -1.0001 && temp <= 1.0001)
				{
					if (temp < -1)
						temp = -1; 
					else if (temp > 1) 
						temp = 1;

					theta = acos(temp) / 3;
					ts[i++] = xN+2*delta*cos(theta);
					ts[i++] = xN+2*delta*cos(2.0943951 + theta);	// 2*pi/3
					ts[i++] = xN+2*delta*cos(4.1887902 + theta);	// 4*pi/3
				}
			}
		} 
		else if ( /* d==0 && */ delta2 != 0)
		{
			delta = yN/(2*sa);
			delta = delta==0 ? 0 : delta>0 ? pow(delta,1./3.) : -pow(-delta,1./3.);
			ts[i++] = xN + delta;	// this root twice, but that's irrelevant to me
			ts[i++] = xN - 2*delta;
		} 
		else if ( /* d==0 && */ delta2 == 0) 
		{
			if (xN >= -0.0001 && xN <= 1.0001)
				ts[0] = xN;
		}
	}
	else if (sb != 0)
	{
		double d = sc * (double)sc - 4*(double)sb*sd;
		if (d < 0 && floatNear((float)d, 0)) 
			d = 0;
		if (d < 0)
			return false;		// All roots imaginary
		d = sqrt(d);
		ts[0] = (-sc-d)/(2*(double) sb);
		ts[1] = (-sc+d)/(2*(double) sb);
	} 
	else if (sc != 0) 
	{
		ts[0] = -sd/(double) sc;
	} 
	else 
	{
		// If it's a point then either everything is a solution, or nothing
	}

	return ts[0] != -999999;
}

int CubicSolve(const Spline1D* sp, double sought, double ts[3])
{
	double t;
	double ts2[3];
	int i,j;
	// This routine gives us all solutions between [0,1] with -1 as an error flag
	// http://mathforum.org/dr.math/faq/faq.cubic.equations.html

	ts[0] = ts[1] = ts[2] = -1;
	if (!_CubicSolve(sp, sought, ts2))
		return false;

	for (i = j = 0; i < 3; ++i)
	{
		if (ts2[i] > -.0001 && ts2[i] < 1.0001)
		{
			if (ts2[i] < 0) 
				ts[j++] = 0;
			else if (ts2[i] > 1)
				ts[j++] = 1;
			else
				ts[j++] = ts2[i];
		}
	}

	if (j == 0)
		return false;

	if ( ts[0]>ts[2] && ts[2]!=-1 ) 
	{
		t = ts[0]; ts[0] = ts[2]; ts[2] = t;
	}

	if ( ts[0]>ts[1] && ts[1]!=-1 ) 
	{
		t = ts[0]; ts[0] = ts[1]; ts[1] = t;
	}

	if ( ts[1]>ts[2] && ts[2]!=-1 )
	{
		t = ts[1]; ts[1] = ts[2]; ts[2] = t;
	}

	return true;
}

/******************************************************************************
   Function:	CompareSpline
   Author:		Paul Gruzdev
   Purpose:		Does the quadratic spline in ttf approximate the cubic spline in ps 
				within err between tmin and tmax (on ps. presumably ttf between 0&1 
				dim is the dimension in which there is the greatest change).

   Returns:		int
   Parameters:	Spline * ps
   Parameters:	Spline * ttf
   Parameters:	double tmin
   Parameters:	double tmax
   Parameters:	double err
   Notes:       
******************************************************************************/
static int CompareSpline(Spline* ps, Spline* ttf, float tmin, float tmax, float err)
{
	int dim=0, other;
	float dx, dy, ddim, dt, t;
	float d, o;
	float ttf_t, sq, val;
	DBounds bb;
	double ts[3];
	int i;

	// Are all points on ttf near points on ps?
	// This doesn't answer that question, but rules out gross errors
	bb.minx = bb.maxx = ps->from.x; 
	bb.miny = bb.maxy = ps->from.y;

	if (ps->cp1.x > bb.maxx)
		bb.maxx = ps->cp1.x;
	else 
		bb.minx = ps->cp1.x;

	if (ps->cp1.y > bb.maxy)
		bb.maxy = ps->cp1.y;
	else 
		bb.miny = ps->cp1.y;

	if (ps->cp2.x > bb.maxx)
		bb.maxx = ps->cp2.x;
	else if ( ps->cp2.x < bb.minx) 
		bb.minx = ps->cp2.x;

	if ( ps->cp2.y > bb.maxy ) 
		bb.maxy = ps->cp2.y;
	else if ( ps->cp2.y<bb.miny ) 
		bb.miny = ps->cp2.y;

	if ( ps->to.x>bb.maxx ) 
		bb.maxx = ps->to.x;
	else if ( ps->to.x<bb.minx ) 
		bb.minx = ps->to.x;

	if ( ps->to.y>bb.maxy ) 
		bb.maxy = ps->to.y;
	else if ( ps->to.y<bb.miny ) 
		bb.miny = ps->to.y;

	for (t = .1f; t < 1; t += .1f) 
	{
		d = (ttf->splines[0].b*t+ttf->splines[0].c)*t+ttf->splines[0].d;
		o = (ttf->splines[1].b*t+ttf->splines[1].c)*t+ttf->splines[1].d;
		if (d < bb.minx || d > bb.maxx || o < bb.miny || o > bb.maxy)
			return false;
	}

	// Are all points on ps near points on ttf?
	dx = ((ps->splines[0].a*tmax+ps->splines[0].b)*tmax+ps->splines[0].c)*tmax -
		((ps->splines[0].a*tmin+ps->splines[0].b)*tmin+ps->splines[0].c)*tmin ;
	dy = ((ps->splines[1].a*tmax+ps->splines[1].b)*tmax+ps->splines[1].c)*tmax -
		((ps->splines[1].a*tmin+ps->splines[1].b)*tmin+ps->splines[1].c)*tmin ;

	if ( dx<0 ) dx = -dx;
	if ( dy<0 ) dy = -dy;

	if (dx > dy)
	{
		dim = 0;
		ddim = dx;
	} 
	else 
	{
		dim = 1;
		ddim = dy;
	}
	other = !dim;

	t = tmin;
	dt = (tmax-tmin)/ddim;
	for (t = tmin; t <= tmax; t += dt)
	{
		if (t > tmax - dt/8.)
			t = tmax;		// Avoid rounding errors

		d = ((ps->splines[dim].a*t+ps->splines[dim].b)*t+ps->splines[dim].c)*t+ps->splines[dim].d;
		o = ((ps->splines[other].a*t+ps->splines[other].b)*t+ps->splines[other].c)*t+ps->splines[other].d;

		if ( ttf->splines[dim].b == 0 )
		{
			ttf_t = (d-ttf->splines[dim].d)/ttf->splines[dim].c;
		} 
		else 
		{
			sq = ttf->splines[dim].c*ttf->splines[dim].c - 4*ttf->splines[dim].b*(ttf->splines[dim].d-d);

			if (sq < 0)
				return false;

			sq = sqrt(sq);
			ttf_t = (-ttf->splines[dim].c-sq)/(2*ttf->splines[dim].b);
			if ( ttf_t>=-0.1 && ttf_t<=1.1 ) // Optimizer gives us rounding errors
			{
				// And tmin/tmax are no longer exact
				val = (ttf->splines[other].b * ttf_t + ttf->splines[other].c) * ttf_t + ttf->splines[other].d;
				if (val > o - err && val < o + err)
					continue;
			}
			ttf_t = (-ttf->splines[dim].c+sq)/(2*ttf->splines[dim].b);
		}

		if (ttf_t >= -0.1 && ttf_t <= 1.1)
		{
			val = (ttf->splines[other].b*ttf_t+ttf->splines[other].c)*ttf_t + ttf->splines[other].d;
			if ( val>o-err && val<o+err )
				continue;
		}

		return false;
	}

	// Are representative points on ttf near points on ps?
	for (t = .125; t < 1; t += .125)
	{
		d = (ttf->splines[dim].b*t+ttf->splines[dim].c)*t+ttf->splines[dim].d;
		o = (ttf->splines[other].b*t+ttf->splines[other].c)*t+ttf->splines[other].d;
		CubicSolve(&ps->splines[dim],d,ts);
		for (i = 0; i < 3; ++i)
		{
			if ( ts[i]!=-1 ) 
			{
				val = (float)
					(((ps->splines[other].a*ts[i] + ps->splines[other].b)*ts[i] + ps->splines[other].c)*ts[i] + ps->splines[other].d);
				if (val > o-err && val < o+err)
					break;
			}
		}
		if (i == 3)
			return false;
	}

	return true;
}

/******************************************************************************
   Function:	CompareData
   Author:		Paul Gruzdev
   Purpose:		
   Returns:		bool
   Parameters:	Spline * ps
   Parameters:	QPoint * data
   Parameters:	int qfirst
   Parameters:	int qlast
   Parameters:	bool bRoundToInt
   Parameters:	int iTestLevel
   Notes:       
******************************************************************************/
static bool CompareData(Spline* ps, QPoint* data, int qfirst,int qlast, bool bRoundToInt, int iTestLevel)
{
	Spline ttf;
	int i;
	// double err = bRoundToInt ? 52.734375 : 35.15625; // 35.1 in Xara units is 1 unit in ttf space [0..2048]
	// but now we approximate in ttf space
	double err = bRoundToInt ? 1.5 : 1;

	if (qfirst == qlast)		// happened (was a bug)
		return false;

	err *= (iTestLevel + 1);

	// Control points diametrically opposed
	if ((data[qlast-2].cp.x - ps->to.x) * (ps->cp2.x - ps->to.x) + (data[qlast-2].cp.y - ps->to.y) * (ps->cp2.y - ps->to.y) < 0)
		return false;

	if ((data[qfirst-1].cp.x - ps->from.x) * (ps->cp1.x - ps->from.x) + 
		(data[qfirst-1].cp.y - ps->from.y) * (ps->cp1.y - ps->from.y) < 0)
		return false;

	for (i = qfirst; i < qlast; ++i)
	{
		ttf.splines[0].d = data[i-1].bp.x;
		ttf.splines[0].c = 2*(data[i-1].cp.x-data[i-1].bp.x);
		ttf.splines[0].b = data[i-1].bp.x+data[i].bp.x-2*data[i-1].cp.x;
		ttf.splines[1].d = data[i-1].bp.y;
		ttf.splines[1].c = 2*(data[i-1].cp.y-data[i-1].bp.y);
		ttf.splines[1].b = data[i-1].bp.y+data[i].bp.y-2*data[i-1].cp.y;

		if (!CompareSpline(ps, &ttf, (float)data[i - 1].t, (float)data[i].t, (float)err))
			return false;
	}

	return true;
}

static bool SplineWithWellBehavedControlPoints(Spline* ps)
{
	DPoint splineunit;
	double splinelen, npos, ppos;

	splineunit.x = ps->to.x - ps->from.x;
	splineunit.y = ps->to.y - ps->from.y;

	splinelen = sqrt(splineunit.x*splineunit.x + splineunit.y*splineunit.y);

	if (splinelen != 0)
	{
		splineunit.x /= (float)splinelen;
		splineunit.y /= (float)splinelen;
	}

	npos = (ps->cp1.x - ps->from.x) * splineunit.x + (ps->cp1.y - ps->from.y) * splineunit.y;
	ppos = (ps->cp2.x - ps->from.x) * splineunit.x + (ps->cp2.y - ps->from.y) * splineunit.y;
	return npos >= 0 && /* npos<=ppos &&*/ ppos <= splinelen;
}

/******************************************************************************
   Function:	PrettyApprox
   Author:		Paul Gruzdev
   Purpose:		Approximate one cubic bezier segment with a set of B-splines.
   Returns:		int
   Parameters:	Spline * ps - bezier spline holder
   Parameters:	double tmin
   Parameters:	double tmax
   Parameters:	QPoint * data
   Parameters:	int qcnt
   Parameters:	bool bRoundToInt
   Parameters:	int iTestLevel
   Notes:       
******************************************************************************/
static int PrettyApprox(Spline* ps, double tmin, double tmax, QPoint* data, int qcnt, bool bRoundToInt, int iTestLevel)
{
	int ptcnt, q, i;
	double distance, dx, dy, tstart;
	DPoint end, mid, slopemin, slopemid, slopeend;
	DPoint splineunit, start;
	double splinelen, midpos, lastpos, lastpos2, cppos;
	bool bDoGoodSplineCheck;
	QPoint data2[12];

	if (qcnt == -1)
		return -1;

	slopemin.x = (3*ps->splines[0].a*(float)tmin + 2*ps->splines[0].b)*(float)tmin + ps->splines[0].c;
	slopemin.y = (3*ps->splines[1].a*(float)tmin + 2*ps->splines[1].b)*(float)tmin + ps->splines[1].c;

	if (slopemin.x == 0 && slopemin.y == 0)
	{
		float t = (float)(tmin + (tmax-tmin)/256);
		// If there is no control point for this end point, then the slope is 
		//  0/0 at the end point. Which isn't useful, it leads to a quadratic 
		//  control point at the end point, but this one is double because it   
		//  is used to interpolate the next point, but we get all confused    
		//  because we don't expect a double cp to be on the base point.        
		slopemin.x = (3*ps->splines[0].a*t+2*ps->splines[0].b)*t+ps->splines[0].c;
		slopemin.y = (3*ps->splines[1].a*t+2*ps->splines[1].b)*t+ps->splines[1].c;
	}

	end.x = (float)(((ps->splines[0].a*tmax+ps->splines[0].b)*tmax+ps->splines[0].c)*tmax+ps->splines[0].d);
	end.y = (float)(((ps->splines[1].a*tmax+ps->splines[1].b)*tmax+ps->splines[1].c)*tmax+ps->splines[1].d);
	slopeend.x = (float)((3*ps->splines[0].a*tmax+2*ps->splines[0].b)*tmax+ps->splines[0].c);
	slopeend.y = (float)((3*ps->splines[1].a*tmax+2*ps->splines[1].b)*tmax+ps->splines[1].c);

	if (slopemin.x == 0 && slopemin.y == 0) 
	{
		float t = (float)(tmax - (tmax-tmin)/256);
		// Same problem as above, except at the other end
		slopeend.x = (3*ps->splines[0].a*t+2*ps->splines[0].b)*t+ps->splines[0].c;
		slopeend.y = (3*ps->splines[1].a*t+2*ps->splines[1].b)*t+ps->splines[1].c;
	}

	start.x = data[qcnt-1].bp.x;
	start.y = data[qcnt-1].bp.y;
	splineunit.x = end.x - start.x;
	splineunit.y = end.y - start.y;
	splinelen = sqrt(splineunit.x*splineunit.x + splineunit.y*splineunit.y);

	if ( splinelen!=0 ) 
	{
		splineunit.x /= (float)splinelen;
		splineunit.y /= (float)splinelen;
	}

	bDoGoodSplineCheck = SplineWithWellBehavedControlPoints(ps);

	if (bRoundToInt && tmax != 1)
	{
		end.x = (float)rint( end.x );
		end.y = (float)rint( end.y );
	}

	dx = end.x - data[qcnt-1].bp.x;
	dy = end.y - data[qcnt-1].bp.y;
	distance = dx*dx + dy*dy;

	if ( distance < .3 ) 
	{
		// This is meaningless in truetype, use a line
		data[qcnt-1].cp = data[qcnt-1].bp;
		data[qcnt].bp = end;
		data[qcnt].t = 1;
		return qcnt + 1;
	}

	for (ptcnt = 0; ptcnt < 10; ++ptcnt)
	{
		if (ptcnt > 1 && distance / (ptcnt*ptcnt) < 16)
			return -1;			// Points too close for a good approx
		q = qcnt;
		data2[ptcnt+1].bp = end;
		lastpos=0; lastpos2 = splinelen;

		for (i = 0; i <= ptcnt; ++i)
		{
			tstart = (tmin*(ptcnt-i) + tmax*(i+1))/(ptcnt+1);
			mid.x = (float)(((ps->splines[0].a*tstart+ps->splines[0].b)*tstart+ps->splines[0].c)*tstart+ps->splines[0].d);
			mid.y = (float)(((ps->splines[1].a*tstart+ps->splines[1].b)*tstart+ps->splines[1].c)*tstart+ps->splines[1].d);

			if (i == 0)
			{
				slopemid.x = (float)((3*ps->splines[0].a*tstart+2*ps->splines[0].b)*tstart+ps->splines[0].c);
				slopemid.y = (float)((3*ps->splines[1].a*tstart+2*ps->splines[1].b)*tstart+ps->splines[1].c);

				if ( slopemid.x==0 )
					data[q-1].cp.x=mid.x;
				else if ( slopemin.x==0 )
					data[q-1].cp.x=data[q-1].bp.x;
				else if ( floatNear(slopemin.y/slopemin.x,slopemid.y/slopemid.x) )
					break;
				else
					data[q-1].cp.x = -(data[q-1].bp.y-(slopemin.y/slopemin.x)*data[q-1].bp.x-mid.y+(slopemid.y/slopemid.x)*mid.x)/(slopemin.y/slopemin.x-slopemid.y/slopemid.x);

				if ( slopemid.y==0 )
					data[q-1].cp.y=mid.y;
				else if ( slopemin.y==0 )
					data[q-1].cp.y=data[q-1].bp.y;
				else if ( floatNear(slopemin.x/slopemin.y,slopemid.x/slopemid.y) )
					break;
				else
					data[q-1].cp.y = -(data[q-1].bp.x-(slopemin.x/slopemin.y)*data[q-1].bp.y-mid.x+(slopemid.x/slopemid.y)*mid.y)/(slopemin.x/slopemin.y-slopemid.x/slopemid.y);
			} 
			else 
			{
				data[q-1].cp.x = 2*data[q-1].bp.x - data[q-2].cp.x;
				data[q-1].cp.y = 2*data[q-1].bp.y - data[q-2].cp.y;
			}

			midpos = (mid.x-start.x)*splineunit.x + (mid.y-start.y)*splineunit.y;
			cppos  = (data[q-1].cp.x-start.x)*splineunit.x + (data[q-1].cp.y-start.y)*splineunit.y;

			if (((bDoGoodSplineCheck || i != 0) &&  cppos < lastpos) || cppos > midpos)
			{
				i = 0;		// Means we failed
				break;
			}
			lastpos = midpos;

			data[q].bp = mid;
			data[q++].t = tstart;

			tstart = (tmax*(ptcnt-i) + tmin*(i+1))/(ptcnt+1);
			mid.x = (float)(((ps->splines[0].a*tstart+ps->splines[0].b)*tstart+ps->splines[0].c)*tstart+ps->splines[0].d);
			mid.y = (float)(((ps->splines[1].a*tstart+ps->splines[1].b)*tstart+ps->splines[1].c)*tstart+ps->splines[1].d);

			if ( i==0 ) 
			{
				slopemid.x = (float)((3*ps->splines[0].a*tstart+2*ps->splines[0].b)*tstart+ps->splines[0].c);
				slopemid.y = (float)((3*ps->splines[1].a*tstart+2*ps->splines[1].b)*tstart+ps->splines[1].c);

				if ( slopemid.x==0 )
					data2[ptcnt-i].cp.x=mid.x;
				else if ( slopeend.x==0 )
					data2[ptcnt-i].cp.x=data2[ptcnt-i+1].bp.x;
				else if ( floatNear(slopeend.y/slopeend.x,slopemid.y/slopemid.x) )
					break;
				else
					data2[ptcnt-i].cp.x = -(data2[ptcnt-i+1].bp.y-(slopeend.y/slopeend.x)*data2[ptcnt-i+1].bp.x-mid.y+(slopemid.y/slopemid.x)*mid.x)/(slopeend.y/slopeend.x-slopemid.y/slopemid.x);

				if ( slopemid.y==0 )
					data2[ptcnt-i].cp.y=mid.y;
				else if ( slopeend.y==0 )
					data2[ptcnt-i].cp.y=data2[ptcnt-i+1].bp.y;
				else if ( floatNear(slopeend.x/slopeend.y,slopemid.x/slopemid.y) )
					break;
				else
					data2[ptcnt-i].cp.y = -(data2[ptcnt-i+1].bp.x-(slopeend.x/slopeend.y)*data2[ptcnt-i+1].bp.y-mid.x+(slopemid.x/slopemid.y)*mid.y)/(slopeend.x/slopeend.y-slopemid.x/slopemid.y);
			} 
			else 
			{
				data2[ptcnt-i].cp.x = 2*data2[ptcnt-i+1].bp.x - data2[ptcnt-i+1].cp.x;
				data2[ptcnt-i].cp.y = 2*data2[ptcnt-i+1].bp.y - data2[ptcnt-i+1].cp.y;
			}
			data2[ptcnt-i].bp = mid;

			midpos = (mid.x-start.x)*splineunit.x + (mid.y-start.y)*splineunit.y;
			cppos  = (data2[ptcnt-i].cp.x-start.x)*splineunit.x + (data2[ptcnt-i].cp.y-start.y)*splineunit.y;

			if (((bDoGoodSplineCheck || i != 0) && cppos > lastpos2) || cppos < midpos)
			{
				i = 0;		// Means we failed
				break;
			}
			lastpos2 = midpos;
		}

		if ( i==0 )
			continue;

		if ( (data2[ptcnt+1].bp.x-data2[ptcnt].bp.x)*(data2[ptcnt].cp.x-data2[ptcnt].bp.x)<0 ||
			(data2[ptcnt+1].bp.y-data2[ptcnt].bp.y)*(data2[ptcnt].cp.y-data2[ptcnt].bp.y)<0 ) 
		{
			; // data2 are bad ... don't use them
		} 
		else if ( (data[qcnt-1].bp.x-data[qcnt].bp.x)*(data[qcnt-1].cp.x-data[qcnt].bp.x)<0 ||
			(data[qcnt-1].bp.y-data[qcnt].bp.y)*(data[qcnt-1].cp.y-data[qcnt].bp.y)<0 ) 
		{
			; // data are bad
			for (i = 0; i <= ptcnt; ++i)
			{
				data[qcnt+i-1].cp = data2[i].cp;
				data[qcnt+i-1].bp = data2[i].bp;
			}
		} 
		else 
		{
			for ( i=0; i<=ptcnt; ++i ) 
			{
				if ( ptcnt!=0 ) 
				{
					data[qcnt+i-1].cp.x = (data[qcnt+i-1].cp.x*(ptcnt-i) + data2[i].cp.x*i)/ptcnt;
					data[qcnt+i-1].cp.y = (data[qcnt+i-1].cp.y*(ptcnt-i) + data2[i].cp.y*i)/ptcnt;
				}
			}
		}

		if (bRoundToInt) 
		{
			for ( i=0; i<=ptcnt; ++i ) 
			{
				data[qcnt+i-1].cp.x = rint( data[qcnt+i-1].cp.x );
				data[qcnt+i-1].cp.y = rint( data[qcnt+i-1].cp.y );
			}
		}

		for ( i=0; i<ptcnt; ++i ) 
		{
			data[qcnt+i].bp.x = (data[qcnt+i].cp.x + data[qcnt+i-1].cp.x)/2;
			data[qcnt+i].bp.y = (data[qcnt+i].cp.y + data[qcnt+i-1].cp.y)/2;
		}

		if (CompareData(ps, data, qcnt, q, bRoundToInt, iTestLevel))
			return q;
	}

	return -1;
}

static bool BuildTestQuads(Spline* ttf, float xmin, float ymin, float cx, float cy, float x, float y, float tmin, float t, float err, Spline* ps, DBounds* psbb)
{
	float fudge, normal, para;
	DPoint segdir, cpdir;

	// test the control points are reasonable
	fudge = (psbb->maxx-psbb->minx) + (psbb->maxy-psbb->miny);
	if (cx < psbb->minx - fudge || cx > psbb->maxx + fudge)
		return false;

	if (cy < psbb->miny - fudge || cy > psbb->maxy + fudge)
		return false;

	segdir.x = x - xmin;
	segdir.y = y - ymin;
	cpdir.x = cx - xmin; 
	cpdir.y = cy - ymin;
	para = segdir.x*cpdir.x + segdir.y*cpdir.y;
	normal = segdir.x*cpdir.y - segdir.y*cpdir.x;

	if (normal < 0)
		normal = -normal;

	if (para < 0 && -para > 4*normal)
		return false;

	cpdir.x = x - cx;
	cpdir.y = y - cy;
	para = segdir.x*cpdir.x + segdir.y*cpdir.y;
	normal = segdir.x*cpdir.y - segdir.y*cpdir.x;

	if (normal < 0)
		normal = -normal;

	if (para < 0 && -para > 4*normal)
		return false;

	ttf->splines[0].d = xmin;
	ttf->splines[0].c = 2*(cx-xmin);
	ttf->splines[0].b = xmin+x-2*cx;
	ttf->splines[1].d = ymin;
	ttf->splines[1].c = 2*(cy-ymin);
	ttf->splines[1].b = ymin+y-2*cy;

	if (CompareSpline(ps, ttf, tmin, t, err))
	{
//		DumpCmpData(ps, ttf, 1);
		return true;
	}
//DumpCmpData(ps, ttf, 0);
	return false;
}

static DPoint LinearSpline(std::vector<DPoint>& bs, RECT& rcBBox, Spline* ps, float tmax)
{
	float x,y;

	x = ((ps->splines[0].a*tmax + ps->splines[0].b)*tmax + ps->splines[0].c)*tmax + ps->splines[0].d;
	y = ((ps->splines[1].a*tmax + ps->splines[1].b)*tmax + ps->splines[1].c)*tmax + ps->splines[1].d;

	if (tmax == 1)
	{
		x = ps->to.x;
		y = ps->to.y;
	}

	// spline is linear. But since we can't tell up we have only one point, we add two points with the control point equal line
	// point. This will be optimised when writing points to 'glyf' table.
	DPoint dp(x, y);
	bs.push_back(dp);
	UpdateBox(ttf2xar_dp(dp), rcBBox);
	bs.push_back(dp);
	UpdateBox(ttf2xar_dp(dp), rcBBox);

	return dp;
}

static DPoint MakeQuadSpline(std::vector<DPoint>& bs, RECT& rcBBox, Spline *ttf, float x, float y, float tmax, DPoint& oldend)
{
	if (tmax == 1)
	{
		x = oldend.x;
		y = oldend.y;	// Want it to compare exactly
	}

	DPoint end(x, y);

	if (ttf->splines[0].b == 0 && ttf->splines[1].b == 0)
	{
		// spline is linear
		DPoint dp(x, y);
		bs.push_back(dp);
		UpdateBox(ttf2xar_dp(dp), rcBBox);
	} 
	else 
	{
		double cpx = ttf->splines[0].c/2.0 + ttf->splines[0].d;
		double cpy = ttf->splines[1].c/2.0 + ttf->splines[1].d;
		DPoint cp(cpx, cpy);
		bs.push_back(cp);
		UpdateBox(ttf2xar_dp(cp), rcBBox);
		DPoint dp(x, y);
		bs.push_back(dp);
		UpdateBox(ttf2xar_dp(dp), rcBBox);
	}
	return end;
}

static bool TTFApprox(std::vector<DPoint>& bs, RECT& rcBBox, Spline* ps, float tmin, float tmax, DocCoord dcStart, bool bStartOnCurve)
{
	int dim=0;
	float dx, dy, ddim, dt, t, err;
	float x, y, xmin, ymin;
	float dxdtmin, dydtmin, dxdt, dydt;
	DPoint sp;
	float cx, cy;
	Spline ttf;
	int cnt = -1, forceit, unforceable;
	DPoint start, end, rend, dend;
	DBounds bb;

	rend.x = ((ps->splines[0].a * tmax + ps->splines[0].b) * tmax + ps->splines[0].c) * tmax + ps->splines[0].d;
	rend.y = ((ps->splines[1].a * tmax + ps->splines[1].b) * tmax + ps->splines[1].c) * tmax + ps->splines[1].d;
	end.x = rint(rend.x);
	end.y = rint(rend.y);
	dend.x = (3 * ps->splines[0].a * tmax + 2*ps->splines[0].b) * tmax+ps->splines[0].c;
	dend.y = (3 * ps->splines[1].a * tmax + 2*ps->splines[1].b) * tmax+ps->splines[1].c;

	bb.minx = bb.maxx = ps->from.x;
	if (ps->cp1.x > bb.maxx) 
		bb.maxx = ps->cp1.x;
	else if (ps->cp1.x < bb.minx) 
		bb.minx = ps->cp1.x;

	if (ps->cp2.x > bb.maxx)
		bb.maxx = ps->cp2.x;
	else if (ps->cp2.x < bb.minx)
		bb.minx = ps->cp2.x;

	if (ps->to.x > bb.maxx)
		bb.maxx = ps->to.x;
	else if (ps->to.x < bb.minx) 
		bb.minx = ps->to.x;

	bb.miny = bb.maxy = ps->from.y;
	if (ps->cp1.y > bb.maxy)
		bb.maxy = ps->cp1.y;
	else if (ps->cp1.y < bb.miny)
		bb.miny = ps->cp1.y;

	if (ps->cp2.y > bb.maxy)
		bb.maxy = ps->cp2.y;
	else if (ps->cp2.y < bb.miny)
		bb.miny = ps->cp2.y;

	if (ps->to.y > bb.maxy)
		bb.maxy = ps->to.y;
	else if (ps->to.y < bb.miny)
		bb.miny = ps->to.y;

	start = xar2ttf_dcdp(dcStart);

tail_recursion:
	++cnt;

	xmin = start.x;
	ymin = start.y;
	dxdtmin = (3*ps->splines[0].a*tmin + 2*ps->splines[0].b)*tmin + ps->splines[0].c;
	dydtmin = (3*ps->splines[1].a*tmin + 2*ps->splines[1].b)*tmin + ps->splines[1].c;

	dx = ((ps->splines[0].a*tmax + ps->splines[0].b)*tmax + ps->splines[0].c)*tmax -
		 ((ps->splines[0].a*tmin + ps->splines[0].b)*tmin + ps->splines[0].c)*tmin;
	dy = ((ps->splines[1].a*tmax + ps->splines[1].b)*tmax + ps->splines[1].c)*tmax -
		 ((ps->splines[1].a*tmin + ps->splines[1].b)*tmin + ps->splines[1].c)*tmin;

	if (dx < 0)
		dx = -dx;
	if (dy < 0)
		dy = -dy;
	if (dx > dy)
	{
		dim = 0;
		ddim = dx;
	} 
	else 
	{
		dim = 1;
		ddim = dy;
	}

	err = ddim / 3000;
	if (err < 1) 
		err = 1;

	if (ddim < 2 ||
		(dend.x == 0 && rint(start.x) == end.x && dy <= 10 && cnt != 0) ||
		(dend.y == 0 && rint(start.y) == end.y && dx <= 10 && cnt != 0) ) 
	{
		if (cnt == 0 || !bStartOnCurve)
		{
			LinearSpline(bs, rcBBox, ps, tmax);
			return true;
		}
/*
		// If the end point is very close to where we want to be, then just
		//  pretend it's right
		start->prev->splines[0].b += ps->to.x - start.x;
		start->prev->splines[1].b += ps->to.y - start.y;
		start->prevcp.x += rend.x-start->me.x;
		start->prevcp.y += rend.y-start->me.y;

		if ( start->prev!=NULL && !start->prev->from->nonextcp )
			start->prev->from->nextcp = start->prevcp;

		start->me = rend;
		return( start );*/
	}

	dt = (tmax - tmin) / ddim;
	forceit = false;
	// force_end:
	unforceable = false;

	for (t = tmax; t > tmin + dt / 128; t -= dt)
	{
		// dt/128 is a hack to avoid rounding errors
		x = ((ps->splines[0].a * t + ps->splines[0].b) * t + ps->splines[0].c) * t + ps->splines[0].d;
		y = ((ps->splines[1].a * t + ps->splines[1].b) * t + ps->splines[1].c) * t + ps->splines[1].d;
		dxdt = (3*ps->splines[0].a*t + 2*ps->splines[0].b)*t + ps->splines[0].c;
		dydt = (3*ps->splines[1].a*t + 2*ps->splines[1].b)*t + ps->splines[1].c;
		// if the slopes are parallel at the ends there can be no bezier quadratic 
		//  (control point is where the splines intersect. But if they are 
		//  parallel and colinear then there is a line between 'em 
		if ((dxdtmin == 0 && dxdt == 0) || (dydtmin == 0 && dydt == 0) ||
			(dxdt != 0 && dxdtmin != 0 && floatNearish(dydt/dxdt, dydtmin/dxdtmin)))
			continue;

		if (dxdt == 0)
			cx = x;
		else if (dxdtmin == 0)
			cx = xmin;
		else
			cx = -(ymin - (dydtmin/dxdtmin)*xmin - y + (dydt/dxdt)*x)/(dydtmin/dxdtmin - dydt/dxdt);

		if (dydt == 0)
			cy = y;
		else if ( dydtmin==0 )
			cy = ymin;
		else
			cy = -(xmin - (dxdtmin/dydtmin)*ymin - x + (dxdt/dydt)*y)/(dxdtmin/dydtmin - dxdt/dydt);

		if (t == tmax && ((cy == y && cx == x) || (cy == ymin && cx == xmin)))
			unforceable = true;

		// Make the quadratic spline from (xmin,ymin) through (cx,cy) to (x,y)
		if (forceit || BuildTestQuads(&ttf, xmin, ymin, cx, cy, x, y, tmin, t, err, ps, &bb)) 
		{
			sp = MakeQuadSpline(bs, rcBBox, &ttf, x, y, t, ps->to);
			forceit = false;

			if (t == tmax)
				return true;

			tmin = t;
			start = sp;
			bStartOnCurve = true;
			goto tail_recursion;
		}

		ttf.splines[0].d = xmin;
		ttf.splines[0].c = x-xmin;
		ttf.splines[0].b = 0;
		ttf.splines[1].d = ymin;
		ttf.splines[1].c = y-ymin;
		ttf.splines[1].b = 0;

		if (CompareSpline(ps, &ttf, tmin, t, err))
		{
			sp = LinearSpline(bs, rcBBox, ps, t);
			if (t == tmax)
				return true;
			tmin = t;
			start = DocCoord((LONG)rint(sp.x), (LONG)rint(sp.y));
			bStartOnCurve = false;
			goto tail_recursion;
		}
	}

	tmin += dt;
	DPoint pt = LinearSpline(bs, rcBBox, ps, tmin);
	start = DocCoord((LONG)rint(pt.x), (LONG)rint(pt.y));
	bStartOnCurve = false;
	goto tail_recursion;
}

ULONG Spline2DFindPointsOfInflection(const Spline* sp, double poi[2])
{
	ULONG cnt = 0;
	double b2_fourac, t;
	// A POI happens when d2 y/dx2 is zero. This is not the same as d2y/dt2 / d2x/dt2 
	// d2 y/dx^2 = d/dt ( dy/dt / dx/dt ) / dx/dt 
	//		 = ( (dx/dt) * d2 y/dt2 - ((dy/dt) * d2 x/dt2) )/ (dx/dt)^3 
	// (3ax*t^2+2bx*t+cx) * (6ay*t+2by) - (3ay*t^2+2by*t+cy) * (6ax*t+2bx) == 0 
	// (3ax*t^2+2bx*t+cx) * (3ay*t+by) - (3ay*t^2+2by*t+cy) * (3ax*t+bx) == 0 
	//   9*ax*ay*t^3 + (3ax*by+6bx*ay)*t^2 + (2bx*by+3cx*ay)*t + cx*by	   
	// -(9*ax*ay*t^3 + (3ay*bx+6by*ax)*t^2 + (2by*bx+3cy*ax)*t + cy*bx)==0 
	// 3*(ax*by-ay*bx)*t^2 + 3*(cx*ay-cy*ax)*t+ (cx*by-cy*bx) == 0	   
	double a = 3*((double)sp->splines[1].a * sp->splines[0].b - (double)sp->splines[0].a * sp->splines[1].b);
	double b = 3*((double)sp->splines[0].c * sp->splines[1].a - (double)sp->splines[1].c * sp->splines[0].a);
	double c = (double) sp->splines[0].c*sp->splines[1].b-(double) sp->splines[1].c*sp->splines[0].b;

	if (!floatNear((float)a, 0))
	{
		b2_fourac = b*b - 4*a*c;
		poi[0] = poi[1] = -1;

		if (b2_fourac < 0)
			return 0;

		b2_fourac = sqrt(b2_fourac);
		t = (-b + b2_fourac) / (2*a);

		if (t >= 0 && t <= 1.0)
			poi[cnt++] = t;

		t = (-b - b2_fourac) / (2*a);

		if (t >= 0 && t <= 1.0)
		{
			if(cnt == 1 && poi[0] > t)
			{
				poi[1] = poi[0];
				poi[0] = t;
				++cnt;
			} 
			else
				poi[cnt++] = t;
		}
	} 
	else if (!floatNear((float)b, 0))
	{
		t = -c/b;
		if (t >= 0 && t <= 1.0)
			poi[cnt++] = t;
	}

	if (cnt < 2)
		poi[cnt] = -1;

	return cnt;
}

/******************************************************************************
   Function:	ApproximateBezier
   Author:		Paul Gruzdev
   Purpose:		Approximate a bezier defined by Spline coefficients using cubic spline
   Returns:		void
   Parameters:	DocCoord dcCurrent
   Parameters:	Spline & spline
   Parameters:	std::vector<DocCoord> & bs
   Parameters:	RECT & rcBBox
   Parameters:	bool bStartOnCurve
   Notes:       
******************************************************************************/
void ApproximateBezier(DocCoord dcCurrent, Spline& spline, std::vector<DPoint>& bs, RECT& rcBBox, bool bStartOnCurve)
{
	QPoint data[8*10];
	int qcnt = 1;
	data[0].bp = xar2ttf_dcdp(dcCurrent);
	data[0].t = 0;

	// The end points are at integer points, or one coord is at half while
	//  the other is at an integer (ie. condition for ttf interpolated point)
	bool bRoundToInt =
		((spline.from.x == rint(spline.from.x) &&
		spline.from.y == rint(spline.from.y)) ||
		(spline.from.x == rint(spline.from.x) &&
		spline.from.x == spline.cp1.x &&
		spline.from.y != spline.cp1.y &&
		2*spline.from.y == rint(2*spline.from.y)) ||
		(spline.from.y == rint(spline.from.y) &&
		spline.from.y == spline.cp1.y &&
		spline.from.x != spline.cp1.x &&
		2*spline.from.x == rint(2*spline.from.x)) ) &&
		((spline.to.x == rint(spline.to.x) &&
		spline.to.y == rint(spline.to.y)) ||
		(spline.to.x == rint(spline.to.x) &&
		spline.to.x == spline.cp2.x &&
		spline.to.y != spline.cp2.y &&
		2*spline.to.y == rint(2*spline.to.y)) ||
		(spline.to.y == rint(spline.to.y) &&
		spline.to.y == spline.cp2.y &&
		spline.to.x != spline.cp2.x &&
		2*spline.to.x == rint(2*spline.to.x)));

	qcnt = PrettyApprox(&spline, 0, 1, data, qcnt, bRoundToInt, 0);
	
	if(qcnt > 0)
	{
//		DumpQData(&spline, data, qcnt);
		for(size_t j = 1; j < (size_t)qcnt; j++)
		{
			bs.push_back(data[j-1].cp);
			UpdateBox(ttf2xar_dp(data[j-1].cp), rcBBox);

			bs.push_back(data[j].bp);
			UpdateBox(ttf2xar_dp(data[j].bp), rcBBox);
		}
	}
	else
	{
		//AppendFile("d:/temp/ttf/ttf.log", L"%s: glyph %d PretyApprox failed\n", A2W(m_sFullName.c_str()), m_wGlyphBeingProcessed);
		double magicpoints[6];
		ULONG cnt = Spline2DFindPointsOfInflection(&spline, magicpoints);

		// Remove points outside range
		for (ULONG ii = 0; ii < cnt; ++ii)
		{
			if (magicpoints[ii] <= 0 || magicpoints[ii] >= 1)
			{
				for (ULONG j = ii+1; j < cnt; ++j)
					magicpoints[j-1] = magicpoints[j];
				--cnt;
				--ii;
			}
		}

		// Sort points
		for (ULONG ii = 0; ii < cnt; ++ii)
		{
			for (ULONG j = ii+1; j < cnt; ++j)
			{
				if (magicpoints[ii] > magicpoints[j])
				{
					double temp = magicpoints[ii];
					magicpoints[ii] = magicpoints[j];
					magicpoints[j] = temp;
				}
			}
		}

		// Remove duplicates
		for (ULONG ii = 1; ii < cnt; ++ii)
		{
			while (ii < cnt && floatNear((float)magicpoints[ii - 1], (float)magicpoints[ii]))
			{
				--cnt;
				for (ULONG j = ii; j < cnt; ++j)
					magicpoints[j] = magicpoints[j+1];
				magicpoints[cnt] = -1;
			}
		}

		bool bSuccess = false;

		for (int iTestLevel = 0; iTestLevel < 3; ++iTestLevel)
		{
			int qcnt = 1;
			double last = 0;
			for (ULONG ii = 0; ii < cnt; ++ii)
			{
				qcnt = PrettyApprox(&spline, last, magicpoints[ii], data, qcnt, false, iTestLevel);
				last = magicpoints[ii];
			}

			qcnt = PrettyApprox(&spline, last, 1, data, qcnt, false, iTestLevel);

			if (qcnt != -1)
			{
				for(size_t j = 1; j < (size_t)qcnt; j++)
				{
					bs.push_back(data[j-1].cp);
					UpdateBox(ttf2xar_dp(data[j-1].cp), rcBBox);

					bs.push_back(data[j].bp);
					UpdateBox(ttf2xar_dp(data[j].bp), rcBBox);
				}
				bSuccess = true;
				break; // from the iTestLevel loop
			}
		}

		if(!bSuccess)
		{
			TTFApprox(bs, rcBBox, &spline, 0, 1, dcCurrent, bStartOnCurve);
			//AppendFile("d:/temp/ttf/ttf.log", L"%s: glyph %d all further PretyApprox failed\n", A2W(m_sFullName.c_str()), m_wGlyphBeingProcessed);
			// 					ERROR3("PrettyApprox failed several times, need to perform regular interpolation.");
			// 					bs.push_back(xar2ttf_dp(cp1));
			// 					UpdateBox(DocCoord((LONG)rint(cp1.x), (LONG)rint(cp1.y)), rcBBox);
			// 					bs.push_back(xar2ttf_dc(pCoord[i+2]));
			// 					UpdateBox(pCoord[i+2], rcBBox);
		}
	}

}
/******************************************************************************
   Function:	ConvertCubicToQuadratic
   Author:		Paul Gruzdev
   Purpose:		Given a set of bezier verbs/coords, approximate them as a set 
				of B-spline points ready to be written to TTF 'glyf' table.
   Returns:		size_t: the number of processed points.
   Parameters:	DocCoord dcCurrent - current (last) point
   Parameters:	BYTE * verb
   Parameters:	DocCoord * pCoord
   Parameters:	ULONG count
   Parameters:	std::vector<DPoint> & bs - output ttf points
   Parameters:	RECT & rcBBox
   Notes:       
******************************************************************************/
static ULONG ConvertCubicToQuadratic(DocCoord dcCurrent, BYTE* verb, DocCoord* pCoord, ULONG count, std::vector<DPoint>& bs, RECT& rcBBox)
{
	USES_CONVERSION;
	ULONG i = 0;
	for(; i < count; i += 3)
	{
		if((verb[i] & ~PT_CLOSEFIGURE) != PT_BEZIERTO)
			return i;

		DPoint cp1;
		if(AlreadyQuadratic(dcCurrent, &pCoord[i], cp1))
		{
			bs.push_back(xar2ttf_dpdp(cp1));
			UpdateBox(DocCoord((LONG)rint(cp1.x), (LONG)rint(cp1.y)), rcBBox);
			bs.push_back(xar2ttf_dcdp(pCoord[i+2]));
			UpdateBox(pCoord[i+2], rcBBox);
		}
		else
		{
			ASSERT(i+2 < count);
			Spline spline(xar2ttf_dcdp(dcCurrent), xar2ttf_dcdp(pCoord[i]), xar2ttf_dcdp(pCoord[i+1]), xar2ttf_dcdp(pCoord[i+2]));

			if(spline.IsLinear())
			{
				cp1.x = (dcCurrent.x + pCoord[i + 2].x)/2.0f;
				cp1.y = (dcCurrent.y + pCoord[i + 2].y)/2.0f;
				bs.push_back(xar2ttf_dpdp(cp1));
				UpdateBox(DocCoord((LONG)rint(cp1.x), (LONG)rint(cp1.y)), rcBBox);
				bs.push_back(xar2ttf_dcdp(pCoord[i+2]));
				UpdateBox(pCoord[i+2], rcBBox);
			}
			else
			{
				bool bStartOnCurve = i > 0 && (verb[i-1] & ~PT_CLOSEFIGURE) == PT_BEZIERTO;
				ApproximateBezier(dcCurrent, spline, bs, rcBBox, bStartOnCurve);
			}
		}
		dcCurrent = pCoord[i+2];
	}

	return i;
}

/************************************************************************************************
   Function:	AddTTFPoints
   Author:		Paul Gruzdev
   Purpose:		Given a bezier sub-path, create B-spline points ready to save to TTF.
				Each B-spline point is assigned a flag: 'on the curve' or 'off the curve' meaning
				it's either curve point or a control point. The bounding box of all points is 
				calculated (not bbox of the curve itself, but the bbox of the points: ttf thing).
   Returns:		ULONG
   Parameters:	[in] BYTE * verb
   Parameters:	[in] DocCoord * pCoord
   Parameters:	[in] ULONG count
   Parameters:	[out] std::vector<DocCoord> & bp
   Parameters:	[out] std::vector<BYTE> & flags
   Notes:       
*************************************************************************************************/
ULONG SFNTWriter::AddTTFPoints(BYTE* verb, DocCoord* pCoord, ULONG count, std::vector<DocCoord>& bp, std::vector<BYTE>& flags, RECT& rcBBox)
{
	ULONG cnt = 0;
	if(!count || verb[0] != PT_MOVETO)
	{
		ERROR3("AddTTFPoints sequence should contain some points and start with a moveto");
		return 0;
	}
	flags.push_back(_On_Curve);
	bp.push_back(xar2ttf_dc(pCoord[0]));
	UpdateBox(pCoord[0], rcBBox);

	DocCoord dcStart = pCoord[0];
	DocCoord dcCurrent = pCoord[0];

	for(cnt = 1; cnt < count; cnt++)
	{
		DocCoord& dc = pCoord[cnt];
		switch(verb[cnt] & ~PT_CLOSEFIGURE)
		{
		case PT_MOVETO:
			return cnt;
			break;

		case PT_LINETO:
			if(dc != dcStart)
			{
				DocCoord ttf = xar2ttf_dc(dc);

				if(ttf.x == bp.back().x && ttf.y == bp.back().y)
					;
				else
				{
					flags.push_back(_On_Curve);
					bp.push_back(ttf);
					UpdateBox(pCoord[cnt], rcBBox);
					dcCurrent = pCoord[cnt];
				}
			}
			break;
		case PT_BEZIERTO:
			{
				std::vector<DPoint> pt;
				ULONG ulProcessed = ConvertCubicToQuadratic(dcCurrent, &verb[cnt], &pCoord[cnt], count - cnt, pt, rcBBox);
				ASSERT(pt.size() && (pt.size() % 2) == 0);
				for(size_t i = 0; i < pt.size(); i++)
				{
					if(rint(pt[i].x) == bp.back().x && rint(pt[i].y) == bp.back().y)
						;
					else
					{
						flags.push_back(_Off_Curve);
						bp.push_back(dp2dc(pt[i]));
					}

					i++;

					if(i + 1 < pt.size() && CanInterpolate(pt[i-1], pt[i], pt[i+1]))
					{
					}
					else
					{
						if(rint(pt[i].x) == bp.back().x && rint(pt[i].y) == bp.back().y)
						{
							if(flags.back() == _Off_Curve)
								flags.back() = _On_Curve;
						}
						else
						{
							flags.push_back(_On_Curve);
							bp.push_back(dp2dc(pt[i]));
						}
					}
				}

				// (ULONG)pt.size()/2*3 - 1; // 
				cnt += ulProcessed - 1; // skip all the points we just proceeded and jump to the last TO point of the bezier
				dcCurrent = pCoord[cnt];
			}
			break;
		}
	}

	return cnt;
}

/******************************************************************************
   Function:	DumpPointArray
   Author:		Paul Gruzdev
   Purpose:		Save B-spline curve points into TTF file
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	DocCoord * bp
   Parameters:	BYTE * fs
   Parameters:	int pc
   Notes:       
******************************************************************************/
static void DumpPointArray(CCFile* pFile, DocCoord* bp, BYTE* fs, int pc)
{
	DocCoord last;
	int flags;
	int lastflag, flagcnt;

	// flags
	last.x = last.y = 0;
	lastflag = -1; flagcnt = 0;

	for (int i = 0; i < pc; ++i)
	{
		flags = 0;

		if (fs==NULL || fs[i])
			flags = _On_Curve;		// points are on curve

		if (last.x == bp[i].x)
			flags |= _X_Same;
		else if(bp[i].x - last.x > -256 && bp[i].x - last.x < 255)
		{
			flags |= _X_Short;
			if (bp[i].x >= last.x)
				flags |= _X_Same;		// In this context it means positive
		}

		if (last.y == bp[i].y)
			flags |= _Y_Same;
		else if (bp[i].y - last.y > -256 && bp[i].y - last.y < 255)
		{
			flags |= _Y_Short;
			if (bp[i].y >= last.y)
				flags |= _Y_Same;		// In this context it means positive
		}

		last = bp[i];
		if (lastflag == -1)
		{
			lastflag = flags;
			flagcnt = 0;
		} 
		else if ( flags != lastflag ) 
		{
			if ( flagcnt!=0 )
				lastflag |= _Repeat;

			pFile->write((char)lastflag);

			if ( flagcnt != 0 )
				pFile->write((char)flagcnt);

			lastflag = flags;
			flagcnt = 0;
		} 
		else 
		{
			if ( ++flagcnt == 255 ) 
			{
				pFile->write((char)(lastflag | _Repeat));
				pFile->write((char)0xFF);

				lastflag = -1;
				flagcnt = 0;
			}
		}
	}

	if(lastflag != -1)
	{
		if ( flagcnt != 0 )
			lastflag |= _Repeat;
		pFile->write((char)lastflag);
		if ( flagcnt != 0 )
			pFile->write((char)flagcnt);
	}

	// xcoords
	last.x = 0;
	for (int i = 0; i < pc; ++i) 
	{
		if ( last.x==bp[i].x )
			; // Do Nothing
		else if ( bp[i].x - last.x > -256 && bp[i].x - last.x < 255 ) 
		{
			if ( bp[i].x >= last.x )
				pFile->write((char)(bp[i].x-last.x));
			else
				pFile->write((char)(last.x-bp[i].x));
		} 
		else
			putshort(pFile, (short)(bp[i].x-last.x));

		last.x = bp[i].x;
	}

	// ycoords
	last.y = 0;
	for (int i = 0; i < pc; ++i)
	{
		if ( last.y == bp[i].y )
			; // Do Nothing
		else if ( bp[i].y - last.y > -256 && bp[i].y - last.y < 255 ) 
		{
			if ( bp[i].y >= last.y )
				pFile->write((char)(bp[i].y-last.y));
			else
				pFile->write((char)(last.y-bp[i].y));
		} 
		else
			putshort(pFile, (short)(bp[i].y-last.y));

		last.y = bp[i].y;
	}

	if (pFile->tell() & 1)	// Pad the file so that the next glyph will start
		pFile->write('\0');	// on a word boundary
}

/******************************************************************************
   Function:	DumpTTFSpace
   Author:		Paul Gruzdev
   Purpose:		Write a space glyph
   Returns:		void
   Parameters:	CCFile * ttf
   Notes:       
******************************************************************************/
void SFNTWriter::DumpTTFSpace(CCFile* ttf)
{
	// These don't get a glyph header, because there are no contours
	m_PointCount.push_back(0);
	m_loca.push_back(ttf->tell());
}

/******************************************************************************
   Function:	dumpghstruct
   Author:		Paul Gruzdev
   Purpose:		Write glyph header
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	DWORD numContours
   Parameters:	RECT rcGlyph
   Notes:       
******************************************************************************/
static void dumpghstruct(CCFile* pFile, DWORD numContours, RECT rcGlyph) 
{
	putshort(pFile, (short)numContours);
	putshort(pFile, (short)xar2ttf_l(rcGlyph.left));
	putshort(pFile, (short)xar2ttf_l(rcGlyph.top));
	putshort(pFile, (short)xar2ttf_l(rcGlyph.right));
	putshort(pFile, (short)xar2ttf_l(rcGlyph.bottom));
}

/******************************************************************************
   Function:	GetCvtIndex
   Author:		Paul Gruzdev
   Purpose:		Store the value in cvt table and return it's index in it
   Returns:		BYTE
   Parameters:	WORD val
   Notes:       
******************************************************************************/
BYTE SFNTWriter::GetCvtIndex(WORD val)
{
	std::vector<WORD>::iterator it = std::find(m_cvt.begin(), m_cvt.end(), val);
	if(it != m_cvt.end())
		return BYTE(it - m_cvt.begin());
	m_cvt.push_back(val);
	return (BYTE)m_cvt.size() - 1;
}

/******************************************************************************
   Function:	DumpTTFBlankGlyph
   Author:		Paul Gruzdev
   Purpose:		Write a blank glyph
   Returns:		void
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
void SFNTWriter::DumpTTFBlankGlyph(CCFile* pFile)
{
	m_PointCount.push_back(0);
	m_loca.push_back(pFile->tell());
}

/******************************************************************************
   Function:	DumpTTFMissingGlyph
   Author:		Paul Gruzdev
   Purpose:		Write a missing glyph
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	RECT rc
   Notes:       
******************************************************************************/
void SFNTWriter::DumpTTFMissingGlyph(CCFile* pFile, RECT rc)
{
	DocCoord bp[10];
	BYTE instrs[50];
	int stemcvt, stem;

	m_PointCount.push_back(8);
	m_loca.push_back(pFile->tell());

	if(m_dwMaxContours < 2)
		m_dwMaxContours = 2;
	//gi->maxp->maxContours = 2;

	dumpghstruct(pFile, 2, rc);

	rc.left = xar2ttf_l(rc.left);
	rc.top = xar2ttf_l(rc.top);
	rc.right = xar2ttf_l(rc.right);
	rc.bottom = xar2ttf_l(rc.bottom);
	stem = rc.left;

	bp[0].x = stem;		bp[0].y = 0;
	bp[1].x = stem;		bp[1].y = rc.bottom;
	bp[2].x = rc.right;		bp[2].y = rc.bottom;
	bp[3].x = rc.right;		bp[3].y = 0;

	bp[4].x = 2*stem;		bp[4].y = stem;
	bp[5].x = rc.right-stem;	bp[5].y = stem;
	bp[6].x = rc.right-stem;	bp[6].y = rc.bottom-stem;
	bp[7].x = 2*stem;		bp[7].y = rc.bottom-stem;

	stemcvt = GetCvtIndex((WORD)stem);

	instrs[0] = 0xb1;		// Pushb, 2byte 
	instrs[1] = 1;			// Point 1 
	instrs[2] = 0;			// Point 0 
	instrs[3] = 0x2f;		// MDAP, rounded (pt0) 
	instrs[4] = 0x3c;		// ALIGNRP, (pt1 same pos as pt0)
	instrs[5] = 0xb2;		// Pushb, 3byte 
	instrs[6] = 7;			// Point 7 
	instrs[7] = 4;			// Point 4 
	instrs[8] = (BYTE)stemcvt;		// CVT entry for our stem width 
	instrs[9] = 0xe0+0x0d;		// MIRP, don't set rp0, minimum, rounded, black 
	instrs[10] = 0x32;		// SHP[rp2] (pt7 same pos as pt4) 
	instrs[11] = 0xb1;		// Pushb, 2byte 
	instrs[12] = 6;			// Point 6 
	instrs[13] = 5;			// Point 5 
	instrs[14] = 0xc0+0x1c;		// MDRP, set rp0, minimum, rounded, grey 
	instrs[15] = 0x3c;		// ALIGNRP, (pt6 same pos as pt5)
	instrs[16] = 0xb2;		// Pushb, 3byte 
	instrs[17] = 3;			// Point 3 
	instrs[18] = 2;			// Point 2 
	instrs[19] = (BYTE)stemcvt;		// CVT entry for our stem width 
	instrs[20] = 0xe0+0x0d;		// MIRP, dont set rp0, minimum, rounded, black 
	instrs[21] = 0x32;		// SHP[rp2] (pt3 same pos as pt2) 

	instrs[22] = 0x00;		// SVTCA, y axis 

	instrs[23] = 0xb1;		// Pushb, 2byte 
	instrs[24] = 3;			// Point 3 
	instrs[25] = 0;			// Point 0 
	instrs[26] = 0x2f;		// MDAP, rounded 
	instrs[27] = 0x3c;		// ALIGNRP, (pt3 same height as pt0)
	instrs[28] = 0xb2;		// Pushb, 3byte 
	instrs[29] = 5;			// Point 5 
	instrs[30] = 4;			// Point 4 
	instrs[31] = (BYTE)stemcvt;		// CVT entry for our stem width 
	instrs[32] = 0xe0+0x0d;		// MIRP, don't set rp0, minimum, rounded, black 
	instrs[33] = 0x32;		// SHP[rp2] (pt5 same height as pt4) 
	instrs[34] = 0xb2;		// Pushb, 3byte 
	instrs[35] = 7;			// Point 7 
	instrs[36] = 6;			// Point 6 
	instrs[37] = GetCvtIndex((WORD)bp[6].y);	// CVT entry for top height 
	instrs[38] = 0xe0+0x1c;	// MIRP, set rp0, minimum, rounded, grey 
	instrs[39] = 0x3c;		// ALIGNRP (pt7 same height as pt6) 
	instrs[40] = 0xb2;		// Pushb, 3byte 
	instrs[41] = 1;			// Point 1 
	instrs[42] = 2;			// Point 2 
	instrs[43] = (BYTE)stemcvt;		// CVT entry for our stem width 
	instrs[44] = 0xe0+0x0d;		// MIRP, dont set rp0, minimum, rounded, black 
	instrs[45] = 0x32;		// SHP[rp2] (pt1 same height as pt2) 

	// We've touched all points in all dimensions 
	// Don't need any IUP 

	// end contours array 
	putshort(pFile,4-1);
	putshort(pFile,8-1);

	// instruction length + instructions 
	{
		putshort(pFile, 46);
		pFile->write(instrs, 46);
	}

	DumpPointArray(pFile, bp, NULL, 8);
/*
	if (!m_bFixedWidth)
		putshort(gi->hmtx,gh.xmax + 2*stem);
	else
		putshort(gi->hmtx,fixedwidth);

	putshort(gi->hmtx,stem);*/
}

/******************************************************************************
   Function:	DumpTTFGlyph
   Author:		Paul Gruzdev
   Purpose:		Write out glyph outline as a B-spline into ttf file
   Returns:		void
   Parameters:	CCFile * pFile
   Parameters:	BYTE * pVerb
   Parameters:	DocCoord * pCoord
   Parameters:	ULONG ulCount
   Parameters:	RECT rcGlyph
   Notes:       
******************************************************************************/
void SFNTWriter::DumpTTFGlyph(CCFile* pFile, BYTE* pVerb, DocCoord* pCoord, ULONG ulCount, RECT& rcGlyph)
{
	if (ulCount == 0) 
	{
		memset(&rcGlyph, 0, sizeof(RECT));
		DumpTTFSpace(pFile);
		return;
	}

	m_loca.push_back(pFile->tell());

	DWORD dwNumContours = 0;
	FilePos fpNumContours = pFile->tell(); // we return here later
	dumpghstruct(pFile, 0, rcGlyph);

	std::vector<DocCoord> bp;
	std::vector<BYTE> flags;
	RECT rcBBox; // This box is a ttf one - this is a minmax of all ttf coordinates written, not the bbox of an actual shape
	rcBBox.left = rcBBox.top = MAXLONG;
	rcBBox.right = rcBBox.bottom = MINLONG;

	for (ULONG cnt = 0; cnt < ulCount; )
	{
		ULONG uNewCnt = AddTTFPoints(&pVerb[cnt], &pCoord[cnt], ulCount-cnt, bp, flags, rcBBox);

		ASSERT(uNewCnt);
		if(!uNewCnt)
			return;
		putshort(pFile, (short)(bp.size() - 1));
		cnt += uNewCnt;
		dwNumContours++;
	}
//DumpPoints(&bp[0], &flags[0], bp.size());
	rcGlyph = rcBBox;
	FilePos fpCurrent = pFile->tell();
	pFile->seek(fpNumContours);
	dumpghstruct(pFile, dwNumContours, rcBBox);
	
	pFile->seek(fpCurrent);

	if (dwNumContours > m_dwMaxContours) 
		m_dwMaxContours = dwNumContours;

	if (flags.size() > m_dwMaxPoints) 
		m_dwMaxPoints = (DWORD)flags.size();

	m_PointCount.push_back((DWORD)flags.size());

	//dumpinstrs(gi, isc->ttf_instrs, isc->ttf_instrs_len);
	putshort(pFile, 0); // zero-length instructions
	DumpPointArray(pFile, &bp[0], &flags[0], (int)flags.size());
}

static void AlignGlyphFile(CCFile* pFile)
{
	if((pFile->tell() & 3) != 0)
	{
		// Apple says glyphs must be 16bit aligned
		if (pFile->tell() & 1)
			pFile->write('\0');
		// MS says glyphs should be 32bit aligned
		if (pFile->tell() & 2)
			putshort(pFile, 0);
	}
}

/******************************************************************************
   Function:	DumpTTFGlyphs
   Author:		Paul Gruzdev
   Purpose:		Dump all glyphs as B-splines into a 'glyf' table
   Returns:		void
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
void SFNTWriter::DumpTTFGlyphs(CCFile* pFile)
{
	m_OutputFormat = keTrueType;
	LONG stem = (m_lAscent - m_lDescent)/30;
	RECT rc, bbox;
	LONG lCalcAvgCharWidth = 0, lWidthCount = 0;

	rc.top = 0;
	rc.bottom = 2*(m_lAscent - m_lDescent)/3;
	rc.right = 5*stem + (m_lAscent - m_lDescent)/10;
	rc.left = stem;
	rc.right += stem;

	if (rc.bottom > m_lAscent)
		rc.bottom = m_lAscent;
	// Dump .notdef glyph
	DumpTTFMissingGlyph(pFile, rc);
	AlignGlyphFile(pFile);
	m_CharProps[0].lCharWidth = rc.right - rc.left;
	m_CharProps[0].rcBBox = rc;
	bbox = rc;

	// dump null glyph
	DumpTTFBlankGlyph(pFile);
	m_CharProps.insert(m_CharProps.begin() + 1, CharProps(0, 0));

	// For reasons quite obscure to me, glyph 1 has an advance width of 0
	// even in a mono-spaced font like CourierNew.ttf
	LONG lAdvance = m_bFixedWidth ? m_lDefaultWidth : (m_lAscent - m_lDescent)/3;
	// dump carriage return glyph
	DumpTTFBlankGlyph(pFile);
	m_CharProps.insert(m_CharProps.begin() + 2, CharProps(lAdvance, 0));

	bool bAllGood = true;
	m_lXHeight = m_lCapHeight = 0;
	LONG lCapHeightCnt = 0, lXHeightCnt = 0;

	// Now iterate through all 'normal' glyphs
	for (size_t i = 3; i < m_CharProps.size(); i++)
	{
		CharProps* props = &m_CharProps[i];
		WCHAR wUnicode = props->wUnicode;
		m_wGlyphBeingProcessed = wUnicode;
		PathVerb* pVerbs = NULL;
		DocCoord* pCoords = NULL;
		ULONG ulNumCoords = 0;

		VGlyphSource::GlyphState GState = m_pSource->GetGlyphPath(wUnicode, &pVerbs, &pCoords, &ulNumCoords);

		if (GState == VGlyphSource::keRegular || GState == VGlyphSource::keEmbedded)
		{
			RECT rcBox;
			DumpTTFGlyph(pFile, pVerbs, pCoords, ulNumCoords, rcBox);
			AlignGlyphFile(pFile);
			//m_pSource->CalcBBox(pCoords, pVerbs, ulNumCoords, rcBox);
			m_CharProps[i].rcBBox = rcBox;
			::UnionRect(&m_rcFontBounds, &bbox, &rcBox);
			bbox = m_rcFontBounds;
			
			if(rcBox.right - rcBox.left)
			{
				lCalcAvgCharWidth += LONG(xar2ttf_l(rcBox.right) - xar2ttf_l(rcBox.left));
				lWidthCount++;
			}
			// accumulate Cap Height to resolve it later
			if (wUnicode >= 'A' && wUnicode <= 'Z')
			{
				m_lCapHeight += rcBox.bottom;
				lCapHeightCnt++;
			}
			else if (wUnicode >= 'a' && wUnicode <= 'z')
			{
				m_lXHeight += rcBox.bottom;
				lXHeightCnt++;
			}
		}
		else
			bAllGood = false;
	}

	if (lCapHeightCnt)
		m_lCapHeight /= lCapHeightCnt;
	if (lXHeightCnt)
		m_lXHeight /= lXHeightCnt;
	if(lWidthCount)
		lCalcAvgCharWidth /= lWidthCount;

	m_loca.push_back(pFile->tell());
	//return bAllGood;

}

/******************************************************************************
   Function:	WriteCVT
   Author:		Paul Gruzdev
   Purpose:		https://www.microsoft.com/typography/otspec/cvt.htm
   Returns:		LONG
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
LONG SFNTWriter::WriteCVT(CCFile* pFile)
{
	for(std::vector<WORD>::iterator it = m_cvt.begin(); it != m_cvt.end(); ++it)
		putshort(pFile, *it);
	return pFile->tell();
}

/******************************************************************************
   Function:	WriteLOCA
   Author:		Paul Gruzdev
   Purpose:		https://www.microsoft.com/typography/otspec/loca.htm
   Returns:		LONG
   Parameters:	CCFile * pFile
   Notes:       
******************************************************************************/
LONG SFNTWriter::WriteLOCA(CCFile* pFile)
{
	if(!m_loca.size())
		return 0;

	if(m_loca.back() >= 0x20000)
	{
		for(std::list<DWORD>::iterator it = m_loca.begin(); it != m_loca.end(); ++it)
			putlong(pFile, *it);
	}
	else
	{
		for(std::list<DWORD>::iterator it = m_loca.begin(); it != m_loca.end(); ++it)
			putshort(pFile, (short)((WORD)(*it >> 1)));
	}
	return pFile->tell();
}

/******************************************************************************
   Function:	WriteGASP
   Author:		Paul Gruzdev
   Purpose:		https://www.microsoft.com/typography/otspec/gasp.htm
   Returns:		void
   Parameters:	CCFile * pFile
   Notes:       currently only empty 'gasp' table is being written
******************************************************************************/
LONG SFNTWriter::WriteGASP(CCFile* pFile)
{
	putshort(pFile, 0);	// Old version number 
	// For fonts with no instructions always dump a gasp table which 
	//  asks for grey and no grid fit 
	putshort(pFile, 1);
	putshort(pFile, (short)0xffff);	// Upper bound on pixels/em for this range 
	putshort(pFile, 0x2);	// Grey scale, no gridfitting 
	// No hints, so no grids to fit 
	return pFile->tell();
}

} // namespace SFNT
