/**************************************************************************************
	Definitions of methods declared in otfTools.h used for parsing fonts,
	especially GPOS data for glyph kerning.

	Externally, one method is available is for parsing:
	void OTFDataLoader::ParseFontData(CDC *pDC)

	and one for filling data array:
	static void SetKernData(INT* iKernCount, MillipointKerningPair* pKernArray);

	Also lots of auxiliary data is available in structure:
	static OpenTypeFontData OTFData;

***************************************************************************************/
#include "stdafx.h"
#ifdef CAMELOT_PRODUCT
#include "app.h"
#include "camelot.h"
#include "fontman.h"
#include "atmfonts.h"
#include "unicdman.h"
#include "textfuns.h"
#endif

#include "otfTools.h"
#include "otfToolsConsts.h"

#include <sstream>
#include <unordered_map>

// uncomment this if you want to measure code performance
//#define MEASURE_FORMATTING_PERFORMANCE

#ifdef MEASURE_FORMATTING_PERFORMANCE
#define FORMATTING_PERFORMANCE_FUNC_START SIMPLEPERF_FUNCSTART
#define FORMATTING_PERFORMANCE_START SIMPLEPERF_START
#define FORMATTING_PERFORMANCE_END SIMPLEPERF_END
#else
#define FORMATTING_PERFORMANCE_FUNC_START 
#define FORMATTING_PERFORMANCE_START
#define FORMATTING_PERFORMANCE_END
#endif

static bool IsPrivateUseAreaChar(WCHAR c)
{
#ifdef CAMELOT_PRODUCT
	return CharDescription::IsPrivateUseAreaChar((WCHAR)c) ? true : false;
#else
	return (c > 0xE000 && c < 0xF8FF);
#endif
}

/*
	Helper method for containers
*/

#define FRONTORNULL(v) { (v).empty() ? nullptr : (v).front() }
#define BACKORNULL(v) { (v).empty() ? nullptr : (v).back() }
#define DEFAULT_EMSIZE 2048
#define MAX_PUA_SIZE 6400

void TagToChrArr(CHAR* T, ULONG uTag)
{
	T[0] = (CHAR)(uTag >> 24);
	uTag -= T[0] << 24;
	T[1] = (CHAR)(uTag >> 16);
	uTag -= T[1] << 16;
	T[2] = (CHAR)(uTag >> 8);
	uTag -= T[2] << 16;
	T[3] = (CHAR)(uTag);
}

/*
	LigatureDataSet operators
*/

bool LigatureDataSet::operator == (const LigatureDataSet& A)
{
	if (A.vecLigComponents.size() != vecLigComponents.size())
		return false;

	for (size_t i = 0; i < vecLigComponents.size(); ++i)
		if (A.vecLigComponents[i] != vecLigComponents[i])
			return false;

	if (A.LigType != LigType)
		return false;

	return (A.LigUnicode == LigUnicode);
}

/********************************************************************************************

>	OpenTypeFontData::OpenTypeFontData()

	Author:		Greg Watroba
	Created:	23/11/2015
	Inputs:		-
	Outputs		-
	Returns:	-
	Purpose:	Default constructor for OpenTypeFontData structure

********************************************************************************************/

//initialization for static vectors
std::vector<wchar_t> OpenTypeFontData::s_vCharSet;
std::vector<UINT16> OpenTypeFontData::s_vGlyphIndices;
std::vector<UINT32> OpenTypeFontData::s_vScriptsToSupport = { 0, CHR('a','r','m','n'), CHR('b','y','z','m'), CHR('c','a','n','s'),
																 CHR('a','g','h','b'), CHR('c','y','r','l'), CHR('D','F','L','T'),   
																 CHR('g','e','o','r'), CHR('g','o','t','h'), CHR('g','r','e','k'), 
																 CHR('k','a','n','a'), CHR('l','a','t','n'), CHR('l','i','n','a'),
																 CHR('l','i','n','b'), CHR('m','a','t','h'), CHR('m','u','s','c'),
																 CHR('h','u','n','g')};

std::vector<GlyphSetEnum> OpenTypeFontData::s_vSupportedRules =		{ GSET_ANCIENT_LIGATURES, GSET_CONTEXTUAL_LIGATURES, GSET_DISCRETIONARY_LIGATURES, 
																	  GSET_HISTORIC_LIGATURES, GSET_REQUIRED_LIGATURES, GSET_STANDARD_LIGATURES,
																	  GSET_STYLE_SET_1, GSET_STYLE_SET_2, GSET_STYLE_SET_3, GSET_STYLE_SET_4, 
																	  GSET_STYLE_SET_5, GSET_STYLE_SET_6, GSET_STYLE_SET_7, GSET_STYLE_SET_8,
																	  GSET_STYLE_SET_9, GSET_STYLE_SET_10, GSET_STYLE_SET_11, GSET_STYLE_SET_12,
																	  GSET_STYLE_SET_13, GSET_STYLE_SET_14, GSET_STYLE_SET_15, GSET_STYLE_SET_16,
																	  GSET_STYLE_SET_17, GSET_STYLE_SET_18, GSET_STYLE_SET_19, GSET_STYLE_SET_20, 
																	  GSET_LOWERCASE_TO_SMALL_CAPITALS, GSET_CONTEXTUAL_ALTERNATES,
																	  GSET_STYLISTIC_ALTERNATIVES, GSET_TITLING, GSET_CONTEXTUAL_SWASH, GSET_SWASH };

std::vector<GlyphSetEnum> OpenTypeFontData::s_vSupportedSubstitutions = { GSET_ANCIENT_LIGATURES, GSET_CONTEXTUAL_LIGATURES, GSET_DISCRETIONARY_LIGATURES,
																		  GSET_HISTORIC_LIGATURES, GSET_REQUIRED_LIGATURES, GSET_STANDARD_LIGATURES,
																		  GSET_STYLE_SET_1, GSET_STYLE_SET_2, GSET_STYLE_SET_3, GSET_STYLE_SET_4,
																		  GSET_STYLE_SET_5, GSET_STYLE_SET_6, GSET_STYLE_SET_7, GSET_STYLE_SET_8,
																		  GSET_STYLE_SET_9, GSET_STYLE_SET_10, GSET_STYLE_SET_11, GSET_STYLE_SET_12,
																		  GSET_STYLE_SET_13, GSET_STYLE_SET_14, GSET_STYLE_SET_15, GSET_STYLE_SET_16,
																		  GSET_STYLE_SET_17, GSET_STYLE_SET_18, GSET_STYLE_SET_19, GSET_STYLE_SET_20, 
																		  GSET_LOWERCASE_TO_SMALL_CAPITALS, GSET_CONTEXTUAL_ALTERNATES,
																		  GSET_STYLISTIC_ALTERNATIVES, GSET_TITLING, GSET_CONTEXTUAL_SWASH, GSET_SWASH };

OpenTypeFontData::OpenTypeFontData(): hDC(nullptr), iLastSizePos(0), bBadOpenTypeData(FALSE), 
									  bLookupsInitialized(FALSE), bBadGlyphData(FALSE)
{
	if (!s_vCharSet.empty() && !s_vGlyphIndices.empty())
		return;

	s_vCharSet.clear();
	s_vCharSet.resize(UINT16_MAX);

	for (UINT16 i = 0; i < UINT16_MAX; ++i)
		s_vCharSet[i] = (static_cast<wchar_t>(i));

	s_vGlyphIndices.resize(UINT16_MAX);
};

/********************************************************************************************

>	OpenTypeFontData::OpenTypeFontData()

	Author:		Greg Watroba
	Created:	23/11/2015
	Inputs:		-
	Outputs		-
	Returns:	-
	Purpose:	Cleanup method for OpenTypeFontData class;

********************************************************************************************/

void OpenTypeFontData::ClearData()
{
	vFontStyleNames.clear();

	for (auto &pSplineChar : vChars)
		SAFE_DELETE(pSplineChar);

	vChars.clear();

	for (auto &pVKernClass : vVerticalKernClasses)
		SAFE_DELETE(pVKernClass);

	vVerticalKernClasses.clear();

	for (auto &pKernClass : vKernClasses)
		SAFE_DELETE(pKernClass);

	vKernClasses.clear();

	for (auto &pFPST : lstFPST)
		SAFE_DELETE(pFPST);

	lstFPST.clear();

	for (auto i = 0; i < 4; ++i)
		iFontBoundingBoxes[i] = 0;

	for (auto& pDict : vpTopDicts)
		SAFE_DELETE(pDict);
	vpTopDicts.clear();
	for (auto& pDict : vpSubDicts)
		SAFE_DELETE(pDict);
	vpSubDicts.clear();
	vFDSelect.clear();
	vInUse.clear();

	hDC = nullptr;

	bBadCFFData = FALSE;
	bBadOpenTypeData = FALSE;
	bBadGlyphData = FALSE;
	bLookupsInitialized = FALSE;
	bOnlyStrikes = FALSE;
	bIsOrder2Type = FALSE;
	bIsTrueTypeCollection = FALSE;
	bAlreadyFoundBadGlyphData = FALSE;
	bBadBoundingBoxData = FALSE;
	bBadOS2Data = FALSE;
	bUseTypoMetrics = FALSE;
	bWeightWidthSlopeOnly = FALSE;

	iFontTypeFace = 0;
	uBitmapDataStart = 0;
	uBitmapDataLength = 0;
	uBitmapLocationStart = 0;
	uBitmapLocationLength = 0;
	uBSLNStart = 0;
	uCFFStart = 0;
	uCFFLength = 0;
	uDesignSize = 0;
	uDesignRangeBottom = 0;
	uDesignRangeTop = 0;
	uFontStyleId = 0;
	uFontRevision = 0;
	uFontTimeStamp = 0;
	uGASPStart = 0;
	uGDEFStart = 0;
	uGDEFLength = 0;
	uHMetricsStart = 0;
	uJSTFStart = 0;
	uJSTFLength = 0;
	uKernStart = 0;
	uTyp1Start = 0;
	uTyp1Length = 0;
	uVHead = 0;
	uVMetrics = 0;
	uOptimisedForClearType = 0;
	uApplyLSB = 0;
	uIndexToLocIsLong = 0;
	uPostScriptStart = 0;
	uOS2Start = 0;
	uFeatureStart = 0;
	uLcarStart = 0;
	uOPBDStart = 0;
	uPropStart = 0;
	uCVTStart = 0;
	uCVTLen = 0;
	uPrepStart = 0;
	uPrepLength = 0;
	uFPGMStart = 0;
	uFPGMLength = 0;
	uPFEDStart = 0;
	uTexStart = 0;
	uBDFStart = 0;
	uCvarStart = 0;
	uCvarLength = 0;
	uAvarStart = 0;
	uAvarLength = 0;
	iEmSize = 2048;
	iAscent = 0;
	iDescent = 0;
	iMacStyle = 0;
	iWidthCnt = 0;
	iAdvanceWidthMax = 0;
	iBaseStart = 0;
	iGposStart = 0;
	iGposLength = 0;
	iGsubStart = 0;
	iGsubLength = 0;
	iLastSizePos = 0;
	iFeatureCnt = 0;
	iLookupCnt = 0;
	iPositioningSizeBounds = 0;
	iGDefBounds = 0;
	iNumTables = 0;
	iGlyphStart = 0;
	iGlyphLength = 0;
	iGlyphLocationStart = 0;
	iGlyphLocationLength = 0;
	iGlyphCnt = 0;
	iEncodingStart = 0;
	iHeadStart = 0;
	iHHeaStart = 0;
	iMaxPStart = 0;
	iMaxPLen = 0;
	iMathStart = 0;
	iMathLength = 0;
	iMortStart = 0;
	iMorxStart = 0;
	iGVarStart = 0;
	iGVarLength = 0;
	iFVarStart = 0;
	iFVarLength = 0;
	iCopyrightStart = 0;
	eUniInterp = UI_UNSET;
	uOS2Version = 0;
	uAvgWidth = 0;
	LookupSubtableUID = 0;

	iMarkClassCnt = 0;
	iMarkSetCnt = 0;
	vMarkClasses.clear();
	vMarkClassNames.clear();
	vMarkSets.clear();
	vMarkSetNames.clear();

	vGPOSLookups.clear();
	vGSUBLookups.clear();

	mapLookupToGPOSFeature.clear();
	mapLookupToGSUBFeature.clear();

	vecLookupSubTables.clear();
	vecNeededLookups.clear();
	setStandardSubLookups.clear();
}

/********************************************************************************************

>	USHORT OTFDataLoader::ParseUShortFromPos(const BYTE** pPos)

	Author:		Greg Watroba
	Created:	23/11/2015
	Inputs:		pPos - current pointer to BYTE data
	Outputs		USHORT read from BYTE data
	Returns:	-
	Purpose:	Parses USHORT value from given BYTE data

********************************************************************************************/

USHORT OTFDataLoader::ParseUShortFromPos(const BYTE** pPos)
{
	USHORT preVal = 0;
	memcpy(&preVal, (*pPos), sizeof(preVal));
	preVal = _byteswap_ushort(preVal);
	(*pPos) += sizeof(USHORT);
	return preVal;
}

/********************************************************************************************

>	ULONG OTFDataLoader::ParseLongFromPos(const BYTE** pPos)

	Author:		Greg Watroba
	Created:	23/11/2015
	Inputs:		pPos - current pointer to BYTE data
	Outputs		LONG read from BYTE data
	Returns:	-
	Purpose:	Parses LONG value from given BYTE data

********************************************************************************************/

LONG OTFDataLoader::ParseLongFromPos(const BYTE** pPos)
{
	LONG preVal = 0;
	memcpy(&preVal, (*pPos), sizeof(preVal));
	preVal = _byteswap_ulong(preVal);
	(*pPos) += sizeof(LONG);
	return preVal;
}

/********************************************************************************************

>	CHAR OTFDataLoader::ParseCharFromPos(const BYTE** pPos)

	Author:		Greg Watroba
	Created:	8/12/2015
	Inputs:		pPos - current pointer to BYTE data
	Outputs		CHAR 1 byte value read from BYTE data
	Returns:	-
	Purpose:	Parses CHAR 1 byte value from given BYTE data

********************************************************************************************/

UCHAR OTFDataLoader::ParseCharFromPos(const BYTE** pPos)
{
	UCHAR preVal = 0;
	memcpy(&preVal, (*pPos), sizeof(preVal));
	(*pPos) += sizeof(UCHAR);
	return preVal;
}

/********************************************************************************************

>	INT OTFDataLoader::Parse3ByteFromPos(const BYTE** pPos)

	Author:		Greg Watroba
	Created:	8/12/2015
	Inputs:		pPos - current pointer to BYTE data
	Outputs		INT/CHAR 3 byte value read from BYTE data
	Returns:	-
	Purpose:	Parses INT/CHAR 3 byte value from given BYTE data

********************************************************************************************/

INT OTFDataLoader::Parse3ByteFromPos(const BYTE** pPos)
{
	INT iPV1 = ParseCharFromPos(pPos);
	INT iPV2 = ParseCharFromPos(pPos);
	INT iPV3 = ParseCharFromPos(pPos);
	return ((iPV1 << 16) | (iPV2 << 8) | iPV3);
}

/********************************************************************************************

>	INT32 OTFDataLoader::ParseOffsetFromPos(const BYTE** pPos, INT iOffsetSize)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pPos - current pointer to BYTE data
				iOffsetSize - size of needed offset value in bytes
	Outputs		INT32 offset value
	Returns:	-
	Purpose:	Returns offset value depending on size parameter

********************************************************************************************/

INT32 OTFDataLoader::ParseOffsetFromPos(const BYTE** pPos, INT& iOffsetSize)
{
	switch (iOffsetSize)
	{
		case 1: return ParseCharFromPos(pPos);
		case 2: return ParseUShortFromPos(pPos);
		case 3: return Parse3ByteFromPos(pPos);
		default: return ParseLongFromPos(pPos);
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseFontHeader(CDC *pDC)

	Author:		Greg Watroba
	Created:	2/12/2015
	Inputs:		pDC	- DC with currently selected font
	Outputs		Fills apropriate data in OTFData structure
	Returns:	-
	Purpose:	Parses Open Type Font Data

********************************************************************************************/

void OTFDataLoader::ParseFontData(HDC hDC, VFontBase* pBase, ParseOptions options)
{
	if (m_bOTFDataParsingFinished)
		return;

	ERROR2VOIDIF(!hDC, "OTFDataLoader::ParseFontData : Device Context cannot be NULL");

	TRACEUSER("Greg", "*** STARTING TO PARSE NEW FONT ***\n");

	// prepare data
	OTFData.ClearData();
	OTFData.hDC = hDC;

	DWORD DataSize = ::GetFontData(hDC, 0, 0, nullptr, 0);
	unique_ptr<BYTE[]> FontDataTable(new BYTE[DataSize]);

	if (::GetFontData(hDC, 0, 0, FontDataTable.get(), DataSize) != DataSize)
	{
		TRACEUSER("Greg", "Error while reading header data\n");
		return;
	}

	// parsing header
	ParseFontHeader(FontDataTable.get());
	ParseFontPreglyphData(FontDataTable.get());

	if (OTFData.bBadOpenTypeData)
	{
		TRACEUSER("Greg", "Bad Open Type Font data. Ending parsinng operation\n");
		return;
	}

	if ((OTFData.iGlyphLocationStart == 0 || OTFData.iGlyphLength == 0) &&
		OTFData.uCFFStart == 0 && OTFData.uTyp1Start == 0 &&
		OTFData.uBitmapDataStart != 0 && OTFData.uBitmapLocationStart != 0)
	{
		OTFData.bOnlyStrikes = TRUE;
	}

	// we have to go with only one glyph set
	if (!OTFData.bOnlyStrikes && OTFData.iGlyphLocationStart != 0 &&
		OTFData.iGlyphStart != 0 && OTFData.uCFFStart != 0)
	{
		OTFData.uCFFStart = 0;
	}

	BOOL bFontsSameAsNew = FALSE;
	BOOL bFontsInOrder2 = FALSE;

	if (OTFData.bOnlyStrikes)
	{
		OTFData.vChars.resize(OTFData.iGlyphCnt + 1);
		for (UINT16 i = 0; i < OTFData.iGlyphCnt + 1; ++i)
			OTFData.vChars[i] = new SplineChar();
		OTFData.bIsOrder2Type = bFontsInOrder2;
	}
	//We're only interested in TTF/OTF data
	else if (OTFData.iGlyphLocationStart != 0 && OTFData.iGlyphStart != 0)
	{
		OTFData.bIsOrder2Type = (!bFontsSameAsNew || (bFontsSameAsNew && bFontsInOrder2));

		if (OTFData.iGVarStart != 0 && OTFData.iFVarStart != 0)
			OTFData.bIsOrder2Type = TRUE;

		TRACEUSER("Greg", "Parsing TTF Glyphs.\n");
		ParseTTFGlyphs(FontDataTable.get());
		ParseTTFNames(FontDataTable.get());
	}
	else if (OTFData.uCFFStart != 0)
	{
		OTFData.bIsOrder2Type = (bFontsSameAsNew && bFontsInOrder2);
		TRACEUSER("Greg", "Parsing CFF Glyphs.\n");
		ParseCFFGlyphs(FontDataTable.get(), options & ParseOptions::keKeepGlyphData);
	}
	else if (OTFData.uTyp1Start != 0)
	{
		TRACEUSER("Greg", "Parsing Typ1 Glyphs.\n");
		ParseTyp1Glyphs(FontDataTable.get());
	}

	if (OTFData.bBadOpenTypeData)
		return;

	ParseTTFEncodings(FontDataTable.get(), GIUT_NORMAL);
	MapCharEncodings();

	if (OTFData.uOS2Start != 0)
		ParseTTFOS2Metrics(FontDataTable.get());

	ParseTTFPostNames(FontDataTable.get());

	if (OTFData.uGDEFStart != 0)
	{
		ParseTTFGDef(FontDataTable.get());
	}
	else
	{
		if (OTFData.uPropStart != 0)
			ParseTTFProp(FontDataTable.get());

		if (OTFData.uLcarStart != 0)
			ParseTTFLcar(FontDataTable.get());
	}

	/*

	if (OTFData.iBaseStart != 0)
		ParseTTFBase(FontDataTable.get());
	else if (OTFData.uBSLNStart != 0)
		ParseTTFBSLN(FontDataTable.get());

	if (OTFData.uGASPStart != 0)
		ParseTTFGASP(FontDataTable.get());

	if (OTFData.bIsOrder2Type && OTFData.iGVarStart != 0 && OTFData.iFVarStart != 0)
		ParseTTFVariations(FontDataTable.get());*/

	if (OTFData.uKernStart != 0)
	{
		OTFData.uKernLength = ::GetKerningPairsW(hDC, 0, 0);
		if (OTFData.uKernLength > 0)
			TRACEUSER("Greg", "Found %d old type kerning pairs. If no GPOS Table, kern tables will be processed.\n", OTFData.uKernLength);
	}

	if (OTFData.iGposStart != 0 && !OTFData.bBadOpenTypeData)
	{
		TRACEUSER("Greg", "Parsing GPOS data for font kerning\n");
		const BYTE* pGposTableStart = FontDataTable.get() + OTFData.iGposStart;
		OTFDataLoader::ParseGposGsubData(pGposTableStart, OTFData.iGposLength, TRUE, GIUT_NORMAL);
	}

	if (OTFData.iGsubStart != 0 && !OTFData.bBadOpenTypeData && pBase->UseOpenTypeSubstitutionData())
	{
		TRACEUSER("Greg", "Parsing GSUB data for font ligatures\n");
		const BYTE* pGsubTableStart = FontDataTable.get() + OTFData.iGsubStart;
		OTFDataLoader::ParseGposGsubData(pGsubTableStart, OTFData.iGsubLength, FALSE, GIUT_NORMAL);

#ifdef BUILD_TEXT_ADVANCED_OPENTYPE
		TRACEUSER("Greg", "Processing substitution rules\n");
		ProcessSubstitutionRulesData();
		ProcessSubstitutionData();
#endif
	}

	// Not needed for now - can be useful in the future
#if 0
	if (OTFData.iMorxStart != 0 && OTFData.iMortStart != 0)
		ParseTTFMort(FontDataTable.get());

	if (OTFData.uJSTFStart != 0)
		ParseTTFJSTF(FontDataTable.get());

	if (OTFData.uPFEDStart != 0)
		ParseTTFPFED(FontDataTable.get());

	if (OTFData.uTexStart != 0)
		ParseTTFTex(FontDataTable.get());

	if (OTFData.iMathStart != 0)
		ParseTTFMath(FontDataTable.get());

	TTFFixupReferences();
#endif

	if (!OTFData.bBadOpenTypeData)
		m_bOTFDataParsingFinished = TRUE;
}

/********************************************************************************************

>	void OTFDataLoader::ParseFontHeader(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	2/12/2015
	Inputs:		pFontData - pointer to raw font data
	Outputs		Fills apropriate data in OTFData structure
	Returns:	-
	Purpose:	Parses header font data for initial information

********************************************************************************************/

void OTFDataLoader::ParseFontHeader(const BYTE* pFontData)
{
	const BYTE* pDataTablePos = pFontData;
	INT32 iVersion = ParseLongFromPos(&pDataTablePos);
	OTFData.bIsTrueTypeCollection = (iVersion == CHR('t', 't', 'c', 'f'));

	if (iVersion < 0)
	{
		TRACEUSER("Greg", "Wrong format - data not ready yet\n");
		OTFData.bBadOpenTypeData = TRUE;
		return;
	}

	if (iVersion == CHR('t', 'y', 'p', '1') || iVersion == CHR('C', 'I', 'D', ' '))
	{
		TRACEUSER("Greg", "Not an Open Type font format\n");
		return;
	}

	TRACEUSER("Greg", "Font version: %d\n", iVersion);

	OTFData.iNumTables = ParseUShortFromPos(&pDataTablePos);

	//skip search range
	MOVE16BITS(pDataTablePos);
	//skip entry selector
	MOVE16BITS(pDataTablePos);
	//skip range shift
	MOVE16BITS(pDataTablePos);

	for (INT32 i = 0; i < OTFData.iNumTables; ++i)
	{
		UINT32 uTag = ParseLongFromPos(&pDataTablePos);
		//INT32 iChecksum = ParseLongFromPos(&pDataTablePos);
		MOVE32BITS(pDataTablePos);
		UINT32 uOffset = ParseLongFromPos(&pDataTablePos);
		UINT32 uLength = ParseLongFromPos(&pDataTablePos);

		switch (uTag)
		{
			case CHR('B', 'A', 'S', 'E'):
			{
				OTFData.iBaseStart = uOffset;
				break;
			}
			case CHR('B', 'D', 'F', ' '):
			{
				OTFData.uBDFStart = uOffset;
				break;
			}
			case CHR('b', 's', 'l', 'b'):
			{
				OTFData.uBSLNStart = uOffset;
				break;
			}
			case CHR('C', 'F', 'F', ' '):
			{
				OTFData.uCFFStart = uOffset;
				OTFData.uCFFLength = uLength;
				break;
			}
			case CHR('C', 'I', 'D', ' '):
			case CHR('T', 'Y', 'P', '1'):
			{
				OTFData.uTyp1Start = uOffset;
				OTFData.uTyp1Length = uLength;
				break;
			}
			case CHR('c', 'm', 'a', 'p'):
			{
				OTFData.iEncodingStart = uOffset;
				break;
			}
			case CHR('c', 'v', 't', ' '):
			{
				OTFData.uCVTStart = uOffset;
				OTFData.uCVTLen = uLength;
				break;
			}
			case CHR('f', 'e', 'a', 't'):
			{	
				OTFData.uFeatureStart = uOffset;
				break;
			}
			case CHR('F', 'F', 'T', 'M'):
			{
				OTFData.uFontTimeStamp = uOffset;
				break;
			}
			case CHR('f', 'p', 'g', 'm'):
			{
				OTFData.uFPGMStart = uOffset;
				OTFData.uFPGMLength = uLength;
				break;
			}
			case CHR('P', 'f', 'E', 'd'):
			{
				OTFData.uPFEDStart = uOffset;
			}
			case CHR('g', 'a', 's', 'p'):
			{
				OTFData.uGASPStart = uOffset;
				break;
			}
			case CHR('G', 'D', 'E', 'F'):
			{
				OTFData.uGDEFStart = uOffset;
				OTFData.uGDEFLength = uLength;
				break;
			}
			case CHR('g', 'l', 'y', 'f'):
			{
				OTFData.iGlyphStart = uOffset;
				OTFData.iGlyphLength = uLength;
				break;
			}
			case CHR('G', 'P', 'O', 'S'):
			{
				OTFData.iGposStart = uOffset;
				OTFData.iGposLength = uLength;
				break;
			}
			case CHR('G', 'S', 'U', 'B'):
			{
				OTFData.iGsubStart = uOffset;
				OTFData.iGsubLength = uLength;
				break;
			}
			case CHR('h', 'e', 'a', 'd'):
			{
				OTFData.iHeadStart = uOffset;
				break;
			}
			case CHR('h', 'h', 'e', 'a'):
			{
				OTFData.iHHeaStart = uOffset;
				break;
			}
			case CHR('h', 'm', 't', 'x'):
			{
				OTFData.uHMetricsStart = uOffset;
				break;
			}
			case CHR('J', 'S', 'T', 'F'):
			{
				OTFData.uJSTFStart = uOffset;
				OTFData.uJSTFLength = uLength;
				break;
			}
			case CHR('k', 'e', 'r', 'n'):
			{
				OTFData.uKernStart = uOffset;
				break;
			}
			case CHR('l', 'c', 'a', 'r'):
			{
				OTFData.uLcarStart = uOffset;
				break;
			}
			case CHR('l', 'o', 'c', 'a'):
			{
				OTFData.iGlyphLocationStart = uOffset;
				OTFData.iGlyphLocationLength = uLength;
				/* 
					the minus one is because there is one extra entry 
					to give the length of the last glyph 
				*/
				OTFData.iGlyphCnt = uLength / 2 - 1;
				break;
			}
			case CHR('M', 'A', 'T', 'H'):
			{
				OTFData.iMathStart = uOffset;
				OTFData.iMathLength = uLength;
				break;
			}
			case CHR('m', 'a', 'x', 'p'):
			{
				OTFData.iMaxPStart = uOffset;
				OTFData.iMaxPLen = uLength;
				break;
			}
			case CHR('m', 'o', 'r', 't'):
			{
				OTFData.iMortStart = uOffset;
				break;
			}
			case CHR('m', 'o', 'r', 'x'):
			{
				OTFData.iMorxStart = uOffset;
				break;
			}
			case CHR('n', 'a', 'm', 'e'):
			{
				OTFData.iCopyrightStart = uOffset;
				break;
			}
			case CHR('o', 'p', 'b', 'd'):
			{
				OTFData.uOPBDStart = uOffset;
				break;
			}
			case CHR('O', 'S', '/', '2'):
			{
				OTFData.uOS2Start = uOffset;
				break;
			}
			case CHR('p', 'r', 'e', 'p'):
			{
				OTFData.uPrepStart = uOffset;
				OTFData.uPrepLength = uLength;
				break;
			}
			case CHR('p', 'r', 'o', 'p'):
			{
				OTFData.uPropStart = uOffset;
				break;
			}
			case CHR('p', 'o', 's', 't'):
			{
				OTFData.uPostScriptStart = uOffset;
				break;
			}
			case CHR('T', 'e', 'X', ' '):
			{
				OTFData.uTexStart = uOffset;
				break;
			}
			case CHR('v', 'h', 'e', 'a'):
			{
				OTFData.uVHead = uOffset;
				break;
			}
			case CHR('v', 'm', 't', 'x'):
			{
				OTFData.uVMetrics = uOffset;
				break;
			}
			//for Apple fonts
			case CHR('g', 'v', 'a', 'r'):
			{
				OTFData.iGVarStart = uOffset;
				OTFData.iGVarLength = uLength;
				break;
			}
			case CHR('f', 'v', 'a', 'r'):
			{
				OTFData.iFVarStart = uOffset;
				OTFData.iFVarLength = uLength;
				break;
			}
			case CHR('a', 'v', 'a', 'r'):
			{
				OTFData.uAvarStart = uOffset;
				OTFData.uAvarLength = uLength;
				break;
			}
			case CHR('c', 'v', 'a', 'r'):
			{
				OTFData.uCvarStart = uOffset;
				OTFData.uCvarLength = uLength;
				break;
			}
			//for bitmap data - probably will never be used here
			//Apple/MS use a different tag, but the same format.
			case CHR('b', 'd', 'a', 't'):
			case CHR('E', 'B', 'D', 'T'):
			{
				OTFData.uBitmapDataStart = uOffset;
				OTFData.uBitmapDataLength = uLength;
				break;
			}
			case CHR('b', 'l', 'o', 'c'):
			case CHR('E', 'B', 'L', 'C'):
			{
				OTFData.uBitmapLocationStart = uOffset;
				OTFData.uBitmapLocationLength = uLength;
				break;
			}
			default:
			{
				CHAR sMSG[256];
				sprintf_s(sMSG, "Tag omitted: %c%c%c%c\n", uTag >> 24, uTag >> 16, uTag >> 8, uTag);
				TRACEUSER("Greg",sMSG);
			}
		}
	}

	if (OTFData.iGposStart == 0)
		TRACEUSER("Greg", "No Gpos data found.\n");

	if (OTFData.iGlyphStart == 0 && OTFData.uCFFStart != 0)
		TRACEUSER("Greg", "No TTF/OTF glyphs found. Compact Font Format have to be processed\n");

	if (OTFData.iGlyphStart == 0 && OTFData.uCFFStart == 0 && OTFData.uTyp1Start != 0)
		TRACEUSER("Greg", "No TTF/OTF/CFF glyphs found. Type 1 Format have to be processed\n");

	if (OTFData.iGsubStart == 0)
		TRACEUSER("Greg", "No GSUB data found.\n");
}

/********************************************************************************************

>	void OTFDataLoader::ParseFontPreglyphData(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	2/12/2015
	Inputs:		pFontData - pointer to raw font data
	Outputs		Fills apropriate data in OTFData structure
	Returns:	-
	Purpose:	Parses data prior to glyph loading, kerning and ligature data

********************************************************************************************/

void OTFDataLoader::ParseFontPreglyphData(const BYTE* pFontData)
{
	if (OTFData.iHeadStart != 0)
		ParseHeadData(pFontData);

	if (OTFData.iHHeaStart != 0)
		ParseHHeaData(pFontData);

	if (OTFData.iMaxPStart != 0)
		ParseMaxpData(pFontData);

	ParseCopyrightsData(pFontData);
}

/********************************************************************************************

>	void OTFDataLoader::ParseGposGsubData(const BYTE* pPosData, const DWORD& iTableSize,
										 BOOL bIsGPOS, GsubInUseType eType)

	Author:		Greg Watroba
	Created:	23/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iTableSize - size of positioning data array
				bIsGPOS - flag whether its GPOS or GSUB data
				eType - type of GSUB data if relevant
	Outputs		-
	Returns:	-
	Purpose:	Parses GPOS/GSUB data array to kerning pairs
	See also:	FontKerningPairsCacheEntry::CacheFontKerns

********************************************************************************************/

void OTFDataLoader::ParseGposGsubData(const BYTE* pPosData, const DWORD& iTableSize,
									 BOOL bIsGPOS, GsubInUseType eType)
{
	const BYTE* pBasicDataPos = pPosData;
	std::string ParseErrMsgSuffix = { (bIsGPOS) ? " in GPOS Data\n" : " in GSUB Data\n"};
	
	if (bIsGPOS)
		OTFData.iPositioningSizeBounds = (OTFData.iGposLength != 0) ? OTFData.iGposLength : iTableSize;
	else
		OTFData.iPositioningSizeBounds = (OTFData.iGsubLength != 0) ? OTFData.iGsubLength : iTableSize;

	std::vector<Lookup>& vLookups = (bIsGPOS) ? OTFData.vGPOSLookups : OTFData.vGSUBLookups;
	vLookups.clear();
	OTFData.vecLookupSubTables.clear();
	OTFData.LookupSubtableUID = 0;

	//UINT32 uVersion = ParseLongFromPos(&pBasicDataPos);
	MOVE32BITS(pBasicDataPos);

	USHORT iScriptOffset = ParseUShortFromPos(&pBasicDataPos);
	USHORT iFeatureOffset = ParseUShortFromPos(&pBasicDataPos);
	USHORT iLookupOffset = ParseUShortFromPos(&pBasicDataPos);

	std::vector<Scripts> vScripts;
	ParseGposScripts(&vScripts, pPosData, iScriptOffset, bIsGPOS);

	if (OTFData.bBadOpenTypeData)
	{
		std::string err = "Error occurred while parsing scripts" + ParseErrMsgSuffix;
		TRACEUSER("Greg", err.c_str());
		return;
	}

	std::vector<Feature> vFeatures;
	ParseGposFeatures(&vFeatures, pPosData, iFeatureOffset, bIsGPOS);

	if (OTFData.bBadOpenTypeData)
	{
		std::string err = "Error occurred while parsing features" + ParseErrMsgSuffix;
		TRACEUSER("Greg", err.c_str());
		return;
	}

	ParseGposLookups(&vLookups, pPosData, iLookupOffset, bIsGPOS);

	if (OTFData.bBadOpenTypeData)
	{
		std::string err = "Error occurred while parsing lookups" + ParseErrMsgSuffix;
		TRACEUSER("Greg", err.c_str());
		return;
	}

	if (vLookups.empty())
	{
		vScripts.clear();
		vFeatures.clear();
		std::string err = "No lookups found" + ParseErrMsgSuffix;
		TRACEUSER("Greg", err.c_str());
		return;
	}

	TagLookupsWithScript(&vScripts, &vFeatures, &vLookups);

	if (OTFData.bBadOpenTypeData)
	{
		std::string err = "Error occurred while tagging lookups" + ParseErrMsgSuffix;
		TRACEUSER("Greg", err.c_str());
		return;
	}

	if (OTFData.vChars.empty())
	{
		std::string err = "No lookups found" + ParseErrMsgSuffix;
		TRACEUSER("Greg", err.c_str());
		return;
	}

	//populate lookup with actual GPOS data
	for (auto &pLookup : vLookups)
	{
		if (pLookup.vOTLookups.empty())
			continue;

		UINT i = 0;
		for (auto& pSubTable : pLookup.vOTLookups.back()->vSubTables)
		{
			// we filter out not supported subtables
			if (!bIsGPOS && std::find(OpenTypeFontData::s_vSupportedSubstitutions.begin(), 
									  OpenTypeFontData::s_vSupportedSubstitutions.end(), 
									  pSubTable->eGlyphSet) == OpenTypeFontData::s_vSupportedSubstitutions.end() && 
				pSubTable->eGlyphSet != GSET_NONE)
			{
				continue;
			}

			// we look for substitution data that we won't support
			if (pSubTable->pOTLookup)
			{
				BOOL bCanOmit = FALSE;
				for (auto& pFeature : pSubTable->pOTLookup->vFeatures)
				{
					if (pSubTable->eGlyphSet != pFeature->eGlyphSet)
						continue;

					if (pFeature->vScripts.empty())
					{
						bCanOmit = FALSE;
						break;
					}

					for (auto& pScript : pFeature->vScripts)
					{
						auto res = std::find(OpenTypeFontData::s_vScriptsToSupport.begin(), OpenTypeFontData::s_vScriptsToSupport.end(), pScript->uScript);
						if (res == OpenTypeFontData::s_vScriptsToSupport.end())
						{
							bCanOmit = TRUE;
						}
						else
						{
							bCanOmit = FALSE;
							break;
						}
					}
				}

				if (bCanOmit)
					continue;
			}

			INT32 iSubTableOffset = pLookup.vSubTabOffsets[i];

			if (bIsGPOS)
			{
				GPOSLookupSwitch(pPosData, iSubTableOffset, &pLookup, pSubTable, &vLookups);
			}
			else
			{
#ifdef BUILD_TEXT_LIGATURES
				GSUBLookupSwitch(pPosData, iSubTableOffset, &pLookup, pSubTable, &vLookups, eType);
#endif
			}

			++i;
		}
	}

	//set names for lookups
	if (eType == GIUT_NORMAL)
	{
		for (auto &pLookup : vLookups)
			NameOTLookup(&pLookup.vOTLookups);
	}

	if (OTFData.bBadOpenTypeData)
	{
		std::string err = "Error occured while working" + ParseErrMsgSuffix;
		TRACEUSER("Greg", err.c_str());
	}

	//since all data has been rewritten to lookup vector, 
	//we don't need scripts and features anymore
	vScripts.clear();
	vFeatures.clear();
}

#ifdef BUILD_TEXT_ADVANCED_OPENTYPE

/********************************************************************************************

>	void OTFDataLoader::ProcessSubstitutionRulesData()

	Author:		Greg Watroba
	Created:	1/03/2017
	Inputs:		-
	Outputs		-
	Returns:	-
	Purpose:	Processes contextual rules data for substitutions based on tables parsed
				from a font file.
	See also:	https://www.microsoft.com/typography/otspec/gsub.htm#CS

********************************************************************************************/

static void GetLangsFromLookupForGlyphset(std::vector<UINT32>* pOutResult, OTLookup* pLookup, GlyphSetEnum eGlyphSet)
{
	if (!pOutResult)
		return;

	for (auto& pFeat : pLookup->vFeatures)
	{
		if (pFeat->eGlyphSet == eGlyphSet)
		{
			for (auto& pScript : pFeat->vScripts)
			{
				for (int i = 0; i < MAX_LANG; ++i)
				{
					if (pScript->uLangs[i] != 0 && std::find(pOutResult->begin(), pOutResult->end(), pScript->uLangs[i]) == pOutResult->end())
						pOutResult->push_back(pScript->uLangs[i]);
				}

				for (auto& pLang : pScript->vMoreLangs)
				{
					if (pLang != 0 && std::find(pOutResult->begin(), pOutResult->end(), pLang) == pOutResult->end())
						pOutResult->push_back(pLang);
				}
			}
		}
	}
}

void OTFDataLoader::ProcessSubstitutionRulesData()
{
	OTFData.lstFPST.reverse();

	for (auto& pFPST : OTFData.lstFPST)
	{
		for (auto& pSubTable : pFPST->vSubTables)
		{
			if (std::find(OpenTypeFontData::s_vSupportedRules.begin(), OpenTypeFontData::s_vSupportedRules.end(), pSubTable->eGlyphSet) == OpenTypeFontData::s_vSupportedRules.end())
				continue;

			const UINT16 Priority = (UINT16)pSubTable->pOTLookup->iLookupIndex;
			const GlyphSetEnum eGlyphSet = pSubTable->eGlyphSet;
			std::vector<UINT32> vLangs;
			GetLangsFromLookupForGlyphset(&vLangs, pSubTable->pOTLookup, eGlyphSet);

			for (auto& pRule : pFPST->vRules)
			{
				switch (pRule->eType)
				{
					case FPSTR_GLYPH: { ProcessGlyphRule(static_cast<FPSTGlyphRule*>(pRule), eGlyphSet, Priority, vLangs); }					break;
					case FPSTR_CLASS: { ProcessClassRule(static_cast<FPSTClassRule*>(pRule), eGlyphSet, Priority, vLangs); }					break;
					case FPSTR_COVERAGE: { ProcessCoverageRule(static_cast<FPSTCoverageRule*>(pRule), eGlyphSet, Priority, vLangs); }			break;
					case FPSTR_REV_COVERAGE: { ProcessRevCoverageRule(static_cast<FPSTRevCoverageRule*>(pRule), eGlyphSet, Priority, vLangs); }	break;
				}
			}
		}
	}
}

/********************************************************************************************

>	std::list<LookupSubTable*> OTFDataLoader::GetSubtableList(FPSTRule* pRule)

	Author:		Greg Watroba
	Created:	3/03/2017
	Inputs:		-
	Outputs		-
	Returns:	-
	Purpose:	Gets a list of lookup subtable pointers from given rule
	See also:	https://www.microsoft.com/typography/otspec/gsub.htm#CS

********************************************************************************************/

std::list<LookupSubTable*> OTFDataLoader::GetSubtableList(FPSTRule* pRule)
{
	std::list<LookupSubTable*> lstSubtables;

	for (auto& pSeqLookup : pRule->vLookups)
	{
		if (!pSeqLookup)
			continue;

		if (OTFData.vGSUBLookups.size() <= (size_t)pSeqLookup->iLookupIndex)
		{
			TRACEUSER("Greg", "OTFDataLoader::ProcessGlyphRule : Lookup index outside lookups container!\n");
			continue;
		}

		for (auto& pOTLookups : OTFData.vGSUBLookups[pSeqLookup->iLookupIndex].vOTLookups)
		{
			for (auto& pLookupSubTable : pOTLookups->vSubTables)
				lstSubtables.push_back(pLookupSubTable);
		}
	}

	return lstSubtables;
}

void OTFDataLoader::GetCharContainerFromGlyphs(std::vector<std::wstring>* pOutResult, std::vector<std::vector<UINT16>>& srcContainer)
{
	if (!pOutResult)
		return;

	pOutResult->reserve(srcContainer.size());

	// preceeding glyphs
	for (auto& pStr : srcContainer)
	{
		if (pStr.empty())
			continue;

		std::wstring pChars;
		pChars.reserve(pStr.size());
		for (auto& pChr : pStr)
		{
			if (pChr != 0xffff)
				pChars.push_back((WCHAR)OTFData.vChars[pChr]->iUnicodeEnc);
		}

		std::sort(pChars.begin(), pChars.end());

		pOutResult->push_back(pChars);
	}
}

/********************************************************************************************

>	BOOL OTFDataLoader::AllGlyphsInContainerExist(std::vector<UINT16>& srcContainer)

	Author:		Greg Watroba
	Created:	3/03/2017
	Inputs:		srcContainer - source container of glyph indices
	Outputs		-
	Returns:	TRUE if all glyphs in container are correct, FALSE if otherwise
	Purpose:	Determines validity of source container to be processed

********************************************************************************************/

BOOL OTFDataLoader::AllGlyphsInContainerExist(std::vector<UINT16>& srcContainer)
{
	for (auto& pKey : srcContainer)
	{
		if (OTFData.vChars.size() <= pKey || OTFData.vChars[pKey] == nullptr)
		{
			TRACEUSER("Greg", "Incorrect glyph index: %d\n", pKey);
			return FALSE;
		}
	}

	return TRUE;
}

/********************************************************************************************

>	void OTFDataLoader::ProcessGlyphRule(FPSTGlyphRule* pRule, const GlyphSetEnum& eGlyphSet, 
										 UINT16 uPriority)

	Author:		Greg Watroba
	Created:	1/03/2017
	Inputs:		pRule - glyph rule pointer
				eGlyphSet - glyph set to which rule applies
				uPriority - substitution rule priority
	Outputs		-
	Returns:	-
	Purpose:	Process glyph rule data
	See also:	https://www.microsoft.com/typography/otspec/gsub.htm#CS

********************************************************************************************/
void OTFDataLoader::ProcessGlyphRule(FPSTGlyphRule* pRule, const GlyphSetEnum& eGlyphSet, 
									 UINT16 uPriority, std::vector<UINT32>& vLangs)
{
	ERROR2VOIDIF(!pRule, "OTFDataLoader::ProcessGlyphRule : Rule cannot be NULL!\n");
	TRACEUSER("Greg", "OTFDataLoader::ProcessGlyphRule : Trying to parse simple glyph rule\n");
}

/********************************************************************************************

>	void OTFDataLoader::ProcessClassRule(FPSTClassRule* pRule, const GlyphSetEnum& eGlyphSet, 
										 UINT16 uPriority)

	Author:		Greg Watroba
	Created:	1/03/2017
	Inputs:		pRule - class rule pointer
				eGlyphSet - glyph set to which rule applies
				uPriority - substitution rule priority
	Outputs		-
	Returns:	-
	Purpose:	Process class rule data
	See also:	https://www.microsoft.com/typography/otspec/gsub.htm#CS

********************************************************************************************/

void OTFDataLoader::ProcessClassRule(FPSTClassRule* pRule, const GlyphSetEnum& eGlyphSet, 
									 UINT16 uPriority, std::vector<UINT32>& vLangs)
{
	ERROR2VOIDIF(!pRule, "OTFDataLoader::ProcessClassRule : Rule cannot be NULL!\n");
	TRACEUSER("Greg", "OTFDataLoader::ProcessClassRule : Trying to parse class glyph rule\n");
}

/********************************************************************************************

>	void OTFDataLoader::ProcessCoverageRule(FPSTClassRule* pRule, const GlyphSetEnum& eGlyphSet, 
											UINT16 uPriority)

	Author:		Greg Watroba
	Created:	1/03/2017
	Inputs:		pRule - coverage rule pointer
				eGlyphSet - glyph set to which rule applies
				uPriority - substitution rule priority
	Outputs		-
	Returns:	-
	Purpose:	Process coverage rule data
	See also:	https://www.microsoft.com/typography/otspec/gsub.htm#CS

********************************************************************************************/

std::vector<SubstitutionRule> vecRulesToMerge;

void OTFDataLoader::ProcessCoverageRule(FPSTCoverageRule* pRule, const GlyphSetEnum& eGlyphSet, 
										UINT16 uPriority, std::vector<UINT32>& vLangs)
{
	ERROR2VOIDIF(!pRule, "OTFDataLoader::ProcessCoverageRule : Rule cannot be NULL!\n");
	//TRACEUSER("Greg", "OTFDataLoader::ProcessCoverageRule : Trying to parse coverage rule\n");

	// we get list of lookups
	auto lstSubTables = GetSubtableList(pRule);

	UINT8 PreSize = (UINT8)pRule->vecBCoverGlyphs.size();
	UINT8 KeySize = (UINT8)pRule->vecNCoverGlyphs.size();
	UINT8 PostSize = (UINT8)pRule->vecFCoverGlyphs.size();

	// preceeding glyphs - they're stored in reverse
	std::vector<std::wstring> preChars;
	GetCharContainerFromGlyphs(&preChars, pRule->vecBCoverGlyphs);
	
	std::vector<std::wstring> postChars;
	GetCharContainerFromGlyphs(&postChars, pRule->vecFCoverGlyphs);

	std::vector<std::wstring> keyChars;
	GetCharContainerFromGlyphs(&keyChars, pRule->vecNCoverGlyphs);

	if (lstSubTables.empty())
	{
		SubstitutionRule Rule;
		Rule.Pre = preChars;
		Rule.Keys = keyChars;
		Rule.Post = postChars;
		vecRulesToMerge.push_back(Rule);
	}

	// now, it depends on a type of the OTLookup table which
	// type of substitution we will use
	for (auto& pLookupSubTable : lstSubTables)
	{
		for (auto ruleToMerge : vecRulesToMerge)
		{
			ruleToMerge.SubstitutionTableID = UINT16_MAX;
			ruleToMerge.Priority = uPriority;
			ruleToMerge.Langs = vLangs;

			//OTFData.vecNeededLookups.push_back(pLookupSubTable->UniqueID);
			OTFData.SubRuleSet[eGlyphSet].first.push_back(ruleToMerge);
		}

		SubstitutionRule Rule;
		Rule.Pre = preChars;
		Rule.Keys = keyChars;
		Rule.Post = postChars;

		Rule.SubstitutionTableID = pLookupSubTable->UniqueID;
		Rule.Priority = uPriority;
		Rule.Langs = vLangs;

		OTFData.vecNeededLookups.push_back(pLookupSubTable->UniqueID);
		OTFData.SubRuleSet[eGlyphSet].first.push_back(Rule);
	}

	// we must've merged rules so we can clear container
	if (!lstSubTables.empty())
		vecRulesToMerge.clear();

	if (OTFData.SubRuleSet[eGlyphSet].second.Pre < PreSize)
		OTFData.SubRuleSet[eGlyphSet].second.Pre = PreSize;

	if (OTFData.SubRuleSet[eGlyphSet].second.Key < KeySize)
		OTFData.SubRuleSet[eGlyphSet].second.Key = KeySize;

	if (OTFData.SubRuleSet[eGlyphSet].second.Post < PostSize)
		OTFData.SubRuleSet[eGlyphSet].second.Post = PostSize;

	if (OTFData.TotalMaxSizes.Pre < PreSize)
		OTFData.TotalMaxSizes.Pre = PreSize;

	if (OTFData.TotalMaxSizes.Key < KeySize)
		OTFData.TotalMaxSizes.Key = KeySize;

	if (OTFData.TotalMaxSizes.Post < PostSize)
		OTFData.TotalMaxSizes.Post = PostSize;
}

/********************************************************************************************

>	void OTFDataLoader::ProcessRevCoverageRule(FPSTClassRule* pRule, const GlyphSetEnum& eGlyphSet,
											   UINT16 uPriority)

	Author:		Greg Watroba
	Created:	1/03/2017
	Inputs:		pRule - reverse coverage rule pointer
				eGlyphSet - glyph set to which rule applies
				uPriority - substitution rule priority
	Outputs		-
	Returns:	-
	Purpose:	Process reverse coverage rule data
	See also:	https://www.microsoft.com/typography/otspec/gsub.htm#CS

********************************************************************************************/

void OTFDataLoader::ProcessRevCoverageRule(FPSTRevCoverageRule* pRule, const GlyphSetEnum& eGlyphSet, 
										   UINT16 uPriority, std::vector<UINT32>& vLangs)
{
	ERROR2VOIDIF(!pRule, "OTFDataLoader::ProcessRevCoverageRule : Rule cannot be NULL!");
}

/********************************************************************************************

>	void OTFDataLoader::ProcessSubstitutionData()

	Author:		Greg Watroba
	Created:	12/04/2018
	Inputs:		-
	Outputs		-
	Returns:	-
	Purpose:	Process substitution data based on parsed information from GSUB tables
	See also:	https://www.microsoft.com/typography/otspec/gsub.htm#CS

********************************************************************************************/

void OTFDataLoader::ProcessSubstitutionData()
{
	// curently we support from here only stylistic sets and ligatures

	for (auto& pLookup : OTFData.vGSUBLookups)
	{
		for (auto& pOTLookup : pLookup.vOTLookups)
		{
			if (pOTLookup->eLookupType != OTL_TYPE_GSUB_SINGLE &&
				pOTLookup->eLookupType != OTL_TYPE_GSUB_MULTIPLE &&
				pOTLookup->eLookupType != OTL_TYPE_GSUB_ALTERNATE &&
				pOTLookup->eLookupType != OTL_TYPE_GSUB_LIGATURE)
				continue;

			for (auto& pSubTable : pOTLookup->vSubTables)
			{
				const GlyphSetEnum eGlyphSet = pSubTable->eGlyphSet;

				if (std::find(OpenTypeFontData::s_vSupportedSubstitutions.begin(), 
					OpenTypeFontData::s_vSupportedSubstitutions.end(), 
					eGlyphSet) == OpenTypeFontData::s_vSupportedSubstitutions.end())
					continue;

				if (pSubTable->vecPosSubTab.empty())
					continue;

				const UINT16 Priority = (UINT16)pOTLookup->iLookupIndex;
				std::vector<UINT32> vLangs;
				GetLangsFromLookupForGlyphset(&vLangs, pSubTable->pOTLookup, eGlyphSet);

				SubstitutionRule Rule;
				Rule.SubstitutionTableID = pSubTable->UniqueID;
				Rule.Priority = Priority;
				Rule.Langs = vLangs;
				UINT8 maxKey = 0;

				// singular substitution tables entries
				for (auto& pPosSubTab : pSubTable->vecPosSubTab)
				{
					std::wstring key;
					switch (pPosSubTab->uType)
					{
						case PST_SUBSTITUTION:
						case PST_MULTIPLE:
						case PST_ALTERNATE:
						{
							key.push_back((WCHAR)OTFData.vChars[pPosSubTab->GlyphID]->iUnicodeEnc);
						}break;

						case PST_LIGATURE:
						{
							for (auto& pGlyph : pPosSubTab->PSTLig.vecGlyphs)
								key.push_back((WCHAR)OTFData.vChars[pGlyph]->iUnicodeEnc);
						}break;
					}

					if (!key.empty())
					{
						Rule.Keys.push_back(key);

						if (key.size() > maxKey)
							maxKey = (UINT8)key.size();
					}
				}

				OTFData.vecNeededLookups.push_back(pSubTable->UniqueID);
				OTFData.setStandardSubLookups.insert(pSubTable->UniqueID);
				OTFData.SubRuleSet[eGlyphSet].first.push_back(Rule);

				if (OTFData.SubRuleSet[eGlyphSet].second.Key < maxKey)
					OTFData.SubRuleSet[eGlyphSet].second.Key = maxKey;

				if (OTFData.TotalMaxSizes.Key < maxKey)
					OTFData.TotalMaxSizes.Key = maxKey;
			}
		}
	}
}

#endif //BUILD_TEXT_ADVANCED_OPENTYPE

/********************************************************************************************

>	void OTFDataLoader::MapCharEncodings()

	Author:		Greg Watroba
	Created:	13/01/2016
	Inputs:		-
	Outputs		Modifies iUnicodeEnc in vChars vector, in OTFData
	Returns:	-
	Purpose:	Sets character code based on glyph code

********************************************************************************************/

void OTFDataLoader::MapCharEncodings()
{
	std::fill(OTFData.s_vGlyphIndices.begin(), OTFData.s_vGlyphIndices.end(), (UINT16)0);
	GetGlyphIndicesW(OTFData.hDC, &OTFData.s_vCharSet.front(), UINT16_MAX, &OTFData.s_vGlyphIndices.front(), 0);

	// unicode values for specific glyph are in order
	// since we want to take doubled values in ascending order
	// we have to emplace them in descending order
	std::unordered_map<INT32, std::list<UINT16>> mGlyphMap;

	// also we have private use range for glyphs which won't have
	// appropriate unicode representation
	std::vector<UINT16> lFreePUR;
	lFreePUR.reserve(MAX_PUA_SIZE);

	for (UINT16 j = UINT16_MAX - 1; j > 0; --j)
	{
		UINT16 uGlyph = OTFData.s_vGlyphIndices[j];
		if (uGlyph > 0)
			mGlyphMap[uGlyph].push_back(j);
		else if (IsPrivateUseAreaChar((WCHAR)j))
			lFreePUR.push_back(j);
	}

	// sometimes instead of '0' empty glyph can be of different value
	// since in general case one glyph is assigned to one value
	// we check assignments of glyphs that are more than 1
	// we can try to salvage PUA values from those cases
	for (auto& pMapElement : mGlyphMap)
	{
		if (pMapElement.second.size() <= 1)
			continue;

		for (auto& pUnicodeValue : pMapElement.second)
		{
			if (IsPrivateUseAreaChar((WCHAR)pUnicodeValue))
				lFreePUR.push_back(pUnicodeValue);
		}
	}

	std::sort(lFreePUR.begin(), lFreePUR.end());
	auto last = std::unique(lFreePUR.begin(), lFreePUR.end());
	lFreePUR.erase(last, lFreePUR.end());

	//mapping found glyphs to char values
	//vChars have characters in ascending order
	for (UINT16 i = 0; i < OTFData.vChars.size(); ++i)
	{
		if (!OTFData.vChars[i])
			continue;

		INT32 iGlyphIndex = OTFData.vChars[i]->iOrigPos;

		// if we haven't found proper unicode value for glyph
		// we take it from private use range
		if (mGlyphMap[iGlyphIndex].empty())
		{
			if (!lFreePUR.empty())
			{
				OTFData.vChars[i]->iUnicodeEnc = lFreePUR.back();
				lFreePUR.pop_back();
			}
			continue;
		}

		OTFData.vChars[i]->iUnicodeEnc = mGlyphMap[iGlyphIndex].back();
		mGlyphMap[iGlyphIndex].pop_back();
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFOS2Metrics(const BYTE* pPosData)

	Author:		Greg Watroba
	Created:	20/09/2016
	Inputs:		-
	Outputs		Sets PMFInfo data (font metrics)
	Returns:	-
	Purpose:	Load font metrics from file

********************************************************************************************/

void OTFDataLoader::ParseTTFOS2Metrics(const BYTE* pPosData)
{
	const BYTE* pFontMetricsPos = pPosData + OTFData.uOS2Start;

	OTFData.uOS2Version = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.uAvgWidth = ParseUShortFromPos(&pFontMetricsPos);

	OTFData.PfmInfo.iWeight = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iWidth = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iFStype = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SubxSize = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SubySize = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SubxOff = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SubyOff = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SubxSize = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SubySize = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SupxOff = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2SupyOff = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2StriKeySize = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2StriKeyPos = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2FamilyClass = ParseUShortFromPos(&pFontMetricsPos);

	for (int i = 0; i < 10; ++i)
		OTFData.PfmInfo.cPanose[i] = ParseCharFromPos(&pFontMetricsPos);

	OTFData.PfmInfo.uPfmFamily = OTFData.PfmInfo.cPanose[0] == 2 ? 0x11 : OTFData.PfmInfo.cPanose[0] == 3 ? 0x41 : OTFData.PfmInfo.cPanose[0] == 4 ? 0x51 : 0x51;

	for (int i = 0; i < 4; ++i)
		OTFData.PfmInfo.uUnicodeRanges[i] = ParseLongFromPos(&pFontMetricsPos);

	OTFData.PfmInfo.bHasUnicodeRanges = TRUE;

	for (int i = 0; i < 4; ++i)
		OTFData.PfmInfo.cOS2Vendor[i] = ParseCharFromPos(&pFontMetricsPos);

	INT iSel = ParseUShortFromPos(&pFontMetricsPos);

	if (OTFData.uOS2Version >= 4)
	{
		OTFData.bUseTypoMetrics = (iSel & 128) ? TRUE : FALSE;
		OTFData.bWeightWidthSlopeOnly = (iSel & 256) ? TRUE : FALSE;
	}

	/*UINT uFirstChar = */ ParseUShortFromPos(&pFontMetricsPos);
	/*UINT uLastChar = */ ParseUShortFromPos(&pFontMetricsPos);

	OTFData.PfmInfo.iOS2TypoAscent = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2Typodescent = (short) ParseUShortFromPos(&pFontMetricsPos);

	double emSizeScale = OTFData.iEmSize > 0 ? (double)DEFAULT_EMSIZE / (double)OTFData.iEmSize : 1;
	OTFData.PfmInfo.iOS2TypoAscent = (INT16)(emSizeScale * OTFData.PfmInfo.iOS2TypoAscent);
	OTFData.PfmInfo.iOS2Typodescent = (INT16)(emSizeScale * OTFData.PfmInfo.iOS2Typodescent);

	if (OTFData.PfmInfo.iOS2TypoAscent - OTFData.PfmInfo.iOS2Typodescent == OTFData.iEmSize)
	{
		OTFData.iAscent = OTFData.PfmInfo.iOS2TypoAscent;
		OTFData.iDescent = OTFData.PfmInfo.iOS2Typodescent;
	}

	OTFData.PfmInfo.iOS2Typolinegap = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2Typolinegap = (INT16)(emSizeScale * OTFData.PfmInfo.iOS2Typolinegap);

	OTFData.PfmInfo.iOS2WinAscent = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.iOS2WinDescent = ParseUShortFromPos(&pFontMetricsPos);
	OTFData.PfmInfo.uWinAscentAdd = OTFData.PfmInfo.uWinDescentAdd = FALSE;
	OTFData.PfmInfo.uTypoAscentAdd = OTFData.PfmInfo.uTypoDescentAdd = FALSE;
	OTFData.PfmInfo.uPfmSet = TRUE;
	OTFData.PfmInfo.uPanoseSet = TRUE;
	OTFData.PfmInfo.uSubSuperSet = TRUE;

	if (OTFData.uOS2Version >= 1)
	{
		OTFData.PfmInfo.uCodePages[0] = ParseLongFromPos(&pFontMetricsPos);
		OTFData.PfmInfo.uCodePages[1] = ParseLongFromPos(&pFontMetricsPos);
		OTFData.PfmInfo.uHasCodePages = TRUE;
	}

	if (OTFData.uOS2Version == 0)
	{
		TRACEUSER("Greg", "Windows will reject fonts with an OS/2 version number of 0\n");
		OTFData.bBadOS2Data = TRUE;
	}
	else if (OTFData.uOS2Version == 1 && OTFData.uCFFStart != 0)
	{
		TRACEUSER("Greg", "Windows will reject otf (cff) fonts with an OS/2 version number of 1\n");
		OTFData.bBadOS2Data = TRUE;
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFPostNames(const BYTE* pPosData)

	Author:		Greg Watroba
	Created:	20/09/2016
	Inputs:		-
	Outputs		Set additional names for glyphs stored in post script data
	Returns:	-
	Purpose:	Reads additional glyphs names

********************************************************************************************/

void OTFDataLoader::ParseTTFPostNames(const BYTE* pPosData)
{
	
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFGDef(const BYTE* pPosData)

	Author:		Greg Watroba
	Created:	20/09/2016
	Inputs:		-
	Outputs		Set auxiliary data used in GSUB among other further data
	Returns:	-
	Purpose:	Reads GDEF data

********************************************************************************************/

void OTFDataLoader::ParseTTFGDef(const BYTE* pPosData)
{
	const BYTE* pGDefPos = pPosData + OTFData.uGDEFStart;

	INT iVersion = ParseLongFromPos(&pGDefPos);

	if (iVersion != 0x00010000 && iVersion != 0x00010002)
		return;

	OTFData.iGDefBounds = OTFData.uGDEFStart + OTFData.uGDEFLength;
	OTFData.iPositioningSizeBounds = OTFData.iGDefBounds;

	INT iGClass = ParseUShortFromPos(&pGDefPos);
	/* attach list = */ ParseUShortFromPos(&pGDefPos);

	INT iLigCaretOffset = ParseUShortFromPos(&pGDefPos);
	INT iMarkAttachOffset = ParseUShortFromPos(&pGDefPos);

	INT iMarkSetOffset = 0;
	if (iVersion == 0x00010002)
		iMarkSetOffset = ParseUShortFromPos(&pGDefPos);

	if (iGClass != 0)
	{
		std::vector<UINT16> vGClasses;

		GetClassDefinitionTable(&vGClasses, pPosData, OTFData.uGDEFStart + iGClass);
		for (int i = 0; i < OTFData.iGlyphCnt; ++i)
		{
			if (OTFData.vChars[i] && vGClasses[i] != 0)
				OTFData.vChars[i]->uGlyphClass = vGClasses[i] + 1;
		}
	}

	if (iMarkAttachOffset != 0)
	{
		std::string sFormatSpec = "MarkClass-";
		std::vector<UINT16> vMClasses;
		GetClassDefinitionTable(&vMClasses, pPosData, OTFData.uGDEFStart + iMarkAttachOffset);
		OTFData.iMarkClassCnt = ClassFindCnt(vMClasses, OTFData.iGlyphCnt);
		ClassToNames(&OTFData.vMarkClasses, OTFData.iMarkClassCnt, &vMClasses, OTFData.iGlyphCnt);
		OTFData.vMarkClassNames.resize(OTFData.iMarkClassCnt + 1);
		OTFData.vMarkClassNames[0].clear();
		for (int i = 0; i < OTFData.iMarkClassCnt; ++i)
		{
			OTFData.vMarkClassNames[i + 1].append(sFormatSpec);
			OTFData.vMarkClassNames[i + 1].append(std::to_string(i));
		}
	}

	if (iMarkSetOffset != 0)
	{
		std::string sFormatSpec = "MarkSet-";
		const BYTE* pMarkSetPos = pPosData + OTFData.uGDEFStart + iMarkSetOffset;
		INT iVersion = ParseUShortFromPos(&pMarkSetPos);

		if (iVersion == 1)
		{
			std::vector<UINT32> vOffsets;
			std::vector<UINT16> vGlyphs;
			OTFData.iMarkSetCnt = ParseUShortFromPos(&pMarkSetPos);
			vOffsets.resize(OTFData.iMarkSetCnt);

			for (int i = 0; i < OTFData.iMarkSetCnt; ++i)
				vOffsets[i] = ParseLongFromPos(&pMarkSetPos);

			OTFData.vMarkSets.resize(OTFData.iMarkSetCnt);
			OTFData.vMarkSetNames.clear();
			OTFData.vMarkSetNames.resize(OTFData.iMarkSetCnt);

			for (int i = 0; i < OTFData.iMarkSetCnt; ++i)
			{
				OTFData.vMarkSetNames[i].append(sFormatSpec);
				OTFData.vMarkSetNames[i].append(std::to_string(i));

				if (vOffsets[i] != 0)
				{
					ParseCoverageTable(&vGlyphs, pPosData, OTFData.uGDEFStart + iMarkSetOffset + vOffsets[i]);
					OTFData.vMarkSets[i] = GlyphsToNames(&vGlyphs, TRUE);
				}
				else
				{
					OTFData.vMarkSets.clear();
				}
			}
		}
	}

	// NOTE: Ligature positioning currently not supported
	if (iLigCaretOffset != 0)
	{
#if 0
		iLigCaretOffset += OTFData.uGDEFStart;
		const BYTE* pLigPos = pPosData + iLigCaretOffset;

		INT iCoverage = ParseUShortFromPos(&pLigPos);
		INT iCnt = ParseUShortFromPos(&pLigPos);

		if (iCnt == 0)
			return;

		std::vector<UINT16> vLigCaretOffsets;
		std::vector<UINT16> vGlyphs;
		vLigCaretOffsets.resize(iCnt);

		for (int i = 0; i < iCnt; ++i)
			vLigCaretOffsets[i] = ParseUShortFromPos(&pLigPos);

		ParseCoverageTable(&vGlyphs, pPosData, iLigCaretOffset + iCoverage);

		if (vGlyphs.empty())
			return;

		for (int i = 0; i < iCnt; ++i)
		{
			SplineChar* pSplChr = nullptr;
			PosSubTab* pPST = nullptr;

			if (vGlyphs[i] < OTFData.iGlyphCnt && (pSplChr = OTFData.vChars[vGlyphs[i]]) != nullptr)
			{
				const BYTE* pLigCaretPos = pPosData + iLigCaretOffset + vLigCaretOffsets[i];
				PosSubTab* pStartPST = BACKORNULL(pSplChr->vecPosSub);

				for (pPST = pStartPST; pPST != nullptr && pPST->uType != PST_LCARET; pPST = pPST->pNext);

				if (!pPST)
				{
					pPST = new PosSubTab();
					pPST->pNext = BACKORNULL(pSplChr->vecPosSub);
					pPST->uType = PST_LCARET;
					pPST->pSubtable = nullptr;
					pPST->Priority = (UINT16)i;
					pSplChr->vecPosSub.push_back(pPST);
					pSplChr->bLigCaretCntFixed = TRUE;
				}

				const BYTE* pCaretBasePos = pLigCaretPos;

				pPST->PSTLCaret.iCnt = ParseUShortFromPos(&pLigCaretPos);

				if (!pPST->PSTLCaret.vecCaretPos.empty())pPST->PSTLCaret.vecCaretPos.clear();

				std::vector<UINT16> vOffsets;
				vOffsets.resize(pPST->PSTLCaret.iCnt);

				for (int j = 0; j < pPST->PSTLCaret.iCnt; ++j)
					vOffsets[j] = ParseUShortFromPos(&pLigCaretPos);

				pPST->PSTLCaret.vecCaretPos.resize(pPST->PSTLCaret.iCnt);
				
				for (int j = 0; j < pPST->PSTLCaret.iCnt; ++j)
				{
					const BYTE* pLigCaretOffPos = pCaretBasePos + vOffsets[j];
					INT iFormat = ParseUShortFromPos(&pLigCaretOffPos);

					if (iFormat == 1)
					{
						pPST->PSTLCaret.vecCaretPos[j] = ParseUShortFromPos(&pLigCaretOffPos);
					}
					else if (iFormat == 2)
					{
						pPST->PSTLCaret.vecCaretPos[j] = 0;
						/*point = 0*/ParseUShortFromPos(&pLigCaretOffPos);
					}
					else if (iFormat == 3)
					{
						pPST->PSTLCaret.vecCaretPos[j] = ParseUShortFromPos(&pLigCaretOffPos);
						/*device table = 0*/ParseUShortFromPos(&pLigCaretOffPos);
					}
					else
					{
						TRACEUSER("Greg", "!!!! Unknown caret format %d !!!!\n", iFormat);
						return;
					}
				}
			}
		}
#endif
	}

	OTFData.iGDefBounds = 0;
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFProp(const BYTE* pPosData)

	Author:		Greg Watroba
	Created:	20/09/2016
	Inputs:		-
	Outputs		Set auxiliary data used in GSUB among other further data
	Returns:	-
	Purpose:	Reads font Prop data

********************************************************************************************/

void OTFDataLoader::ParseTTFProp(const BYTE* pPosData)
{
	/*TODO: probable caret positioning in some specific situations*/
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFLcar(const BYTE* pPosData)

	Author:		Greg Watroba
	Created:	20/09/2016
	Inputs:		-
	Outputs		Set auxiliary data used in GSUB among other further data
	Returns:	-
	Purpose:	Reads Lcar data

********************************************************************************************/

void OTFDataLoader::ParseTTFLcar(const BYTE* pPosData)
{
	/*TODO: probable caret positioning in some specific situations*/
}

/********************************************************************************************

>	void OTFDataLoader::SetKernData(INT* iKernCount, MillipointKerningPair* pKernArray, INT* iEmSize)

	Author:		Greg Watroba
	Created:	8/12/2015
	Inputs:		-
	Outputs		iKernCount - counter for kerning pairs
				pKernArray - array for storing kerning pairs
				iEmSize - font element size
	Returns:	-
	Purpose:	Converts kerning pairs parsed from GPOS data into Camelot's format

********************************************************************************************/

void OTFDataLoader::SetKernData(INT* iKernCount, MillipointKerningPair** pKernArray, INT* iEmSize)
{
	if (!DataParsingFinished())
	{
		(*iKernCount) = 0;
		return;
	}

	//creating pairs for new type of kerning pairs
	std::vector<MillipointKerningPair> vKernPairs;
	for (UINT i = 0; i < OTFData.vChars.size(); ++i)
	{
		if (!OTFData.vChars[i] || OTFData.vChars[i]->iUnicodeEnc == -1)
			continue;

		if (OTFData.vChars[i]->sName.empty())
			TRACEUSER("Greg", "No name found for character: I: %d UNICODE: %c GLYPH: %d\n", i, OTFData.vChars[i]->iUnicodeEnc, OTFData.vChars[i]->iOrigPos);

		for (auto &pPair : OTFData.vChars[i]->vecKernPairs)
		{
			MillipointKerningPair MKP;
			MKP.wFirst = static_cast<WORD>(OTFData.vChars[i]->iUnicodeEnc);
			MKP.wSecond = static_cast<WORD>(pPair.pSC->iUnicodeEnc);
			MKP.iKernAmount = pPair.iOff;
			vKernPairs.push_back(MKP);
		}
	}

	//character mapping from string values to unicode values
	std::unordered_map<std::string, UINT> mCharMap;
	mCharMap.reserve(OTFData.vChars.size());
	for (UINT i = 0; i < OTFData.vChars.size(); ++i)
	{
		if (!OTFData.vChars[i] || OTFData.vChars[i]->iUnicodeEnc <= 0 || OTFData.vChars[i]->sName.empty())
			continue;

		mCharMap[OTFData.vChars[i]->sName] = OTFData.vChars[i]->iUnicodeEnc;
	}

	//single character description shouldn't be bigger than this
	CHAR sTemp[128] = { 0 };

	//class based conversion to pairs
	for (UINT iKernClass = 0; iKernClass < OTFData.vKernClasses.size(); ++iKernClass)
	{
		if (!OTFData.vKernClasses[iKernClass])
			continue;

		for (UINT iFirst = 0; iFirst < OTFData.vKernClasses[iKernClass]->vecKernFirsts.size(); ++iFirst)
		{
			const std::string& sFirst = OTFData.vKernClasses[iKernClass]->vecKernFirsts[iFirst];

			//it turned out that char-by-char approach is quicker than
			//any other method of string tokenization

			//we have to store 'firsts' for later usage
			std::vector<std::string> sFirstElements;
			INT iLastChar = 0;
			for (auto &pChar : sFirst)
			{
				//sometimes end of string characters may occur
				if (pChar == '\0')
					continue;

				if (iLastChar >= 128)
				{
					ERROR3("Unexpectedly long character description - check parser for errors");
					iLastChar = 0;
				}

				//space is our delimiter
				if (pChar == ' ')
				{
					sTemp[iLastChar] = '\0';
					iLastChar = 0;
					sFirstElements.push_back(sTemp);
				}
				else
				{
					sTemp[iLastChar++] = pChar;
				}
			}

			for (UINT iSecond = 0; iSecond < OTFData.vKernClasses[iKernClass]->vecKernSeconds.size(); ++iSecond)
			{
				INT iKernIndex = static_cast<INT>(iFirst * OTFData.vKernClasses[iKernClass]->vecKernSeconds.size() + iSecond);
				MILLIPOINT iKernAmount = OTFData.vKernClasses[iKernClass]->vecKernOffsets[iKernIndex];

				//0 is default value for kerning in Camelot, so we don't have to parse characters in this case
				if (iKernAmount == 0)
					continue;

				//we don't have to store vector for 'seconds' so we only do parsing...
				const std::string& sSecond = OTFData.vKernClasses[iKernClass]->vecKernSeconds[iSecond];
				iLastChar = 0;

				for (auto &pChar : sSecond)
				{
					if (pChar == '\0')
						continue;

					if (iLastChar >= 128)
					{
						ERROR3("Unexpectedly long character description - check parser for errors");
						iLastChar = 0;
					}

					//and when we'll find complete character description
					//we directly add kerning pair
					if (pChar == ' ')
					{
						sTemp[iLastChar] = '\0';
						iLastChar = 0;

						const auto pSecondChar = mCharMap.find(sTemp);

						if (pSecondChar == mCharMap.end())
							continue;

						//since the only change is the value from stored
						//vector of strings, we can instantiate kerning pair class earlier

						MillipointKerningPair MKP;
						MKP.wSecond = static_cast<WORD>((*pSecondChar).second);
						MKP.iKernAmount = iKernAmount;

						for (auto &pFirst : sFirstElements)
						{
							const auto pFirstChar = mCharMap.find(pFirst);

							if (pFirstChar == mCharMap.end())
								continue;
							
							MKP.wFirst = static_cast<WORD>((*pFirstChar).second);
							
							vKernPairs.push_back(MKP);
						}
					}
					else
					{
						sTemp[iLastChar++] = pChar;
					}
				}
			}
		}
	}

	//if we don't have any new type kerning - we skip to the old way outside this method
	if (vKernPairs.empty())
	{
		(*iKernCount) = 0;
		return;
	}

	(*iEmSize) = OTFData.iEmSize;
	(*iKernCount) = static_cast<INT>(vKernPairs.size());
	(*pKernArray) = new MillipointKerningPair[(*iKernCount)];
	memcpy((*pKernArray), &vKernPairs.front(), sizeof(MillipointKerningPair) * (*iKernCount));
}

/********************************************************************************************

>	void OTFDataLoader::SetSubstitutionData(CachedFontItem* pFontEntry)

	Author:		Greg Watroba
	Created:	22/09/2016
	Inputs:		pFontEntry - cached font entry class we will map data into
	Outputs		-
	Returns:	-
	Purpose:	Converts ligature classes into table for easier access in Camelot's renderer

********************************************************************************************/

void OTFDataLoader::SetSubstitutionData(VFontBase* pBase)
{
#ifndef BUILD_TEXT_LIGATURES
	return;
#else	
	ERROR2VOIDIF(!pBase, "OTFDataLoader::SetSubstitutionData : FontBase cannot be NULL!");

	if (!DataParsingFinished())
		return;

	// list of all parsed ligatures
	std::list<LigatureDataSet> mLigatureList;

	// array of stylistic set lists
	std::array<std::list<StylisticSetEntry>, 20> mStylisticSets;

	// unicode to glyph
	std::unordered_map<WCHAR, DWORD> mUnicodeToGlyph;

	for (UINT i = 0; i < OTFData.vChars.size(); ++i)
	{
		if (!OTFData.vChars[i] || OTFData.vChars[i]->iUnicodeEnc <= 0)
			continue;

		mUnicodeToGlyph[(WCHAR)OTFData.vChars[i]->iUnicodeEnc] = (DWORD)OTFData.vChars[i]->iOrigPos;
	}

	size_t MaxLigatureSize = 1;

	for (UINT i = 0; i < OTFData.vChars.size(); ++i)
	{
		if (!OTFData.vChars[i] || OTFData.vChars[i]->iUnicodeEnc <= 0)
			continue;

		for (auto& posSub : OTFData.vChars[i]->vecPosSub)
		{
			std::wstring vecSubComponents;

			if (posSub->uType == PST_SUBSTITUTION || posSub->uType == PST_ALTERNATE || posSub->uType == PST_MULTIPLE)
			{
				if (MaxLigatureSize < posSub->PSTSubs.vecGlyphs.size())
					MaxLigatureSize = (INT)posSub->PSTSubs.vecGlyphs.size();

				for (auto& pGlyph : posSub->PSTSubs.vecGlyphs)
					vecSubComponents.push_back((WCHAR)OTFData.vChars[pGlyph]->iUnicodeEnc);
			}
			else if (posSub->uType == PST_LIGATURE)
			{
				if (MaxLigatureSize < posSub->PSTLig.vecGlyphs.size())
					MaxLigatureSize = (INT)posSub->PSTLig.vecGlyphs.size();

				for (auto& pGlyph : posSub->PSTLig.vecGlyphs)
					vecSubComponents.push_back((WCHAR)OTFData.vChars[pGlyph]->iUnicodeEnc);
			}
			
			if (!vecSubComponents.empty())
			{
				GlyphSetEnum eGlyphType = (posSub->pSubtable) ? posSub->pSubtable->eGlyphSet : GSET_NONE;
				LigatureSubsetFlag eLigSubFlag = LIG_SUB_NONE;
				SHORT sStylisticSet = -1;

				switch (eGlyphType)
				{
					case GSET_STANDARD_LIGATURES:
					{
						eLigSubFlag = LIG_SUB_STANDARD;
					} break;

					case GSET_ANCIENT_LIGATURES:
					case GSET_HISTORIC_LIGATURES:
					{
						eLigSubFlag = LIG_SUB_HISTORIC;
					} break;

					case GSET_DISCRETIONARY_LIGATURES:
					{
						eLigSubFlag = LIG_SUB_DISCRETIONARY;
					} break;

					case GSET_CONTEXTUAL_LIGATURES:
					{
						eLigSubFlag = LIG_SUB_CONTEXTUAL;
					} break;

					case GSET_STYLE_SET_1:
					case GSET_STYLE_SET_2:
					case GSET_STYLE_SET_3:
					case GSET_STYLE_SET_4:
					case GSET_STYLE_SET_5:
					case GSET_STYLE_SET_6:
					case GSET_STYLE_SET_7:
					case GSET_STYLE_SET_8:
					case GSET_STYLE_SET_9:
					case GSET_STYLE_SET_10:
					case GSET_STYLE_SET_11:
					case GSET_STYLE_SET_12:
					case GSET_STYLE_SET_13:
					case GSET_STYLE_SET_14:
					case GSET_STYLE_SET_15:
					case GSET_STYLE_SET_16:
					case GSET_STYLE_SET_17:
					case GSET_STYLE_SET_18:
					case GSET_STYLE_SET_19:
					case GSET_STYLE_SET_20:
					{
						// we need index value [0,19]
						sStylisticSet = (SHORT)(eGlyphType - GSET_STYLE_SET_1);
						break;
					}
				}

				if (eLigSubFlag != LIG_SUB_NONE)
				{
					LigatureDataSet LigEntry;
					LigEntry.LigType = eLigSubFlag;
					LigEntry.vecLigComponents = vecSubComponents;
					LigEntry.LigUnicode = (WCHAR)OTFData.vChars[i]->iUnicodeEnc;
					LigEntry.Priority = posSub->Priority;
					LigEntry.GlyphID = mUnicodeToGlyph[LigEntry.LigUnicode];

					if (!mLigatureList.empty() && mLigatureList.back() == LigEntry)
						continue;

					mLigatureList.push_back(LigEntry);
				}
				else if (sStylisticSet >= 0)
				{
					StylisticSetEntry SetEntry;
					SetEntry.Key = (WCHAR)OTFData.vChars[i]->iUnicodeEnc;
					SetEntry.StylisticChar = vecSubComponents[0];
					SetEntry.SubGlyphID = mUnicodeToGlyph[SetEntry.StylisticChar];
					mStylisticSets[sStylisticSet].push_back(SetEntry);
				}
			}
		}
	}

	// we set unicode to glyph map
	pBase->SetUnicodeMap(mUnicodeToGlyph);

	// we set max ligature size
	pBase->SetMaxLigatureSize((INT)MaxLigatureSize);

	// we prepare and setup ligature data
	SetupLigatureData(&mLigatureList, pBase);

	// we prepare and setup stylistic set data
	SetupStylisticSetData(&mStylisticSets, pBase);

	TypoMetrics ParsedMetrics;
	if (OTFData.bUseTypoMetrics)
	{
		ParsedMetrics.bUseTypoMetrics = OTFData.bUseTypoMetrics;
		ParsedMetrics.TypoAscent = OTFData.PfmInfo.iOS2TypoAscent;
		ParsedMetrics.TypoDescent = OTFData.PfmInfo.iOS2Typodescent;
		ParsedMetrics.TypoLineGap = OTFData.PfmInfo.iOS2Typolinegap;
		pBase->SetTypoMetricsData(ParsedMetrics);
	}

#ifdef BUILD_TEXT_ADVANCED_OPENTYPE
	// we prepare and setup substitution rules
	SetupSubstitutionRules(pBase);

	// we prepare and setup substitution tables
	SetupSubstitutionTables(pBase);

	if (!OTFData.mapLookupToGSUBFeature.empty())
	{
		auto upMapPriorityToFeature = std::make_unique<std::unordered_map<UINT16, GlyphSetEnum>>();
		upMapPriorityToFeature->insert(OTFData.mapLookupToGSUBFeature.begin(), OTFData.mapLookupToGSUBFeature.end());
		pBase->SetPriorityToFeatureMap(std::move(upMapPriorityToFeature));
	}
#endif //BUILD_TEXT_ADVANCED_OPENTYPE

#endif // !BUILD_TEXT_LIGATURES
}

/********************************************************************************************

>	void OTFDataLoader::SetupLigatureData(std::list<LigatureDataSet>* pLigatureList, 
										  FontBase* pFontBase)

	Author:		Greg Watroba
	Created:	16/02/2017
	Inputs:		pLigatureList - parsed ligature list
				pFontBase - pointer to font instance for ligature setup
	Outputs		-
	Returns:	-
	Purpose:	Prepares ligature data for specific font instance

********************************************************************************************/

void OTFDataLoader::SetupLigatureData(std::list<LigatureDataSet>* pLigatureList, 
									  VFontBase* pFontBase)
{
	ERROR2VOIDIF(!pFontBase, "OTFDataLoader::SetupLigatureData : font base cannot be NULL!");

	// sort table by key and priority
	// all keys have the same size, so 'f' 'i' ' ' will be smaller than 'f' 'j' ' ' or 'f' 'j' 'a'
	pLigatureList->sort([&](const LigatureDataSet& A, const LigatureDataSet& B)
	{
		size_t maxSize = (A.vecLigComponents.size() > B.vecLigComponents.size()) ? A.vecLigComponents.size() : B.vecLigComponents.size();

		for (size_t i = 0; i < maxSize; ++i)
		{
			if (i >= A.vecLigComponents.size() || i >= B.vecLigComponents.size())
				return (A.vecLigComponents.size() < B.vecLigComponents.size());

			if (A.vecLigComponents[i] != B.vecLigComponents[i])
				return (A.vecLigComponents[i] < B.vecLigComponents[i]);
		}

		return (A.Priority < B.Priority);
	});

	UINT LigListSize = (UINT)pLigatureList->size();
	TRACEUSER("Greg", "Found %d ligatures\n", LigListSize);

	// rewrite map to table
	auto pLigEntry = std::make_unique<LigatureEntry[]>(LigListSize);
	
	INT i = 0;
	for (auto& pLig : (*pLigatureList))
	{
		pLigEntry[i].Key = pLig.vecLigComponents;
		pLigEntry[i].KeyLth = (BYTE)pLig.vecLigComponents.size();
		pLigEntry[i].Ligature = pLig.LigUnicode;
		pLigEntry[i].LigType = pLig.LigType;
		pLigEntry[i].Priority = pLig.Priority;
		pLigEntry[i].GlyphID = pLig.GlyphID;

		++i;
	}

	// we set ligature data
	pFontBase->SetLigatureData(std::move(pLigEntry), LigListSize);
}

/********************************************************************************************

>	void OTFDataLoader::SetupStylisticSetData(std::array<std::list<StylisticDataSet>, 20>* pStylisticSets,
											  FontBase* pFontBase)

	Author:		Greg Watroba
	Created:	22/02/2017
	Inputs:		pStylisticSets - parsed stylistic set data
				pFontBase - pointer to font instance for ligature setup
	Outputs		-
	Returns:	-
	Purpose:	Prepares stylistic set data for specific font instance

********************************************************************************************/

void OTFDataLoader::SetupStylisticSetData(std::array<std::list<StylisticSetEntry>, 20>* pStylisticSets,
										  VFontBase* pFontBase)
{
	ERROR2VOIDIF(!pFontBase, "OTFDataLoader::SetupLigatureData : font base cannot be NULL!");

	auto arrStyleSetEntries = std::make_unique<StylisticSetDataSet>();
	auto arrStyleSetSizes = std::make_unique<StylisticSetSizeSet>();
	arrStyleSetSizes.get()->fill(0);

	auto arrStylisticSetNames = std::make_unique<StylisticSetNameSet>();

	for (INT iStyleSet = 0; iStyleSet < STYLISTIC_SETS_MAX_SIZE; ++iStyleSet)
	{
		// sort similarily to ligatures
		(*pStylisticSets)[iStyleSet].sort([&](const StylisticSetEntry& A, const StylisticSetEntry& B)
		{
			return (A.Key < B.Key);
		});

		// get size of the current size
		const ULONG StylisticSetSize = (ULONG)(*pStylisticSets)[iStyleSet].size();
		//TRACEUSER("Greg", "Stylistic set: %d has %d substitutions\n", iStyleSet, StylisticSetSize);
		(*arrStyleSetSizes.get())[iStyleSet] = StylisticSetSize;

		GlyphSetEnum curSSetEnum = (GlyphSetEnum)(GSET_STYLE_SET_1 + iStyleSet);
		auto res = OTFData.mapUserFriendlyGlyphSetNames.find(curSSetEnum);

		if (res != OTFData.mapUserFriendlyGlyphSetNames.end())
			(*arrStylisticSetNames.get())[iStyleSet] = res->second;

		// rewrite map to table
		if (StylisticSetSize == 0)
			continue;

		for (auto& pStyleSetEntry : (*pStylisticSets)[iStyleSet])
			(*arrStyleSetEntries.get())[iStyleSet][pStyleSetEntry.Key] = pStyleSetEntry.StylisticChar;
	}

	// set stylistic set data
	pFontBase->SetStylisticSetData(std::move(arrStyleSetEntries), std::move(arrStyleSetSizes));

	// set stylistic set names
	pFontBase->SetStylisticSetNames(std::move(arrStylisticSetNames));
}

/********************************************************************************************

>	void OTFDataLoader::SetupSubstitutionRules(VFontBase* pFontBase)

	Author:		Greg Watroba
	Created:	09/03/2017
	Inputs:		pFontBase - pointer to font instance for ligature setup
	Outputs		-
	Returns:	-
	Purpose:	Prepares substitution rules data for specific font instance

********************************************************************************************/

void OTFDataLoader::SetupSubstitutionRules(VFontBase* pFontBase)
{
#ifdef BUILD_TEXT_ADVANCED_OPENTYPE
	ERROR2VOIDIF(!pFontBase, "OTFDataLoader::SetupLigatureData : font base cannot be NULL!");
	pFontBase->SetSubstitutionRuleSet(std::make_unique<SubstitutionRuleSet>(OTFData.SubRuleSet));

	TRACEUSER("Greg", "!@# Max sizes: %d %d %d\n", OTFData.TotalMaxSizes.Pre, OTFData.TotalMaxSizes.Key, OTFData.TotalMaxSizes.Post);
	pFontBase->SetTotalMaxSizes(OTFData.TotalMaxSizes);
#endif
}

/********************************************************************************************

>	void OTFDataLoader::SetupSubstitutionTables(VFontBase* pFontBase)

	Author:		Greg Watroba
	Created:	10/03/2017
	Inputs:		pFontBase - pointer to font instance for ligature setup
	Outputs		-
	Returns:	-
	Purpose:	Prepares substitution tables data for specific font instance

********************************************************************************************/

void OTFDataLoader::SetupSubstitutionTables(VFontBase* pFontBase)
{
#ifdef BUILD_TEXT_ADVANCED_OPENTYPE
	ERROR2VOIDIF(!pFontBase, "OTFDataLoader::SetupLigatureData : font base cannot be NULL!");

	if (OTFData.vecNeededLookups.empty())
		return;

	std::sort(OTFData.vecNeededLookups.begin(), OTFData.vecNeededLookups.end());
	auto last = std::unique(OTFData.vecNeededLookups.begin(), OTFData.vecNeededLookups.end());
	OTFData.vecNeededLookups.erase(last, OTFData.vecNeededLookups.end());

	auto pSubstitutionTableMap = std::make_unique<SubstitutionTableSet>();
	pSubstitutionTableMap->reserve(OTFData.vecNeededLookups.size());
	pSubstitutionTableMap->max_load_factor(0.2f);

	for (auto& pSubTableID : OTFData.vecNeededLookups)
	{
		if (pSubTableID >= OTFData.vecLookupSubTables.size())
			continue;

		auto pLookupSubTable = OTFData.vecLookupSubTables[pSubTableID];

		if (!pLookupSubTable || !pLookupSubTable->pOTLookup)
			continue;

		SubstitutionTable Table;
		Table.UID = pLookupSubTable->UniqueID;

		auto setRes = OTFData.setStandardSubLookups.find(pLookupSubTable->UniqueID);
		const BOOL isStandard = setRes != OTFData.setStandardSubLookups.end();

		switch (pLookupSubTable->pOTLookup->eLookupType)
		{
			case OTL_TYPE_GSUB_SINGLE: { Table.SubstitutionType = isStandard ? GSUB_SINGLE_STANDARD : GSUB_SINGLE; }break;
			case OTL_TYPE_GSUB_MULTIPLE: { Table.SubstitutionType = isStandard ? GSUB_MULTIPLE_STANDARD : GSUB_MULTIPLE; }break;
			case OTL_TYPE_GSUB_LIGATURE: { Table.SubstitutionType = isStandard ? GSUB_LIGATURE_STANDARD : GSUB_LIGATURE; }break;
		}

		Table.SubstitutionMap.reserve(pLookupSubTable->vecPosSubTab.size());
		Table.SubstitutionMap.max_load_factor(0.2f);
		for (auto& pSubstitution : pLookupSubTable->vecPosSubTab)
		{
			std::wstring vecKey;
			std::vector<std::pair<WCHAR, DWORD>> vecValues;

			switch (pSubstitution->uType)
			{
				case PST_SUBSTITUTION: 
				case PST_MULTIPLE: 
				case PST_ALTERNATE:
				{
					for (auto& pGlyph : pSubstitution->PSTSubs.vecGlyphs)
						vecValues.push_back(std::make_pair((WCHAR)OTFData.vChars[pGlyph]->iUnicodeEnc, (DWORD)pGlyph));

					vecKey.push_back((WCHAR)OTFData.vChars[pSubstitution->GlyphID]->iUnicodeEnc);
				}break;

				case PST_LIGATURE: 
				{ 
					for (auto& pGlyph : pSubstitution->PSTLig.vecGlyphs)
						vecKey.push_back((WCHAR)OTFData.vChars[pGlyph]->iUnicodeEnc);

					vecValues.push_back(std::make_pair((WCHAR)OTFData.vChars[pSubstitution->GlyphID]->iUnicodeEnc, (DWORD)pSubstitution->GlyphID));
				}break;
			}
		
			Table.SubstitutionMap[vecKey] = vecValues;
		}

		pSubstitutionTableMap.get()->insert(std::make_pair(Table.UID, Table));
	}

	pFontBase->SetSubstitutionTable(std::move(pSubstitutionTableMap));
#endif //BUILD_TEXT_ADVANCED_OPENTYPE
}



/********************************************************************************************

>	void OTFDataLoader::ParseHeadData(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	3/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs		Fills apropriate data in OTFData structure
	Returns:	-
	Purpose:	Parses header font data for initial information stored with HEAD tag

********************************************************************************************/

void OTFDataLoader::ParseHeadData(const BYTE* pFontData)
{
	/* Here I want units per em, and size of loca entries */
	/* oh... also creation/modification times */

	const BYTE* pHeadDataPos = pFontData + OTFData.iHeadStart;

	//skip version number
	MOVE32BITS(pHeadDataPos);

	OTFData.uFontRevision = ParseLongFromPos(&pHeadDataPos);

	//skip checkSumAdjustment
	MOVE32BITS(pHeadDataPos);

	//skip magicNumber
	MOVE32BITS(pHeadDataPos);

	INT32 iFlags = ParseUShortFromPos(&pHeadDataPos);
	OTFData.uOptimisedForClearType = (iFlags & (1 << 13) ) ? 1 : 0;
	OTFData.uApplyLSB = !(iFlags & (1 << 1));
	OTFData.iEmSize = ParseUShortFromPos(&pHeadDataPos);

	OTFData.iAscent = (INT)(OTFData.iEmSize * 0.8);
	OTFData.iDescent = OTFData.iEmSize - OTFData.iAscent;

	//skip created & modified datetimes
	for (INT i = 0; i < 8; ++i)
		MOVE16BITS(pHeadDataPos);

	for (INT i = 0; i < 4; ++i)
		OTFData.iFontBoundingBoxes[i] = (SHORT)ParseUShortFromPos(&pHeadDataPos);

	OTFData.iMacStyle = ParseUShortFromPos(&pHeadDataPos);

	//skip lowestRecPPEM & fontDirectionHint
	for (INT i = 0; i < 2; ++i)
		MOVE16BITS(pHeadDataPos);

	OTFData.uIndexToLocIsLong = ParseUShortFromPos(&pHeadDataPos);

	if (OTFData.uIndexToLocIsLong)
		OTFData.iGlyphCnt = OTFData.iGlyphLocationLength / 4 - 1;

	if (OTFData.iGlyphCnt < 0)
		OTFData.iGlyphCnt = 0;
}

/********************************************************************************************

>	void OTFDataLoader::ParseHHeaData(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	3/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs		Fills apropriate data in OTFData structure
	Returns:	-
	Purpose:	Parses header font data for initial information stored with HHEA tag

********************************************************************************************/

void OTFDataLoader::ParseHHeaData(const BYTE* pFontData)
{
	//Here I want ascent, descent and the number of horizontal metrics
	const BYTE* pHHeaDataPos = pFontData + OTFData.iHHeaStart;

	//skip version number
	MOVE32BITS(pHHeaDataPos);

	OTFData.PfmInfo.iHHeadAscent = ParseUShortFromPos(&pHHeaDataPos);
	OTFData.PfmInfo.iHHeadDescent = ParseUShortFromPos(&pHHeaDataPos);

	OTFData.PfmInfo.uHHeadDescentAdd = 0;
	OTFData.PfmInfo.uHHeadAscentAdd = 0;

	OTFData.PfmInfo.iLineGap = ParseUShortFromPos(&pHHeaDataPos);
	OTFData.iAdvanceWidthMax = ParseUShortFromPos(&pHHeaDataPos);

	OTFData.PfmInfo.uHHeadSet = TRUE;

	//skip caret slopes & other fillers we don't need
	for (INT32 i = 0; i < 11; ++i)
		MOVE16BITS(pHHeaDataPos);

	OTFData.iWidthCnt = ParseUShortFromPos(&pHHeaDataPos);
}

/********************************************************************************************

>	void OTFDataLoader::ParseMaxpData(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	3/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs		Fills apropriate data in OTFData structure
	Returns:	-
	Purpose:	Parses header font data for initial information stored with MAXP tag

********************************************************************************************/

void OTFDataLoader::ParseMaxpData(const BYTE* pFontData)
{
	//All I want here is the number of glyphs
	const BYTE* pMaxpDataPos = pFontData + OTFData.iMaxPStart;

	//skip version number
	MOVE32BITS(pMaxpDataPos);

	INT iCnt = ParseUShortFromPos(&pMaxpDataPos);

	if (OTFData.iGlyphCnt == 0 && OTFData.iGlyphLength == 0 &&
		OTFData.iGlyphLocationLength <= 4 && OTFData.uCFFLength == 0 &&
		OTFData.uCFFLength == 0 && OTFData.uTyp1Start == 0)
	{
		OTFData.bOnlyStrikes = TRUE;
	}
	else if (iCnt != OTFData.iGlyphCnt && OTFData.iGlyphLocationLength != 0)
	{
		OTFData.bBadGlyphData = TRUE;
		if (iCnt > OTFData.iGlyphCnt)
			iCnt = OTFData.iGlyphCnt;
	}

	/* Open Type fonts have no loca table, so we can't calculate the glyph */
	/*  count from it */
	OTFData.iGlyphCnt = iCnt;
	if (iCnt < 0)
		OTFData.iGlyphCnt = 0;
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFGlyphs(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs		Fills apropriate data in OTFData structure
	Returns:	-
	Purpose:	Parses glyph data used for kerning

********************************************************************************************/

void OTFDataLoader::ParseTTFGlyphs(const BYTE* pFontData)
{
	std::vector<UINT32> vGlyphOffsets(OTFData.iGlyphCnt + 1);

	/* First we read all the locations. This might not be needed, they may */
	/*  just follow one another, but nothing I've noticed says that so let's */
	/*  be careful */
	const BYTE* pGlyphDataPos = pFontData + OTFData.iGlyphLocationStart;

	if (OTFData.uIndexToLocIsLong)
	{
		for (INT32 i = 0; i <= OTFData.iGlyphCnt; ++i)
			vGlyphOffsets[i] = ParseLongFromPos(&pGlyphDataPos);
	}
	else
	{
		for (INT32 i = 0; i <= OTFData.iGlyphCnt; ++i)
			vGlyphOffsets[i] = 2 * ParseUShortFromPos(&pGlyphDataPos);
	}

	OTFData.vChars.resize(OTFData.iGlyphCnt);

	//If it's not an TrueTypeCollection or all glyphs stored in one place
	if (!OTFData.bIsTrueTypeCollection || (OTFData.OTFlags & OTF_ALL_GLYPHS_IN_TTC))
	{
		for (INT32 i = 0; i < OTFData.iGlyphCnt; ++i)
		{
			OTFData.vChars[i] = ParseTTFGlyph(pFontData, vGlyphOffsets[i], vGlyphOffsets[i + 1], i);
			if (OTFData.vChars[i])
				OTFData.vChars[i]->iOrigPos = i;
		}
	}
	else
	{
		/* only read the glyphs we actually use in this font */
		/* this is complicated by references (and substitutions), */
		/* we can't just rely on the encoding to tell us what is used */
		OTFData.vChars.resize(OTFData.iGlyphCnt);
		OTFData.vInUse.resize(OTFData.iGlyphCnt);

		ParseTTFEncodings(pFontData, GIUT_JUST_IN_USE);

		/* Some glyphs may appear in substitutions and not in the encoding... */
		if (OTFData.iGsubStart != 0)
			ParseTTFGsubUsed(pFontData);

		if (OTFData.iMathStart != 0)
			ParseOTFMathUsed(pFontData);

		if (OTFData.iMorxStart != 0 || OTFData.iMortStart != 0)
			ParseTTFMortUsed(pFontData);

		BOOL bAnyParse = TRUE;

		while (bAnyParse)
		{
			bAnyParse = FALSE;
			for (INT32 i = 0; i < OTFData.iGlyphCnt; ++i)
			{
				if (OTFData.vInUse[i] && OTFData.vChars[i] == nullptr)
				{
					OTFData.vChars[i] = ParseTTFGlyph(pFontData, vGlyphOffsets[i], 
														vGlyphOffsets[i + 1], i);
					if (OTFData.vChars[i])
						OTFData.vChars[i]->iOrigPos = i;
					bAnyParse = OTFData.vChars[i] != nullptr;
				}
			}
		}

		OTFData.vInUse.clear();
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFNames(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	13/01/2016
	Inputs:		pFontData - raw BYTE font data table
	Outputs		Sets sName field in vChars character vector in OTFData
	Returns:	-
	Purpose:	Sets character names needed for glyph mapping in class-based kerning

********************************************************************************************/

void OTFDataLoader::ParseTTFNames(const BYTE* pFontData)
{
	std::string sName;
	CHAR Buffer[30];
	std::vector<UINT16> vIndexes;
	INT bNotDefWarned = false;
	INT bAnyNames = false;

	if (OTFData.uPostScriptStart != 0) 
	{
		const BYTE* pPostScriptPos = pFontData + OTFData.uPostScriptStart;
		INT iFormat = ParseLongFromPos(&pPostScriptPos);
		//italic angle field - not needed
		MOVE32BITS(pPostScriptPos);
		//upos - not needed
		MOVE16BITS(pPostScriptPos);
		//uwidth - not needed
		MOVE16BITS(pPostScriptPos);
		//'post' defn of this field is different from FontInfo defn and I didn't notice
		MOVE32BITS(pPostScriptPos);
		//skip some more data
		MOVE32BITS(pPostScriptPos);
		MOVE32BITS(pPostScriptPos);
		MOVE32BITS(pPostScriptPos);
		MOVE32BITS(pPostScriptPos);
		
		if (iFormat == 0x00020000)
		{
			INT iGCnt = ParseUShortFromPos(&pPostScriptPos);
			vIndexes.resize(UINT16_MAX);

			//the index table is backwards from the way I want to use it
			INT iGCntBig = 0;
			
			for (INT i = 0; i < iGCnt; ++i)
			{
				INT iVal = ParseUShortFromPos(&pPostScriptPos);
				vIndexes[iVal] = static_cast<UINT16>(i);

				if (iVal >= 258)
					++iGCntBig;
			}

			//if we are only loading bitmaps, we can get holes in our data
			for (INT i = 0; i < 258; ++i)
			{
				if (vIndexes[i] != 0 || i == 0)
				{
					if (vIndexes[i] < OTFData.iGlyphCnt && vIndexes[i] < OTFData.vChars.size() &&
						OTFData.vChars[vIndexes[i]] != NULL)
					{
						OTFData.vChars[vIndexes[i]]->sName = TTFNames[i];
					}
				}
			}

			iGCntBig += 258;

			for (INT i = 258; i < iGCntBig; ++i)
			{
				std::string sTemp;
				INT len = ParseCharFromPos(&pPostScriptPos);
				if (len < 0)
					break;

				sTemp.resize(len);
				for (INT j = 0; j < len; ++j)
					sTemp[j] = ParseCharFromPos(&pPostScriptPos);

				if (vIndexes[i] < OTFData.iGlyphCnt && vIndexes[i] < OTFData.vChars.size() &&
					OTFData.vChars[vIndexes[i]] != NULL && OTFData.vChars[vIndexes[i]]->sName.empty())
				{
					OTFData.vChars[vIndexes[i]]->sName = sTemp;
				}
			}

			vIndexes.clear();
			bAnyNames = TRUE;
		}
	}

	/*
		Here should be Type42 Adobe character nameset parsing.
		Might be necessary when parsing unicode values 
		from glyphs won't be enough
	*/

	//If couldn't find standard names, then look for them in Adobe tables
	for (INT i = 0; i < OTFData.iGlyphCnt; ++i)
	{
		if (OTFData.vChars[i] != NULL)
		{
			if (i != 0 && !OTFData.vChars[i]->sName.empty() && OTFData.vChars[i]->sName == ".notdef")
			{
				/*
					For some reason MS puts out fonts where several characters are called .notdef (and only one is a real notdef).
					So if we find a glyph other than 0 called ".notdef" then pretend it had no name
				*/
				if (!bNotDefWarned)
				{
					bNotDefWarned = true;
					TRACEUSER("Greg", "Glyph %d is called \".notdef\", a singularly inept choice of name (only glyph 0\n may be called .notdef)\nIt should be renamed.\n", i);
				}
				OTFData.vChars[i]->sName.clear();
			}

			if (!OTFData.vChars[i]->sName.empty() && OTFData.vChars[i]->sName.find('\0') == std::string::npos)
				continue;

			//If it's a null string get rid of it
			OTFData.vChars[i]->sName.clear();

			if (i == 0)
				sName = ".notdef";
			else if (OTFData.vChars[i]->iUnicodeEnc == -1)
				sName.clear();
			else
			{
				sName = ReadGlyphName(Buffer, OTFData.vChars[i]->iUnicodeEnc, OTFData.eUniInterp, NULL);
				if (bAnyNames)
				{
					for (INT j = 0; j < OTFData.iGlyphCnt; ++j)
					{
						if (OTFData.vChars[j] != NULL && j != i && !OTFData.vChars[j]->sName.empty())
						{
							if (OTFData.vChars[j]->sName == sName)
							{
								sName.clear();
								break;
							}
						}
					}
				}
			}

			if (OTFData.vChars[i]->sName.empty())
				OTFData.vChars[i]->sName = sName;
		}
	}

	//supplement for char string names from other tables
	//at this point - not needed
#if 0
	for (INT i = OTFData.iGlyphCnt - 1; i >= 0; --i)
	{
		if (OTFData.vChars[i] != NULL && OTFData.vChars[i]->sName.empty())
			break;

		//currently VariationSequenceSubtable is not needed so we won't be parsing it
		if (i >= 0 && OTFData.vs_start != 0)
			ApplyVariationSequenceSubtable(ttf, OTFData.vs_start, info, git_findnames);

		//necessary when ligatures processing will be on
		if (i >= 0 && OTFData.iGsubStart != 0)
			GuessNamesFromGSUB(ttf, info);
		
		if (i >= 0 && OTFData.iMathStart != 0)
			GuessNamesFromMATH(ttf, info);
	}
#endif

	for (INT i = 0; i < OTFData.iGlyphCnt; ++i) 
	{
		if (OTFData.vChars[i] == NULL)
			continue;
		if (!OTFData.vChars[i]->sName.empty())
			continue;
		if (!OTFData.sOrdering.empty())
			sprintf_s(Buffer, 30 * sizeof(CHAR),  "%.20s-%d", OTFData.sOrdering, i);
		//if standard encoding method won't suffice we will have to create EncodingMap
		//else if (OTFData.map != NULL && OTFData.map->backmap[i] != -1)
		//	sprintf_s(Buffer, 30 * sizeof(CHAR), "nounicode.%d.%d.%x", OTFData.platform, OTFData.specific, (int)OTFData.map->backmap[i]);
		else
			sprintf_s(Buffer, 30 * sizeof(CHAR), "glyph%d", i);

		OTFData.vChars[i]->sName = Buffer;
	}
}

std::string OTFDataLoader::ReadGlyphName(CHAR* pBuffer, INT iUnicode, UnicodeIterp eUniInterp, NameList* pThisFont)
{
	std::string sResult;
	int up, ub, uc;

	if (pThisFont == nullptr)
		pThisFont = ALGNameList;
	else if (pThisFont == (NameList*)-1)
		pThisFont = &AGL;

	if ((iUnicode >= 0 && iUnicode < ' ') || (iUnicode >= 0x7f && iUnicode < 0xa0));
		// standard controls
	else if (iUnicode != -1) 
	{
		if (iUnicode >= 0xe000 && iUnicode <= 0xf8ff && (eUniInterp == UI_TRAD_CHINESE || pThisFont == &AGLAms)) 
		{
			const int *pua = eUniInterp == UI_TRAD_CHINESE ? CNS14PUA : AMSPUA;
			if (pua[iUnicode - 0xe000] != 0)
				iUnicode = pua[iUnicode - 0xe000];
		}

		up = iUnicode >> 16;
		ub = (iUnicode & 0xff00) >> 8;
		uc = (iUnicode & 0xff);

		if (up < 17)
		{
			for (NameList* pNameList = pThisFont; pNameList != nullptr; pNameList = pNameList->pBasedOn)
			{
				const CHAR* pTempStr = nullptr;
				if (pNameList->pUnicode[up] != nullptr && pNameList->pUnicode[up][ub] != nullptr && (pTempStr = pNameList->pUnicode[up][ub][uc]) != nullptr)
				{
					sResult = pTempStr;
					break;
				}
			}
		}
	}

	if (sResult.empty()) 
	{
		if (iUnicode >= 0x10000)
			sprintf_s(pBuffer, 30 * sizeof(CHAR), "u%04X", iUnicode);
		else
			sprintf_s(pBuffer, 30 * sizeof(CHAR), "uni%04X", iUnicode);

		sResult = pBuffer;
	}

	return sResult;
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFGlyph(const BYTE* pFontData, UINT32 iStart, UINT32 iEnd,
										 UINT32 iGid)

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
				iStart - start position
				iEnd - end position
				iGid - glyph id
	Returns:	SplineChar* which has been read
	Purpose:	Parse specific glyph stored of TTF type

********************************************************************************************/

SplineChar* OTFDataLoader::ParseTTFGlyph(const BYTE* pFontData, UINT32 iStart, UINT32 iEnd, 
											UINT32 iGid)
{
	SplineChar* pSC = new SplineChar();

	pSC->iUnicodeEnc = -1;
	pSC->iVWidth = (INT16)OTFData.iEmSize;
	pSC->iOrigPos = iGid;

	if (iEnd > (UINT32)OTFData.iGlyphLength)
	{
		if (!OTFData.bAlreadyFoundBadGlyphData)
			TRACEUSER("Greg", "Bad glyph (%d), its definition extends beyond the end of the glyf table\n", iGid);

		OTFData.bBadGlyphData = TRUE;
		OTFData.bAlreadyFoundBadGlyphData = TRUE;
		SAFE_DELETE(pSC);
		return nullptr;
	}
	else if (iEnd < iStart)
	{
		TRACEUSER("Greg", "Bad glyph (%d), its data length is negative\n", iGid);
		SAFE_DELETE(pSC);
		return nullptr;
	}

	if (iStart == iEnd)
	{
		//This isn't mentioned, but we seem to get some glyphs with no size,
		//not even a path cnt. They appear to be empty glyphs 
		return pSC;
	}

	// Note: loading spline data is not neccessary
#if 0

	const BYTE* pGlyphPos = pFontData + OTFData.iGlyphStart + iStart;
	INT iPathCnt = (SHORT) ParseUShortFromPos(&pGlyphPos);
	

	INT iGlyphBoundingBox[4];
	iGlyphBoundingBox[0] = (SHORT) ParseUShortFromPos(&pGlyphPos);
	iGlyphBoundingBox[1] = (SHORT) ParseUShortFromPos(&pGlyphPos);
	iGlyphBoundingBox[2] = (SHORT) ParseUShortFromPos(&pGlyphPos);
	iGlyphBoundingBox[3] = (SHORT) ParseUShortFromPos(&pGlyphPos);

	pSC->iLSideBearing = (SHORT) iGlyphBoundingBox[0];

	if (OTFData.iHeadStart != 0 &&
		(iGlyphBoundingBox[0] < OTFData.iFontBoundingBoxes[0] ||
		iGlyphBoundingBox[1] < OTFData.iFontBoundingBoxes[1] ||
		iGlyphBoundingBox[2] > OTFData.iFontBoundingBoxes[2] ||
		iGlyphBoundingBox[3] > OTFData.iFontBoundingBoxes[3]))
	{
		if (!OTFData.bBadBoundingBoxData || (OTFData.OTFlags & OTF_FONT_LINT))
		{
			//TRACEUSER("Greg", "Glyph bounding box data exceeds font bounding box data for GID %d\n", iGid);
			OTFData.bBadGlyphData = TRUE;

			if (!(OTFData.OTFlags & OTF_FONT_LINT))
				TRACEUSER("Greg", "Subsequent errors will not be reported.\n");

			OTFData.bBadBoundingBoxData = TRUE;
		}
	}

	if (iPathCnt >= 0)
		ParseTTFSimpleGlyph(pFontData, pSC, iPathCnt, iGlyphBoundingBox);
	else
		ParseTTFCompositeGlyph(pFontData, pSC, OTFData.iGlyphStart + iEnd);

	//I don't check that composite glyphs fit in the bounding box
	//because the components may not have been read in yet
	//I'll check against the font bb later, if validation mode
	ptrdiff_t iDiff = pGlyphPos - pFontData;

	if (iStart > iEnd)
	{
		TRACEUSER("Greg", "Glyph bounding box data exceeds font bounding box data for GID %d\n", iGid);
		OTFData.bBadGlyphData = TRUE;
	}
	else if (iDiff > OTFData.iGlyphStart + iEnd)
	{
		TRACEUSER("Greg", "Bad glyph (%d), its definition extends beyond the space allowed for it\n", iGid);
		OTFData.bBadGlyphData = TRUE;
	}
#endif

	return pSC;
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFEncodings(const BYTE* pFontData, INT iJustInUse)

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
				iJustInUse - flag if actions should be commenced only on part of
				font in use
	Outputs:	Set encoding data in OTFData structure
	Purpose:	Parses font encodings


********************************************************************************************/
void OTFDataLoader::ParseTTFEncodings(const BYTE* pFontData, INT iJustInUse)
{
	//TODO: implement when necessary ParseTTFEncodings (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFGsubUsed(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs:	Set Gsub data in OTFData structure
	Purpose:	Parses Gsub data of currently used part of the font


********************************************************************************************/

void OTFDataLoader::ParseTTFGsubUsed(const BYTE* pFontData)
{
	//TODO: implement when necessary ParseTTFGsubUsed
}

/********************************************************************************************

>	void OTFDataLoader::ParseOTFMathUsed(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs:	Set Math data in OTFData structure
	Purpose:	Parses Math data of currently used part of the font


********************************************************************************************/

void OTFDataLoader::ParseOTFMathUsed(const BYTE* pFontData)
{
	//TODO: implement when necessary ParseOTFMathUsed
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFMortUsed(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs:	Set Mort data in OTFData structure
	Purpose:	Parses Mort data of currently used part of the font


********************************************************************************************/

void OTFDataLoader::ParseTTFMortUsed(const BYTE* pFontData)
{
	//TODO: implement when necessary ParseTTFMortUsed
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFSimpleGlyph(const BYTE* pFontData, SplineChar* pSC, INT iPathCnt, 
												INT iGlyphBoundingBox[4])

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
				pSC - pointer to SplineChar
				iPathCnt - number of paths for glyph
				iGlyphBoundingBox - current glyph's bounding box
	Outputs:	Glyph data stored in OTFData
	Purpose:	Parses glyph data with simpler structure


********************************************************************************************/

void OTFDataLoader::ParseTTFSimpleGlyph(const BYTE* pFontData, SplineChar* pSC, INT& iPathCnt, 
											INT iGlyphBoundingBox[4])
{
	//TODO: implement when necessary ParseTTFSimpleGlyph (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFCompositeGlyph(const BYTE* pFontData, SplineChar* pSC, 
												  INT iEndOffset)

	Author:		Greg Watroba
	Created:	4/12/2015
	Inputs:		pFontData - raw BYTE font data table
				pSC - pointer to SplineChar
				iEndOffset - glyph offset data
	Outputs:	Glyph data stored in OTFData
	Purpose:	Parses glyph data with complex structure


********************************************************************************************/

void OTFDataLoader::ParseTTFCompositeGlyph(const BYTE* pFontData, SplineChar* pSC, INT iEndOffset)
{
	//TODO: implement when necessary ParseTTFCompositeGlyph (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseCopyrightsData(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	7/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs:	Set Copyright data in OTFData structure
	Purpose:	Parses Copyright data from font

********************************************************************************************/

void OTFDataLoader::ParseCopyrightsData(const BYTE* pFontData)
{
	//TODO: implement when necessary ParseCopyrightsData (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseCFFGlyphs(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	7/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs:	Set Glyph data in OTFData structure
	Purpose:	Parses Glyph data from font stored in Composite Font Format

********************************************************************************************/

void OTFDataLoader::ParseCFFGlyphs(const BYTE* pFontData, bool bKeepGlyphData)
{
	const BYTE* pCFFPos = pFontData + OTFData.uCFFStart;

	if (ParseCharFromPos(&pCFFPos) != '\1')
	{
		TRACEUSER("Greg", "CFF version mismatch\n");
		OTFData.bBadCFFData = TRUE;

		//failsafe for now to prevent crashes
		OTFData.vChars.resize(OTFData.iGlyphCnt);

		return;
	}

	MOVE8BITS(pCFFPos);

	INT iHdrSize = ParseCharFromPos(&pCFFPos);
	//INT iOffSize = ParseCharFromPos(&pCFFPos);
	MOVE8BITS(pCFFPos);

	if (iHdrSize != 4)
		pCFFPos = pFontData + OTFData.uCFFStart + iHdrSize;

	std::vector<std::string> vsFontNames;
	ReadCFFFontNames(&vsFontNames, &pCFFPos, nullptr);
	if (vsFontNames.size())
		OTFData.strCFFFontName = vsFontNames[0];

	INT iWhich = 0;

	std::vector<TopDicts*> vpTemp;
	std::vector<TopDicts*>& vpTopDicts = bKeepGlyphData ? OTFData.vpTopDicts : vpTemp;
	ReadCFFTopDicts(&vpTopDicts, &pCFFPos, &vsFontNames, OTFData.uCFFStart, nullptr);

	vpTopDicts[iWhich]->iStringsOff = (INT)(pCFFPos - pFontData - OTFData.uCFFStart);
	//String index is just the same as fontname index
	std::vector<std::string> vsStrings;
	INT iStringCnt;
	ReadCFFFontNames(&vsStrings, &pCFFPos, &iStringCnt);
	vpTopDicts[iWhich]->iStringsSize = (INT)(pCFFPos - pFontData - OTFData.uCFFStart - vpTopDicts[iWhich]->iStringsOff);

	INT iOffset = (INT)(pCFFPos - pFontData - OTFData.uCFFStart);
	PSChars GSubs;
	ReadCFFSubs(&pCFFPos, &GSubs);

	if (GSubs.iCnt)
	{
		vpTopDicts[iWhich]->iGSubsOff = iOffset;
		vpTopDicts[iWhich]->iGSubsSize = (INT)(pCFFPos - pFontData - OTFData.uCFFStart - vpTopDicts[iWhich]->iGSubsOff);
	}

	std::vector<UINT8> vFDTemp;
	std::vector<UINT8>& vFDSelect = bKeepGlyphData ? OTFData.vFDSelect : vFDTemp;

	//Can be many fonts here. Only decompose the one
	if (vpTopDicts[iWhich]->iCharStringsOff > -1)
	{
		const BYTE* pCharStringPos = pFontData + OTFData.uCFFStart + vpTopDicts[iWhich]->iCharStringsOff;
		ReadCFFSubs(&pCharStringPos, &vpTopDicts[iWhich]->Glyphs);
	}

	if (vpTopDicts[iWhich]->iPrivateOffset > -1)
		ReadCFFPrivate(pFontData, vpTopDicts[iWhich]);

	if (vpTopDicts[iWhich]->iCharsetOff > -1)
		ReadCFFSet(pFontData, vpTopDicts[iWhich]);

	std::vector<TopDicts*> vpSDTemp;
	std::vector<TopDicts*>& vpSubDicts = bKeepGlyphData ? OTFData.vpSubDicts : vpSDTemp;

	if (vpTopDicts[iWhich]->iFDArrayOff == -1)
	{
		ReadCFFFigure(vpTopDicts[iWhich], vsStrings, iStringCnt, &GSubs);
	}
	else
	{
		const BYTE* pFDArrayPos = pFontData + OTFData.uCFFStart + vpTopDicts[iWhich]->iFDArrayOff;
		ReadCFFTopDicts(&vpSubDicts, &pFDArrayPos, nullptr, OTFData.uCFFStart, vpTopDicts[iWhich]);

		const BYTE* pFDSelectPos = pFontData + OTFData.uCFFStart + vpTopDicts[iWhich]->iFDSelectOff;
		ReadFDSelect(&vFDSelect, &pFDSelectPos, vpTopDicts[iWhich]->Glyphs.iCnt);

		for (INT j = 0; vpSubDicts[j] != nullptr; ++j)
		{
			if (vpSubDicts[j]->iPrivateOffset > -1)
				ReadCFFPrivate(pFontData, vpSubDicts[j]);

			if (vpSubDicts[j]->iCharsetOff > -1)
				ReadCFFSet(pFontData, vpSubDicts[j]);
		}

		CIDFigure(vpTopDicts[iWhich], vsStrings, iStringCnt, &GSubs, vpSubDicts, vFDSelect);
	}

	if (vpTopDicts[iWhich]->iEncodingOff > -1)
		ReadCFFEncoding(pFontData, vpTopDicts[iWhich], vsStrings, iStringCnt);

	if (vpTopDicts[iWhich]->iFDArrayOff == -1)
	{
		for (INT i = 0; i < OTFData.iGlyphCnt; ++i)
		{
			if (OTFData.vChars[i])
				OTFData.vChars[i]->iOrigPos = i;
		}
	}

	if (OTFData.bIsOrder2Type)
	{
		for (INT i = 0; i < OTFData.iGlyphCnt; ++i)
			SplineCharConvertToOrder2Type(OTFData.vChars[i]);
	}

	if (!bKeepGlyphData)
	{
		for (auto& pDict : vpTemp)
			SAFE_DELETE(pDict);
		vpTemp.clear();

		for (auto& pSubDict : vpSubDicts)
			SAFE_DELETE(pSubDict);

		vpSubDicts.clear();
	}
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFFontNames(std::vector<std::string> *pvsFontNames, 
											const BYTE** pCFFPos, INT* piCnt)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pCFFPos - position in BYTE data table
				piCnt - pointer to counter needed for parsing
	Outputs:	pvsFontNames - names stored in CFF format
	Purpose:	Parse CFF font names data

********************************************************************************************/

void OTFDataLoader::ReadCFFFontNames(std::vector<std::string> *pvsFontNames, 
									 const BYTE** pCFFPos, INT* piCnt)
{
	UINT16 iCnt = ParseUShortFromPos(pCFFPos);

	if (piCnt)
		(*piCnt) = iCnt;

	if (iCnt == 0)
		return;

	std::vector<UINT32> vOffsets(iCnt + 1);
	INT iOffsize = ParseCharFromPos(pCFFPos);

	for (INT i = 0; i <= iCnt; ++i)
		vOffsets[i] = ParseOffsetFromPos(pCFFPos, iOffsize);

	pvsFontNames->resize(iCnt + 1);

	for (INT i = 0; i < iCnt; ++i)
	{
		if (vOffsets[i + 1] < vOffsets[i])
		{
			/*
				The CFF font type contains a thing called a name INDEX, and that INDEX
				is bad. It is an index of many of the names used in the CFF font.
				GT: We hope the user will never see this.
			*/
			TRACEUSER("Greg", "Bad CFF name INDEX\n");
			OTFData.bBadCFFData = TRUE;
			while (i < iCnt)
			{
				(*pvsFontNames)[i] = "";
				++i;
			}
			--i;
		}
		else
		{
			INT iFNameSize = vOffsets[i + 1] - vOffsets[i];
			(*pvsFontNames)[i].resize(iFNameSize);
			for (INT j = 0; j < iFNameSize; ++j)
				(*pvsFontNames)[i][j] = (CHAR)ParseCharFromPos(pCFFPos);
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFTopDicts(std::vector<TopDicts*>* vpTopDicts, const BYTE** pDictPos,
											std::vector<std::string>& vsFontNames,
											INT32 iCFFOffset, TopDicts* pParentDict)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pDictPos - pointer to raw BYTE font data
				pvsFontNames - pointer to vector of font names
				iCFFOffset - offset to wanted part of data
				pParentDict - point to parent dictionary
	Outputs:	vpTopDicts - pointer to vector of dictionaries for CFF
	Purpose:	Parse CFF font dictionaries for futher data reading

********************************************************************************************/

void OTFDataLoader::ReadCFFTopDicts(std::vector<TopDicts*>* vpTopDicts, const BYTE** pDictPos, 
										std::vector<std::string>* pvsFontNames, 
										INT32 iCFFOffset, TopDicts* pParentDict)
{
	UINT16 iCnt = ParseUShortFromPos(pDictPos);

	if (iCnt == 0)
		return;

	std::vector<UINT32> vOffsets;
	vOffsets.resize(iCnt + 1);

	INT iOffSize = ParseCharFromPos(pDictPos);

	for (INT i = 0; i <= iCnt; ++i)
		vOffsets[i] = ParseOffsetFromPos(pDictPos, iOffSize);

	vpTopDicts->resize(iCnt + 1);

	for (INT i = 0; i < iCnt; ++i)
	{
		ReadCFFTopSingleDict(&(*vpTopDicts)[i], pDictPos, (pvsFontNames != nullptr) ? &(*pvsFontNames)[i] : nullptr,
							 vOffsets[i + 1] - vOffsets[i]);

		if (pParentDict != nullptr && pParentDict->iFontMatrixSet)
			MatMultiply(pParentDict->fFontMatrix, (*vpTopDicts)[i]->fFontMatrix, (*vpTopDicts)[i]->fFontMatrix);

		(*vpTopDicts)[i]->iCFFStart = iCFFOffset;
	}
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFTopSingleDict(TopDicts** pTopDict, const BYTE** pDictPos,
												std::string* sFontName, INT iLen)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pDictPos - pointer to raw BYTE font data
				sFontName - pointer to font name
				iLen - data length
	Outputs:	pTopDict - pointer to dictionary for CFF
	Purpose:	Parse single dictionary data for CFF

********************************************************************************************/

void OTFDataLoader::ReadCFFTopSingleDict(TopDicts** pTopDict, const BYTE** pDictPos, 
											std::string* sFontName, INT iLen)
{
	(*pTopDict) = new TopDicts();

	INT iStackValue;
	INT iStackIndex;
	INT iReturnValue;
	INT iOperatorValue;
	FLOAT fStack[50];

	(*pTopDict)->sFontName = (sFontName) ? (*sFontName) : "";
	(*pTopDict)->fUnderlinePos = -100;
	(*pTopDict)->fUnderlineWidth = 50;
	(*pTopDict)->iCharStringType = 2;
	(*pTopDict)->fFontMatrix[0] = (*pTopDict)->fFontMatrix[3] = 0.001f;

	(*pTopDict)->iNotice = -1;
	(*pTopDict)->iCopyright = -1;
	(*pTopDict)->iFullName = -1;
	(*pTopDict)->iFamilyName = -1;
	(*pTopDict)->iWeight = -1;
	(*pTopDict)->iVersion = -1;
	(*pTopDict)->iPostScriptCode = -1;
	(*pTopDict)->iBaseFontName = -1;
	(*pTopDict)->iSyntheticBase = -1;
	(*pTopDict)->iRosRegistry = -1;
	(*pTopDict)->iFDArrayOff = -1;
	(*pTopDict)->iFDSelectOff = -1;
	(*pTopDict)->iSidFontname = -1;
	(*pTopDict)->iBlendAxisTypes[0] = -1;

	const BYTE* pEndPos = (*pDictPos) + iLen;

	while ((*pDictPos) < pEndPos)
	{
		iStackIndex = 0;
		
		while ((iReturnValue = ReadCffThing(pDictPos, &iStackValue, &fStack[iStackIndex], &iOperatorValue)) != 3 && (*pDictPos) < pEndPos) 
		{
			if (iReturnValue == 1)
				fStack[iStackIndex] = (FLOAT)iStackValue;

			if (iReturnValue != 0 && iStackIndex < 45)
				++iStackIndex;
		}

		if (iReturnValue == 3 && iOperatorValue == 31) 
		{
			SkipCFF2Thing(pDictPos);
		}
		else if (iStackIndex == 0) 
		{
			TRACEUSER("Greg", "No argument to operator\n");
			OTFData.bBadCFFData = TRUE;
		}
		else if (iReturnValue == 3)
		{
			switch (iOperatorValue) 
			{
				case 0:
				{
					(*pTopDict)->iVersion = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 1:
				{
					(*pTopDict)->iNotice = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 0:
				{
					(*pTopDict)->iCopyright = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 2:
				{
					(*pTopDict)->iFullName = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 3:
				{	
					(*pTopDict)->iFamilyName = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 4:
				{
					(*pTopDict)->iWeight = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 1:
				{
					(*pTopDict)->iIsFixedPitch = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 2:
				{
					(*pTopDict)->fItalicAngle = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 3:
				{
					(*pTopDict)->fUnderlinePos = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 4:
				{
					(*pTopDict)->fUnderlineWidth = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 5:
				{
					(*pTopDict)->iPaintType = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 6:
				{
					(*pTopDict)->iCharStringType = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 7:
				{
					memcpy((*pTopDict)->fFontMatrix, fStack, (iStackIndex >= 6 ? 6 : iStackIndex)*sizeof(FLOAT));
					(*pTopDict)->iFontMatrixSet = 1;
					break;
				}
				case 13:
				{
					(*pTopDict)->iUniqueID = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 5:
				{
					memcpy((*pTopDict)->fFontBoundingBox, fStack, (iStackIndex >= 4 ? 4 : iStackIndex)*sizeof(FLOAT));
					break;
				}
				case (12 << 8) + 8:
				{
					(*pTopDict)->fStrokeWidth = fStack[iStackIndex - 1];
					break;
				}
				case 14:
				{
					for (auto i = 0; i < iStackIndex && i < 20; ++i)
						(*pTopDict)->iXuid[i] = static_cast<INT>(fStack[i]);
					break;
				}
				case 15:
				{
					(*pTopDict)->iCharsetOff = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 16:
				{
					(*pTopDict)->iEncodingOff = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 17:
				{
					(*pTopDict)->iCharStringsOff = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 18:
				{
					(*pTopDict)->iPrivateSize = static_cast<INT>(fStack[0]);
					(*pTopDict)->iPrivateOffset = static_cast<INT>(fStack[1]);
					break;
				}
				case (12 << 8) + 20:
				{
					TRACEUSER("Greg", "We're not supporting synthetic fonts\n");
					(*pTopDict)->iSyntheticBase = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 21:
				{
					(*pTopDict)->iPostScriptCode = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 22:
				{
					(*pTopDict)->iBaseFontName = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 23:
				{
					for (auto i = 0; i < iStackIndex && i < 16; ++i)
						(*pTopDict)->fBaseFontBlend[i] = fStack[i];

					break;
				}
				case (12 << 8) + 24:
				{
					TRACEUSER("Greg", "Camelot does not support type2 multiple master fonts\n");
					OTFData.bBadCFFData = TRUE;

					(*pTopDict)->iNMasters = static_cast<INT>(fStack[0]);
					(*pTopDict)->iNAxes = iStackIndex - 4;
					memcpy((*pTopDict)->iWeightVector, fStack + 1, (iStackIndex - 4) * sizeof(FLOAT));
					(*pTopDict)->iLenBuildCharArray = static_cast<INT>(fStack[iStackIndex - 3]);
					(*pTopDict)->iNormalizeDesignVector = static_cast<INT>(fStack[iStackIndex - 2]);
					(*pTopDict)->iConvertDesignVector = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 26:
				{
					INT i = 0;
					for (i = 0; i < iStackIndex && i < 16; ++i)
						(*pTopDict)->iBlendAxisTypes[i] = static_cast<INT>(fStack[i]);

					(*pTopDict)->iBlendAxisTypes[i] = -1;
					break;
				}
				case (12 << 8) + 30:
				{
					(*pTopDict)->iRosRegistry = static_cast<INT>(fStack[0]);
					(*pTopDict)->iRosOrdering = static_cast<INT>(fStack[1]);
					(*pTopDict)->iRosSupplement = static_cast<INT>(fStack[2]);
					break;
				}
				case (12 << 8) + 31:
				{
					(*pTopDict)->fCidFontVersion = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 32:
				{
					(*pTopDict)->iCidFontRevision = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 33:
				{
					(*pTopDict)->iCidFontType = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 34:
				{
					(*pTopDict)->iCidCount = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 35:
				{
					(*pTopDict)->iUidBase = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 36:
				{
					(*pTopDict)->iFDArrayOff = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 37:
				{
					(*pTopDict)->iFDSelectOff = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 38:
				{
					(*pTopDict)->iSidFontname = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 39:
				{
					TRACEUSER("Greg", "Camelot does not support Chameleon fonts\n");
					break;
				}
				default:
				{
					TRACEUSER("Greg", "Unknown operator: %d", iOperatorValue);
					OTFData.bBadCFFData = TRUE;
					break;
				}
			}
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::SkipCFF2Thing(const BYTE** pDictPos)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pDictPos - pointer to raw BYTE font data
	Outputs:	-
	Purpose:	Skip redundant data in BYTE table

********************************************************************************************/

void OTFDataLoader::SkipCFF2Thing(const BYTE** pDictPos)
{
	INT iVal;

	for (;;)
	{
		iVal = ParseCharFromPos(pDictPos);
		if (iVal >= 247 && iVal <= 254)
			ParseCharFromPos(pDictPos);
		else if (iVal == 255) 
			MOVE32BITS((*pDictPos));
		else if (iVal == 28) 
			MOVE16BITS((*pDictPos));
		else if (iVal == 12) 
			MOVE8BITS((*pDictPos));
		else if (iVal == 14)
			return;
	}
}

/********************************************************************************************

>	INT OTFDataLoader::ReadCffThing(const BYTE**pDictPos, INT* iStackValue, FLOAT* fStackValue,
										INT* iOperatorValue)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pDictPos - pointer to raw BYTE font data
				iStackValue - integer value from previous parsing
	Outputs:	iOperatorValue - operator value
				fStackValue - float stack value
	Purpose:	Read CFF value from BYTE table

********************************************************************************************/

INT OTFDataLoader::ReadCffThing(const BYTE** pDictPos, INT* iStackValue, FLOAT* fStackValue, 
									INT* iOperatorValue)
{
	CHAR Buffer[50] = { 0 }, *pt;
	INT iValue;

	INT iChar = ParseCharFromPos(pDictPos);

	if (iChar == 12) 
	{
		*iOperatorValue = (12 << 8) | ParseCharFromPos(pDictPos);
		return(3);
	}
	else if (iChar <= 21) 
	{
		*iOperatorValue = iChar;
		return(3);
	}
	else if (iChar == 30) 
	{
		//fixed format doesn't exist in dict data but does in type2 strings
		pt = Buffer;
		do 
		{
			iChar = ParseCharFromPos(pDictPos);
			if (pt < Buffer + 44 || (iChar & 0xf) == 0xf || (iChar & 0xf0) == 0xf0) 
			{
				pt = AddNibble(pt, iChar >> 4);
				pt = AddNibble(pt, iChar & 0xf);
			}

		} while (*(pt - 1) != '\0');
		
		*fStackValue = (FLOAT)strtod(Buffer, NULL);
		return(2);
	}
	else if (iChar >= 32 && iChar <= 246) 
	{
		*iStackValue = iChar - 139;
		return(1);
	}
	else if (iChar >= 247 && iChar <= 250) 
	{
		*iStackValue = ((iChar - 247) << 8) + ParseCharFromPos(pDictPos) + 108;
		return(1);
	}
	else if (iChar >= 251 && iChar <= 254) 
	{
		*iStackValue = -((iChar - 251) << 8) - ParseCharFromPos(pDictPos) - 108;
		return(1);
	}
	else if (iChar == 28) 
	{
		iValue = ParseCharFromPos(pDictPos) << 8;
		*iStackValue = (short)(iValue | ParseCharFromPos(pDictPos));
		return(1);
	}
	else if (iChar == 29) 
	{
		//4 byte integers exist in dict data but not in type2 strings
		iValue = ParseCharFromPos(pDictPos) << 24;
		iValue = iValue | ParseCharFromPos(pDictPos) << 16;
		iValue = iValue | ParseCharFromPos(pDictPos) << 8;
		*iStackValue = (int)(iValue | ParseCharFromPos(pDictPos));
		return(1);
	}

	TRACEUSER("Greg", "Unexpected value in dictionary %d\n", iChar);
	OTFData.bBadCFFData = TRUE;
	*iStackValue = 0;
	return(0);
}

/********************************************************************************************

>	CHAR* OTFDataLoader::AddNibble(CHAR* pChar, int iNib)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pChar - pointed CHAR array
				iNib - flag which char should we add
	Returns:	Changed CHAR array
	Purpose:	Add value to pointer char array depending on iNib value

********************************************************************************************/

CHAR* OTFDataLoader::AddNibble(CHAR* pChar, int iNib)
{
	if (iNib <= 9)
		*pChar++ = (CHAR)iNib + '0';
	else if (iNib == 10)
		*pChar++ = '.';
	else if (iNib == 11)
		*pChar++ = 'E';
	else if (iNib == 12) 
	{
		*pChar++ = 'E';
		*pChar++ = '-';
	}
	else if (iNib == 14)
		*pChar++ = '-';
	else if (iNib == 15)
		*pChar++ = '\0';

	return(pChar);
}

/********************************************************************************************

>	INT OTFDataLoader::ClassFindCnt(std::vector<UINT16>& vClass, INT iTotalOffset)

	Author:		Greg Watroba
	Created:	21/09/2016
	Inputs:		vClass - class vector
				iTOtalOffset - total offset for class vector
	Returns:	Class counter
	Purpose:	Returns class counter in passed vector

********************************************************************************************/

INT OTFDataLoader::ClassFindCnt(std::vector<UINT16>& vClass, INT iTotalOffset)
{
	UINT16 uMax = 0;

	for (int i = 0; i < iTotalOffset; ++i)
	{
		if (vClass[i] > uMax)
			uMax = vClass[i];
	}

	return (uMax + 1);
}

/********************************************************************************************

>	std::string OTFDataLoader::GlyphsToNames(std::vector<UINT16>* pvGlyphs, BOOL bMakeUnique)

	Author:		Greg Watroba
	Created:	21/09/2016
	Inputs:		pvGlyphs - pointer to glyph vector
				bMakeUnique - create unique name for glyph
	Returns:	Glyph name
	Purpose:	Returns name for given glyph (unique name if requested)

********************************************************************************************/

std::string OTFDataLoader::GlyphsToNames(std::vector<UINT16>* pvGlyphs, BOOL bMakeUnique)
{
	std::string sResult;
	return sResult;
	/*std::string sResult;

	if (!pvGlyphs)
		return sResult;

	if (bMakeUnique)
	{
		std::sort((*pvGlyphs).begin(), (*pvGlyphs).end());
		auto last = std::unique((*pvGlyphs).begin(), (*pvGlyphs).end());
		std::fill(last, (*pvGlyphs).end(), (UINT16)0xffff);
	}

	INT iLen = 0;

	for (int i = iLen = 0; (*pvGlyphs)[i] != 0xffff; ++i)
	{
		if ((*pvGlyphs)[i] >= OTFData.iGlyphCnt)
		{
			if (!OTFData.bBadOpenTypeData)
				TRACEUSER("Greg", "GID out of range\n");

			return sResult;
		}

		if (OTFData.vChars[(*pvGlyphs)[i]])
			iLen += (INT)OTFData.vChars[(*pvGlyphs)[i]]->sName.length();
	}

	for (int i = 0; (*pvGlyphs)[i] != 0xffff; ++i)
	{
		if (OTFData.vChars[(*pvGlyphs)[i]])
		{
			sResult.append(OTFData.vChars[(*pvGlyphs)[i]]->sName);
			sResult.append(" ");
		}
	}

	return sResult;*/
}

/********************************************************************************************

>	void OTFDataLoader::MatMultiply(FLOAT fMat1[6], FLOAT fMat2[6], FLOAT fMatTo[6])

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		fMat1 - first input matrix
				fMat2 - second input matrix
	Outputs:	fMatTo - result matrix
	Purpose:	Matrix multiplication used for cff glyphs

********************************************************************************************/

void OTFDataLoader::MatMultiply(FLOAT fMat1[6], FLOAT fMat2[6], FLOAT fMatTo[6])
{
	FLOAT mTrans[6];
	mTrans[0] = fMat1[0] * fMat2[0] + fMat1[1] * fMat2[2];
	mTrans[1] = fMat1[0] * fMat2[1] + fMat1[1] * fMat2[3];
	mTrans[2] = fMat1[2] * fMat2[0] + fMat1[3] * fMat2[2];
	mTrans[3] = fMat1[2] * fMat2[1] + fMat1[3] * fMat2[3];
	mTrans[4] = fMat1[4] * fMat2[0] + fMat1[5] * fMat2[2] + fMat2[4];
	mTrans[5] = fMat1[4] * fMat2[1] + fMat1[5] * fMat2[3] + fMat2[5];
	memcpy(fMatTo, mTrans, sizeof(mTrans));
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFSubs(const BYTE** pCFFPos, PSChars* GSubs)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pCFFPos - position in BYTE data table
	Outputs:	pGSubs - pointer to PSChars structure
	Purpose:	Parse PSChar data from CFF font

********************************************************************************************/

void OTFDataLoader::ReadCFFSubs(const BYTE** pCFFPos, PSChars* pGSubs)
{
	UINT16 iCount = ParseUShortFromPos(pCFFPos);

	if (iCount == 0)
		return;

	BOOL bError = FALSE;

	pGSubs->iCnt = iCount;
	pGSubs->vLens.resize(iCount);
	pGSubs->vValues.resize(iCount);
	std::vector<UINT32> vOffsets(iCount + 1);
	INT iOffSize = ParseCharFromPos(pCFFPos);

	for (INT i = 0; i <= iCount; ++i)
		vOffsets[i] = ParseOffsetFromPos(pCFFPos, iOffSize);

	const BYTE* pCurPos = (*pCFFPos) - 1;

	for (INT i = 0; i < iCount; ++i)
	{
		if (vOffsets[i + 1] > vOffsets[i] && vOffsets[i + 1] - vOffsets[i] < 0x10000)
		{
			pGSubs->vLens[i] = vOffsets[i + 1] - vOffsets[i];
			pGSubs->vValues[i].resize(vOffsets[i + 1] - vOffsets[i]);
			UINT32 j;
			for (j = 0; j < vOffsets[i + 1] - vOffsets[i]; ++j)
				pGSubs->vValues[i][j] = (UINT8)ParseCharFromPos(pCFFPos);
		}
		else
		{
			if (!bError)
				TRACEUSER("Greg", "Bad subroutine INDEX in cff font.\n");

			OTFData.bBadCFFData = TRUE;
			bError = TRUE;

			pGSubs->vLens[i] = 1;
			pGSubs->vValues[i].resize(2);
			pGSubs->vValues[i][0] = 11;
			pGSubs->vValues[i][1] = '\0';

			(*pCFFPos) = pCurPos + vOffsets[i + 1];
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFPrivate(const BYTE* pFontData, TopDicts* pTopDicts)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pFontData - pointer to font data
	Outputs:	pTopDicts - pointer to CFF dictionary
	Purpose:	Parse data for pointed dictionary

********************************************************************************************/

void OTFDataLoader::ReadCFFPrivate(const BYTE* pFontData, TopDicts* pTopDicts)
{
	INT i;
	INT iStackValue;
	INT iReturnValue;
	INT iOperatorValue;
	INT iStackIndex;
	
	FLOAT fStack[50];
	memset(fStack, 0, 50 * sizeof(FLOAT));

	const BYTE* pPrivatePos = pFontData + pTopDicts->iCFFStart + pTopDicts->iPrivateOffset;
	const BYTE* pEndPos = pFontData + pTopDicts->iCFFStart + pTopDicts->iPrivateOffset + pTopDicts->iPrivateSize;

	pTopDicts->iSubrsOff = -1;
	pTopDicts->fExpansionFactor = 0.06f;
	pTopDicts->fBlueFuzz = 1.0f;
	pTopDicts->fBlueShift = 7.0f;
	pTopDicts->fBlueScale = 0.039625f;

	while (pPrivatePos < pEndPos) 
	{
		iStackIndex = 0;
		
		while ((iReturnValue = ReadCffThing(&pPrivatePos, &iStackValue, &fStack[iStackIndex], &iOperatorValue)) != 3 && pPrivatePos < pEndPos) 
		{
			if (iReturnValue == 1)
				fStack[iStackIndex] = (FLOAT)iStackValue;
			if (iReturnValue != 0 && iStackIndex < 45)
				++iStackIndex;
		}
		if (iReturnValue == 3 && iOperatorValue == 31 /* "T2" operator, can have 0 arguments */) 
		{
			SkipCFF2Thing(&pPrivatePos);
		}
		else if (iStackIndex == 0 && iOperatorValue != 6 && iOperatorValue != 7 && 
					iOperatorValue != 8 && iOperatorValue != 9 && 
					iOperatorValue != (12 << 8) + 12 && iOperatorValue != (12 << 8) + 13) 
		{
			TRACEUSER("Greg", "No argument to operator %d in private dict\n", iOperatorValue);
			OTFData.bBadCFFData = TRUE;
		}
		else if (iReturnValue == 3)
		{
			switch (iOperatorValue)
			{
				case 6:
				{
					for (i = 0; i < iStackIndex && i < 14; ++i)
					{
						pTopDicts->fBlueValues[i] = fStack[i];
						if (i != 0)
							pTopDicts->fBlueValues[i] += pTopDicts->fBlueValues[i - 1];
					}
					//Marker for an empty array, which is legal, and different from no array
					if (i == 0) 
						pTopDicts->fBlueValues[0] = 1234567;
					break;
				}
				case 7:
				{
					for (i = 0; i < iStackIndex && i < 10; ++i) 
					{
						pTopDicts->fOtherBlues[i] = fStack[i];
						if (i != 0)
							pTopDicts->fOtherBlues[i] += pTopDicts->fOtherBlues[i - 1];
					}

					if (i == 0) 
						pTopDicts->fOtherBlues[0] = 1234567;

					break;
				}
				case 8:
				{
					for (i = 0; i < iStackIndex && i < 14; ++i) 
					{
						pTopDicts->fFamilyBlues[i] = fStack[i];
						if (i != 0)
							pTopDicts->fFamilyBlues[i] += pTopDicts->fFamilyBlues[i - 1];
					}

					if (i == 0) 
						pTopDicts->fFamilyBlues[0] = 1234567;

					break;
				}
				case 9:
				{
					for (i = 0; i < iStackIndex && i < 10; ++i) 
					{
						pTopDicts->fFamilyOtherBlues[i] = fStack[i];
						if (i != 0)
							pTopDicts->fFamilyOtherBlues[i] += pTopDicts->fFamilyOtherBlues[i - 1];
					}
				
					if (i == 0) 
						pTopDicts->fFamilyOtherBlues[0] = 1234567;

					break;
				}
				case (12 << 8) + 9:
				{
					pTopDicts->fBlueScale = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 10:
				{
					pTopDicts->fBlueShift = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 11:
				{
					pTopDicts->fBlueFuzz = fStack[iStackIndex - 1];
					break;
				}
				case 10:
				{
					pTopDicts->iStdHw = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 11:
				{
					pTopDicts->iStdVw = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 12:
				{
					for (i = 0; i < iStackIndex && i < 10; ++i)
					{
						pTopDicts->fStemsNapH[i] = fStack[i];
						if (i != 0)
							pTopDicts->fStemsNapH[i] += pTopDicts->fStemsNapH[i - 1];
					}

					if (i == 0)
						pTopDicts->fStemsNapH[0] = 1234567;

					break;
				}
				case (12 << 8) + 13:
				{
					for (i = 0; i < iStackIndex && i < 10; ++i)
					{
						pTopDicts->fStemsNapV[i] = fStack[i];
						if (i != 0)
							pTopDicts->fStemsNapV[i] += pTopDicts->fStemsNapV[i - 1];
					}

					if (i == 0)
						pTopDicts->fStemsNapV[0] = 1234567;

					break;
				}
				case (12 << 8) + 14:
				{
					pTopDicts->iForceBold = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 15:
				{
					pTopDicts->fForceBoldThreshold = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 16:
				{
					break;
				}
				case (12 << 8) + 17:
				{
					pTopDicts->iLanguageGroup = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case (12 << 8) + 18:
				{
					pTopDicts->fExpansionFactor = fStack[iStackIndex - 1];
					break;
				}
				case (12 << 8) + 19:
				{
					pTopDicts->iInitialRandomSeed = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 19:
				{
					pTopDicts->iSubrsOff = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 20:
				{
					pTopDicts->iDefaultWidthX = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				case 21:
				{
					pTopDicts->iNominalWidthX = static_cast<INT>(fStack[iStackIndex - 1]);
					break;
				}
				default:
				{
					TRACEUSER("Greg", "Unknown operator in %s: %x\n", pTopDicts->sFontName.c_str(), iOperatorValue);
					OTFData.bBadCFFData = TRUE;
					break;
				}
			}
		}
	}

	if (pTopDicts->iSubrsOff > -1)
	{
		const BYTE* pSubsData = pFontData + pTopDicts->iCFFStart + pTopDicts->iPrivateOffset + pTopDicts->iSubrsOff;
		const BYTE* pOrgPos = pSubsData;
		ReadCFFSubs(&pSubsData, &pTopDicts->LocalSub);
		pTopDicts->iLocalSubSize = (INT)(pSubsData - pOrgPos);
	}
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFPrivate(const BYTE* pFontData, TopDicts* pTopDicts)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pFontData - pointer to font data
	Outputs:	pTopDicts - pointer to CFF dictionary
	Purpose:	Parse CFF set for pointed dictionary

********************************************************************************************/

void OTFDataLoader::ReadCFFSet(const BYTE* pFontData, TopDicts* pTopDicts)
{
	UINT16 iLength = static_cast<UINT16>(pTopDicts->Glyphs.iCnt);
	UINT16 iSetIterator = 0;

	if (pTopDicts->iCharsetOff == 0) 
	{
		//ISO Adobe charset
		pTopDicts->vecTDCharset.resize(iLength);
		for (iSetIterator = 0; iSetIterator < iLength && iSetIterator <= 228; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator;
	}
	else if (pTopDicts->iCharsetOff == 1) 
	{
		//Expert charset
		pTopDicts->vecTDCharset.resize(iLength < 162 ? 162 : iLength);
		pTopDicts->vecTDCharset[0] = 0;
		pTopDicts->vecTDCharset[1] = 1;
		for (iSetIterator = 2; iSetIterator < iLength && iSetIterator <= 238 - 227; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 227;
		pTopDicts->vecTDCharset[12] = 13;
		pTopDicts->vecTDCharset[13] = 14;
		pTopDicts->vecTDCharset[14] = 15;
		pTopDicts->vecTDCharset[15] = 99;
		for (iSetIterator = 16; iSetIterator < iLength && iSetIterator <= 248 - 223; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 223;
		pTopDicts->vecTDCharset[25] = 27;
		pTopDicts->vecTDCharset[26] = 28;
		for (iSetIterator = 27; iSetIterator < iLength && iSetIterator <= 266 - 222; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 222;
		pTopDicts->vecTDCharset[44] = 109;
		pTopDicts->vecTDCharset[45] = 110;
		for (iSetIterator = 46; iSetIterator < iLength && iSetIterator <= 318 - 221; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 221;
		pTopDicts->vecTDCharset[96] = 158;
		pTopDicts->vecTDCharset[97] = 155;
		pTopDicts->vecTDCharset[98] = 163;
		for (iSetIterator = 99; iSetIterator < iLength && iSetIterator <= 326 - 220; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 220;
		pTopDicts->vecTDCharset[107] = 150;
		pTopDicts->vecTDCharset[108] = 164;
		pTopDicts->vecTDCharset[109] = 169;
		for (iSetIterator = 110; iSetIterator < iLength && iSetIterator <= 378 - 217; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 217;
	}
	else if (pTopDicts->iCharsetOff == 2) 
	{
		//Expert subset charset
		pTopDicts->vecTDCharset.resize((iLength < 130) ? 130 : iLength);
		pTopDicts->vecTDCharset[0] = 0;
		pTopDicts->vecTDCharset[1] = 1;
		for (iSetIterator = 2; iSetIterator < iLength && iSetIterator <= 238 - 227; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 227;
		pTopDicts->vecTDCharset[12] = 13;
		pTopDicts->vecTDCharset[13] = 14;
		pTopDicts->vecTDCharset[14] = 15;
		pTopDicts->vecTDCharset[15] = 99;
		for (iSetIterator = 16; iSetIterator < iLength && iSetIterator <= 248 - 223; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 223;
		pTopDicts->vecTDCharset[25] = 27;
		pTopDicts->vecTDCharset[26] = 28;
		for (iSetIterator = 27; iSetIterator < iLength && iSetIterator <= 266 - 222; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 222;
		pTopDicts->vecTDCharset[44] = 109;
		pTopDicts->vecTDCharset[45] = 110;
		for (iSetIterator = 46; iSetIterator < iLength && iSetIterator <= 272 - 221; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 221;
		pTopDicts->vecTDCharset[51] = 300;
		pTopDicts->vecTDCharset[52] = 301;
		pTopDicts->vecTDCharset[53] = 302;
		pTopDicts->vecTDCharset[54] = 305;
		pTopDicts->vecTDCharset[55] = 314;
		pTopDicts->vecTDCharset[56] = 315;
		pTopDicts->vecTDCharset[57] = 158;
		pTopDicts->vecTDCharset[58] = 155;
		pTopDicts->vecTDCharset[59] = 163;
		for (iSetIterator = 60; iSetIterator < iLength && iSetIterator <= 326 - 260; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 260;
		pTopDicts->vecTDCharset[67] = 150;
		pTopDicts->vecTDCharset[68] = 164;
		pTopDicts->vecTDCharset[69] = 169;
		for (iSetIterator = 110; iSetIterator < iLength && iSetIterator <= 346 - 217; ++iSetIterator)
			pTopDicts->vecTDCharset[iSetIterator] = iSetIterator + 217;
	}
	else 
	{
		pTopDicts->vecTDCharset.resize(iLength);
		pTopDicts->vecTDCharset[0] = 0;

		const BYTE* pCFFSubPos = pFontData + pTopDicts->iCFFStart + pTopDicts->iCharsetOff;

		INT iFormat = ParseCharFromPos(&pCFFSubPos);
		if (iFormat == 0) 
		{
			for (iSetIterator = 1; iSetIterator < iLength; ++iSetIterator)
				pTopDicts->vecTDCharset[iSetIterator] = ParseUShortFromPos(&pCFFSubPos);
		}
		else if (iFormat == 1) 
		{
			for (iSetIterator = 1; iSetIterator < iLength;) 
			{
				UINT16 iFirst = pTopDicts->vecTDCharset[iSetIterator++] = ParseUShortFromPos(&pCFFSubPos);
				INT iCnt = ParseCharFromPos(&pCFFSubPos);
				for (INT j = 0; j < iCnt; ++j)
					pTopDicts->vecTDCharset[iSetIterator++] = ++iFirst;
			}
		}
		else if (iFormat == 2) 
		{
			for (iSetIterator = 1; iSetIterator < iLength;) 
			{
				UINT16 iFirst = pTopDicts->vecTDCharset[iSetIterator++] = ParseUShortFromPos(&pCFFSubPos);
				INT iCnt = ParseUShortFromPos(&pCFFSubPos);
				for (INT j = 0; j < iCnt; ++j)
					pTopDicts->vecTDCharset[iSetIterator++] = ++iFirst;
			}
		}
		else 
		{
			TRACEUSER("Greg", "Unexpected charset format in cff: %d\n", iFormat);
			OTFData.bBadCFFData = TRUE;
		}
	}

	while (iSetIterator < iLength) 
		pTopDicts->vecTDCharset[iSetIterator++] = 0;
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFFigure(TopDicts* pTopDicts, std::vector<std::string>& vsStrings,
										 INT iStringCnt, PSChars* pGSubs)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pTopDicts - pointer to CFF dictionary
				vsStrings - vector of strings
				iStringCnt - string counter
				pGSubs - pointer to GSub structure
	Outputs:	most of the data is stored in OTFData
	Purpose:	Parse and create appropriate CFF-based glyphs

********************************************************************************************/

void OTFDataLoader::ReadCFFFigure(TopDicts* pTopDicts, std::vector<std::string>& vsStrings, 
									 INT& iStringCnt, PSChars* pGSubs)
{
	PSChars *pCharSub;
	PSContext PsContext;

	CFFInfoFillUp(pTopDicts, vsStrings, iStringCnt);

	INT iCharsType = pTopDicts->iCharStringType;
	PsContext.iIsType2 = iCharsType - 1;
	PsContext.iPaintType = pTopDicts->iPaintType;
	pGSubs->iBias = iCharsType==1 ? 0 : pGSubs->iCnt < 1240 ? 107 : pGSubs->iCnt <33900 ? 1131 : 32768;
	pCharSub = &pTopDicts->LocalSub;
	pCharSub->iBias = iCharsType==1 ? 0 : pCharSub->iCnt < 1240 ? 107 : pCharSub->iCnt <33900 ? 1131 : 32768;

	OTFData.vChars.resize(OTFData.iGlyphCnt);

	for (INT i = 0; i < OTFData.iGlyphCnt; ++i ) 
	{
		OTFData.vChars[i] = PSCharStringToSplines(pTopDicts->Glyphs.vValues[i], pTopDicts->Glyphs.vLens[i], 
													&PsContext, pCharSub, pGSubs, 
													GetSID(pTopDicts->vecTDCharset[i], vsStrings, iStringCnt));

		OTFData.vChars[i]->iVWidth = static_cast<INT16>(OTFData.iEmSize);
		
		if ( iCharsType == 2 ) 
		{
			if ( OTFData.vChars[i]->iWidth == (INT16)0x8000 )
				OTFData.vChars[i]->iWidth = static_cast<INT16>(pTopDicts->iDefaultWidthX);
			else
				OTFData.vChars[i]->iWidth += static_cast<INT16>(pTopDicts->iNominalWidthX);
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::CFFInfoFillUp(TopDicts* pTopDicts, std::vector<std::string>& vsStrings, 
											INT& iStringCnt)

	Author:		Greg Watroba
	Created:	6/01/2016
	Inputs:		pTopDicts - pointer to dictionary structure
				vsString - vector of loaded strings
				iStringCnt - strings quantifier
	Outputs:	Data stored in OTFData structure
	Purpose:	Fill CFF specific data

********************************************************************************************/

void OTFDataLoader::CFFInfoFillUp(TopDicts* pTopDicts, std::vector<std::string>& vsStrings, 
									  INT& iStringCnt)
{
	OTFData.iGlyphCnt = pTopDicts->Glyphs.iCnt;

	if (OTFData.iGlyphCnt < 0)
		OTFData.iGlyphCnt = 0;

	if (pTopDicts->fFontMatrix[0] == 0)
		OTFData.iEmSize = 1000;
	else
		OTFData.iEmSize = static_cast<INT>(rint(1.f / pTopDicts->fFontMatrix[0]));

	OTFData.iAscent = static_cast<INT>(0.8 * OTFData.iEmSize);
	OTFData.iDescent = OTFData.iEmSize - OTFData.iAscent;

	if (!pTopDicts->sFontName.empty()) 
		OTFData.sFontName = pTopDicts->sFontName;
}

/********************************************************************************************

>	std::string OTFDataLoader::GetSID(INT iSID,std::vector<std::string>& vsStrings, INT iSCnt)

	Author:		Greg Watroba
	Created:	6/01/2016
	Inputs:		iSID - string id we want to extract
				vsString - vector of loaded strings
				iSCnt - strings quantifier
	Outputs:	Name of associated character
	Purpose:	Get specific character's name by string ID

********************************************************************************************/

std::string OTFDataLoader::GetSID(INT iSID,std::vector<std::string>& vsStrings, INT iSCnt)
{
	if (iSID == -1)
		return nullptr;
	else if (iSID < nStdStrings)
		return g_pszCFFNames[iSID];
	else if (iSID - nStdStrings > iSCnt)
	{
		TRACEUSER("Greg", "Bad sid %d (must be less than %d)\n", iSID, iSCnt + nStdStrings);
		OTFData.bBadCFFData = TRUE;
		return nullptr;
	}
	else
		return vsStrings[iSID - nStdStrings];
}

/********************************************************************************************

>	SplineChar* OTFDataLoader::PSCharStringToSplines(UINT8 iType1, INT iLen, PSContext* pContext,
													PSChars* pCharSub, PSChars* pGSubs, std::string sName)

	Author:		Greg Watroba
	Created:	6/01/2016
	Inputs:		iType1 - character type
				iLen - character length
				pContext - pointer to context structure
				pCharSub - pointer to character sub records
				pGSubs - pointer to glyph sub records
				sName - name of character
	Outputs:	SplineChar object containing glyph data
	Purpose:	Create SplineChar for glyph identification

********************************************************************************************/

SplineChar* OTFDataLoader::PSCharStringToSplines(std::vector<UINT8>& vType1, INT iLen, PSContext* pContext,
													PSChars* pCharSub, PSChars* pGSubs, std::string sName)
{
	SplineChar* pResultChar = new SplineChar();

	pResultChar->sName = sName;
	pResultChar->iUnicodeEnc = -1;
	pResultChar->iWidth = (INT16)0x8000;

	if (sName.empty())
		sName = "unnamed";

	return pResultChar;
}

/********************************************************************************************

>	void OTFDataLoader::ReadFDSelect(std::vector<UINT8>* pvFDSelect, const BYTE** pFDPos,
										INT32 iGlyphCnt)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pvFDSelect - vector for storing font description data
				pFDPos - pointer to font description data
				iGlyphCnt - glyph counter
	Outputs:	most of the data is stored in OTFData
	Purpose:	Parse font description data

********************************************************************************************/

void OTFDataLoader::ReadFDSelect(std::vector<UINT8>* pvFDSelect, const BYTE** pFDPos, 
									INT32& iGlyphCnt)
{
	pvFDSelect->resize(iGlyphCnt);

	INT iFormat = ParseCharFromPos(pFDPos);
	if (iFormat == 0) 
	{
		for (INT i = 0; i < iGlyphCnt; ++i)
			(*pvFDSelect)[i] = static_cast<UINT8>(ParseCharFromPos(pFDPos));
	}
	else if (iFormat == 3) 
	{
		INT iNr = ParseUShortFromPos(pFDPos);
		INT iFirst = ParseUShortFromPos(pFDPos);
		for (INT i = 0; i < iNr; ++i) 
		{
			INT iFontDescriptor = ParseCharFromPos(pFDPos);
			INT iEnd = ParseUShortFromPos(pFDPos);
			for (INT j = iFirst; j < iEnd; ++j) 
			{
				if (j >= iGlyphCnt) 
				{
					TRACEUSER("Greg", "Bad fdselect\n");
					OTFData.bBadCFFData = TRUE;
				}
				else
					(*pvFDSelect)[j] = static_cast<UINT8>(iFontDescriptor);
			}
			iFirst = iEnd;
		}
	}
	else 
	{
		TRACEUSER("Greg", "Didn't understand format for fdselect %d\n", iFormat);
		OTFData.bBadCFFData = TRUE;
	}
}

/********************************************************************************************

>	void OTFDataLoader::CIDFigure(TopDicts* pTopDicts, std::vector<std::string>& vsStrings,
									INT iStringCnt, PSChars* pGSubs, std::vector<TopDicts*>& vpSubDicts,
									std::vector<UINT8>& pvFDSelect)

	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pTopDicts - pointer to CFF dictionary
				vsStrings - vector of strings
				iStringCnt - string counter
				pGSubs - pointer to GSub structure
				vpSubDicts - vector of subdictionaries
				pvFDSelect - vector for storing font description data
	Outputs:	most of the data is stored in OTFData
	Purpose:	Parse glyph data for alternative format if offset exists

********************************************************************************************/

void OTFDataLoader::CIDFigure(TopDicts* pTopDicts, std::vector<std::string>& vsStrings,
								 INT& iStringCnt, PSChars* pGSubs, std::vector<TopDicts*>& vpSubDicts, 
								 std::vector<UINT8>& pvFDSelect)
{
	//TODO: Implement when necessary CIDFigure
}

/********************************************************************************************

>	void OTFDataLoader::ReadCFFEncoding(const BYTE* pFontData, TopDicts* pTopDicts,
											std::vector<std::string>& vsStrings, INT iStringCnt)


	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		pFontData - pointer to font data
				pTopDicts - pointer to CFF dictionary
				vsStrings - vector of strings
				iStringCnt - string counter
	Outputs:	most of the data is stored in OTFData
	Purpose:	Parse glyph data CFF encodings

********************************************************************************************/

void OTFDataLoader::ReadCFFEncoding(const BYTE* pFontData, TopDicts* pTopDicts, 
										std::vector<std::string>& vsStrings, INT& iStringCnt)
{
	//TODO: implement when necessary ReadCFFEncoding (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::SplineCharConvertToOrder2Type(SplineChar* pSC)


	Author:		Greg Watroba
	Created:	9/12/2015
	Inputs:		-
	Outputs:	pSC - pointer to Spline Char
	Purpose:	Convert each layer of SplineChar to Order2 Type

********************************************************************************************/

void OTFDataLoader::SplineCharConvertToOrder2Type(SplineChar* pSC)
{
	//TODO: implement when necessary SplineCharConvertToOrder2Type (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseTyp1Glyphs(const BYTE* pFontData)

	Author:		Greg Watroba
	Created:	7/12/2015
	Inputs:		pFontData - raw BYTE font data table
	Outputs:	Set Glyph data in OTFData structure
	Purpose:	Parses Glyph data from font stored in Typ1 (old font format)

********************************************************************************************/

void OTFDataLoader::ParseTyp1Glyphs(const BYTE* pFontData)
{
	//TODO: implement when necessary ParseTyp1Glyphs
}

/********************************************************************************************

>	void OTFDataLoader::ParseGposScripts(std::vector<Scripts> *pvScripts, const BYTE* pPosData,
											const USHORT& iScriptsOffset, BOOL bIsGPOS)

	Author:		Greg Watroba
	Created:	23/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iScriptsOffset - offset in BYTE data to start reading
				bIsGPOS - flag whether it's GPOS/GSUB data
	Outputs:	pvScripts - vector of Script data read from BYTE data
	Purpose:	Parses GPOS script data from given BYTE data

********************************************************************************************/

void OTFDataLoader::ParseGposScripts(std::vector<Scripts> *pvScripts, const BYTE* pPosData,
										const USHORT& iScriptsOffset, BOOL bIsGPOS)
{
	if (iScriptsOffset >= OTFData.iPositioningSizeBounds)
	{
		if(bIsGPOS)
			TRACEUSER("Greg", "Script offset is out of GPOS table size bounds\n");
		else
			TRACEUSER("Greg", "Script offset is out of GSUB table size bounds\n");

		OTFData.bBadOpenTypeData = TRUE;
		return;
	}

	const BYTE* pScriptPos = pPosData + iScriptsOffset;
	INT32 iScriptCnt = ParseUShortFromPos(&pScriptPos);

	if (iScriptCnt <= 0)
		return;

	if (iScriptCnt > 1000)
	{
		TRACEUSER("Greg", "Wrong amount of scripts: %d\n", iScriptCnt);
		OTFData.bBadOpenTypeData = TRUE;
		return;
	}

	pvScripts->resize(iScriptCnt);

	for (INT32 i = 0; i < iScriptCnt; ++i)
	{
		(*pvScripts)[i].uTag = ParseLongFromPos(&pScriptPos);
		(*pvScripts)[i].uOffset = ParseUShortFromPos(&pScriptPos);
	}

	for (INT32 i = 0; i < iScriptCnt; ++i)
	{
		const BYTE* pLangPos = pPosData + iScriptsOffset + (*pvScripts)[i].uOffset;
		INT32 iLangDef = ParseUShortFromPos(&pLangPos);
		INT32 iLangCnt = ParseUShortFromPos(&pLangPos);

		iLangCnt += (iLangDef != 0);
		iLangCnt = (iLangCnt > MAX_LANG * 8) ? MAX_LANG * 8 : iLangCnt;

		(*pvScripts)[i].iLangCnt = iLangCnt;
		(*pvScripts)[i].vLanguages.resize(iLangCnt);

		INT32 j = 0;

		if (iLangDef != 0) 
		{
			(*pvScripts)[i].vLanguages[0].uTag = CHR('d', 'f', 'l', 't');
			(*pvScripts)[i].vLanguages[0].uOffset = iLangDef + (*pvScripts)[i].uOffset;
			++j;
		}
		for (; j < iLangCnt; ++j) 
		{
			(*pvScripts)[i].vLanguages[j].uTag = ParseLongFromPos(&pLangPos);
			(*pvScripts)[i].vLanguages[j].uOffset = (*pvScripts)[i].uOffset + ParseUShortFromPos(&pLangPos);
		}
		for (j = 0; j < iLangCnt; ++j) 
		{
			/*CHAR T[4];
			TagToChrArr(T, (*pvScripts)[i].vLanguages[j].uTag);
			TRACEUSERPOS("Greg", "LANG CODE: %c %c %c %c\n", T[0], T[1], T[2], T[3]);*/

			if ((INT32)(iScriptsOffset + (*pvScripts)[i].vLanguages[j].uOffset) >= OTFData.iPositioningSizeBounds)
			{
				if(bIsGPOS)
					TRACEUSER("Greg", "Language offset is out of GPOS table size bounds\n");
				else
					TRACEUSER("Greg", "Language offset is out of GSUB table size bounds\n");

				OTFData.bBadOpenTypeData = TRUE;
				return;
			}

			const BYTE* pFeaturesPos = pPosData + iScriptsOffset + (*pvScripts)[i].vLanguages[j].uOffset;
			
			/*USHORT iOrder = */ParseUShortFromPos(&pFeaturesPos);

			(*pvScripts)[i].vLanguages[j].uReq = ParseUShortFromPos(&pFeaturesPos);
			USHORT uFeatureCnt = ParseUShortFromPos(&pFeaturesPos);

			if (uFeatureCnt > 1000)
			{
				TRACEUSER("Greg", "Feature count limit exceeded\n");
				uFeatureCnt = 1000;
			}

			(*pvScripts)[i].vLanguages[j].iFCnt = uFeatureCnt;

			(*pvScripts)[i].vLanguages[j].vFeatures.resize((*pvScripts)[i].vLanguages[j].iFCnt);

			for (INT32 k = 0; k < (*pvScripts)[i].vLanguages[j].iFCnt; ++k)
				(*pvScripts)[i].vLanguages[j].vFeatures[k] = ParseUShortFromPos(&pFeaturesPos);
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGposFeatures(std::vector<Feature> *pvFeatures, const BYTE* pPosData,
										const USHORT& iFeaturesOffset, BOOL bIsGPOS)

	Author:		Greg Watroba
	Created:	23/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iFeaturesOffset - offset in BYTE data to start reading
				bIsGPOS - flag whether it's GPOS/GSUB data
	Outputs:	pvFeatures - vector of Feature data read from BYTE data
	Purpose:	Parses GPOS features data from given BYTE data

********************************************************************************************/

void OTFDataLoader::ParseGposFeatures(std::vector<Feature> *pvFeatures, const BYTE* pPosData,
										const USHORT& iFeaturesOffset, BOOL bIsGPOS)
{
	if (iFeaturesOffset >= OTFData.iPositioningSizeBounds)
	{
		if(bIsGPOS)
			TRACEUSER("Greg", "Feature offset is out of GPOS table size bounds\n");
		else
			TRACEUSER("Greg", "Feature offset is out of GSUB table size bounds\n");

		OTFData.bBadOpenTypeData = TRUE;
		return;
	}

	const BYTE* pFeaturePos = pPosData + iFeaturesOffset;

	INT32 iFeatureCnt = ParseUShortFromPos(&pFeaturePos);
	OTFData.iFeatureCnt = iFeatureCnt;

	if (iFeatureCnt <= 0)
		return;

	if (iFeatureCnt > 1000)
	{
		TRACEUSER("Greg", "Wrong amount of features: %d \n", iFeatureCnt);
		OTFData.bBadOpenTypeData = TRUE;
		return;
	}

	pvFeatures->resize(iFeatureCnt);

	for (INT32 i = 0; i < iFeatureCnt; ++i)
	{
		(*pvFeatures)[i].uTag = ParseLongFromPos(&pFeaturePos);
		(*pvFeatures)[i].uOffset = ParseUShortFromPos(&pFeaturePos);

		for (auto& pNameEntry : s_FriendlyNames)
		{
			if (pNameEntry.iTag == (*pvFeatures)[i].uTag)
			{
				(*pvFeatures)[i].eGlyphSet = pNameEntry.eGlyphSet;
				break;
			}
		}
	}

	for (INT32 i = 0; i < iFeatureCnt; ++i)
	{
		if ((INT32)(iFeaturesOffset + (*pvFeatures)[i].uOffset) >= OTFData.iPositioningSizeBounds)
		{
			if (bIsGPOS)
				TRACEUSER("Greg", "Feature offset is out of GPOS table size bounds\n");
			else
				TRACEUSER("Greg", "Feature offset is out of GSUB table size bounds\n");

			OTFData.bBadOpenTypeData = TRUE;
			return;
		}

		const BYTE* pParamsPos = pPosData + iFeaturesOffset + (*pvFeatures)[i].uOffset;
		INT32 iParameters = ParseUShortFromPos(&pParamsPos);

		if ((*pvFeatures)[i].uTag == CHR('s', 'i', 'z', 'e') && iParameters != 0)
		{
			ParseTTFSizeParameters(pPosData, iFeaturesOffset + iParameters, iFeaturesOffset + iParameters + (*pvFeatures)[i].uOffset);
		} 
		else if ((*pvFeatures)[i].eGlyphSet >= GSET_STYLE_SET_1 && (*pvFeatures)[i].eGlyphSet <= GSET_STYLE_SET_20 && iParameters != 0)
		{
			ParseTTFFeatureNameParameters(pPosData, iFeaturesOffset + iParameters + (*pvFeatures)[i].uOffset, (*pvFeatures)[i].uTag,(*pvFeatures)[i].eGlyphSet);
		}

		(*pvFeatures)[i].iLCnt = ParseUShortFromPos(&pParamsPos);
		(*pvFeatures)[i].vLookups.resize((*pvFeatures)[i].iLCnt);

		for (INT32 j = 0; j < (*pvFeatures)[i].iLCnt; ++j)
		{
			const UINT16 uLookupIndex = ParseUShortFromPos(&pParamsPos);
			(*pvFeatures)[i].vLookups[j] = uLookupIndex;
			if (bIsGPOS)
				OTFData.mapLookupToGPOSFeature[uLookupIndex] = (*pvFeatures)[i].eGlyphSet;
			else
				OTFData.mapLookupToGSUBFeature[uLookupIndex] = (*pvFeatures)[i].eGlyphSet;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::FindAllLangEntries(std::vector<OTFName>* pvOTFNames, 
											  const BYTE* pPosData, INT32 iID)

	Author:		Greg Watroba
	Created:	26/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iID - id of language entry to find names for
	Outputs:	pvOTFNames - names read from OTF data for specific ID
	Purpose:	Parses and stores OTF names for specific ID entry

********************************************************************************************/

void OTFDataLoader::FindAllLangEntries(std::vector<OTFName>* pvOTFNames, 
									   const BYTE* pPosData, INT32 iID)
{
	if (!pvOTFNames)
		return;

	if (iID == 0)
		return;

	UINT size = ::GetFontData(OTFData.hDC, *(DWORD*)"name", 0, nullptr, 0);
	if (size <= 0)
		return;

	unique_ptr<BYTE[]> FontDataTable(new BYTE[size]);

	if (::GetFontData(OTFData.hDC, *(DWORD*)"name", 0, FontDataTable.get(), size) != size)
		return;

	const BYTE* pNameDataPos = FontDataTable.get();

	/*UINT uFormat = */ParseUShortFromPos(&pNameDataPos);
	INT32 iCnt = ParseUShortFromPos(&pNameDataPos);
	USHORT uTabOffset = ParseUShortFromPos(&pNameDataPos);

	//pointer for reading encoded strings - probably there is lot better way to do it
	const BYTE* pTablePos = FontDataTable.get() + uTabOffset;

	for (INT32 i = 0; i < iCnt; ++i)
	{
		INT32 iPlatform = ParseUShortFromPos(&pNameDataPos);
		/*INT32 iEncoding = */ParseUShortFromPos(&pNameDataPos);
		UINT16 iLanguage = ParseUShortFromPos(&pNameDataPos);
		INT32 iName = ParseUShortFromPos(&pNameDataPos);
		INT32 iStrLen = ParseUShortFromPos(&pNameDataPos);
		INT32 iStrOffset = ParseUShortFromPos(&pNameDataPos);

		if (iPlatform == 3 && iName == iID)
		{
			const BYTE* pString = pTablePos + iStrOffset;
			std::string sTemp;

			if (iStrLen % 2 || iStrLen < 2)
			{
				if (iStrLen == 0)
					continue;

				sTemp.reserve(iStrLen);
				for (INT c = 0; c < iStrLen; ++c)
					sTemp.push_back(ParseCharFromPos(&pString));
			}
			else
			{
				sTemp.reserve(iStrLen / 2);
				for (INT c = 0; c < (iStrLen / 2); ++c)
					sTemp.push_back((ParseCharFromPos(&pString) << 8) | ParseCharFromPos(&pString));
			}

			if (!sTemp.empty())
			{
				OTFName CurOTFName;
				CurOTFName.iLang = iLanguage;
				CurOTFName.sName = sTemp;
				pvOTFNames->push_back(CurOTFName);
			}
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFSizeParameters(const BYTE* pPosData, INT32 iBrokenOffsetPos, 
												   INT32 iCorrectOffsetPos)

	Author:		Greg Watroba
	Created:	26/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iBrokenOffsetPos - offset for size parameter in GPOS table which is incorrect
				in terms of OTF standard
				iCorrectOffsetPos - offset for size parameter in GPOS table which is correct
				in terms of OTF standard
	Outputs:	-
	Purpose:	Parses GPOS size parameter - it looks for it in two different places,
				because some fonts do not follow strictly standard
				Data is stored in static structure - probably won't be needed in the end

********************************************************************************************/

void OTFDataLoader::ParseTTFSizeParameters(const BYTE* pPosData, INT32 iBrokenOffsetPos, 
											   INT32 iCorrectOffsetPos)
{
	if (OTFData.iLastSizePos == iBrokenOffsetPos || OTFData.iLastSizePos == iCorrectOffsetPos)
		return;

	if (OTFData.iLastSizePos != 0)
	{
		TRACEUSER("Greg", "Wrong 'size' feature in GPOS table: broken: %d correct: %d\n", 
				 iBrokenOffsetPos, iCorrectOffsetPos);
		return;
	}

	INT32 Test[2];
	INT32 i;

	Test[0] = iCorrectOffsetPos;
	Test[1] = iBrokenOffsetPos;

	for (i = 0; i < 2; ++i) 
	{
		const BYTE* pSizePos = pPosData + Test[i]; //tutaj trzeba startowac od ZERA!!!
		OTFData.iLastSizePos = Test[i];
		OTFData.uDesignSize = ParseUShortFromPos(&pSizePos);

		if (OTFData.uDesignSize == 0)
			continue;

		OTFData.uFontStyleId = ParseUShortFromPos(&pSizePos);
		UINT16 uNid = ParseUShortFromPos(&pSizePos);
		OTFData.uDesignRangeBottom = ParseUShortFromPos(&pSizePos);
		OTFData.uDesignRangeTop = ParseUShortFromPos(&pSizePos);

		if (OTFData.uFontStyleId == 0 && uNid == 0 && OTFData.uDesignSize != 0 &&
			OTFData.uDesignRangeBottom == 0 && OTFData.uDesignRangeTop == 0)
		{
			OTFData.vFontStyleNames.clear();
			break;
		}

		if (OTFData.uDesignSize == 0 || OTFData.uDesignSize < OTFData.uDesignRangeBottom ||
			OTFData.uDesignSize > OTFData.uDesignRangeTop || 
			OTFData.uDesignRangeBottom > OTFData.uDesignRangeTop)
			continue;

		if (OTFData.uFontStyleId == 0 && uNid == 0)
		{
			TRACEUSER("Greg", "This font contains a 'size' feature with a design size and design range but no stylename.\n");
			OTFData.vFontStyleNames.clear();
			break;
		}

		if (uNid < 256 || uNid > 32767)
			continue;
		
		FindAllLangEntries(&OTFData.vFontStyleNames, pPosData, uNid);

		if (OTFData.vFontStyleNames.empty())
			continue;
		else
			break;
	}

	if (i == 2) 
	{
		TRACEUSER("Greg", "The 'size' feature does not seem to follow the standard.\n");

		OTFData.uDesignSize = 0;
		OTFData.uDesignRangeBottom = 0;
		OTFData.uDesignRangeTop = 0;
		OTFData.uFontStyleId = 0;
		OTFData.vFontStyleNames.clear();

		return;
	} 
	else if (i == 1) 
	{
		TRACEUSER("Greg", "The 'size' feature of this font conforms to Adobe's early misinterpretation of the otf standard.\n");
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseTTFFeatureNameParameters(const BYTE* pPosData, INT32 iPos, 
													  UINT32 uTag, const GlyphSetEnum& eGSet)

	Author:		Greg Watroba
	Created:	26/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iPos - offset to start reading Feature names parameters
				uTag - tag of feature to read name parameters
	Outputs:	-
	Purpose:	Parses GPOS feature name parameters
				Data is stored in static structure - probably won't be needed in the end
********************************************************************************************/

void OTFDataLoader::ParseTTFFeatureNameParameters(const BYTE* pPosData, INT32 iPos, 
												  UINT32 uTag, const GlyphSetEnum& eGSet)
{
	const BYTE* pFeatureNamePos = pPosData + iPos;

	INT32 iVersion = ParseUShortFromPos(&pFeatureNamePos);
	UINT16 uNid = ParseUShortFromPos(&pFeatureNamePos);

	if (iVersion >= 10 || uNid < 256 || uNid > 32767)
	{

		CHAR sMSG[256];
		
		if (uNid < 256 || uNid > 32767)
		{
			sprintf_s(sMSG, "The name parameter of the '%c%c%c%c' feature does not contain a valid name id.\n", uTag >> 24, uTag >> 16, uTag >> 8, uTag);
			TRACEUSER("Greg",sMSG);
		}
		else
		{
			sprintf_s(sMSG, "The name parameter of the '%c%c%c%c' feature has an unlikely version number %d.\n", uTag >> 24, uTag >> 16, uTag >> 8, uTag, iVersion);
			TRACEUSER("Greg",sMSG);
		}

		//OTFData.bBadOpenTypeData = TRUE;

		return;
	}

	auto res = OTFData.mapFeatNames.find(eGSet);

	if (res != OTFData.mapFeatNames.end() && uTag == res->second.uTag)
	{
		if (res->second.uNid == uNid)
			return;

		CHAR sMSG[256];
		sprintf_s(sMSG, "There are multiple name ids naming the '%c%c%c%c' feature\n this is technically legitimate, but Camelot can't handle it.\n", uTag >> 24, uTag >> 16, uTag >> 8, uTag);
		TRACEUSER("Greg",sMSG);
	}

	OTFFeatName FeatureName;
	FeatureName.uTag = uTag;
	FeatureName.uNid = uNid;
	FindAllLangEntries(&FeatureName.vNames, pPosData, uNid);

	OTFData.mapFeatNames[eGSet] = FeatureName;

	for (auto& pName : FeatureName.vNames)
		OTFData.mapUserFriendlyGlyphSetNames[eGSet][pName.iLang] = pName.sName;
}

/********************************************************************************************

>	void OTFDataLoader::ParseGposLookups(std::vector<OTLookup> *pvLookups, const BYTE* pPosData,
										const USHORT& iLookupsOffset, BOOL bIsGPOS)

	Author:		Greg Watroba
	Created:	24/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iLookupsOffset - offset in BYTE data to start reading
				bIsGPOS - flag whether it's GPOS/GSUB data
	Outputs:	pvLookups - vector of OTLookup data read from BYTE data
	Purpose:	Parses GPOS lookup data from given BYTE data

********************************************************************************************/


void OTFDataLoader::ParseGposLookups(std::vector<Lookup> *pvLookups, const BYTE* pPosData,
										const USHORT& iLookupsOffset, BOOL bIsGPOS)
{
	if (iLookupsOffset >= OTFData.iPositioningSizeBounds)
	{
		if (bIsGPOS)
			TRACEUSER("Greg", "Lookup offset is out of GPOS table size bounds.\n");
		else
			TRACEUSER("Greg", "Lookup offset is out of GSUB table size bounds.\n");

		OTFData.bBadOpenTypeData = TRUE;
		return;
	}

	const BYTE* pLookupPos = pPosData + iLookupsOffset;
	INT32 iLookupCnt = ParseUShortFromPos(&pLookupPos);

	OTFData.iLookupCnt = iLookupCnt;

	if (iLookupCnt <= 0)
		return;

	if (iLookupCnt > 1000)
	{
		TRACEUSER("Greg", "Wrong amount of lookups: %d \n", iLookupCnt);
		OTFData.bBadOpenTypeData = TRUE;
		return;
	}

	pvLookups->clear();
	pvLookups->resize(iLookupCnt);

	for (INT32 i = 0; i < iLookupCnt; ++i)
		(*pvLookups)[i].uOffset = ParseUShortFromPos(&pLookupPos);

	for (INT32 i = 0; i < iLookupCnt; ++i)
	{
		if ((INT32)(iLookupsOffset + (*pvLookups)[i].uOffset) >= OTFData.iPositioningSizeBounds)
		{
			if (bIsGPOS)
				TRACEUSER("Greg", "Lookup offset is out of GPOS table size bounds.\n");
			else
				TRACEUSER("Greg", "Lookup offset is out of GSUB table size bounds.\n");

			OTFData.bBadOpenTypeData = TRUE;
			return;
		}

		const BYTE* pLookupSubPos = pPosData + iLookupsOffset + (*pvLookups)[i].uOffset;
		
		(*pvLookups)[i].iType = ParseUShortFromPos(&pLookupSubPos);
		(*pvLookups)[i].iFlags = ParseUShortFromPos(&pLookupSubPos);
		(*pvLookups)[i].iSubTabCnt = ParseUShortFromPos(&pLookupSubPos);
		(*pvLookups)[i].vSubTabOffsets.resize((*pvLookups)[i].iSubTabCnt);

		for (INT32 j = 0; j < (*pvLookups)[i].iSubTabCnt; ++j)
			(*pvLookups)[i].vSubTabOffsets[j] = iLookupsOffset + (*pvLookups)[i].uOffset + ParseUShortFromPos(&pLookupSubPos);

		if ((*pvLookups)[i].iFlags & PST_FLAG_USE_MARK_FILTERING_SET)
			(*pvLookups)[i].iFlags |= (ParseUShortFromPos(&pLookupSubPos) << 16);

		OTLookup* pOTLookup = new OTLookup();
		pOTLookup->pNext = BACKORNULL((*pvLookups)[i].vOTLookups);
		(*pvLookups)[i].vOTLookups.push_back(pOTLookup);
		pOTLookup->iLookupIndex = i;

		pOTLookup->eLookupType = (OTLookupType)(((bIsGPOS > 0) << 8) | (*pvLookups)[i].iType);
		pOTLookup->uLookupFlags = (*pvLookups)[i].iFlags;
		pOTLookup->iLookupIndex = i;
		pOTLookup->eGlyphSet = (bIsGPOS) ? OTFData.mapLookupToGPOSFeature[(UINT16)i] : OTFData.mapLookupToGSUBFeature[(UINT16)i];

		for (INT32 j = 0; j < (*pvLookups)[i].iSubTabCnt; ++j)
		{
			LookupSubTable *pSubTable = new LookupSubTable();
			pSubTable->pNext = FRONTORNULL(pOTLookup->vSubTables);
			pSubTable->pOTLookup = pOTLookup;
			pOTLookup->vSubTables.push_front(pSubTable);
			pSubTable->UniqueID = OTFData.LookupSubtableUID;
			pSubTable->eGlyphSet = pOTLookup->eGlyphSet;
			OTFData.vecLookupSubTables.push_back(pSubTable);
			++OTFData.LookupSubtableUID;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::TagLookupsWithScript(std::vector<Scripts> *pvScripts,
												 std::vector<Feature> *pvFeatures,
												 std::vector<Lookup> *pvLookups)

	Author:		Greg Watroba
	Created:	25/11/2015
	Inputs:		pvScripts - parsed Scripts vector from GPOS data
				pvFeatures - parsed Feature vector from GPOS data
				pvLookups - parsed Lookup vector from GPOS data
	Outputs:	-
	Purpose:	Tags parsed data lookups with apropriate scripts and features

********************************************************************************************/

void OTFDataLoader::TagLookupsWithScript(std::vector<Scripts> *pvScripts,
											 std::vector<Feature> *pvFeatures,
											 std::vector<Lookup> *pvLookups)
{
	if (pvScripts->empty() || pvFeatures->empty())
		return;

	for (auto &pScript : (*pvScripts))
	{
		for (auto &pLang : pScript.vLanguages)
		{
			if (pLang.uReq == 0xffff)
			{
				//Do Nothing
			}
			else if (pLang.uReq >= OTFData.iFeatureCnt)
			{
				TRACEUSER("Greg", "Wrong amount of required features: %d.\n", pLang.uReq);
				OTFData.bBadOpenTypeData = TRUE;
			}
			else
			{
				TagLookupsWithFeature(pScript.uTag, pLang.uTag, TRUE, &(*pvFeatures)[pLang.uReq], pvLookups);
			}

			for (INT32 i = 0; i < pLang.iFCnt; ++i)
			{
				if (pLang.vFeatures[i] >= OTFData.iFeatureCnt)
				{
					TRACEUSER("Greg", "Wrong amount of required features: %d.\n", pLang.vFeatures[i]);
					OTFData.bBadOpenTypeData = TRUE;
				}
				else
				{
					TagLookupsWithFeature(pScript.uTag, pLang.uTag, FALSE, &(*pvFeatures)[pLang.vFeatures[i]], pvLookups);
				}
			}
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::TagLookupsWithFeature(UINT32 uScriptTag, UINT32 uLangTag, BOOL bRequiredFeature, 
												 Feature* pFeature, std::vector<Lookup> *pvLookups)

	Author:		Greg Watroba
	Created:	25/11/2015
	Inputs:		uScriptTag - tag of required script
				uLangTag - tag of required language
				bRequiredFeature - flag whether feature is required
				pFeature - pointer to Feature that we will tag lookup with
				pLookup - pointer to Lookup that we will tag with Feature
	Outputs:	-
	Purpose:	Tags parsed data lookups with apropriate feature

********************************************************************************************/

void OTFDataLoader::TagLookupsWithFeature(UINT32& uScriptTag, UINT32& uLangTag, BOOL bRequiredFeature, 
									      Feature* pFeature, std::vector<Lookup> *pvLookups)
{
	UINT32 iFeatureTag = bRequiredFeature ? REQUIRED_FEATURE : pFeature->uTag;
	
	/*
		The otf docs are ambiguous as to the capitalization of the default
		script. The capitalized version is correct (uncapitalized is used for languages)
	*/
	/*if (uLangTag == DEFAULT_LANG)
		uScriptTag = DEFAULT_SCRIPT;*/

	for (UINT16 i = 0; i < pFeature->vLookups.size(); ++i) 
	{
		if (pFeature->vLookups[i] >= OTFData.iLookupCnt) 
		{
			TRACEUSER("Greg", "Wrong amount of required lookups: %d.\n", pFeature->vLookups[i]);
			OTFData.bBadOpenTypeData = TRUE;
		} 
		else 
		{
			OTLookup* pOTLookup = BACKORNULL((*pvLookups)[pFeature->vLookups[i]].vOTLookups);

			if (!pOTLookup)
				continue;

			FeatureScriptLangList* pFSL = nullptr;

			auto res = std::find_if(pOTLookup->vFeatures.begin(), pOTLookup->vFeatures.end(), [iFeatureTag](const FeatureScriptLangList* pFSLL)
			{
				return (pFSLL && pFSLL->uFeatureTag == iFeatureTag);
			});

			if (res != pOTLookup->vFeatures.end())
				pFSL = (*res);

			if (!pFSL)
			{
				pFSL = new FeatureScriptLangList();
				pFSL->uFeatureTag = iFeatureTag;
				pFSL->eGlyphSet = pFeature->eGlyphSet;
				pFSL->pNext = BACKORNULL(pOTLookup->vFeatures);
				pOTLookup->vFeatures.push_back(pFSL);
			}

			AppendScriptLangOnFeatureListIteratively(pFSL,uScriptTag,uLangTag);
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::AppendScriptLangOnFeatureList(FeatureScriptLangList *pFSL, 
														  UINT32 uScriptTag, UINT32 uLangTag)

	Author:		Greg Watroba
	Created:	25/11/2015
	Inputs:		pFSL - pointer to FeatureScriptLangList on which operation will be commenced
				uScriptTag - tag of script to append
				uLangTag - tag of language to append
	Outputs:	-
	Purpose:	Appends script language to features on specific list

********************************************************************************************/

void OTFDataLoader::AppendScriptLangOnFeatureList(FeatureScriptLangList *pFSL, 
												  UINT32& uScriptTag, UINT32& uLangTag)
{
	if (!pFSL)
		return;

	ScriptLangList* pScriptLangL = nullptr;

	auto pRes = std::find_if(pFSL->vScripts.begin(), pFSL->vScripts.end(), [uScriptTag](const ScriptLangList* pSLL)
	{
		return (pSLL && pSLL->uScript == uScriptTag);
	});

	if (pRes != pFSL->vScripts.end())
		pScriptLangL = (*pRes);

	if (!pScriptLangL)
	{
		pScriptLangL = new ScriptLangList;
		pScriptLangL->uScript = uScriptTag;
		pFSL->vScripts.push_back(pScriptLangL);
	}

	INT iLangIt = 0;
	for (iLangIt = 0; iLangIt < MAX_LANG && iLangIt < pScriptLangL->iLangCnt && 
		 pScriptLangL->uLangs[iLangIt] != uLangTag; ++iLangIt);

	if (iLangIt >= MAX_LANG && iLangIt < pScriptLangL->iLangCnt)
	{
		while (iLangIt < pScriptLangL->iLangCnt && 
				(UINT)(iLangIt - MAX_LANG) < pScriptLangL->vMoreLangs.size() && 
				pScriptLangL->vMoreLangs[iLangIt - MAX_LANG] != uLangTag)
			++iLangIt;
	}
	
	if (iLangIt >= pScriptLangL->iLangCnt)
	{
		if (iLangIt < MAX_LANG)
		{
			pScriptLangL->uLangs[iLangIt] = uLangTag;
		}
		else 
		{
			if (iLangIt % MAX_LANG == 0)
				pScriptLangL->vMoreLangs.resize(iLangIt);
			/* We've just allocated MAX_LANG-1 more than we need */
			/*  so we don't do quite some many allocations */
			pScriptLangL->vMoreLangs[iLangIt - MAX_LANG] = uLangTag;
		}
		++pScriptLangL->iLangCnt;
	}
}

/********************************************************************************************

>	void OTFDataLoader::AppendScriptLangOnFeatureListIteratively(FeatureScriptLangList *pFSL, 
																	UINT32 uScriptTag, UINT32 uLangTag)

	Author:		Greg Watroba
	Created:	25/11/2015
	Inputs:		pFSL - pointer to FeatureScriptLangList on which operation will be commenced
				uScriptTag - tag of script to append
				uLangTag - tag of language to append
	Outputs:	-
	Purpose:	Appends script language to features on specific list and it's sliblings

********************************************************************************************/

void OTFDataLoader::AppendScriptLangOnFeatureListIteratively(FeatureScriptLangList *pFSL, 
														     UINT32& uScriptTag, UINT32& uLangTag)
{
	while (pFSL)
	{
		AppendScriptLangOnFeatureList(pFSL, uScriptTag, uLangTag);
		pFSL = pFSL->pNext;
	}
}

/********************************************************************************************

>	void OTFDataLoader::GPOSLookupSwitch(const BYTE* pPosData, INT32 iSubTableOffset, 
											Lookup* pLookup, LookupSubTable* pSubTable, 
											std::vector<Lookup>* pvLookups)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				pvLookups - pointer to Lookup vector
	Outputs:	-
	Purpose:	Parse actual GPOS data

********************************************************************************************/

void OTFDataLoader::GPOSLookupSwitch(const BYTE* pPosData, INT32& iSubTableOffset, Lookup* pLookup,
										 LookupSubTable* pSubTable, std::vector<Lookup>* pvLookups)
{
	switch (pLookup->iType | 0x100)
	{
		case OTL_TYPE_GPOS_SINGLE:
		{
			ParseGPOSSimplePos(pPosData, iSubTableOffset, pLookup, pSubTable);
			break;
		}
		case OTL_TYPE_GPOS_PAIR:
		{
			ParseGPOSKernSubTables(pPosData, iSubTableOffset, pLookup, pSubTable);
			break;
		}
		case OTL_TYPE_GPOS_CURSIVE:
		{
			ParseGPOSCursiveSubTable(pPosData, iSubTableOffset, pLookup, pSubTable);
			break;
		}
		case OTL_TYPE_GPOS_MARK_2_BASE: 
		case OTL_TYPE_GPOS_MARK_2_LIGATURE: 
		case OTL_TYPE_GPOS_MARK_2_MARK:
		{
			ParseGPOSMarkSubTable(pPosData, iSubTableOffset, pLookup, pSubTable);
			break;
		}
		case OTL_TYPE_GPOS_CONTEXT:
		{
			ParseGPOSContextSubTable(pPosData, iSubTableOffset, pLookup, pSubTable, pvLookups);
			break;
		}
		case OTL_TYPE_GPOS_CONTEXT_CHAIN:
		{
			ParseGPOSChainingSubTable(pPosData, iSubTableOffset, pLookup, pSubTable, pvLookups);
			break;
		}
		case 0x109:
		{
			ParseGPOSExtensionSubTable(pPosData, iSubTableOffset, pLookup, pSubTable, pvLookups);
			break;
		}
		/* Any cases added here also need to go in the gposExtensionSubTable */
		default:
		{
			if(!pLookup->vOTLookups.empty())
				TRACEUSER("Greg", "Unknown GPOS sub-table type: %d.\n", pLookup->vOTLookups.back()->eLookupType);

			break;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::GSUBLookupSwitch(const BYTE* pGsubData, INT32 iSubTableOffset, Lookup* pLookup,
											LookupSubTable* pSubTable,
											std::vector<Lookup>* pvLookups, GsubInUseType eType)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				pvLookups - pointer to Lookup vector
				eType - GSUB usage type
	Outputs:	-
	Purpose:	Parse actual GSUB data

********************************************************************************************/

void OTFDataLoader::GSUBLookupSwitch(const BYTE* pGsubData, INT32& iSubTableOffset, Lookup* pLookup,
										LookupSubTable* pSubTable, 
										std::vector<Lookup>* pvLookups, GsubInUseType eType)
{
	switch (pLookup->iType)
	{
		case OTL_TYPE_GSUB_SINGLE:
		{
			ParseGSUBSimpleSubTable(pGsubData, iSubTableOffset, pLookup, pSubTable, eType);
			break;
		}
		case OTL_TYPE_GSUB_MULTIPLE:
		case OTL_TYPE_GSUB_ALTERNATE:
		{
			ParseGSUBMultipleSubTable(pGsubData, iSubTableOffset, pLookup, pSubTable, eType);
			break;
		}
		case OTL_TYPE_GSUB_LIGATURE:
		{
			ParseGSUBLigatureSubTable(pGsubData, iSubTableOffset, pLookup, pSubTable, eType);
			break;
		}
		case OTL_TYPE_GSUB_CONTEXT:
		{
			ParseGSUBContextSubTable(pGsubData, iSubTableOffset, pLookup, pSubTable, eType, pvLookups);
			break;
		}
		case OTL_TYPE_GSUB_CONTEXT_CHAIN:
		{
			ParseGSUBChainSubTable(pGsubData, iSubTableOffset, pLookup, pSubTable, eType, pvLookups);
			break;
		}
		case 7:
		{
			ParseGSUBExtensionSubTable(pGsubData, iSubTableOffset, pLookup, pSubTable, eType, pvLookups);
			break;
		}
		// Note: currently not supported
#if 0
		case OTL_TYPE_GSUB_REVERSE_C_CHAIN:
		{
			ParseGSUBReverseChainSubTable(pGsubData, iSubTableOffset, pLookup, pSubTable, eType);
			break;
		}
#endif
		default:
		{
			TRACEUSER("Greg", "Unknown GSUB Table type: %d\n", pLookup->iType);
			break;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBSimpleSubTable(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
	Outputs:	-
	Purpose:	Parse GSUB Simple substitution data

********************************************************************************************/

void OTFDataLoader::ParseGSUBSimpleSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, 
												Lookup* pLookup, LookupSubTable* pSubTable, 
												GsubInUseType eType)
{
	if (!pSubTable)
		return;

	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	UINT16 uFormat = ParseUShortFromPos(&pLookupData);

	if (uFormat != 1 && uFormat != 2)
		return;

	//TRACEUSER("Greg", "\t\tParseGSUBSimpleSubTable\n");

	INT iCoverage = ParseUShortFromPos(&pLookupData);
	INT32 iDelta = 0, iCnt = 0, iWhich = 0;

	std::vector<UINT16> vGlyphs2;
	std::vector<UINT16> vGlyphs;

	if (uFormat == 1)
	{
		iDelta = ParseUShortFromPos(&pLookupData);
	}
	else
	{
		iCnt = ParseUShortFromPos(&pLookupData);
		vGlyphs2.resize(iCnt);

		for (int i = 0; i < iCnt; ++i)
			vGlyphs2[i] = ParseUShortFromPos(&pLookupData);
	}

	ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad simple substitution table, ignoring\n");
		return;
	}

	if (eType == GIUT_FIND_NAMES)
	{
		OTLookup* pOTL = BACKORNULL(pLookup->vOTLookups);
		if (!pOTL)
			return;

		FeatureScriptLangList* pFL = BACKORNULL(pOTL->vFeatures);
		if (!pFL)
			return;

		for (int i = 0; vGlyphs[i] != 0xffff; ++i)
		{
			if (vGlyphs[i] >= OTFData.iGlyphCnt)
				continue;

			if (OTFData.vChars[vGlyphs[i]]->sName.empty())
				continue;

			iWhich = (uFormat == 1) ? (UINT16)(vGlyphs[i] + iDelta) : vGlyphs2[i];

			if (iWhich < OTFData.iGlyphCnt && 
				iWhich >= 0 && 
				OTFData.vChars[iWhich] != nullptr && 
				OTFData.vChars[iWhich]->sName.empty())
			{
				std::string baseName = OTFData.vChars[vGlyphs[i]]->sName;
				std::string str, pt;
				char tag[5];

				int j = 0;
				for (j = 0; GSUBTagStr[j].tag != 0 && GSUBTagStr[j].tag != pFL->uFeatureTag; ++j);

				if (GSUBTagStr[j].tag != 0)
				{
					pt = GSUBTagStr[j].str;
				}
				else
				{
					tag[0] = pFL->uFeatureTag >> 24;
					if ( (tag[1] = (pFL->uFeatureTag >> 16) & 0xff) == ' ' ) tag[1] = '\0';
					if ( (tag[2] = (pFL->uFeatureTag >> 8) & 0xff) == ' ' ) tag[2] = '\0';
					if ( (tag[3] = (pFL->uFeatureTag) & 0xff) == ' ' ) tag[3] = '\0';
					tag[4] = '\0';
					pt = tag;
				}

				str.append(baseName);
				str.append(".");
				str.append(pt);
				OTFData.vChars[iWhich]->sName = str;
			}
		}
	}
	else if (eType == GIUT_JUST_IN_USE)
	{
		for (int i = 0; vGlyphs[i] != 0xffff; ++i)
		{
			if (vGlyphs[i] >= OTFData.iGlyphCnt)
				continue;

			OTFData.vInUse[vGlyphs[i]] = TRUE;
			iWhich = (uFormat == 1) ? (UINT16)(vGlyphs[i] + iDelta) : vGlyphs2[i];
			OTFData.vInUse[iWhich] = TRUE;
		}
	}
	else if (eType == GIUT_NORMAL)
	{
		for (int i = 0; vGlyphs[i] != 0xffff; ++i)
		{
			if (vGlyphs[i] >= OTFData.iGlyphCnt || OTFData.vChars[vGlyphs[i]] == nullptr)
				continue;

			iWhich = (uFormat == 1) ? (UINT16)(vGlyphs[i] + iDelta) : vGlyphs2[i];

			if (iWhich >= OTFData.iGlyphCnt)
			{
				TRACEUSER("Greg", "OTFDataLoader::ParseGSUBSimpleSubTable: Bad substitution glyph: GID %d not less than %d\n", iWhich, OTFData.iGlyphCnt);
				iWhich = 0;
			}

			if (OTFData.vChars[iWhich] && OTFData.vChars[vGlyphs[i]])
			{
				PosSubTab* pPos = new PosSubTab();
				pPos->uType = PST_SUBSTITUTION;
				pPos->pSubtable = pSubTable;
				pPos->pNext = BACKORNULL(OTFData.vChars[vGlyphs[i]]->vecPosSub);
				pPos->PSTSubs.sVariant = OTFData.vChars[iWhich]->sName;
				pPos->PSTSubs.vecGlyphs.push_back((UINT16)iWhich);
				pPos->Priority = (UINT16)i;
				pPos->GlyphID = (DWORD)vGlyphs[i];
				if (pPos->pSubtable) pPos->pSubtable->vecPosSubTab.push_back(pPos);

				OTFData.vChars[vGlyphs[i]]->vecPosSub.push_back(pPos);
			}
		}
	}

	if (pSubTable)
		pSubTable->bPerGlyphPstOrKern = TRUE;
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBMultipleSubTable(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
	Outputs:	-
	Purpose:	Parse GSUB Multiple substitution data

********************************************************************************************/

void OTFDataLoader::ParseGSUBMultipleSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, 
												Lookup* pLookup, LookupSubTable* pSubTable, 
												GsubInUseType eType)
{
	if (!pSubTable)
		return;

	if (eType == GIUT_FIND_NAMES)
		return;

	//TRACEUSER("Greg", "\t\tParseGSUBMultipleSubTable\n");

	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	UINT16 uFormat = ParseUShortFromPos(&pLookupData);

	if (uFormat != 1)
		return;

	INT iCoverage = ParseUShortFromPos(&pLookupData);
	INT iCnt = ParseUShortFromPos(&pLookupData);

	std::vector<UINT16> vOffsets;
	std::vector<UINT16> vGlyphs2;
	std::vector<UINT16> vGlyphs;

	vOffsets.resize(iCnt);
	for(int i = 0; i < iCnt; ++i)
		vOffsets[i] = ParseUShortFromPos(&pLookupData);

	ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad multiple substitution table, ignoring...\n");
		return;
	}

	INT i = 0;
	for (i = 0; vGlyphs[i] != 0xffff; ++i);

	if (i != iCnt)
	{
		TRACEUSER("Greg", "Coverage table specifies a different number of glyphs than the sub-table expects.\n");
		if (iCnt < i)
			vGlyphs[iCnt] = 0xffff;
		else
			iCnt = i;
	}
	
	INT iMax = 20;
	size_t iLen = 0;
	BOOL bBad = FALSE;
	vGlyphs2.resize(iMax);

	for (i = 0; vGlyphs[i] != 0xffff; ++i)
	{
		const BYTE* pOffData = pGsubData + iSubTableOffset + vOffsets[i];
		iCnt = ParseUShortFromPos(&pOffData);

		if (iCnt > iMax)
		{
			iMax = iCnt + 30;
			vGlyphs2.resize(iMax);
		}

		iLen = 0;
		bBad = FALSE;
		
		for (int j = 0; j < iCnt; ++j)
		{
			vGlyphs2[j] = ParseUShortFromPos(&pOffData);

			if (vGlyphs2[j] >= OTFData.iGlyphCnt)
			{
				if (eType != GIUT_NORMAL)
				{
					TRACEUSER("Greg", "Bad Multiple/Alternate substitution glyph. GID %d not less than %d\n", vGlyphs2[j], OTFData.iGlyphCnt);
					vGlyphs2[j] = 0;
					return;
				}
			}

			if (eType == GIUT_JUST_IN_USE)
			{
				/* Do Nothing */;
			}
			else if (!OTFData.vChars[vGlyphs2[j]])
			{
				TRACEUSER("Greg", "Glyph: %d not found in parsed characters!\n", vGlyphs2[j]);
				bBad = TRUE;
			}
			else
			{
				iLen += OTFData.vChars[vGlyphs2[j]]->sName.length();
			}
		}

		if (eType == GIUT_JUST_IN_USE)
		{
			OTFData.vInUse[vGlyphs[i]] = TRUE;
			for (int j = 0; j < iCnt; ++j)
				OTFData.vInUse[vGlyphs2[j]] = TRUE;
		}
		else if (OTFData.vChars[vGlyphs[i]] && !bBad)
		{
			PosSubTab* pAlt = new PosSubTab();
			OTLookup* pOTLookup = BACKORNULL(pLookup->vOTLookups);
			BOOL bIsGSUBMult = pOTLookup && pOTLookup->eLookupType == OTLookupType::OTL_TYPE_GSUB_MULTIPLE;
			pAlt->uType = (bIsGSUBMult) ? PosSubType::PST_MULTIPLE : PosSubType::PST_ALTERNATE;
			pAlt->pSubtable = pSubTable;
			pAlt->pNext = BACKORNULL(OTFData.vChars[vGlyphs[i]]->vecPosSub);
			pAlt->Priority = (UINT16)i;
			pAlt->GlyphID = (DWORD)vGlyphs[i];
			if (pAlt->pSubtable) pAlt->pSubtable->vecPosSubTab.push_back(pAlt);

			OTFData.vChars[vGlyphs[i]]->vecPosSub.push_back(pAlt);

			for(int j = 0; j < iCnt; ++j)
			{
				pAlt->PSTSubs.sVariant.append(OTFData.vChars[vGlyphs2[j]]->sName);
				pAlt->PSTSubs.sVariant.append(" ");
				pAlt->PSTSubs.vecGlyphs.push_back(vGlyphs2[j]);
			}
		}
	}

	if (pSubTable)
		pSubTable->bPerGlyphPstOrKern = TRUE;
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBLigatureSubTable(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
	Outputs:	-
	Purpose:	Parse GSUB Ligature substitution data

********************************************************************************************/

void OTFDataLoader::ParseGSUBLigatureSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, 
												Lookup* pLookup, LookupSubTable* pSubTable, 
												GsubInUseType eType)
{
	//TRACEUSER("Greg", "\t\tParseGSUBLigatureSubTable\n");

	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	/*UINT16 uFormat = */ParseUShortFromPos(&pLookupData);
	INT iCoverage = ParseUShortFromPos(&pLookupData);
	INT iCnt = ParseUShortFromPos(&pLookupData);

	std::vector<UINT16> vLsOffsets;
	std::vector<UINT16> vGlyphs;
	std::vector<UINT16> vLigOffsets;
	std::vector<UINT16> vLigGlyphs;

	vLsOffsets.resize(iCnt);

	for (int i = 0; i < iCnt; ++i)
		vLsOffsets[i] = ParseUShortFromPos(&pLookupData);

	ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad ligature table, ignoring... \n");
		return;
	}

	INT iLigCnt = 0;

	for (int i = 0; i < iCnt; ++i)
	{
		const BYTE* pOffData = pGsubData + iSubTableOffset + vLsOffsets[i];
		iLigCnt = ParseUShortFromPos(&pOffData);
		vLigOffsets.resize(iLigCnt);

		for (int j = 0; j < iLigCnt; ++j)
			vLigOffsets[j] = ParseUShortFromPos(&pOffData);

		for (int j = 0; j < iLigCnt; ++j)
		{
			const BYTE* pLigData = pGsubData + iSubTableOffset + vLsOffsets[i] + vLigOffsets[j];
			UINT16 uLig = ParseUShortFromPos(&pLigData);

			if (uLig >= OTFData.iGlyphCnt)
			{
				TRACEUSER("Greg", "Bad ligature glyph. GID %d not less than %d\n", uLig, OTFData.iGlyphCnt);
				uLig = 0;
				return;
			}

			INT iCC = ParseUShortFromPos(&pLigData);

			if (iCC < 0 || iCC > 100)
			{
				TRACEUSER("Greg", "Unlikely count of ligature components (%d), I suspect this ligature sub-\n table is garbage, I'm giving up on it.\n", iCC);
				return;
			}

			vLigGlyphs.resize(iCC + 1);
			vLigGlyphs[0] = vGlyphs[i];

			INT k = 0;

			for (k = 0; k < iCC; ++k)
			{
				vLigGlyphs[k + 1] = ParseUShortFromPos(&pLigData);
				if (vLigGlyphs[k + 1] >= OTFData.iGlyphCnt)
				{
					if (eType == GIUT_NORMAL)
						TRACEUSER("Greg", "Bad ligature component glyph. GID %d not less than %d (in ligature %d)\n", vLigGlyphs[k + 1], OTFData.iGlyphCnt, uLig);

					vLigGlyphs[k + 1] = 0;
				}
			}

			if (eType == GIUT_JUST_IN_USE)
			{
				OTFData.vInUse[uLig] = TRUE;
				for (int k = 0; k < iCC; ++k)
					OTFData.vInUse[vLigGlyphs[k]] = TRUE;
			}
			else if (eType == GIUT_FIND_NAMES)
			{
				OTLookup* pOTL = BACKORNULL(pLookup->vOTLookups);
				FeatureScriptLangList* pFSLL = nullptr;
				if (pOTL && !pOTL->vFeatures.empty())
					pFSLL = pOTL->vFeatures.back();

				if (pFSLL && OTFData.vChars[uLig] && OTFData.vChars[uLig]->sName.empty())
				{
					size_t iLen = 0;
					INT k = 0;

					for (k = 0; k < iCC; ++k)
					{
						if (!OTFData.vChars[vLigGlyphs[k]] || OTFData.vChars[vLigGlyphs[k]]->sName.empty())
							break;

						iLen += OTFData.vChars[vLigGlyphs[k]]->sName.length();
					}

					if (k == iCC)
					{
						std::string tag; tag.resize(5);
						tag[0] = pFSLL->uFeatureTag >> 24;
						if ( (tag[1] = (pFSLL->uFeatureTag >> 16) & 0xff) == ' ' ) tag[1] = '\0';
						if ( (tag[2] = (pFSLL->uFeatureTag >> 8) & 0xff) == ' ' ) tag[2] = '\0';
						if ( (tag[3] = (pFSLL->uFeatureTag) & 0xff) == ' ' ) tag[3] = '\0';
						tag[4] = '\0';

						std::string sStr;
						for (int k = 0; k < iCC; ++k)
						{
							sStr.append(OTFData.vChars[vLigGlyphs[k]]->sName);
							sStr.append("_");
						}

						sStr.append(".");
						sStr.append(tag);
						OTFData.vChars[uLig]->sName = sStr;
					}
				}
			}
			else if (OTFData.vChars[uLig])
			{
				BOOL bErr = FALSE;
				size_t iLen = 0;
				for (int k = 0; k < iCC; ++k)
				{
					if (vLigGlyphs[k] < OTFData.iGlyphCnt && OTFData.vChars[vLigGlyphs[k]])
						iLen += OTFData.vChars[vLigGlyphs[k]]->sName.length();
					else
						bErr = TRUE;
				}

				if (!bErr)
				{
					PosSubTab* pLig = new PosSubTab();
					pLig->uType = PosSubType::PST_LIGATURE;
					pLig->pSubtable = pSubTable;
					pLig->pNext = BACKORNULL(OTFData.vChars[uLig]->vecPosSub);
					pLig->PSTLig.pLig = OTFData.vChars[uLig];
					pLig->Priority = (UINT16)j;
					pLig->GlyphID = (DWORD)uLig;
					if (pLig->pSubtable) pLig->pSubtable->vecPosSubTab.push_back(pLig);

					OTFData.vChars[uLig]->vecPosSub.push_back(pLig);

					pLig->PSTLig.vecGlyphs.reserve((size_t)iCC);
					for (int k = 0; k < iCC; ++k)
					{
						if (vLigGlyphs[k] < OTFData.iGlyphCnt && OTFData.vChars[vLigGlyphs[k]])
						{
							pLig->PSTLig.sComponents.append(OTFData.vChars[vLigGlyphs[k]]->sName);
							pLig->PSTLig.sComponents.append(" ");
							pLig->PSTLig.vecGlyphs.push_back(vLigGlyphs[k]);
						}
					}
				}
			}
		}
	}

	if (pSubTable)
		pSubTable->bPerGlyphPstOrKern = TRUE;
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBContextSubTable(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
	Outputs:	-
	Purpose:	Parse GSUB Context substitution data

********************************************************************************************/

void OTFDataLoader::ParseGSUBContextSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, 
												Lookup* pLookup, LookupSubTable* pSubTable, 
												GsubInUseType eType, std::vector<Lookup>* pvLookups)
{
	if (eType == GIUT_FIND_NAMES)
		return;

	//TRACEUSER("Greg", "\t\tParseGSUBContextSubTable\n");

	const BYTE* pLookupData = pGsubData + iSubTableOffset;

	UINT uSubType = ParseUShortFromPos(&pLookupData);

	switch (uSubType)
	{
		case 1:
		{
			ParseGSUBContextSubTableType1(pGsubData, iSubTableOffset, pLookup, pSubTable, eType, pvLookups, FALSE);
		}break;

		case 2:
		{
			ParseGSUBContextSubTableType2(pGsubData, iSubTableOffset, pLookup, pSubTable, eType, pvLookups, FALSE);
		}break;

		case 3:
		{
			ParseGSUBContextSubTableType3(pGsubData, iSubTableOffset, pLookup, pSubTable, eType, pvLookups, FALSE);
		}break;
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBContextSubTableType1(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)

	Author:		Greg Watroba
	Created:	27/10/2016
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
				bGPOS - if TRUE we're parsing GPOS data, GSUB in the other case
	Outputs:	-
	Purpose:	Parse GSUB Context substitution data of subtype 1

********************************************************************************************/

void OTFDataLoader::ParseGSUBContextSubTableType1(const BYTE* pGsubData, INT32 iSubTableOffset,
												  Lookup* pLookup, LookupSubTable* pSubTable,
												  GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)
{
	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	// omit type
	ParseUShortFromPos(&pLookupData);

	std::vector<UINT16> vGlyphs;
	UINT16 iCoverage = ParseUShortFromPos(&pLookupData);
	INT iRCnt = ParseUShortFromPos(&pLookupData);

	std::vector<Rule> vRules;
	vRules.resize(iRCnt);
	for (int i = 0; i < iRCnt; ++i)
		vRules[i].iOffset = ParseUShortFromPos(&pLookupData) + iSubTableOffset;

	ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad contextual table, ignoring \n");
		return;
	}

	UINT16 iCnt = 0;
	BOOL bWarned = FALSE, bWarned2 = FALSE;

	for (INT i = 0; i < iRCnt; ++i)
	{
		const BYTE* pRuleData = pGsubData + vRules[i].iOffset;
		vRules[i].iSCnt = ParseUShortFromPos(&pRuleData);
		iCnt += (UINT16)vRules[i].iSCnt;

		vRules[i].vSubRules.resize(vRules[i].iSCnt);

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
			vRules[i].vSubRules[j].iOffset = ParseUShortFromPos(&pRuleData) + vRules[i].iOffset;

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			const BYTE* pSubRuleData = pGsubData + vRules[i].iOffset;
			vRules[i].vSubRules[j].iGCnt = ParseUShortFromPos(&pSubRuleData);
			vRules[i].vSubRules[j].iSCnt = ParseUShortFromPos(&pSubRuleData);
			vRules[i].vSubRules[j].vGlyphs.resize(vRules[i].vSubRules[j].iGCnt + 1);
			vRules[i].vSubRules[j].vGlyphs[0] = vGlyphs[i];

			for (INT k = 0; k < vRules[i].vSubRules[j].iGCnt; ++k)
			{
				vRules[i].vSubRules[j].vGlyphs[k + 1] = ParseUShortFromPos(&pSubRuleData);

				if (vRules[i].vSubRules[j].vGlyphs[k + 1] >= OTFData.iGlyphCnt)
				{
					if (!bWarned)
						TRACEUSER("Greg", "Bad contextual or chaining sub table. Glyph %d out of range [0,%d)\n", vRules[i].vSubRules[j].vGlyphs[k + 1], OTFData.iGlyphCnt);

					bWarned = TRUE;
					vRules[i].vSubRules[j].vGlyphs[k + 1] = 0;
				}
			}

			vRules[i].vSubRules[j].vGlyphs[vRules[i].vSubRules[j].iGCnt] = 0xffff;
			vRules[i].vSubRules[j].vSeqLookups.resize(vRules[i].vSubRules[j].iSCnt);

			for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
				vRules[i].vSubRules[j].vSeqLookups[k] = new SeqLookup();

			for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
			{
				vRules[i].vSubRules[j].vSeqLookups[k]->iSeq = ParseUShortFromPos(&pSubRuleData);

				if (vRules[i].vSubRules[j].vSeqLookups[k]->iSeq >= vRules[i].vSubRules[j].iGCnt)
				{
					if (!bWarned2)
					{
						TRACEUSER("Greg", "Bad contextual or chaining sub table. Glyph %d out of range [0,%d)\n", vRules[i].vSubRules[j].vSeqLookups[k]->iSeq, vRules[i].vSubRules[j].iGCnt);
						bWarned2 = TRUE;
					}
				}

				vRules[i].vSubRules[j].vSeqLookups[k]->iLookupIndex = ParseUShortFromPos(&pSubRuleData);
			}
		}
	}

	/* Nothing to do. This lookup doesn't really reference any glyphs */
	/*  any lookups it invokes will be processed on their own */
	if (eType == GIUT_JUST_IN_USE)
	{
		for (INT i = 0; i < iRCnt; ++i)
			for (INT j = 0; j < vRules[i].iSCnt; ++j)
				for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
					delete (vRules[i].vSubRules[j].vSeqLookups[k]);
		return;
	}

	FeatureParamSubTab* pFpst = new FeatureParamSubTab();

	pFpst->uType = (UINT16)((bGPOS) ? PST_CONTEXT_POS : PST_CONTEXT_SUB);
	pFpst->uFormat = FPST_GLYPHS;
	
	if (pSubTable)
		pSubTable->pFPST = pFpst;

	pFpst->vSubTables = { pSubTable };

	pFpst->pNext = FRONTORNULL(OTFData.lstFPST);
	OTFData.lstFPST.push_front(pFpst);

	pFpst->vRules.resize(iCnt);
	pFpst->uRuleCnt = iCnt;

	iCnt = 0;

	for (INT i = 0; i < iRCnt; ++i)
	{
		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			auto pRule = new FPSTGlyphRule();

			pRule->sNames = GlyphsToNames(&vRules[i].vSubRules[j].vGlyphs, FALSE);
			pRule->vecNameGlyphs = vRules[i].vSubRules[j].vGlyphs;

			pRule->iLookupCnt = vRules[i].vSubRules[j].iSCnt;
			pRule->vLookups = vRules[i].vSubRules[j].vSeqLookups;

			vRules[i].vSubRules[j].vSeqLookups.clear();

			for (INT k = 0; k < pRule->iLookupCnt; ++k)
				ProcessSubLookups(pvLookups, pRule->vLookups[k], bGPOS);

			pFpst->vRules[iCnt] = pRule;

			++iCnt;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBContextSubTableType2(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)

	Author:		Greg Watroba
	Created:	27/10/2016
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
				bGPOS - if TRUE we're parsing GPOS data, GSUB in the other case
	Outputs:	-
	Purpose:	Parse GSUB Context substitution data of subtype 2

********************************************************************************************/

void OTFDataLoader::ParseGSUBContextSubTableType2(const BYTE* pGsubData, INT32 iSubTableOffset,
												  Lookup* pLookup, LookupSubTable* pSubTable,
												  GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)
{
	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	// omit type
	ParseUShortFromPos(&pLookupData);

	std::vector<UINT16> vGlyphs;
	UINT16 iCoverage = ParseUShortFromPos(&pLookupData);
	UINT16 iClassOffsets = ParseUShortFromPos(&pLookupData);
	INT iRCnt = ParseUShortFromPos(&pLookupData);

	std::vector<Rule> vRules;
	vRules.resize(iRCnt);
	for (INT i = 0; i < iRCnt; ++i)
		vRules[i].iOffset = iSubTableOffset + ParseUShortFromPos(&pLookupData);

	INT iCnt = 0;
	BOOL bWarned = FALSE;

	for (INT i = 0; i < iRCnt; ++i)
	{
		if (vRules[i].iOffset == (UINT32)iSubTableOffset)
			continue;

		const BYTE* pRulePos = pGsubData + vRules[i].iOffset;
		vRules[i].iSCnt = ParseUShortFromPos(&pRulePos);

		if (vRules[i].iSCnt < 0)
		{
			TRACEUSER("Greg", "Bad count in context chaining sub-table.\n");
			return;
		}

		iCnt += vRules[i].iSCnt;
		vRules[i].vSubRules.resize(vRules[i].iSCnt);

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
			vRules[i].vSubRules[j].iOffset = ParseUShortFromPos(&pRulePos) + vRules[i].iOffset;

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			const BYTE* pSubRulePos = pGsubData + vRules[i].vSubRules[j].iOffset;

			vRules[i].vSubRules[j].iCCnt = ParseUShortFromPos(&pSubRulePos);
			vRules[i].vSubRules[j].iSCnt = ParseUShortFromPos(&pSubRulePos);

			if (vRules[i].vSubRules[j].iCCnt < 0)
			{
				TRACEUSER("Greg", "Bad class count in contextual chaining sub-table.\n");
				return;
			}

			vRules[i].vSubRules[j].vClassIndices.resize(vRules[i].vSubRules[j].iCCnt + 1);
			vRules[i].vSubRules[j].vClassIndices[0] = (UINT16)i;

			for (INT k = 0; k < vRules[i].vSubRules[j].iCCnt; ++k)
				vRules[i].vSubRules[j].vClassIndices[k + 1] = ParseUShortFromPos(&pSubRulePos);

			if (vRules[i].vSubRules[j].iSCnt < 0)
			{
				TRACEUSER("Greg", "Bad count in contextual chaining sub-table.\n");
				return;
			}

			vRules[i].vSubRules[j].vSeqLookups.resize(vRules[i].vSubRules[j].iSCnt);

			for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
			{
				vRules[i].vSubRules[j].vSeqLookups[k] = new SeqLookup();
				vRules[i].vSubRules[j].vSeqLookups[k]->iSeq = ParseUShortFromPos(&pSubRulePos);

				if (vRules[i].vSubRules[j].vSeqLookups[k]->iSeq >= vRules[i].vSubRules[j].iCCnt)
				{
					if (!bWarned) 
					{
						TRACEUSER("Greg", "Attempt to apply a lookup to a location out of the range of this contextual\n lookup seq=%d max=%d\n", vRules[i].vSubRules[j].vSeqLookups[k]->iSeq, vRules[i].vSubRules[j].iCCnt);
						bWarned = TRUE;
					}
				}

				vRules[i].vSubRules[j].vSeqLookups[k]->iLookupIndex = ParseUShortFromPos(&pSubRulePos);
			}
		}

	}

	/* Nothing to do. This lookup doesn't really reference any glyphs */
	/*  any lookups it invokes will be processed on their own */
	if (eType == GIUT_JUST_IN_USE)
	{
		for (INT i = 0; i < iRCnt; ++i)
			for (INT j = 0; j < vRules[i].iSCnt; ++j)
				for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
					delete (vRules[i].vSubRules[j].vSeqLookups[k]);
		return;
	}

	FeatureParamSubTab* pFpst = new FeatureParamSubTab();

	pFpst->uType = (UINT16)((bGPOS) ? PST_CONTEXT_POS : PST_CONTEXT_SUB);
	pFpst->uFormat = FPST_GLYPHS;

	if (pSubTable) pSubTable->pFPST = pFpst;
	pFpst->vSubTables = { pSubTable };

	pFpst->pNext = FRONTORNULL(OTFData.lstFPST);
	OTFData.lstFPST.push_front(pFpst);

	pFpst->vRules.resize(iCnt);
	pFpst->uRuleCnt = (UINT16)iCnt;

	std::vector<UINT16> vClasses;
	GetClassDefinitionTable(&vClasses, pGsubData, iSubTableOffset + iClassOffsets);
	pFpst->uNcCnt = (UINT16)ClassFindCnt(vClasses, OTFData.iGlyphCnt);
	INT32 iClassCnt = (INT32)pFpst->uNcCnt;
	ClassToNames(&pFpst->vsNClass, iClassCnt, &vClasses, OTFData.iGlyphCnt);
	pFpst->vsNClassNames.resize(pFpst->uNcCnt);
	ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad contextual substitution table, ignored\n");
		return;
	}

	CoverageMinusClasses(&pFpst->vsNClass[0], &vGlyphs, &vClasses);

	iCnt = 0;

	for (INT i = 0; i < iRCnt; ++i)
	{
		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			auto pRule = new FPSTClassRule();

			pRule->pNClasses = vRules[i].vSubRules[j].vClassIndices;
			pRule->iNCnt = vRules[i].vSubRules[j].iCCnt;

			vRules[i].vSubRules[j].vClassIndices.clear();

			pRule->iLookupCnt = vRules[i].vSubRules[j].iSCnt;
			pRule->vLookups = vRules[i].vSubRules[j].vSeqLookups;

			vRules[i].vSubRules[j].vSeqLookups.clear();
			
			for (INT k = 0; k < pRule->iLookupCnt; ++k)
				ProcessSubLookups(pvLookups, pRule->vLookups[k], bGPOS);

			pFpst->vRules[iCnt] = pRule;

			++iCnt;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBContextSubTableType3(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)

	Author:		Greg Watroba
	Created:	27/10/2016
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
				bGPOS - if TRUE we're parsing GPOS data, GSUB in the other case
	Outputs:	-
	Purpose:	Parse GSUB Context substitution data of subtype 3

********************************************************************************************/

void OTFDataLoader::ParseGSUBContextSubTableType3(const BYTE* pGsubData, INT32 iSubTableOffset,
												  Lookup* pLookup, LookupSubTable* pSubTable,
												  GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)
{
	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	// omit type
	ParseUShortFromPos(&pLookupData);

	INT iGlyphCnt = ParseUShortFromPos(&pLookupData);
	INT iSeqCnt = ParseUShortFromPos(&pLookupData);

	std::vector<UINT16> vCoverage;
	vCoverage.resize(iGlyphCnt);

	for (INT i = 0; i < iGlyphCnt; ++i)
		vCoverage[i] = ParseUShortFromPos(&pLookupData);

	std::vector<SeqLookup*> vSeqLookups;
	vSeqLookups.resize(iSeqCnt);

	BOOL bWarned = FALSE;

	for (INT k = 0; k < iSeqCnt; ++k)
	{
		vSeqLookups[k] = new SeqLookup();
		vSeqLookups[k]->iSeq = ParseUShortFromPos(&pLookupData);

		if (vSeqLookups[k]->iSeq >= iGlyphCnt && !bWarned)
		{
			TRACEUSER("Greg", "Attempt to apply a lookup to a location out of the range of this contextual\n lookup seq=%d, max=%d\n", vSeqLookups[k]->iSeq, iGlyphCnt - 1);
			bWarned = TRUE;
		}

		vSeqLookups[k]->iLookupIndex = ParseUShortFromPos(&pLookupData);
	}

	if (eType == GIUT_JUST_IN_USE)
	{
		for (INT k = 0; k < iSeqCnt; ++k)
			delete (vSeqLookups[k]);

		return;
	}

	FeatureParamSubTab* pFpst = new FeatureParamSubTab();

	pFpst->uType = (UINT16)((bGPOS) ? PST_CONTEXT_POS : PST_CONTEXT_SUB);
	pFpst->uFormat = FPST_COVERAGE;

	if (pSubTable) pSubTable->pFPST = pFpst;
	pFpst->vSubTables = { pSubTable };

	pFpst->pNext = FRONTORNULL(OTFData.lstFPST);
	OTFData.lstFPST.push_front(pFpst);

	pFpst->vRules.resize(1);
	pFpst->uRuleCnt = 1;
	auto pRule = new FPSTCoverageRule();
	pRule->iNCnt = iGlyphCnt;
	pRule->psNCovers.resize(iGlyphCnt);
	pRule->vecNCoverGlyphs.resize(iGlyphCnt);

	for (INT i = 0; i < iGlyphCnt; ++i)
	{
		std::vector<UINT16> vGlyphs;
		ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + vCoverage[i]);
		pRule->psNCovers[i] = GlyphsToNames(&vGlyphs, TRUE);
		pRule->vecNCoverGlyphs[i] = vGlyphs;
	}

	pRule->iLookupCnt = iSeqCnt;
	pRule->vLookups = vSeqLookups;

	for (INT k = 0; k < iSeqCnt; ++k)
		ProcessSubLookups(pvLookups, pRule->vLookups[k], bGPOS);

	pFpst->vRules[0] = pRule;
}

/********************************************************************************************

>	void OTFDataLoader::ProcessSubLookups(std::vector<Lookup>* pvLookups, SeqLookup* pSeqLookup, BOOL bGPOS)

	Author:		Greg Watroba
	Created:	27/10/2016
	Inputs:		pvLookups - container of subtable lookups
				pSeqLookup - pointer to sequence lookup
				bGPOS - if TRUE we're parsing GPOS data, GSUB in the other case
	Outputs:	-
	Purpose:	Sets appropriate lookup to sequence lookups

********************************************************************************************/

void OTFDataLoader::ProcessSubLookups(std::vector<Lookup>* pvLookups, SeqLookup* pSeqLookup, BOOL bGPOS)
{
	if (!pvLookups || !pSeqLookup)
		return;

	if (pSeqLookup->vLookup.empty())
		return;

	INT i = pSeqLookup->iLookupIndex;

	if (i < 0 || i > OTFData.iLookupCnt)
	{
		TRACEUSER("Greg", "Attempt to reference lookup %d (within a contextual lookup), but there are\n only %d lookups in %s\n", i, OTFData.iLookupCnt, bGPOS ? "'GPOS'" : "'GSUB'");
		pSeqLookup->vLookup.clear();
		return;
	}

	pSeqLookup->vLookup = (*pvLookups)[i].vOTLookups;
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBChainSubTable(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
	Outputs:	-
	Purpose:	Parse GSUB Chaining substitution data

********************************************************************************************/

void OTFDataLoader::ParseGSUBChainSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, 
												Lookup* pLookup, LookupSubTable* pSubTable, 
												GsubInUseType eType, std::vector<Lookup>* pvLookups)
{
	const BYTE* pLookupData = pGsubData + iSubTableOffset;

	//TRACEUSER("Greg", "\t\tParseGSUBChainSubTable\n");

	UINT uSubType = ParseUShortFromPos(&pLookupData);

	switch (uSubType)
	{
		case 1:
		{
			ParseGSUBChainSubTableType1(pGsubData, iSubTableOffset, pLookup, pSubTable, GIUT_NORMAL, pvLookups, FALSE);
		}break;

		case 2:
		{
			ParseGSUBChainSubTableType2(pGsubData, iSubTableOffset, pLookup, pSubTable, GIUT_NORMAL, pvLookups, FALSE);
		}break;

		case 3:
		{
			ParseGSUBChainSubTableType3(pGsubData, iSubTableOffset, pLookup, pSubTable, GIUT_NORMAL, pvLookups, FALSE);
		}break;
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBChainSubTableType1(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)

	Author:		Greg Watroba
	Created:	31/10/2016
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
				bGPOS - if TRUE we're parsing GPOS data, GSUB in the other case
	Outputs:	-
	Purpose:	Parse GSUB Chain substitution data of subtype 1

********************************************************************************************/

void OTFDataLoader::ParseGSUBChainSubTableType1(const BYTE* pGsubData, INT32 iSubTableOffset,
												Lookup* pLookup, LookupSubTable* pSubTable,
												GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)
{
	BOOL bWarned = FALSE, bWarned2 = FALSE;

	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	// omit type
	ParseUShortFromPos(&pLookupData);
	UINT16 iCoverage = ParseUShortFromPos(&pLookupData);
	INT iRCnt = ParseUShortFromPos(&pLookupData);

	std::vector<Rule> vRules;
	vRules.resize(iRCnt);

	for (INT i = 0; i < iRCnt; ++i)
		vRules[i].iOffset = iSubTableOffset + ParseUShortFromPos(&pLookupData);

	std::vector<UINT16> vGlyphs;
	ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad contextual chaining table, ignored\n");
		return;
	}

	INT iCnt = 0;
	INT iWhich = 0;

	for (INT i = 0; i < iRCnt; ++i)
	{
		const BYTE* pRuleData = pGsubData + vRules[i].iOffset;
		vRules[i].iSCnt = ParseUShortFromPos(&pRuleData);
		iCnt += (UINT16)vRules[i].iSCnt;
		vRules[i].vSubRules.resize(vRules[i].iSCnt);

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
			vRules[i].vSubRules[j].iOffset = vRules[i].iOffset + ParseUShortFromPos(&pRuleData);

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			const BYTE* pSubRuleData = pGsubData + vRules[i].vSubRules[j].iOffset;
			vRules[i].vSubRules[j].iBCnt = ParseUShortFromPos(&pSubRuleData);
			vRules[i].vSubRules[j].vBGlyphs.resize(vRules[i].vSubRules[j].iBCnt + 1);

			for (INT k = 0; k < vRules[i].vSubRules[j].iBCnt; ++k)
				vRules[i].vSubRules[j].vBGlyphs[k] = ParseUShortFromPos(&pSubRuleData);

			vRules[i].vSubRules[j].vBGlyphs[vRules[i].vSubRules[j].iBCnt] = 0xffff;
			vRules[i].vSubRules[j].iGCnt = ParseUShortFromPos(&pSubRuleData);

			vRules[i].vSubRules[j].vGlyphs.resize(vRules[i].vSubRules[j].iGCnt + 1);
			vRules[i].vSubRules[j].vGlyphs[0] = vGlyphs[i];

			for (INT k = 0; k < vRules[i].vSubRules[j].iGCnt; ++k)
				vRules[i].vSubRules[j].vGlyphs[k + 1] = ParseUShortFromPos(&pSubRuleData);

			vRules[i].vSubRules[j].vGlyphs[vRules[i].vSubRules[j].iGCnt] = 0xffff;


			vRules[i].vSubRules[j].iFCnt = ParseUShortFromPos(&pSubRuleData);
			vRules[i].vSubRules[j].vFGlyphs.resize(vRules[i].vSubRules[j].iFCnt + 1);
			for (INT k = 0; k < vRules[i].vSubRules[j].iFCnt; ++k)
				vRules[i].vSubRules[j].vFGlyphs[k] = ParseUShortFromPos(&pSubRuleData);

			vRules[i].vSubRules[j].vFGlyphs[vRules[i].vSubRules[j].iFCnt] = 0xffff;

			for (iWhich = 0; iWhich > 3; ++iWhich)
			{
				for (INT k = 0; k < (&vRules[i].vSubRules[j].iGCnt)[iWhich]; ++k)
				{
					if ((&vRules[i].vSubRules[j].vGlyphs)[iWhich][k] >= OTFData.iGlyphCnt)
					{
						if (!bWarned)
							TRACEUSER("Greg", "Bad contextual or chaining sub table. Glyph %d out of range [0,%d)\n", (&vRules[i].vSubRules[j].vGlyphs)[iWhich][k], OTFData.iGlyphCnt);

						bWarned = TRUE;
						(&vRules[i].vSubRules[j].vGlyphs)[iWhich][k] = 0;
					}
				}
			}

			vRules[i].vSubRules[j].iSCnt = ParseUShortFromPos(&pSubRuleData);
			vRules[i].vSubRules[j].vSeqLookups.resize(vRules[i].vSubRules[j].iSCnt);
			for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
			{
				vRules[i].vSubRules[j].vSeqLookups[k] = new SeqLookup();
				vRules[i].vSubRules[j].vSeqLookups[k]->iSeq = ParseUShortFromPos(&pSubRuleData);

				if (vRules[i].vSubRules[j].vSeqLookups[k]->iSeq >= vRules[i].vSubRules[j].iGCnt + 1)
				{
					if (!bWarned2)
					{
						TRACEUSER("Greg", "Attempt to apply a lookup to a location out of the range of this contextual\n lookup seq=%d max=%d\n", vRules[i].vSubRules[j].vSeqLookups[k]->iSeq, vRules[i].vSubRules[j].iGCnt);
						bWarned2 = TRUE;
						break;
					}
				}

				vRules[i].vSubRules[j].vSeqLookups[k]->iLookupIndex = ParseUShortFromPos(&pSubRuleData);
			}
		}
	}


	if (eType == GIUT_JUST_IN_USE)
	{
		for (INT i = 0; i < iRCnt; ++i)
			for (INT j = 0; j < vRules[i].iSCnt; ++j)
				for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
					delete (vRules[i].vSubRules[j].vSeqLookups[k]);
		return;
	}

	FeatureParamSubTab* pFPST = new FeatureParamSubTab();
	pFPST->uType = (UINT16)((bGPOS) ? PST_CHAIN_POS : PST_CHAIN_SUB);
	pFPST->uFormat = FPST_GLYPHS;
	
	if (pSubTable) pSubTable->pFPST = pFPST;
	pFPST->vSubTables = { pSubTable };

	pFPST->pNext = FRONTORNULL(OTFData.lstFPST);
	OTFData.lstFPST.push_front(pFPST);

	pFPST->vRules.resize(iCnt);
	pFPST->uRuleCnt = (UINT16)iCnt;

	iCnt = 0;

	for (INT i = 0; i < iRCnt; ++i)
	{
		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			auto pRule = new FPSTGlyphRule();

			pRule->sBack = GlyphsToNames(&vRules[i].vSubRules[j].vBGlyphs, FALSE);
			pRule->vecBackGlyphs = vRules[i].vSubRules[j].vBGlyphs;
			
			pRule->sNames = GlyphsToNames(&vRules[i].vSubRules[j].vGlyphs, FALSE);
			pRule->vecNameGlyphs = vRules[i].vSubRules[j].vGlyphs;
			
			pRule->sFore = GlyphsToNames(&vRules[i].vSubRules[j].vFGlyphs, FALSE);
			pRule->vecForeGlyphs = vRules[i].vSubRules[j].vFGlyphs;

			pRule->iLookupCnt = vRules[i].vSubRules[j].iSCnt;
			pRule->vLookups = vRules[i].vSubRules[j].vSeqLookups;
			vRules[i].vSubRules[j].vSeqLookups.clear();

			for (INT k = 0; k < pRule->iLookupCnt; ++k)
				ProcessSubLookups(pvLookups, pRule->vLookups[k], bGPOS);

			pFPST->vRules[iCnt] = pRule;

			++iCnt;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBChainSubTableType2(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)

	Author:		Greg Watroba
	Created:	31/10/2016
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
				bGPOS - if TRUE we're parsing GPOS data, GSUB in the other case
	Outputs:	-
	Purpose:	Parse GSUB Chain substitution data of subtype 2

********************************************************************************************/

void OTFDataLoader::ParseGSUBChainSubTableType2(const BYTE* pGsubData, INT32 iSubTableOffset,
												Lookup* pLookup, LookupSubTable* pSubTable,
												GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)
{
	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	// omit type
	ParseUShortFromPos(&pLookupData);
	UINT16 iCoverage = ParseUShortFromPos(&pLookupData);
	UINT16 iBClassOffset = ParseUShortFromPos(&pLookupData);
	UINT16 iClassOffset = ParseUShortFromPos(&pLookupData);
	UINT16 iFClassOffset = ParseUShortFromPos(&pLookupData);
	INT iRCnt = ParseUShortFromPos(&pLookupData);

	std::vector<Rule> vRules;
	vRules.resize(iRCnt);

	for (INT i = 0; i < iRCnt; ++i)
	{
		UINT16 iOffset = ParseUShortFromPos(&pLookupData);
		vRules[i].iOffset = (iOffset == 0) ? 0 : (iSubTableOffset + iOffset);
	}

	INT iCnt = 0;
	BOOL bWarned = FALSE;
	for (INT i = 0; i < iRCnt; ++i)
	{
		if (vRules[i].iOffset == 0)
			continue;

		const BYTE* pRulePos = pGsubData + vRules[i].iOffset;
		vRules[i].iSCnt = ParseUShortFromPos(&pRulePos);

		if (vRules[i].iSCnt < 0)
		{
			TRACEUSER("Greg", "Bad count in context chaining sub-table.\n");
			return;
		}

		iCnt += vRules[i].iSCnt;
		vRules[i].vSubRules.resize(vRules[i].iSCnt);

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
			vRules[i].vSubRules[j].iOffset = vRules[i].iOffset + ParseUShortFromPos(&pRulePos);

		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			const BYTE* pSubRulePos = pGsubData + vRules[i].vSubRules[j].iOffset;
			vRules[i].vSubRules[j].iBCnt = ParseUShortFromPos(&pSubRulePos);

			if (vRules[i].vSubRules[j].iBCnt < 0)
			{
				TRACEUSER("Greg", "Bad class count in contextual chaining sub-table.\n");
				return;
			}

			vRules[i].vSubRules[j].vBClassIndices.resize(vRules[i].vSubRules[j].iBCnt);

			for (INT k = 0; k < vRules[i].vSubRules[j].iBCnt; ++k)
				vRules[i].vSubRules[j].vBClassIndices[k] = ParseUShortFromPos(&pSubRulePos);

			vRules[i].vSubRules[j].iCCnt = ParseUShortFromPos(&pSubRulePos);

			if (vRules[i].vSubRules[j].iCCnt < 0)
			{
				TRACEUSER("Greg", "Bad class count in contextual chaining sub-table.\n");
				return;
			}

			vRules[i].vSubRules[j].vClassIndices.resize(vRules[i].vSubRules[j].iCCnt + 1);
			vRules[i].vSubRules[j].vClassIndices[0] = (UINT16)i;

			for (INT k = 0; k < vRules[i].vSubRules[j].iCCnt; ++k)
				vRules[i].vSubRules[j].vClassIndices[k + 1] = ParseUShortFromPos(&pSubRulePos);

			vRules[i].vSubRules[j].iFCnt = ParseUShortFromPos(&pSubRulePos);
			if (vRules[i].vSubRules[j].iFCnt < 0)
			{
				TRACEUSER("Greg", "Bad class count in contextual chaining sub-table.\n");
				return;
			}

			vRules[i].vSubRules[j].vFClassIndices.resize(vRules[i].vSubRules[j].iFCnt);
			for (INT k = 0; k < vRules[i].vSubRules[j].iFCnt; ++k)
				vRules[i].vSubRules[j].vFClassIndices[k] = ParseUShortFromPos(&pSubRulePos);

			vRules[i].vSubRules[j].iSCnt = ParseUShortFromPos(&pSubRulePos);
			if (vRules[i].vSubRules[j].iSCnt < 0)
			{
				TRACEUSER("Greg", "Bad count in contextual chaining sub-table.\n");
				return;
			}

			vRules[i].vSubRules[j].vSeqLookups.resize(vRules[i].vSubRules[j].iSCnt);
			for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
			{
				vRules[i].vSubRules[j].vSeqLookups[k] = new SeqLookup();
				vRules[i].vSubRules[j].vSeqLookups[k]->iSeq = ParseUShortFromPos(&pSubRulePos);

				if (vRules[i].vSubRules[j].vSeqLookups[k]->iSeq >= vRules[i].vSubRules[j].iCCnt)
				{
					if (!bWarned)
					{
						TRACEUSER("Greg", "Attempt to apply a lookup to a location out of the range of this contextual\n lookup seq=%d max=%d\n", vRules[i].vSubRules[j].vSeqLookups[k]->iSeq, vRules[i].vSubRules[j].iCCnt);
						bWarned = TRUE;
					}
				}

				vRules[i].vSubRules[j].vSeqLookups[k]->iLookupIndex = ParseUShortFromPos(&pSubRulePos);
			}
		}
	}

	if (eType == GIUT_JUST_IN_USE)
	{
		for (INT i = 0; i < iRCnt; ++i)
			for (INT j = 0; j < vRules[i].iSCnt; ++j)
				for (INT k = 0; k < vRules[i].vSubRules[j].iSCnt; ++k)
					delete (vRules[i].vSubRules[j].vSeqLookups[k]);

		return;
	}
	
	FeatureParamSubTab* pFpst = new FeatureParamSubTab();
	pFpst->uType = (UINT16)((bGPOS) ? PST_CHAIN_POS : PST_CHAIN_SUB);
	pFpst->uFormat = FPST_CLASS;

	if (pSubTable) pSubTable->pFPST = pFpst;
	pFpst->vSubTables = { pSubTable };

	pFpst->pNext = FRONTORNULL(OTFData.lstFPST);
	OTFData.lstFPST.push_front(pFpst);

	pFpst->vRules.resize(iCnt);
	pFpst->uRuleCnt = (UINT16)iCnt;

	std::vector<UINT16> vClass;
	GetClassDefinitionTable(&vClass, pGsubData, iSubTableOffset + iClassOffset);
	INT iClassCnt = ClassFindCnt(vClass, OTFData.iGlyphCnt);
	pFpst->uNcCnt = (UINT16)iClassCnt;
	ClassToNames(&pFpst->vsNClass, iClassCnt, &vClass, OTFData.iGlyphCnt);
	pFpst->vsNClassNames.resize(pFpst->uNcCnt);

	std::vector<UINT16> vGlyphs;
	ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad contextual chaining substitution table, ignored\n");
		return;
	}

	CoverageMinusClasses(&pFpst->vsNClass[0], &vGlyphs, &vClass);

	vClass.clear();

	if (iBClassOffset != 0)
		GetClassDefinitionTable(&vClass, pGsubData, iSubTableOffset + iBClassOffset);
	else
		vClass.resize(OTFData.iGlyphCnt);

	INT iBClassCnt = ClassFindCnt(vClass, OTFData.iGlyphCnt);
	pFpst->uBcCnt = (UINT16)iBClassCnt;
	ClassToNames(&pFpst->vsBClass, iBClassCnt, &vClass, OTFData.iGlyphCnt);
	pFpst->vsBClassNames.resize(pFpst->uBcCnt);

	vClass.clear();

	if (iFClassOffset != 0)
		GetClassDefinitionTable(&vClass, pGsubData, iSubTableOffset + iFClassOffset);
	else
		vClass.resize(OTFData.iGlyphCnt);

	INT iFClassCnt = ClassFindCnt(vClass, OTFData.iGlyphCnt);
	pFpst->uFcCnt = (UINT16)iFClassCnt;
	ClassToNames(&pFpst->vsFClass, iFClassCnt, &vClass, OTFData.iGlyphCnt);
	pFpst->vsFClassNames.resize(pFpst->uFcCnt);

	iCnt = 0;
	for (INT i = 0; i < iRCnt; ++i)
	{
		for (INT j = 0; j < vRules[i].iSCnt; ++j)
		{
			auto pRule = new FPSTClassRule();
			pRule->pNClasses = vRules[i].vSubRules[j].vClassIndices;
			pRule->iNCnt = vRules[i].vSubRules[j].iCCnt;
			pRule->pBClasses = vRules[i].vSubRules[j].vBClassIndices;
			pRule->iBCnt = vRules[i].vSubRules[j].iBCnt;
			pRule->pFClasses = vRules[i].vSubRules[j].vFClassIndices;
			pRule->iFCnt = vRules[i].vSubRules[j].iFCnt;
			pRule->iLookupCnt = vRules[i].vSubRules[j].iSCnt;
			pRule->vLookups = vRules[i].vSubRules[j].vSeqLookups;

			for (INT k = 0; k < pRule->iLookupCnt; ++k)
				ProcessSubLookups(pvLookups, pRule->vLookups[k], bGPOS);

			pFpst->vRules[iCnt] = pRule;

			++iCnt;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBChainSubTableType3(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)

	Author:		Greg Watroba
	Created:	31/10/2016
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
				bGPOS - if TRUE we're parsing GPOS data, GSUB in the other case
	Outputs:	-
	Purpose:	Parse GSUB Chain substitution data of subtype 3

********************************************************************************************/

void OTFDataLoader::ParseGSUBChainSubTableType3(const BYTE* pGsubData, INT32 iSubTableOffset,
												Lookup* pLookup, LookupSubTable* pSubTable,
												GsubInUseType eType, std::vector<Lookup>* pvLookups, BOOL bGPOS)
{
	const BYTE* pLookupData = pGsubData + iSubTableOffset;
	// omit type
	ParseUShortFromPos(&pLookupData);

	INT iBCnt = ParseUShortFromPos(&pLookupData);
	std::vector<UINT16> vBCoverage;
	vBCoverage.resize(iBCnt);
	for (INT i = 0; i < iBCnt; ++i)
		vBCoverage[i] = ParseUShortFromPos(&pLookupData);

	INT iGCnt = ParseUShortFromPos(&pLookupData);
	std::vector<UINT16> vCoverage;
	vCoverage.resize(iGCnt);
	for (INT i = 0; i < iGCnt; ++i)
		vCoverage[i] = ParseUShortFromPos(&pLookupData);

	INT iFCnt = ParseUShortFromPos(&pLookupData);
	std::vector<UINT16> vFCoverage;
	vFCoverage.resize(iFCnt);
	for (INT i = 0; i < iFCnt; ++i)
		vFCoverage[i] = ParseUShortFromPos(&pLookupData);

	INT iSCnt = ParseUShortFromPos(&pLookupData);

	std::vector<SeqLookup*> vSeqLookups;
	vSeqLookups.resize(iSCnt);
	BOOL bWarned = FALSE;

	for (INT k = 0; k < iSCnt; ++k)
	{
		vSeqLookups[k] = new SeqLookup();
		vSeqLookups[k]->iSeq = ParseUShortFromPos(&pLookupData);

		if (vSeqLookups[k]->iSeq >= iGCnt && !bWarned)
		{
			TRACEUSER("Greg", "Attempt to apply a lookup to a location out of the range of this contextual\n lookup seq=%d, max=%d\n", vSeqLookups[k]->iSeq, iGCnt - 1);
			bWarned = TRUE;
		}

		vSeqLookups[k]->iLookupIndex = ParseUShortFromPos(&pLookupData);
	}

	if (eType == GIUT_JUST_IN_USE)
	{
		for (INT k = 0; k < iSCnt; ++k)
			delete (vSeqLookups[k]);

		return;
	}

	FeatureParamSubTab* pFpst = new FeatureParamSubTab();
	pFpst->uType = (UINT16)((bGPOS) ? PST_CHAIN_POS : PST_CHAIN_SUB);
	pFpst->uFormat = FPST_COVERAGE;
	
	if (pSubTable) pSubTable->pFPST = pFpst;
	pFpst->vSubTables = { pSubTable };

	pFpst->pNext = FRONTORNULL(OTFData.lstFPST);
	OTFData.lstFPST.push_front(pFpst);

	pFpst->vRules.resize(1);
	pFpst->uRuleCnt = 1;
	auto pRule = new FPSTCoverageRule();
	pRule->iBCnt = iBCnt;
	pRule->psBCovers.resize(iBCnt);
	pRule->vecBCoverGlyphs.resize(iBCnt);

	std::vector<UINT16> vGlyphs;
	for (INT i = 0; i < iBCnt; ++i)
	{
		ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + vBCoverage[i]);
		pRule->psBCovers[i] = GlyphsToNames(&vGlyphs, TRUE);
		pRule->vecBCoverGlyphs[i] = vGlyphs;
		vGlyphs.clear();
	}

	pRule->iNCnt = iGCnt;
	pRule->psNCovers.resize(iGCnt);
	pRule->vecNCoverGlyphs.resize(iGCnt);

	for (INT i = 0; i < iGCnt; ++i)
	{
		ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + vCoverage[i]);
		pRule->psNCovers[i] = GlyphsToNames(&vGlyphs, TRUE);
		pRule->vecNCoverGlyphs[i] = vGlyphs;
		vGlyphs.clear();
	}
	
	pRule->iFCnt = iFCnt;
	pRule->psFCovers.resize(iFCnt);
	pRule->vecFCoverGlyphs.resize(iFCnt);

	for (INT i = 0; i < iFCnt; ++i)
	{
		ParseCoverageTable(&vGlyphs, pGsubData, iSubTableOffset + vFCoverage[i]);
		pRule->psFCovers[i] = GlyphsToNames(&vGlyphs, TRUE);
		pRule->vecFCoverGlyphs[i] = vGlyphs;
		vGlyphs.clear();
	}

	pRule->iLookupCnt = iSCnt;
	pRule->vLookups = vSeqLookups;

	for (INT k = 0; k < iSCnt; ++k)
		ProcessSubLookups(pvLookups, pRule->vLookups[k], bGPOS);

	pFpst->vRules[0] = pRule;
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBExtensionSubTable(const BYTE* pGsubData, INT32 iSubTableOffset,
													Lookup* pLookup, LookupSubTable* pSubTable,
													GsubInUseType eType, std::vector<Lookup>* pvLookups)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
				pvLookups - pointer to Lookup vector
	Outputs:	-
	Purpose:	Parse GSUB Extension substitution data

********************************************************************************************/

void OTFDataLoader::ParseGSUBExtensionSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, 
												Lookup* pLookup, LookupSubTable* pSubTable, 
												GsubInUseType eType, std::vector<Lookup>* pvLookups)
{
	OTLookup* pOTL = BACKORNULL(pLookup->vOTLookups);
	if (!pOTL)
		return;

	//TRACEUSER("Greg", "ParseGSUBExtensionSubTable\n");

	const BYTE* pLookupData = pGsubData + iSubTableOffset;

	/*UINT16 uFormat = */ParseUShortFromPos(&pLookupData);
	INT iType = ParseUShortFromPos(&pLookupData);
	UINT32 uOffset = ParseLongFromPos(&pLookupData);

	pOTL->eLookupType = (OTLookupType)iType;
	INT32 iSubTableExtOffset = iSubTableOffset + uOffset;

	if (iSubTableExtOffset > OTFData.iGsubStart + OTFData.iGsubLength)
	{
		TRACEUSER("Greg", "OTFDataLoader::ParseGSUBExtensionSubTable : Offset cannot go out of GSUB range!\n");
		return;
	}

	switch (iType)
	{
		case 1:
		{
			ParseGSUBSimpleSubTable(pGsubData, iSubTableExtOffset, pLookup, pSubTable, eType);
		}break;

		case 2:
		case 3:
		{
			ParseGSUBMultipleSubTable(pGsubData, iSubTableExtOffset, pLookup, pSubTable, eType);
		}break;

		case 4:
		{
			ParseGSUBLigatureSubTable(pGsubData, iSubTableExtOffset, pLookup, pSubTable, eType);
		}break;

		case 5:
		{
			ParseGSUBContextSubTable(pGsubData, iSubTableExtOffset, pLookup, pSubTable, eType, pvLookups);
		}break;

		case 6:
		{
			ParseGSUBChainSubTable(pGsubData, iSubTableExtOffset, pLookup, pSubTable, eType, pvLookups);
		}break;

		case 7:
		{
			TRACEUSER("Greg", "OTFDataLoader::ParseGSUBExtensionSubTable: Error - font shouldn't have extension table in extension table\n");
			return;
		}break;
			
		case 8:
		{
			ParseGSUBReverseChainSubTable(pGsubData, iSubTableExtOffset, pLookup, pSubTable, eType);
		}break;

		default:
		{
			TRACEUSER("Greg", "OTFDataLoader::ParseGSUBExtensionSubTable: Uknown GSUB SubTable Type: %d\n", iType);
		}break;
	};
}

/********************************************************************************************

>	void OTFDataLoader::ParseGSUBReverseChainSubTable(const BYTE* pGsubData, INT32 iSubTableOffset,
														Lookup* pLookup, LookupSubTable* pSubTable,
														GsubInUseType eType)

	Author:		Greg Watroba
	Created:	10/12/2015
	Inputs:		pPosData - raw GSUB data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				eType - GSUB usage type
	Outputs:	-
	Purpose:	Parse GSUB Ligature substitution data

********************************************************************************************/

void OTFDataLoader::ParseGSUBReverseChainSubTable(const BYTE* pGsubData, INT32 iSubTableOffset, 
													Lookup* pLookup, LookupSubTable* pSubTable, 
													GsubInUseType eType)
{
	if (eType >= 0)
		return;

	/* Reverse chaining is used only in Arabic substitutions since sequences are parsed from right to left
	if (eType == GIUT_FIND_NAMES)
		return;*/

	const BYTE* pLookupData = pGsubData + iSubTableOffset;

	UINT16 uFormat = ParseUShortFromPos(&pLookupData);

	if (uFormat != 1)
		return;

	//TRACEUSER("Greg", "\t\tParseGSUBReverseChainSubTable\n");

	std::vector<UINT16> Coverage;
	std::vector<UINT16> BCoverage;
	std::vector<UINT16> FCoverage;
	std::vector<UINT16> SGlyphs;
	std::vector<UINT16> Glyphs;

	UINT16 uCoverage = ParseUShortFromPos(&pLookupData);

	INT iBCnt = ParseUShortFromPos(&pLookupData);
		BCoverage.resize(iBCnt);
	for (int i = 0; i < iBCnt; ++i)
		BCoverage[i] = ParseUShortFromPos(&pLookupData);
	
	INT iFCnt = ParseUShortFromPos(&pLookupData);
	FCoverage.resize(iFCnt);
	for (int i = 0; i < iFCnt; ++i)
		FCoverage[i] = ParseUShortFromPos(&pLookupData);

	INT iSCnt = ParseUShortFromPos(&pLookupData);
	SGlyphs.resize(iSCnt + 1);
	for (int i = 0; i < iSCnt; ++i)
	{
		SGlyphs[i] = ParseUShortFromPos(&pLookupData);
		if (SGlyphs[i] >= OTFData.iGlyphCnt)
		{
			TRACEUSER("Greg", "Bad reverse contextual chaining substitution glyph: %d is not less than %d\n", SGlyphs[i], OTFData.iGlyphCnt);
			SGlyphs[i] = 0;
		}
	}
	SGlyphs[iSCnt] = 0xffff;

	if (eType == GIUT_JUST_IN_USE)
	{
		for (int i = 0; i < iSCnt; ++i)
			OTFData.vInUse[SGlyphs[i]] = 1;
	}
	else
	{
		FeatureParamSubTab* pFpst = new FeatureParamSubTab();

		pFpst->uType = PST_REVERSE_SUB;
		pFpst->uFormat = FPST_REVERSECOVERAGE;

		if (pSubTable) pSubTable->pFPST = pFpst;
		pFpst->vSubTables = { pSubTable };

		pFpst->pNext = FRONTORNULL(OTFData.lstFPST);
		OTFData.lstFPST.push_front(pFpst);

		auto pRule = new FPSTRevCoverageRule();
		pFpst->uRuleCnt = 1;
		pFpst->vRules.push_back(pRule);

		pRule->iAlwaysL = 1;
		pRule->iBCnt = iBCnt;
		pRule->iFCnt = iFCnt;
		
		pRule->sReplacements = GlyphsToNames(&SGlyphs, FALSE);
		pRule->vecReplacementGlyphs = SGlyphs;

		ParseCoverageTable(&Glyphs, pGsubData, iSubTableOffset + uCoverage);
		pRule->psNCovers.push_back(GlyphsToNames(&Glyphs, FALSE));
		pRule->vecNCoverGlyphs.push_back(Glyphs);
		Glyphs.clear();

		for (int i = 0; i < iBCnt; ++i)
		{
			ParseCoverageTable(&Glyphs, pGsubData, iSubTableOffset + BCoverage[i]);
			pRule->psBCovers.push_back(GlyphsToNames(&Glyphs, TRUE));
			pRule->vecBCoverGlyphs.push_back(Glyphs);
			Glyphs.clear();
		}

		for (int i = 0; i < iFCnt; ++i)
		{
			ParseCoverageTable(&Glyphs, pGsubData, iSubTableOffset + FCoverage[i]);
			pRule->psFCovers.push_back(GlyphsToNames(&Glyphs, TRUE));
			pRule->vecFCoverGlyphs.push_back(Glyphs);
			Glyphs.clear();
		}

		pRule->iLookupCnt = 0;
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseValueRecord(ValueRecord* pValueRecord, INT32 iValueFormat, 
											const BYTE** pLookupData)

	Author:		Greg Watroba
	Created:	30/11/2015
	Inputs:		iValueFormat - value format data to identify values to read
				pLookupData - pointer to data to read from
	Outputs:	pValueRecord - pointer to ValueRecord structure
	Purpose:	Parse Value Record data from raw BYTE array

********************************************************************************************/

void OTFDataLoader::ParseValueRecord(ValueRecord* pValueRecord, INT32& iValueFormat, 
										const BYTE** pLookupData)
{
	memset(pValueRecord,'\0',sizeof(ValueRecord));

	if (iValueFormat & 1)
		pValueRecord->iXPlacement = ParseUShortFromPos(pLookupData);

	if (iValueFormat & 2)
		pValueRecord->iYPlacement = ParseUShortFromPos(pLookupData);

	if (iValueFormat & 4)
		pValueRecord->iXAdvance = ParseUShortFromPos(pLookupData);

	if (iValueFormat & 8)
		pValueRecord->iYAdvance = ParseUShortFromPos(pLookupData);

	if (iValueFormat & 0x10)
		pValueRecord->uOffXPlaceDev = ParseUShortFromPos(pLookupData);

	if (iValueFormat & 0x20)
		pValueRecord->uOffYPlaceDev = ParseUShortFromPos(pLookupData);

	if (iValueFormat & 0x40)
		pValueRecord->uOffXAdvanceDev = ParseUShortFromPos(pLookupData);

	if (iValueFormat & 0x80)
		pValueRecord->uOffYAdvanceDev = ParseUShortFromPos(pLookupData);
}

/********************************************************************************************

>	void OTFDataLoader::ParseCoverageTable(std::vector<UINT16>* pvGlyphs, const BYTE* pPosData, 
											  INT32 iCoverageOffset)

	Author:		Greg Watroba
	Created:	30/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iCoverageOffset - offset for coverage table data
	Outputs:	pvGlyphs - pointer to glyphs vector
	Purpose:	Parse glyph coverage data

********************************************************************************************/

void OTFDataLoader::ParseCoverageTable(std::vector<UINT16>* pvGlyphs, const BYTE* pPosData, 
										  INT32 iCoverageOffset)
{
	const BYTE* pCoverageOffsetPos = pPosData + iCoverageOffset;

	INT32 iCnt = 0;
	INT32 iFormat = ParseUShortFromPos(&pCoverageOffsetPos);

	if (iFormat == 1)
	{
		iCnt = ParseUShortFromPos(&pCoverageOffsetPos);
		pvGlyphs->resize(iCnt + 1);

		if (iCoverageOffset + 2 * iCnt > OTFData.iPositioningSizeBounds)
		{
			TRACEUSER("Greg", "Coverage table extends beyond end of table.\n");
			OTFData.bBadOpenTypeData = TRUE;

			if (iCoverageOffset > OTFData.iPositioningSizeBounds)
			{
				pvGlyphs->clear();
				return;
			}

			iCnt = (OTFData.iPositioningSizeBounds - iCoverageOffset) / 2;
		}

		for (INT32 i = 0; i < iCnt; ++i)
		{
			if (iCnt & 0xffff0000)
			{
				TRACEUSER("Greg", "Bad count.\n");
				OTFData.bBadOpenTypeData = TRUE;
			}

			(*pvGlyphs)[i] = ParseUShortFromPos(&pCoverageOffsetPos);

			if ((*pvGlyphs)[i] >= OTFData.iGlyphCnt)
			{
				TRACEUSER("Greg", "Bad coverage table. Glyph %d out of range [0,%d).\n", 
						 (*pvGlyphs)[i], OTFData.iGlyphCnt);
				(*pvGlyphs)[i] = 0;
			}
		}

	} 
	else if (iFormat == 2) 
	{
		INT32 iMax = 256;
		pvGlyphs->resize(iMax);
		iCnt = 0;

		INT32 iSecCnt = ParseUShortFromPos(&pCoverageOffsetPos);

		if (iCoverageOffset + 6 * iSecCnt > OTFData.iPositioningSizeBounds)
		{
			TRACEUSER("Greg", "Coverage table extends beyond end of table. \n");
			OTFData.bBadOpenTypeData = TRUE;
			iSecCnt = (OTFData.iPositioningSizeBounds - iCoverageOffset) / 6;
		}

		for (INT32 i = 0; i < iSecCnt; ++i)
		{
			INT32 iStart = ParseUShortFromPos(&pCoverageOffsetPos);
			INT32 iEnd = ParseUShortFromPos(&pCoverageOffsetPos);
			INT32 iInd = ParseUShortFromPos(&pCoverageOffsetPos);

			if (iStart > iEnd || iEnd >= OTFData.iGlyphCnt)
			{
				TRACEUSER("Greg", "Bad coverage table. Glyph range %d-%d out of range [0,%d).\n", 
						 iStart, iEnd, OTFData.iGlyphCnt);
				iStart = 0;
				iEnd = 0;
			}

			if (iInd + iEnd - iStart + 2 >= iMax)
			{
				iMax = iInd + iEnd - iStart + 2;
				pvGlyphs->resize(iMax, (UINT16)0);
			}

			for (INT32 j = iStart; j <= iEnd; ++j)
			{
				(*pvGlyphs)[j - iStart + iInd] = (UINT16)j;
				if (j >= OTFData.iGlyphCnt)
					(*pvGlyphs)[j - iStart + iInd] = 0;
			}

			if (iInd + iEnd - iStart + 1 > iCnt)
				iCnt = iInd + iEnd - iStart + 1;
		}
	} 
	else 
	{
		TRACEUSER("Greg", "Bad format for coverage table %d.\n", iFormat);
		pvGlyphs->clear();
		return;
	}

	(*pvGlyphs)[iCnt] = 0xffff;
	pvGlyphs->erase(pvGlyphs->begin() + iCnt + 1, pvGlyphs->end());
	std::sort((*pvGlyphs).begin(), (*pvGlyphs).end());
	auto last = std::unique((*pvGlyphs).begin(), (*pvGlyphs).end());
	pvGlyphs->erase(last, pvGlyphs->end());
}

/********************************************************************************************

>	void OTFDataLoader::ParseGPOSSimplePos(const BYTE* pPosData, INT32 iSubTableOffset, 
											   Lookup* pLookup, LookupSubTable* pSubTable)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
	Outputs:	-
	Purpose:	Parse Single Adjustment data

********************************************************************************************/

void OTFDataLoader::ParseGPOSSimplePos(const BYTE* pPosData, INT32& iSubTableOffset, 
										   Lookup* pLookup, LookupSubTable* pSubTable)
{
	const BYTE* pLookupData = pPosData + iSubTableOffset;

	UINT16 uFormat = ParseUShortFromPos(&pLookupData);
	if (uFormat != 1 && uFormat != 2)
		return;

	INT32 iCoverage = ParseUShortFromPos(&pLookupData);
	INT32 iValueFormat = ParseUShortFromPos(&pLookupData);

	/* Not interested in things whose data just live in device tables */
	if((iValueFormat & 0xf) == 0)
		return;

	ValueRecord sValueRecord;
	std::vector<ValueRecord> vValueRecords;

	if (uFormat == 1)
	{
		ParseValueRecord(&sValueRecord, iValueFormat, &pLookupData);
	}
	else
	{
		INT32 iCnt = ParseUShortFromPos(&pLookupData);
		vValueRecords.resize(iCnt);
		for (INT32 i = 0; i < iCnt; ++i)
			ParseValueRecord(&vValueRecords[i], iValueFormat, &pLookupData);
	}

	std::vector<UINT16> vGlyphs;
	ParseCoverageTable(&vGlyphs, pPosData, iSubTableOffset + iCoverage);

	if (vGlyphs.empty())
	{
		TRACEUSER("Greg", "Bad simple positioning table - ignoring.\n");
		return;
	}

	for (UINT i = 0; i < vGlyphs.size(); ++i)
	{
		if (vGlyphs[i] >= OTFData.iGlyphCnt || 
			vGlyphs[i] >= OTFData.vChars.size() ||
			!OTFData.vChars[vGlyphs[i]] || 
			i >= vValueRecords.size())
			continue;

		PosSubTab* pPos = new PosSubTab();
		pPos->Priority = (UINT16)i;
		pPos->uType = PST_POSITION;
		pPos->pSubtable = pSubTable;
		pPos->pNext = BACKORNULL(OTFData.vChars[vGlyphs[i]]->vecPosSub);
		pPos->GlyphID = (DWORD)vGlyphs[i];

		OTFData.vChars[vGlyphs[i]]->vecPosSub.push_back(pPos);
		
		pPos->Pos.iXOff = (uFormat == 1) ? sValueRecord.iXPlacement : vValueRecords[i].iXPlacement;
		pPos->Pos.iYOff = (uFormat == 1) ? sValueRecord.iYPlacement : vValueRecords[i].iYPlacement;
		pPos->Pos.iHAdvOff = (uFormat == 1) ? sValueRecord.iXAdvance : vValueRecords[i].iXAdvance;
		pPos->Pos.iVAdvOff = (uFormat == 1) ? sValueRecord.iYAdvance : vValueRecords[i].iYAdvance;
	}
	
	if(!pSubTable) 
		pSubTable->bPerGlyphPstOrKern = TRUE;
}

/********************************************************************************************

>	void OTFDataLoader::ParseGPOSKernSubTables(const BYTE* pPosData, INT32 iSubTableOffset, 
												   Lookup* pLookup, LookupSubTable* pSubTable)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
	Outputs:	-
	Purpose:	Parse Pair Adjustment data

********************************************************************************************/

void OTFDataLoader::ParseGPOSKernSubTables(const BYTE* pPosData, INT32& iSubTableOffset, 
											   Lookup* pLookup, LookupSubTable* pSubTable)
{
	FORMATTING_PERFORMANCE_FUNC_START;

	if (!pSubTable)
	{
		TRACEUSER("Greg", "OTFDataLoader::ParseGPOSKernSubTables : pSubTable cannot be NULL!");
		return;
	}

	const BYTE* pLookupData = pPosData + iSubTableOffset;

	UINT16 uFormat = ParseUShortFromPos(&pLookupData);
	if (uFormat != 1 && uFormat != 2)
		return;

	INT32 iCoverage = ParseUShortFromPos(&pLookupData);
	INT32 iValueFormat1 = ParseUShortFromPos(&pLookupData);
	INT32 iValueFormat2 = ParseUShortFromPos(&pLookupData);
	INT32 iRight2Left = 0;
	INT32 iIsVertical = 0;

	std::vector<UINT16> vOffsets;
	KernClass *pKC = nullptr;

	//Accept forms both with and without device tables
	if ((iValueFormat1 == 0x0008 || iValueFormat1 == 0x0088) && iValueFormat2 == 0x0000)
	{
		//Top to bottom
		iIsVertical = 1;
	}
	else if ( iValueFormat1==0x0000 && (iValueFormat2==0x0004 || iValueFormat2==0x0044) && (pLookup->iFlags & PST_FLAG_R2L)) 
	{
		//Right to left
		iIsVertical = 0;
		iRight2Left = 1;
	} 
	else if ((iValueFormat1 == 0x0004 || iValueFormat1 == 0x0044) && iValueFormat2 == 0x0000 && !(pLookup->iFlags & PST_FLAG_R2L))
	{
		//Left to right
		iIsVertical = 0;
	}
	else
	{
		//Can't optimize, store all 8 settings
		iIsVertical = 2;
	}

	if (uFormat == 1 && pSubTable)
	{
		pSubTable->bPerGlyphPstOrKern = TRUE;
		INT32 iCnt = ParseUShortFromPos(&pLookupData);
		vOffsets.resize(iCnt);

		for (INT32 i = 0; i < iCnt; ++i)
			vOffsets[i] = ParseUShortFromPos(&pLookupData);

		std::vector<UINT16> vGlyphs;
		ParseCoverageTable(&vGlyphs, pPosData, iSubTableOffset + iCoverage);

		if (vGlyphs.empty())
		{
			TRACEUSER("Greg", "Bad pairwise kerning table, ignored.\n");
			return;
		}

		if (OTFData.iGlyphCnt > 0)
		{
			for (INT32 i = 0; i < iCnt; ++i)
			{
				if ((size_t)i < vGlyphs.size() && vGlyphs[i] < OTFData.iGlyphCnt)
				{
					const BYTE* pKernPos = pPosData + iSubTableOffset + vOffsets[i];
					INT32 iPairCnt = ParseUShortFromPos(&pKernPos);

					for (INT32 j = 0; j < iPairCnt; ++j)
					{
						UINT16 iGlyph = ParseUShortFromPos(&pKernPos);
						ValueRecord pValRec1, pValRec2;
						ParseValueRecord(&pValRec1, iValueFormat1, &pKernPos);
						ParseValueRecord(&pValRec2, iValueFormat2, &pKernPos);

						if (iIsVertical == 2)
						{
							AddPairPos(vGlyphs[i], iGlyph, pLookup, pSubTable, &pValRec1, &pValRec2);
						}
						else if (iIsVertical)
						{
							if (AddKernPair(vGlyphs[i], iGlyph, pValRec1.iYAdvance, pLookup, pSubTable, iIsVertical))
								AddPairPos(vGlyphs[i], iGlyph, pLookup, pSubTable, &pValRec1, &pValRec2);
						}
						/*
							If we've already got kern data for this pair of
							glyphs, then we can't make it be a true KernPair
							but we can save the info as a PST_PAIR
							*/
						else if (iRight2Left)
						{
							if (AddKernPair(vGlyphs[i], iGlyph, pValRec2.iXAdvance, pLookup, pSubTable, iIsVertical))
								AddPairPos(vGlyphs[i], iGlyph, pLookup, pSubTable, &pValRec1, &pValRec2);
						}
						else
						{
							if (AddKernPair(vGlyphs[i], iGlyph, pValRec1.iXAdvance, pLookup, pSubTable, iIsVertical))
								AddPairPos(vGlyphs[i], iGlyph, pLookup, pSubTable, &pValRec1, &pValRec2);
						}

					}
				}
			}
		}
	} 
	/* Class-based kerning */
	else if (uFormat == 2)
	{
		INT32 iClassDefOffset1 = ParseUShortFromPos(&pLookupData);
		INT32 iClassDefOffset2 = ParseUShortFromPos(&pLookupData);

		std::vector<UINT16> vClass1;
		GetClassDefinitionTable(&vClass1, pPosData, iSubTableOffset + iClassDefOffset1);

		std::vector<UINT16> vClass2;
		GetClassDefinitionTable(&vClass2, pPosData, iSubTableOffset + iClassDefOffset2);

		std::vector<UINT16> vGlyphs;
		ParseCoverageTable(&vGlyphs, pPosData, iSubTableOffset + iCoverage);
		if (vGlyphs.empty())
		{
			TRACEUSER("Greg", "Bad kerning class table, ignored.\n");
			return;
		}

		INT32 iClassCnt1 = ParseUShortFromPos(&pLookupData);
		INT32 iClassCnt2 = ParseUShortFromPos(&pLookupData);

		if (iIsVertical != 2)
		{
			if (iIsVertical)
			{
				if (OTFData.vVerticalKernClasses.empty())
				{
					pKC = new KernClass();
					OTFData.vVerticalKernClasses.push_back(pKC);
				}
				else
				{
					pKC = new KernClass();
					KernClass *pLastKC = OTFData.vVerticalKernClasses.back();
					OTFData.vVerticalKernClasses.push_back(pKC);
					pLastKC->pNext = OTFData.vVerticalKernClasses.back();
				}
			}
			else
			{
				if (OTFData.vKernClasses.empty())
				{
					pKC = new KernClass();
					OTFData.vKernClasses.push_back(pKC);
				}
				else
				{
					pKC = new KernClass();
					KernClass *pLastKC = OTFData.vKernClasses.back();
					OTFData.vKernClasses.push_back(pKC);
					pLastKC->pNext = OTFData.vKernClasses.back();
				}
			}

			if (pSubTable)
			{
				pSubTable->uVerticalKerning = iIsVertical;
				pSubTable->pKernClass = pKC;
				pKC->iFirstCnt = iClassCnt1;
				pKC->iSecondCnt = iClassCnt2;
				pKC->pSubTable = pSubTable;
				pKC->vecKernOffsets.resize(iClassCnt1 * iClassCnt2);
			}

			ClassToNames(&pKC->vecKernFirsts, iClassCnt1, &vClass1, OTFData.iGlyphCnt);
			ClassToNames(&pKC->vecKernSeconds, iClassCnt2, &vClass2, OTFData.iGlyphCnt);

			/* Now if the coverage table contains entries which are not in */
			/*  the list of first classes, then those glyphs are the real */
			/*  values for kc->firsts[0] */

			CoverageMinusClasses(&pKC->vecKernFirsts[0], &vGlyphs, &vClass1);

			for (INT32 i = 0; i < iClassCnt1; ++i)
			{
				for (INT32 j = 0; j < iClassCnt2; ++j)
				{
					ValueRecord pValRec1, pValRec2;
					ParseValueRecord(&pValRec1, iValueFormat1, &pLookupData);
					ParseValueRecord(&pValRec2, iValueFormat2, &pLookupData);

					if (iIsVertical)
						pKC->vecKernOffsets[i * iClassCnt2 + j] = pValRec1.iYAdvance;
					else if (iRight2Left)
						pKC->vecKernOffsets[i * iClassCnt2 + j] = pValRec2.iXAdvance;
					else
						pKC->vecKernOffsets[i * iClassCnt2 + j] = pValRec1.iXAdvance;
				}
			}
		}
		/* 
			This happens when we have a feature which is neither 'kern' nor 'vkrn' 
			we don't know what to do with it so we make it into kern pairs 
		*/
		else if (pSubTable)
		{
			pSubTable->bPerGlyphPstOrKern = TRUE;

			for (INT32 i = 0; i < iClassCnt1; ++i)
			{
				for (INT32 j = 0; j < iClassCnt2; ++j)
				{
					ValueRecord pValRec1, pValRec2;
					ParseValueRecord(&pValRec1, iValueFormat1, &pLookupData);
					ParseValueRecord(&pValRec2, iValueFormat2, &pLookupData);

					BOOL bVR1ParamsCorrect = pValRec1.iXAdvance != 0 || pValRec1.iXPlacement != 0 || 
											 pValRec1.iYAdvance != 0 || pValRec1.iYPlacement != 0;
					BOOL bVR2ParamsCorrect = pValRec2.iXAdvance != 0 || pValRec2.iXPlacement != 0 || 
											 pValRec2.iYAdvance != 0 || pValRec2.iYPlacement != 0;

					if (bVR1ParamsCorrect || bVR2ParamsCorrect)
					{
						for (UINT16 k = 0; k < OTFData.iGlyphCnt; ++k)
						{
							if (vClass1[k] == i)
							{
								for (UINT16 m = 0; m < OTFData.iGlyphCnt; ++m)
								{
									if (vClass2[m] == j)
									{
										AddPairPos(k, m, pLookup, pSubTable, &pValRec1, &pValRec2);
									}
								}
							}
						}
					}
				}
			}
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ParseGPOSCursiveSubTable(const BYTE* pPosData, INT32 iSubTableOffset, 
													 Lookup* pLookup, LookupSubTable* pSubTable)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
	Outputs:	-
	Purpose:	Parse Cursive Adjustment data

********************************************************************************************/

void OTFDataLoader::ParseGPOSCursiveSubTable(const BYTE* pPosData, INT32& iSubTableOffset, 
												 Lookup* pLookup, LookupSubTable* pSubTable)
{
	//TODO: implement when necessary ParseGPOSCursiveSubTable (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseGPOSMarkSubTable(const BYTE* pPosData, INT32 iSubTableOffset, 
												  Lookup* pLookup, LookupSubTable* pSubTable)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
	Outputs:	-
	Purpose:	Parse Mark-To-Base, Mark-To-Ligature and Mark-To-Mark Attachment data

********************************************************************************************/

void OTFDataLoader::ParseGPOSMarkSubTable(const BYTE* pPosData, INT32& iSubTableOffset, 
											  Lookup* pLookup, LookupSubTable* pSubTable)
{
	//TODO: implement when necessary ParseGPOSMarkSubTable (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseGPOSContextSubTable(const BYTE* pPosData, INT32 iSubTableOffset, 
													 Lookup* pLookup, LookupSubTable* pSubTable,
													 std::vector<Lookup>* pvLookups)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				pvLookups - pointer to Lookup vector
	Outputs:	-
	Purpose:	Parse Contextual Positioning data

********************************************************************************************/

void OTFDataLoader::ParseGPOSContextSubTable(const BYTE* pPosData, INT32& iSubTableOffset, 
												 Lookup* pLookup, LookupSubTable* pSubTable,
												 std::vector<Lookup>* pvLookups)
{
	//TODO: implement when necessary ParseGPOSContextSubTable  (probably not)
}

/********************************************************************************************

>	void OTFDataLoader::ParseGPOSChainingSubTable(const BYTE* pPosData, INT32 iSubTableOffset, 
													  Lookup* pLookup, LookupSubTable* pSubTable,
													  std::vector<Lookup>* pvLookups)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				pvLookups - pointer to Lookup vector
	Outputs:	-
	Purpose:	Parse Chained Contextual Positioning data

********************************************************************************************/

void OTFDataLoader::ParseGPOSChainingSubTable(const BYTE* pPosData, INT32& iSubTableOffset, 
												  Lookup* pLookup, LookupSubTable* pSubTable,
												  std::vector<Lookup>* pvLookups)
{
	//TODO: implement when necessary ParseGPOSChainingSubTable (probably not)
}

/********************************************************************************************

>	OTFDataLoader::ParseGPOSExtensionSubTable(const BYTE* pPosData, INT32 iSubTableOffset, 
												  Lookup* pLookup, LookupSubTable* pSubTable,
												  std::vector<Lookup>* pvLookups)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pPosData - raw GPOS data array to parse
				iSubTableOffset - offset for sub table data
				pLookup - pointer to Lookup data
				pSubTable - pointer to SubTables vector
				pvLookups - pointer to Lookup vector
	Outputs:	-
	Purpose:	Parse Extension Positioning data

********************************************************************************************/


void OTFDataLoader::ParseGPOSExtensionSubTable(const BYTE* pPosData, INT32& iSubTableOffset, 
												   Lookup* pLookup, LookupSubTable* pSubTable,
												   std::vector<Lookup>* pvLookups)
{
	OTLookup* pOTL = BACKORNULL(pLookup->vOTLookups);
	if (!pOTL)
		return;

	const BYTE* pLookupData = pPosData + iSubTableOffset;

	/*UINT16 uFormat = */ParseUShortFromPos(&pLookupData);
	INT iType = ParseUShortFromPos(&pLookupData);
	UINT32 uOffset = ParseLongFromPos(&pLookupData);

	pOTL->eLookupType = (OTLookupType)iType;
	INT32 iSubTableExtOffset = iSubTableOffset + uOffset;

	if (iSubTableExtOffset > OTFData.iGposStart + OTFData.iGposLength)
	{
		TRACEUSER("Greg", "OTFDataLoader::ParseGSUBExtensionSubTable : Offset cannot go out of GSUB range!\n");
		return;
	}

	switch (iType)
	{
		case 1:
		{
			ParseGPOSSimplePos(pPosData, iSubTableExtOffset, pLookup, pSubTable);
		}break;

		case 2:
		{
			ParseGPOSKernSubTables(pPosData, iSubTableExtOffset, pLookup, pSubTable);
		}break;

		case 3:
		{
			ParseGPOSCursiveSubTable(pPosData, iSubTableExtOffset, pLookup, pSubTable);
		}break;

		case 4:
		case 5:
		case 6:
		{
			ParseGPOSMarkSubTable(pPosData, iSubTableExtOffset, pLookup, pSubTable);
		}break;

		case 7:
		{
			ParseGPOSContextSubTable(pPosData, iSubTableExtOffset, pLookup, pSubTable, pvLookups);
		}break;
			
		case 8:
		{
			ParseGPOSChainingSubTable(pPosData, iSubTableExtOffset, pLookup, pSubTable, pvLookups);
		}break;

		case 9:
		{
			TRACEUSER("Greg", "OTFDataLoader::ParseGPOSExtensionSubTable: Error - font shouldn't have extension table in extension table\n");
			return;
		}break;

		default:
		{
			TRACEUSER("Greg", "OTFDataLoader::ParseGPOSExtensionSubTable: Uknown GPOS SubTable Type: %d\n", iType);
		}break;
	};
}

/********************************************************************************************

>	INT OTFDataLoader::DefaultLangTagInOneScriptList(ScriptLangList* pSLL)

	Author:		Greg Watroba
	Created:	30/09/2016
	Inputs:		pSLL - script lang list
	Return:		TRUE if tag is default, FALSE otherwise
	Outputs:	-
	Purpose:	Determine, whether Script Language List has default tag

********************************************************************************************/

BOOL OTFDataLoader::DefaultLangTagInOneScriptList(ScriptLangList* pSLL)
{
	if (!pSLL)
		return FALSE;

	for (INT i = 0; i < pSLL->iLangCnt; ++i)
	{
		UINT32 uLang = i < MAX_LANG ? pSLL->uLangs[i] : pSLL->vMoreLangs[i - MAX_LANG];
		if (uLang == DEFAULT_LANG)
			return TRUE;
	}

	return FALSE;
}

/********************************************************************************************

>	ScriptLangList* OTFDataLoader::DefaultLangTagInScriptList(std::vector<ScriptLangList*>& pvScripts, 
															  BOOL iDefault)

	Author:		Greg Watroba
	Created:	30/09/2016
	Inputs:		pvScripts - script lang list
				bDefault - do we look for default tag?
	Return:		Script language list
	Outputs:	-
	Purpose:	Get language list for lookup table

********************************************************************************************/

ScriptLangList* OTFDataLoader::DefaultLangTagInScriptList(std::vector<ScriptLangList*>& pvScripts, 
														  BOOL bDefault)
{
	for (auto& pScript : pvScripts)
	{
		if (bDefault || pScript->uScript != DEFAULT_SCRIPT)
		{
			if (DefaultLangTagInOneScriptList(pScript))
				return pScript;
		}
	}

	return nullptr;
}

/********************************************************************************************

>	void OTFDataLoader::TagFullName(std::string* pStrFullName, GlyphSetEnum* pTagEnum,
									UINT32 uTag, BOOL bIsMac, BOOL bOnlyIfKnown)

	Author:		Greg Watroba
	Created:	30/09/2016
	Inputs:		uTag - tag which name we want to get
				bIsMac - flag to determine whether we have mac font
				bOnlyIfKnown - return name only if it's know
	Outputs:	pStrFullName - full tag name
				pTagEnum - tag's glyph set enum
	Return:		-
	Outputs:	-
	Purpose:	Get tag's full name and glyph set indicator

********************************************************************************************/

void OTFDataLoader::TagFullName(std::string* pStrFullName, GlyphSetEnum* pTagEnum,
								UINT32 uTag, UINT32 uIsMac, BOOL bOnlyIfKnown)
{
	if (!pStrFullName || !pTagEnum)
		return;

	pStrFullName->clear();
	
	std::string sTagFullName;

	if (uIsMac == -1)
		uIsMac = (uTag >> 24) < ' ' || (uTag >> 24) > 0x7e;

	// for now we will support only non-mac fonts
	uIsMac = 0;

	if (uIsMac)
	{
		sTagFullName.append("<");
		sTagFullName.append(std::to_string(uTag >> 16));
		sTagFullName.append(",");
		sTagFullName.append(std::to_string(uTag & 0xffff));
		sTagFullName.append(">");
		/*
		Mac tagging
		FindMacSettingName(sf,tag>>16,tag&0xffff);
		std::string SetName = PickNameFromMacName();
		*/
	}
	else
	{
		UINT sTag = uTag;

		if (uTag == CHR('n', 'u', 't', 'f'))	/* early name that was standardize later as... */
			sTag = CHR('a', 'f', 'r', 'c');	/*  Stood for nut fractions. "nut" meaning "fits in an en" in old typography-speak => vertical fractions rather than diagonal ones */

		if (uTag == REQUIRED_FEATURE)
		{
			sTagFullName = "Required Feature";
		}
		else
		{
			INT k = 0;
			for (k = 0; s_FriendlyNames[k].iTag != 0; ++k)
			{
				if (s_FriendlyNames[k].iTag == sTag)
					break;
			}

			sTagFullName.clear();
			sTagFullName.resize(7);
			sTagFullName[0] = '\'';
			sTagFullName[1] = uTag >> 24;
			sTagFullName[2] = (uTag >> 16) & 0xff;
			sTagFullName[3] = (uTag >> 8) & 0xff;
			sTagFullName[4] = uTag & 0xff;
			sTagFullName[5] = '\'';
			sTagFullName[6] = ' ';

			if (s_FriendlyNames[k].iTag != 0)
			{
				sTagFullName.append(s_FriendlyNames[k].sFriendlyName);
				(*pTagEnum) = s_FriendlyNames[k].eGlyphSet;
			}
			else if (bOnlyIfKnown)
				return;
		}
	}

	(*pStrFullName) = sTagFullName;
}

/********************************************************************************************

>	std::string OTFDataLoader::SuffixFromTags(std::vector<FeatureScriptLangList*>* pvFeatureList)

	Author:		Greg Watroba
	Created:	30/09/2016
	Inputs:		pvFeatureList - feature script lang list container
	Return:		Suffix of given tag
	Outputs:	-
	Purpose:	Get suffix for subtable

********************************************************************************************/

std::string OTFDataLoader::SuffixFromTags(std::vector<FeatureScriptLangList*>* pvFeatureList)
{
	if (!pvFeatureList)
		return "";

	for (auto& pFSLL : (*pvFeatureList))
	{
		for (UINT32 i = 0; i < s_Tags2Suffix[i].uTag != 0; ++i)
		{
			if (s_Tags2Suffix[i].uTag == pFSLL->uFeatureTag)
				return s_Tags2Suffix[i].sSuffix;
		}
	}
	
	return "";
}

/********************************************************************************************

>	void OTFDataLoader::NameOTLookup(std::vector<OTLookup*>* pvOTLookups)

	Author:		Greg Watroba
	Created:	27/11/2015
	Inputs:		pvOTLookups - pointer to OTLookup vector
	Outputs:	-
	Purpose:	Set names for parsed lookups

********************************************************************************************/

void OTFDataLoader::NameOTLookup(std::vector<OTLookup*>* pvOTLookups)
{
	OTLookup* pOTL = BACKORNULL((*pvOTLookups));

	if (!pOTL)
		return;

	std::string sUserFriendly, sLookupType, sScript, sFormat;
	GlyphSetEnum eTagEnum = GSET_NONE;

	if (pOTL->sLookupName.empty())
	{
		FeatureScriptLangList* pFSLL = nullptr;

		for (int k = 0; k < 2; ++k)
		{
			for (auto& pFeature : pOTL->vFeatures)
			{
				if (k == 1 || DefaultLangTagInScriptList(pFeature->vScripts, FALSE))
				{
					TagFullName(&sUserFriendly, &eTagEnum, pFeature->uFeatureTag, pFeature->uIsMac, TRUE);
				}
			}

			if (!sUserFriendly.empty())
				break;
		}

		if (sUserFriendly.empty())
		{
			if ((pOTL->eLookupType & 0xff) >= 0xf0)
				sLookupType = "State Machine";
			else if ((pOTL->eLookupType >> 8) < 2 && (pOTL->eLookupType & 0xff) < 10)
				sLookupType = s_LookupTypeNames[pOTL->eLookupType >> 8][pOTL->eLookupType & 0xff];
			else
				sLookupType = "LookupType|Unknown";

			for (FeatureScriptLangList* pFeatureScript : pOTL->vFeatures)
			{
				if (pFeatureScript->uIsMac)
					pFSLL = pFeatureScript;
			}

			if (pFSLL == nullptr)
			{
				sUserFriendly = sLookupType;
			}
			else
			{
				sUserFriendly.clear();
				sUserFriendly.append(sLookupType);
				sUserFriendly.append(" ");
				sUserFriendly.append(std::to_string(pFSLL->uFeatureTag >> 24));
				sUserFriendly.append(std::to_string(pFSLL->uFeatureTag >> 16));
				sUserFriendly.append(std::to_string(pFSLL->uFeatureTag >> 8));
				sUserFriendly.append(std::to_string(pFSLL->uFeatureTag));
			}
		}

		sScript.clear();
		if (pFSLL == nullptr)
			pFSLL = BACKORNULL(pOTL->vFeatures);

		ScriptLangList* pSLL = nullptr;
		if (pFSLL)
			pSLL = BACKORNULL(pFSLL->vScripts);

		ScriptLangList* pFound = nullptr;
		ScriptLangList* pFound2 = nullptr;
		auto pTemp = std::make_unique<ScriptLangList>(-1);

		if (pFSLL && pSLL)
		{
			UINT32 uScriptTag = pSLL->uScript;

			for (auto& pScriptLang : pFSLL->vScripts)
			{
				if (pScriptLang->uScript == DEFAULT_SCRIPT)
				{
				}
				else if (DefaultLangTagInOneScriptList(pScriptLang))
				{
					if (pFound == nullptr)
					{
						pFound = pScriptLang;
					}
					else
					{
						pFound = pFound2 = nullptr;
						break;
					}
				}
				else if (pFound2 == nullptr)
				{
					pFound2 = pScriptLang;
				}
				else
				{
					pFound2 = pTemp.get();
				}
			}

			if (!pFound && pFound2 && pFound2->iLangCnt != -1)
				pFound = pFound2;

			if (pFound)
			{
				uScriptTag = pFound->uScript;

				INT j = 0;
				for (j = 0; !s_LocalScripts[j].sText.empty() && uScriptTag != s_LocalScripts[j].uTag; ++j);

				if (!s_LocalScripts[j].sText.empty())
				{
					sScript = s_LocalScripts[j].sText;
				}
				else
				{
					sScript.append("\\");
					sScript.append(std::to_string(pSLL->uScript >> 24));
					sScript.append(std::to_string( (pSLL->uScript >> 16) & 0xff ));
					sScript.append(std::to_string( (pSLL->uScript >> 8) & 0xff ));
					sScript.append(std::to_string( pSLL->uScript & 0xff ));
					sScript.append("\\");
				}
			}
		}

		if (!sScript.empty())
		{
			pOTL->sLookupName.append(sUserFriendly);
			pOTL->sLookupName.append(" in ");
			pOTL->sLookupName.append(sScript);
			pOTL->sLookupName.append(" lookup ");
			pOTL->sLookupName.append(std::to_string(pOTL->iLookupIndex));
		}
		else
		{
			pOTL->sLookupName.append(sUserFriendly);
			pOTL->sLookupName.append(" lookup ");
			pOTL->sLookupName.append(std::to_string(pOTL->iLookupIndex));
		}
	}

	if (eTagEnum == GSET_NONE)
	{
		for (auto& pFeature : pOTL->vFeatures)
		{
			if (pFeature->eGlyphSet != GSET_NONE)
				eTagEnum = pFeature->eGlyphSet;
		}
	}

	if (pOTL->vSubTables.empty())
	{
		TRACEUSER("Greg", "No subtables to name setup\n");
	}
	else
	{
		INT iCnt = 0;
		for (auto& pSubTable : pOTL->vSubTables)
		{
			if (pSubTable->eGlyphSet == GSET_NONE)
				pSubTable->eGlyphSet = eTagEnum;

			if (pSubTable->sSubtableName.empty())
			{
				if (pSubTable == pOTL->vSubTables.front() && !pSubTable->pNext)
					sFormat = " subtable ";
				else if (pSubTable->bPerGlyphPstOrKern)
					sFormat = " per glyph data ";
				else if (pSubTable->pKernClass)
					sFormat = " kerning class ";
				else if (pSubTable->pFPST)
					sFormat = " contextual ";
				else if (pSubTable->uAnchorClasses)
					sFormat = " anchor ";
				else if (eTagEnum == GSET_NONE) 
				{
					//this means we don't know subtable and the glyphs set type
					//TRACEUSER("Greg", "Subtable naming error in lookup: %s \n", pOTL->sLookupName.c_str());
					sFormat = " !!!!!!!!!! ";
				}
				else
				{
					//this means we don't know subtable but we know the glyph set type
					sFormat = " !!!!!!!!!! ";
				}

				pSubTable->sSubtableName.append(pOTL->sLookupName);
				pSubTable->sSubtableName.append(sFormat);
				pSubTable->sSubtableName.append(std::to_string(iCnt));
			}

			++iCnt;
		}
	}

	if (pOTL->eLookupType == OTL_TYPE_GSUB_LIGATURE)
	{
		for (auto& pFeature : pOTL->vFeatures)
		{
			if (pFeature->uFeatureTag == CHR('l', 'i', 'g', 'a') || pFeature->uFeatureTag == CHR('r', 'l', 'i', 'g'))
				pOTL->uStoreInAfm = 1;
		}
	}

	if (pOTL->eLookupType == OTL_TYPE_GSUB_SINGLE)
	{
		for (auto& pSubTable : pOTL->vSubTables)
			pSubTable->sSuffix = SuffixFromTags(&pOTL->vFeatures);
	}
}

/********************************************************************************************

>	void SwapBytePairs(BYTE* pDestBytes, const BYTE* pSrcBytes, int iBytesToSwap)

	Author: 	Marc, shamelessly stolen by Greg W.
	Created:	28/09/15
	Inputs:		pSrcBytes - the source bytes
				iBytesToSwap - how many bytes to swap
	Outputs:	pDestBytes - the destination buffer
	Returns:	The bytes of the font table as a unique_ptr, null_ptr if failed
	Purpose:	Copies swapped byte pairs to the destination array. Used for
				converting big -> little endian wide char character array.
	Notes:		The caller is responsible for ensuring both bufferss are the
				required length, to avoid overruns.

********************************************************************************************/

void OTFDataLoader::SwapBytePairs(BYTE* pDestBytes, const BYTE* pSrcBytes, int& iBytesToSwap)
{
	if (pDestBytes && pSrcBytes && iBytesToSwap)
	{
		for (int i = 0; i < iBytesToSwap; i = i + 2)
		{
			// Copy and swap the byte order at the same time
			pDestBytes[i] = pSrcBytes[i + 1];
			pDestBytes[i + 1] = pSrcBytes[i];
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::AddPairPos(UINT16 &uGlyph1, UINT16 &uGlyph2, Lookup* pLookup, 
								   LookupSubTable* pSubTable, ValueRecord* pValRec1, 
								   ValueRecord* pValRec2)

	Author: 	Greg Watroba
	Created:	1/12/15
	Inputs:		uGlyph1 - first glyph
				uGlyph2 - second glyph
				pLookup - pointer to specific Lookup
				pSubTable - pointer to specific LookupSubTable
				pValRec1 - first ValueRecord
				pValRec2 - second ValueRecord
	Outputs:	Result is set into OTFDataLoader::OTFData
	Purpose:	Sets pair position between two glyphs

********************************************************************************************/

void OTFDataLoader::AddPairPos(UINT16& uGlyph1, UINT16& uGlyph2, Lookup* pLookup, 
								   LookupSubTable* pSubTable, ValueRecord* pValRec1, 
								   ValueRecord* pValRec2)
{
	if (uGlyph1 >= OTFData.iGlyphCnt || uGlyph2 >= OTFData.iGlyphCnt || 
		!OTFData.vChars[uGlyph1] || !OTFData.vChars[uGlyph2])
	{
		TRACEUSER("Greg", "Bad pair position: glyphs %d & %d should have been < %d.\n", 
				 uGlyph1, uGlyph2, OTFData.iGlyphCnt);
		return;
	}

	PosSubTab *pPosSubTab = new PosSubTab();
	pPosSubTab->uType = PST_PAIR;
	pPosSubTab->pSubtable = pSubTable;
	pPosSubTab->Priority = (UINT16)((pSubTable && pSubTable->pOTLookup) ? pSubTable->pOTLookup->iLookupIndex : 0);
	pPosSubTab->pNext = BACKORNULL(OTFData.vChars[uGlyph1]->vecPosSub);
	OTFData.vChars[uGlyph1]->vecPosSub.push_back(pPosSubTab);
	
	pPosSubTab->PstPair.sPaired = OTFData.vChars[uGlyph2]->sName;

	pPosSubTab->PstPair.pVR[0].iXOff = pValRec1->iXPlacement;
	pPosSubTab->PstPair.pVR[0].iYOff = pValRec1->iYPlacement;
	pPosSubTab->PstPair.pVR[0].iHAdvOff = pValRec1->iXAdvance;
	pPosSubTab->PstPair.pVR[0].iVAdvOff = pValRec1->iYAdvance;
	pPosSubTab->PstPair.pVR[1].iXOff = pValRec2->iXPlacement;
	pPosSubTab->PstPair.pVR[1].iYOff = pValRec2->iYPlacement;
	pPosSubTab->PstPair.pVR[1].iHAdvOff = pValRec2->iXAdvance;
	pPosSubTab->PstPair.pVR[1].iVAdvOff = pValRec2->iYAdvance;
}

/********************************************************************************************

>	BOOL OTFDataLoader::AddKernPair(UINT16 &uGlyph1, UINT16 &uGlyph2, INT16& iOffset, 
									Lookup* pLookup, LookupSubTable* pSubTable, 
									INT32 iIsVertical)

	Author: 	Greg Watroba
	Created:	1/12/15
	Inputs:		uGlyph1 - first glyph
				uGlyph2 - second glyph
				iOffset - kerning offset
				pLookup - pointer to specific Lookup
				pSubTable - pointer to specific LookupSubTable
				iIsVertical - checkflag for vertical positioning
	Outputs:	Result is set into OTFDataLoader::OTFData
	Returns:    TRUE if kerning pair has been added successfully
	Purpose:	Sets pair kerning between two glyphs

********************************************************************************************/

BOOL OTFDataLoader::AddKernPair(UINT16& uGlyph1, UINT16& uGlyph2, INT16& iOffset, 
									Lookup* pLookup, LookupSubTable* pSubTable, 
									INT32& iIsVertical)
{
	if (uGlyph1 >= OTFData.iGlyphCnt || uGlyph2 >= OTFData.iGlyphCnt ||
		uGlyph1 >= OTFData.vChars.size() || uGlyph2 >= OTFData.vChars.size())
	{
		TRACEUSER("Greg", "Bad kern pair: glyphs %d & %d should have been < %d.\n", 
				 uGlyph1, uGlyph2, OTFData.iGlyphCnt);
		return FALSE;
	}

	if (!OTFData.vChars[uGlyph1] || !OTFData.vChars[uGlyph2])
		return FALSE;

	auto vecToSearch = (iIsVertical) ? &OTFData.vChars[uGlyph1]->vecVerticalKernPairs : &OTFData.vChars[uGlyph1]->vecKernPairs;
	auto pCharToSearch = OTFData.vChars[uGlyph2];
	auto ret = std::find_if(vecToSearch->begin(), vecToSearch->end(), [pCharToSearch](const KernPair& pair)
			   {
					return pair.pSC == pCharToSearch;
			   });

	if (ret == vecToSearch->end())
	{
		KernPair pKernPair;
		pKernPair.pSC = OTFData.vChars[uGlyph2];
		pKernPair.iOff = iOffset;
		pKernPair.pSubTable = pSubTable;
		vecToSearch->push_back(pKernPair);
	}
	else if ((*ret).pSubTable != pSubTable)
	{
		return TRUE;
	}

	return FALSE;
}

/********************************************************************************************

>	void OTFDataLoader::GetClassDefinitionTable(std::vector<UINT16> *pvClass, const BYTE* pPosData, 
												INT32 iClassDefOffset)

	Author: 	Greg Watroba
	Created:	1/12/15
	Inputs:		pvClass - pointer to vector of class values
				pPosData - raw GPOS data array to parse
				iClassDefOffset - class definition offset
	Outputs:	Vector of class definition values
	Returns:    -
	Purpose:	Fills class definition values

********************************************************************************************/

void OTFDataLoader::GetClassDefinitionTable(std::vector<UINT16> *pvClass, const BYTE* pPosData, 
												INT32 iClassDefOffset)
{
	//Class 0 is default
	const BYTE* pClassDefinitionPos = pPosData + iClassDefOffset;
	INT32 iFormat = ParseUShortFromPos(&pClassDefinitionPos);

	pvClass->resize(OTFData.iGlyphCnt);

	if (iFormat == 1)
	{
		INT32 iStart = ParseUShortFromPos(&pClassDefinitionPos);
		INT32 iGlyphCnt = ParseUShortFromPos(&pClassDefinitionPos);
		ptrdiff_t iDiff = pClassDefinitionPos - pPosData;

		if (iStart + iGlyphCnt > OTFData.iGlyphCnt)
		{
			TRACEUSER("Greg", "Bad class def table. start=%d OTFData.iGlyphCnt=%d, max glyph=%d\n", 
					 iStart, iGlyphCnt, OTFData.iGlyphCnt);
			OTFData.bBadOpenTypeData = TRUE;

			iGlyphCnt = OTFData.iGlyphCnt - iStart;
		} 
		else if (iDiff + 2 * iGlyphCnt > OTFData.iPositioningSizeBounds)
		{
			TRACEUSER("Greg", "Class definition sub-table extends beyond end of table.\n");
			OTFData.bBadOpenTypeData = TRUE;

			iGlyphCnt = (OTFData.iPositioningSizeBounds - (INT32)iDiff) / 2;
		}

		for (INT32 i = 0; i < iGlyphCnt; ++i)
			(*pvClass)[iStart + i] = ParseUShortFromPos(&pClassDefinitionPos);
	} 
	else if (iFormat == 2)
	{
		INT32 iRangeCnt = ParseUShortFromPos(&pClassDefinitionPos);
		ptrdiff_t iDiff = pClassDefinitionPos - pPosData;

		if (iDiff + 6 * iRangeCnt > OTFData.iPositioningSizeBounds)
		{
			TRACEUSER("Greg", "Class definition sub-table extends beyond end of table.\n");
			OTFData.bBadOpenTypeData = TRUE;
			iRangeCnt = (OTFData.iPositioningSizeBounds - (INT32)iDiff) / 6;
		}

		for (INT32 i = 0; i < iRangeCnt; ++i ) 
		{
			UINT16 uStart = ParseUShortFromPos(&pClassDefinitionPos);
			UINT16 uEnd = ParseUShortFromPos(&pClassDefinitionPos);

			if (uStart > uEnd || uEnd >= OTFData.iGlyphCnt)
			{
				TRACEUSER("Greg", "Bad class def table. Glyph range %d-%d out of range [0,%d).\n", 
						 uStart, uEnd, OTFData.iGlyphCnt);
				OTFData.bBadOpenTypeData = TRUE;
			}
		
			UINT16 uClass = ParseUShortFromPos(&pClassDefinitionPos);
		
			for (UINT16 j = uStart; j <= uEnd; ++j)
			{
				if (j < OTFData.iGlyphCnt)
					(*pvClass)[j] = uClass;
			}
		}
	} 
	else
	{
		//Put everything in class 0 and return that
		TRACEUSER("Greg", "Unknown class table format: %d.\n", iFormat);
		OTFData.bBadOpenTypeData = TRUE;
	}

	//Do another validity test
	BOOL bWarned = FALSE;
	for (INT32 i = 0; i < OTFData.iGlyphCnt; ++i)
	{
		if ((*pvClass)[i] >= OTFData.iGlyphCnt + 1)
		{
			if (!bWarned) 
			{
				TRACEUSER("Greg", "Nonsensical class assigned to a glyph-- class=%d is too big. Glyph=%d\n", 
						 (*pvClass)[i], i);
				OTFData.bBadOpenTypeData = TRUE;
				bWarned = TRUE;
			}
			(*pvClass)[i] = 0;
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::ClassToNames(std::vector<std::string>* pvNames, INT32 iClassCnt, 
										 std::vector<UINT16>* pvClass, INT32 iGlyphCnt)

	Author: 	Greg Watroba
	Created:	1/12/15
	Inputs:		pvNames - pointer to vector of strings
				iClassCnt - number of classes
				pvClass - pointer to vector of class values
				iGlyphCnt - number of glyphs
	Outputs:	Vector of class names
	Returns:    -
	Purpose:	Converts class values to names

********************************************************************************************/

void OTFDataLoader::ClassToNames(std::vector<std::string>* pvNames, INT32& iClassCnt, 
									 std::vector<UINT16>* pvClass, INT32& iGlyphCnt)
{
	pvNames->resize(iClassCnt);

	for (UINT16 i = 0; i < iGlyphCnt; ++i)
	{
		if ((*pvClass)[i] != 0 && OTFData.vChars[i] != nullptr)
		{
			if ((*pvClass)[i] < iClassCnt)
			{
				(*pvNames)[(*pvClass)[i]].append(OTFData.vChars[i]->sName);
				(*pvNames)[(*pvClass)[i]].append(" ");
			}
			else
			{
				TRACEUSER("Greg", "Class index out of range %d (must be <%d)\n", (*pvClass)[i], iClassCnt);
				return;
			}
		}
	}
}

/********************************************************************************************

>	void OTFDataLoader::CoverageMinusClasses(std::string *sName, std::vector<UINT16>* pvGlyphs, 
												 std::vector<UINT16>* pvClass)

	Author: 	Greg Watroba
	Created:	1/12/15
	Inputs:		pvClass - pointer to vector of class values
				pvGlyphs - pointer to vector of glyph values
	Outputs:	sName - will be filled with proper name
	Returns:    -
	Purpose:	Set proper name with given coverage and classes

********************************************************************************************/

void OTFDataLoader::CoverageMinusClasses(std::string* sName, std::vector<UINT16>* pvGlyphs, 
											 std::vector<UINT16>* pvClass)
{
	std::vector<UINT8> vSubGlyphs(OTFData.iGlyphCnt);

	for (INT32 i = 0; (*pvGlyphs)[i] != 0xffff; ++i)
	{
		if ((*pvGlyphs)[i] < OTFData.iGlyphCnt)
			vSubGlyphs[(*pvGlyphs)[i]] = 1;
	}

	for (INT32 i = 0; i < OTFData.iGlyphCnt; ++i)
	{
		if ((*pvClass)[i] != 0)
			vSubGlyphs[i] = 0;
	}

	INT32 i;
	for (i = 0; i < OTFData.iGlyphCnt; ++i)
	{
		if (vSubGlyphs[i] != 0)
			break;
	}
	
	//coverage table matches glyphs in classes. No need for special treatment
	if (i == OTFData.iGlyphCnt)
	{
		return;
	}

	//Otherwise we need to generate a class string of glyph names in the coverage
	//table but not in any class. These become the glyphs in class 0
	sName->clear();

	for (UINT16 i = 0; i < OTFData.iGlyphCnt; ++i)
	{
		if (vSubGlyphs[i] != 0 && OTFData.vChars[i])
		{
			sName->append(OTFData.vChars[i]->sName);
			sName->append(" ");
		}
	}
}
